/**CFile***********************************************************************

  FileName [bmcMH.c]

  PackageName [bmc]

  Synopsis [Functions for solving PSL/ABA problems with BMC and symbolic Miyano-Hayashi]

  Description []

  SeeAlso  []

  Author   [Tommi Junttila]

  Copyright [
  Copyright (C) 2006 by Tommi Junttila. Released under GNU LGPL ]
 
******************************************************************************/

#include "bmcBmc.h"
#include "dag/dag.h"

#include "node/node.h"
#include "be/be.h"
#include "mc/mc.h" /* for print_spec */
#include "sat/sat.h" /* for solver and result */
#include "sat/SatSolver.h"
#include "sat/SatIncSolver.h"
#include "prop/prop.h"

#include "bmcInt.h"
#include "bmcGen.h"
#include "bmcSatTrace.h"
#include "bmcDump.h"
#include "bmcModel.h"
#include "bmcWff.h"
#include "bmcConv.h"
#include "bmcUtils.h"
#include "bmcTJIncrSat.h"
#include "bmcSatTrace.h"

#include "mc/mc.h" /* for print_{explanation, spec, invar} */
#include "utils/error.h"

#include "utils/assoc.h"

#include "trace/Trace.h"
#include "trace/TraceManager.h"

#include "compile/symb_table/SymbType.h"
#include "compile/symb_table/SymbLayer.h"
#include "compile/type_checking/type_checking.h"

#include "enc/enc.h"
#include "enc/be/BeEnc.h"
#include "utils/error.h"
#include "bmcMH.h"
#include "sugarInt.h"
#include "parser/psl/pslInt.h"
#include "parser/psl/pslExpr.h"
#include "parser/psl/pslNode.h"
#include "parser/psl/psl_grammar.h"
#include "bmcPartitionMap.h"
#include "bmcPslToSugar.h"
#include "parser/symbols.h"
#include "utils/str_assoc.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/



static void findAtomicProps(Sugar_Formula_t *formula, lsList aplist);


static Trace_ptr _generate_and_print_cntexample(BeEnc_ptr be_enc, 
						SymbLayer_ptr psl_layer,
						Bmc_IncrSat_ptr solver, 
						BddEnc_ptr bdd_enc, 
						const int k, 
						const char* trace_name,
						be_ptr l_var);

static be_ptr _equal_vectors_formula(const BeEnc_ptr be_enc,
				     lsList vars,
				     const unsigned int i,
				     const unsigned int j);

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


EXTERN node_ptr boolean_range;

#ifdef BENCHMARKING
#include <unistd.h>
#include <sys/times.h>
struct tms time_bmc_start;
struct tms time_gen_start;
struct tms time_solver_start;
struct tms time_temp;
static double time_diff(struct tms *start, struct tms *stop)
{
  return ((((double)stop->tms_utime + (double)stop->tms_stime) -
	   ((double)start->tms_utime + (double)start->tms_stime)) /
	  (double)CLK_TCK);
}
#endif


/*
 * Auxiliary wrapper stuff needed to use both the standard and my incremental
 * SAT solver interfaces
 */
typedef struct _MetaSolver_TAG
{
  BeEnc_ptr        be_enc;

  int              using_volatile_group;

  Bmc_IncrSat_ptr  TJ_solver;

  SatIncSolver_ptr S_solver;
  SatSolverGroup   S_permanent_group;
  SatSolverGroup   S_volatile_group;
} _MetaSolver;

static _MetaSolver *_MS_create(BeEnc_ptr be_enc,
			       const int use_TJ_solver)
{
  assert(be_enc);

  _MetaSolver *ms = ALLOC(_MetaSolver, 1);
  ms->be_enc = be_enc;
  ms->using_volatile_group = 0;
  ms->TJ_solver = 0;
  ms->S_solver = 0;
  if(use_TJ_solver)
    {
      ms->TJ_solver = Bmc_IncrSat_init(get_sat_solver(options));
      if(ms->TJ_solver == 0)
	{
	  fprintf(nusmv_stderr,
		  "Incremental sat solver '%s' is not available.\n", 
		  get_sat_solver(options));
	  FREE(ms);
	  return 0;
	}
    }
  else
    {
      ms->S_solver = Sat_CreateIncSolver(get_sat_solver(options));
      if(ms->S_solver == SAT_INC_SOLVER(NULL)) {
	fprintf(nusmv_stderr,
		"Incremental sat solver '%s' is not available.\n", 
		get_sat_solver(options));
	FREE(ms);
	return 0;
      }
      ms->S_permanent_group =
	SatSolver_get_permanent_group(SAT_SOLVER(ms->S_solver));
    }
  return ms;
}
static void _MS_destroy(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_release(ms->TJ_solver); ms->TJ_solver = 0;
    }
  else
    {
      SatIncSolver_destroy(ms->S_solver); ms->S_solver = 0;
    }
}
static void _MS_goto_permanent_group(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  assert(ms->using_volatile_group);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_pop_frame(ms->TJ_solver);
    }
  else
    {
      SatIncSolver_destroy_group(ms->S_solver, ms->S_volatile_group);
    }
  ms->using_volatile_group = 0;
}
static void _MS_goto_volatile_group(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  assert(!ms->using_volatile_group);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_new_frame(ms->TJ_solver);
    }
  else
    {
      ms->S_volatile_group = SatIncSolver_create_group(ms->S_solver);
    }
  ms->using_volatile_group = 1;
}
static void _MS_force_true(_MetaSolver *ms,
			   be_ptr be_constraint)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);

  if(ms->TJ_solver)
    {
      Bmc_IncrSat_insert(ms->TJ_solver, be_constraint);
      Bmc_IncrSat_force_true(ms->TJ_solver, be_constraint);
    }
  else
    {
      SatSolver_ptr solver = SAT_SOLVER(ms->S_solver);
      Be_Cnf_ptr cnf = Be_ConvertToCnf(BeEnc_get_be_manager(ms->be_enc),
				       be_constraint);
      if(ms->using_volatile_group)
	{
	  SatSolver_add(solver, cnf, ms->S_volatile_group);
	  SatSolver_set_polarity(solver, cnf, 1, ms->S_volatile_group);
	}
      else
	{
	  SatSolver_add(solver, cnf, ms->S_permanent_group);
	  SatSolver_set_polarity(solver, cnf, 1, ms->S_permanent_group);
	}
      Be_Cnf_Delete(cnf);
    }
}
static void _MS_force_constraint_list(_MetaSolver *ms, lsList constraints)
{
  lsGen  iterator;
  be_ptr be_constraint;
  
  lsForEachItem(constraints, iterator, be_constraint)
    _MS_force_true(ms, be_constraint);
}
static Bmc_IncrSat_result _MS_solve(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);

  if(ms->TJ_solver)
    {
      return Bmc_IncrSat_solve(ms->TJ_solver);
    }
  else
    {
      return SatSolver_solve_all_groups(SAT_SOLVER(ms->S_solver));
    }
}

static char rcsid[] UTIL_UNUSED = "$Id: bmcMH.c,v 1.1 2006/05/05 11:56:07 tjunttil Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/


/*
 * Routines for the state indexing scheme
 * State 0 is the L state and 1 is the E state
 * The first real state is state 2.
 */
static int _L_state() {return 0; }
static int _E_state() {return 1; }
static int _real_k(int k) {return k+2; }
static unsigned int _model_k(int k) {assert(k >= 2); return k-2; }

/*
 * Get the initial state and invariant constraint at time i.
 * Similar to Bmc_Model_GetInit0(be_fsm) except for the i argument.
 */
static be_ptr Bmc_TJ_GetInitI(const BeFsm_ptr be_fsm, const int i)
{
  BeEnc_ptr be_enc = BeFsm_get_be_encoding(be_fsm);
  be_ptr init = BeEnc_untimed_expr_to_timed(be_enc,
			      Be_And(BeEnc_get_be_manager(be_enc), 
				     BeFsm_get_init(be_fsm),
				     BeFsm_get_invar(be_fsm)),
					    i);
  return init;
}

/*
 * An auxiliary routine to add a new state variable with 'name' to 'layer'
 */
static node_ptr _add_new_state_variable(SymbLayer_ptr layer, char *name)
{
  SymbType_ptr symbolicType;
  node_ptr node;

  assert(name && name[0] == '#');
  
  node = find_node(DOT, Nil, sym_intern(name));
  assert(node);

  if(!SymbLayer_can_declare_var(layer, node))
    error_redefining(node); 
    
  symbolicType = SymbType_create(SYMB_TYPE_ENUM, boolean_range);
  SymbLayer_declare_state_var(layer, node, symbolicType);
  return node;
}


static be_ptr _equal_vectors_formula(const BeEnc_ptr be_enc,
				     lsList vars,
				     const unsigned int i,
				     const unsigned int j)
{
  lsGen    iterator;
  node_ptr node_sv;

  be_ptr be_constraint;

  Be_Manager_ptr be_mgr;

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Initialize constraint */
  be_constraint = Be_Truth(be_mgr);

  lsForEachItem(vars, iterator, node_sv)
    {
      /* Get state variable sv at time i */
      be_ptr be_sv_i = BeEnc_name_to_timed(be_enc, node_sv, i);
      /* Get state variable sv at time j */
      be_ptr be_sv_j = BeEnc_name_to_timed(be_enc, node_sv, j);

      be_constraint = Be_And(be_mgr,
			     be_constraint,
			     Be_Iff(be_mgr, be_sv_i, be_sv_j));
    }

  return be_constraint;
}






/*
 *
 * Miyano-Hayashi stuff by TJ
 *
 */

struct _MH_TAG
{
  Sugar_Automaton_t *aba;
  
  /* All state and input vars of the system, a list of node_ptr */
  lsList   system_state_and_input_vars;
  
  /* The timed loop selector variable */
  node_ptr l_var;
  /* The LoopExists variable, only time index 0 is used */
  node_ptr LoopExists_var;

  /* array[model_time] of be */
  array_t *InLoop_bes;

  lsList ABA_Q_vars;
  lsList ABA_P_vars;

  hash_ptr sugar_id_to_Q_var;
  hash_ptr sugar_id_to_P_var;

  /* How many model fairnesses? */
  int model_fairness_nof;
  /* array[fairness_set_num] of be_ptr */
  array_t *model_fairness_untimed_bes;
  /* array[fairness_set_num] of array[model_time] of be */
  array_t *model_fairness_timed_bes;

  /* Automata fairness: array[model_time] of be */
  array_t *automata_fairness_timed_bes;

  /* System state variables occurring in formula */
  lsList   formula_state_vars;
  /* System input variables occurring in formula */
  lsList   formula_input_vars;
};
typedef struct _MH_TAG _MH_info_t;

static _MH_info_t *_MH_info_create()
{
  _MH_info_t *info = ALLOC(_MH_info_t, 1);
  info->aba = 0;
  info->system_state_and_input_vars = lsCreate();
  info->l_var = (node_ptr)NULL;
  info->LoopExists_var = (node_ptr)NULL;
  info->InLoop_bes = array_alloc(be_ptr, 0);
  info->model_fairness_nof = 0;
  info->model_fairness_untimed_bes = 0;
  info->model_fairness_timed_bes = 0;
  info->automata_fairness_timed_bes = array_alloc(be_ptr, 0);
  info->ABA_Q_vars = lsCreate();
  info->ABA_P_vars = lsCreate();

  info->sugar_id_to_Q_var = 0;
  info->sugar_id_to_P_var = 0;
  return info;
}

static void _MH_info_destroy(_MH_info_t * info)
{
  assert(info);
  info->aba = 0;
  lsDestroy(info->system_state_and_input_vars, NULL);
  info->l_var = (node_ptr)NULL;
  info->LoopExists_var = (node_ptr)NULL;
  assert(info->InLoop_bes); array_free(info->InLoop_bes); info->InLoop_bes = 0;
  info->model_fairness_nof = 0;
  if(info->model_fairness_untimed_bes) {
    array_free(info->model_fairness_untimed_bes);
    info->model_fairness_untimed_bes = 0;
  }
  if(info->model_fairness_timed_bes) {
    int i;
    for(i = 0; i < array_n(info->model_fairness_timed_bes); i++) {
      array_t *timed_bes = array_fetch(array_t*,
				       info->model_fairness_timed_bes,
				       i);
      if(timed_bes)
	array_free(timed_bes);
      array_insert(array_t*, info->model_fairness_timed_bes, i, 0);
    }
    array_free(info->model_fairness_timed_bes);
    info->model_fairness_timed_bes = 0;
  }
  if(info->automata_fairness_timed_bes) {
    array_free(info->automata_fairness_timed_bes);
    info->automata_fairness_timed_bes = 0;
  }
  if(info->ABA_Q_vars)
    {
      lsDestroy(info->ABA_Q_vars, 0);
    }
  if(info->ABA_P_vars)
    {
      lsDestroy(info->ABA_P_vars, 0);
    }
  if(info->sugar_id_to_Q_var)
    {
      free_assoc(info->sugar_id_to_Q_var);
      info->sugar_id_to_Q_var = 0;
    }
  if(info->sugar_id_to_P_var)
    {
      free_assoc(info->sugar_id_to_P_var);
      info->sugar_id_to_P_var = 0;
    }
}

static void _MH_set_Q_var(_MH_info_t *info,
			  int vertex_id,
			  node_ptr node)
{
  assert(info);
  assert(vertex_id >= 0);
  assert(node);
  insert_assoc(info->sugar_id_to_Q_var, (node_ptr)(vertex_id+1), node);
}
static node_ptr _MH_get_Q_var(_MH_info_t * const info, const int vertex_id)
{
  node_ptr node = 0;
  assert(info);
  assert(vertex_id >= 0);
  node = find_assoc(info->sugar_id_to_Q_var, (node_ptr)(vertex_id+1));
  assert(node);
  return node;
}
static void _MH_set_P_var(_MH_info_t *info,
			  int vertex_id,
			  node_ptr node)
{
  assert(info);
  assert(vertex_id >= 0);
  insert_assoc(info->sugar_id_to_P_var, (node_ptr)(vertex_id+1), node);
}
static node_ptr _MH_get_P_var(_MH_info_t * const info, const int vertex_id)
{
  node_ptr node = 0;
  assert(info);
  assert(vertex_id >= 0);
  node = find_assoc(info->sugar_id_to_P_var, (node_ptr)(vertex_id+1));
  return node;
}


static node_ptr _MH_new_Q_var(SymbLayer_ptr layer, const int index)
{
  char new_var_name[16];
  sprintf(new_var_name, "#Q_%d", index);
  return _add_new_state_variable(layer, new_var_name);
}

static node_ptr _MH_new_P_var(SymbLayer_ptr layer, const int index)
{
  char new_var_name[16];
  sprintf(new_var_name, "#P_%d", index);
  return _add_new_state_variable(layer, new_var_name);
}

static void
_MH_init_model_fairness(const BeEnc_ptr be_enc,
			const SymbLayer_ptr layer,
			_MH_info_t *info, 
			Prop_ptr coi_prop)
{  
  SexpFsm_ptr sexp_fsm = Prop_get_bool_sexp_fsm(coi_prop);
  assert(sexp_fsm != (SexpFsm_ptr)NULL);

  {
    node_ptr compassion_list = SexpFsm_get_compassion(sexp_fsm);
    if(!is_list_empty(compassion_list))
      internal_error("%s:%d: Compassion not handled",__FILE__,__LINE__);
  }

  {
    /* Initialize model fairness stuff */
    int fairness_num = 0;
    node_ptr justice_list = SexpFsm_get_justice(sexp_fsm);
    assert(info->model_fairness_nof == 0);
    assert(info->model_fairness_untimed_bes == 0);
    assert(info->model_fairness_timed_bes == 0);
    info->model_fairness_untimed_bes = array_alloc(be_ptr, 0);
    info->model_fairness_timed_bes = array_alloc(array_t *, 0);
    while(justice_list != LS_NIL)
      {
	be_ptr be_expr = Bmc_Conv_Bexp2Be(be_enc, Bmc_Wff_MkNnf((node_ptr)car(justice_list)));
	array_insert(be_ptr, info->model_fairness_untimed_bes,
		     fairness_num, be_expr);
	array_t *timed_array = array_alloc(be_ptr, 0);
	array_insert(array_t *, info->model_fairness_timed_bes,
		     fairness_num, timed_array);
	fairness_num++;
	justice_list = cdr(justice_list);
      }
    info->model_fairness_nof = fairness_num;
  }
  return;
}


static void _MH_init_state_vars(SymbLayer_ptr layer,
				_MH_info_t *info,
				Sugar_Automaton_t *aba)
{
  int index = 0;
  lsGen  iterator;
  Sugar_Vertex_t *vertex = 0;
  hash_ptr accepting_states = 0;
  int i;

  assert(info);
  assert(aba);

  info->aba = aba;

  /* Find accepting states */
  accepting_states = new_assoc();
  for(i = 0; i < array_n(aba->accept_states); i++)
    {
      Sugar_Vertex_t* v = array_fetch(Sugar_Vertex_t*, aba->accept_states, i);
      insert_assoc(accepting_states, (node_ptr)v, (node_ptr)v);
    }

  assert(lsLength(info->ABA_Q_vars) == 0);
  assert(lsLength(info->ABA_P_vars) == 0);

  info->sugar_id_to_Q_var = new_assoc();
  info->sugar_id_to_P_var = new_assoc();

  index = 0;
  lsForEachItem(aba->state_list, iterator, vertex)
    {
      node_ptr Q_var, P_var;
      Q_var = _MH_new_Q_var(layer, index);
      lsNewEnd(info->ABA_Q_vars, (lsGeneric)Q_var, LS_NH);
      if(find_assoc(accepting_states, (node_ptr)vertex))
	{
	  P_var = 0;
	}
      else
	{
	  P_var = _MH_new_P_var(layer, index);
	  lsNewEnd(info->ABA_P_vars, (lsGeneric)P_var, LS_NH);
	}

      _MH_set_Q_var(info, vertex->id, Q_var);
      _MH_set_P_var(info, vertex->id, P_var);

      index++;
    }

  free_assoc(accepting_states);

  return;
}

static lsList _MH_make_P_subset_Q_minus_A_contraints(BeEnc_ptr be_enc,
						     _MH_info_t *info,
						     const int i_real)
{
  lsList result = lsCreate();

  lsGen iterator;
  Sugar_Vertex_t *vertex;
  lsForEachItem(info->aba->state_list, iterator, vertex)
    {
      node_ptr Q_var, P_var;
      Q_var = _MH_get_Q_var(info, vertex->id);
      assert(Q_var);
      P_var = _MH_get_P_var(info, vertex->id);
      if(P_var)
	{
	  /* Make constraint P_i -> Q_i for this aba state */
	  be_ptr be_Q_i = 0;
	  be_ptr be_P_i = 0;
	  be_ptr be_constraint = 0;
	  be_Q_i = BeEnc_name_to_timed(be_enc, Q_var, i_real);
	  assert(be_Q_i);
	  be_P_i = BeEnc_name_to_timed(be_enc, P_var, i_real);
	  assert(be_P_i);
	  be_constraint = Be_Implies(BeEnc_get_be_manager(be_enc),
				     be_P_i,
				     be_Q_i);
	  assert(be_constraint);
	  lsNewEnd(result, (lsGeneric)be_constraint, LS_NH);
	}
    }
  return result;
}


int Bmc_SolveMH(Prop_ptr pslprop,
		Sugar_Automaton_t *given_aba,
		str_hash_ptr given_aba_map_sugarap_to_ap,
		const int k)
{
  BeFsm_ptr be_fsm; /* The corresponding be fsm */
  Be_Manager_ptr be_mgr;
  BoolEnc_ptr bool_enc;
  BeEnc_ptr be_enc;
  Prop_ptr coi_pslprop; /*Virtual property for COI*/
  int previous_k;
  int current_k;
  _MH_info_t *info = 0;
  Sugar_Formula_t *negated_sugar_prop = (Sugar_Formula_t *)NULL;
#ifdef MS
  _MetaSolver *zolver = 0;
#else
  /* incremental sat solver instance */
  Bmc_IncrSat_ptr solver = 0;
#endif

  
  /* Symbol table and layer stuff needed to allocate new vars */
  SymbTable_ptr symb_table = (SymbTable_ptr)NULL;
  SymbLayer_ptr mh_layer = (SymbLayer_ptr)NULL;

  boolean found_solution = false;
  /*Map ap names to pslNode_ptr*/
  str_hash_ptr ap_table = 0;
  Sugar_Automaton_t *aba = (Sugar_Automaton_t *)NULL;

#ifdef BENCHMARKING
  times(&time_bmc_start);
#endif

  /* checks that a PSL property or ABA was given: */
  nusmv_assert(pslprop != PROP(NULL) || given_aba);
  nusmv_assert(!(pslprop != PROP(NULL) && given_aba));


  /*
   *
   * If a PSL property was given, construct the corresponding (W)ABA
   *
   */
  if(pslprop != PROP(NULL))
    {
      Sugar_Formula_t *sugar_prop = (Sugar_Formula_t *)NULL;
      const PslNode_ptr pslformula = Prop_get_expr(pslprop);

      /* Some debugging output */
      fprintf(nusmv_stderr, "PSL formula: ");
      PslNode_print(nusmv_stderr, pslformula);
      fprintf(nusmv_stderr, "\n");
      
      /* Give pslformula to sugar */
      ap_table = new_str_assoc();
      sugar_prop = Bmc_pslToSugar(pslformula, ap_table);

      /* Some debugging output */
      fprintf(nusmv_stderr, "Sugar formula: ");
      Sugar_FormulaPrint(nusmv_stderr, sugar_prop);
      fprintf(nusmv_stderr, "\n");

      /* Negate the sugar formula */
      negated_sugar_prop = Sugar_FormulaCreate(Sugar_FL_Not_c,sugar_prop,NULL);

      /* Some debugging output */
      fprintf(nusmv_stderr, "Negated sugar formula: ");
      Sugar_FormulaPrint(nusmv_stderr, negated_sugar_prop);
      fprintf(nusmv_stderr, "\n");

      /* Create the automaton */
      aba = Sugar_AutomatonGenerate(negated_sugar_prop);

      /* Some debugging output */
      Sugar_AutomatonPrint(aba,5);
    }
  else
    {
      assert(given_aba);
      assert(given_aba_map_sugarap_to_ap);
      aba = given_aba;
      ap_table = given_aba_map_sugarap_to_ap;
    }

#if 0
  if(optimisations.buchi)
    {
      Sugar_Automaton_t *buchi = Sugar_AutomatonExpand(aba);
      Sugar_AutomatonPrint(buchi, 5);
      Sugar_AutomatonFree(aba);
      aba = buchi;
    }
#endif

  /*
   *
   * Cone of influence reduction
   *
   */
  /* Build coi_pslprop to be the conjunction of the atomic propositions
     appearing in the automaton */
  assert(aba != (Sugar_Automaton_t *)NULL);
  {
    PslNode_ptr virtual_pslprop = psl_node_make_true();
    lsGen state_iter;
    Sugar_Vertex_t *source_vertex = 0;
    hash_ptr set_of_ap_already_seen = new_assoc();
    lsForEachItem(aba->state_list, state_iter, source_vertex)
      {
	lsGen edge_iter;
	Sugar_Edge_t *edge = 0;
	lsForEachItem(source_vertex->edge_out_list, edge_iter, edge)
	  {
	    Sugar_Label_t *label = 0;
	    int label_index = 0;
	    arrayForEachItem(Sugar_Label_t*,
			     edge->label,
			     label_index,
			     label)
	      {
		if(label->name != NIL(char))
		  {
		    PslNode_ptr ap = find_str_assoc(ap_table, label->name);
		    assert(ap);
		    if(find_assoc(set_of_ap_already_seen, ap))
		      continue;
		    insert_assoc(set_of_ap_already_seen, ap, ap);
		    virtual_pslprop = psl_new_node(TKAMPERSAND,
						   virtual_pslprop, 
						   ap);
		    
		  }
	      }
	  }
      }
    coi_pslprop = Prop_create_partial(virtual_pslprop, Prop_Psl);

    fprintf(nusmv_stderr, "Virtual PSL formula for COI: ");
    Prop_print(coi_pslprop, nusmv_stderr);
    fprintf(nusmv_stderr, "\n");
    
    free_assoc(set_of_ap_already_seen); set_of_ap_already_seen = 0;
  }
  /* Semantic check */
  if(!TypeCheckingPkg_check_property(TypeCheckingPkg_get_global_type_checker(),
				     coi_pslprop))
    {
      return -1; /* type violation */
    }
  /* Apply COI */
  if(opt_cone_of_influence(options) == true)
    {
      Prop_apply_coi_for_bmc(coi_pslprop, global_fsm_builder);
    }
  /* If no COI was applied, set the fms of coi_pslprop to the master */
  be_fsm = Prop_get_be_fsm(coi_pslprop);
  if(be_fsm == (BeFsm_ptr) NULL)
    {
      Prop_set_fsm_to_master(coi_pslprop);
      be_fsm = Prop_get_be_fsm(coi_pslprop);
      nusmv_assert(be_fsm != (BeFsm_ptr) NULL);
    }



  /* Get be encoding */
  be_enc = BeFsm_get_be_encoding(be_fsm);
  assert(be_enc);

  /* Get Boolean encoding */
  bool_enc = BoolEncClient_get_bool_enc(BOOL_ENC_CLIENT(be_enc));
  assert(bool_enc);
  
  /* Find manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);


  /*
   * Incremental SAT solver construction
   */
#ifdef MS
  zolver = _MS_create(be_enc, opt_use_standard_solver == 0);
  if(zolver == 0)
    /* Something went wrong */
    return 1;
#else
  solver = Bmc_IncrSat_init(get_sat_solver(options));
  if(solver == 0)
    {
      fprintf(nusmv_stderr, "Incremental sat solver '%s' is not available.\n", 
	      get_sat_solver(options));
      return 1;
    }
#endif

  /*
   * Initialize info structure
   */
  info = _MH_info_create();
  assert(info);


#if 0
  /*
   * Get the state variables occurring in the transition relation and
   * add them to info->transition_state_vars
   */
  {
    lsGen iterator; 
    be_ptr be_var = 0;
    lsList tr_be_vars = Bmc_Utils_get_vars_list_for_uniqueness(be_enc,
							       coi_pslprop);
    lsForEachItem(tr_be_vars, iterator, be_var)
      {
	node_ptr node = BeEnc_var_to_name(be_enc, be_var);
	lsNewEnd(info->transition_state_vars, (lsGeneric)node, LS_NH);
      }
    lsDestroy(tr_be_vars, 0); tr_be_vars = 0;
  }
#endif
  {
    int index = BeEnc_get_first_untimed_var_index(be_enc,
						  BE_VAR_TYPE_CURR |
						  BE_VAR_TYPE_INPUT);
    while(BeEnc_is_var_index_valid(be_enc, index))
      {
	node_ptr var = BeEnc_index_to_name(be_enc, index);
	assert(var);
	lsNewEnd(info->system_state_and_input_vars, (lsGeneric)var, LS_NH);
	index = BeEnc_get_next_var_index(be_enc, index,
					 BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
      }
  }


  /* Get the symbol table */
  symb_table = Compile_get_global_symb_table();

  /* Add a new layer for translation variables */
  mh_layer = SymbTable_create_layer(symb_table, "MH translation vars",
				    SYMB_LAYER_POS_TOP);

  /* Allocate the l_i variable */
  info->l_var = _add_new_state_variable(mh_layer, "#l");

  /* Allocate the LoopExists variable */
  info->LoopExists_var = _add_new_state_variable(mh_layer, "#LoopExists");



  /* */
  _MH_init_model_fairness(be_enc, mh_layer, info, coi_pslprop);

  /*
   * Adds Q and P variables for each aba state
   */
  _MH_init_state_vars(mh_layer, info, aba);

  /*
   * After introducing all new variables, commit mh_layer
   */
  BaseEnc_commit_layer(BASE_ENC(bool_enc), SymbLayer_get_name(mh_layer));
  BaseEnc_commit_layer(BASE_ENC(be_enc), SymbLayer_get_name(mh_layer));
  /*
  BaseEnc_commit_layer(BASE_ENC(Enc_get_bdd_encoding()),
		       SymbLayer_get_name(mh_layer));
  */
  
  {
    /* P subseteq Q\A at time L */
    lsGen  iterator;
    be_ptr be_constraint;
    lsList constraints =
      _MH_make_P_subset_Q_minus_A_contraints(be_enc, info, _L_state());
    lsForEachItem(constraints, iterator, be_constraint)
      {
#ifdef MS
	_MS_force_true(zolver, be_constraint;
#else
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif
     }      
  }
  
  {
    /* P subseteq Q\A at time E */
    lsGen  iterator;
    be_ptr be_constraint;
    lsList constraints =
      _MH_make_P_subset_Q_minus_A_contraints(be_enc, info, _E_state());
    lsForEachItem(constraints, iterator, be_constraint)
      {
#ifdef MS
	_MS_force_true(zolver, be_constraint;
#else
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif
     }      
  }
  
  {
    /* P subseteq Q\A at time 0 */
    lsGen  iterator;
    be_ptr be_constraint;
    lsList constraints =
      _MH_make_P_subset_Q_minus_A_contraints(be_enc, info, _real_k(0));
    lsForEachItem(constraints, iterator, be_constraint)
      {
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif
     }      
  }

  /*
   * Automata initial state constraints
   */
  {
    int index1, index2;
    array_t *initial_state;
    be_ptr be_initial_cond = Be_Falsity(be_mgr);
    
    arrayForEachItem(array_t*, aba->initial_states, index1, initial_state)
      {
	Sugar_Vertex_t *vertex;
	be_ptr be_and_states = Be_Truth(be_mgr);
	arrayForEachItem(Sugar_Vertex_t*, initial_state, index2, vertex)
	  {
	    node_ptr Q_var = _MH_get_Q_var(info, vertex->id);
	    node_ptr P_var = _MH_get_P_var(info, vertex->id);
	    assert(Q_var);
	    be_and_states = Be_And(be_mgr,
				   be_and_states,
				   BeEnc_name_to_timed(be_enc,
						       Q_var,
						       _real_k(0)));
	    if(P_var)
	      be_and_states = Be_And(be_mgr,
				     be_and_states,
				     BeEnc_name_to_timed(be_enc,
							 P_var,
							 _real_k(0)));
	    
	  }
	be_initial_cond = Be_Or(be_mgr, be_initial_cond, be_and_states);
      }
#ifdef MS
    _MS_force_true(zolver, be_initial_cond);
#else
    Bmc_IncrSat_insert(solver, be_initial_cond);
    Bmc_IncrSat_force_true(solver, be_initial_cond);
#endif
  }


  /*
   * In the real thing where automata may have final states, too,
   * we should check whether the empty string is accepted here
   * before forcing (the possibly unsat) initial state condition
   */


  /*
   * Insert the initial condition into the sat solver permanently
   */
  {
    be_ptr be_init = Bmc_TJ_GetInitI(be_fsm, _real_k(0));
#ifdef MS
    _MS_force_true(zolver, be_init);
#else
    Bmc_IncrSat_insert(solver, be_init);
    Bmc_IncrSat_force_true(solver, be_init);
#endif

    if(opt_verbose_level_ge(options, 6))
      {
	fprintf(nusmv_stderr, "Forced initial condition: ");
	Be_DumpSexpr(be_mgr, be_init, nusmv_stderr);
	fprintf(nusmv_stderr, "\n");
      }
  }
  

  if(pslprop)
    {
      /* This should actually have been done much earlier somewhere... */
      Prop_set_type(pslprop, Prop_Psl);
    }

  /* Start from deep prehistory (at least <= -2) */
  previous_k = -1000;

  /* Start problems generation: */
  found_solution = false;
  for(current_k = 0;
      current_k <= k && !found_solution;
      current_k = current_k + 1)
    {
      int i = 0;
      Bmc_IncrSat_result satResult;
      
      if(opt_verbose_level_gt(options, 0))
	{
	  fprintf(nusmv_stderr, "\nGenerating problem with bound %d, "
		  "all possible loopbacks...\n", current_k);
	  fflush(nusmv_stderr);
	}
      

      
      /********************************************************************
       *
       * The k-invariant part of the translation
       *
       ********************************************************************/

      /* Unroll P subseteq Q\A upto and including current_k+1 */ 
      for(i = max(previous_k+2, 1); i <= current_k+1; i++)
	{
	  lsGen  iterator;
	  be_ptr be_constraint;
	  lsList constraints =
	    _MH_make_P_subset_Q_minus_A_contraints(be_enc, info, _real_k(i));
	  lsForEachItem(constraints, iterator, be_constraint)
	    {
#ifdef MS
	      _MS_force_true(zolver, be_constraint);
#else
	      Bmc_IncrSat_insert(solver, be_constraint);
	      Bmc_IncrSat_force_true(solver, be_constraint);
#endif
	    }      
	}
      /*
       * Unroll the model transition relation up to and including current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      for(i = max(previous_k, 0); i < current_k; i++)
	{
	  be_ptr be_TR = Bmc_Model_GetUnrolling(be_fsm, 
						_real_k(i),
						_real_k(i+1));
#ifdef MS
	  _MS_force_true(zolver, be_TR);
#else
	  Bmc_IncrSat_insert(solver, (Rbc_t*)be_TR);
	  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_TR);
#endif
	
	  if(opt_verbose_level_gt(options, 1))
	    {
	      fprintf(nusmv_stderr, "Forced T(%d,%d)", i, i+1);
	      if(opt_verbose_level_ge(options, 7)) {
		fprintf(nusmv_stderr, ": ");
		Be_DumpSexpr(be_mgr, be_TR, nusmv_stderr);
	      }
	      fprintf(nusmv_stderr, "\n");
	    }
	}

      /* Unroll automaton transition relation upto and including current_k+1
       * Q_s_i -> (Q' models delta(s, sigma_i)) */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  lsGen state_iter;
	  Sugar_Vertex_t *source_vertex;
	  lsForEachItem(aba->state_list, state_iter, source_vertex)
	    {
	      lsGen edge_iter;
	      Sugar_Edge_t *edge;
	      node_ptr source_Q_var;
	      be_ptr be_source_Q_i;
	      be_ptr be_constraint = 0;
	      be_ptr be_or = Be_Falsity(be_mgr);
	      source_Q_var =  _MH_get_Q_var(info, source_vertex->id);
	      assert(source_Q_var);
	      be_source_Q_i = BeEnc_name_to_timed(be_enc,
						  source_Q_var,
						  _real_k(i));
	      lsForEachItem(source_vertex->edge_out_list, edge_iter, edge)
		{
		  Sugar_Label_t *label = 0;
		  int label_index;
		  Sugar_Vertex_t *dest_vertex = 0;
		  lsGen dest_vertex_iter;
		  be_ptr be_and = Be_Truth(be_mgr);
		  arrayForEachItem(Sugar_Label_t*,
				   edge->label,
				   label_index,
				   label)
		    {
		      if(label->name == NIL(char))
			{
			  be_and = Be_And(be_mgr,
					  be_and,
					  (label->isNegate) ? 
					  Be_Falsity(be_mgr):
					  Be_Truth(be_mgr));
			}
		      else
			{
			  be_ptr be_label;
			  be_ptr be_label_i;
			  PslNode_ptr ap = find_str_assoc(ap_table,
							  label->name);
			  assert(ap);
			  be_label = Bmc_Conv_Bexp2Be(be_enc, Bmc_Wff_MkNnf(detexpr2bexpr(Enc_get_bdd_encoding(), PslNode_convert_to_node_ptr(PslNode_pslltl2ltl(ap, PSL2SMV)))));
			  be_label_i = BeEnc_untimed_expr_to_timed(be_enc,
								   be_label,
								   _real_k(i));
			  be_and = Be_And(be_mgr,
					  be_and,
					  (label->isNegate)?
					  Be_Not(be_mgr, be_label_i):
					  be_label_i);
			}
		    }
		  lsForEachItem(edge->dest_vertex_list, dest_vertex_iter,
				dest_vertex)
		    {
		      node_ptr dest_Q_var;
		      be_ptr be_dest_Q_iP1;
		      dest_Q_var = _MH_get_Q_var(info, dest_vertex->id);
		      assert(dest_Q_var);
		      be_dest_Q_iP1 = BeEnc_name_to_timed(be_enc,
							  dest_Q_var,
							  _real_k(i+1));
		      be_and = Be_And(be_mgr,
				      be_and,
				      be_dest_Q_iP1);
		    }
		  be_or = Be_Or(be_mgr, be_or, be_and);
		}
	      be_constraint = Be_Implies(be_mgr, be_source_Q_i, be_or);
#ifdef MS
	      _MS_force_true(zolver, be_constraint);
#else
	      Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	      Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
	    }
	
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced delta(%d,%d)\n", i, i+1);
	}
      
      /*
       * Unroll constraints for the P set upto and including current_k+1
       * P_i empty -> P'=Q'\A
       * P_i not empty -> TO DO
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  be_ptr be_P_empty_i = 0;

	  be_P_empty_i = Be_Truth(be_mgr);
	  {
	    lsGen P_iter;
	    node_ptr P_var;
	    lsForEachItem(info->ABA_P_vars, P_iter, P_var)
	      {
		be_ptr be_P_i = BeEnc_name_to_timed(be_enc, P_var, _real_k(i));
		be_P_empty_i = Be_And(be_mgr,
				      be_P_empty_i,
				      Be_Not(be_mgr, be_P_i));
	      }
	  }

	  {
	    /* Case 1: P empty implies P'=Q'\A */
	    be_ptr be_constraint = 0;
	    be_ptr be_and = Be_Truth(be_mgr);
	    lsGen iterator;
	    Sugar_Vertex_t *vertex;
	    lsForEachItem(info->aba->state_list, iterator, vertex)
	      {
		node_ptr Q_var, P_var;
		be_ptr be_Q_iP1 = 0;
		be_ptr be_P_iP1 = 0;
		Q_var = _MH_get_Q_var(info, vertex->id);
		assert(Q_var);
		P_var = _MH_get_P_var(info, vertex->id);
		if(!P_var)
		  continue;
		be_Q_iP1 = BeEnc_name_to_timed(be_enc, Q_var, _real_k(i+1));
		assert(be_Q_iP1);
		be_P_iP1 = BeEnc_name_to_timed(be_enc, P_var, _real_k(i+1));
		assert(be_P_iP1);
		/* We already have P' subseteq Q'\A so now it is sufficient
		 * to force P' supseteq Q'\A to have P'=Q'\A */
		be_and = Be_And(be_mgr,
				be_and,
				Be_Implies(be_mgr, be_Q_iP1, be_P_iP1));
		
	      }
	    be_constraint = Be_Implies(be_mgr,
				       be_P_empty_i,
				       be_and);
#ifdef MS
	    _MS_force_true(zolver, be_constraint);
#else
	    Bmc_IncrSat_insert(solver, be_constraint);
	    Bmc_IncrSat_force_true(solver, be_constraint);
#endif
	  } /* End of case 1*/
	  {
	    /* Case 2:
	     * P_i not empty ->
	     * P' \cup (Q' \cap A) models (forall s\inP : delta(s,sigma_i))*/
	    be_ptr be_constraint = Be_Truth(be_mgr);
	    lsGen state_iter;
	    Sugar_Vertex_t *source_vertex;
	    lsForEachItem(aba->state_list, state_iter, source_vertex)
	      {
		lsGen edge_iter;
		Sugar_Edge_t *edge;
		node_ptr source_P_var;
		be_ptr be_source_P_i;
		be_ptr be_or = Be_Falsity(be_mgr);
		source_P_var = _MH_get_P_var(info, source_vertex->id);
		if(!source_P_var)
		  continue;
		be_source_P_i = BeEnc_name_to_timed(be_enc,
						    source_P_var,
						    _real_k(i));
		lsForEachItem(source_vertex->edge_out_list, edge_iter, edge)
		  {
		    Sugar_Label_t *label = 0;
		    int label_index;
		    Sugar_Vertex_t *dest_vertex = 0;
		    lsGen dest_vertex_iter;
		    be_ptr be_and = Be_Truth(be_mgr);
		    arrayForEachItem(Sugar_Label_t*,
				     edge->label,
				     label_index,
				     label)
		      {
			if(label->name == NIL(char))
			  {
			    be_and = Be_And(be_mgr,
					    be_and,
					    (label->isNegate) ? 
					    Be_Falsity(be_mgr):
					    Be_Truth(be_mgr));
			  }
			else
			  {
			    be_ptr be_label;
			    be_ptr be_label_i;
			    PslNode_ptr ap = find_str_assoc(ap_table,
							    label->name);
			    assert(ap);
			    be_label = Bmc_Conv_Bexp2Be(be_enc, Bmc_Wff_MkNnf(detexpr2bexpr(Enc_get_bdd_encoding(), PslNode_convert_to_node_ptr(PslNode_pslltl2ltl(ap, PSL2SMV)))));
			    be_label_i = BeEnc_untimed_expr_to_timed(be_enc,
								     be_label,
								     _real_k(i));
			    be_and = Be_And(be_mgr,
					    be_and,
					    (label->isNegate)?
					    Be_Not(be_mgr, be_label_i):
					    be_label_i);
			  }
		      }
		    lsForEachItem(edge->dest_vertex_list, dest_vertex_iter,
				  dest_vertex)
		      {
			node_ptr dest_QP_var;
			be_ptr be_dest_QP_iP1;
			dest_QP_var = _MH_get_P_var(info, dest_vertex->id);
			if(!dest_QP_var)
			  dest_QP_var = _MH_get_Q_var(info, dest_vertex->id);
			assert(dest_QP_var);
			be_dest_QP_iP1 = BeEnc_name_to_timed(be_enc,
							     dest_QP_var,
							     _real_k(i+1));
			be_and = Be_And(be_mgr,
					be_and,
					be_dest_QP_iP1);
		      }
		    be_or = Be_Or(be_mgr, be_or, be_and);
		  } /* iter over edges */
		be_constraint = Be_And(be_mgr,
				       be_constraint,
				       Be_Implies(be_mgr,
						  be_source_P_i,
						  be_or));
	      } /* iter over states */
	    
	    be_constraint = Be_Implies(be_mgr,
				       Be_Not(be_mgr, be_P_empty_i),
				       be_constraint);
#ifdef MS
	    _MS_force_true(zolver, be_constraint);
#else
	    Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	    Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
	  } /* End of case 2 */
	}


      
      /*
       * Define InLoop_i := (InLoop_{i-1} || l_i)
       * and build constraint InLoop_{i-1} => !l_i up to current_k
       * Force l_0 <=> FALSE
       * Assumes that they have already been build up to previous_k
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  be_ptr be_l_i = 0;
	  be_ptr be_constraint = 0;
	  be_ptr be_InLoop_i = 0;
	  
	  be_l_i = BeEnc_name_to_timed(be_enc, info->l_var, _real_k(i));
	  assert(be_l_i);

	  if(i == 0)
	    {
	      /* InLoop_0 := false */
	      be_InLoop_i = Be_Falsity(be_mgr);
	      /* Force constraint: l_0 <=> false */
	      be_constraint = Be_Iff(be_mgr, be_l_i, Be_Falsity(be_mgr));
	      assert(be_constraint);
	    }
	  else
	    {
	      /* InLoop_i := InLoop_{i-1} | l_i */
	      be_ptr be_InLoop_iM1 = 0;
	      be_InLoop_iM1 = array_fetch(be_ptr, info->InLoop_bes, i-1);
	      assert(be_InLoop_iM1);
	      be_InLoop_i = Be_Or(be_mgr, be_InLoop_iM1, be_l_i);
	      assert(be_InLoop_i);
	      /* Force constraint: InLoop_{i-1} => !l_i */
	      be_constraint = Be_Implies(be_mgr,
					 be_InLoop_iM1,
					 Be_Not(be_mgr, be_l_i));
	      assert(be_constraint);
	    }
	  /* Store InLoop_i */
	  assert(be_InLoop_i);
	  assert(array_n(info->InLoop_bes) == i);
	  array_insert(be_ptr, info->InLoop_bes, i, be_InLoop_i);
#ifndef MS
	  /* also give (not force) it to the solver if using TJ interface */
	  Bmc_IncrSat_insert(solver, be_InLoop_i);
#endif
	  /* Force constraint */
	  assert(be_constraint);
#ifdef MS
	  _MS_force_true(zolver, be_constraint);
#else	  
	  Bmc_IncrSat_insert(solver, be_constraint);
	  Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced InLoop_%d stuff\n", i);
	}


      /*
       * Unroll the (l_i => (s_{i-1} = s_E)) constraint up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  be_ptr be_l_i = 0;
	  be_ptr be_equal_states = 0;
	  be_ptr be_constraint = 0;
	  
	  be_l_i = BeEnc_name_to_timed(be_enc, info->l_var, _real_k(i));
	  assert(be_l_i);

	  if(i == 0)
	    {
	      /* As l_0 == false, we can skip this */
	      continue;
	    }

	  be_equal_states =
	    _equal_vectors_formula(be_enc, info->system_state_and_input_vars,
				   _real_k(i-1), _E_state());
	  assert(be_equal_states);
	  be_constraint = Be_Implies(be_mgr, be_l_i, be_equal_states);
	  assert(be_constraint);
#ifdef MS
	  _MS_force_true(zolver, be_constraint);
#else	  
	  Bmc_IncrSat_insert(solver, be_constraint);
	  Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced (l_%d => (s_%d = s_E))\n", i, i-1);
	}
      /*
       * Unroll the (l_i => (Q_{i} = Q_L)) and (l_i => (P_{i} = P_L))
       * constraints up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  be_ptr be_l_i = 0;
	  be_ptr be_equal_states = 0;
	  be_ptr be_constraint = 0;
	  
	  be_l_i = BeEnc_name_to_timed(be_enc, info->l_var, _real_k(i));
	  assert(be_l_i);

	  if(i == 0)
	    {
	      /* As l_0 == false, we can skip this */
	      continue;
	    }

	  be_equal_states = _equal_vectors_formula(be_enc,
						   info->ABA_Q_vars,
						   _real_k(i),
						   _L_state());
	  assert(be_equal_states);
	  be_constraint = Be_Implies(be_mgr, be_l_i, be_equal_states);
	  assert(be_constraint);
#ifdef MS
	  _MS_force_true(zolver, be_constraint);
#else	  
	  Bmc_IncrSat_insert(solver, be_constraint);
	  Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced (l_%d => (Q_%d = Q_L))\n", i, i);

	  be_equal_states = _equal_vectors_formula(be_enc,
						   info->ABA_P_vars,
						   _real_k(i),
						   _L_state());
	  assert(be_equal_states);
	  be_constraint = Be_Implies(be_mgr, be_l_i, be_equal_states);
	  assert(be_constraint);
#ifdef MS
	  _MS_force_true(zolver, be_constraint);
#else	  
	  Bmc_IncrSat_insert(solver, be_constraint);
	  Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced (l_%d => (P_%d = P_L))\n", i, i);
	}      


      /*
       * Model fairness constraints
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  int fairness_num;
	  for(fairness_num = 0;
	      fairness_num < info->model_fairness_nof;
	      fairness_num++)
	    {
	      be_ptr be_expr = 0;
	      array_t *timed_bes = 0;

	      be_expr = array_fetch(be_ptr,
				    info->model_fairness_untimed_bes,
				    fairness_num);
	      assert(be_expr);

	      timed_bes = array_fetch(array_t*,
				      info->model_fairness_timed_bes,
				      fairness_num);
	      assert(timed_bes);
	      assert(array_n(timed_bes) == i);

	      if(i == 0)
		{
		  /* FAIR_0 = false */
		  array_insert(be_ptr, timed_bes, i, Be_Falsity(be_mgr));
		}
	      else
		{
		  /* FAIR_i := FAIR_{i-1} | (InLoop_i & be_expr_i)*/
		  be_ptr be_expr_i = 0;
		  be_ptr be_fair_i = 0;
		  be_ptr be_fair_iM1 = 0;
		  be_ptr be_InLoop_i = 0;

		  be_expr_i = BeEnc_untimed_expr_to_timed(be_enc,
							  be_expr,
							  _real_k(i));
		  assert(be_expr_i);

		  be_fair_iM1 = array_fetch(be_ptr, timed_bes, i-1);
		  assert(be_fair_iM1);
		  
		  be_InLoop_i = array_fetch(be_ptr, info->InLoop_bes, i);
		  assert(be_InLoop_i);
		  
		  be_fair_i = Be_Or(be_mgr,
				    be_fair_iM1,
				    Be_And(be_mgr,
					   be_InLoop_i,
					   be_expr_i));
		  assert(be_fair_i);

		  array_insert(be_ptr, timed_bes, i, be_fair_i);
#ifndef MS
		  /* Give (not force) it to the solver if using TJ interface */
		  Bmc_IncrSat_insert(solver, be_fair_i);
#endif
		}
	    }
	}

      /*
       * Automata fairness constraints
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  be_ptr be_P_empty_i = 0;
	  be_ptr be_InLoop_i = 0;
	  be_ptr be_fair_i = 0;
	  be_ptr be_fair_iM1 = 0;
	  lsGen P_iter;
	  node_ptr P_var;

	  assert(array_n(info->automata_fairness_timed_bes) == i);

	  if(i == 0)
	    {
	      /* FAIR_0 = false */
	      array_insert(be_ptr,
			   info->automata_fairness_timed_bes,
			   i,
			   Be_Falsity(be_mgr));
	      continue;
	    }
	  /* FAIR_i := FAIR_{i-1} | (InLoop_i & P empty)*/
	  be_InLoop_i = array_fetch(be_ptr, info->InLoop_bes, i);
	  assert(be_InLoop_i);
		  
	  be_fair_iM1 = array_fetch(be_ptr,
				    info->automata_fairness_timed_bes,
				    i-1);
	  assert(be_fair_iM1);
		  
	  be_P_empty_i = Be_Truth(be_mgr);
	  lsForEachItem(info->ABA_P_vars, P_iter, P_var)
	    {
	      be_ptr be_P_i = BeEnc_name_to_timed(be_enc, P_var, _real_k(i));
	      be_P_empty_i = Be_And(be_mgr,
				    be_P_empty_i,
				    Be_Not(be_mgr, be_P_i));
	    }
	  be_fair_i = Be_Or(be_mgr,
			    be_fair_iM1,
			    Be_And(be_mgr,
				   be_InLoop_i,
				   be_P_empty_i));
	  /* Store fairness constraint */
	  array_insert(be_ptr,
		       info->automata_fairness_timed_bes,
		       i,
		       be_fair_i);
#ifndef MS
	  /* Give (not force) it to the solver if using TJ interface */
	  Bmc_IncrSat_insert(solver, be_fair_i);
#endif
	}


      
      /********************************************************************
       *
       * The k-dependent part of the translation
       *
       ********************************************************************/
      
      
      if(opt_verbose_level_ge(options, 2))
	fprintf(nusmv_stderr, "Adding k-dependent stuff\n");
      
      
      /*
       * Go to the volatile frame 1
       */
#ifdef MS
      _MS_goto_volatile_group(zolver);
#else
      Bmc_IncrSat_new_frame(solver);
#endif
      
      
      /*
       * Add k-dependent constraints
       */
      {
	/* Add LoopExists <=> InLoop_k */
	be_ptr be_LoopExists;
	be_ptr be_InLoop_k;
	be_ptr be_constraint;
	
	be_LoopExists = BeEnc_name_to_timed(be_enc, info->LoopExists_var, 0);
	assert(be_LoopExists);

	be_InLoop_k = array_fetch(be_ptr, info->InLoop_bes, current_k);
	assert(be_InLoop_k);

	be_constraint = Be_Iff(be_mgr, be_LoopExists, be_InLoop_k);

#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
      }
      {
	/* Add LoopExists <=> True */
	be_ptr be_LoopExists;
	be_ptr be_constraint;

	fprintf(nusmv_stderr, "WARNING: finite executions are ignored\n");
	
	be_LoopExists = BeEnc_name_to_timed(be_enc, info->LoopExists_var, 0);
	assert(be_LoopExists);
	be_constraint = Be_Iff(be_mgr, be_LoopExists, Be_Truth(be_mgr));
	assert(be_constraint);
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
      }
      {
	/* Add s_k = s_E */
	be_ptr be_constraint = 0;
	be_constraint =
	  _equal_vectors_formula(be_enc, info->system_state_and_input_vars,
				 _real_k(current_k), _E_state());
	assert(be_constraint);
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	if(opt_verbose_level_gt(options, 1))
	  fprintf(nusmv_stderr, "Forced (s_%d = s_E)\n", current_k);
	
      }
      {
	/* Add Q_k+1 = Q_L */
	be_ptr be_constraint = 0;
	be_constraint = _equal_vectors_formula(be_enc,
					       info->ABA_Q_vars,
					       _real_k(current_k+1),
					       _L_state());
	assert(be_constraint);
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	if(opt_verbose_level_gt(options, 1))
	  fprintf(nusmv_stderr, "Forced (Q_%d = Q_L)\n", current_k+1);
      }
      {
	/* Add P_k+1 = P_L */
	be_ptr be_constraint = 0;
	be_constraint = _equal_vectors_formula(be_enc,
					       info->ABA_P_vars,
					       _real_k(current_k+1),
					       _L_state());
	assert(be_constraint);
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	if(opt_verbose_level_gt(options, 1))
	  fprintf(nusmv_stderr, "Forced (P_%d = P_L)\n", current_k+1);
      }
      {
	/* Force automata fairness by requiring FAIR_k <=> true */
	be_ptr be_fair_k = 0;
	be_ptr be_constraint = 0; 
	    
	assert(array_n(info->automata_fairness_timed_bes) == current_k+1);

	be_fair_k = array_fetch(be_ptr,
				info->automata_fairness_timed_bes,
				current_k);
	assert(be_fair_k);
	be_constraint = Be_Iff(be_mgr, be_fair_k, Be_Truth(be_mgr));
	assert(be_constraint);
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif
	if(info->model_fairness_nof > 0)
	  fprintf(nusmv_stderr, "WARNING: finite executions are ignored "
		  "due to requiring automata fairness\n");
      } 
      {
	/* Force possible model fairness by requiring FAIR_k <=> true
	 * for each model fairness set */
	int fairness_num;
	for(fairness_num = 0;
	    fairness_num < info->model_fairness_nof;
	    fairness_num++)
	  {
	    array_t *timed_bes = 0;
	    be_ptr be_fair_k = 0;
	    be_ptr be_constraint = 0; 
	    
	    timed_bes = array_fetch(array_t*,
				    info->model_fairness_timed_bes,
				    fairness_num);
	    assert(timed_bes);
	    assert(array_n(timed_bes) == current_k+1);

	    be_fair_k = array_fetch(be_ptr, timed_bes, current_k);
	    assert(be_fair_k);
	    be_constraint = Be_Iff(be_mgr, be_fair_k, Be_Truth(be_mgr));
	    assert(be_constraint);
#ifdef MS
	    _MS_force_true(zolver, be_constraint);
#else	  
	    Bmc_IncrSat_insert(solver, be_constraint);
	    Bmc_IncrSat_force_true(solver, be_constraint);
#endif
	  }
	if(info->model_fairness_nof > 0)
	  fprintf(nusmv_stderr, "WARNING: finite executions are ignored "
		  "due to requiring model fairness\n");
      }


      
#ifdef BENCHMARKING
      fprintf(nusmv_stderr, ":START:benchmarking Solving\n");
      times(&time_solver_start);
#endif  
      
      
      /*
       * Actually solve the problem
       */
#ifdef MS
      satResult = _MS_solve(zolver);
#else
      satResult = Bmc_IncrSat_solve(solver);
#endif
      
      
#ifdef BENCHMARKING
      {
	times(&time_temp);
	fprintf(nusmv_stderr, ":UTIME = %.4f secs.\n",
		time_diff(&time_solver_start, &time_temp));
	  fprintf(nusmv_stderr, ":STOP:benchmarking Solving\n");
	  fprintf(nusmv_stderr, "TJ: counterexample, k=%d, solver time=%.4f, "
		  "cumulative time=%.4f.\n",
		  current_k,
		  time_diff(&time_solver_start, &time_temp),
		  time_diff(&time_bmc_start, &time_temp));
      }
#endif
      
      
      /* Processes the result: */
      switch (satResult)
	{
	case BMC_INCRSAT_UNSATISFIABLE_PROBLEM:
	  {
	    fprintf(nusmv_stderr,
		    "-- no counterexample found with bound %d",
		    current_k);
	    if (opt_verbose_level_gt(options, 2)) {
	      fprintf(nusmv_stderr, " for "); 		  
	      if(pslprop)
		print_spec(nusmv_stderr, pslprop);
	      else
		fprintf(nusmv_stderr, "given nABA");
	    }
	    fprintf(nusmv_stderr, "\n");
	    
	    break;
	  }
	  case BMC_INCRSAT_SATISFIABLE_PROBLEM:
	    fprintf(nusmv_stderr, "-- ");
	    if(pslprop)
	      print_spec(nusmv_stderr, pslprop);
	    else
	      fprintf(nusmv_stderr, "given nABA");
	    fprintf(nusmv_stderr, "  is false\n");
	    if(pslprop)
	      Prop_set_status(pslprop, Prop_False);
	    
	    found_solution = true;
	    
	    if(0) {
	      int t;
	      for(t = 0; t <= current_k; t++)
		{
		  int v;
		  lsGen iterator;
		  Sugar_Vertex_t *vertex;
		  fprintf(nusmv_stderr, "Time %d: ", t);
		  v = 0;
		  lsForEachItem(info->aba->state_list, iterator, vertex)
		    {
		      node_ptr var = _MH_get_Q_var(info, vertex->id);
		      be_ptr be_v_i = BeEnc_name_to_timed(be_enc, var, _real_k(t));
		      int result = Bmc_IncrSat_get_solution_value(solver, be_v_i);
		      fprintf(nusmv_stderr, "Q_%d = %s ", v, result?"T":"F");
		      v++;
		    }
		  v = 0;
		  lsForEachItem(info->aba->state_list, iterator, vertex)
		    {
		      be_ptr be_v_i;
		      int result;
		      node_ptr var = _MH_get_P_var(info, vertex->id);
		      if(!var)
			continue;
		      be_v_i = BeEnc_name_to_timed(be_enc, var, _real_k(t));
		      result = Bmc_IncrSat_get_solution_value(solver, be_v_i);
		      fprintf(nusmv_stderr, "P_%d = %s ", v, result?"T":"F");
		      v++;
		    }
		  fprintf(nusmv_stderr, "\n");
		}
	    }


	    if (opt_counter_examples(options)) {
#ifdef MS
	      if(opt_use_standard_solver)
		{
		  /* Counter-examples with standard incremental SAT solver
		     not supported yet */
		}
	      else
		{
		  Trace_ptr trace = 
		    _generate_and_print_cntexample(be_enc, mh_layer,
						   zolver->TJ_solver,
						   Enc_get_bdd_encoding(), 
						   current_k, 
						   "BMC Counterexample",
						   info->l_var);
		  if(pslprop)
		    Prop_set_trace(pslprop, Trace_get_id(trace));
		}
#else
	      Trace_ptr trace = 
		_generate_and_print_cntexample(be_enc, mh_layer,
					       solver,
					       Enc_get_bdd_encoding(), 
					       current_k, 
					       "BMC Counterexample",
					       info->l_var);
	      if(pslprop)
		Prop_set_trace(pslprop, Trace_get_id(trace));
#endif
	    }
	    
	    break;
	    
	case BMC_INCRSAT_INTERNAL_ERROR:
	  internal_error("Sorry, solver answered with a fatal Internal " 
			 "Failure during problem solving.\n");
	  break;
	  
	case BMC_INCRSAT_TIMEOUT:
	case BMC_INCRSAT_MEMOUT:
	    internal_error("Sorry, solver ran out of resources and aborted "
			   "the execution.\n");
	    break;
	    
	default:
	  internal_error("Bmc_GenSolveLtlTJ: Unexpected value in satResult");
	  
	} /* switch */
      
      
      if(opt_verbose_level_gt(options, 1))
	fprintf(nusmv_stderr, "Deleting k-dependent stuff\n");
      
      
      /*
       * Release the volatile frame 1
       */
#ifdef MS
      _MS_goto_permanent_group(zolver);
#else
      Bmc_IncrSat_pop_frame(solver);
#endif
      
      /*
       * Remember the current_k
       */
      previous_k = current_k;
    
    }
  /*
   * Remove extra encoding layers
   */
  /*
  BaseEnc_remove_layer(BASE_ENC(Enc_get_bdd_encoding()),
		       SymbLayer_get_name(mh_layer));
  */
  BaseEnc_remove_layer(BASE_ENC(be_enc), SymbLayer_get_name(mh_layer));
  BaseEnc_remove_layer(BASE_ENC(bool_enc), SymbLayer_get_name(mh_layer));
  SymbTable_remove_layer(symb_table, mh_layer);
  mh_layer = 0;
  
  /*
   * Destroy the sat solver instance
   */
#ifdef MS
  _MS_destroy(zolver); zolver = 0;
#else
  Bmc_IncrSat_release(solver); solver = 0;
#endif

  /*Additional clean up*/
  _MH_info_destroy(info); info = 0;
  
  if(pslprop != PROP(NULL))
    {
      clear_str_assoc(ap_table);
      Sugar_AutomatonFree(aba); aba = 0;
      Sugar_FormulaFree(negated_sugar_prop); negated_sugar_prop = 0;
    }
  return 0;
}




/**Function********************************************************************

  Synopsis           [Find the unique atomic propositions and return a
  map of sugar atomic propositions mapped to psl atomic propositions.}

  Description []

  SideEffects        []

  SeeAlso            []

******************************************************************************/


static void findAtomicProps(Sugar_Formula_t *formula, lsList idlist) 
{
  if (formula == NIL(Sugar_Formula_t)) {
    return;
  }
  /* The formula is a leaf. */
  if (formula->type == Sugar_IDENTIFIER_c) {
    lsNewEnd (idlist, (lsGeneric) formula, LS_NH);
    return;
  }

  switch(formula->type) {
  
  case Sugar_FL_Releases_c:
  case Sugar_FL_Until_c:
  case Sugar_FL_Until__c:
  case Sugar_FL_Before_c:
  case Sugar_FL_Before__c: 
  case Sugar_FL_And_c:
  case Sugar_FL_Or_c:
  case Sugar_FL_Imply_c:
  case Sugar_FL_Equiv_c:
  case Sugar_Seq_FLProp_c:
  case Sugar_Seq_Imply_c:
  case Sugar_Seq_Imply2_c:
  case Sugar_Seq_Within_c:
  case Sugar_Seq_Within__c:
  case Sugar_Seq_Whilenot_c:
  case Sugar_Seq_Whilenot__c:
  case Sugar_RegExpr_Concat_c:
  case Sugar_RegExpr_Compose_c:
  case Sugar_RegExpr_Or_c:
  case Sugar_RegExpr_And_c:
  case Sugar_RegExpr_AndEq_c:
    findAtomicProps(formula->left, idlist);
    findAtomicProps(formula->right, idlist);
    break;
  case Sugar_FL_Not_c:
  case Sugar_FL_Next_c:
  case Sugar_FL_Always_c:
  case Sugar_FL_Never_c:
  case Sugar_FL_Eventually_c:
  case Sugar_RegExpr_Plus_c:
  case Sugar_RegExpr_Star_c:
  case Sugar_RegExpr_Equal_c:
  case Sugar_RegExpr_Goto_c:
  case Sugar_Seq_Always_c:
  case Sugar_Seq_Never_c:
  case Sugar_Seq_Eventually_c:
    findAtomicProps(formula->left, idlist);
    break;
  case Sugar_TRUE_c:
  case Sugar_FALSE_c:
    return;

  default:
    fprintf(nusmv_stderr, "Unknown op: %d\n", formula->type);
    nusmv_assert(false);
  }
}
/*
 * Similar to Bmc_Utils_generate_and_print_cntexample but modified
 * for my incremental SAT solver interface
 */
static Trace_ptr _generate_and_print_cntexample(BeEnc_ptr be_enc,
						SymbLayer_ptr psl_layer,
						Bmc_IncrSat_ptr solver, 
						BddEnc_ptr bdd_enc, 
						const int k, 
						const char* trace_name,
						be_ptr l_var)
{
  BmcSatTrace_ptr sat_trace;
  Trace_ptr trace;
  lsList be_model;
  node_ptr path;
  int bmc_tr;
  hash_ptr psl_layer_vars = 0;

  assert(l_var);

  /*	  
  The original
  be_model = Be_CnfModelToBeModel(BeEnc_get_be_manager(be_enc), 
				  SatSolver_get_model(solver));
  is replaced by the following:
  */
  {
    /* Find all vars that are in the PSL layer */
    psl_layer_vars = new_assoc();
    NodeList_ptr l = SymbLayer_get_all_vars(psl_layer);
    ListIter_ptr iter = NodeList_get_first_iter(l);
    while(!ListIter_is_end(iter))
      {
	node_ptr node = NodeList_get_elem_at(l, iter);
	insert_assoc(psl_layer_vars, node, node);
	iter = ListIter_get_next(iter);
      }
  }

  be_model = lsCreate();
  {
    int i = 0;
    /* For each time step ... */
    for(i = _real_k(0); i <= _real_k(k); i++)
      {
	/* get values of state and input variables */
	int index = BeEnc_get_first_untimed_var_index(be_enc, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
	while(BeEnc_is_var_index_valid(be_enc, index))
	  {
	    node_ptr var_node;
	    be_ptr var;
	    be_ptr var_real_i;
	    int result;

	    /* Ignore vars that belong to the PSL layer */
	    var_node = BeEnc_index_to_name(be_enc, index);
	    if(find_assoc(psl_layer_vars, var_node))
	      {
		index = BeEnc_get_next_var_index(be_enc, index, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
		continue;
	      }

	    var = BeEnc_index_to_var(be_enc, index);
	    var_real_i = BeEnc_var_to_timed(be_enc, var, i);

	    result = Bmc_IncrSat_get_solution_value(solver, var_real_i);
	    if(result == 0 || result == 1)
	      {
		be_ptr var_model_i =
		  BeEnc_var_to_timed(be_enc, var, _model_k(i));
		int rbc_index = Be_Var2Index(BeEnc_get_be_manager(be_enc), var_model_i);
		  /*int rbc_index = BeEnc_var_to_index(be_enc, var_model_i);*/
		if(result == 0)
		  rbc_index = -rbc_index;
		lsNewEnd(be_model, (lsGeneric)rbc_index, LS_NH);
	      }

	    /* Go to next var */
	    index = BeEnc_get_next_var_index(be_enc, index, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
	  }
      }
  }
  /* Free PSL layer vars */
  free_assoc(psl_layer_vars); psl_layer_vars = 0;


  /* As no real be_prob was ever generated,
     it is replaced by a dummy, non-constant be */
  sat_trace = BmcSatTrace_create(BeEnc_name_to_timed(be_enc, l_var, 0),
				 be_model);
  /*sat_trace = BmcSatTrace_create(be_prob, be_model);*/
  path = NodeList_to_node_ptr(
	    BmcSatTrace_get_symbolic_model(sat_trace, be_enc, k));
	  
  trace = Trace_create_from_state_input_list(bdd_enc,
					     trace_name,
					     TRACE_TYPE_CNTEXAMPLE,
					     path);

  bmc_tr = TraceManager_register_trace(global_trace_manager, trace);
	  
  /* Print the trace using default plugin */
  fprintf(nusmv_stdout,
	  "-- as demonstrated by the following execution sequence\n");

  /* New code: find and print the loopback point */
  {
    int i = 0;
    int loopback = -1;
    for(i = 0; i <= k; i++)
      {
	be_ptr l_var_real_i = BeEnc_name_to_timed(be_enc, l_var, _real_k(i));
	int result = Bmc_IncrSat_get_solution_value(solver, l_var_real_i);
	assert(result == 0 || result == 1);
	if(result == 1)
	  {
	    assert(loopback == -1);
	    loopback = i;
	  }
      }
    if(loopback != -1)
      {
	/* NuSMV prints traces so that first state is numbered 1 */
	fprintf(nusmv_stdout,
		"   the loop starts at state %d (that is redundantly printed also as state %d)\n", loopback+1-1, k+1);
      }
    else
      {
	fprintf(nusmv_stdout, "   the execution has no loop\n");
      }
  }

  TraceManager_execute_plugin(global_trace_manager, 
		      TraceManager_get_default_plugin(global_trace_manager),
		      bmc_tr);

  BmcSatTrace_destroy(&sat_trace);
  lsDestroy(be_model, 0);

  return trace;
}
