#include "bmcPartitionMap.h"

static void 
visit(lsList fatherlist,int* low, int* dfsnum, int* N,Sugar_Vertex_t *node, struct BmcPartitionMap *map); 


static int
min(int x1, int x2);

/*Allocate memory for the partition map and intialise everything to zero*/

EXTERN struct BmcPartitionMap*
Bmc_PartitionMapCreate(const Sugar_Automaton_t *aut)
{
  int i=0;
  struct BmcPartitionMap* map=(struct BmcPartitionMap*)malloc(sizeof(struct BmcPartitionMap));
 
  nusmv_assert(aut!=NULL);
  
  map->numPartitions=0;
  map->aut=aut;
  map->map=(int*) malloc (sizeof(int) * lsLength(aut->state_list));
  for (i=0; i<lsLength(aut->state_list); i++) {
    (map->map)[i]=0;
  }
  map->vertexmap=(int*) malloc (sizeof(int) * lsLength(aut->state_list));
  for (i=0; i<lsLength(aut->state_list); i++) {
    (map->vertexmap)[i]=0;
  }
  map->partsize =(int*) malloc (sizeof(int) * lsLength(aut->state_list));
  for (i=0; i<lsLength(aut->state_list); i++) {
    (map->partsize)[i]=0;
  }
  map->idToVertexId=Bmc_Hash_new_htable();
  return map;
}

/*Delete a partition map*/

EXTERN void 
Bmc_PartitionMapDelete(struct BmcPartitionMap* map)
{
  nusmv_assert(map!=NULL);
  free(map->map);
  free(map->vertexmap);
  free(map->partsize);
  Bmc_Hash_delete_table(map->idToVertexId);
  free(map);
  return;
}

/*Get the partition id of a vertex vertex_id*/

EXTERN int
Bmc_PartitionMapGet (struct BmcPartitionMap *map, int vertex_id)
{
  nusmv_assert(map!=NULL);
  nusmv_assert(lsLength(map->aut->state_list)>vertex_id);
  return (map->map)[vertex_id];
}

/*Get the partition id of a vertex with the automaton id*/

EXTERN int
Bmc_PartitionMapGet_AutId (struct BmcPartitionMap *map, int aut_id)
{
  nusmv_assert(map!=NULL);
  int vertex_id=Bmc_PartitionMap_AutToId(map,aut_id);
  return Bmc_PartitionMapGet(map,vertex_id);
}

/*Set the partition id of a vertex vertex_id*/

EXTERN void
Bmc_PartitionMapSet (struct BmcPartitionMap *map, int vertex_id, int partitionId)
{
  nusmv_assert(map!=NULL);
  nusmv_assert(lsLength(map->aut->state_list)>vertex_id);
  nusmv_assert(map->numPartitions>partitionId);
  (map->map)[vertex_id]=partitionId;
}

EXTERN int
Bmc_PartitionMap_AutToId (struct BmcPartitionMap *map, int aut_id)
{  
  int result=Bmc_Hash_find(map->idToVertexId, aut_id);
  nusmv_assert(result!=BMC_HASH_NOTFOUND);
  return result;
}

EXTERN int
Bmc_PartitionMap_IdToAut (struct BmcPartitionMap *map, int vertex_id)
{
  nusmv_assert(map!=NULL);
  nusmv_assert(lsLength(map->aut->state_list)>vertex_id);
  return (map->vertexmap)[vertex_id];
}


/*Compute the partitions of a newly created partition map.
  States that belong to the same MSCC are part of 
  the same partition. The partitions are computed using a
  nested depth first search procedure due to Tarjan. States
  that are furthest from the initial state will have the smallest
  partition id.
*/

EXTERN void
Bmc_PartitionMapInit (struct BmcPartitionMap *map)
{
  nusmv_assert(map!=NULL);
  
  int index1;
  int index2;
  int N=1;
  array_t *item;
  Sugar_Vertex_t *node;
  /*Stack for storing MSCCs*/
  lsList fatherlist=lsCreate();
  /*Arrays for storing info on states*/
  int *dfsnum=calloc(lsLength(map->aut->state_list),sizeof(int));
  int *low=calloc(lsLength(map->aut->state_list),sizeof(int));

  /*Start a DFS for each unvisited initial state*/
  arrayForEachItem(array_t*, map->aut->initial_states, index1, item) {
    arrayForEachItem(Sugar_Vertex_t*, item, index2, node) {
      if(Bmc_Hash_find(map->idToVertexId, node->id)==BMC_HASH_NOTFOUND) {
	visit(fatherlist,low,dfsnum,&N,node,map);
      }
    }
  }
  free(low);
  free(dfsnum);
  lsDestroy(fatherlist,NULL);
  
  return;
}

/*Get the size of the partition a vertex belongs to*/
EXTERN int
Bmc_PartitionMapPartSize (struct BmcPartitionMap *map, int vertex_id)
{
  nusmv_assert(map!=NULL);
  nusmv_assert(lsLength(map->aut->state_list)>vertex_id);
  return (map->partsize)[vertex_id];
}

/*Set the size of the partition a vertex belongs to*/
EXTERN void
Bmc_PartitionMapPartSizeSet (struct BmcPartitionMap *map, int vertex_id, int size)
{
  nusmv_assert(map!=NULL);
  nusmv_assert(lsLength(map->aut->state_list)>vertex_id);
   nusmv_assert(lsLength(map->aut->state_list)>=size);
  (map->partsize)[vertex_id]=size;
}

static int
min(int x1, int x2)
{
  if(x1<x2) {
    return x1;
  }    
  return x2;
}


static void
visit(lsList fatherlist, int* low, int* dfsnum, 
int *N, Sugar_Vertex_t *node, struct BmcPartitionMap *map)
{
  nusmv_assert(*N <= lsLength(map->aut->state_list));
  nusmv_assert(Bmc_Hash_find(map->idToVertexId, node->id)==BMC_HASH_NOTFOUND);

  lsNewBegin(fatherlist, (lsGeneric) node, LS_NH);
  Bmc_Hash_insert(map->idToVertexId, node->id, *N-1);
  (map->vertexmap)[*N-1]=node->id;
  dfsnum[*N-1]=*N;
  *N=*N+1;
  low[Bmc_Hash_find(map->idToVertexId, node->id)] = dfsnum[Bmc_Hash_find(map->idToVertexId, node->id)];
  /*Visit all unvisited successors*/
  {
    lsGen  edge_iter;
    Sugar_Edge_t *edge;
    lsForEachItem(node->edge_out_list, edge_iter, edge) {
      lsGen node_iter;
      Sugar_Vertex_t *successor;      
      lsForEachItem(edge->dest_vertex_list, node_iter, successor) {
	if (Bmc_Hash_find(map->idToVertexId, successor->id)==BMC_HASH_NOTFOUND) { 
	  visit(fatherlist,low,dfsnum,N,successor,map);
	  low[Bmc_Hash_find(map->idToVertexId, node->id)] = 
	    min(low[Bmc_Hash_find(map->idToVertexId, node->id)], low[Bmc_Hash_find(map->idToVertexId, successor->id)]);
	} 
	else {
	  if (dfsnum[Bmc_Hash_find(map->idToVertexId, successor->id)]!=0) { 
	    low[Bmc_Hash_find(map->idToVertexId, node->id)] = 
	      min(low[Bmc_Hash_find(map->idToVertexId, node->id)], dfsnum[Bmc_Hash_find(map->idToVertexId, successor->id)]);
	  }
	}
      }
    }
  }
  if (low[Bmc_Hash_find(map->idToVertexId, node->id)]==dfsnum[Bmc_Hash_find(map->idToVertexId, node->id)]) {    
    /*Assign the partition number to the new MSCC*/
    lsGen iter=lsStart(fatherlist);
    int size=0;
    Sugar_Vertex_t *vertex;
    do {
      lsNext(iter,(lsGeneric *) &vertex, LS_NH);
      (map->map)[Bmc_Hash_find(map->idToVertexId, vertex->id)]=map->numPartitions;  
      dfsnum[Bmc_Hash_find(map->idToVertexId, vertex->id)]=0; /*Reset dfsnum*/
      size++; /*The size of the component*/
    } while (node!=vertex); 
    map->numPartitions++;
    vertex=0;
    /*Assign size of new MSCC to the vertex id*/
    do {
      lsDelBegin(fatherlist,(lsGeneric *) &vertex);
      Bmc_PartitionMapPartSizeSet (map, 
				   Bmc_Hash_find(map->idToVertexId, vertex->id), 
				   size); 
    } while (node!=vertex); 
  }
  return;
}
