/**CFile***********************************************************************

  FileName [bmcPsl.c]

  PackageName [bmc]

  Synopsis [Functions for solving PSL problems with BMC]

  Description []

  SeeAlso  []

  Author   [Timo Latvala]

  Copyright [ This file is part of the ``bmc'' package of NuSMV
  version 2.  Copyright (C) 2004 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it> or
  email to <nusmv-users@irst.itc.it>.  Please report bugs to
  <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to
  <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "bmcBmc.h"
#include "dag/dag.h"

#include "node/node.h"
#include "be/be.h"
#include "mc/mc.h" /* for print_spec */
#include "sat/sat.h" /* for solver and result */
#include "sat/SatSolver.h"
#include "sat/SatIncSolver.h"
#include "prop/prop.h"

#include "bmcInt.h"
#include "bmcGen.h"
#include "bmcSatTrace.h"
#include "bmcDump.h"
#include "bmcModel.h"
#include "bmcWff.h"
#include "bmcConv.h"
#include "bmcUtils.h"
#include "bmcTJIncrSat.h"
#include "bmcSatTrace.h"

#include "mc/mc.h" /* for print_{explanation, spec, invar} */
#include "utils/error.h"

#include "trace/Trace.h"
#include "trace/TraceManager.h"

#include "compile/symb_table/SymbType.h"
#include "compile/symb_table/SymbLayer.h"
#include "compile/type_checking/type_checking.h"

#include "enc/enc.h"
#include "enc/be/BeEnc.h"
#include "utils/error.h"

#include "bmcPsl.h"
#include "sugarInt.h"
#include "parser/psl/pslInt.h"
#include "parser/psl/pslExpr.h"
#include "parser/psl/pslNode.h"
#include "parser/psl/psl_grammar.h"
#include "bmcPartitionMap.h"
#include "bmcPslToSugar.h"
#include "parser/symbols.h"
#include "utils/str_assoc.h"
#include "utils/assoc.h"

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

struct WABA_TAG 
{
  /* lsLists below are lists of node_ptr */

  /* System state variables occurring in transition relation of the model*/
  lsList   transition_state_vars;

  /* Auxiliary variables needed in the PLTL BMC encoding */
  node_ptr *waba_vars;  /* [[s_q]]_i^d and [[s_q]]_L^d vars */
  node_ptr l_var;                 /* The loop selector variable */
#ifdef _OLD_INLOOP
  node_ptr InLoop_var;                /* The InLoop variable*/
#else
  /* array[model_time] of be */
  array_t *InLoop_bes;
#endif

  /* TJ: How many model fairnesses? */
  int model_fairness_nof;
  /* TJ: array[fairness_set_num] of be_ptr */
  array_t *model_fairness_untimed_bes;
  /* TJ: array[fairness_set_num] of array[model_time] of be */
  array_t *model_fairness_timed_bes;

  /*Automata fairness: list of (variable, vertexid) pairs */
  lsList automata_fairness;
  /* Buchi "one state" condition, untimed be_ptr */
  be_ptr buchi_one_state_condition;
  /* System state variables occurring in formula */
  lsList   formula_state_vars;
  /* System input variables occurring in formula */
  lsList   formula_input_vars;
};

typedef struct WABA_TAG WABA_state_vars_t;


/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/



static void findAtomicProps(Sugar_Formula_t *formula, lsList aplist);

static int checkMemberShip(struct BmcPartitionMap *partmap, array_t * accept_states, int vertexid);

static be_ptr LoopSelectVar(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
			    const int time, const int k);

static be_ptr AutomataStateVarTime(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
				   struct BmcPartitionMap *partmap, const int time, const int depth, 
				   const int stateid, const int k);

static be_ptr AutomataStateVarUntimed(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
				      struct BmcPartitionMap *partmap, const int depth, const int stateid);

static be_ptr AutomataStateVarL(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
				struct BmcPartitionMap *partmap, const int depth, const int stateid, 
				const int k);

static be_ptr AutomataInstantiate(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
				  struct BmcPartitionMap *partmap, str_hash_ptr ap_table, 
				  const int time, const int k);

static be_ptr InLoop(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, const int time, const int k);

static be_ptr LoopConstraints(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
			    const int previous_k, const int k);

static be_ptr BaseConditions(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
			     struct BmcPartitionMap *partmap, const int k);


static be_ptr kDependentConstraints(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
				   struct BmcPartitionMap *partmap, const int k);
     
static be_ptr EquivalentToSE(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, const int time, const int k);

static be_ptr labelToBe(const BeEnc_ptr be_enc,const int time, const int k, PslNode_ptr psl_ap);
 
static be_ptr labelToBeAux(const BeEnc_ptr be_enc,const node_ptr ap_wff, const int time, const int k);

static be_ptr _make_buchi_one_state_cond(const BeEnc_ptr be_enc,
					 WABA_state_vars_t *waba_vars, 
					 struct BmcPartitionMap *partmap);

static Sugar_Automaton_t *bmcTestAut(int testcase);

static Trace_ptr _generate_and_print_cntexample(BeEnc_ptr be_enc,
						SymbLayer_ptr psl_layer,
						Bmc_IncrSat_ptr solver, 
						BddEnc_ptr bdd_enc, 
						const int k, 
						const char* trace_name,
						be_ptr l_var);

static void WABA_init_state_vars(SymbLayer_ptr layer, WABA_state_vars_t * waba_vars, 
				 struct BmcPartitionMap *partmap);

static void WABA_delete_state_vars(WABA_state_vars_t *waba_vars);

static be_ptr _equal_vectors_formula(const BeEnc_ptr be_enc,
				     lsList vars,
				     const unsigned int i,
				     const unsigned int j);
static int unroll_depth(struct BmcPartitionMap *partmap, int vertexid);

static be_ptr be_implies_iff(const Be_Manager_ptr be_mgr, be_ptr lhs, be_ptr rhs); 

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

typedef struct opt_TAG {
  /*Use implications instead of bi-implications where possible. Default is false.*/
  boolean implications;
  /*Add monotonicity constraints to encoding for faster propagation. Default is false.*/
  boolean mono_constraints;
  /*No unrolling for accepting components in the waba. Default is true.*/
  boolean no_acc_unrolling;
  /*Perform only partial unrolling. 0=unroll all, 1=only a single unrolling, 2=two unrollings, etc.
   The default value is 0.*/
  int partial_unrolling;
  /*Unfold the WABA to a Buchi automaton*/
  boolean buchi;
} opt_options_t;

static opt_options_t optimisations;

static int max_partition_size(struct BmcPartitionMap *partmap)
{
  if(optimisations.buchi) {
    /*Partition size does not matter for Buchi, but we need a set auxilliary vars for
      the exactly one constraint*/
    return 2;
  }
  if(optimisations.partial_unrolling) {
    return optimisations.partial_unrolling+1;
  }

  return (lsLength(partmap->aut->state_list)-partmap->numPartitions)+1;
}


static void init_opt_options() 
{
  switch(opt_waba_optimize(options)) {
  case 0:
    optimisations.implications=false;
    optimisations.mono_constraints=false;
    optimisations.no_acc_unrolling=false;
    break;
  case 1:
    optimisations.implications=false;
    optimisations.mono_constraints=false;
    optimisations.no_acc_unrolling=true;
    break;
  case 2:
    optimisations.implications=true;
    optimisations.mono_constraints=false;
    optimisations.no_acc_unrolling=true;
    break;
  case 3:
    optimisations.implications=true;
    optimisations.mono_constraints=true;
    optimisations.no_acc_unrolling=true;
    break;
  default:
    fprintf(nusmv_stderr, "Unrecognised waba optimisation value, %d", opt_waba_optimize(options));
  }
  optimisations.partial_unrolling=opt_waba_unroll(options);
  optimisations.buchi=opt_waba_buchi(options);
  return; 
}


EXTERN node_ptr boolean_range;

#ifdef BENCHMARKING
#include <unistd.h>
#include <sys/times.h>
struct tms time_bmc_start;
struct tms time_gen_start;
struct tms time_solver_start;
struct tms time_temp;
static double time_diff(struct tms *start, struct tms *stop)
{
  return ((((double)stop->tms_utime + (double)stop->tms_stime) -
	   ((double)start->tms_utime + (double)start->tms_stime)) /
	  (double)CLK_TCK);
}
#endif


/*
 * Auxiliary wrapper stuff needed to use both the standard and my incremental
 * SAT solver interfaces
 */
typedef struct _MetaSolver_TAG
{
  BeEnc_ptr        be_enc;

  int              using_volatile_group;

  Bmc_IncrSat_ptr  TJ_solver;

  SatIncSolver_ptr S_solver;
  SatSolverGroup   S_permanent_group;
  SatSolverGroup   S_volatile_group;
} _MetaSolver;

static _MetaSolver *_MS_create(BeEnc_ptr be_enc,
			       const int use_TJ_solver)
{
  assert(be_enc);

  _MetaSolver *ms = ALLOC(_MetaSolver, 1);
  ms->be_enc = be_enc;
  ms->using_volatile_group = 0;
  ms->TJ_solver = 0;
  ms->S_solver = 0;
  if(use_TJ_solver)
    {
      ms->TJ_solver = Bmc_IncrSat_init(get_sat_solver(options));
      if(ms->TJ_solver == 0)
	{
	  fprintf(nusmv_stderr,
		  "Incremental sat solver '%s' is not available.\n", 
		  get_sat_solver(options));
	  FREE(ms);
	  return 0;
	}
    }
  else
    {
      ms->S_solver = Sat_CreateIncSolver(get_sat_solver(options));
      if(ms->S_solver == SAT_INC_SOLVER(NULL)) {
	fprintf(nusmv_stderr,
		"Incremental sat solver '%s' is not available.\n", 
		get_sat_solver(options));
	FREE(ms);
	return 0;
      }
      ms->S_permanent_group =
	SatSolver_get_permanent_group(SAT_SOLVER(ms->S_solver));
    }
  return ms;
}
static void _MS_destroy(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_release(ms->TJ_solver); ms->TJ_solver = 0;
    }
  else
    {
      SatIncSolver_destroy(ms->S_solver); ms->S_solver = 0;
    }
}
static void _MS_goto_permanent_group(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  assert(ms->using_volatile_group);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_pop_frame(ms->TJ_solver);
    }
  else
    {
      SatIncSolver_destroy_group(ms->S_solver, ms->S_volatile_group);
    }
  ms->using_volatile_group = 0;
}
static void _MS_goto_volatile_group(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  assert(!ms->using_volatile_group);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_new_frame(ms->TJ_solver);
    }
  else
    {
      ms->S_volatile_group = SatIncSolver_create_group(ms->S_solver);
    }
  ms->using_volatile_group = 1;
}
static void _MS_force_true(_MetaSolver *ms,
			   be_ptr be_constraint)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);

  if(ms->TJ_solver)
    {
      Bmc_IncrSat_insert(ms->TJ_solver, be_constraint);
      Bmc_IncrSat_force_true(ms->TJ_solver, be_constraint);
    }
  else
    {
      SatSolver_ptr solver = SAT_SOLVER(ms->S_solver);
      Be_Cnf_ptr cnf = Be_ConvertToCnf(BeEnc_get_be_manager(ms->be_enc),
				       be_constraint);
      if(ms->using_volatile_group)
	{
	  SatSolver_add(solver, cnf, ms->S_volatile_group);
	  SatSolver_set_polarity(solver, cnf, 1, ms->S_volatile_group);
	}
      else
	{
	  SatSolver_add(solver, cnf, ms->S_permanent_group);
	  SatSolver_set_polarity(solver, cnf, 1, ms->S_permanent_group);
	}
      Be_Cnf_Delete(cnf);
    }
}
static void _MS_force_constraint_list(_MetaSolver *ms, lsList constraints)
{
  lsGen  iterator;
  be_ptr be_constraint;
  
  lsForEachItem(constraints, iterator, be_constraint)
    _MS_force_true(ms, be_constraint);
}
static Bmc_IncrSat_result _MS_solve(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);

  if(ms->TJ_solver)
    {
      return Bmc_IncrSat_solve(ms->TJ_solver);
    }
  else
    {
      return SatSolver_solve_all_groups(SAT_SOLVER(ms->S_solver));
    }
}

static char rcsid[] UTIL_UNUSED = "$Id: bmcPsl.c,v 1.31 2006/03/09 10:36:19 tjunttil Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef struct var_be_TAG {
  node_ptr var;
  be_ptr exp;
} var_be_pair_t;


/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/


/*
 * Routines for the state indexing scheme
 * State 0 is the L state and 1 is the E state
 * The first real state is state 2.
 */
static int _L_state() {return 0; }
static int _E_state() {return 1; }
static int _real_k(int k) {return k+2; }
static unsigned int _model_k(int k) {assert(k >= 2); return k-2; }

/*Return lhs -> rhs if the implication optimisation is active, 
 * otherwise return lhs <-> rhs
 */
static be_ptr be_implies_iff(const Be_Manager_ptr be_mgr, be_ptr lhs, be_ptr rhs) 
{
  if(optimisations.implications)
    {
      return Be_Implies(be_mgr, lhs, rhs);
    }
  return Be_Iff(be_mgr, lhs, rhs);
}




/*Return the maximum unrolling depth for a state taking
  into account currently active optimisations*/
static int unroll_depth(struct BmcPartitionMap *partmap, int vertexid)
{
  /*No unrolling for Buchi automata checking*/
  if (optimisations.buchi)
    return 0;

  if(!optimisations.no_acc_unrolling && optimisations.partial_unrolling==0) {
    return Bmc_PartitionMapPartSize (partmap, vertexid);
  }
  
  if(optimisations.partial_unrolling==0) {
      return Bmc_PartitionMapPartSize (partmap, vertexid);
  } 
  else {
    return (optimisations.partial_unrolling < Bmc_PartitionMapPartSize (partmap, vertexid)) ? 
      optimisations.partial_unrolling : Bmc_PartitionMapPartSize (partmap, vertexid);
  }
  /*Always valid over approximation for maximum size of partition*/
  return max_partition_size(partmap);
}

/*
 * Get the initial state and invariant constraint at time i.
 * Similar to Bmc_Model_GetInit0(be_fsm) except for the i argument.
 */
static be_ptr Bmc_TJ_GetInitI(const BeFsm_ptr be_fsm, const int i)
{
  BeEnc_ptr be_enc = BeFsm_get_be_encoding(be_fsm);
  be_ptr init = BeEnc_untimed_expr_to_timed(be_enc,
			      Be_And(BeEnc_get_be_manager(be_enc), 
				     BeFsm_get_init(be_fsm),
				     BeFsm_get_invar(be_fsm)),
					    i);
  return init;
}

/*
 * An auxiliary routine to add a new state variable with 'name' to 'layer'
 */
static node_ptr _add_new_state_variable(SymbLayer_ptr layer, char *name)
{
  SymbType_ptr symbolicType;
  node_ptr node;
  
  node = find_node(DOT, Nil, sym_intern(name));
  assert(node);

  if(!SymbLayer_can_declare_var(layer, node))
    error_redefining(node); 
    
  symbolicType = SymbType_create(SYMB_TYPE_ENUM, boolean_range);
  SymbLayer_declare_state_var(layer, node, symbolicType);
  return node;
}

/*
 * Returns a fresh state variable allocated to layer with name #WABA_t'index'.
 * index is increased by one.
 */

static node_ptr _alloc_Q_state_var(SymbLayer_ptr layer, unsigned int index)
{
  char new_var_name[16];
  node_ptr new_var = 0;
  
  sprintf(new_var_name, "#WABA_Q_%u", index);
  new_var = _add_new_state_variable(layer, new_var_name);
  return new_var;
}

/* Model fairness state var allocator*/
static node_ptr _alloc_MF_state_var(SymbLayer_ptr layer, unsigned int index)
{
  char new_var_name[16];
  node_ptr new_var = 0;
  
  sprintf(new_var_name, "#WABA_MF_%u", index);
  new_var = _add_new_state_variable(layer, new_var_name);
  return new_var;
}

/* Automata fairness state var allocator*/
static node_ptr _alloc_AF_state_var(SymbLayer_ptr layer, unsigned int index)
{
  char new_var_name[16];
  node_ptr new_var = 0;
  
  sprintf(new_var_name, "#WABA_AF_%u", index);
  new_var = _add_new_state_variable(layer, new_var_name);
  return new_var;
}

/*
 * Create an empty waba_state_vars_t
 */
static WABA_state_vars_t *_WABA_state_vars_create()
{
  WABA_state_vars_t *WABA_vars = ALLOC(WABA_state_vars_t, 1);
  assert(WABA_vars);

  WABA_vars->transition_state_vars = lsCreate();
  WABA_vars->l_var = NULL;
#ifdef _OLD_INLOOP
  WABA_vars->InLoop_var=NULL;
#else
  WABA_vars->InLoop_bes = array_alloc(be_ptr, 0);
#endif
  WABA_vars->waba_vars= NULL;
  WABA_vars->formula_state_vars = lsCreate();
  WABA_vars->formula_input_vars = lsCreate();
  WABA_vars->model_fairness_nof = 0;
  WABA_vars->model_fairness_untimed_bes = 0;
  WABA_vars->model_fairness_timed_bes = 0;
  WABA_vars->automata_fairness = lsCreate();
  WABA_vars->buchi_one_state_condition = NULL;

  return WABA_vars;
}


static void
WABA_init_state_vars(SymbLayer_ptr layer,
		     WABA_state_vars_t * waba_vars, 
		     struct BmcPartitionMap *partmap)
{
  int i;
  waba_vars->waba_vars = (node_ptr *)malloc(sizeof(node_ptr)*
					    lsLength(partmap->aut->state_list)*
					    (max_partition_size(partmap)+2));
  for (i = 0;
       i < lsLength(partmap->aut->state_list)*(max_partition_size(partmap)+2);
       i++)
    {
      (waba_vars->waba_vars)[i]= _alloc_Q_state_var(layer, i);
    }
  return;
}

static void WABA_delete_state_vars(WABA_state_vars_t *waba_vars)
{
  lsGen iterator;
  var_be_pair_t *pair_be;

  free(waba_vars->waba_vars);
  lsDestroy(waba_vars->formula_state_vars, LS_NIL);
  lsDestroy(waba_vars->formula_input_vars, LS_NIL);
  lsDestroy(waba_vars->transition_state_vars, LS_NIL);
#ifndef _OLD_INLOOP
  assert(waba_vars->InLoop_bes); array_free(waba_vars->InLoop_bes);
  waba_vars->InLoop_bes = 0;
#endif
  lsForEachItem(waba_vars->automata_fairness, iterator, pair_be)
    {
      free(pair_be);
    }
  waba_vars->model_fairness_nof = 0;
  if(waba_vars->model_fairness_untimed_bes) {
    array_free(waba_vars->model_fairness_untimed_bes);
    waba_vars->model_fairness_untimed_bes = 0;
  }
  if(waba_vars->model_fairness_timed_bes) {
    int i;
    for(i = 0; i < array_n(waba_vars->model_fairness_timed_bes); i++) {
      array_t *timed_bes = array_fetch(array_t*,
				       waba_vars->model_fairness_timed_bes,
				       i);
      if(timed_bes)
	array_free(timed_bes);
      array_insert(array_t*, waba_vars->model_fairness_timed_bes, i, 0);
    }
    array_free(waba_vars->model_fairness_timed_bes);
    waba_vars->model_fairness_timed_bes = 0;
  }
  lsDestroy(waba_vars->automata_fairness, LS_NIL);
  free(waba_vars);
}


static be_ptr _equal_vectors_formula(const BeEnc_ptr be_enc,
				     lsList vars,
				     const unsigned int i,
				     const unsigned int j)
{
  lsGen    iterator;
  node_ptr node_sv;

  be_ptr be_constraint;

  Be_Manager_ptr be_mgr;

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Initialize constraint */
  be_constraint = Be_Truth(be_mgr);

  lsForEachItem(vars, iterator, node_sv)
    {
      /* Get state variable sv at time i */
      be_ptr be_sv_i = BeEnc_name_to_timed(be_enc, node_sv, i);
      /* Get state variable sv at time j */
      be_ptr be_sv_j = BeEnc_name_to_timed(be_enc, node_sv, j);

      be_constraint = Be_And(be_mgr,
			     be_constraint,
			     Be_Iff(be_mgr, be_sv_i, be_sv_j));
    }

  return be_constraint;
}



static void
_init_model_fairness(const BeEnc_ptr be_enc,
		     const SymbLayer_ptr layer,
		     WABA_state_vars_t * const waba_vars,
		     Prop_ptr coi_prop)
{  
  SexpFsm_ptr sexp_fsm = Prop_get_bool_sexp_fsm(coi_prop);
  assert(sexp_fsm != (SexpFsm_ptr)NULL);

  {
    node_ptr compassion_list = SexpFsm_get_compassion(sexp_fsm);
    if(!is_list_empty(compassion_list))
      internal_error("%s:%d: Compassion not handled",__FILE__,__LINE__);
  }

  {
    /* Initialize model fairness stuff */
    int fairness_num = 0;
    node_ptr justice_list = SexpFsm_get_justice(sexp_fsm);
    assert(waba_vars->model_fairness_nof == 0);
    assert(waba_vars->model_fairness_untimed_bes == 0);
    assert(waba_vars->model_fairness_timed_bes == 0);
    waba_vars->model_fairness_untimed_bes = array_alloc(be_ptr, 0);
    waba_vars->model_fairness_timed_bes = array_alloc(array_t *, 0);
    while(justice_list != LS_NIL)
      {
	be_ptr be_expr = Bmc_Conv_Bexp2Be(be_enc, Bmc_Wff_MkNnf((node_ptr)car(justice_list)));
	array_insert(be_ptr, waba_vars->model_fairness_untimed_bes,
		     fairness_num, be_expr);
	array_t *timed_array = array_alloc(be_ptr, 0);
	array_insert(array_t *, waba_vars->model_fairness_timed_bes,
		     fairness_num, timed_array);
	fairness_num++;
	justice_list = cdr(justice_list);
      }
    waba_vars->model_fairness_nof = fairness_num;
  }
  return;
} 


static be_ptr BaseConditions(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
			     struct BmcPartitionMap *partmap, const int k)
{
  be_ptr result=Be_Truth(BeEnc_get_be_manager(be_enc));
  int stateid;
  /*Loop conditions base */
  /*l_0 <=> \bot*/
  result=Be_And(BeEnc_get_be_manager(be_enc),
		result,
		Be_Iff(BeEnc_get_be_manager(be_enc),
		       LoopSelectVar(be_enc,waba_vars,0,k),
		       Be_Falsity(BeEnc_get_be_manager(be_enc)))
		);
#ifdef _OLD_INLOOP
  /*InLoop_0 <=> \bot*/
  result=Be_And(BeEnc_get_be_manager(be_enc),
		result,
		Be_Iff(BeEnc_get_be_manager(be_enc),
		       InLoop(be_enc,waba_vars,0,k),
		       Be_Falsity(BeEnc_get_be_manager(be_enc)))
		);
#endif

  /*Automata Base conditions*/

  /*Ignore for Buchi model checking*/
  if(!optimisations.buchi) {
    for(stateid=0;stateid<lsLength(partmap->aut->state_list); stateid++) {
      /*|s_q|_L^{Q+1} <=> \bot if q\not\in F
	<=> \top if q\in F when unrolling final components*/
      if(!checkMemberShip(partmap, partmap->aut->accept_states,stateid)) {
	result=Be_And(BeEnc_get_be_manager(be_enc),
		      result,
		      Be_Iff(BeEnc_get_be_manager(be_enc),
			     AutomataStateVarL(be_enc,waba_vars,partmap,unroll_depth(partmap,stateid)+1,stateid,k),
			     Be_Falsity(BeEnc_get_be_manager(be_enc)))
		    );
      }
      /*If we are not optimising unrolling of final components they are unrolled*/
      else if (!optimisations.no_acc_unrolling) {
	result=Be_And(BeEnc_get_be_manager(be_enc),
		      result,
		      Be_Iff(BeEnc_get_be_manager(be_enc),
			     AutomataStateVarL(be_enc,waba_vars,partmap,unroll_depth(partmap,stateid)+1,stateid,k),
			     Be_Truth(BeEnc_get_be_manager(be_enc)))
		      );
      }
      
    }
  }

  /*Automata Initial state constraints*/
  {
    int index1, index2;
    array_t * item;
    Sugar_Vertex_t *node;
    be_ptr initial_cond=Be_Falsity(BeEnc_get_be_manager(be_enc));
    
    arrayForEachItem(array_t*, partmap->aut->initial_states, index1, item) {
      be_ptr and_states=Be_Truth(BeEnc_get_be_manager(be_enc));    
      arrayForEachItem(Sugar_Vertex_t*, item, index2, node) {
	and_states=Be_And(BeEnc_get_be_manager(be_enc), 
			  and_states,
			  AutomataStateVarTime(be_enc,waba_vars,partmap,0,0,
					       Bmc_PartitionMap_AutToId(partmap,node->id),k)
			  );
      }
      initial_cond=Be_Or(BeEnc_get_be_manager(be_enc), initial_cond, and_states);
    }
    result=Be_And(BeEnc_get_be_manager(be_enc),
		  result, 
		  Be_Iff(BeEnc_get_be_manager(be_enc),
			 initial_cond, 
			 Be_Truth(BeEnc_get_be_manager(be_enc))));
  }

  /*Automata fairness constraints*/
  if(lsLength(waba_vars->automata_fairness)!=0) {
    lsGen iter;
    var_be_pair_t *pair;
    lsForEachItem(waba_vars->automata_fairness,iter,pair) {
      result=Be_And(BeEnc_get_be_manager(be_enc),
		    result, 
		    Be_Iff(BeEnc_get_be_manager(be_enc),
			   BeEnc_name_to_timed(be_enc, pair->var, _real_k(0)),
			   Be_Falsity(BeEnc_get_be_manager(be_enc))
			   )
		    );
    }
  }


  /*Optimising constraints: if partial unrolling is active we must 
   * add the stabilisation constraint.
   */
  if (optimisations.partial_unrolling!=0 && !optimisations.buchi) {
    for(stateid=0;stateid<lsLength(partmap->aut->state_list); stateid++) {
      /*(|s_q|_L^0 <=> |s_q|_L^1 )*/
      result=Be_And(BeEnc_get_be_manager(be_enc),
		    result,
		    Be_Iff(BeEnc_get_be_manager(be_enc),
			   AutomataStateVarL(be_enc,waba_vars,partmap,
					     0,stateid,k),
			   AutomataStateVarL(be_enc,waba_vars,partmap,
					     1,stateid,k))
			   );
    }
  }

  



  return result;

}


static be_ptr kDependentConstraints(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
				    struct BmcPartitionMap *partmap, const int k)
{
  be_ptr result = Be_Truth(BeEnc_get_be_manager(be_enc));
  int depth,stateid;

  /*Loop conditions*/
  /*l_{k+1} <=> FALSE*/
  result = Be_And(BeEnc_get_be_manager(be_enc),
		  result,
		  Be_Iff(BeEnc_get_be_manager(be_enc),
			 LoopSelectVar(be_enc,waba_vars,k+1,k),
			 Be_Falsity(BeEnc_get_be_manager(be_enc)))
		  );

  /*s_E <=> s_k*/
  result=Be_And(BeEnc_get_be_manager(be_enc),
		result,
		_equal_vectors_formula(be_enc,waba_vars->transition_state_vars,
				       _real_k(k),_E_state()));

#ifdef _OLD_INLOOP
  /*InLoop_k <=> \top*/
  result=Be_And(BeEnc_get_be_manager(be_enc),
		result,
		Be_Iff(BeEnc_get_be_manager(be_enc),
		       InLoop(be_enc,waba_vars,k,k), 
		       Be_Truth(BeEnc_get_be_manager(be_enc)))
		);
#endif


  /*Unrolling conditions*/
  for(stateid=0; stateid<lsLength(partmap->aut->state_list); stateid++) {
    /*Final states are not unrolled when optimising, same goes for all states in Buchi model checking*/
    if((optimisations.no_acc_unrolling && checkMemberShip(partmap,partmap->aut->accept_states,stateid))
       || optimisations.buchi) {
      /*|s_q|_L^0 <=> |s_q|_{k+1}^0*/
      result=Be_And(BeEnc_get_be_manager(be_enc),
		    result,
		    Be_Iff(BeEnc_get_be_manager(be_enc),
			   AutomataStateVarL(be_enc,waba_vars,partmap,0,stateid,k),
			   AutomataStateVarTime(be_enc,waba_vars,partmap,k+1,0,stateid,k)
			   )
		    );
    }
    else {
      for (depth=0; depth<=unroll_depth(partmap,stateid); depth++) {
	/*|s_q|_L^{d+1} <=> |s_q|_{k+1}^d */
	result=Be_And(BeEnc_get_be_manager(be_enc),
		      result,
		      Be_Iff(BeEnc_get_be_manager(be_enc),
			     AutomataStateVarL(be_enc,waba_vars,partmap,depth+1,stateid,k),
			     AutomataStateVarTime(be_enc,waba_vars,partmap,k+1,depth,stateid,k)
			     )
		      );
      }
    }
  }
   
  /*Automata fairness constraints*/
  if(lsLength(waba_vars->automata_fairness)!=0)
    {
      lsGen iter;
      var_be_pair_t *pair;
      lsForEachItem(waba_vars->automata_fairness,iter,pair)
	{
	  /*Fair(var)_k <=> \top*/
	  result = 
	    Be_And(BeEnc_get_be_manager(be_enc),
		   result, 
		   Be_Iff(BeEnc_get_be_manager(be_enc),
			  BeEnc_name_to_timed(be_enc, pair->var, _real_k(k)),
			  Be_Truth(BeEnc_get_be_manager(be_enc))
			  )
		   );
	}
    }
  return result;
}


static void
addFairnessVars(SymbLayer_ptr layer, const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, 
		struct BmcPartitionMap* partmap, Prop_ptr coi_prop) {
  Be_Manager_ptr be_mgr=BeEnc_get_be_manager(be_enc);
  int var_index;
  SexpFsm_ptr sexp_fsm = Prop_get_bool_sexp_fsm(coi_prop);
  assert(sexp_fsm != (SexpFsm_ptr)NULL);  
  node_ptr list = SexpFsm_get_justice(sexp_fsm);
  node_ptr compassion_list = SexpFsm_get_compassion(sexp_fsm);
  if(!is_list_empty(compassion_list))
    internal_error("%s:%d: Compassion not handled",__FILE__,__LINE__);
  
  /*Only for Buchi model checking*/
  var_index = 0;
  if(optimisations.buchi)
    {
      var_be_pair_t *pair=ALLOC(var_be_pair_t,1);  
      pair->var = _alloc_AF_state_var(layer, var_index++);
      pair->exp = Be_Falsity(be_mgr);   
      lsNewEnd(waba_vars->automata_fairness,(lsGeneric) pair, LS_NH);
    }
  return;
}

/* Generate the extra condition needed for Buchi model checking.
 * The condition requires that exactly one state appears at each 
 * time point in a run.
 * Returns an untimed be_ptr. */
static be_ptr _make_buchi_one_state_cond(const BeEnc_ptr be_enc,
					 WABA_state_vars_t *waba_vars, 
					 struct BmcPartitionMap *partmap)
{
  Be_Manager_ptr be_mgr = BeEnc_get_be_manager(be_enc);
  be_ptr result = Be_Truth(be_mgr);
  const int autsize = lsLength(partmap->aut->state_list);
  int stateid;
  array_t *already_one = array_alloc(be_ptr, autsize);
  assert(already_one);
  assert(autsize >= 1);

  /* <<s_{q_0}>>_i <=> |[s_{q_0}]|_i */
  array_insert(be_ptr, already_one, 0,
	       AutomataStateVarUntimed(be_enc,waba_vars,partmap,0,0));
  /* For each vertex of the automaton except the first  */
  for(stateid = 1; stateid < autsize; stateid++)
    {
      /* <<s_{q_j}>>_i := <<s_{q_{j-1}}>>_i \vee |[s_{q_j}]|_i */
      array_insert(be_ptr, already_one, stateid,
		   Be_Or(be_mgr, 
			 array_fetch(be_ptr, already_one, stateid-1),
			 AutomataStateVarUntimed(be_enc,waba_vars,partmap,
						 0,stateid)));
    }
  for(stateid = 1; stateid < autsize; stateid++)
    {
      /* <<s_{q_{j-1}}>>_i  => \neg |[s_{q_j}]|_i*/
      result=Be_And(be_mgr, 
		    result,
		    Be_Implies(be_mgr, 
			       array_fetch(be_ptr, already_one, stateid-1),
			       Be_Not(be_mgr, 
				      AutomataStateVarUntimed(be_enc,waba_vars,
							      partmap,0,
							      stateid))));
    }
  /*<<s_{q_n}>>_i <=> \top*/
  result = Be_And(be_mgr, 
		  result,
		  Be_Iff(be_mgr, 
			 array_fetch(be_ptr, already_one, autsize-1),
			 Be_Truth(be_mgr)));

  array_free(already_one);
  return result;
}
#if 0
/*Generate the extra condition needed for Buchi model checking.
 *The condition requires that exactly one state appears at each 
 *time point in a run. */
static be_ptr buchi_one_state_cond(const BeEnc_ptr be_enc,
				   WABA_state_vars_t *waba_vars, 
				   struct BmcPartitionMap *partmap,
				   const int time,
				   const int k)
{
  Be_Manager_ptr be_mgr = BeEnc_get_be_manager(be_enc);
  be_ptr result = Be_Truth(be_mgr);
  const int autsize = lsLength(partmap->aut->state_list);
  int stateid;
  array_t *already_one = array_alloc(be_ptr, autsize);
  assert(already_one);
  assert(autsize >= 1);

  /* <<s_{q_0}>>_i <=> |[s_{q_0}]|_i */
  array_insert(be_ptr, already_one, 0,
	       AutomataStateVarTime(be_enc,waba_vars,partmap,time,0,0,k));
  /* For each vertex of the automaton except the first  */
  for(stateid = 1; stateid < autsize; stateid++)
    {
      /* <<s_{q_j}>>_i := <<s_{q_{j-1}}>>_i \vee |[s_{q_j}]|_i */
      array_insert(be_ptr, already_one, stateid,
		   Be_Or(be_mgr, 
			 array_fetch(be_ptr, already_one, stateid-1),
			 AutomataStateVarTime(be_enc,waba_vars,partmap,
					      time,0,stateid,k)));
    }
  for(stateid = 1; stateid < autsize; stateid++)
    {
      /* <<s_{q_{j-1}}>>_i  => \neg |[s_{q_j}]|_i*/
      result=Be_And(be_mgr, 
		    result,
		    Be_Implies(be_mgr, 
			       array_fetch(be_ptr, already_one, stateid-1),
			       Be_Not(be_mgr, 
				      AutomataStateVarTime(be_enc,waba_vars,
							   partmap,time,0,
							   stateid,k))));
    }
  /*<<s_{q_n}>>_i <=> \top*/
  result = Be_And(be_mgr, 
		  result,
		  Be_Iff(be_mgr, 
			 array_fetch(be_ptr, already_one, autsize-1),
			 Be_Truth(be_mgr)));

  array_free(already_one);
  return result;
}
/*Generate the extra condition needed for Buchi model checking.
 *The condition requires that exactly one state appears at each 
 *time point in a run. We implement this condition on state
 * variables |s_q|_i by using auxilliary variables <<s_q>>_i.
 * j  = 0:        <<s_{q_0}>>_i <=> |[s_{q_0}]|_i
 * j >= 1:        <<s_{q_j}>>_i <=> <<s_{q_{j-1}}>>_i \vee |[s_{q_j}]|_i
 * j >= 1:    <<s_{q_{j-1}}>>_i  => \neg |[s_{q_j}]|_i
 * j  = n:        <<s_{q_n}>>_i <=> \top
 * To implement the auxillary variables we use the depth=1 variables 
 *normally used on the WABA encoding.
 */
static be_ptr buchi_one_state_cond(const BeEnc_ptr be_enc,
				   WABA_state_vars_t *waba_vars, 
				   struct BmcPartitionMap *partmap,
				   const int time,
				   const int k)
{
  Be_Manager_ptr be_mgr = BeEnc_get_be_manager(be_enc);
  be_ptr result = Be_Truth(be_mgr);
  const int autsize = lsLength(partmap->aut->state_list);
  int stateid;
  
  /* <<s_{q_0}>>_i <=> |[s_{q_0}]|_i */
  result=Be_And(be_mgr,
		result, 
		Be_Iff(be_mgr, 
		       AutomataStateVarTime(be_enc,waba_vars,partmap,
					    time,1,0,k),
		       AutomataStateVarTime(be_enc,waba_vars,partmap,
					    time,0,0,k)));
  /* For each vertex of the automaton except the first  */
  for(stateid = 1; stateid < autsize; stateid++)
    {
      /* j >= 1: <<s_{q_j}>>_i <=> <<s_{q_{j-1}}>>_i \vee |[s_{q_j}]|_i */
      result=Be_And(be_mgr,
		    result,
		    Be_Iff(be_mgr,
			   AutomataStateVarTime(be_enc,waba_vars,partmap,
						time,1,stateid,k),
			   Be_Or(be_mgr, 
				 AutomataStateVarTime(be_enc,waba_vars,partmap,
						      time,1,stateid-1,k),
				 AutomataStateVarTime(be_enc,waba_vars,partmap,
						      time,0,stateid,k))));


      /* j >= 1: <<s_{q_{j-1}}>>_i  => \neg |[s_{q_j}]|_i*/
      result=Be_And(be_mgr, 
		    result,
		    Be_Implies(be_mgr, 
			       AutomataStateVarTime(be_enc,waba_vars,partmap,
						    time,1,stateid-1,k),
			       Be_Not(be_mgr, 
				      AutomataStateVarTime(be_enc,waba_vars,
							   partmap,time,0,
							   stateid,k))));
    }
  /*<<s_{q_n}>>_i <=> \top*/
  result = Be_And(be_mgr, 
		  result,
		  Be_Iff(be_mgr, 
			 AutomataStateVarTime(be_enc,waba_vars,partmap,
					      time,1,autsize-1,k),
			 Be_Truth(be_mgr)));
  return result;
}
#endif




/**Function********************************************************************

  Synopsis           [Given a PSL property generates and solve the problems
  for all Ki (k_min<=i<=k_max). If bIncreaseK is 0 then k_min==k_max==k and
  only one problem is generated. If bIncreaseK is 1 then k_min == 0 and
  k_max == k.
  Each problem Ki takes into account of all possible loops from k_min to Ki
  if loopback is '*' (BMC_ALL_LOOPS). <BR>
  Returns 1 if solver could not be created, 0 if everything went smooth]

  Description [Returns 1 if solver could not be created, 0 if
  everything went smooth]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int Bmc_PslSolve(Prop_ptr pslprop,
		 const int k, const int relative_loop,
		 const boolean must_inc_length,
		 const boolean must_solve, 
		 const Bmc_DumpType dump_type,
		 const char* dump_fname_template)
{
  Prop_ptr coi_pslprop; /*Virtual property for COI*/
  BeFsm_ptr be_fsm; /* The corresponding be fsm */
  Be_Manager_ptr be_mgr;
  BoolEnc_ptr bool_enc;
  BeEnc_ptr be_enc;
  WABA_state_vars_t *waba_vars;
  struct BmcPartitionMap *partmap;
  int previous_k, current_k;

#ifdef MS
  _MetaSolver *zolver = 0;
#else
  /* incremental sat solver instance */
  Bmc_IncrSat_ptr solver = 0;
#endif

  
 /* Symbol table and layer stuff needed to allocate new vars */
  SymbTable_ptr symb_table = (SymbTable_ptr)NULL;
  SymbLayer_ptr ltl_layer = (SymbLayer_ptr)NULL;

  boolean found_solution;
  /*Map ap names to pslNode_ptr*/
  str_hash_ptr ap_table =new_str_assoc(); 
  PslNode_ptr pslformula=Prop_get_expr(pslprop);
  Sugar_Formula_t *sugar_prop=Bmc_pslToSugar(pslformula, ap_table);
  Sugar_Formula_t *negated_sugar_prop;
  Sugar_Automaton_t *waba;

  /* PslNode_print(nusmv_stderr,pslformula); */
  /* fprintf(nusmv_stderr, "\nSugar formula: \n"); */

#ifdef BENCHMARKING
  times(&time_bmc_start);
#endif

  /* checks that a property was selected: */
  nusmv_assert(pslprop != PROP(NULL));

  found_solution = false;

  {
    lsList atomicprops=lsCreate();
    lsGen  iterator;
    Sugar_Formula_t *sugarAP;
    PslNode_ptr virtual_pslprop=psl_node_make_true();

    findAtomicProps(sugar_prop,atomicprops);
 
    /*Create a virtual psl property for semantic checks and BMC COI reduction*/
    /*Does not recognise duplicate atomic propositions e.g. p \vee !p !*/
    lsForEachItem(atomicprops, iterator, sugarAP) {
      virtual_pslprop=psl_new_node(TKAMPERSAND, virtual_pslprop, 
				   (PslNode_ptr) find_str_assoc(ap_table, (char*) sugarAP->left));
    }
    
    coi_pslprop=Prop_create_partial(virtual_pslprop, Prop_Psl);

    lsDestroy(atomicprops, NULL);
  }
  
  /*Semantic check*/
  if (!TypeCheckingPkg_check_property(
				      TypeCheckingPkg_get_global_type_checker(), coi_pslprop)) {
    return -1; /* type violation */
  }
  
  /*Intialise the optimisations options structure*/
  init_opt_options();

  /*Apply COI*/
  if (opt_cone_of_influence(options) == true) {
    Prop_apply_coi_for_bmc(coi_pslprop, global_fsm_builder);
  }

  be_fsm = Prop_get_be_fsm(coi_pslprop);
  if (be_fsm == (BeFsm_ptr) NULL) {
    Prop_set_fsm_to_master(coi_pslprop);
    be_fsm = Prop_get_be_fsm(coi_pslprop);
    nusmv_assert(be_fsm != (BeFsm_ptr) NULL);
  }

  /* Get be encoding */
  be_enc = BeFsm_get_be_encoding(be_fsm);
  assert(be_enc);

  /* Get Boolean encoding */
  bool_enc = BoolEncClient_get_bool_enc(BOOL_ENC_CLIENT(be_enc));
  assert(bool_enc);
  
  /* Find manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);


  /*
   * Incremental SAT solver construction
   */
#ifdef MS
  zolver = _MS_create(be_enc, opt_use_standard_solver == 0);
  if(zolver == 0)
    /* Something went wrong */
    return 1;
#else
  solver = Bmc_IncrSat_init(get_sat_solver(options));
  if(solver == 0)
    {
      fprintf(nusmv_stderr, "Incremental sat solver '%s' is not available.\n", 
	      get_sat_solver(options));
      return 1;
    }
#endif

  /*
   * Initialize the state variable structure
   */
  waba_vars = _WABA_state_vars_create();
  assert(waba_vars);

  /*
   * Get the state variables occurring in the transition relation and
   * add them to HJL_state_vars->transition_state_vars
   */
#if 0  
  {
    lsGen iterator; 
    be_ptr be_var = 0;
    /*Do we need to have the same check as Tommi here?*/
    lsList tr_be_vars = Bmc_Utils_get_vars_list_for_uniqueness(be_enc,
							       coi_pslprop);
    lsForEachItem(tr_be_vars, iterator, be_var)
      {
	node_ptr node = BeEnc_var_to_name(be_enc, be_var);
	lsNewEnd(waba_vars->transition_state_vars,
		 (lsGeneric)node, LS_NH);
      }
    lsDestroy(tr_be_vars, 0); tr_be_vars = 0;
  }
#endif 
  {
    int index = BeEnc_get_first_untimed_var_index(be_enc,
						  BE_VAR_TYPE_CURR |
						  BE_VAR_TYPE_INPUT);
    while(BeEnc_is_var_index_valid(be_enc, index))
      {
	node_ptr var = BeEnc_index_to_name(be_enc, index);
	assert(var);
	lsNewEnd(waba_vars->transition_state_vars, (lsGeneric)var, LS_NH);
	index = BeEnc_get_next_var_index(be_enc, index,
					 BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
      }
  }


  /* Get the symbol table */
  symb_table = Compile_get_global_symb_table();

  /* Add a new layer for translation variables */
  ltl_layer = SymbTable_create_layer(symb_table, "LTL translation vars",
				     SYMB_LAYER_POS_TOP);

  /* Allocate the l_i variable */
  waba_vars->l_var = _add_new_state_variable(ltl_layer, "#WABA_l");


#ifdef _OLD_INLOOP
  /* Allocate the InLoop variable */
  waba_vars->InLoop_var =
    _add_new_state_variable(ltl_layer, "#WABA_InLoop");
#endif

  {
    Sugar_Automaton_t *buchi;
    negated_sugar_prop=Sugar_FormulaCreate(Sugar_FL_Not_c, sugar_prop, NULL);
    /*Generate weak alternating automaton*/
    Sugar_FormulaPrint(nusmv_stderr,negated_sugar_prop);
    if(0) {
      waba=bmcTestAut(3);
    } 
    else {
      waba=Sugar_AutomatonGenerate(negated_sugar_prop);
    }
    Sugar_AutomatonPrint(waba,5);
    if(optimisations.buchi) {
      buchi=Sugar_AutomatonExpand(waba);
      Sugar_AutomatonPrint(buchi,5);
      Sugar_AutomatonFree(waba);
      waba=buchi;
    }
    partmap=Bmc_PartitionMapCreate(waba);
    Bmc_PartitionMapInit(partmap);
    fprintf(nusmv_stderr, "Number of partitions: %d\n", partmap->numPartitions);
    int i;
    for(i=0;i<lsLength(partmap->aut->state_list); i++) {
      fprintf(stderr, "Aut Id: %d, partition: %d\n", Bmc_PartitionMap_IdToAut(partmap,i),
	      Bmc_PartitionMapGet(partmap,i));

    }

    /* */
    _init_model_fairness(be_enc, ltl_layer, waba_vars, coi_pslprop);

    /*
     * Adds a variable for each waba state and each unrolling
     */
    WABA_init_state_vars(ltl_layer,waba_vars,partmap);
    /*Add fairness vars*/
    addFairnessVars(ltl_layer,be_enc,waba_vars,partmap,coi_pslprop);
  }
 
  /*
   * After introducing all new variables, commit ltl_layer
   */
  BaseEnc_commit_layer(BASE_ENC(bool_enc), SymbLayer_get_name(ltl_layer));
  BaseEnc_commit_layer(BASE_ENC(be_enc), SymbLayer_get_name(ltl_layer));
  /*
  BaseEnc_commit_layer(BASE_ENC(Enc_get_bdd_encoding()),
		       SymbLayer_get_name(ltl_layer));
  */
  if (optimisations.buchi) {   
    /* Initialize one state condition */
    waba_vars->buchi_one_state_condition =
      _make_buchi_one_state_cond(be_enc, waba_vars, partmap);

    lsGen iter;
    var_be_pair_t *pair;
    /*For all fairness sets*/
    lsForEachItem(waba_vars->automata_fairness, iter, pair) {
      Sugar_Vertex_t *node;
      int index;
      arrayForEachItem(Sugar_Vertex_t*,partmap->aut->accept_states,index,node)
	{
	  const int stateid = Bmc_PartitionMap_AutToId(partmap, node->id);
	  /*fprintf(stderr, "Stateid: %d\n", stateid);*/
	  pair->exp=Be_Or(be_mgr,
			  pair->exp,
			  AutomataStateVarUntimed(be_enc,waba_vars,partmap,
						  0,stateid));
	}
    }
  }
  /*
   * Insert the initial condition into the sat solver permanently
   */
  {
    be_ptr be_init = Bmc_TJ_GetInitI(be_fsm, _real_k(0));
#ifdef MS
    _MS_force_true(zolver, be_init);
#else
    Bmc_IncrSat_insert(solver, be_init);
    Bmc_IncrSat_force_true(solver, be_init);
#endif

    if(opt_verbose_level_ge(options, 6)) {
      fprintf(nusmv_stderr, "Forced initial condition: ");
      Be_DumpSexpr(be_mgr, be_init, nusmv_stderr);
      fprintf(nusmv_stderr, "\n");
    }
  }
  
  /*
   * Make base stuff
   */
  {
    be_ptr new_constraints = BaseConditions(be_enc,waba_vars,partmap,k);
    /* Force constraints to be true in the fixed frame */
#ifdef MS
    _MS_force_true(zolver, new_constraints);
#else
    Bmc_IncrSat_insert(solver, new_constraints);
    Bmc_IncrSat_force_true(solver, new_constraints);
#endif
 
  }
  
  /* Start problems generation: */
  Prop_set_type(pslprop,Prop_Psl);
 
  previous_k = -1000; /* Start from deep prehistory (at least <= -2) */
  for(current_k = 0;
      current_k <= k && !found_solution;
      current_k = current_k + 1)
    {
      int i = 0;
      Bmc_IncrSat_result satResult;
      
      if(opt_verbose_level_gt(options, 0)) {
	fprintf(nusmv_stderr,
		"\nGenerating problem with bound %d, all possible loopbacks...\n",
                current_k);
      }


      
      /********************************************************************
       *
       * The k-invariant part of the translation
       *
       ********************************************************************/

      /*
       * Unroll the model transition relation up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      for(i = max(previous_k, 0); i < current_k; i++)
	{
	  be_ptr be_TR = Bmc_Model_GetUnrolling(be_fsm, 
						_real_k(i),
						_real_k(i+1));
#ifdef MS
	  _MS_force_true(zolver, be_TR);
#else
	  Bmc_IncrSat_insert(solver, (Rbc_t*)be_TR);
	  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_TR);
#endif
	  
	  if(opt_verbose_level_gt(options, 1)) {
	    fprintf(nusmv_stderr, "Forced T(%d,%d)", i, i+1);
	    if(opt_verbose_level_ge(options, 7)) {
	      fprintf(nusmv_stderr, ": ");
	      Be_DumpSexpr(be_mgr, be_TR, nusmv_stderr);
	    }
	    fprintf(nusmv_stderr, "\n");
	  }
	}
      

#ifndef _OLD_INLOOP      
      /*
       * Define InLoop_i := (InLoop_{i-1} || l_i)
       * and build constraint InLoop_{i-1} => !l_i up to current_k
       * Force l_0 <=> FALSE
       * Assumes that they have already been build up to previous_k
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  be_ptr be_l_i = 0;
	  be_ptr be_constraint = 0;
	  be_ptr be_InLoop_i = 0;
	  
	  be_l_i = BeEnc_name_to_timed(be_enc, waba_vars->l_var, _real_k(i));
	  assert(be_l_i);

	  if(i == 0)
	    {
	      /* InLoop_0 := false */
	      be_InLoop_i = Be_Falsity(be_mgr);
	      /* Force constraint: l_0 <=> false */
	      be_constraint = Be_Iff(be_mgr, be_l_i, Be_Falsity(be_mgr));
	      assert(be_constraint);
	    }
	  else
	    {
	      /* InLoop_i := InLoop_{i-1} | l_i */
	      be_ptr be_InLoop_iM1 = 0;
	      be_InLoop_iM1 = array_fetch(be_ptr, waba_vars->InLoop_bes, i-1);
	      assert(be_InLoop_iM1);
	      be_InLoop_i = Be_Or(be_mgr, be_InLoop_iM1, be_l_i);
	      assert(be_InLoop_i);
	      /* Force constraint: InLoop_{i-1} => !l_i */
	      be_constraint = Be_Implies(be_mgr,
					 be_InLoop_iM1,
					 Be_Not(be_mgr, be_l_i));
	      assert(be_constraint);
	    }
	  /* Store InLoop_i */
	  assert(be_InLoop_i);
	  assert(array_n(waba_vars->InLoop_bes) == i);
	  array_insert(be_ptr, waba_vars->InLoop_bes, i, be_InLoop_i);
#ifndef MS
	  /* also give (not force) it to the solver if using TJ interface */
	  Bmc_IncrSat_insert(solver, be_InLoop_i);
#else
#error "You shouldn't be using MS at the moment (it is unoptimal)"
#endif
	  /* Force constraint */
	  assert(be_constraint);
#ifdef MS
	  _MS_force_true(zolver, be_constraint);
#else	  
	  Bmc_IncrSat_insert(solver, be_constraint);
	  Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced InLoop_%d stuff\n", i);
	}
#endif /* #ifndef _OLD_INLOOP */


      /*
       * Unroll the (l_i => (s_{i-1} = s_E)) constraint up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      /*
       * Define InLoop_i := (InLoop_{i-1} || l_i)
       * and build constraint InLoop_{i-1} => !l_i up to current_k
       * Assumes that they have already been build up to previous_k
       */
      {
	be_ptr be_constraint=LoopConstraints(be_enc,waba_vars,previous_k,current_k);
	  
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else
	Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
      }


      /* Automata loop constraints
       * Unroll the (l_i => (|s_q|_L <=> |s_q|_i)) constraint from 
       * previous_k+1 up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */

      {
	int time;
	for(time = max(previous_k+1,1); time <= current_k; time++)
	  {
	    int stateid;
	    for(stateid=0;stateid<lsLength(partmap->aut->state_list);stateid++)
	      {
		int depth;
		for (depth=0;depth<=unroll_depth(partmap,stateid);depth++) {
		  /*l_i => (|s_q|_L^d <=> |s_q|_i^d )*/
		  be_ptr be_constraint = 
		    Be_Implies(be_mgr,
			       LoopSelectVar(be_enc,waba_vars,time,current_k),
			       Be_Iff(be_mgr,
				      AutomataStateVarL(be_enc,waba_vars,
							partmap,depth,stateid,
							current_k),
				      AutomataStateVarTime(be_enc,waba_vars,
							   partmap,time,depth,
							   stateid,current_k))
			       );
#ifdef MS
		  _MS_force_true(zolver, be_constraint);
#else
		  Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
		  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
		}
	      }
	  }
      }


      
      /*
       * Unroll automata transition relation to current_k. Assume it has
       * been unrolled to previous_k
       */
      for(i = max(previous_k+1,0); i <= current_k; i++)
	{
	  be_ptr be_aut_tr = AutomataInstantiate(be_enc,waba_vars,partmap,ap_table,i,current_k);
#ifdef MS
	  _MS_force_true(zolver, be_aut_tr);
#else	  
	  Bmc_IncrSat_insert(solver, be_aut_tr);
	  Bmc_IncrSat_force_true(solver, be_aut_tr);
#endif
	  
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "Conditions for the automatons transition relation: ");
	    Be_DumpSexpr(be_mgr, be_aut_tr, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}
      
      /*
       * Model fairness constraints
       */
      for(i = max(previous_k+1, 0); i <= current_k; i++)
	{
	  int fairness_num;
	  for(fairness_num = 0;
	      fairness_num < waba_vars->model_fairness_nof;
	      fairness_num++)
	    {
	      be_ptr be_expr = 0;
	      array_t *timed_bes = 0;
	      
	      be_expr = array_fetch(be_ptr,
				    waba_vars->model_fairness_untimed_bes,
				    fairness_num);
	      assert(be_expr);
	      
	      timed_bes = array_fetch(array_t*,
				      waba_vars->model_fairness_timed_bes,
				      fairness_num);
	      assert(timed_bes);
	      assert(array_n(timed_bes) == i);

	      if(i == 0)
		{
		  /* FAIR_0 = false */
		  array_insert(be_ptr, timed_bes, i, Be_Falsity(be_mgr));
		}
	      else
		{
		  /* FAIR_i := FAIR_{i-1} | (InLoop_i & be_expr_i)*/
		  be_ptr be_expr_i = 0;
		  be_ptr be_fair_i = 0;
		  be_ptr be_fair_iM1 = 0;
		  be_ptr be_InLoop_i = 0;

		  be_expr_i = BeEnc_untimed_expr_to_timed(be_enc,
							  be_expr,
							  _real_k(i));
		  assert(be_expr_i);

		  be_fair_iM1 = array_fetch(be_ptr, timed_bes, i-1);
		  assert(be_fair_iM1);
		  
		  be_InLoop_i = InLoop(be_enc, waba_vars, i, current_k);
		  assert(be_InLoop_i);
		  
		  be_fair_i = Be_Or(be_mgr,
				    be_fair_iM1,
				    Be_And(be_mgr,
					   be_InLoop_i,
					   be_expr_i));
		  assert(be_fair_i);

		  array_insert(be_ptr, timed_bes, i, be_fair_i);
#ifndef MS
		  /* Give (not force) it to the solver if using TJ interface */
		  Bmc_IncrSat_insert(solver, be_fair_i);
#else
#error "You shouldn't use MS at the moment due to the obvious reason (see above)"
#endif
		}
	    }
	}

      /*Automata fairness constraints; exists only for the Buchi case*/
      if(lsLength(waba_vars->automata_fairness) != 0)
	{
	  nusmv_assert(optimisations.buchi);
	  for(i = max(previous_k+1, 1); i <= current_k; i++)
	    {
	      lsGen iter;
	      var_be_pair_t *pair;
	      lsForEachItem(waba_vars->automata_fairness,iter,pair)
		{
		  /* Fair_i <=>
		   * Fair_{i-1} \vee (InLoop_i \wedge fairnessexpr_i) */
		  be_ptr be_constraint = 
		    Be_Iff(BeEnc_get_be_manager(be_enc),
			   BeEnc_name_to_timed(be_enc, pair->var, _real_k(i)),
			   Be_Or(be_mgr,
				 BeEnc_name_to_timed(be_enc,
						     pair->var,
						     _real_k(i-1)),
				 Be_And(be_mgr,
					InLoop(be_enc,waba_vars,i,current_k),
					BeEnc_untimed_expr_to_timed(be_enc,
								    pair->exp,
								    _real_k(i)
								    )
					)
				 )
			   );
#ifdef MS
		  _MS_force_true(zolver, be_constraint);
#else
		  Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
		  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
		}
	    }
	}

      
      /*For the Buchi case, add the 'exactly one state per run' constraint*/
      if (optimisations.buchi)
	{
	  for(i = max(previous_k+1,0); i <= current_k; i++)
	    {
	      assert(waba_vars->buchi_one_state_condition);
	      be_ptr one_state =
		BeEnc_untimed_expr_to_timed(be_enc,waba_vars->buchi_one_state_condition,_real_k(i));
	      /*
		be_ptr one_state = buchi_one_state_cond(be_enc, waba_vars,
		partmap, i, current_k);
	      */
#ifdef MS
	      _MS_force_true(zolver, one_state);
#else
	      Bmc_IncrSat_insert(solver, (Rbc_t*)one_state);
	      Bmc_IncrSat_force_true(solver, (Rbc_t*)one_state);
#endif
	    }
	}
      
      /*Optimising constraints:
       * add monotonicity constraints if the optimisation is enabled.
       */
      if(optimisations.mono_constraints && !optimisations.buchi) {
	int time;
	for(time=(previous_k<0)?0:previous_k+2; time<=current_k+1; time++) {
	  int stateid;
	  for(stateid=0;stateid<lsLength(partmap->aut->state_list);stateid++) {
	    int depth;
	    for (depth=1;depth<=unroll_depth(partmap,stateid);depth++) {
	      /*When unrolling final components we can add
		the monotoncity constraint*/
	      if(!optimisations.no_acc_unrolling &&
		 checkMemberShip(partmap,partmap->aut->accept_states, stateid))
		{
		  /*(|s_q|_i^{d-1} => |s_q|_i^d )*/
		  const be_ptr be_constraint =
		    Be_Implies(be_mgr,
			       AutomataStateVarTime(be_enc,waba_vars,partmap,
						    time,depth-1,stateid,
						    current_k),
			       AutomataStateVarTime(be_enc,waba_vars,partmap,
						    time,depth,stateid,
						    current_k));
#ifdef MS
		  _MS_force_true(zolver, be_constraint);
#else
		  Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
		  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
		}
	      else
		{
		  /*(|s_q|_i^d => |s_q|_i^{d-1} )*/
		  const be_ptr be_constraint =
		    Be_Implies(be_mgr,
			       AutomataStateVarTime(be_enc,waba_vars,partmap,
						    time,depth,stateid,
						    current_k),
			       AutomataStateVarTime(be_enc,waba_vars,partmap,
						    time,depth-1,stateid,
						    current_k));
#ifdef MS
		  _MS_force_true(zolver, be_constraint);
#else
		  Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
		  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
		}
	    }
	  }
	}
      }


      /********************************************************************
       *
       * The k-dependent part of the translation
       *
       ********************************************************************/
      
      
      if(opt_verbose_level_gt(options, 1))
	fprintf(nusmv_stderr, "Adding k-dependent stuff\n");
      
      
      /*
       * Go to the volatile frame 1
       */
#ifdef MS
      _MS_goto_volatile_group(zolver);
#else
      Bmc_IncrSat_new_frame(solver);
#endif
      
      
      /*
       * Add k-dependent constraints
       */
      {
	be_ptr new_constraints=kDependentConstraints(be_enc,waba_vars,partmap,current_k);
#ifdef MS
	_MS_force_true(zolver, new_constraints);
#else	  
	Bmc_IncrSat_insert(solver, new_constraints);
	Bmc_IncrSat_force_true(solver, new_constraints);
#endif
      }
#ifndef _OLD_INLOOP
      {
	/* Force loop */
	be_ptr be_constraint;
      	be_ptr be_InLoop_k = InLoop(be_enc, waba_vars, current_k, current_k);
	assert(be_InLoop_k);
	be_constraint = Be_Iff(be_mgr, be_InLoop_k, Be_Truth(be_mgr));
#ifdef MS
	_MS_force_true(zolver, be_constraint);
#else	  
	Bmc_IncrSat_insert(solver, be_constraint);
	Bmc_IncrSat_force_true(solver, be_constraint);
#endif	
      }
#endif
      {
	/* Force possible model fairness by requiring FAIR_k <=> true
	 * for each model fairness set */
	int fairness_num;
	for(fairness_num = 0;
	    fairness_num < waba_vars->model_fairness_nof;
	    fairness_num++)
	  {
	    array_t *timed_bes = 0;
	    be_ptr be_fair_k = 0;
	    be_ptr be_constraint = 0; 
	    
	    timed_bes = array_fetch(array_t*,
				    waba_vars->model_fairness_timed_bes,
				    fairness_num);
	    assert(timed_bes);
	    assert(array_n(timed_bes) == current_k+1);

	    be_fair_k = array_fetch(be_ptr, timed_bes, current_k);
	    assert(be_fair_k);
	    be_constraint = Be_Iff(be_mgr, be_fair_k, Be_Truth(be_mgr));
	    assert(be_constraint);
#ifdef MS
	    _MS_force_true(zolver, be_constraint);
#else	  
	    Bmc_IncrSat_insert(solver, be_constraint);
	    Bmc_IncrSat_force_true(solver, be_constraint);
#endif
	  }
	if(waba_vars->model_fairness_nof > 0)
	  fprintf(nusmv_stderr, "WARNING: finite executions are ignored "
		  "due to requiring model fairness\n");
      }

      
#ifdef BENCHMARKING
      fprintf(nusmv_stderr, ":START:benchmarking Solving\n");
      times(&time_solver_start);
#endif  
      
      
      /*
       * Actually solve the problem
       */
#ifdef MS
      satResult = _MS_solve(zolver);
#else
      satResult = Bmc_IncrSat_solve(solver);
#endif
      
      
#ifdef BENCHMARKING
      {
	times(&time_temp);
	fprintf(nusmv_stderr, ":UTIME = %.4f secs.\n",
		time_diff(&time_solver_start, &time_temp));
	  fprintf(nusmv_stderr, ":STOP:benchmarking Solving\n");
	  fprintf(nusmv_stderr, "TJ: counterexample, k=%d, solver time=%.4f, "
		  "cumulative time=%.4f.\n",
		  current_k,
		  time_diff(&time_solver_start, &time_temp),
		  time_diff(&time_bmc_start, &time_temp));
      }
#endif
      
      
      /* Processes the result: */
      switch (satResult)
	{
	case BMC_INCRSAT_UNSATISFIABLE_PROBLEM:
	  {
	    fprintf(nusmv_stderr,
		    "-- no counterexample found with bound %d",
		    current_k);
	    if (opt_verbose_level_gt(options, 2)) {
	      fprintf(nusmv_stderr, " for "); 		  
	      print_spec(nusmv_stderr, pslprop);
	    }
	    fprintf(nusmv_stderr, "\n");
	    
	    break;
	  }
	  case BMC_INCRSAT_SATISFIABLE_PROBLEM:
	    fprintf(nusmv_stderr, "-- ");
	    print_spec(nusmv_stderr, pslprop);
	    fprintf(nusmv_stderr, "  is false\n");
	    Prop_set_status(pslprop, Prop_False);
	    
	    found_solution = true;
	    
	    if (opt_counter_examples(options)) {
#ifdef MS
	      if(opt_use_standard_solver)
		{
		  /* Counter-examples with standard incremental SAT solver
		     not supported yet */
		}
	      else
		{
		  Trace_ptr trace = 
		    _generate_and_print_cntexample(be_enc, ltl_layer,
						   zolver->TJ_solver,
						   Enc_get_bdd_encoding(), 
						   current_k, 
						   "BMC Counterexample",
						   waba_vars->l_var);
		  Prop_set_trace(pslprop, Trace_get_id(trace));
		}
#else
	      Trace_ptr trace = 
		_generate_and_print_cntexample(be_enc, ltl_layer,
					       solver,
					       Enc_get_bdd_encoding(), 
					       current_k, 
					       "BMC Counterexample",
					       waba_vars->l_var);
	      Prop_set_trace(pslprop, Trace_get_id(trace));
#endif
	    }
	    
	    break;
	    
	case BMC_INCRSAT_INTERNAL_ERROR:
	  internal_error("Sorry, solver answered with a fatal Internal " 
			 "Failure during problem solving.\n");
	  break;
	  
	case BMC_INCRSAT_TIMEOUT:
	case BMC_INCRSAT_MEMOUT:
	    internal_error("Sorry, solver ran out of resources and aborted "
			   "the execution.\n");
	    break;
	    
	default:
	  internal_error("Bmc_GenSolveLtlTJ: Unexpected value in satResult");
	  
	} /* switch */
      
      
      if(opt_verbose_level_gt(options, 1))
	fprintf(nusmv_stderr, "Deleting k-dependent stuff\n");
      
      
      /*
       * Release the volatile frame 1
       */
#ifdef MS
      _MS_goto_permanent_group(zolver);
#else
      Bmc_IncrSat_pop_frame(solver);
#endif
      
      /*
       * Remember the current_k
       */
      previous_k = current_k;
    
    }
  /*
   * Remove extra encoding layers
   */
  /*
  BaseEnc_remove_layer(BASE_ENC(Enc_get_bdd_encoding()),
		       SymbLayer_get_name(ltl_layer));
  */
  BaseEnc_remove_layer(BASE_ENC(be_enc), SymbLayer_get_name(ltl_layer));
  BaseEnc_remove_layer(BASE_ENC(bool_enc), SymbLayer_get_name(ltl_layer));
  SymbTable_remove_layer(symb_table, ltl_layer);
  ltl_layer = 0;
  
  /*
   * Destroy the sat solver instance
   */
#ifdef MS
  _MS_destroy(zolver); zolver = 0;
#else
  Bmc_IncrSat_release(solver); solver = 0;
#endif

  /*Additional clean up*/
  WABA_delete_state_vars(waba_vars);
  clear_str_assoc(ap_table);
  Sugar_FormulaFree(negated_sugar_prop);
  Sugar_AutomatonFree(waba);
  Bmc_PartitionMapDelete(partmap);
  return 0;
}


static 
Sugar_Automaton_t *bmcTestAut(int testcase)
{
  Sugar_Automaton_t *aut;

  if(testcase==0) {
    int index1, index2;
    Sugar_Vertex_t *node;
    array_t * item;
    /*Automaton accepting F p_0*/
    aut=SugarAutomatonIdentifier("p_0", 0, 1);
    arrayForEachItem(array_t*, aut->initial_states, index1, item) {
      arrayForEachItem(Sugar_Vertex_t*, item, index2, node) {
	SugarAutomatonNewEdgeLoop(aut, node);
      }
    }
  }
  else if (testcase==1) {
    /*Automaton accepting G p_0*/
    array_t *dest_array = array_alloc(Sugar_Vertex_t*, 1);
    array_t *label_array=0;
    lsGen vertex_gen, edge_gen;
    Sugar_Vertex_t *vertex;
    Sugar_Vertex_t *destvertex;
    Sugar_Edge_t *edge;
    aut=SugarAutomatonIdentifier("p_0", 0, 0);
  

    lsForEachItem(aut->state_list, vertex_gen, vertex) {
       int index;
       Sugar_Vertex_t *node;     
       arrayForEachItem(Sugar_Vertex_t*,aut->accept_states, index, node) {
	 if (node->id==vertex->id) { 
	   destvertex=vertex;
	   array_insert_last(Sugar_Vertex_t*,dest_array,vertex);	
	 }
	 else {
	   lsForEachItem(vertex->edge_out_list, edge_gen, edge) {
	     label_array=array_dup(edge->label);
	   }
	 }
       }
    }    
    SugarAutomatonNewEdge(aut, destvertex, dest_array, label_array,1);
    array_free(label_array);
    array_free(dest_array);
  }
  else if (testcase==2) {
    /*Automaton accepting F (p_0 \wedge G p_1)*/
    array_t *term_array = array_alloc(Sugar_Vertex_t*, 1);
    array_t *dest_array;
    array_t *prop_array;
    Sugar_Vertex_t* node1;
    Sugar_Vertex_t* node2;
    Sugar_Vertex_t* node3;
    Sugar_Label_t* labl;

    aut=Sugar_AutomatonCreate();
    node1=SugarAutomatonNewNode(aut);
    node2=SugarAutomatonNewNode(aut);
    node3=SugarAutomatonNewNode(aut);

    /* make node1 the initial state */ 
    array_insert(Sugar_Vertex_t*, term_array, 0, node1);
    array_insert_last(array_t*, aut->initial_states, term_array);
    /*Add true loop to node1 and node2*/
    SugarAutomatonNewEdgeLoop(aut, node1);
    SugarAutomatonNewEdgeLoop(aut, node2);
    /*Add a p_0&p1 edge from node1 to node2&node3*/
    dest_array = array_alloc(Sugar_Vertex_t*, 2);
    array_insert(Sugar_Vertex_t*, dest_array, 0, node2);
    array_insert_last(Sugar_Vertex_t*, dest_array, node3);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_0"); labl->isNegate = 0;
    prop_array = array_alloc(Sugar_Label_t*, 2);
    array_insert(Sugar_Label_t*, prop_array, 0, labl);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_1"); labl->isNegate = 0;
    array_insert_last(Sugar_Label_t*, prop_array, labl);
    SugarAutomatonNewEdge(aut, node1, dest_array, prop_array, 0);
    array_free(dest_array);
    array_free(prop_array);
    /*Add a p_1 edge from node3 to node3*/
    dest_array = array_alloc(Sugar_Vertex_t*, 1);
    array_insert(Sugar_Vertex_t*, dest_array, 0, node3);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_1"); labl->isNegate = 0;
    prop_array = array_alloc(Sugar_Label_t*, 1);
    array_insert(Sugar_Label_t*, prop_array, 0, labl);
    SugarAutomatonNewEdge(aut, node3, dest_array, prop_array, 0);
    array_free(dest_array);
    array_free(prop_array);
    /* make node2 and node3 accepting */
    array_insert_last(Sugar_Vertex_t*, aut->accept_states, node2);
    array_insert_last(Sugar_Vertex_t*, aut->accept_states, node3);
  }
    else if (testcase==3) {
    /*Automaton accepting F [G p0 \wedge G p_1]*/
    array_t *term_array = array_alloc(Sugar_Vertex_t*, 1);
    array_t *dest_array;
    array_t *prop_array;
    Sugar_Vertex_t* node1;
    Sugar_Vertex_t* node2;
    Sugar_Vertex_t* node3;
    Sugar_Label_t* labl;

    aut=Sugar_AutomatonCreate();
    node1=SugarAutomatonNewNode(aut);
    node2=SugarAutomatonNewNode(aut);
    node3=SugarAutomatonNewNode(aut);

    /* make node1 the initial state */ 
    array_insert(Sugar_Vertex_t*, term_array, 0, node1);
    array_insert_last(array_t*, aut->initial_states, term_array);
    /*Add true loop to node1*/
    SugarAutomatonNewEdgeLoop(aut, node1);
    /*Add a p_0&p1 edge from node1 to node2&node3*/
    dest_array = array_alloc(Sugar_Vertex_t*, 2);
    array_insert(Sugar_Vertex_t*, dest_array, 0, node2);
    array_insert_last(Sugar_Vertex_t*, dest_array, node3);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_0"); labl->isNegate = 0;
    prop_array = array_alloc(Sugar_Label_t*, 2);
    array_insert(Sugar_Label_t*, prop_array, 0, labl);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_1"); labl->isNegate = 0;
    array_insert_last(Sugar_Label_t*, prop_array, labl);
    SugarAutomatonNewEdge(aut, node1, dest_array, prop_array, 0);
    array_free(dest_array);
    array_free(prop_array);
    /*Add a p_0 edge from node2 to node2*/
    dest_array = array_alloc(Sugar_Vertex_t*, 1);
    array_insert(Sugar_Vertex_t*, dest_array, 0, node2);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_0"); labl->isNegate = 0;
    prop_array = array_alloc(Sugar_Label_t*, 1);
    array_insert(Sugar_Label_t*, prop_array, 0, labl);
    SugarAutomatonNewEdge(aut, node2, dest_array, prop_array, 0);
    array_free(dest_array);
    array_free(prop_array);
    /*Add a p_1 edge from node3 to node3*/
    dest_array = array_alloc(Sugar_Vertex_t*, 1);
    array_insert(Sugar_Vertex_t*, dest_array, 0, node3);
    labl = ALLOC(Sugar_Label_t, 1);
    labl->name = util_strsav("p_1"); labl->isNegate = 0;
    prop_array = array_alloc(Sugar_Label_t*, 1);
    array_insert(Sugar_Label_t*, prop_array, 0, labl);
    SugarAutomatonNewEdge(aut, node3, dest_array, prop_array, 0);
    array_free(dest_array);
    array_free(prop_array);
    /* make node2 and node3 accepting */
    array_insert_last(Sugar_Vertex_t*, aut->accept_states, node2);
    array_insert_last(Sugar_Vertex_t*, aut->accept_states, node3);
  }

  return aut;
}


/**Function********************************************************************

  Synopsis           [Find the unique atomic propositions and return a
  map of sugar atomic propositions mapped to psl atomic propositions.}

  Description []

  SideEffects        []

  SeeAlso            []

******************************************************************************/


static void findAtomicProps(Sugar_Formula_t *formula, lsList idlist) 
{
  if (formula == NIL(Sugar_Formula_t)) {
    return;
  }
  /* The formula is a leaf. */
  if (formula->type == Sugar_IDENTIFIER_c) {
    lsNewEnd (idlist, (lsGeneric) formula, LS_NH);
    return;
  }

  switch(formula->type) {
  
  case Sugar_FL_Releases_c:
  case Sugar_FL_Until_c:
  case Sugar_FL_Until__c:
  case Sugar_FL_Before_c:
  case Sugar_FL_Before__c: 
  case Sugar_FL_And_c:
  case Sugar_FL_Or_c:
  case Sugar_FL_Imply_c:
  case Sugar_FL_Equiv_c:
  case Sugar_Seq_FLProp_c:
  case Sugar_Seq_Imply_c:
  case Sugar_Seq_Imply2_c:
  case Sugar_Seq_Within_c:
  case Sugar_Seq_Within__c:
  case Sugar_Seq_Whilenot_c:
  case Sugar_Seq_Whilenot__c:
  case Sugar_RegExpr_Concat_c:
  case Sugar_RegExpr_Compose_c:
  case Sugar_RegExpr_Or_c:
  case Sugar_RegExpr_And_c:
  case Sugar_RegExpr_AndEq_c:
    findAtomicProps(formula->left, idlist);
    findAtomicProps(formula->right, idlist);
    break;
  case Sugar_FL_Not_c:
  case Sugar_FL_Next_c:
  case Sugar_FL_Always_c:
  case Sugar_FL_Never_c:
  case Sugar_FL_Eventually_c:
  case Sugar_RegExpr_Plus_c:
  case Sugar_RegExpr_Star_c:
  case Sugar_RegExpr_Equal_c:
  case Sugar_RegExpr_Goto_c:
  case Sugar_Seq_Always_c:
  case Sugar_Seq_Never_c:
  case Sugar_Seq_Eventually_c:

    findAtomicProps(formula->left, idlist);
    break;
  case Sugar_TRUE_c:
  case Sugar_FALSE_c:
    return;

  default:
    fprintf(nusmv_stderr, "Unknown op: %d\n", formula->type);
    nusmv_assert(false);
  }
}

static int checkMemberShip(struct BmcPartitionMap *partmap, array_t *accept_states, int vertexid)
{
  int index;
  Sugar_Vertex_t *node;

  arrayForEachItem(Sugar_Vertex_t*, accept_states, index, node) {
    if (Bmc_PartitionMap_AutToId(partmap,node->id)==vertexid) 
      return 1;
  } 
  return 0;
}


/*Get a loop_select(i) variable*/

static be_ptr
LoopSelectVar(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, const int time, const int k)
{
  nusmv_assert((time < k+2) && (time>=0));

  return BeEnc_name_to_timed(be_enc,waba_vars->l_var,_real_k(time));
}

/*Get the InLoop_i variable*/

#ifdef _OLD_INLOOP
static be_ptr 
InLoop(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
       const int time, const int k)
{
  nusmv_assert(time>=0 && time<= k);
  return BeEnc_name_to_timed(be_enc,waba_vars->InLoop_var,_real_k(time));
}
#else
static be_ptr 
InLoop(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
       const int time, const int k)
{
  be_ptr result = 0;
  nusmv_assert(time>=0 && time<= k);
  result = array_fetch(be_ptr, waba_vars->InLoop_bes, time);
  assert(result);
  return result;
}
#endif

/*Get |s_q|_L^d variable */

static be_ptr AutomataStateVarL(const BeEnc_ptr be_enc,
				WABA_state_vars_t *waba_vars,  
				struct BmcPartitionMap *partmap,
				const int depth,
				const int stateid, 
				const int k)
{
  nusmv_assert(depth>=0 && depth<(max_partition_size(partmap))+2);
  nusmv_assert(stateid>=0 && stateid<lsLength(partmap->aut->state_list));
  /*Offset for depth and stateid*/
  int offset= stateid*(max_partition_size(partmap)+2)+depth;
  return BeEnc_name_to_timed(be_enc,(waba_vars->waba_vars)[offset],_L_state());
}

/*Get |s_q|_i^d variable*/

static be_ptr 
AutomataStateVarTime(const BeEnc_ptr be_enc,
		     WABA_state_vars_t *waba_vars, 
		     struct BmcPartitionMap *partmap,
		     const int time,
		     const int depth,
		     const int stateid,
		     const int k)
{
  nusmv_assert(depth>=0 && depth<max_partition_size(partmap)+1);
  nusmv_assert(stateid>=0 && stateid<lsLength(partmap->aut->state_list));
  nusmv_assert(time>=0 && time<=k+1);
  /*Offset for stateid, time and depth*/
  int offset=stateid*(max_partition_size(partmap)+2) + depth;
  return BeEnc_name_to_timed(be_enc,(waba_vars->waba_vars)[offset],_real_k(time));

}

static be_ptr AutomataStateVarUntimed(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars,
				      struct BmcPartitionMap *partmap, const int depth, const int stateid)
{
  nusmv_assert(depth>=0 && depth<max_partition_size(partmap)+1);
  nusmv_assert(stateid>=0 && stateid<lsLength(partmap->aut->state_list));
  /*Offset for stateid, time and depth*/
  int offset=stateid*(max_partition_size(partmap)+2) + depth;
  /*fprintf(stderr, "Offset: %d\n", offset);*/
  return BeEnc_name_to_untimed(be_enc,(waba_vars->waba_vars)[offset]);

}

static be_ptr 
LoopConstraints(const BeEnc_ptr be_enc, WABA_state_vars_t *waba_vars, const int prevK, const int k)
{
  be_ptr result=Be_Truth(BeEnc_get_be_manager(be_enc));
  int i=0;

  /*Encode the k-invariant InLoop and at most one LoopSelectorVar constraints*/
  
  
  /*k-invariant*/
  for(i=(prevK<0) ? 1 : prevK+1; i<=k; i++) {
    /*InLoop_i <=> InLoop_{i-1}\vee l_i*/
    result=Be_And(BeEnc_get_be_manager(be_enc),
		  result,
		  Be_Iff(BeEnc_get_be_manager(be_enc),
			 InLoop(be_enc,waba_vars,i,k),
			 Be_Or(BeEnc_get_be_manager(be_enc),
			       InLoop(be_enc,waba_vars,i-1,k),
			       LoopSelectVar(be_enc,waba_vars,i,k))
			 )
		  );
    
    /*l_i => (s_{i-1} <=> s_E)*/
    result=Be_And(BeEnc_get_be_manager(be_enc),
		  result,
		  Be_Implies(BeEnc_get_be_manager(be_enc),
			     LoopSelectVar(be_enc,waba_vars,i,k),
			     _equal_vectors_formula(be_enc,waba_vars->transition_state_vars,
						    _real_k(i-1),_E_state())
			     )
		  );
    /*InLoop_{i-1} => \neg l_i*/
    result=Be_And(BeEnc_get_be_manager(be_enc),
		  result,
		  Be_Implies(BeEnc_get_be_manager(be_enc),
			     InLoop(be_enc,waba_vars,i-1,k),
			     Be_Not(BeEnc_get_be_manager(be_enc),
				    LoopSelectVar(be_enc,waba_vars,i,k))
			     )
		  );
    
  }

   return result;
}

/*Encode the transition relation for the waba at a given time, for all
  relevant depths |s_q|_i^d <-> |\delta(q)|_i^d*/

static be_ptr 
AutomataInstantiate(const BeEnc_ptr be_enc,
		    WABA_state_vars_t *waba_vars,
		    struct BmcPartitionMap *partmap, 
		    str_hash_ptr ap_table,
		    const int time,
		    const int k)
{
  const Be_Manager_ptr be_mgr = BeEnc_get_be_manager(be_enc);
  be_ptr result=Be_Truth(be_mgr); 
  Sugar_Vertex_t *sourcenode;
  lsGen state_gen;
  
  /*For each vertex of the automaton  */
  int depth=0;
  lsForEachItem(partmap->aut->state_list, state_gen, sourcenode)
    {
      for(depth = 0;
	  depth <= unroll_depth(partmap,Bmc_PartitionMap_AutToId(partmap,sourcenode->id));
	  depth++)
	{
	  lsGen edge_gen;
	  Sugar_Edge_t *edge;
	  be_ptr and_or_edge=Be_Falsity(be_mgr);      
	  /*For each out-going edge of the vertex*/
	  lsForEachItem(sourcenode->edge_out_list, edge_gen, edge) {
	    be_ptr and_edge=Be_Truth(be_mgr);
	    int index;
	    Sugar_Label_t *labl=0;
	    Sugar_Vertex_t *destnode=0;
	    lsGen destnode_gen;
	    
	    /*First generate the formula corresponding to the edge*/
	    arrayForEachItem(Sugar_Label_t*, edge->label, index, labl) {
	      if (labl->name == NIL(char)) {
		and_edge=Be_And(be_mgr,
				and_edge,
				(labl->isNegate) ? 
				Be_Falsity(be_mgr) : Be_Truth(be_mgr)
				);
	      }
	      else {
		be_ptr be_label;
		be_ptr be_label_i;
		PslNode_ptr ap=find_str_assoc(ap_table, labl->name);
		/* fprintf(nusmv_stderr, "AP: %s\n", labl->name);  */
		nusmv_assert(ap!=NULL);
		be_label=Bmc_Conv_Bexp2Be(be_enc, Bmc_Wff_MkNnf(detexpr2bexpr(Enc_get_bdd_encoding(), PslNode_convert_to_node_ptr(PslNode_pslltl2ltl(ap, PSL2SMV)))));
		be_label_i = BeEnc_untimed_expr_to_timed(be_enc,
							 be_label,
							 _real_k(time));
		and_edge=Be_And(be_mgr,
				and_edge,
				(labl->isNegate) ? Be_Not(be_mgr, be_label_i) : be_label_i 
				);
	      }
	    }      
	    /*Then generate the formula corresponding to the destination*/
	    lsForEachItem(edge->dest_vertex_list, destnode_gen, destnode) {
	      int curr_depth=depth;
	      if (optimisations.buchi || Bmc_PartitionMapGet_AutId(partmap, sourcenode->id)!=Bmc_PartitionMapGet_AutId(partmap, destnode->id)) {
		/*States in different partitions than the source state are evaluated at depth 0*/
		curr_depth=0;
	      }
	      and_edge=Be_And(be_mgr,
			      and_edge,
			      AutomataStateVarTime(be_enc,waba_vars,partmap,
						   time+1,curr_depth,
						   Bmc_PartitionMap_AutToId(partmap,destnode->id),k)
			      );
	      
	    }
	    /*Combine all outgoing edges from the state*/
	    and_or_edge=Be_Or(be_mgr,and_or_edge, and_edge);
	  }
	  if(optimisations.buchi)
	    {
	      result =
		Be_And(be_mgr, 
		       result,
		       Be_Implies(be_mgr,
				  AutomataStateVarTime(be_enc,waba_vars,partmap,
						       time,depth,
						       Bmc_PartitionMap_AutToId(partmap,sourcenode->id),k),
				  and_or_edge)
		       );
	    }
	  else
	    {
	      result = Be_And(be_mgr, 
			      result,
			      be_implies_iff(be_mgr,
					     AutomataStateVarTime(be_enc,waba_vars,partmap,time,depth,
								  Bmc_PartitionMap_AutToId(partmap,sourcenode->id),k),
					     and_or_edge)
			      );
	    }
	} /*For all depths */
      
    } /*For all states...*/
  
  return result;
}

/*
 * Similar to Bmc_Utils_generate_and_print_cntexample but modified
 * for my incremental SAT solver interface
 */
static Trace_ptr _generate_and_print_cntexample(BeEnc_ptr be_enc,
						SymbLayer_ptr psl_layer,
						Bmc_IncrSat_ptr solver, 
						BddEnc_ptr bdd_enc, 
						const int k, 
						const char* trace_name,
						be_ptr l_var)
{
  BmcSatTrace_ptr sat_trace;
  Trace_ptr trace;
  lsList be_model;
  node_ptr path;
  int bmc_tr;
  hash_ptr psl_layer_vars = 0;

  assert(l_var);

  /*	  
  The original
  be_model = Be_CnfModelToBeModel(BeEnc_get_be_manager(be_enc), 
				  SatSolver_get_model(solver));
  is replaced by the following:
  */
  {
    /* Find all vars that are in the PSL layer */
    psl_layer_vars = new_assoc();
    NodeList_ptr l = SymbLayer_get_all_vars(psl_layer);
    ListIter_ptr iter = NodeList_get_first_iter(l);
    while(!ListIter_is_end(iter))
      {
	node_ptr node = NodeList_get_elem_at(l, iter);
	insert_assoc(psl_layer_vars, node, node);
	iter = ListIter_get_next(iter);
      }
  }

  be_model = lsCreate();
  {
    int i = 0;
    /* For each time step ... */
    for(i = _real_k(0); i <= _real_k(k); i++)
      {
	/* get values of state and input variables */
	int index = BeEnc_get_first_untimed_var_index(be_enc, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
	while(BeEnc_is_var_index_valid(be_enc, index))
	  {
	    node_ptr var_node;
	    be_ptr var;
	    be_ptr var_real_i;
	    int result;

	    /* Ignore vars that belong to the PSL layer */
	    var_node = BeEnc_index_to_name(be_enc, index);
	    if(find_assoc(psl_layer_vars, var_node))
	      {
		index = BeEnc_get_next_var_index(be_enc, index, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
		continue;
	      }

	    var = BeEnc_index_to_var(be_enc, index);
	    var_real_i = BeEnc_var_to_timed(be_enc, var, i);

	    result = Bmc_IncrSat_get_solution_value(solver, var_real_i);
	    if(result == 0 || result == 1)
	      {
		be_ptr var_model_i =
		  BeEnc_var_to_timed(be_enc, var, _model_k(i));
		int rbc_index = Be_Var2Index(BeEnc_get_be_manager(be_enc), var_model_i);
		  /*int rbc_index = BeEnc_var_to_index(be_enc, var_model_i);*/
		if(result == 0)
		  rbc_index = -rbc_index;
		lsNewEnd(be_model, (lsGeneric)rbc_index, LS_NH);
	      }

	    /* Go to next var */
	    index = BeEnc_get_next_var_index(be_enc, index, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
	  }
      }
  }
  /* Free PSL layer vars */
  free_assoc(psl_layer_vars); psl_layer_vars = 0;


  /* As no real be_prob was ever generated,
     it is replaced by a dummy, non-constant be */
  sat_trace = BmcSatTrace_create(BeEnc_name_to_timed(be_enc, l_var, 0),
				 be_model);
  /*sat_trace = BmcSatTrace_create(be_prob, be_model);*/
  path = NodeList_to_node_ptr(
	    BmcSatTrace_get_symbolic_model(sat_trace, be_enc, k));
	  
  trace = Trace_create_from_state_input_list(bdd_enc,
					     trace_name,
					     TRACE_TYPE_CNTEXAMPLE,
					     path);

  bmc_tr = TraceManager_register_trace(global_trace_manager, trace);
	  
  /* Print the trace using default plugin */
  fprintf(nusmv_stdout,
	  "-- as demonstrated by the following execution sequence\n");

  /* New code: find and print the loopback point */
  {
    int i = 0;
    int loopback = -1;
    for(i = 0; i <= k; i++)
      {
	be_ptr l_var_real_i = BeEnc_name_to_timed(be_enc, l_var, _real_k(i));
	int result = Bmc_IncrSat_get_solution_value(solver, l_var_real_i);
	assert(result == 0 || result == 1);
	if(result == 1)
	  {
	    assert(loopback == -1);
	    loopback = i;
	  }
      }
    if(loopback != -1)
      {
	/* NuSMV prints traces so that first state is numbered 1 */
	fprintf(nusmv_stdout,
		"   the loop starts at state %d (that is redundantly printed also as state %d)\n", loopback+1-1, k+1);
      }
    else
      {
	fprintf(nusmv_stdout, "   the execution has no loop\n");
      }
  }

  TraceManager_execute_plugin(global_trace_manager, 
		      TraceManager_get_default_plugin(global_trace_manager),
		      bmc_tr);

  BmcSatTrace_destroy(&sat_trace);
  lsDestroy(be_model, 0);

  return trace;
}
