/**CHeaderFile*****************************************************************

  FileName    [sugar.h]

  PackageName [sugar]

  Synopsis    [Routines for parsing, writing and accessing SUGAR formulas.]

  Description [This package implements a parser for Sugar's LTL (Linear Time
  Logic) formulas and SERE's (Sugar Extended Regular Expression).  LTL is a
  language used to describe properties of systems.]

  SeeAlso     [mc]

  Author      [Christian Josef Kargl]

  Copyright   [Copyright (c) 2003
               University of Technology Graz, Institute for Softwaretechnology]

  All rights reserved.

  Permission is hereby granted, without written agreement and without license
  or royalty fees, to use, copy, modify, and distribute this software and its
  documentation for any purpose, provided that the above copyright notice and
  the following two paragraphs appear in all copies of this software.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]

  Revision    [$Id: sugar.h,v 1.1 2005/12/22 23:59:08 timo Exp $]

******************************************************************************/

#ifndef _SUGAR
#define _SUGAR

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/
#include "utils/utils.h" /*For EXTERN*/
#include "utils/array.h" /*For array_t*/
/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

EXTERN FILE* nusmv_stdout;


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/**Enum************************************************************************

  Synopsis    [Types of operators allowed in SUGAR formulas.]

  Description [The types of nodes in a SUGAR formula parse tree. ID, TRUE, and
  FALSE are leaves, and all others are internal nodes.]

  SeeAlso     [Sugar_FormulaReadType]

******************************************************************************/
typedef enum {
  Sugar_FL_Next_c,           /* formula next X f */
  Sugar_FL_Always_c,         /* formula always G f */
  Sugar_FL_Never_c,          /* formula never is equal to !(G f) */
  Sugar_FL_Eventually_c,     /* formula finally F f */
  Sugar_FL_Releases_c,       /* formula releases f V g */
  Sugar_FL_Until_c,          /* formula until construct f U g */
  Sugar_FL_Until__c,         /* formula until construct */
  Sugar_FL_Before_c,         /* formula before construct */
  Sugar_FL_Before__c,        /* formula before construct */
  Sugar_FL_Not_c,            /* boolean not of property !f */
  Sugar_FL_And_c,            /* boolean and of properties f && g */
  Sugar_FL_Or_c,             /* boolean or of properties f || g */
  Sugar_FL_Imply_c,          /* boolean implication of properties f -> g */
  Sugar_FL_Equiv_c,          /* boolean equivalence of properties f <-> g */
  Sugar_Formula_sep,         /* ---- seperator for negation ---- */
  Sugar_Seq_Always_c,        /* formula for SERE version of G f is always r */
  Sugar_Seq_Never_c,         /* formula for SERE version of !(G f) is never r */
  Sugar_Seq_Eventually_c,    /* formula for SERE version of F f is always r */
  Sugar_Seq_FLProp_c,        /* formula suffix imply with property r(f) */
  Sugar_Seq_Imply_c,         /* formula suffix imply with one overlap r |-> s */
  Sugar_Seq_Imply2_c,        /* formula suffix imply without overlap r |=> s */
  Sugar_Seq_Within_c,        /* SERE witin construct (TODO) */
  Sugar_Seq_Within__c,       /* SERE witin construct (TODO) */
  Sugar_Seq_Whilenot_c,      /* SERE whilenot construct (TODO) */
  Sugar_Seq_Whilenot__c,     /* SERE whilenot construct (TODO) */
  Sugar_Sequence_sep,        /* ---- seperator for negation ---- */
  Sugar_RegExpr_Concat_c,    /* SERE concatination r ; s */
  Sugar_RegExpr_Compose_c,   /* SERE composition with one overlap r : s */
  Sugar_RegExpr_Or_c,        /* SERE or r | s */
  Sugar_RegExpr_And_c,       /* SERE and r & s */
  Sugar_RegExpr_AndEq_c,     /* SERE length matching and r && s*/
  Sugar_RegExpr_Plus_c,      /* SERE at least one repeating [+] */
  Sugar_RegExpr_Star_c,      /* SERE repeating construct [* count ] */
  Sugar_RegExpr_Equal_c,     /* SERE counting construct [= count ]*/
  Sugar_RegExpr_Goto_c,      /* SERE goto construct [-> count ] */
  Sugar_IDENTIFIER_c,        /* an atomic proposition */
  Sugar_TRUE_c,              /* tautology (used only for translation to exist. form) */
  Sugar_FALSE_c,             /* empty (used only for translation to exist. form) */
} Sugar_FormulaType;

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
typedef struct SugarAutomataStruct   Sugar_Automaton_t;
typedef struct SugarEdgeStruct       Sugar_Edge_t;
typedef struct SugarLabelStruct      Sugar_Label_t;
typedef struct SugarVertexStruct     Sugar_Vertex_t;
typedef struct SugarArrayPairStruct  Sugar_Array_Pair_t;
typedef struct SugarFormulaStruct    Sugar_Formula_t;

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

EXTERN void Sugar_Init(void);
EXTERN void Sugar_End(void);

/* /\* EXTERN array_t * Sugar_FileParseFormulaArray(FILE *fp); *\/ */
EXTERN char * Sugar_FormulaConvertToString(Sugar_Formula_t *formula);
EXTERN void Sugar_FormulaPrint(FILE *fp, Sugar_Formula_t *formula);
EXTERN char* Sugar_FormulaRandom(int maxLevel, int numProps, int makeFormula);

EXTERN void Sugar_FormulaFree(Sugar_Formula_t *formula);
/* EXTERN void Sugar_FormulaArrayFree(array_t *formulaArray); */
EXTERN Sugar_Formula_t * Sugar_FormulaCreate(Sugar_FormulaType type, void *left, void *right);
/* EXTERN boolean Sugar_FormulaStaticSemanticCheckOnNetwork(Sugar_Formula_t *formula, Ntk_Network_t *network); */

EXTERN Sugar_Automaton_t * Sugar_AutomatonCreate(void);
EXTERN void Sugar_AutomatonFree(Sugar_Automaton_t *aut);
EXTERN void Sugar_AutomatonPrint(Sugar_Automaton_t *aut, int verbosity);
EXTERN void Sugar_AutomatonPrintDot(Sugar_Automaton_t *aut, FILE *fp);
EXTERN Sugar_Automaton_t* Sugar_AutomatonGenerate(Sugar_Formula_t *formula);
EXTERN Sugar_Automaton_t* Sugar_AutomatonExpand(Sugar_Automaton_t *aut);
EXTERN Sugar_Automaton_t* Sugar_AutomatonDeterministic(Sugar_Automaton_t *aut);
/* EXTERN Ltl_Automaton_t* Sugar_AutomatonToLtlAutomaton(Sugar_Automaton_t *aut); */

/**AutomaticEnd***************************************************************/

#endif /* _SUGAR */
