/**CFile***********************************************************************

FileName    [sugarAutomaton.c]

PackageName [sugar]

Synopsis    [Translate SUGAR formula to alternating automaton.]

Author      [Christian Josef Kargl]

Copyright   [This file was created at the University of Technology Graz.
The University of Technology Graz makes no warranty about the suitability
of this software for any purpose. It is presented on an AS IS basis.]

History    [Initial Version: Jul, 08. 2003
Last Modified:   Dec, 12. 2003]

******************************************************************************/

#include "sugarInt.h"
/* #include "ltlSet.h" */
/* #include "ctlspInt.h" */
/* #include "ltlInt.h" */

/* static char rcsid[] UNUSED = "$Id: sugarAutomaton.c,v 1.3 2006/02/24 13:33:17 tjunttil Exp $"; */

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int bitRange(int x);
static int lsCompareVertex(lsGeneric obj1, lsGeneric obj2);
static int compareVertex(void *obj1, void* obj2);
static int compareLabel(void *obj1, void* obj2);
static int compareEdge(void *obj1, void* obj2);
static array_t* ArrayMerge(array_t *a1, array_t *a2);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

Synopsis    [Allocate the automaton data structure.]

Description [Initializes and empty alternating automaton.]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     [Sugar_AutomatonFree]

******************************************************************************/
Sugar_Automaton_t*
Sugar_AutomatonCreate(void)
{
  Sugar_Automaton_t *aut = ALLOC(Sugar_Automaton_t, 1);

  aut->name            = NIL(char);
  aut->state_list      = lsCreate();
  aut->prop_list       = lsCreate();
  aut->edge_list       = lsCreate();
  aut->initial_states  = array_alloc(array_t*, 0);
  aut->accept_states   = array_alloc(Sugar_Vertex_t*, 0);
  aut->num_state_index = 0;
  return aut;
}

/**Function********************************************************************

Synopsis    [Free the automaton data structure.]

Description []

SideEffects []

SeeAlso     [Sugar_AutomatonCreate]

******************************************************************************/
void
Sugar_AutomatonFree(
  Sugar_Automaton_t *aut)
{
  int index;
  char *name;
  array_t *item;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;

  /* nothing to do */
  if (aut == NIL(Sugar_Automaton_t))
    return;

  /* delete name */
  if (aut->name)
    FREE(aut->name);

  /* delete state list */
  while (lsDelBegin(aut->state_list, (lsGeneric *) &node) != LS_NOMORE) {
    lsDestroy(node->edge_in_list, (void (*)(lsGeneric)) NULL);
    lsDestroy(node->edge_out_list, (void (*)(lsGeneric)) NULL);
    FREE(node);
  }
  lsDestroy(aut->state_list, (void (*)(lsGeneric)) NULL);

  /* delete proposition name list */
  while (lsDelBegin(aut->prop_list, (lsGeneric *) &name) != LS_NOMORE) {
    FREE(name);
  }
  lsDestroy(aut->prop_list, (void (*)(lsGeneric)) NULL);

  /* delete edge list */
  while (lsDelBegin(aut->edge_list, (lsGeneric *) &edge) != LS_NOMORE) {
    arrayForEachItem(Sugar_Label_t*, edge->label, index, labl) {
      FREE(labl);
    }
    array_free(edge->label);
    lsDestroy(edge->dest_vertex_list, (void (*)(lsGeneric)) NULL);
    FREE(edge);
  }
  lsDestroy(aut->edge_list, (void (*)(lsGeneric)) NULL);

  /* delete initial states array and accepting states */
  arrayForEachItem(array_t*, aut->initial_states, index, item) {
    array_free(item);
  }
  array_free(aut->initial_states);
  array_free(aut->accept_states);

  /* delete automaton */
  FREE(aut);
}

/**Function********************************************************************

Synopsis    [Print the automaton.]

Description [Prints the alternating automaton to VIS standard output.
Use verbosity for the amount of information: 0 prints only the formula
name, 1 prints adds some statistic values, 2 prints edges and accepting
states and 3 also prints the states with number of in-/out-edges.]

SideEffects []

SeeAlso     []

******************************************************************************/
void
Sugar_AutomatonPrint(
  Sugar_Automaton_t *aut,
  int verbosity)
{
  int first;
  int index1;
  int index2;
  char *name;
  char *temp;
  char *numbers;
  array_t *item;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  lsGen gen1, gen2;

  /* just in case */
  if (aut == NIL(Sugar_Automaton_t))
    return;

  /* formula */
  fprintf(nusmv_stdout, "\nAutomaton:\n");
  fprintf(nusmv_stdout, "-----------------------------------------------\n");
  if (aut->name)
    fprintf(nusmv_stdout, "Name: %s \n", aut->name);
  fprintf(nusmv_stdout, "-----------------------------------------------\n");
  /* print more ? */
  if (verbosity > 0) {
    if (verbosity > 1) {
      if (verbosity > 2) {
        /* all states */
        fprintf(nusmv_stdout, "\nStates:\n");
        lsForEachItem(aut->state_list, gen1, node) {
          fprintf(nusmv_stdout, "  q%d:   IN %d / OUT %d (%i)\n", node->id,
                  lsLength(node->edge_in_list), lsLength(node->edge_out_list), node);
        }
      }
      /* Initial states */
      fprintf(nusmv_stdout, "\nArcs:\n");
      arrayForEachItem(array_t*, aut->initial_states, index1, item) {
        first = 1;
        arrayForEachItem(Sugar_Vertex_t*, item, index2, node) {
          fprintf(nusmv_stdout, " %s q%d(%i)", (first)?("->"):("&"), node->id, node);
          first = 0;
        }
        fprintf(nusmv_stdout, "\n");
      }
      /* Edges with label */
      lsForEachItem(aut->edge_list, gen1, edge) {
        first = 1;
        fprintf(nusmv_stdout, "    q%d", edge->source_vertex->id);
        lsForEachItem(edge->dest_vertex_list, gen2, node) {
          fprintf(nusmv_stdout, "%sq%d", (first)?(" -> "):(","), node->id);
          first = 0;
        }
        first = 1;
        fprintf(nusmv_stdout, "  Label: ");
        arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
          if (labl->name == NIL(char)) {
            fprintf(nusmv_stdout, "%s", (labl->isNegate)?"FALSE":"TRUE");
          }
          else {
            fprintf(nusmv_stdout, "%s%s'%s'", (first)?(""):("&"), (labl->isNegate)?("!"):(""), labl->name);
          }
          first = 0;
        }
        fprintf(nusmv_stdout, "\n");
      }
      /* Accepting states */
      first = 1;
      fprintf(nusmv_stdout, "\nAccept States:\n  {");
      arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index1, node) {
        fprintf(nusmv_stdout, "%sq%d(%i)", (first)?(""):(","), node->id, node);
        first = 0;
      }
      fprintf(nusmv_stdout, "}\n");
    }
    /* statistics */
    first = 1;
    temp = ALLOC(char, 12);
    numbers = util_strsav("");
    arrayForEachItem(array_t*, aut->initial_states, index1, item) {
      sprintf(temp, "%s%d", (first)?(""):(","), array_n(item));
      name = util_strcat3(numbers, temp, "");
      FREE(numbers); numbers = name; first = 0;
    }
    fprintf(nusmv_stdout, "\n");
    fprintf(nusmv_stdout, "Number of states:         %d\n", lsLength(aut->state_list));
    fprintf(nusmv_stdout, "Number of symbols:        %d\n", lsLength(aut->prop_list));
    fprintf(nusmv_stdout, "Number of arcs:           %d\n", lsLength(aut->edge_list));
    fprintf(nusmv_stdout, "Number of initial states: {%s}\n", numbers);
    fprintf(nusmv_stdout, "Number of accept states:  %d\n\n", array_n(aut->accept_states));
    FREE(numbers); FREE(temp);

  }
}

/**Function********************************************************************

Synopsis    [Print non-alterating automaton in DOT format.]

Description [Write the automaton to file in DOT format. Because it is not
possible to display alternating edges - edges with multiple destinations -
the automaton must be tranformed with Sugar_AutomatonExpand() before this
function is called. Otherwise the automaton is incomplete!]

SideEffects []

SeeAlso     [Sugar_AutomatonExpand]

******************************************************************************/
void
Sugar_AutomatonPrintDot(
  Sugar_Automaton_t *aut,
  FILE *fp)
{
  int first;
  int index1;
  int index2;
  int not_found;
  array_t *item;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *node2;
  lsGen gen;

  /* just in case */
  if (aut == NIL(Sugar_Automaton_t))
    return;

  /* formula */
  fprintf(fp, "digraph \"%s\" {\n", aut->name);
  fprintf(fp, "  center = true;\n");
  fprintf(fp, "  \"title\" [label=\"%s\", shape=plaintext]\n", aut->name);
  fprintf(fp, "  edge [fontname=\"Courier\", fontsize=10];\n");

  /* Initial states */
  arrayForEachItem(array_t*, aut->initial_states, index1, item) {
    node = array_fetch(Sugar_Vertex_t*, item, 0);
    not_found = 1;
    arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index2, node2) {
      if (node == node2) { not_found = 0; break; }
    }
    if (not_found) {
      fprintf(fp, "  q%d [shape=box];\n", node->id);
    }
    else {
      fprintf(fp, "  q%d [shape=box, style=bold];\n", node->id);
    }
  }
  /* Accepting states */
  arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index1, node) {
    not_found = 1;
    arrayForEachItem(array_t*, aut->initial_states, index2, item) {
      node2 = array_fetch(Sugar_Vertex_t*, item, 0);
      if (node == node2) { not_found = 0; break; }
    }
    if (not_found) {
      fprintf(fp, "  q%d [style=bold];\n", node->id);
    }
  }
  /* Edges with label */
  lsForEachItem(aut->edge_list, gen, edge) {
    first = 1;
    lsFirstItem(edge->dest_vertex_list, (lsGeneric*)&node, NULL);
    fprintf(fp, "  q%d -> q%d [label=\"", edge->source_vertex->id, node->id);
    arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
      if (labl->name == NIL(char)) {
        fprintf(fp, "%s", (labl->isNegate)?"FALSE":"TRUE");
      }
      else {
        fprintf(fp, "%s%s%s", (first)?(""):(" & "), (labl->isNegate)?("!"):(""), labl->name);
      }
      first = 0;
    }
    fprintf(fp, "\"];\n");
  }
  fprintf(fp, "}\n");
}

/**Function********************************************************************

Synopsis    [Generate alternating automaton from the formula.]

Description [Starting with identifiers, an alternating automaton is built
recursivly from the formula. For negations of a sub-formula the function
switches to SugarAutomatonNot().]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     [SugarAutomatonIdentifer, SugarAutomatonNot]

******************************************************************************/
Sugar_Automaton_t*
Sugar_AutomatonGenerate(
  Sugar_Formula_t *formula)
{
  int isEqualExpr = 0;
  int index, nStr, nEnd;
  array_t *new_ini_array;
  array_t *new_acc_array;
  Sugar_Edge_t* edge;
  Sugar_Vertex_t *node;
  Sugar_Automaton_t *A_temp;
  Sugar_Automaton_t *A_left;
  Sugar_Automaton_t *A_right;
  Sugar_Automaton_t *A_left_clone;
  Sugar_Automaton_t *A_right_clone;
  lsGen gen;

/*   fprintf(nusmv_stdout, "\033[21;32m\033[07m******************************************\n* creating Automaton for %s\n\033[0m", Sugar_FormulaConvertToString(formula)); */

  /* no formula returns no automaton */
  if (formula == NIL(Sugar_Formula_t)) {
    return NIL(Sugar_Automaton_t);
  }
  /* type TRUE returns single edge with no label */
  if (formula->type == Sugar_TRUE_c) {
    return SugarAutomatonIdentifier(NIL(char), 0, formula->isStrongOperator);
  }
  /* type FALSE returns single edge with no label and flag */
  if (formula->type == Sugar_FALSE_c) {
    return SugarAutomatonIdentifier(NIL(char), 1, formula->isStrongOperator);
  }
  /* type IDENTIFIER returns single edge with name as label */
  if (formula->type == Sugar_IDENTIFIER_c) {
    return SugarAutomatonIdentifier((char*)(formula->left), 0, formula->isStrongOperator);
  }
  /* build NEGATION from formula, not with automaton */
  if (formula->type == Sugar_FL_Not_c) {
    return SugarAutomatonNot(formula->left);
  }
  /* build NEVER f with (G !f), but from formula */
  if (formula->type == Sugar_FL_Never_c) {
    A_temp = SugarAutomatonNot(formula->left);
    return SugarAutomatonAlways(A_temp);
  }
  /*
   * The formula is an inner node: create left and right automata
   */
  A_left = Sugar_AutomatonGenerate(formula->left);
  A_right = Sugar_AutomatonGenerate(formula->right);
  /*
   * Connect both automata depending on type. Depending on the
   * type some parts of the right automata gets deleted.
   */
  switch(formula->type) {
  case Sugar_FL_Next_c:
    nStr = formula->numRangeStr;
    if (nStr == SUGAR_RANGE_UNUSED) nStr = 1;
    /* create automata by inserting TRUE edges at the start */
    while (nStr-- > 0) {
      A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
      SugarAutomatonConcat(A_temp, A_left, 0);
      /* weak version may not have next cycle -> new initial state is accepting */
      if (!formula->isStrongOperator) {
	new_ini_array = array_fetch(array_t*, A_temp->initial_states, 0);
	node = array_fetch(Sugar_Vertex_t*, new_ini_array, 0);
	array_insert_last(Sugar_Vertex_t*, A_temp->accept_states, node);
      }
      A_left = A_temp;
    }
    break;
  case Sugar_FL_Always_c:
    A_left = SugarAutomatonAlways(A_left);
    break;
  case Sugar_FL_Eventually_c:
    A_left = SugarAutomatonEventually(A_left);
    break;
  case Sugar_FL_Releases_c:
    /* build automaton [f V g] as [(g U (f and g))] */
    A_right_clone = Sugar_AutomatonGenerate(formula->right);
    SugarAutomatonAndEq(A_left, A_right_clone);
    A_left = SugarAutomatonUntil(A_right, A_left, 0); // weak
    break;
  case Sugar_FL_Until_c:
    /* build automaton [f U! g] */
    A_left = SugarAutomatonUntil(A_left, A_right, formula->isStrongOperator);
    break;
  case Sugar_FL_Until__c:
    /* build automaton as [f U! (f and g)] */
    A_left_clone = Sugar_AutomatonGenerate(formula->left);
    SugarAutomatonAndEq(A_right, A_left_clone);
    A_left = SugarAutomatonUntil(A_left, A_right, formula->isStrongOperator);
    break;
  case Sugar_FL_Before_c:
    fprintf(nusmv_stdout, "Construct 'before' not implemented yet! IGNORED!\n");
    break;
  case Sugar_FL_Before__c:
    fprintf(nusmv_stdout, "Construct 'before_' not implemented yet! IGNORED!\n");
    break;
  case Sugar_FL_And_c:
    SugarAutomatonAndEq(A_left, A_right);      
    break;
  case Sugar_FL_Or_c:
    SugarAutomatonOr(A_left, A_right);
    break;
  case Sugar_FL_Imply_c:
    /* build automaton as [not(f1) or f2] */
    Sugar_AutomatonFree(A_left);
    A_left = SugarAutomatonNot(formula->left);
    SugarAutomatonOr(A_left, A_right);
    break;
  case Sugar_FL_Equiv_c:
    /* build automaton as [(f1 and f2) or (not(f1) and not(f2))] */
    SugarAutomatonAndEq(A_left, A_right);
    A_left_clone = SugarAutomatonNot(formula->left);
    A_right_clone = SugarAutomatonNot(formula->right);
    SugarAutomatonAndEq(A_left_clone, A_right_clone);
    SugarAutomatonOr(A_left, A_left_clone);
    break;
  case Sugar_Seq_Always_c:
    /*
     * if semantics for weak suffix implication are defined use this code:
     * #==--
     * |  /* for SERE build always{r} with {T+}|->{r}
     * |  A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
     * |  SugarAutomatonPlus(A_temp);
     * |  A_left = SugarAutomatonImply(A_temp, A_left, 1, 0); // weak operator
     * #==--
     */
    /* for now we use formula version (G {r}) instead */
    A_left = SugarAutomatonAlways(A_left);
    /* add loops on accepting states without outgoing edges */
    arrayForEachItem(Sugar_Vertex_t*, A_left->accept_states, index, node) {
      if (lsLength(node->edge_out_list) == 0) {
	SugarAutomatonNewEdgeLoop(A_left, node);
      }
    }
    break;
  case Sugar_Seq_Never_c:
    /*
     * if semantics for weak suffix implication are defined use this code:
     * #==--
     * |  /* for SERE build never{r} with {T*;r}|->{F} equals {r}|->{F} here
     * |  A_right = SugarAutomatonIdentifier(NIL(char), 1, 0);
     * |  A_left = SugarAutomatonImply(A_left, A_right, 1, 0); // weak operator
     * #==--
     */
    /* add self loops to accepting states for infinite words */
    arrayForEachItem(Sugar_Vertex_t*, A_left->accept_states, index, node) {
      if (lsLength(node->edge_out_list) > 0) {
	/* delete outgoing edges from accepting state */
	lsForEachItem(node->edge_out_list, gen, edge) {
	  SugarAutomatonFreeEdge(edge);
	}
      }
      SugarAutomatonNewEdgeLoop(A_left, node);
    }
    /* for now we use formula version (G !{r}) instead */
    A_temp = SugarAutomatonNegate(A_left);
    Sugar_AutomatonFree(A_left);
    A_left = SugarAutomatonAlways(A_temp);
    break;
  case Sugar_Seq_Eventually_c:
    /* for SERE build eventually!{r} with {T}|->{T*:r}!*/
    A_right = SugarAutomatonIdentifier(NIL(char), 0, 0);
    SugarAutomatonStar(A_right);
    SugarAutomatonConcat(A_right, A_left, 1);
    A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
    A_left = SugarAutomatonImply(A_temp, A_right, 1, 1); // strong operator
    break;
  case Sugar_Seq_FLProp_c: //formula suffix implication
    //      A_left = SugarAutomatonImply(A_left, A_right, 0, formula->isStrongOperator);
/*     fprintf(nusmv_stdout, "\033[01;34m\033[07m*********************\n** sugar FL prop !**\033[0m\033[22;34m\n"); */

    A_left = SugarAutomatonImplyNFA(A_left, A_right);
    Sugar_AutomatonPrint(A_left, 1);
    break;
  case Sugar_Seq_Imply_c:
    // AG preliminary call for testing cases
    A_left=SugarAutomatonImplyNFA(A_left, A_right);
    // end call
      
    //      A_left = SugarAutomatonImply(A_left, A_right, 1, formula->isStrongOperator);
    break;
  case Sugar_Seq_Imply2_c:
    /* construct {r1}|=>{r2} as {r1}|->{T;r2} */
    A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
    SugarAutomatonConcat(A_temp, A_right, 0);
    //      A_left = SugarAutomatonImply(A_left, A_temp, 1, formula->isStrongOperator);
    A_left = SugarAutomatonImplyNFA(A_left, A_temp);
    break;
  case Sugar_Seq_Within_c:
    fprintf(nusmv_stdout, "Construct 'within' not implemented yet! IGNORED!\n");
    break;
  case Sugar_Seq_Within__c:
    fprintf(nusmv_stdout, "Construct 'within_' not implemented yet! IGNORED!\n");
    break;
  case Sugar_Seq_Whilenot_c:
    fprintf(nusmv_stdout, "Construct 'whilenot' not implemented yet! IGNORED!\n");
    break;
  case Sugar_Seq_Whilenot__c:
    fprintf(nusmv_stdout, "Construct 'whilenot_' not implemented yet! IGNORED!\n");
    break;
  case Sugar_RegExpr_Concat_c:
    SugarAutomatonConcat(A_left, A_right, 0);
    break;
  case Sugar_RegExpr_Compose_c:
    SugarAutomatonConcat(A_left, A_right, 1);
    break;
  case Sugar_RegExpr_Or_c:
    SugarAutomatonOr(A_left, A_right);
    break;
  case Sugar_RegExpr_And_c:
    /* duplicate left and right formula */
    A_left_clone = Sugar_AutomatonGenerate(formula->left);
    A_right_clone = Sugar_AutomatonGenerate(formula->right);
    /* build (r1)&&(r2;T*) */
    A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
    SugarAutomatonStar(A_temp);
    SugarAutomatonConcat(A_right_clone, A_temp, 0);
    SugarAutomatonAndEq(A_left_clone, A_right_clone);
    /* build (r1;T*)&&(r2) */
    A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
    SugarAutomatonStar(A_temp);
    SugarAutomatonConcat(A_left, A_temp, 0);
    SugarAutomatonAndEq(A_left, A_right);
    /* finally or them and make automaton non-deterministic */
    SugarAutomatonOr(A_left, A_left_clone);
    A_left_clone = Sugar_AutomatonExpand(A_left);
    Sugar_AutomatonFree(A_left);
    A_left = A_left_clone;
    break;
  case Sugar_RegExpr_AndEq_c:
    fprintf(nusmv_stdout, "before and automatons:\n");
    Sugar_AutomatonPrint(A_left, 3);
    Sugar_AutomatonPrint(A_right, 3);      

    SugarAutomatonAndEq(A_left, A_right);

    fprintf(nusmv_stdout, "result of and\n");
    Sugar_AutomatonPrint(A_left, 3);

    /* make automaton non-deterministic */
    /*       A_left_clone = Sugar_AutomatonExpand(A_left); */
    /*       Sugar_AutomatonFree(A_left); */
    /*       A_left = A_left_clone; */
    break;
  case Sugar_RegExpr_Plus_c:
    SugarAutomatonPlus(A_left);
    break;
  case Sugar_RegExpr_Star_c:
    /* read limits */
    nStr = formula->numRangeStr;
    nEnd = formula->numRangeEnd;
    /* check normal case r[*] */
    if ((nStr == SUGAR_RANGE_UNUSED) && (nEnd == SUGAR_RANGE_UNUSED)) {
      SugarAutomatonStar(A_left);
    }
    else {
      if (nStr != SUGAR_RANGE_MIN_VAL) {
	/* build r[*i] */
	for (index=1; index<nStr; index++) {
	  A_left_clone = Sugar_AutomatonGenerate(formula->left);
	  SugarAutomatonConcat(A_left, A_left_clone, 0);
	}
      }
      if (nEnd != SUGAR_RANGE_UNUSED) {
	if (nEnd == SUGAR_RANGE_MAX_VAL) {
	  /* need to add r* for r[*i..] */
	  A_right = Sugar_AutomatonGenerate(formula->left);
	  SugarAutomatonStar(A_right);
	}
	else {
	  /* need to add r[*..j-i] for r[*..j] and r[*i..j] */
	  A_right = Sugar_AutomatonGenerate(formula->left);
	  for (index=nStr+1; index<nEnd; index++) {
	    A_right_clone = Sugar_AutomatonGenerate(formula->left);
	    SugarAutomatonConcat(A_right, A_right_clone, 0);
	  }
	  new_acc_array = array_alloc(Sugar_Vertex_t*, 0);
	  lsForEachItem(A_right->state_list, gen, node) {
	    array_insert_last(Sugar_Vertex_t*, new_acc_array, node);
	  }
	  array_free(A_right->accept_states);
	  A_right->accept_states = new_acc_array;
	}
	if (nStr == SUGAR_RANGE_MIN_VAL) {
	  Sugar_AutomatonFree(A_left);
	  A_left = A_right;
	}
	else {
	  SugarAutomatonConcat(A_left, A_right, 0);
	}
      }
    }
    break;
  case Sugar_RegExpr_Equal_c:
    /* set flag for [=] expression and fall through to [->] expression */
    isEqualExpr = 1;
  case Sugar_RegExpr_Goto_c:
    /* read limits */
    nStr = formula->numRangeStr;
    nEnd = formula->numRangeEnd;
    Sugar_AutomatonFree(A_left);
    /* check normal case b[->] */
    if ((nStr == SUGAR_RANGE_UNUSED) && (nEnd == SUGAR_RANGE_UNUSED)) {
      A_left = SugarAutomatonGoto((char*)formula->left->left);
    }
    else {
      if (nStr != SUGAR_RANGE_MIN_VAL) {
	/* build b[->i] as b[->];b[->(i-1)] same for b[=] */
	A_left = SugarAutomatonGoto((char*)formula->left->left);
	for (index=0; index<nStr-1; index++) {
	  A_temp = SugarAutomatonGoto((char*)formula->left->left);
	  SugarAutomatonConcat(A_left, A_temp, 0);
	}
	/* for b[=i] we need to add !b* to the end of the automaton */
	if (isEqualExpr) {
	  if (nEnd == SUGAR_RANGE_UNUSED) {
	    A_temp = SugarAutomatonIdentifier((char*)formula->left->left, 1, 0);
	    SugarAutomatonStar(A_temp);
	    SugarAutomatonConcat(A_left, A_temp, 0);
	  }
	}
      }
      if (nEnd != SUGAR_RANGE_UNUSED) {
	if (nEnd == SUGAR_RANGE_MAX_VAL) {
	  if (isEqualExpr == 0) {
	    /* need to add (T*;b) for b[->i..] */
	    A_right = SugarAutomatonIdentifier(NIL(char), 0, 0);
	    SugarAutomatonStar(A_right);
	    A_temp = SugarAutomatonIdentifier((char*)formula->left->left, 0, 0);
	  }
	  else {
	    /* need to add (!b*;T*) for b[=i..] */
	    A_right = SugarAutomatonIdentifier((char*)formula->left->left, 1, 0);
	    SugarAutomatonStar(A_right);
	    A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
	    SugarAutomatonStar(A_temp);
	  }
	  SugarAutomatonConcat(A_right, A_temp, 0);
	}
	else {
	  /* need to add b[->..j-i] for b[->..j] and b[->i..j] same for b[=] */
	  A_right = SugarAutomatonGoto((char*)formula->left->left);
	  new_acc_array = array_dup(A_right->accept_states);
	  for (index=nStr; index<nEnd-1; index++) {
	    A_temp = SugarAutomatonGoto((char*)formula->left->left);
	    SugarAutomatonConcat(A_right, A_temp, 0);
	    array_append(new_acc_array, A_right->accept_states);
	  }
	  array_free(A_right->accept_states);
	  A_right->accept_states = new_acc_array;
	  /* add !b* to end of automaton for b[=..j] or b[=i..j] */
	  if (isEqualExpr) {
	    A_temp = SugarAutomatonIdentifier((char*)formula->left->left, 1, 0);
	    SugarAutomatonStar(A_temp);
	    SugarAutomatonConcat(A_right, A_temp, 0);
	  }
	}
	if (nStr == SUGAR_RANGE_MIN_VAL) {
	  A_left = A_right; /* A_left already deleted */
	}
	else {
	  /* build (b[->];(eps|T*;b)) otherwise build (b[=];T*) */
	  if (nEnd == SUGAR_RANGE_MAX_VAL) {
	    SugarAutomatonConcat(A_left, A_right, 0);
	  }
	  else {
	    new_acc_array = array_dup(A_left->accept_states);
	    SugarAutomatonConcat(A_left, A_right, 0);
	    array_append(new_acc_array, A_left->accept_states);
	    array_free(A_left->accept_states);
	    A_left->accept_states = new_acc_array;
	  }
	}
      }
    }
    break;
  default:
    fail("Unexpected type");
  }
/*   fprintf(nusmv_stdout, "\033[22;32m\033[07m******************************************\n* created Automaton for %s\n\033[0m", Sugar_FormulaConvertToString(formula)); */
/*   Sugar_AutomatonPrint(A_left, 3); */
  SugarAutomatonFlush(A_left);
/*   fprintf(nusmv_stdout, "\033[22;32m\033[07m******************************************\n* flushed"); */
/*   Sugar_AutomatonPrint(A_left, 1); */
/*   fprintf(nusmv_stdout, "\n\033[0m"); */
  return A_left;
}

/**Function********************************************************************

Synopsis    [Remove alternation from alternating automaton.]

Description [Removes multiple initial states and all edges with multiple
destinations. The resulting automaton has a set of single initial states,
and only edges from source to one destination node. Labels of the new edges
are concatenated as neccessary. Uses algorithm of Miyano and Hayashi.]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     [Sugar_AutomatonToLtlAut]

******************************************************************************/
Sugar_Automaton_t*
Sugar_AutomatonExpand(
  Sugar_Automaton_t *aut)
{
  int index1;
  int index2;
  int index3;
  int index4;
  int is_accept;
  int not_found;
  char *tmpStr;
  char *dest_name;
  array_t *duplicate;
  array_t *composite;
  array_t *temp, *dup;
  array_t *disj_array;
  array_t *konj_array;
  array_t *state_array;
  array_t *label_array;
  array_t *dest_array;
  array_t *dest_acc_array;
  array_t *edge_base_array;
  array_t *current_states;
  array_t *current_acc_states;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *ini_node;
  Sugar_Vertex_t *acc_node;
  Sugar_Vertex_t *curr_node;
  Sugar_Vertex_t *new_node;
  Sugar_Automaton_t *new_aut;
  lsList node_queue;
  lsGen gen1, gen2;

  node_queue = lsCreate();
  new_aut = Sugar_AutomatonCreate();
  /* for all possible initial states in OR-list */
  arrayForEachItem(array_t*, aut->initial_states, index1, disj_array) {
#ifdef DEBUG_SUGAR
    printf("INITIAL NODES\n");
#endif
    /* create node sorted by index for each disjunct initial states */
    dest_name = util_strsav("NODE");
    dest_array = array_alloc(Sugar_Vertex_t*, 0);
    dest_acc_array = array_alloc(Sugar_Vertex_t*, 0);
    arrayForEachItem(Sugar_Vertex_t*, disj_array, index2, node) {
      not_found = 1;
      arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index3, acc_node) {
        if (node == acc_node) { not_found = 0; break; }
      }
      if (not_found) {
        array_insert_last(Sugar_Vertex_t*, dest_array, node);
      }
      else {
        array_insert_last(Sugar_Vertex_t*, dest_acc_array, node);
      }
    }
    /* create name from nodes: first, the non-accepting ones */
    array_sort(dest_array, SugarCompareNodeID);
    arrayForEachItem(Sugar_Vertex_t*, dest_array, index2, node) {
      tmpStr = util_strcat3(dest_name, "_", util_inttostr(node->id));
      FREE(dest_name); dest_name = tmpStr;
    }
    /* insert seperator for readability */
    tmpStr = util_strcat3(dest_name, "_|", "");
    FREE(dest_name); dest_name = tmpStr;
    /* second, the already accepting states */
    array_sort(dest_acc_array, SugarCompareNodeID);
    arrayForEachItem(Sugar_Vertex_t*, dest_acc_array, index2, node) {
      tmpStr = util_strcat3(dest_name, "_", util_inttostr(node->id));
      FREE(dest_name); dest_name = tmpStr;
    }
    /* create node and add to 'to-be-processed' queue */
    ini_node = SugarAutomatonNewNode(new_aut);
    ini_node->destination = dest_array;
    ini_node->dest_accept = dest_acc_array;
    ini_node->name = dest_name;
    lsNewEnd(node_queue, (lsGeneric)ini_node, 0);
    /* insert new node as initial state */
    state_array = array_alloc(Sugar_Vertex_t*, 1);
    array_insert(Sugar_Vertex_t*, state_array, 0, ini_node);
    array_insert_last(array_t*, new_aut->initial_states, state_array);
    /* check if initial state should be accepting */
    if (array_n(dest_array) == 0) {
#ifdef DEBUG_SUGAR
      printf("    ACCEPTING Initial Node '%s'\n", dest_name);
#endif
      array_insert_last(Sugar_Vertex_t*, new_aut->accept_states, ini_node);
      ini_node->destination = dest_acc_array; ini_node->dest_accept = dest_array;
    }
    /* run through all reachable states */
    while (lsLength(node_queue) > 0) {
      lsDelBegin(node_queue, (lsGeneric*)&curr_node);
      current_states = curr_node->destination;
      current_acc_states = curr_node->dest_accept;
#ifdef DEBUG_SUGAR
      printf("  NODE: 0x%08x  Id=%d  Name='%s'  (ToDo:%d)\n", curr_node, curr_node->id, curr_node->name, lsLength(node_queue));
#endif
      /* union all states (prevent duplicates) */
      state_array = array_dup(current_states);
      arrayForEachItem(Sugar_Vertex_t*, current_acc_states, index2, acc_node) {
        not_found = 1;
        arrayForEachItem(Sugar_Vertex_t*, state_array, index3, node) {
          if (node == acc_node) { not_found = 0; break; }
        }
        if (not_found) {
          array_insert_last(Sugar_Vertex_t*, state_array, acc_node);
        }
      }
      /* evaluate all new edges */
      edge_base_array = array_alloc(array_t*, 0);
      temp = array_alloc(Sugar_Edge_t*, 0);
      array_insert_last(array_t*, edge_base_array, temp);
      arrayForEachItem(Sugar_Vertex_t*, state_array, index2, node) {
        composite = array_alloc(array_t*, 0);
        lsForEachItem(node->edge_out_list, gen1, edge) {
          /* setup new base array for new outgoing edge */
          duplicate = array_alloc(array_t*, 0);
          arrayForEachItem(array_t*, edge_base_array, index3, temp) {
            dup = array_dup(temp);
            array_insert_last(array_t*, duplicate, dup);
          }
          /* add edge to all items of current AND-list */
          arrayForEachItem(array_t*, duplicate, index3, temp) {
            array_insert_last(Sugar_Edge_t*, temp, edge);
          }
          array_append(composite, duplicate);
          array_free(duplicate);
        }
        /* set new base */
        array_free(edge_base_array);
        edge_base_array = composite;
      }
      /* insert new edges to new non-alternating automaton */
      arrayForEachItem(array_t*, edge_base_array, index2, konj_array) {
        dest_name = util_strsav("NODE");
        dest_array = array_alloc(Sugar_Vertex_t*, 0);
        dest_acc_array = array_alloc(Sugar_Vertex_t*, 0);
        label_array = array_alloc(Sugar_Label_t*, 0);
        /* generate node name and merge labels (no duplicates) */
        arrayForEachItem(Sugar_Edge_t*, konj_array, index3, edge) {
          /* check if source node of edge is in accepting set */
          is_accept = 0;
          arrayForEachItem(Sugar_Vertex_t*, current_acc_states, index4, acc_node) {
            if (edge->source_vertex == acc_node) { is_accept = 1; break; }
          }
          /* all states go into accepting set */
          if (is_accept) {
            lsForEachItem(edge->dest_vertex_list, gen1, node) {
              not_found = 1;
              arrayForEachItem(Sugar_Vertex_t*, dest_acc_array, index4, acc_node) {
                if (node == acc_node) { not_found = 0; break; }
              }
              if (not_found) {
                array_insert_last(Sugar_Vertex_t*, dest_acc_array, node);
                /* check if we have a node that is both: accepting and non-accepting */
                arrayForEachItem(Sugar_Vertex_t*, current_states, index4, acc_node) {
                  if (edge->source_vertex == acc_node) {
                    /* yes, also keep this node in the set of non-accepting states */
                    array_insert_last(Sugar_Vertex_t*, dest_array, node);
                    break;
                  }
                }
              }
            }
          }
          else {
            /* for each destination node add to ... */
            lsForEachItem(edge->dest_vertex_list, gen1, node) {
              not_found = 1;
              arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index4, acc_node) {
                if (node == acc_node) { not_found = 0; break; }
              }
              if (not_found) {
                /* ... set of non-accepting states if node is not accepting */
                not_found = 1;
                arrayForEachItem(Sugar_Vertex_t*, dest_array, index4, acc_node) {
                  if (node == acc_node) { not_found = 0; break; }
                }
                if (not_found) {
                  array_insert_last(Sugar_Vertex_t*, dest_array, node);
                }
              }
              else {
                /* ... set of accepting states if node is accepting */
                not_found = 1;
                arrayForEachItem(Sugar_Vertex_t*, dest_acc_array, index4, acc_node) {
                  if (node == acc_node) { not_found = 0; break; }
                }
                if (not_found) {
                  array_insert_last(Sugar_Vertex_t*, dest_acc_array, node);
                }
              }
            }
          }
          /* create label */
          temp = SugarAutomatonMergeLabelArray(edge->label, label_array);
          arrayForEachItem(Sugar_Label_t*, label_array, index4, labl) FREE(labl);
          label_array = temp;
        }
        /* create name from nodes (first the non-accepting ones) */
        array_sort(dest_array, SugarCompareNodeID);
        arrayForEachItem(Sugar_Vertex_t*, dest_array, index3, node) {
          tmpStr = util_strcat3(dest_name, "_", util_inttostr(node->id));
          FREE(dest_name); dest_name = tmpStr;
        }
        tmpStr = util_strcat3(dest_name, "_|", "");
        FREE(dest_name); dest_name = tmpStr;
        array_sort(dest_acc_array, SugarCompareNodeID);
        arrayForEachItem(Sugar_Vertex_t*, dest_acc_array, index3, node) {
          tmpStr = util_strcat3(dest_name, "_", util_inttostr(node->id));
          FREE(dest_name); dest_name = tmpStr;
        }
        /* use node with the same name else ... */
        not_found = 1;
        lsForEachItem(new_aut->state_list, gen1, node) {
          if (strcmp(node->name, dest_name) == 0) {
            new_node = node; not_found = 0; break;
          }
        }
        if (not_found) {
          /* ... create new node and ... */
          new_node = SugarAutomatonNewNode(new_aut);
          new_node->destination = dest_array;
          new_node->dest_accept = dest_acc_array;
          new_node->name = dest_name;
          lsNewEnd(node_queue, (lsGeneric)new_node, 0);
          /* ... check for acceptance */
          if (array_n(dest_array) == 0) {
#ifdef DEBUG_SUGAR
            printf("    ACCEPTING Node '%s'\n", dest_name);
#endif
            array_insert_last(Sugar_Vertex_t*, new_aut->accept_states, new_node);
            new_node->destination = dest_acc_array; new_node->dest_accept = dest_array;
          }
        }
        /* add edge and clean up labels */
        temp = array_alloc(Sugar_Vertex_t*, 1);
        array_insert(Sugar_Vertex_t*, temp, 0, new_node);
        SugarAutomatonNewEdge(new_aut, curr_node, temp, label_array, 1);
        arrayForEachItem(Sugar_Label_t*, label_array, index3, labl) FREE(labl);
        array_free(label_array); array_free(temp);
      }
      /* node finished, go to next one */
      array_free(current_states);
      array_free(current_acc_states);
    }
    /* start with next initial states */
  }
  lsDestroy(node_queue, (void (*)(lsGeneric)) NULL);
  /* add labels from source automaton to new one */
  lsForEachItem(aut->prop_list, gen1, tmpStr) {
    dest_name = util_strsav(tmpStr);
    lsNewEnd(new_aut->prop_list, (lsGeneric)dest_name, 0);
    /* update edge labels of new automaton */
    lsForEachItem(new_aut->edge_list, gen2, edge) {
      arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
        if (labl->name != NIL(char)) {
          if (strcmp(labl->name, dest_name) == 0) { labl->name = dest_name; }
        }
      }
    }
  }
  /* free state names (not needed anymore) */
  lsForEachItem(new_aut->state_list, gen1, node) {
    FREE(node->name); node->name = NIL(char);
  }
  SugarAutomatonFlush(new_aut);
#ifdef DEBUG_SUGAR
  printf("AUT->Expanded DONE!\n");
#endif
  return new_aut;
}

/**Function********************************************************************

Synopsis    [Convert expanded automaton to deterministic automaton.]

Description [Transforms the alternating automaton into a deterministic
automaton. The automata will be tranformed with Sugar_AutomatonExpand().
This version uses no epsilon-edges, but expands the label terms.]

SideEffects [The returned automaton should be freed by the caller. Changes
destination, destacctpt, and name field of the vertices]

SeeAlso     [Sugar_AutomatonExpand]

******************************************************************************/
Sugar_Automaton_t*
Sugar_AutomatonDeterministic(
  Sugar_Automaton_t* aut)
{
  int index1;
  int index2;
  int index3;
  int not_found;
  char *tmpStr;
  char *dest_name;
  array_t *temp;
  array_t *new_labls;
  array_t *disj_array;
  array_t *label_array;
  array_t *new_dest_array;
  array_t *new_label_array;
  array_t *current_states;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Label_t *new_labl;
  Sugar_Label_t *dup_labl;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *ini_node;
  Sugar_Vertex_t *acc_node;
  Sugar_Vertex_t *curr_node;
  Sugar_Vertex_t *dest_node;
  Sugar_Vertex_t *new_node;
  Sugar_Automaton_t *new_aut;
  Sugar_Automaton_t *expanded_aut;
  Sugar_Array_Pair_t *item;
  Sugar_Array_Pair_t *new_item;
  Sugar_Array_Pair_t *dup_item;
  lsList node_queue;
  lsGen gen1, gen2;

  /* setup new automaton and 'ToDo'-list*/
  node_queue = lsCreate();
  new_aut = Sugar_AutomatonCreate();
  expanded_aut = Sugar_AutomatonExpand(aut);

  /* get all initial nodes */
  dest_name = util_strsav("NODE");
  new_dest_array = array_alloc(Sugar_Vertex_t*, 0);
  arrayForEachItem(array_t*, expanded_aut->initial_states, index1, disj_array) {
    node = array_fetch(Sugar_Vertex_t*, disj_array, 0);
    array_insert_last(Sugar_Vertex_t*, new_dest_array, node);
  }
  array_sort(new_dest_array, SugarCompareNodeID);
  arrayForEachItem(Sugar_Vertex_t*, new_dest_array, index1, node) {
    tmpStr = util_strcat3(dest_name, "_", util_inttostr(node->id));
    FREE(dest_name); dest_name = tmpStr;
  }
#ifdef DEBUG_SUGAR
  printf("INITIAL STATE\n");
#endif
  /* create single initial state */
  ini_node = SugarAutomatonNewNode(new_aut);
  ini_node->destination = new_dest_array;
  ini_node->name = dest_name;
  /* insert this new node as initial state and in 'ToDo'-list */
  temp = array_alloc(Sugar_Vertex_t*, 1);
  array_insert(Sugar_Vertex_t*, temp, 0, ini_node);
  array_insert_last(array_t*, new_aut->initial_states, temp);
  lsNewEnd(node_queue, (lsGeneric)ini_node, 0);
  /* check if initial state should be accepting */
  not_found = 1;
  arrayForEachItem(Sugar_Vertex_t*, new_dest_array, index1, node) {
    arrayForEachItem(Sugar_Vertex_t*, expanded_aut->accept_states, index2, acc_node) {
      if (node == acc_node) { not_found = 0; break; }
    }
    if (not_found == 0) break;
  }
  if (not_found == 0) {
    array_insert_last(Sugar_Vertex_t*, new_aut->accept_states, ini_node);
  }
  /* run through all reachable states */
  while (lsLength(node_queue) > 0) {
    lsDelBegin(node_queue, (lsGeneric*)&curr_node);
    current_states = curr_node->destination;
    curr_node->destination = NIL(array_t);
#ifdef DEBUG_SUGAR
    printf("  NODE: 0x%08x  Id=%d  Name='%s' (ToDo:%d)\n", curr_node, curr_node->id, curr_node->name, lsLength(node_queue));
#endif
    /* allocate empty data structures */
    label_array = array_alloc(Sugar_Array_Pair_t*, 1);
    item = ALLOC(Sugar_Array_Pair_t, 1);
    item->label_array = array_alloc(Sugar_Label_t*, 0);
    item->dest_array = array_alloc(Sugar_Vertex_t*, 0);
    array_insert(Sugar_Array_Pair_t*, label_array, 0, item);
    /* use TRUE label for labels as neutral element */
    new_labl = ALLOC(Sugar_Label_t, 1);
    new_labl->name = NIL(char); new_labl->isNegate = 0;
    array_insert_last(Sugar_Label_t*, item->label_array, new_labl);
    /* generate array of all possible labels */
    arrayForEachItem(Sugar_Vertex_t*, current_states, index1, node) {
      lsForEachItem(node->edge_out_list, gen1, edge) {
        /* add label of edge with destination node and its negation (without node) to all items */
        lsFirstItem(edge->dest_vertex_list, (lsGeneric*)&dest_node, NULL);
        new_label_array = array_alloc(Sugar_Array_Pair_t*, 0);
        arrayForEachItem(Sugar_Array_Pair_t*, label_array, index2, item) {
          /* clone current item (also the label structs!!) */
          dup_item = ALLOC(Sugar_Array_Pair_t, 1);
          dup_item->label_array = array_alloc(Sugar_Label_t*, 0);
          arrayForEachItem(Sugar_Label_t*, item->label_array, index3, labl) {
            new_labl = ALLOC(Sugar_Label_t, 1);
            new_labl->name = labl->name; new_labl->isNegate = labl->isNegate;
            array_insert_last(Sugar_Label_t*, dup_item->label_array, new_labl);
          }
          dup_item->dest_array = array_dup(item->dest_array);
          /* merge label of current item with label of edge and ... */
          new_labls = SugarAutomatonMergeLabelArray(item->label_array, edge->label);
          arrayForEachItem(Sugar_Label_t*, item->label_array, index3, labl) FREE(labl);
          array_free(item->label_array); item->label_array = new_labls;
          /* ... add destination node (without duplicates) */
          not_found = 1;
          arrayForEachItem(Sugar_Vertex_t*, item->dest_array, index3, node) {
            if (node == dest_node) { not_found = 0; break; }
          }
          if (not_found) {
            array_insert_last(Sugar_Vertex_t*, item->dest_array, dest_node);
          }
          array_insert_last(Sugar_Array_Pair_t*, new_label_array, item);
          /* insert negation of label of edge from cloned item */
          arrayForEachItem(Sugar_Label_t*, edge->label, index3, labl) {
            new_item = ALLOC(Sugar_Array_Pair_t, 1);
            /* clone and negate one literal from label of edge */
            dup_labl = ALLOC(Sugar_Label_t, 1);
            dup_labl->name = labl->name; dup_labl->isNegate = 1 - labl->isNegate;
            temp = array_alloc(Sugar_Label_t*, 1);
            array_insert(Sugar_Label_t*, temp, 0, dup_labl);
            /* merge label of current item with the negated literal and add new item */
            new_item->label_array = SugarAutomatonMergeLabelArray(dup_item->label_array, temp);
            new_item->dest_array = array_dup(dup_item->dest_array); array_free(temp);
            array_insert_last(Sugar_Array_Pair_t*, new_label_array, new_item);
          }
          /* free clone item data */
          arrayForEachItem(Sugar_Label_t*, dup_item->label_array, index3, labl) FREE(labl);
          array_free(dup_item->label_array); array_free(dup_item->dest_array); FREE(dup_item);
        }
        /* free old label array (we reuse old items) */
        array_free(label_array);
        label_array = array_alloc(Sugar_Array_Pair_t*, 0);
        arrayForEachItem(Sugar_Array_Pair_t*, new_label_array, index2, item) {
          /* if label is not FALSE add it to new list else delete the item */
          not_found = 1;
          if (array_n(item->label_array) == 1) {
            labl = array_fetch(Sugar_Label_t*, item->label_array, 0);
            if (IsLabelFalse(labl)) {
              not_found = 0;
              arrayForEachItem(Sugar_Label_t*, item->label_array, index3, labl) FREE(labl);
              array_free(item->label_array); array_free(item->dest_array); FREE(item);
            }
          }
          if (not_found) {
            array_insert_last(Sugar_Array_Pair_t*, label_array, item);
          }
        }
        array_free(new_label_array);
      }
    }
    /* run through all possible labels */
    arrayForEachItem(Sugar_Array_Pair_t*, label_array, index1, item) {
      /* if we have no destination nodes, skip this label */
      if (array_n(item->dest_array) > 0) {
        /* otheriwse create name from destination nodes sorted by index */
        dest_name = util_strsav("NODE");
        new_label_array = item->label_array;
        new_dest_array = item->dest_array;
        array_sort(new_dest_array, SugarCompareNodeID);
        arrayForEachItem(Sugar_Vertex_t*, new_dest_array, index2, node) {
          tmpStr = util_strcat3(dest_name, "_", util_inttostr(node->id));
          FREE(dest_name); dest_name = tmpStr;
        }
        /* use node with the same name else ... */
        not_found = 1;
        lsForEachItem(new_aut->state_list, gen1, node) {
          if (strcmp(node->name, dest_name) == 0) {
            new_node = node; not_found = 0; break;
          }
        }
        if (not_found) {
          /* ... create new node and ... */
          new_node = SugarAutomatonNewNode(new_aut);
          new_node->destination = array_dup(new_dest_array);
          new_node->name = dest_name;
          lsNewEnd(node_queue, (lsGeneric)new_node, 0);
          /* ... check for acceptance */
          not_found = 1;
          arrayForEachItem(Sugar_Vertex_t*, new_dest_array, index2, node) {
            arrayForEachItem(Sugar_Vertex_t*, expanded_aut->accept_states, index3, acc_node) {
              if (node == acc_node) { not_found = 0; break; }
            }
            if (not_found == 0) break;
          }
          if (not_found == 0) {
#ifdef DEBUG_SUGAR
            printf("    ACCEPTING Node '%s'\n", dest_name);
#endif
            array_insert_last(Sugar_Vertex_t*, new_aut->accept_states, new_node);
          }
        }
        /* add edge and ... */
        temp = array_alloc(Sugar_Vertex_t*, 1);
        array_insert(Sugar_Vertex_t*, temp, 0, new_node);
        SugarAutomatonNewEdge(new_aut, curr_node, temp, new_label_array, 1);
      }
      /* ... clean up temporary data of current item */
      arrayForEachItem(Sugar_Label_t*, item->label_array, index2, labl) FREE(labl);
      array_free(item->label_array); array_free(item->dest_array); FREE(item);
    }
    /* clean up temporary data and go to next node */
    array_free(label_array);
    array_free(current_states);
  }
  lsDestroy(node_queue, (void (*)(lsGeneric)) NULL);
  /* add labels from source automaton to new one */
  lsForEachItem(expanded_aut->prop_list, gen1, tmpStr) {
    dest_name = util_strsav(tmpStr);
    lsNewEnd(new_aut->prop_list, (lsGeneric)dest_name, 0);
    /* update edge labels of new automaton */
    lsForEachItem(new_aut->edge_list, gen2, edge) {
      arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
        if (labl->name != NIL(char)) {
          if (strcmp(labl->name, dest_name) == 0) { labl->name = dest_name; }
        }
      }
    }
  }
  /* free state names (not needed anymore) */
  lsForEachItem(new_aut->state_list, gen1, node) {
    FREE(node->name);
  }
#ifdef DEBUG_SUGAR
  printf("AUT->Deterministic DONE!\n");
#endif
  return new_aut;
}

/**Function********************************************************************

Synopsis    [Convert Sugar automaton to LTL automaton for model checking.]

Description [Transforms the expanded automaton into an LTL automaton. The
labels from the edges are moved into the states of LTL automata. A simple
explanation is that edges and states are exchanged. Alternating automata
must be tranformed with Sugar_AutomatonExpand(), otherwise the returned
automaton is incomplete or wrong!]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     [Sugar_AutomatonExpand]

******************************************************************************/
/* Ltl_Automaton_t* */
/* Sugar_AutomatonToLtlAutomaton( */
/*   Sugar_Automaton_t *aut) */
/* { */
/*   LtlSet_t* ltl_label; */
/*   vertex_t *vtx1, *vtx2; */
/*   Ltl_Automaton_t *ltl_aut; */
/*   Ltl_AutomatonNode_t *ltl_node; */
/*   Ctlsp_Formula_t *F_id, *notF_id; */

/*   int index1; */
/*   int index2; */
/*   int not_found; */
/*   char *name; */
/*   char *name_1; */
/*   char *name_id; */
/*   char *dest_name; */
/*   array_t *disj_array; */
/*   array_t *graph_array; */
/*   Sugar_Edge_t *edge; */
/*   Sugar_Vertex_t *node; */
/*   Sugar_Vertex_t *acc_node; */
/*   Sugar_Vertex_t *new_node; */
/*   Sugar_Vertex_t *curr_node; */
/*   Sugar_Vertex_t *dest_node; */
/*   Sugar_Automaton_t *expanded_aut; */
/*   st_table *fair_tbl; */
/*   lsList node_queue; */
/*   lsGen gen1; */
/*   fprintf(nusmv_stdout, "\033[22;32m\033[07m******************************************\n* inltl"); */
/*   Sugar_AutomatonPrint(aut, 1); */
/*   fprintf(nusmv_stdout, "\n\033[0m"); */

/*   remove alternation from automaton */
/*   ltl_aut = Ltl_AutomatonCreate(); */
/*   expanded_aut = Sugar_AutomatonExpand(aut); */
/* /\*   fprintf(nusmv_stdout, "\033[22;32m\033[07m******************************************\n* after expand"); *\/ */
/* /\*   Sugar_AutomatonPrint(expanded_aut, 3); *\/ */
/* /\*   fprintf(nusmv_stdout, "\n\033[0m"); *\/ */

/*   /\* create FALSE automaton if no accepting states exist *\/ */
/*   if (array_n(expanded_aut->accept_states) == 0) { */
/* #ifdef DEBUG_SUGAR */
/*     printf("Abort: No accepting states found!\n"); */
/*     Sugar_AutomatonPrint(expanded_aut, 3); */
/* #endif */
/*     /\* create label table with labels FALSE and TRUE *\/ */
/*     ltl_aut->labelTable = array_alloc(Ctlsp_Formula_t *, 2); */
/*     ltl_aut->labelTableNegate = array_alloc(int, 2); */
/*     F_id = Ctlsp_FormulaCreate(Ctlsp_FALSE_c, NIL(Ctlsp_Formula_t), NIL(Ctlsp_Formula_t)); */
/*     array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, 0, F_id); */
/*     array_insert(int, ltl_aut->labelTableNegate, 0, 1); */
/*     F_id = Ctlsp_FormulaCreate(Ctlsp_TRUE_c, NIL(Ctlsp_Formula_t), NIL(Ctlsp_Formula_t)); */
/*     array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, 1, F_id); */
/*     array_insert(int, ltl_aut->labelTableNegate, 1, 0); */
/*     /\* fill some tableau fields (declared as optional, but used everywhere ?!) *\/ */
/*     ltl_aut->tableau = LtlTableauCreate(); */
/*     ltl_aut->tableau->labelIndex = 2; */
/*     ltl_aut->tableau->labelTable = ALLOC(Ctlsp_Formula_t*, ltl_aut->tableau->labelIndex); */
/*     arrayForEachItem(Ctlsp_Formula_t*, ltl_aut->labelTable, index1, F_id) { */
/*       ltl_aut->tableau->labelTable[index1] = F_id; */
/*     } */
/*     ltl_aut->tableau->labelTableNegate = array_dup(ltl_aut->labelTableNegate); */
/*     /\* returning... *\/ */
/*     Sugar_AutomatonFree(expanded_aut); */
/*     return ltl_aut; */
/*   } */

/*   /\* initialize working data *\/ */
/*   node_queue = lsCreate(); */
/*   dest_name = ALLOC(char, 16); */
/*   graph_array = array_alloc(Sugar_Vertex_t*, 0); */
/*   fair_tbl = st_init_table(st_ptrcmp, st_ptrhash); */

/*   /\* create label table *\/ */
/*   index1 = lsLength(expanded_aut->prop_list); index2 = 2*index1 + 1; */
/*   ltl_aut->labelTable = array_alloc(Ctlsp_Formula_t *, index2+1); */
/*   ltl_aut->labelTableNegate = array_alloc(int, index2+1); */
/*   /\* insert TRUE *\/ */
/*   F_id = Ctlsp_FormulaCreate(Ctlsp_TRUE_c, NIL(Ctlsp_Formula_t), NIL(Ctlsp_Formula_t)); */
/*   array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, index1, F_id); */
/*   array_insert(int, ltl_aut->labelTableNegate, index1, index2); */
/*   /\* insert FALSE *\/ */
/*   F_id = Ctlsp_FormulaCreate(Ctlsp_FALSE_c, NIL(Ctlsp_Formula_t), NIL(Ctlsp_Formula_t)); */
/*   array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, index2, F_id); */
/*   array_insert(int, ltl_aut->labelTableNegate, index2, index1); */
/*   /\* insert propositionals *\/ */
/*   index2 = index1 + 1; index1 = 0; */
/*   lsForEachItem(expanded_aut->prop_list, gen1, name) { */
/*     /\* insert ID *\/ */
/*     name_1 = util_strsav("1"); name_id = util_strsav(name); */
/*     F_id = Ctlsp_FormulaCreate(Ctlsp_ID_c, name_id, name_1); */
/*     array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, index1, F_id); */
/*     array_insert(int, ltl_aut->labelTableNegate, index1, index2); */
/*     /\* insert not(ID) *\/ */
/*     name_1 = util_strsav("1"); name_id = util_strsav(name); */
/*     F_id = Ctlsp_FormulaCreate(Ctlsp_ID_c, name_id, name_1); */
/*     notF_id = Ctlsp_FormulaCreate(Ctlsp_NOT_c, F_id, NIL(Ctlsp_Formula_t)); */
/*     array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, index2, notF_id); */
/*     array_insert(int, ltl_aut->labelTableNegate, index2, index1); */
/*     index1++; index2++; */
/*   } */
/* #ifdef DEBUG_SUGAR */
/*   printf("LABELTABLE {\n"); */
/*   arrayForEachItem(Ctlsp_Formula_t*, ltl_aut->labelTable, index1, F_id) { */
/*     printf("%4d : ", index1); */
/*     Ctlsp_FormulaPrint(nusmv_stdout, F_id); */
/*     index2 = array_fetch(int, ltl_aut->labelTableNegate, index1); */
/*     printf("  (notF=%d)\n", index2); */
/*   } */
/*   printf("}\n"); */
/* #endif */

/*   /\* for all possible initial states in OR-list *\/ */
/*   arrayForEachItem(array_t*, expanded_aut->initial_states, index1, disj_array) { */
/* #ifdef DEBUG_SUGAR */
/*     printf("INITIAL NODE\n"); */
/* #endif */
/*     /\* read single initial node *\/ */
/*     node = array_fetch(Sugar_Vertex_t*, disj_array, 0); */
/*     lsForEachItem(node->edge_out_list, gen1, edge) { */
/*       /\* create initial state for LTL automaton *\/ */
/*       ltl_node = Ltl_AutomatonNodeCreate(ltl_aut); */
/*       ltl_node->Labels = SugarAutomatonLabelToLtlSet(ltl_aut, edge->label); */
/*       vtx1 = g_add_vertex(ltl_aut->G); */
/*       vtx1->user_data = (gGeneric)ltl_node; */
/*       st_insert(ltl_aut->Init, (char*)vtx1, (char*)vtx1); */
/*       /\* get destination node *\/ */
/*       lsFirstItem(edge->dest_vertex_list, (lsGeneric*)&dest_node, NULL); */
/*       sprintf(dest_name, "DNODE@%08x", dest_node); */
/*       /\* check if new initial state should also be accepting *\/ */
/*       arrayForEachItem(Sugar_Vertex_t*, expanded_aut->accept_states, index2, acc_node) { */
/*         if (dest_node == acc_node) { */
/*           st_insert(fair_tbl, (char*)vtx1, (char*)vtx1); */
/*           break; */
/*         } */
/*       } */
/*       /\* insert new node into to-be-processed queue *\/ */
/*       new_node = ALLOC(Sugar_Vertex_t, 1); */
/*       new_node->id = (int)vtx1; */
/*       new_node->edge_out_list = dest_node->edge_out_list; */
/*       new_node->name = util_strsav(dest_name); */
/*       lsNewEnd(node_queue, (lsGeneric)new_node, 0); */
/*       array_insert_last(Sugar_Vertex_t*, graph_array, new_node); */
/* #ifdef DEBUG_SUGAR */
/*       printf("  NODE: 0x%08x --- LTL-NODE: 0x%08x (id=%d) for q%d\n", new_node, vtx1, ltl_node->index, dest_node->id); */
/* #endif */
/*     } */
/*     /\* run through all reachable states *\/ */
/*     while (lsLength(node_queue) > 0) { */
/*       /\* read next node and its source node in ltl graph *\/ */
/*       lsDelBegin(node_queue, (lsGeneric*)&curr_node); */
/*       vtx1 = (vertex_t*)curr_node->id; */
/* #ifdef DEBUG_SUGAR */
/*       printf("    Read NODE: 0x%08x  Name='%s'  LTL-SRC:0x%08x\n", curr_node, curr_node->name, vtx1); */
/* #endif */
/*       if (lsLength(curr_node->edge_out_list) > 0) { */
/*         lsForEachItem(curr_node->edge_out_list, gen1, edge) { */
/*           /\* read the one (and only) destination node *\/ */
/*           lsFirstItem(edge->dest_vertex_list, (lsGeneric*)&dest_node, NULL); */
/*           sprintf(dest_name, "DNODE@%08x", dest_node); */
/*           /\* search in LTL graph for already existing node *\/ */
/*           not_found = 1; */
/*           ltl_label = SugarAutomatonLabelToLtlSet(ltl_aut, edge->label); */
/*           arrayForEachItem(Sugar_Vertex_t*, graph_array, index2, node) { */
/*             if (strcmp(node->name, dest_name) == 0) { */
/*               vtx2 = (vertex_t *)node->id; */
/*               ltl_node = (Ltl_AutomatonNode_t*)vtx2->user_data; */
/*               if (LtlSetEqual(ltl_node->Labels, ltl_label)) { */
/*                 /\* Okay, node has same destination and label *\/ */
/*                 not_found = 0; break; */
/*               } */
/*             } */
/*           } */
/*           if (not_found) { */
/*             /\* create new node for LTL automaton *\/ */
/*             ltl_node = Ltl_AutomatonNodeCreate(ltl_aut); */
/*             ltl_node->Labels = ltl_label; */
/*             vtx2 = g_add_vertex(ltl_aut->G); */
/*             vtx2->user_data = (gGeneric)ltl_node; */
/*             /\* check if new LTL node should be accepting *\/ */
/*             arrayForEachItem(Sugar_Vertex_t*, expanded_aut->accept_states, index2, acc_node) { */
/*               if (dest_node == acc_node) { */
/*                 st_insert(fair_tbl, (char*)vtx2, (char*)vtx2); */
/*                 break; */
/*               } */
/*             } */
/*             /\* insert new node into to-be-processed queue *\/ */
/*             new_node = ALLOC(Sugar_Vertex_t, 1); */
/*             new_node->id = (int)vtx2; */
/*             new_node->edge_out_list = dest_node->edge_out_list; */
/*             new_node->name = util_strsav(dest_name); */
/*             lsNewEnd(node_queue, (lsGeneric)new_node, 0); */
/*             array_insert_last(Sugar_Vertex_t*, graph_array, new_node); */
/* #ifdef DEBUG_SUGAR */
/*             printf("      New NODE: 0x%08x --- LTL-NODE: 0x%08x (id=%d) for q%d\n", new_node, vtx2, ltl_node->index, dest_node->id); */
/* #endif */
/*           } */
/*           /\* add edge to LTL automaton *\/ */
/*           g_add_edge(vtx1, vtx2); */
/* #ifdef DEBUG_SUGAR */
/*           printf("      Add Edge: 0x%08x to 0x%08x\n", vtx1, vtx2); */
/* #endif */
/*         } */
/*       } */
/*       else { */
/*         ltl_node = Ltl_AutomatonNodeCreate(ltl_aut); */
/*         ltl_node->Labels = LtlSetNew(0); */
/*         vtx2 = g_add_vertex(ltl_aut->G); */
/*         vtx2->user_data = (gGeneric)ltl_node; */
/*         g_add_edge(vtx1, vtx2); */
/*       } */
/*     } */
/*   } */
/*   /\* if accepting states added to fair set use it, else ... *\/ */
/*   if (st_count(fair_tbl) > 0) { */
/*     lsNewEnd(ltl_aut->Fair, (lsGeneric)fair_tbl, (lsHandle*)0); */
/*   } */
/*   else { */
/* #ifdef DEBUG_SUGAR */
/*     printf("Kill: Empty fair set detected!\n"); */
/* #endif */
/*     /\* ... delete generated ltl automaton and create empty one *\/ */
/*     Ltl_AutomatonFree((gGeneric)ltl_aut); */
/*     ltl_aut = Ltl_AutomatonCreate(); */
/*     /\* create label table with labels FALSE and TRUE *\/ */
/*     ltl_aut->labelTable = array_alloc(Ctlsp_Formula_t *, 2); */
/*     ltl_aut->labelTableNegate = array_alloc(int, 2); */
/*     F_id = Ctlsp_FormulaCreate(Ctlsp_FALSE_c, NIL(Ctlsp_Formula_t), NIL(Ctlsp_Formula_t)); */
/*     array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, 0, F_id); */
/*     array_insert(int, ltl_aut->labelTableNegate, 0, 1); */
/*     F_id = Ctlsp_FormulaCreate(Ctlsp_TRUE_c, NIL(Ctlsp_Formula_t), NIL(Ctlsp_Formula_t)); */
/*     array_insert(Ctlsp_Formula_t*, ltl_aut->labelTable, 1, F_id); */
/*     array_insert(int, ltl_aut->labelTableNegate, 1, 0); */
/*   } */

/*   /\* fill some tableau fields (declared as optional, but used everywhere ?!) *\/ */
/*   ltl_aut->tableau = LtlTableauCreate(); */
/*   ltl_aut->tableau->labelIndex = array_n(ltl_aut->labelTable); */
/*   ltl_aut->tableau->labelTable = ALLOC(Ctlsp_Formula_t*, ltl_aut->tableau->labelIndex); */
/*   arrayForEachItem(Ctlsp_Formula_t*, ltl_aut->labelTable, index1, F_id) { */
/*     ltl_aut->tableau->labelTable[index1] = F_id; */
/*   } */
/*   ltl_aut->tableau->labelTableNegate = array_dup(ltl_aut->labelTableNegate); */

/*   /\* free additional data *\/ */
/*   Sugar_AutomatonFree(expanded_aut); */
/*   lsDestroy(node_queue, (void (*)(lsGeneric)) NULL); */
/*   arrayForEachItem(Sugar_Vertex_t*, graph_array, index1, node) { */
/*     FREE(node->name); */
/*     FREE(node); */
/*   } */
/*   array_free(graph_array); */
/*   FREE(dest_name); */
/* #ifdef DEBUG_SUGAR */
/*   printf("AUT->LTL DONE!\n"); */
/* #endif */
/*   return ltl_aut; */
/* } */

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

Synopsis    [Create LTL label from Sugar edge label.]

Description [Create a new label set from the Sugar label. It uses the label
table of the LTL automaton. The label must come from the expanded automaton
that was used to create the label table and the LTL automaton.]

SideEffects [The returned label set should be freed by the caller.]

SeeAlso     []

******************************************************************************/
/* LtlSet_t* */
/* SugarAutomatonLabelToLtlSet( */
/*   Ltl_Automaton_t* ltl_aut, */
/*   array_t* label_array) */
/* { */
/*   int idx; */
/*   int index1; */
/*   int index2; */
/*   char *name; */
/*   LtlSet_t *vset; */
/*   Sugar_Label_t* labl; */
/*   Ctlsp_Formula_t *F; */

/*   vset = LtlSetNew(array_n(ltl_aut->labelTable)); */
/*   arrayForEachItem(Sugar_Label_t*, label_array, index1, labl) { */
/*     if (labl->name != NIL(char)) { */
/*       arrayForEachItem(Ctlsp_Formula_t*, ltl_aut->labelTable, index2, F) { */
/*         name = (char*)(F->left); */
/*         if (name != NIL(char)) { */
/*           if (strcmp(labl->name, name) == 0) { */
/*             if (labl->isNegate) */
/*               idx = array_fetch(int, ltl_aut->labelTableNegate, index2); */
/*             else */
/*               idx = index2; */
/*             var_set_set_elt(vset->e, idx); */
/*             break; */
/*           } */
/*         } */
/*       } */
/*     } */
/*   } */
/*   return vset; */
/* } */

/**Function********************************************************************

Synopsis    [Create new node for automaton.]

Description [Add a new node to the alternating automaton. The node gets
deleted with the automaton or use SugarAutomatonFreeNode().]

SideEffects []

SeeAlso     [Sugar_AutomatonFree, SugarAutomatonFreeNode]

******************************************************************************/
Sugar_Vertex_t*
SugarAutomatonNewNode(
  Sugar_Automaton_t *aut)
{
  Sugar_Vertex_t *node;
  lsHandle handle;

  /* allocate node structure */
  node = ALLOC(Sugar_Vertex_t, 1);
  lsNewEnd(aut->state_list, (lsGeneric)node, (lsHandle*)&handle);
  /* initialize required fields */
  node->id             = aut->num_state_index++;
  node->edge_in_list   = lsCreate();
  node->edge_out_list  = lsCreate();
  node->handle         = handle;
  /* clear optional fields */
  node->destination    = NIL(array_t);
  node->name           = NIL(char);

  return node;
}

/**Function********************************************************************

Synopsis    [Removes node from automaton.]

Description [Delete all incomming edges with SugarAutomatonFreeEdge() before
deleting the node. If there are incoming edges the node will not be deleted.
Outgoing edges will be deleted. Does not care about the optional data for
expanding alternation.]

SideEffects []

SeeAlso     [SugarAutomatonNewNode, SugarAutomatonFreeEdge]

******************************************************************************/
void
SugarAutomatonFreeNode(
  Sugar_Vertex_t *node)
{
  Sugar_Edge_t *edge;
  Sugar_Vertex_t *dummy;
  lsGen gen;

  /* no incoming edges allowed */
  if (lsLength(node->edge_in_list) == 0) {
    /* remove all outgoing edges */
    lsForEachItem(node->edge_out_list, gen, edge) {
      SugarAutomatonFreeEdge(edge);
    }
    /* delete additional data and struct */
    lsDestroy(node->edge_in_list, (void (*)(lsGeneric)) NULL);
    lsDestroy(node->edge_out_list, (void (*)(lsGeneric)) NULL);
    lsRemoveItem(node->handle, (lsGeneric*)&dummy);
#ifdef DEBUG_SUGAR
#if 0
    printf("FreeNode: %s\n", (node==dummy)?("Pointer OK"):("Wrong Pointer!!!"));
#endif
#endif
    FREE(node);
  }
#ifdef DEBUG_SUGAR
  else {
    printf("Edge IN-LIST not empty! (%d elements left)\n", lsLength(node->edge_in_list));
    lsForEachItem(node->edge_in_list, gen, edge) { printf("  0x%08x\n", edge); }
  }
#endif
}

/**Function********************************************************************

Synopsis    [Create new edge for automaton.]

Description [First check if an edge with the same source node, destination
nodes and label exists. If not it creates one. If the label is empty, the
edge will be labeled with TRUE.]

SideEffects []

SeeAlso     [SugarAutomatonFreeEdge]

******************************************************************************/
Sugar_Edge_t*
SugarAutomatonNewEdge(
  Sugar_Automaton_t *aut,
  Sugar_Vertex_t *source,
  array_t *dest,
  array_t *label,
  int checkUnique)
{
  int count;
  int index1;
  int index2;
  int is_equal;
  int not_found;
  array_t* new_dest;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl1;
  Sugar_Label_t *labl2;
  Sugar_Vertex_t *node1;
  Sugar_Vertex_t *node2;
  lsGen gen1, gen2;
  lsHandle handle;

  /* remove multiple destination nodes */
  if (checkUnique > 1) {
    new_dest = array_alloc(Sugar_Vertex_t*, 0);
    arrayForEachItem(Sugar_Vertex_t*, dest, index1, node1) {
      not_found = 1;
      arrayForEachItem(Sugar_Vertex_t*, new_dest, index2, node2) {
        if (node1 == node2) {
          //printf("Duplicate destination node 'q%d'\n", node2->id);
          not_found = 0; break;
        }
      }
      if (not_found) {
        array_insert_last(Sugar_Vertex_t*, new_dest, node1);
      }
    }
  }
  else {
    new_dest = array_dup(dest);
  }
  /* search if edge already exists */
  if (checkUnique > 0) {
    /* for all existing edges */
    lsForEachItem(aut->edge_list, gen1, edge) {
      /* check if source node is equal */
      if (edge->source_vertex == source) {
        /* check if all destination nodes are equal; compare ALL pairs */
        if (array_n(new_dest) == lsLength(edge->dest_vertex_list)) {
          count = 0;
          arrayForEachItem(Sugar_Vertex_t*, new_dest, index1, node1) {
            lsForEachItem(edge->dest_vertex_list, gen2, node2) {
              if (node1 == node2) count++;
            }
          }
          if (count == array_n(new_dest)) {
            /* check if label names are equal */
            is_equal = 1;
            if (array_n(label) == array_n(edge->label)) {
              count = 0;
              /* do not break out these loops; we MUST compare all pairs */
              arrayForEachItem(Sugar_Label_t*, label, index1, labl1) {
                arrayForEachItem(Sugar_Label_t*, edge->label, index2, labl2) {
                  if ((labl1->name == NIL(char)) && (labl2->name == NIL(char))) {
                    if (labl1->isNegate != labl2->isNegate) { is_equal = 0; break; }
                  }
                  else if ((labl1->name != NIL(char)) && (labl2->name != NIL(char))) {
                    if (strcmp(labl1->name, labl2->name) == 0) {
                      if (labl1->isNegate == labl2->isNegate) count++; else is_equal = 0;
                    }
                    else {
                      is_equal = 0;
                    }
                  }
                  else {
                    is_equal = 0;
                  }
                }
              }
              /* if so, then edge already exists and abort function */
              if (is_equal || (array_n(label) == count)) {
                //printf("Edge already exists!\n");
                return NIL(Sugar_Edge_t);
              }
            }
          }
        }
      }
    }
  }
  /* allocate edge and add to automaton */
  edge = ALLOC(Sugar_Edge_t, 1);
  lsNewEnd(aut->edge_list, (lsGeneric)edge, &handle);
  /* clone labels (or set to TRUE) ... */
  edge->label = array_alloc(Sugar_Label_t*, 0);
  if (array_n(label) > 0) {
    arrayForEachItem(Sugar_Label_t*, label, index1, labl1) {
      labl2 = ALLOC(Sugar_Label_t, 1);
      labl2->name = labl1->name;
      labl2->isNegate = labl1->isNegate;
      array_insert_last(Sugar_Label_t*, edge->label, labl2);
    }
  }
  else {
    labl2 = ALLOC(Sugar_Label_t, 1);
    labl2->name = NIL(char); labl2->isNegate = 0;
    array_insert_last(Sugar_Label_t*, edge->label, labl2);
  }
  /* ... and initialize fields */
  edge->source_vertex = source;
  edge->dest_vertex_list = lsCreate();
  edge->handle = handle;
  /* set source/destination and update nodes */
  lsNewEnd(source->edge_out_list, (lsGeneric)edge, 0);
  arrayForEachItem(Sugar_Vertex_t*, new_dest, index1, node1) {
    lsNewEnd(edge->dest_vertex_list, (lsGeneric)node1, 0);
    lsNewEnd(node1->edge_in_list, (lsGeneric)edge, 0);
  }
  array_free(new_dest);
  return edge;
}

/**Function********************************************************************

Synopsis    [Removed edge from automaton.]

Description []

SideEffects [Update lists of source and destination nodes.]

SeeAlso     [SugarAutomatonNewEdge]

******************************************************************************/
void
SugarAutomatonFreeEdge(
  Sugar_Edge_t *edge)
{
  int index;
  Sugar_Edge_t *search;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  lsGen gen_loop;
  lsGen generator;
  lsHandle handle;

  /* ????
   * Must be done here, because lsRemoveItem() destroys the pointer somehow
   */
  arrayForEachItem(Sugar_Label_t*, edge->label, index, labl) {
    FREE(labl);
  }
  array_free(edge->label);
  /* remove edge from source vertex outgoing list */
  generator = lsStart(edge->source_vertex->edge_out_list);
  lsFirstItem(edge->source_vertex->edge_out_list, (lsGeneric*)&search, (lsHandle*)&handle);
  while (lsNext(generator, (lsGeneric*)&search, (lsHandle*)&handle) == LS_OK) {
    if (search == edge) {
      //printf("Deleting from OUT list: item=0x%08x, handle=0x%08x\n", search, handle);
      lsRemoveItem(handle, (lsGeneric*)search);
      break;
    }
  }
  lsFinish(generator);
  /* remove edge from incoming list of all destination vertices */
  lsForEachItem(edge->dest_vertex_list, gen_loop, node) {
    generator = lsStart(node->edge_in_list);
    lsFirstItem(node->edge_in_list, (lsGeneric*)&search, (lsHandle*)&handle);
    while (lsNext(generator, (lsGeneric*)&search, (lsHandle*)&handle) == LS_OK) {
      if (search == edge) {
        //printf("Deleting from IN list:  item=0x%08x, handle=0x%08x\n", search, handle);
        lsRemoveItem(handle, (lsGeneric*)search);
        break;
      }
    }
    lsFinish(generator);
  }
  /* delete additional data and edge struct */
  lsDestroy(edge->dest_vertex_list, (void (*)(lsGeneric)) NULL);
  lsRemoveItem(edge->handle, (lsGeneric*)&search);
#ifdef DEBUG_SUGAR
#if 0
  printf("FreeEdge: %s\n", (edge==search)?("Pointer OK"):("Wrong Pointer!!!"));
#endif
#endif
  FREE(edge);
}

/**Function********************************************************************

Synopsis    [Adds TRUE-loop to node of an automaton.]

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
Sugar_Edge_t*
SugarAutomatonNewEdgeLoop(
  Sugar_Automaton_t *aut,
  Sugar_Vertex_t *node)
{
  array_t *dest_array;
  array_t *prop_array;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;

  /* setup TRUE label */
  labl = ALLOC(Sugar_Label_t, 1);
  labl->name = NIL(char); labl->isNegate = 0;
  /* add TRUE loop to node */
  dest_array = array_alloc(Sugar_Vertex_t*, 1);
  array_insert(Sugar_Vertex_t*, dest_array, 0, node);
  prop_array = array_alloc(Sugar_Label_t*, 1);
  array_insert(Sugar_Label_t*, prop_array, 0, labl);
  edge = SugarAutomatonNewEdge(aut, node, dest_array, prop_array, 1);
  array_free(dest_array);
  array_free(prop_array);
  FREE(labl);
  return edge;
}

/**Function********************************************************************

Synopsis    [Add aborting edges to node of an automaton.]

Description [Read the labels and destinations of the node and create new
edges that negates the formula for the outgoing edges. A node with one
edge labeled 'a&b' will get two new edges to the abort state. One labeled
'!a' and one labeled '!b'. A node with two edges, labeled with 'a' and 'b'
will get one new edge labeled '!a&!b'. Any possible combination of labels
and outgoing edges are handled.]

SideEffects []

SeeAlso     []

******************************************************************************/
void
SugarAutomatonAddAbortEdge(
  Sugar_Automaton_t *aut,
  Sugar_Vertex_t *node,
  Sugar_Vertex_t *abort_node)
{
  int index1;
  int index2;
  int index3;
  array_t *base;
  array_t *temp, *dup;
  array_t *composite;
  array_t *duplicate;
  array_t *dest_array;
  array_t *label_array;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Label_t *labl2;
  Sugar_Label_t *new_labl;
  lsGen gen;

  /* for all outgoing edges of node */
  base = array_alloc(array_t*, 0);
  temp = array_alloc(Sugar_Label_t*, 0);
  array_insert_last(array_t*, base, temp);
  lsForEachItem(node->edge_out_list, gen, edge) {
    composite = array_alloc(array_t*, 0);
    arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
      /* clone base array with labels for new labels */
      duplicate = array_alloc(array_t*, 0);
      arrayForEachItem(array_t*, base, index2, temp) {
        dup = array_alloc(Sugar_Label_t*, 0);
        arrayForEachItem(Sugar_Label_t*, temp, index3, labl2) {
          new_labl = ALLOC(Sugar_Label_t, 1);
          new_labl->name = labl2->name; new_labl->isNegate = labl2->isNegate;
          array_insert_last(Sugar_Label_t*, dup, new_labl);
        }
        array_insert_last(array_t*, duplicate, dup);
      }
      /* add labels to all items of current OR-list */
      arrayForEachItem(array_t*, duplicate, index2, temp) {
        /* clone current label and negate for merging */
        label_array = array_alloc(Sugar_Label_t*, 1);
        new_labl = ALLOC(Sugar_Label_t, 1);
        new_labl->name = labl->name;
        new_labl->isNegate = 1 - labl->isNegate;
        array_insert(Sugar_Label_t*, label_array, 0, new_labl);
        /* merge and replace old array with result */
        dup = SugarAutomatonMergeLabelArray(temp, label_array);
        array_insert(array_t*, duplicate, index2, dup);
        arrayForEachItem(Sugar_Label_t*, temp, index3, labl2) FREE(labl2);
        array_free(temp); array_free(label_array); FREE(new_labl);
      }
      /* add new label array to current OR-list */
      array_append(composite, duplicate);
      array_free(duplicate);
    }
    /* free old base and use composite as new base for OR-list */
    arrayForEachItem(array_t*, base, index1, label_array) {
      arrayForEachItem(Sugar_Label_t*, label_array, index2, labl) FREE(labl);
      array_free(label_array);
    }
    array_free(base);
    base = composite;
  }
  /* add new edge for each item in OR-list */
  dest_array = array_alloc(Sugar_Vertex_t*, 1);
  array_insert(Sugar_Vertex_t*, dest_array, 0, abort_node);
  arrayForEachItem(array_t*, base, index1, temp) {
    SugarAutomatonNewEdge(aut, node, dest_array, temp, 1);
    /* delete all cloned label and arrays */
    arrayForEachItem(Sugar_Label_t*, temp, index2, labl) FREE(labl);
    array_free(temp);
  }
  array_free(base);
}

/**Function********************************************************************

Synopsis    [Deletes FALSE edges and unreachable states.]

Description [First goes through all edges and delete the ones with label
FALSE. Second goes thorugh all nodes and delete the ones with no incoming
edges, except initial states.]

SideEffects [The automaton shrinks.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonFlush(
  Sugar_Automaton_t *aut)
{
  int index1;
  int index2;
  int count = 0;
  array_t *disj_array;
  array_t *new_acc_array;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *acc_node;
  Sugar_Vertex_t *curr_node;
  lsList node_queue;
  lsGen gen1, gen2;

  /* first try to minimize DNF of initial states */
  aut->initial_states = SugarAutomatonMinimizeTermArray(aut->initial_states);

  /* delete edges with label FALSE from automaton before returning it */
  lsForEachItem(aut->edge_list, gen1, edge) {
    arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
      if ((labl->name == NIL(char)) && (labl->isNegate == 1)) {
        SugarAutomatonFreeEdge(edge); count++; break;
      }
    }
  }
#ifdef DEBUG_SUGAR
  if (count) {
    printf("Removed %d edge%s with label FALSE\n", count, (count!=1)?("s"):(""));
  }
#endif
  /* make all node unvisited to ...*/
  lsForEachItem(aut->state_list, gen1, node) { node->name = (char*)1; }
  /* ... remove unreachable states with reachablility check */
  node_queue = lsCreate();
  arrayForEachItem(array_t*, aut->initial_states, index1, disj_array) {
    /* insert all unmarked initial nodes in 'ToDo'-list and mark it */
    arrayForEachItem(Sugar_Vertex_t*, disj_array, index2, node) {
      if (node->name) {
        lsNewEnd(node_queue, (lsGeneric)node, 0);
        node->name = NIL(char);
      }
    }
    /* run through all reachable states */
    while (lsLength(node_queue) > 0) {
      lsDelBegin(node_queue, (lsGeneric*)&curr_node);
      /* insert all unmarked destination nodes and mark it */
      lsForEachItem(curr_node->edge_out_list, gen1, edge) {
        lsForEachItem(edge->dest_vertex_list, gen2, node) {
          if (node->name) {
            lsNewEnd(node_queue, (lsGeneric)node, 0);
            node->name = NIL(char);
          }
        }
      }
    }
  }
  lsDestroy(node_queue, (void (*)(lsGeneric)) NULL);
  /* now delete all unmarked nodes with edges */
  count = 0;
  lsForEachItem(aut->state_list, gen1, node) {
    if (node->name) {
      /* remove node from accepting states */
      new_acc_array = array_alloc(Sugar_Vertex_t*, 0);
      arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index1, acc_node) {
        if (node != acc_node) {
          array_insert_last(Sugar_Vertex_t*, new_acc_array, acc_node);
        }
      }
      array_free(aut->accept_states);
      aut->accept_states = new_acc_array;
      /* delete all incoming edges, then delete node */
      lsForEachItem(node->edge_in_list, gen2, edge) {
        SugarAutomatonFreeEdge(edge);
      }
      SugarAutomatonFreeNode(node); count++;
    }
  }
#ifdef DEBUG_SUGAR
  if (count) {
    printf("Removed %d unreachable state%s!\n", count, (count!=1)?("s"):(""));
  }
#endif
}

/**Function********************************************************************

Synopsis    [Moves states, propositions and edges from the right automaton
to the left automaton.]

Description [Copys only references from the lists. Identical propositional
names are updated to the names of the new automaton.]

SideEffects [The left automaton grows, right automaton gets deleted.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonMoveRightToLeft(
  Sugar_Automaton_t *aut_left,
  Sugar_Automaton_t *aut_right)
{
  int index;
  char *old_name;
  char *new_name;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  lsGen gen1, gen2;
  lsHandle handle;

  /* add states from aut_right to aut_left */
  lsForEachItem(aut_right->state_list, gen1, node) {
    lsNewEnd(aut_left->state_list, (lsGeneric)node, &handle);
    node->handle = handle;
    node->id = aut_left->num_state_index;
    aut_left->num_state_index++;
  }
  /* add labels from aut_right to aut_left */
  lsForEachItem(aut_right->prop_list, gen1, old_name) {
    index = 1;
    lsForEachItem(aut_left->prop_list, gen2, new_name) {
      if (strcmp(old_name, new_name) == 0) { index = 0; break; }
    }
    if (index) {
      lsNewEnd(aut_left->prop_list, (lsGeneric)old_name, 0);
    }
  }
  /* add edges from aut_right to aut_left */
  lsForEachItem(aut_right->edge_list, gen1, edge) {
    lsForEachItem(aut_left->prop_list, gen2, new_name) {
      arrayForEachItem(Sugar_Label_t*, edge->label, index, labl) {
        if (labl->name != NIL(char)) {
          if (strcmp(labl->name, new_name) == 0) { labl->name = new_name; }
        }
      }
    }
    lsNewEnd(aut_left->edge_list, (lsGeneric)edge, &handle);
    edge->handle = handle;
  }
  /* clean up: delete right automata */
  lsDestroy(aut_right->state_list, (void (*)(lsGeneric)) NULL);
  lsDestroy(aut_right->prop_list, (void (*)(lsGeneric)) NULL);
  lsDestroy(aut_right->edge_list, (void (*)(lsGeneric)) NULL);
  FREE(aut_right);
}

/**Function********************************************************************

Synopsis    [Merges labels from source array and destination array]

Description [Create labels from source and destination without duplicates.
Also filters TRUE labels and if one label is FALSE it returns array with
label FALSE.]

SideEffects [The returned array of labels should be freed by the caller.]

SeeAlso     []

******************************************************************************/
array_t*
SugarAutomatonMergeLabelArray(
  array_t *label_src_array,
  array_t *label_dst_array)
{
  int index1;
  int index2;
  int add_to_array;
  int set_false_array = 0;
  Sugar_Label_t *labl_src;
  Sugar_Label_t *labl_dst;
  Sugar_Label_t *new_labl;
  array_t *new_dest_array;

  /* clone destination array and all labels; we add new labels to it */
  new_dest_array = array_alloc(Sugar_Label_t*, 0);
  arrayForEachItem(Sugar_Label_t*, label_dst_array, index1, labl_dst) {
    new_labl = ALLOC(Sugar_Label_t, 1);
    new_labl->name = labl_dst->name;
    new_labl->isNegate = labl_dst->isNegate;
    array_insert_last(Sugar_Label_t*, new_dest_array, new_labl);
  }
  /* for each label in source array */
  arrayForEachItem(Sugar_Label_t*, label_src_array, index1, labl_src) {
    /* special case TRUE and FALSE */
    if (labl_src->name == NIL(char)) {
      if (labl_src->isNegate) {
        set_false_array = 1; break;
      }
      else {
        arrayForEachItem(Sugar_Label_t*, new_dest_array, index2, labl_dst) {
          if (labl_dst->name == NIL(char)) {
            if (labl_dst->isNegate) { set_false_array = 1; break; }
          }
        }
      }
    }
    else {
      /* search destination if label already there */
      add_to_array = 1;
      arrayForEachItem(Sugar_Label_t*, new_dest_array, index2, labl_dst) {
        /* again check TRUE and FALSE case */
        if (labl_dst->name == NIL(char)) {
          add_to_array = 0;
          if (labl_dst->isNegate) {
            set_false_array = 1; break;
          }
          else {
            labl_dst->name = labl_src->name;
            labl_dst->isNegate = labl_src->isNegate;
	  }
        }
        else {
          /* don't add if label names are equal; check for (a && !a) case */
          if (strcmp(labl_src->name, labl_dst->name) == 0) {
            if (labl_src->isNegate != labl_dst->isNegate) set_false_array = 1;
            add_to_array = 0; break;
          }
        }
      }
      if (add_to_array) {
        /* clone source label and add to destination array */
        new_labl = ALLOC(Sugar_Label_t, 1);
        new_labl->name = labl_src->name;
        new_labl->isNegate = labl_src->isNegate;
        array_insert_last(Sugar_Label_t*, new_dest_array, new_labl);
      }
    }
    if (set_false_array) break;
  }
  /* check if label should be FALSE */
  if (set_false_array) {
    /* free destination labels */
    arrayForEachItem(Sugar_Label_t*, new_dest_array, index1, labl_dst) {
      FREE(labl_dst);
    }
    array_free(new_dest_array);
    /* allocate new label with FALSE */
    labl_dst = ALLOC(Sugar_Label_t, 1);
    labl_dst->name = NIL(char); labl_dst->isNegate = 1;
    new_dest_array = array_alloc(Sugar_Label_t*, 1);
    array_insert(Sugar_Label_t*, new_dest_array, 0, labl_dst);
  }
  return new_dest_array;
}

/**Function********************************************************************

Synopsis    [Removes unneccessary terms from DNF.]

Description [Removes all terms that are covered by other, shorter terms.
The input array must be an array of terms, where a term is an array of
pointer to the vertex structures. From formula (q_1 && q_6) || (q_6) the
first term (q_1 && q_6) is not required and will be deleted.]

SideEffects [The input array gets deleted; the returned array should be
freed by the caller.]

SeeAlso     [SugarAutomatonNegate]

******************************************************************************/
array_t*
SugarAutomatonMinimizeTermArray(
  array_t* term_array)
{
  int index1;
  int index2;
  int index3;
  int index4;
  int counter;
  int not_found;
  array_t *temp;
  array_t *duplicate;
  array_t *disj_array;
  array_t *dest_array;
  array_t *new_term_array;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *dest_node;

  /* compare all terms with each other */
  duplicate = array_alloc(array_t*, 0);
  arrayForEachItem(array_t*, term_array, index1, disj_array) {
    not_found = 0;
    arrayForEachItem(array_t*, term_array, index2, dest_array) {
      /* test if a shorter term is part of the current term */
      not_found = 2;
      if (array_n(disj_array) > array_n(dest_array)) {
        counter = 0;
        arrayForEachItem(Sugar_Vertex_t*, disj_array, index3, node) {
          arrayForEachItem(Sugar_Vertex_t*, dest_array, index4, dest_node) {
            if (node == dest_node) { counter++; }
          }
        }
        if (counter > 0) {
          not_found = (counter==array_n(dest_array))?(0):(1);
          break;
        }
      }
    }
    /* no shorter term found, so add this current term */
    if (not_found) {
      arrayForEachItem(array_t*, duplicate, index2, temp) {
        if (temp == disj_array) { not_found = 0; break; }
      }
      if (not_found) {
        array_insert_last(array_t*, duplicate, disj_array);
      }
    }

  }
#ifdef DEBUG_SUGAR
  if (array_n(duplicate) < array_n(term_array)) {
    printf("Removed %d (out of %d) terms from DNF.\n", array_n(term_array) - array_n(duplicate), array_n(term_array));
  }
#endif
  /* duplicate the required terms, ... */
  new_term_array = array_alloc(array_t*, 0);
  arrayForEachItem(array_t*, duplicate, index2, temp) {
    array_insert_last(array_t*, new_term_array, array_dup(temp));
  }
  array_free(duplicate);
  /* ... delete old terms and return */
  arrayForEachItem(array_t*, term_array, index2, temp) {
    array_free(temp);
  }
  array_free(term_array);
  return new_term_array;
}

/**Function********************************************************************

Synopsis    [Creates automaton for identidier.]

Description [Creates the base case for all automata: Two nodes with one
edge with name as label. If isNegate is not zero the label is negated.]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     []

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonIdentifier(
  char *name,
  int isNegate,
  int fromFormula)
{
  char *prop_name;
  array_t *dest_array;
  array_t *prop_array;
  array_t *term_array;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node1;
  Sugar_Vertex_t *node2;
  Sugar_Automaton_t *aut;

  /* add identifer as new proposition */
  aut = Sugar_AutomatonCreate();
  if (name != NIL(char)) {
    prop_name = util_strsav(name);
    lsNewEnd(aut->prop_list, (lsGeneric)prop_name, 0);
  }
  else {
    prop_name = NIL(char);
  }
  /* generate two nodes */
  node1 = SugarAutomatonNewNode(aut);
  node2 = SugarAutomatonNewNode(aut);
  /* generate edge from node1 to node2 */
  dest_array = array_alloc(Sugar_Vertex_t*, 1);
  array_insert(Sugar_Vertex_t*, dest_array, 0, node2);
  labl = ALLOC(Sugar_Label_t, 1);
  labl->name = prop_name; labl->isNegate = isNegate;
  prop_array = array_alloc(Sugar_Label_t*, 1);
  array_insert(Sugar_Label_t*, prop_array, 0, labl);
  SugarAutomatonNewEdge(aut, node1, dest_array, prop_array, 0);
  array_free(dest_array);
  array_free(prop_array);
  /* make node1 the initial state */
  term_array = array_alloc(Sugar_Vertex_t*, 1);
  array_insert(Sugar_Vertex_t*, term_array, 0, node1);
  array_insert_last(array_t*, aut->initial_states, term_array);
  /* finally make node2 accepting */
  array_insert_last(Sugar_Vertex_t*, aut->accept_states, node2);
  /* if it is a formula propositional, add TRUE loop */
  if (fromFormula) SugarAutomatonNewEdgeLoop(aut, node2);
  return aut;
}

/**Function********************************************************************

Synopsis    [Concatenate two alternating automaton.]

Description [This important routine combines two alternating automata.
Depending on the value of typeOverlap there are five possible combination:
First, no overlapping. The right automaton is added to the end of the
left automaton. Second, with overlapping. The outgoing edges from the
initial states of the right automaton are ANDed with the incoming edges
of the accepting states of the left automaton. Third, overlapping with
alternation. This is like the second case, but the destination nodes
from the left automaton are not replaced but combined with the destination
nodes from the right automaton. Case four and five are special cases
from case three. The following table summarizes this:

typeOverlap | overlap |   edges   |   accepting
------------+---------+-----------+---------------
0     |    no   |   normal  |   replaced
1     |   yes   |   normal  |   replaced
2     |   yes   | alternate |   replaced
3     |   yes   | alternate |   appended
4     |   yes   | alternate | replaced new

Finally all additional data from the right automaton are moved to the
left automaton. Most of the references in the nodes and edges of the righ
automaton are not changed.]

SideEffects [The left automaton grows, the right automaton gets deleted.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonConcat(
  Sugar_Automaton_t *aut_left,
  Sugar_Automaton_t *aut_right,
  int typeOverlap)
{
  int index1;
  int index2;
  int index3;
  int index4;
  int index5;
  int not_found;
  int has_initial_accept_states;
  Sugar_Edge_t *edge;
  Sugar_Edge_t *in_edge;
  Sugar_Label_t *labl;
  Sugar_Label_t *new_labl;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *acc_node;
  Sugar_Vertex_t *ini_node;
  array_t *disj_array;
  array_t *final, *base;
  array_t *composite;
  array_t *duplicate;
  array_t *temp, *dup;
  array_t *dest_array;
  array_t *label_array;
  array_t *new_label_array;
  array_t *new_init_array;
  lsGen gen;

  /* set all incoming edges of right automaton to accepting states of left automaton */
  arrayForEachItem(array_t*, aut_right->initial_states, index1, disj_array) {
    arrayForEachItem(Sugar_Vertex_t*, disj_array, index2, ini_node) {
      lsForEachItem(ini_node->edge_in_list, gen, edge) {
        arrayForEachItem(Sugar_Vertex_t*, aut_left->accept_states, index3, acc_node) {
          temp = array_alloc(Sugar_Vertex_t*, 1);
          array_insert(Sugar_Vertex_t*, temp, 0, acc_node);
          /* source is still in right automaton; will be copied to left automaton later */
          SugarAutomatonNewEdge(aut_left, edge->source_vertex, temp, edge->label, 0);
          array_free(temp);
        }
      }
      /* delete incoming edges from initial states */
      lsForEachItem(ini_node->edge_in_list, gen, edge) {
        SugarAutomatonFreeEdge(edge);
      }
    }
  }

  /* generate outlist array for all initial states */
  final = array_alloc(array_t*, 0);
  arrayForEachItem(array_t*, aut_right->initial_states, index1, disj_array) {
    base = array_alloc(array_t*, 0);
    temp = array_alloc(Sugar_Edge_t*, 0);
    array_insert_last(array_t*, base, temp);
    arrayForEachItem(Sugar_Vertex_t*, disj_array, index2, ini_node) {
      composite = array_alloc(array_t*, 0);
      lsForEachItem(ini_node->edge_out_list, gen, edge) {
        /* setup new base array for new outgoing edge */
        duplicate = array_alloc(array_t*, 0);
        arrayForEachItem(array_t*, base, index3, temp) {
          dup = array_dup(temp);
          array_insert_last(array_t*, duplicate, dup);
        }
        /* add edge to all items of current AND-list */
        arrayForEachItem(array_t*, duplicate, index3, temp) {
          array_insert_last(Sugar_Edge_t*, temp, edge);
        }
        array_append(composite, duplicate);
        array_free(duplicate);
      }
      /* set new base */
      array_free(base);
      base = composite;
    }
    /* add base to final OR-list */
    array_append(final, base);
    array_free(base);
  }

  /* now we add edges to the destination nodes of all edges in the 'final'-array */
  if (typeOverlap > 0) {
    /* for all incoming edge of accepting states in left automaton... */
    arrayForEachItem(Sugar_Vertex_t*, aut_left->accept_states, index1, acc_node) {
      /* duplicate in-list of accepting state, because we may modify edge data */
      duplicate = array_alloc(Sugar_Edge_t*, 0);
      lsForEachItem(acc_node->edge_in_list, gen, in_edge) {
        array_insert_last(Sugar_Edge_t*, duplicate, in_edge);
      }
      /* run through a backup of in-list, because we may add edges to it */
      arrayForEachItem(Sugar_Edge_t*, duplicate, index2, in_edge) {
        /* backup labels of edge as base */
        base = array_dup(in_edge->label);
        arrayForEachItem(array_t*, final, index3, temp) {
          /* clone the base labels */
          label_array = array_alloc(Sugar_Label_t*, 0);
          arrayForEachItem(Sugar_Label_t*, base, index4, labl) {
            new_labl = ALLOC(Sugar_Label_t, 1);
            new_labl->name = labl->name; new_labl->isNegate = labl->isNegate;
            array_insert_last(Sugar_Label_t*, label_array, new_labl);
          }
          /* for LTL formulas we need to add alternate edges */
          dest_array = array_alloc(Sugar_Vertex_t*, 0);
          if (typeOverlap > 1) {
            lsForEachItem(in_edge->dest_vertex_list, gen, node) {
              array_insert_last(Sugar_Vertex_t*, dest_array, node);
            }
          }
          /* for each edge in an accepting state add new destination nodes */
          arrayForEachItem(Sugar_Edge_t*, temp, index4, edge) {
            lsForEachItem(edge->dest_vertex_list, gen, node) {
              array_insert_last(Sugar_Vertex_t*, dest_array, node);
            }
            /* prevent duplicate label names and handle TRUE and FALSE */
            new_label_array = SugarAutomatonMergeLabelArray(edge->label, label_array);
            /* delete previous label and replace it with new label */
            arrayForEachItem(Sugar_Label_t*, label_array, index5, labl) {
              FREE(labl);
            }
            label_array = new_label_array;
          }
          /* ... add new edge from its source to all required nodes in right automaton */
          SugarAutomatonNewEdge(aut_left, in_edge->source_vertex, dest_array, label_array, 2);
          /* free temporary data */
          arrayForEachItem(Sugar_Label_t*, label_array, index4, labl) {
            FREE(labl);
          }
          array_free(label_array);
          array_free(dest_array);
        }
        array_free(base);
      }
      /* for LTL formulas remove old incoming edges, we have new alternate edges */
      if (typeOverlap > 1) {
        arrayForEachItem(Sugar_Edge_t*, duplicate, index2, in_edge) {
          SugarAutomatonFreeEdge(in_edge);
        }
      }
      array_free(duplicate);
      /* if accepting state has no outgoing edges ... */
      if (lsLength(acc_node->edge_out_list) == 0) {
#ifdef DEBUG_SUGAR
        printf("Concat: DELETE accepting state q%d\n", acc_node->id);
#endif
        /* ... (TODO?) check (TODO?) if we can delete the incoming edges */
        lsForEachItem(acc_node->edge_in_list, gen, in_edge) {
          SugarAutomatonFreeEdge(in_edge);
        }
        /* if it is an initial state, then also delete the whole initial term */
        new_init_array = array_alloc(array_t*, 0);
        arrayForEachItem(array_t*, aut_left->initial_states, index2, disj_array) {
          not_found = 1;
          arrayForEachItem(Sugar_Vertex_t*, disj_array, index3, ini_node) {
            if (acc_node == ini_node) { not_found = 0; break; }
          }
          if (not_found) {
            array_insert_last(array_t*, new_init_array, disj_array);
          }
        }
        /* set new initial states only if at least one is left */
        if ((array_n(new_init_array) > 0) &&
            (array_n(new_init_array) < array_n(aut_left->initial_states))) {
          array_free(aut_left->initial_states);
          aut_left->initial_states = new_init_array;
        }
        else {
          array_free(new_init_array);
        }
      }
    }
  }
  else {
    /* create outlist and add edges to all accept states in aut_left */
    arrayForEachItem(array_t*, final, index1, temp) {
      /* allocate composite label array and initialize with TRUE */
      labl = ALLOC(Sugar_Label_t, 1);
      labl->name = NIL(char); labl->isNegate = 0;
      label_array = array_alloc(Sugar_Label_t*, 1);
      array_insert(Sugar_Label_t*, label_array, 0, labl);
      /* allocate composite array for destination nodes */
      dest_array = array_alloc(Sugar_Vertex_t*, 0);
      arrayForEachItem(Sugar_Edge_t*, temp, index2, edge) {
        lsForEachItem(edge->dest_vertex_list, gen, node) {
          array_insert_last(Sugar_Vertex_t*, dest_array, node);
        }
        /* prevent duplicate references in label array of edge */
        new_label_array = SugarAutomatonMergeLabelArray(edge->label, label_array);
        /* delete previous label and replace it with new label */
        arrayForEachItem(Sugar_Label_t*, label_array, index3, labl) {
          FREE(labl);
        }
        label_array = new_label_array;
      }
      arrayForEachItem(Sugar_Vertex_t*, aut_left->accept_states, index2, acc_node) {
        SugarAutomatonNewEdge(aut_left, acc_node, dest_array, label_array, 2);
      }
      /* free temporary data */
      arrayForEachItem(Sugar_Label_t*, label_array, index2, labl) {
        FREE(labl);
      }
      array_free(label_array);
      array_free(dest_array);
    }
  }

  /* delete all elements of array 'final' */
  arrayForEachItem(array_t*, final, index1, temp) {
    array_free(temp);
  }
  array_free(final);

  /* check if we have initial accepting states */
  has_initial_accept_states = 0;
  composite = array_alloc(Sugar_Vertex_t*, 0);
  arrayForEachItem(Sugar_Vertex_t*, aut_right->accept_states, index1, acc_node) {
    not_found = 1;
    arrayForEachItem(array_t*, aut_right->initial_states, index2, disj_array) {
      arrayForEachItem(Sugar_Vertex_t*, disj_array, index3, ini_node) {
        if (ini_node == acc_node) {
          //printf("Concat: Accepting initial state q%d!\n", ini_node->id);
          has_initial_accept_states = 1; not_found = 0; break;
        }
      }
      if (not_found == 0) break;
    }
    if (not_found) {
      array_insert_last(Sugar_Vertex_t*, composite, acc_node);
    }
  }
  /* delete all initial states from right automaton */
  base = array_alloc(Sugar_Vertex_t*, 0);
  arrayForEachItem(array_t*, aut_right->initial_states, index1, disj_array) {
    arrayForEachItem(Sugar_Vertex_t*, disj_array, index2, ini_node) {
      not_found = 1;
      arrayForEachItem(Sugar_Vertex_t*, base, index3, node) {
        if (node == ini_node) { not_found = 0; break; }
      }
      if (not_found) {
        array_insert_last(Sugar_Vertex_t*, base, ini_node);
      }
    }
  }
  arrayForEachItem(Sugar_Vertex_t*, base, index1, ini_node) {
    SugarAutomatonFreeNode(ini_node);
  }
  array_free(base);

  /* add or set new accepting states from right automaton */
  if (has_initial_accept_states) {
    if ((typeOverlap == 0) || typeOverlap == 3) {
      //printf("Accept states APPENDED\n");
      array_append(aut_left->accept_states, composite);
    }
    else {
      //printf("Accept states REPLACED with NEW\n");
      array_free(aut_left->accept_states);
      aut_left->accept_states = array_dup(composite);
    }
  }
  else {
    if (typeOverlap > 3) {
      //printf("Accept states REPLACED with NEW\n");
      array_free(aut_left->accept_states);
      aut_left->accept_states = array_dup(composite);
    }
    else if (typeOverlap > 2) {
      //printf("Accept states APPENDED\n");
      array_append(aut_left->accept_states, composite);
    }
    else {
      //printf("Accept states REPLACED with RIGHT\n");
      array_free(aut_left->accept_states);
      aut_left->accept_states = aut_right->accept_states;
    }
  }
  array_free(composite);
  /* finally move additional data from right to left automaton */
  SugarAutomatonMoveRightToLeft(aut_left, aut_right);
}

/**Function********************************************************************

Synopsis    [Concatinate two alternating automaton with OR.]

Description [Only append the initial and accepting states from the right
automaton to the left automaton and copy all other data. The special case
with two simple automata with just one edge and idendifer is optimized.
There only the new edge from the right automaton is added to the left
automaton. For formulas the accepting states from the right automaton
are marked for the LTL operator (G(f)).]

SideEffects [The right automaton gets deleted.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonOr(
  Sugar_Automaton_t *aut_left,
  Sugar_Automaton_t *aut_right)
{
  char *new_name;
  array_t *init_array;
  array_t *label_array;
  Sugar_Edge_t *edge_l;
  Sugar_Edge_t *edge_r;
  Sugar_Label_t *labl_l;
  Sugar_Label_t *labl_r;
  Sugar_Label_t *new_labl;
  Sugar_Vertex_t *ini_node;

  /* special case two automatons with single arcs: ID | ID */
  if (lsLength(aut_left->state_list)==2 && lsLength(aut_right->state_list)==2
      && lsLength(aut_left->prop_list)==1 && lsLength(aut_right->prop_list)==1
      && lsLength(aut_left->edge_list)==1 && lsLength(aut_right->edge_list)==1) {
    lsFirstItem(aut_left->edge_list, (lsGeneric*)&edge_l, NULL);
    lsFirstItem(aut_right->edge_list, (lsGeneric*)&edge_r, NULL);
    labl_l = array_fetch(Sugar_Label_t*, edge_l->label, 0);
    labl_r = array_fetch(Sugar_Label_t*, edge_r->label, 0);
    /* normal case with two propositionals */
    if ((labl_l->name != NIL(char)) && (labl_r->name != NIL(char))) {
      /* add new label to left automaton if names are different */
      if (strcmp(labl_l->name, labl_r->name) != 0) {
        new_name = util_strsav(labl_r->name);
        lsNewEnd(aut_left->prop_list, (lsGeneric)new_name, 0);
        /* add new edge in left automaton */
        new_labl = ALLOC(Sugar_Label_t, 1);
        new_labl->name = new_name; new_labl->isNegate = labl_r->isNegate;
        label_array = array_alloc(Sugar_Label_t*, 1);
        array_insert(Sugar_Label_t*, label_array, 0, new_labl);
        init_array = array_fetch(array_t*, aut_left->initial_states, 0);
        ini_node = array_fetch(Sugar_Vertex_t*, init_array, 0);
        SugarAutomatonNewEdge(aut_left, ini_node, aut_left->accept_states, label_array, 0);
        array_free(label_array);
      }
      else {
        /* check if (a OR !a) should be TRUE */
        if (labl_l->isNegate != labl_r->isNegate) {
          labl_l->name = NIL(char); labl_l->isNegate = 0;
        }
      }
    }
    else {
      /* special cases for labels TRUE and FALSE */
      if (IsLabelTrue(labl_l) || IsLabelTrue(labl_r)) {
        /* left or right label is TRUE >> set to TRUE */
        labl_l->name = NIL(char); labl_l->isNegate = 0;
      }
      else if (IsLabelFalse(labl_l) && !IsLabelFalse(labl_r)) {
        /* left label is FALSE and right label is a propositonal >> copy it */
        new_name = util_strsav(labl_r->name);
        lsNewEnd(aut_left->prop_list, (lsGeneric)new_name, 0);
        labl_l->name = new_name; labl_l->isNegate = labl_r->isNegate;
      }
    }
    Sugar_AutomatonFree(aut_right);
  }
  else {
    /* add initial states from aut_right to aut_left (ORing !!!) */
    array_append(aut_left->initial_states, aut_right->initial_states);
    array_free(aut_right->initial_states);
    /* add accept states from aut_right to aut_left */
    array_append(aut_left->accept_states, aut_right->accept_states);
    array_free(aut_right->accept_states);
    /* move nodes and edges from right to left automaton */
    SugarAutomatonMoveRightToLeft(aut_left, aut_right);
  }
}

/**Function********************************************************************

Synopsis    [Concatinate two alternating automaton with length matching AND.]

Description [The initial states are calculated by cross product. The
accepting states are appended and all other data eill be copied. The
special case with two simple automata with just one edge and idendifer
is optimized. There only the label from the right automaton is added to
the label of the edge from the left automaton.]

SideEffects [The right automaton gets deleted.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonAndEq(
  Sugar_Automaton_t *aut_left,
  Sugar_Automaton_t *aut_right)
{
  int index1;
  int index2;
  char *new_name;
  array_t *item_l;
  array_t *item_r;
  array_t *new_array;
  array_t *temp_disj;
  Sugar_Edge_t *edge_l;
  Sugar_Edge_t *edge_r;
  Sugar_Label_t *labl_l;
  Sugar_Label_t *labl_r;
  Sugar_Label_t *new_labl;

  /* special case two automatons with single arcs: ID && ID */
  if (lsLength(aut_left->state_list)==2 && lsLength(aut_right->state_list)==2
      && lsLength(aut_left->prop_list)==1 && lsLength(aut_right->prop_list)==1
      && lsLength(aut_left->edge_list)==1 && lsLength(aut_right->edge_list)==1) {
    lsFirstItem(aut_left->edge_list, (lsGeneric*)&edge_l, NULL);
    lsFirstItem(aut_right->edge_list, (lsGeneric*)&edge_r, NULL);
    labl_l = array_fetch(Sugar_Label_t*, edge_l->label, 0);
    labl_r = array_fetch(Sugar_Label_t*, edge_r->label, 0);
    /* normal case with two propositionals */
    if ((labl_l->name != NIL(char)) && (labl_r->name != NIL(char))) {
      /* add new label to left automaton if names are different */
      if (strcmp(labl_l->name, labl_r->name) != 0) {
        new_name = util_strsav(labl_r->name);
        lsNewEnd(aut_left->prop_list, (lsGeneric)new_name, 0);
        /* add label to label list of edge in left automaton */
        new_labl = ALLOC(Sugar_Label_t, 1);
        new_labl->name = new_name; new_labl->isNegate = labl_r->isNegate;
        array_insert_last(Sugar_Label_t*, edge_l->label, new_labl);
      }
      else {
        /* check if (a AND !a) should be FALSE */
        if (labl_l->isNegate != labl_r->isNegate) {
          labl_l->name = NIL(char); labl_l->isNegate = 1;
        }
      }
    }
    else {
      /* special cases for labels TRUE and FALSE */
      if (IsLabelFalse(labl_l) || IsLabelFalse(labl_r)) {
        /* left or right label is FALSE >> set to FALSE */
        labl_l->name = NIL(char); labl_l->isNegate = 1;
      }
      else if (IsLabelTrue(labl_l) && !IsLabelTrue(labl_r)) {
        /* left label is TRUE and right label is a propositonal >> copy it */
        new_name = util_strsav(labl_r->name);
        lsNewEnd(aut_left->prop_list, (lsGeneric)new_name, 0);
        labl_l->name = new_name; labl_l->isNegate = labl_r->isNegate;
      }
    }
    Sugar_AutomatonFree(aut_right);
  }
  else {
    /* calculate initial states: DNF needs cross product (ANDing !!!) */
    new_array = array_alloc(array_t*, 0);
    arrayForEachItem(array_t*, aut_left->initial_states, index1, item_l) {
      arrayForEachItem(array_t*, aut_right->initial_states, index2, item_r) {
        temp_disj = array_join(item_l, item_r);
        array_insert_last(array_t*, new_array, temp_disj);
      }
    }
    /* delete old initial states and set new one for aut_left */
    arrayForEachItem(array_t*, aut_left->initial_states, index1, item_l) {
      array_free(item_l);
    }
    array_free(aut_left->initial_states);
    arrayForEachItem(array_t*, aut_right->initial_states, index1, item_r) {
      array_free(item_r);
    }
    array_free(aut_right->initial_states);
    aut_left->initial_states = new_array;
    /* add accept states from aut_right to aut_left */
    array_append(aut_left->accept_states, aut_right->accept_states);
    array_free(aut_right->accept_states);
    /* move data from right to left automaton */
    SugarAutomatonMoveRightToLeft(aut_left, aut_right);
  }
}

/**Function********************************************************************

Synopsis    [Generate automaton for SERE star operator]

Description [Construct automaton with plus operator and add and accepting
initial state for the empty word.]

SideEffects [The automaton changes.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonStar(
  Sugar_Automaton_t *aut)
{
  array_t *disj_array;
  Sugar_Vertex_t *eps_init;

  SugarAutomatonPlus(aut);
  eps_init = SugarAutomatonNewNode(aut);
  disj_array = array_alloc(array_t*, 1);
  array_insert(Sugar_Vertex_t*, disj_array, 0, eps_init);
  array_insert_last(array_t*, aut->initial_states, disj_array);
  array_insert_last(Sugar_Vertex_t*, aut->accept_states, eps_init);
}

/**Function********************************************************************

Synopsis    [Generate automaton for SERE plus operator]

Description [Insert edges from all accepting states to all successors
of initial states. A set of multiple initial states create alternating
edges to all possible succesors.]

SideEffects [The automaton grows with new edges.]

SeeAlso     []

******************************************************************************/
void
SugarAutomatonPlus(
  Sugar_Automaton_t *aut)
{
  int index1;
  int index2;
  Sugar_Edge_t *out_edge;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *acc_node;
  Sugar_Vertex_t *ini_node;
  array_t *disj_array;
  array_t *dest_array;
  array_t *label_array;
  array_t *init_out_array;
  array_t *label;
  lsGen gen1, gen2;

  /* generate source nodes from initial states */
  label_array = array_alloc(array_t*, 0);
  init_out_array = array_alloc(array_t*, 0);
  arrayForEachItem(array_t*, aut->initial_states, index1, disj_array) {
    arrayForEachItem(Sugar_Vertex_t*, disj_array, index2, ini_node) {
      lsForEachItem(ini_node->edge_out_list, gen1, out_edge) {
        dest_array = array_alloc(Sugar_Vertex_t*, 0);
        lsForEachItem(out_edge->dest_vertex_list, gen2, node) {
          array_insert_last(Sugar_Vertex_t*, dest_array, node);
        }
        array_insert_last(array_t*, init_out_array, dest_array);
        array_insert_last(array_t*, label_array, out_edge->label);
      }
    }
  }
  /* insert new edges from all accepting nodes to edges' destinations */
  arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index1, acc_node) {
    arrayForEachItem(array_t*, init_out_array, index2, dest_array) {
      label = array_fetch(array_t*, label_array, index2);
      SugarAutomatonNewEdge(aut, acc_node, dest_array, label, 1);
    }
  }
  /* clean up temporary arrays */
  array_free(label_array);
  arrayForEachItem(array_t*, init_out_array, index1, dest_array) {
    array_free(dest_array);
  }
  array_free(init_out_array);
}

/**Function********************************************************************

Synopsis    [Generate automaton for formula b[->].]

Description [See SUGAR reference manual for the operators definition.]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     []

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonGoto(
  char *bool_prop)
{
  Sugar_Automaton_t *A_temp;
  Sugar_Automaton_t *A_return;

  /* build expression (!b*;b) */
  A_temp = SugarAutomatonIdentifier(bool_prop, 0, 0);
  A_return = SugarAutomatonIdentifier(bool_prop, 1, 0);
  SugarAutomatonStar(A_return);
  SugarAutomatonConcat(A_return, A_temp, 0);
  return A_return;
}

/**Function********************************************************************
 
Synopsis    [Generate complement automaton.]

Description [This function is only used on SEREs. The input automaton
should be nondeterministic. First we create a deterministic automaton.
Next we add abort edges to make the automaton complete, and finally
we exchange accepting and non-accepting states. Because we have no
alternating automata to negate, the dualization of the transitions
is not neccessary.]

SideEffects [The returned automaton should be freed by the caller.]

SeeAlso     [SugarAutomatonNot]

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonNegate(
  Sugar_Automaton_t *aut)
{
  int index;
  int not_found;
  array_t *node_array;
  array_t *new_acc_array;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *acc_node;
  Sugar_Automaton_t *aut_neg;
  lsGen gen;

  /* make nondeterministic automaton deterministic */
  aut_neg = Sugar_AutomatonDeterministic(aut);
  /* create array of non-accepting states */
  node_array = array_alloc(Sugar_Vertex_t*, 0);
  new_acc_array = array_alloc(Sugar_Vertex_t*, 0);
  lsForEachItem(aut_neg->state_list, gen, node) {
    not_found = 1;
    arrayForEachItem(Sugar_Vertex_t*, aut_neg->accept_states, index, acc_node) {
      if (node == acc_node) { not_found = 0; break; }
    }
    if (not_found) {
      array_insert_last(Sugar_Vertex_t*, new_acc_array, node);
    }
  }
  /* add abort state with abort edges */
  aut_neg = SugarAutomatonImply(aut_neg, NIL(Sugar_Automaton_t), 1, 0);
  /* set new accepting states; read and add abort node */
  acc_node = array_fetch_last(Sugar_Vertex_t*, aut_neg->accept_states);
  array_insert_last(Sugar_Vertex_t*, new_acc_array, acc_node);
  array_free(aut_neg->accept_states);
  aut_neg->accept_states = new_acc_array;
  /* remove (a lot of ?) FALSE edges before returning */
  SugarAutomatonFlush(aut_neg);
  return aut_neg;
}

/**Function********************************************************************
 
Synopsis    [copy a label array]

Description [creates a new array and fills it with duplicates of the labels
in the given array]
   
SideEffects []

SeeAlso     []

******************************************************************************/
array_t* labelDup(array_t* label){  
  array_t* new_label;
  int index;
  Sugar_Label_t* literal, *new_lit;
  
  new_label = array_alloc(Sugar_Label_t*, 0);
  arrayForEachItem(Sugar_Label_t*, label, index, literal){
    new_lit = ALLOC(Sugar_Label_t,1);
    new_lit->name = literal->name;
    new_lit->isNegate = literal->isNegate;
    array_insert_last(Sugar_Label_t*, new_label, new_lit);
  }
  return new_label;
}

/**Function********************************************************************
 
Synopsis [copy a list with pointers to vertices - the vertices are not
copied]

Description []
   
SideEffects []

SeeAlso     []

******************************************************************************/
lsList destinationDup(lsList oldlist){
  Sugar_Vertex_t* vertex;
  lsList newlist = lsCreate();
  lsGen gen;
  
  lsForEachItem(oldlist, gen, vertex){
    lsNewEnd(newlist, (lsGeneric) vertex->destination, LS_NH);
  }
  return newlist;
}

/**Function********************************************************************
 
Synopsis    [Build an automaton for the Imply-operator]

Description [The left hand side automaton has to be a NFA, right hand
automaton can be of any type, output automaton is a AFA.
   
First we invert the left hand automaton. the former accepting states are
then connected to the input states of the right hand side automaton.]
   
SideEffects [The returned automaton should be freed by the caller. Field
destination of the nodes of the old automaton links to the new nodes in the
new automaton. Field name is set to 1 if the state was accepting.]

SeeAlso     [SugarAutomatonNegate]

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonImplyNFA(
  Sugar_Automaton_t *left, Sugar_Automaton_t* right)
{
  lsGen gen, gen2;
  Sugar_Vertex_t *lnode, *rnode, *new_node;
  Sugar_Edge_t* edge;
  Sugar_Edge_t* new_edge;
  Sugar_Label_t *label;
  lsHandle handle;
  lsList new_edges_list;
  int index, index2;
  array_t* node_label_array, *label_combinations, *left_connect_states;

/*   fprintf(nusmv_stdout, "\033[01;34m\033[07m*********************\n** right automaton **\033[0m\033[22;34m\n"); */
/*   Sugar_AutomatonPrint(right,3); */
/*   fprintf(nusmv_stdout, "\033[0m"); */

  /* add TRUE loops to accepting states if there is no other edge */
  label = ALLOC(Sugar_Label_t, 1);
  label->name = NIL(char); label->isNegate = 0;
  node_label_array = array_alloc(Sugar_Label_t*, 1);
  array_insert(Sugar_Label_t*, node_label_array, 0, label);
  array_t* dest_array = array_alloc(Sugar_Vertex_t*, 1);
  array_t* edge_true_array = array_alloc(Sugar_Edge_t*, 0);
  arrayForEachItem(Sugar_Vertex_t*, left->accept_states, index, lnode) {
    if (lsLength(lnode->edge_out_list) == 0) {
      array_insert(Sugar_Vertex_t*, dest_array, 0, lnode);
      edge = SugarAutomatonNewEdge(left, lnode, dest_array, node_label_array, 0);
      array_insert_last(Sugar_Edge_t*, edge_true_array, edge);
    }
  }

  FREE(label);
  array_free(node_label_array);
  array_free(dest_array);

  /** prepare the left automaton */
  Sugar_Automaton_t *negated_left = SugarAutomatonNegateNFA(left);


  
  /* find the states in the left automaton to connect the right automaton to by
     marking the accepting states */
  lsForEachItem(negated_left->state_list, gen, lnode){
    lnode->name= (char *)0;
  }
  arrayForEachItem(Sugar_Vertex_t*, negated_left->accept_states, index, lnode){
    lnode->name= (char *)1;
  }
  left_connect_states=array_alloc(Sugar_Vertex_t*, 0);
/*   fprintf(nusmv_stdout, "left_connectStates"); */

  array_free(negated_left->accept_states);
  negated_left->accept_states = array_alloc(Sugar_Vertex_t*, 0);

  lsForEachItem(negated_left->state_list, gen, lnode){
    array_insert_last(Sugar_Vertex_t*, negated_left->accept_states, lnode);
    if(lnode->name == (char *)0){
      array_insert_last(Sugar_Vertex_t*, left_connect_states, lnode);
    }
  }  


  
  /* mark the accepting states of the right automaton to adjust the edges for weakness */
  lsForEachItem(right->state_list, gen, rnode){
    rnode->name= (char *)0;
  }
  arrayForEachItem(Sugar_Vertex_t*, right->accept_states, index, rnode){
    rnode->name= (char *)1;
  }
  
  /** copy the states of the right automaton into the left one */
  lsForEachItem(right->state_list, gen, rnode){
    new_node = SugarAutomatonNewNode(negated_left);
    rnode->destination = (array_t*)new_node;
  }
  
  /* gather the initial states of the right automaton*/
  array_t *right_initial_states;
  right_initial_states = array_alloc(Sugar_Vertex_t*, 0);
  array_t *initial_state;
  arrayForEachItem(array_t*, right->initial_states, index, initial_state){
    array_append(right_initial_states, initial_state);    
  }
  array_sort(right_initial_states, (int(*)())compareVertex);
  array_uniq(right_initial_states, (int(*)())compareVertex, NULL);

  /* add selfloops to accepting states */
  arrayForEachItem(Sugar_Vertex_t*, right->accept_states, index, rnode){
    new_node = (Sugar_Vertex_t *) rnode->destination;
    Sugar_Label_t* truelabel = ALLOC(Sugar_Label_t, 1);
    truelabel->name = NIL(char); truelabel->isNegate=0;
    
    /* create the selfloop (edge) */
    Sugar_Edge_t *selfloop = ALLOC(Sugar_Edge_t, 1);
    array_t *true_labels = array_alloc(Sugar_Label_t*, 1);
    array_insert(Sugar_Label_t*, true_labels, 0, truelabel);
    lsList dest_vertex = lsCreate();
    lsNewEnd(dest_vertex, (lsGeneric)new_node, LS_NH);

    selfloop->label = true_labels;
    selfloop->source_vertex=new_node;
    selfloop->dest_vertex_list=dest_vertex;
    /*add to the node*/
#ifdef SUGAR_DEBUG
    printEdgeList(new_node->edge_in_list);
    printEdgeList(new_node->edge_out_list);
#endif
    lsNewEnd(new_node->edge_in_list, (lsGeneric)selfloop, LS_NH);
    lsNewEnd(new_node->edge_out_list, (lsGeneric)selfloop, LS_NH);

    /* add to the new automaton */
    lsNewEnd(negated_left->edge_list, (lsGeneric)selfloop, &selfloop->handle);

  }
  
  /* adjust the edge lists for nonaccepting states*/
  lsForEachItem(right->edge_list, gen, edge) {
    if(edge->source_vertex->name == (char *)1)
      continue;			/* do nothing for edges from accepting states */
    Sugar_Edge_t* new_edge = ALLOC(Sugar_Edge_t, 1);
    array_t* label_array = array_alloc(Sugar_Label_t*, 0);    
    arrayForEachItem(Sugar_Label_t*, edge->label, index, label){ /* copy propositions
								    to the new automaton*/
      Sugar_Label_t* new_label = ALLOC(Sugar_Label_t, 1);
      if(label->name == NIL(char))
	new_label->name = NIL(char);
      else{
	int set=0;
	char* tmpStr;
	lsForEachItem(negated_left->prop_list, gen2, tmpStr){
	  if(strcmp(tmpStr, label->name)==0){
	    new_label->name=tmpStr;
	    set=1;
	  }
	}
	if(!set){
	  tmpStr = util_strsav(label->name);
	  lsNewEnd(negated_left->prop_list, (lsGeneric)tmpStr, LS_NH);
	  new_label->name=tmpStr;
	}
      }
      new_label->isNegate = label->isNegate;
      array_insert_last(Sugar_Label_t*, label_array, new_label);
    }
    
    new_edge->label = label_array;
    new_edge->source_vertex = (Sugar_Vertex_t*) edge->source_vertex->destination;
    lsNewEnd(new_edge->source_vertex->edge_out_list,(lsGeneric) new_edge, LS_NH);
    
    lsList destination = lsCreate();
    lsForEachItem(edge->dest_vertex_list, gen2, rnode){
      lsNewEnd(destination, (lsGeneric)rnode->destination, LS_NH);
      lsNewEnd(((Sugar_Vertex_t*) rnode->destination)->edge_in_list, (lsGeneric)new_edge, LS_NH);
    }
    new_edge->dest_vertex_list = destination;

    lsNewEnd(negated_left->edge_list, (lsGeneric)new_edge, &handle);
    new_edge->handle=handle;
  }

/*   fprintf(nusmv_stdout, "\033[01;31m\033[07m*********************\n** copied nodes **\033[0m\033[22;31m\n"); */
/*   lsForEachItem(right->state_list, gen, rnode){ */
/*     printNode((Sugar_Vertex_t*)rnode->destination); */
/*   } */
/*   Sugar_AutomatonPrint(negated_left, 3); */
/*   fprintf(nusmv_stdout, "\033[0m"); */

  
  /** connect the automatons **/
  /* determinize the edges out of the initial states (right automaton) */
  /*   array_t *label_combiniations; */
  /*   arrayForEachItem(Sugar_Vertex_t*, right_initial_states, index, rnode) { */
  /*     new_node = (Sugar_Vertex_t*)rnode->destination; */
  /*     /\* collect used labels *\/ */
  /*     node_label_array = array_alloc(Sugar_Label_t*, 0); */
  /*     lsForEachItem(new_node->edge_out_list, gen2, edge){ */
  /*       array_append(node_label_array, edge->label); */
  /*     } */
  /*     array_sort(node_label_array, (int (*)()) compareLabel); */
  /*     array_uniq(node_label_array, (int (*)()) compareLabel, NULL); */

  /*     /\*create label combinations*\/ */
  /*     label_combinations = array_alloc(array_t*, 0); */
  /*     int num_edge = bitRange(array_n(node_label_array)); */
  /*     while(num_edge > index2){ */
  /*       array_t *new_labels = array_alloc(Sugar_Label_t *, 0); */
  /*       int edgebits = array_n(node_label_array); */
  /*       while(edgebits>0){ */
  /* 	edgebits--; */
  /* 	Sugar_Label_t* new_label = ALLOC(Sugar_Label_t, 1); */
  /* 	array_insert_last(Sugar_Label_t *, new_labels, new_label); */
  /* 	new_label->name = array_fetch(Sugar_Label_t*, node_label_array, edgebits)->name; */
  /* 	if((index2>>edgebits)&1==1) */
  /* 	  new_label->isNegate=0; */
  /* 	else */
  /* 	  new_label->isNegate=1; */
  /*       }       */
  /*       array_insert_last(array_t*, label_combinations, new_labels); */
  /*       index2++;       */
  /*     } */

  /*     /\* for each label combination and edge: if there is no conflict, create a */
  /*        new edge *\/ */
  /*     new_edges_list = lsCreate(); */
  /*     lsForEachItem(new_node->edge_out_list, gen, edge){ */
  /*       array_t *label_combi; */
  /*       arrayForEachItem(array_t*, label_combinations, index2, label_combi){ */
  /* 	if(!conflict(label_combi, edge->label)){	   */
  /* 	  /\* create new edge *\/ */
  /* 	  new_edge=ALLOC(Sugar_Edge_t, 1); */
  /* 	  new_edge->label = labelDup(label_combi); */
  /* 	  new_edge->source_vertex=new_node; */
  /* 	  new_edge->dest_vertex_list=lsCopy(edge->dest_vertex_list, NULL);//destinationDup(edge->dest_vertex_list); */
  /* 	  lsNewEnd(negated_left->edge_list, (lsGeneric)new_edge, &(new_edge->handle)); */

  /* 	  /\* register in source and destination nodes *\/ */
  /* 	  lsNewEnd(new_edge->source_vertex->edge_out_list, (lsGeneric)new_edge, LS_NH); */
  /* 	  Sugar_Vertex_t* destnode; */
  /* 	  lsForEachItem(new_edge->dest_vertex_list, gen2, destnode){ */
  /* 	    lsNewEnd(destnode->edge_in_list, (lsGeneric) new_edge, LS_NH); */
  /* 	  } */

  /* 	  lsNewEnd(new_edges_list, (lsGeneric)new_edge, LS_NH); */
  /* 	} */
  /*       } */
  /*       SugarAutomatonFreeEdge(edge); */
  /*     } */
  /*     lsDestroy(new_node->edge_out_list, (void (*) (lsGeneric)) NULL); */
  /*     new_node->edge_out_list=new_edges_list;     */
  /*   } */
  
  /* build the cross product of the incoming and outgoing edges of the initial
     states of the right automaton and combine them with the final states of
     the left automaton */
  array_t* to_delete=array_alloc(Sugar_Edge_t*, 0);
  arrayForEachItem(array_t*, right->initial_states, index, initial_state) {
    array_t* destinations, summary_destinations;
    Sugar_Vertex_t* istate;
    Sugar_Edge_t* destit;
    int index3;
/*     fprintf(nusmv_stdout, "**initial edge with %i nodes:\n", array_n(initial_state));    */

    
    destinations = array_alloc(Sugar_Edge_t*, 0); /* labels and destinations of
						     the new edges*/
    arrayForEachItem(Sugar_Vertex_t*, initial_state, index2, rnode) {
      lnode = (Sugar_Vertex_t*) rnode->destination; /* corresponding state at
						       the negated_left automaton */
/*       fprintf(nusmv_stdout, "create edges for initial state:\n"); */
#ifdef DEBUG_SUGAR
      printNode(lnode);
#endif
      if(array_n(destinations)==0){
	lsForEachItem(lnode->edge_out_list, gen, edge){
/* 	  fprintf(nusmv_stdout, " initial:\n");  */
/* 	  printLabelArray(edge->label); */
	  array_insert_last(Sugar_Edge_t*, destinations, edge);
	}
      }
      else{
/* 	fprintf(nusmv_stdout, " multiply:\n");   	   */
	array_t* destbuf;
	destbuf = array_alloc(Sugar_Edge_t*, 0);
	arrayForEachItem(Sugar_Edge_t*, destinations, index3, destit){
	  lsForEachItem(lnode->edge_out_list, gen, edge){
	    array_t *combined_set=array_alloc(Sugar_Label_t*, 0);
	    if(!conflictc(edge->label, destit->label, combined_set)){
/* 	      fprintf(nusmv_stdout, "create edge for new label:\n "); */
/* 	      printLabelArray(combined_set); */
	      new_edge= ALLOC(Sugar_Edge_t, 1);
	      new_edge->label=combined_set;
	      new_edge->dest_vertex_list = lsCopy(destit->dest_vertex_list, NULL);
	      Sugar_Vertex_t* additional_dest;
	      lsForEachItem(edge->dest_vertex_list, gen2, additional_dest){
		lsNewEnd(new_edge->dest_vertex_list, (lsGeneric) additional_dest,
			 LS_NH);
	      }
#ifdef DEBUG_SUGAR
	      printLabelArray(new_edge->label);
#endif
	      array_insert_last(Sugar_Edge_t*, destbuf, new_edge);
	    }
	    else
	      array_free(combined_set);		 	    
	  }
	  /* clean up */
	  int index4;
/* 	  fprintf(nusmv_stdout, "freeing destinationlabel:"); */
	  /* 	  printEdge(destit); */
	  /* 	  arrayForEachItem(Sugar_Label_t*, destit->label, index4, label) { */
	  /* 	    FREE(label); */
	  /* 	  } */
	  /*  	  array_free(destit->label);  */
	  /*  	  lsDestroy(destit->dest_vertex_list, NULL);  //TODO: control if the */
	  /* 						      //content is to be freed */
	  /* 	  FREE(destit); */
	  /* TODO free the recources !!! take care of duplicated edges ! */
	}
	array_free(destinations);
	destinations = destbuf;
      }
    }

    array_t* new_edges=array_alloc(Sugar_Edge_t*,0);    
    Sugar_Vertex_t* destvertex;
    lsGen gen3;
    /* connect left to right automaton */
    arrayForEachItem(Sugar_Edge_t *, destinations, index3, destit) {
/*       fprintf(nusmv_stdout, "\033[22;35m\033[07m create edges for: \n "); */
/*       printLabelArray(destit->label); */
/*       fprintf(nusmv_stdout, "\033[0m"); */

      int index4;		
      Sugar_Vertex_t* connect_state;
      
      arrayForEachItem(Sugar_Vertex_t*, left_connect_states, index4, connect_state){
	Sugar_Edge_t* in_edge;
	int index5;
#ifdef DEBUG_SUGAR
	printEdgeList(connect_state->edge_in_list);
#endif
	lsForEachItem(connect_state->edge_in_list, gen2, in_edge){	  
	  array_t * combined_set = array_alloc(Sugar_Label_t*, 0);
	  if(!conflictc(in_edge->label, destit->label, combined_set)){
	    new_edge = ALLOC(Sugar_Edge_t, 1);
	    new_edge->label=combined_set;
	    new_edge->source_vertex=in_edge->source_vertex;
	    lsNewEnd(new_edge->source_vertex->edge_out_list, (lsGeneric) new_edge, LS_NH);

	    new_edge->dest_vertex_list = lsCreate();
	    lsForEachItem(destit->dest_vertex_list, gen3, destvertex){
	      lsNewEnd(new_edge->dest_vertex_list, (lsGeneric) destvertex, LS_NH);
	    }
	    lsForEachItem(in_edge->dest_vertex_list, gen3, destvertex){
	      lsNewEnd(new_edge->dest_vertex_list, (lsGeneric) destvertex, LS_NH);
	    }
	    lsNewEnd(negated_left->edge_list, (lsGeneric)new_edge, &(new_edge->handle));
/* 	    fprintf(nusmv_stdout, "\033[22;35m\033[07m new edges: \n "); */
/* 	    printEdge(new_edge); */
/* 	    fprintf(nusmv_stdout, "\033[0m"); */
	    array_insert_last(Sugar_Edge_t*, new_edges, new_edge);
	  }
	  else{
	    array_free(combined_set);
	  }
	  array_insert_last(Sugar_Vertex_t*, to_delete, in_edge);
	  //	    SugarAutomatonFreeEdge(in_edge); /* remove the original edge */

	  /* TODO there are edges which are in the list
	     twice! why !!! - problems with free. */

	}
      }      
    }
    
    arrayForEachItem(Sugar_Edge_t*, new_edges, index3, new_edge){
      lsForEachItem(new_edge->dest_vertex_list, gen3, destvertex){
	lsNewEnd(destvertex->edge_in_list, (lsGeneric) new_edge, LS_NH);
      }
    }

    lsForEachItem(negated_left->edge_list, gen2, edge){
#ifdef DEBUG_SUGAR
      printEdge(edge);
#endif
      lsSort(edge->dest_vertex_list, (int(*)())lsCompareVertex); 
      lsUniq(edge->dest_vertex_list, (int(*)())lsCompareVertex, NULL); 
    }
  }

  array_sort(to_delete, (int (*)()) compareEdge); 
  array_uniq(to_delete, (int (*)()) compareEdge, NULL); 
  
  arrayForEachItem(array_t*, to_delete, index, edge){
    SugarAutomatonFreeEdge(edge);
  }
    

/*   array_free(negated_left->accept_states); */
/*   negated_left->accept_states = array_alloc(Sugar_Vertex_t*, 0); */
  /* adjust accepting states */
  arrayForEachItem(Sugar_Vertex_t*, right->accept_states, index, rnode){
    array_insert_last(Sugar_Vertex_t*, negated_left->accept_states, (Sugar_Vertex_t*)rnode->destination);
  }

/*   arrayForEachItem(Sugar_Vertex_t*, left->accept_states, index, rnode){ */
/*     array_insert_last(Sugar_Vertex_t*, negated_left->accept_states, (Sugar_Vertex_t*)rnode->destination); */
/*   } */
  
  return negated_left;
}

/**Function********************************************************************
 
Synopsis    [Generate complement of an NFA.]

Description [This function is only used on NFAs. The input automaton
has to be an NFA, output automaton is an UFA.
   
First we change disjunctive (or) edges to conjunctive (and). Next we swap
accepting and non acceptiong states.]
   
SideEffects [The returned automaton should be freed by the caller. Changes
field name of the vertices to invert the accepting set. changes field
destination for the connection between old and new automaton.]

SeeAlso     [SugarAutomatonNegate]

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonNegateNFA(
  Sugar_Automaton_t *aut)
{
  array_t *node_array;
  array_t *new_acc_array;
  array_t *node_label_array;
  lsList new_edges_list;
  Sugar_Automaton_t *new_aut;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *acc_node;
  Sugar_Edge_t *edge;
  Sugar_Edge_t *oldedge;
  Sugar_Label_t *label;
  lsGen gen;
  lsGen innergen;
  lsGen gen3;
  lsHandle handle;
  int not_found;
  int num_edge;
  int index;

  new_aut = Sugar_AutomatonCreate();

  Sugar_Vertex_t* new_node;
  /* for each node in the old automaton, create a corresponding node in the new
     automaton */
  lsForEachItem(aut->state_list, gen, node){
    new_node = SugarAutomatonNewNode(new_aut);
    node->destination = (array_t*)new_node;
  }

#ifdef DEBUG_SUGAR
  fprintf(nusmv_stdout, "** SugarAutomatonNegateNFA\n* automaton to negate: \n");
  Sugar_AutomatonPrint(aut,3);
#endif

  /* add a node for completion */
  Sugar_Vertex_t *complete_node = SugarAutomatonNewNode(new_aut); /* node for completion */
  complete_node->name = (char*)0;
  /* create the TRUE label */

  Sugar_Label_t* truelabel = ALLOC(Sugar_Label_t, 1);
  truelabel->name = NIL(char); truelabel->isNegate=0;

  /* create the selfloop (edge) */
  Sugar_Edge_t *selfloop = ALLOC(Sugar_Edge_t, 1);
  array_t *true_labels = array_alloc(Sugar_Label_t*, 1);
  array_insert(Sugar_Label_t*, true_labels, 0, truelabel);
  lsList dest_vertex = lsCreate();
  lsNewEnd(dest_vertex, (lsGeneric)complete_node, LS_NH);
  selfloop->label = true_labels;
  selfloop->source_vertex=complete_node;
  selfloop->dest_vertex_list=dest_vertex;
  /* add to the nodelists */
  lsNewEnd(complete_node->edge_in_list, (lsGeneric)selfloop, LS_NH);
  lsNewEnd(complete_node->edge_out_list, (lsGeneric)selfloop, LS_NH);
  /* add to the automaton */
  lsNewEnd(new_aut->edge_list, (lsGeneric)selfloop, &handle);
  selfloop->handle=handle;
  
  /*mark the accepting states of the old automaton */
  lsForEachItem(aut->state_list, gen, node){
    node->name = (char*)0;
  }
  arrayForEachItem(Sugar_Vertex_t*, aut->accept_states, index, node){
    node->name = (char*)1;
  }
  
  new_acc_array = array_alloc(Sugar_Vertex_t*, 0);
  array_insert_last(Sugar_Vertex_t*, new_acc_array, complete_node);


  Sugar_Vertex_t* printnode;
  lsGen printgen;
  
  /* print the nodes of the new automaton */
  /*   fprintf(nusmv_stdout, "111111111111111\nStates:\n"); */
  /*   lsForEachItem(new_aut->state_list, printgen, printnode) { */
  /*     fprintf(nusmv_stdout, "  q%d:   IN %d / OUT %d\n", printnode->id, */
  /* 	    lsLength(printnode->edge_in_list), lsLength(printnode->edge_out_list)); */
  /*   } */

  
  /*create the new  edges*/
  lsForEachItem(aut->state_list, gen, node){
/*     fprintf(nusmv_stdout, "created edges for old node %i\n", node->id); */
    
    /* adjust the accepting array*/
    if(!((int)node->name))
      array_insert_last(Sugar_Vertex_t*, new_acc_array, (Sugar_Vertex_t *)node->destination);
    
    /* collect used labels */
    node_label_array = array_alloc(Sugar_Label_t*, 0);
    lsForEachItem(node->edge_out_list, innergen, edge){
      int succ=array_append(node_label_array, edge->label);
      assert(succ!=ARRAY_OUT_OF_MEM);
    }
    array_sort(node_label_array, (int (*)()) compareLabel);
    array_uniq(node_label_array, (int (*)()) compareLabel, NULL);
    

    int tfremove=0;
    
    if(array_n(node_label_array)>1){ /* remove T/F props */
      if(array_fetch(Sugar_Label_t*, node_label_array, 0)->name==NIL(char)){
/* 	fprintf(nusmv_stdout, "remove tflabel for !"); */
/* 	printLabelArray(node_label_array); */
	tfremove=1;
      }
    }
    /*create edge for each assignment*/
    new_edges_list = lsCreate();
    num_edge = bitRange(array_n(node_label_array)-tfremove);
    while(num_edge > 0){
      lsNewEnd(new_edges_list, (lsGeneric)ALLOC(Sugar_Edge_t, 1), LS_NH);
      num_edge--;      
    }

    index=0;
    lsForEachItem(new_edges_list, innergen, edge) {
      /* set labels for new edges */
      int edgebits = array_n(node_label_array)-tfremove;
      array_t *new_labels = array_alloc(Sugar_Label_t*, 0);
      while(edgebits>0){
	edgebits--;
	Sugar_Label_t* new_label = ALLOC(Sugar_Label_t, 1);
	array_insert_last(Sugar_Label_t*, new_labels, new_label);
	new_label->name = array_fetch(Sugar_Label_t*, node_label_array, edgebits+tfremove)->name;
	if((index>>edgebits)&1==1)
	  new_label->isNegate=0;
	else
	  new_label->isNegate=1;
      }
      if(array_n(new_labels)==0)
	fprintf(nusmv_stdout, "error!!!!!!!!!!!!!!!!!!!!1\n*!*!*!**!*!*!*!\n*!\n*!\n*!\nempty label");
      edge->label=new_labels;
      edge->source_vertex = (Sugar_Vertex_t* )node->destination;
      edge->dest_vertex_list = lsCreate();
      lsNewEnd(new_aut->edge_list, (lsGeneric) edge, &handle);
      edge->handle=handle;

      /* set the destination states for the new edges */
      int assigned_node=0;
      lsForEachItem(node->edge_out_list, gen3, oldedge){
	if(!conflict(edge->label, oldedge->label)){
/* 	  fprintf(nusmv_stdout, "no conflict between"); */
/* 	  printLabelArray(edge->label); */
/* 	  printLabelArray(oldedge->label);	     */
	  
	  assigned_node=1;
	  Sugar_Vertex_t *destination;
	  lsFirstItem(oldedge->dest_vertex_list, (lsGeneric *) &destination, LS_NH);
	  destination = ((Sugar_Vertex_t*)destination->destination);
	  lsNewEnd(destination->edge_in_list, (lsGeneric)edge, LS_NH);
	  lsNewEnd(edge->dest_vertex_list, (lsGeneric)destination, LS_NH);
	}
/* 	else{ */
/* 	  fprintf(nusmv_stdout, "conflict between"); */
/* 	  printLabelArray(edge->label); */
/* 	  printLabelArray(oldedge->label);	     */
/* 	} */

      }
      if(!assigned_node){
	lsNewEnd(complete_node->edge_in_list, (lsGeneric)edge, LS_NH);
	lsNewEnd(edge->dest_vertex_list, (lsGeneric)complete_node, LS_NH);
      }
	
#ifdef DEBUG_SUGAR
      printEdge(edge);
#endif
      index++;
    }
    ((Sugar_Vertex_t*)node->destination)->edge_out_list = new_edges_list;
    array_free(node_label_array);
  }
  array_free(aut->accept_states);
  new_aut->accept_states = new_acc_array;

  /*create the initial states*/
  array_t *initial_states;
  initial_states = array_alloc(Sugar_Vertex_t*, 0);
  array_t *initial_item;
  arrayForEachItem(array_t*, aut->initial_states, index, initial_item){
    array_append(initial_states, initial_item);    
  }
  array_sort(initial_states, (int(*)())compareVertex);
  array_uniq(initial_states, (int(*)())compareVertex, NULL);

  array_t* new_initial = array_alloc(Sugar_Vertex_t*, 0);
  arrayForEachItem(Sugar_Vertex_t*, initial_states, index, node){
    array_insert_last(Sugar_Vertex_t*, new_initial, (Sugar_Vertex_t*) node->destination);    
  }
  array_free(initial_states);

  array_free(new_aut->initial_states);
  new_aut->initial_states=array_alloc(array_t*, 0);
  array_insert_last(array_t*, new_aut->initial_states, new_initial);

#ifdef DEBUG_SUGAR
  fprintf(nusmv_stdout, "\033[01;34m\033[07m*********************\n** result of negation\033[0m\033[22;34m\n");
  Sugar_AutomatonPrint(new_aut, 3);
  fprintf(nusmv_stdout, "\033[0m");
#endif  

  return new_aut;
}

static int conflict(array_t* set1, array_t* set2){
  return conflictc(set1, set2, NULL);
}

/**Function********************************************************************

Synopsis    [ precond: we assume that partset is a subset of fullset]

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
/* static int conflict(array_t* fullset, array_t* partset){ */
/*   int index, index2=0, comparebuf, size; */
/*   Sugar_Label_t* label; */
  
/*   fullset = array_dup(fullset); */
/*   partset = array_dup(partset); */
/*   array_sort(fullset, (int (*) ()) compareLabel); */
/*   array_sort(partset, (int (*) ()) compareLabel); */


/*   size=array_n(fullset); */
/*   arrayForEachItem(Sugar_Label_t*, partset, index, label){ */
/*     if(label->name==NULL){ */
/*       return label->isNegate; */
/*     } */
    
/*     while(index2 < size){ */
/*       if(compareLabel(&label, &array_fetch(Sugar_Label_t*, fullset, index2))==0) */
/* 	break; */
/*       index2++; */
/*     } */

/*     if(index2<size){ */
/*        if(!(label->isNegate)!=  */
/*  	 !(array_fetch(Sugar_Label_t*, fullset, index2)->isNegate)){  */
/*  	return 1;	  */
/*        }  */
/*     } */
/*   } */
/*   return 0; */
/* } */
static int conflictc(array_t* set1, array_t* set2, array_t* combined_set){
  int index =0, s1index=0, s2index=0, index2=0, comparebuf, s1size, s2size;
  Sugar_Label_t* label;
  array_t* fullset;

  set1 = array_dup(set1);
  set2 = array_dup(set2);
  array_sort(set1, (int (*) ()) compareLabel);
  array_sort(set2, (int (*) ()) compareLabel);
  fullset=array_dup(set1);
/*   fprintf(nusmv_stdout, "set1: ");  */
/*   printLabelArray(set1);  */
/*   fprintf(nusmv_stdout, "set2: ");  */
/*   printLabelArray(set2);    */
  
  array_append(fullset, set2);
  array_sort(fullset, (int (*)()) compareLabel);
  array_uniq(fullset, (int (*)()) compareLabel, NULL);
/*   fprintf(nusmv_stdout, "combined: ");  */
/*   printLabelArray(fullset);    */

  s1size=array_n(set1);
  s2size=array_n(set2);
  
  arrayForEachItem(Sugar_Label_t*, fullset, index, label){
    if(combined_set != NULL){
      Sugar_Label_t * new_label = ALLOC(Sugar_Label_t, 1);
      new_label->name = label->name;
      new_label->isNegate = label->isNegate;
      array_insert_last(Sugar_Label_t*, combined_set, new_label);
    }

    if(label->name==NULL)
      return label->isNegate;
    
    while(s1index < s1size){
      if (compareLabel(&label, &array_fetch(Sugar_Label_t*, set1, s1index))<=0){
	//	fprintf(nusmv_stdout, "found label %s at set 1.%i\n", array_fetch(Sugar_Label_t*, set1, s1index)->name, s1index);
	break;
      }
      s1index++;
    }

    while(s2index < s2size){
      if (compareLabel(&label, &array_fetch(Sugar_Label_t*, set2, s2index))<=0){
	//	fprintf(nusmv_stdout, "found label %s at set 2.%i\n", array_fetch(Sugar_Label_t*, set2, s2index)->name, s2index);
	break;
      }
      s2index++;
    }
        
    if(s1index<s1size && s2index<s2size){
      Sugar_Label_t *l1 = array_fetch(Sugar_Label_t*, set1, s1index);
      Sugar_Label_t *l2 = array_fetch(Sugar_Label_t*, set2, s2index);
      
      if(strcmp(l1->name, l2->name)==0){
	if(l1->isNegate != l2->isNegate){
	  return 1;
	}
      }
    }
  }
  if(combined_set != NULL){
    array_sort(combined_set, (int (*)()) compareLabel);
    array_uniq(combined_set, (int (*)()) compareLabel, NULL);
  }
  array_free(fullset);
  array_free(set1);
  array_free(set2);
  return 0;
}


/**Function********************************************************************

Synopsis    []

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
static int bitRange(int x){
  int result = (x>0);

  while(x > 0){
    result *= 2;
    x--;
  }
  return result;
}

/**Function********************************************************************

Synopsis    [Generate automaton for negated formula]

Description [Simply moves the negation of the fomula inside. For example
AND and OR operator use DeMorgan's Law. For LTL operators there exist
quite the same laws. Details can be found in the corresponding master
thesis of the author. Negation of a negation generates the basic
forumula again. For SEREs the automaton is constructed an then negated.]

SideEffects []

SeeAlso     [SugarAutomatonNegate, Sugar_AutomatonGenerate]

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonNot(
  Sugar_Formula_t *formula)
{
  int nStr;
  int index;
  array_t *new_ini_array;
  Sugar_Edge_t *edge;
  Sugar_Vertex_t *node;
  Sugar_Automaton_t *A_temp;
  Sugar_Automaton_t *A_left_neg;
  Sugar_Automaton_t *A_right_neg;
  Sugar_Automaton_t *A_left_clone;
  Sugar_Automaton_t *A_right_clone;
  lsGen gen;

  /* no formula returns no automaton */
  if (formula == NIL(Sugar_Formula_t)) {
    return NIL(Sugar_Automaton_t);
  }
  /* type TRUE returns single edge with label FALSE */
  if (formula->type == Sugar_TRUE_c) {
    return SugarAutomatonIdentifier(NIL(char), 1, formula->isStrongOperator);
  }
  /* type FALSE returns single edge with label TRUE */
  if (formula->type == Sugar_FALSE_c) {
    return SugarAutomatonIdentifier(NIL(char), 0, formula->isStrongOperator);
  }
  /* type IDENTIFIER returns single edge with !IDENTIFIER */
  if (formula->type == Sugar_IDENTIFIER_c) {
    return SugarAutomatonIdentifier((char*)(formula->left), 1, formula->isStrongOperator);
  }
  /* type NEGATION returns the formula again  --  !(!f) = f */
  if (formula->type == Sugar_FL_Not_c) {
    return Sugar_AutomatonGenerate(formula->left);
  }
  /* type NEVER f returns EVENTUALLY f  --  !(G !f) = F f */
  if (formula->type == Sugar_FL_Never_c) {
    A_temp = Sugar_AutomatonGenerate(formula->left);
    return SugarAutomatonEventually(A_temp);
  }

  /*
   * check if we have a formula or an SERE
   */
  if (formula->type < Sugar_Formula_sep) {
    /*
     * The formula is an inner node: create negated left and right automata
     */
    A_left_neg = SugarAutomatonNot(formula->left);
    A_right_neg = SugarAutomatonNot(formula->right);
    /*
     * Create negation of formula by pushing the negation inside and generate
     * the automaton with the dual operator. (Law of DeMorgan for Booleans)
     */
    switch(formula->type) {
    case Sugar_FL_Next_c:
      nStr = formula->numRangeStr;
      if (nStr == SUGAR_RANGE_UNUSED) nStr = 1;
      /* create automata by inserting TRUE edges at the start */
      while (nStr-- > 0) {
	A_temp = SugarAutomatonIdentifier(NIL(char), 0, 0);
	SugarAutomatonConcat(A_temp, A_left_neg, 0);
	/* weak version may not have next cycle -> new initial state is accepting */
	if (!formula->isStrongOperator) {
	  new_ini_array = array_fetch(array_t*, A_temp->initial_states, 0);
	  node = array_fetch(Sugar_Vertex_t*, new_ini_array, 0);
	  array_insert_last(Sugar_Vertex_t*, A_temp->accept_states, node);
	}
	A_left_neg = A_temp;
      }
      break;
    case Sugar_FL_Always_c:
      /* create !(G f) with (F !f) */
      A_left_neg = SugarAutomatonEventually(A_left_neg);
      break;
    case Sugar_FL_Eventually_c:
      /* create !(F f) with (G !f) */
      A_left_neg = SugarAutomatonAlways(A_left_neg);
      break;
    case Sugar_FL_Releases_c:
      /* create !(f R g) with (!f U! !g) */
      A_left_neg = SugarAutomatonUntil(A_left_neg, A_right_neg, 1); // strong
      break;
    case Sugar_FL_Until_c:
      /* create !(f U! g) with [(!g U! (!g and !f)) or G !g] */
      A_temp = SugarAutomatonNot(formula->right);
      SugarAutomatonAndEq(A_left_neg, A_temp);
      A_left_neg = SugarAutomatonUntil(A_right_neg, A_left_neg, 0); // weak
      /* create !(f U g) with [((!g U! (!g and !f)) or G !g) and F !f] */
      if (!formula->isStrongOperator) {
	A_temp = SugarAutomatonNot(formula->left);
	A_temp = SugarAutomatonEventually(A_temp);
	SugarAutomatonAndEq(A_left_neg, A_temp);
      }
      break;
    case Sugar_FL_Until__c:
      /* create !(f U!_ g) with [((!f or !g) U! !f) or G(!f or !g)] */
      SugarAutomatonOr(A_left_neg, A_right_neg);
      A_temp = SugarAutomatonNot(formula->left);
      A_left_neg = SugarAutomatonUntil(A_left_neg, A_temp, 0); // weak
      /* create !(f U_ g) with [((!f or !g) U! !f) or G(!f or !g) and F !f] */
      if (!formula->isStrongOperator) {
	A_temp = SugarAutomatonNot(formula->left);
	A_temp = SugarAutomatonEventually(A_temp);
	SugarAutomatonAndEq(A_left_neg, A_temp);
      }
      break;
    case Sugar_FL_Before_c:
      fprintf(nusmv_stdout, "Construct 'before' not implemented yet! IGNORED!\n");
      break;
    case Sugar_FL_Before__c:
      fprintf(nusmv_stdout, "Construct 'before_' not implemented yet! IGNORED!\n");
      break;
    case Sugar_FL_Not_c:
      A_left_neg = Sugar_AutomatonGenerate(formula->left);
      break;
    case Sugar_FL_And_c: 
      SugarAutomatonOr(A_left_neg, A_right_neg);
      break;
    case Sugar_FL_Or_c:
      SugarAutomatonAndEq(A_left_neg, A_right_neg);
      break;
    case Sugar_FL_Imply_c:
      /* create !(f->g) with (f and !g) */
      Sugar_AutomatonFree(A_left_neg);
      A_left_neg = Sugar_AutomatonGenerate(formula->left);
      SugarAutomatonAndEq(A_left_neg, A_right_neg);
      break;
    case Sugar_FL_Equiv_c:
      /* create !(f<->g) with (f or g) and (!f or !g) */
      SugarAutomatonOr(A_left_neg, A_right_neg);
      A_left_clone = Sugar_AutomatonGenerate(formula->left);
      A_right_clone = Sugar_AutomatonGenerate(formula->right);
      SugarAutomatonOr(A_left_clone, A_right_clone);
      SugarAutomatonAndEq(A_left_neg, A_left_clone);
      break;
    default:
      fail("Unexpected type");
    }
  }
  else {
    /*
     * check for suffix implication types
     */
    if (formula->type < Sugar_Sequence_sep) {
      if (formula->type == Sugar_Seq_FLProp_c) {
        /* build !({r}(f)) with {r}:{!f} */
        A_temp = Sugar_AutomatonGenerate(formula->left);
        A_left_neg = Sugar_AutomatonDeterministic(A_temp);
        Sugar_AutomatonFree(A_temp);
/* 	fprintf(nusmv_stdout, "after autofree\n"); */
        A_temp = SugarAutomatonNot(formula->right);
/* 	fprintf(nusmv_stdout, "**negated right:\n"); */
/* 	Sugar_AutomatonPrint(A_temp, 3); */
        A_right_neg = Sugar_AutomatonExpand(A_temp);
        Sugar_AutomatonFree(A_temp);
        SugarAutomatonConcat(A_left_neg, A_right_neg, 1);
/* 	fprintf(nusmv_stdout, "after concat\n"); */
/* 	Sugar_AutomatonPrint(A_left_neg, 3); */
      }
      else {
        switch (formula->type) {
	case Sugar_Seq_Always_c:
	  /* build !(always{r}) with {T+}:{neg(r;T*)} */
	  A_left_neg = SugarAutomatonIdentifier(NIL(char), 0, 0);
	  SugarAutomatonPlus(A_left_neg);
	  A_right_neg = Sugar_AutomatonGenerate(formula->left);
	  break;
	case Sugar_Seq_Eventually_c:
	  /* build !(eventually!{r}) with {T}:{neg(T*;r;T*)} */
	  A_left_neg = SugarAutomatonIdentifier(NIL(char), 0, 0);
	  A_right_neg = SugarAutomatonIdentifier(NIL(char), 0, 0);
	  SugarAutomatonStar(A_right_neg);
	  A_temp = Sugar_AutomatonGenerate(formula->left);
	  SugarAutomatonConcat(A_right_neg, A_temp, 0);
	  break;
	case Sugar_Seq_Never_c:
	  /* buil !(never{r}) with {T*;r}:{neg(F;T*)} */
	  A_left_neg = SugarAutomatonIdentifier(NIL(char), 0, 0);
	  SugarAutomatonStar(A_left_neg);
	  A_temp = Sugar_AutomatonGenerate(formula->left);
	  SugarAutomatonConcat(A_left_neg, A_temp, 0);
	  A_right_neg = SugarAutomatonIdentifier(NIL(char), 1, 0);
	  break;
	case Sugar_Seq_Imply_c:
	  /* build !({r1}|->{r2}!) with {r1}:{neg{r2;T*}} */
	  A_left_neg = Sugar_AutomatonGenerate(formula->left);
	  A_right_neg = Sugar_AutomatonGenerate(formula->right);
	  break;
	case Sugar_Seq_Imply2_c:
	  /* build !({r1}|=>{r2}!) with {r1}:{neg{T;r2;T*} */
	  A_left_neg = Sugar_AutomatonGenerate(formula->left);
	  A_right_neg = SugarAutomatonIdentifier(NIL(char), 0, 0);
	  A_temp = Sugar_AutomatonGenerate(formula->right);
	  SugarAutomatonConcat(A_right_neg, A_temp, 0);
	  break;
	default:
	  fail("Unexpected type");
        }
        /* add self loops to accepting states of former right automaton */
        arrayForEachItem(Sugar_Vertex_t*, A_right_neg->accept_states, index, node) {
          if (lsLength(node->edge_out_list) > 0) {
            /* delete outgoing edges from accepting state */
            lsForEachItem(node->edge_out_list, gen, edge) {
              SugarAutomatonFreeEdge(edge);
            }
          }
          SugarAutomatonNewEdgeLoop(A_right_neg, node);
        }
        /* the final construction {r1}:{neg{r2;T*}} starts here */
        A_temp = SugarAutomatonNegate(A_right_neg);
        Sugar_AutomatonFree(A_right_neg); A_right_neg = A_temp;
        /* r1 needs to be determinisitic and r2 non-determinisitic */
        A_left_clone = Sugar_AutomatonDeterministic(A_left_neg);
        Sugar_AutomatonFree(A_left_neg);
        A_right_clone = Sugar_AutomatonExpand(A_right_neg);
        Sugar_AutomatonFree(A_right_neg);
        SugarAutomatonConcat(A_left_clone, A_right_clone, 1);
        A_left_neg = A_left_clone;
      }
    }
    else {
      /*
       * For SEREs build the automaton and negate the network
       */
      A_temp = Sugar_AutomatonGenerate(formula);
      A_left_neg = SugarAutomatonNegate(A_temp);
      Sugar_AutomatonFree(A_temp);
    }
  }
  SugarAutomatonFlush(A_left_neg);
  return A_left_neg;
}

/**Function********************************************************************

Synopsis    [Generate automaton for formula ALWAYS]

Description [Adds a TRUE loop with alternating connection to the start
of the alternating automaton.]

SideEffects [The old automaton gets deleted. The new automaton should be
freed by the caller.]

SeeAlso     []

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonAlways(
  Sugar_Automaton_t *aut)
{
  Sugar_Automaton_t *aut_new;

  /* build Automaton as (T[+]):(r) but with alternate edges */
  aut_new = SugarAutomatonIdentifier(NIL(char), 0, 0);
  SugarAutomatonPlus(aut_new);
  SugarAutomatonConcat(aut_new, aut, 3);
  return aut_new;
}

/**Function********************************************************************

Synopsis    [Generate automaton for formula EVENTUALLY!]

Description [Adds a TRUE loop to the start of the alternating automaton.]

SideEffects [The old automaton gets deleted. The new automaton should be
freed by the caller.]

SeeAlso     []

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonEventually(
  Sugar_Automaton_t *aut)
{
  Sugar_Automaton_t *aut_new;

  /* build automaton as (T[+]):(r) */
  aut_new = SugarAutomatonIdentifier(NIL(char), 0, 0);
  SugarAutomatonPlus(aut_new);
  SugarAutomatonConcat(aut_new, aut, 1);
  return aut_new;
}

/**Function********************************************************************

Synopsis    [Generate automaton for formula UNTIL]

Description [Add the left automaton with star operation to the start of
the right automaton.]

SideEffects [Both old automata gets deleted. The new automaton should be
freed by the caller.]

SeeAlso     []

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonUntil(
  Sugar_Automaton_t *aut_left,
  Sugar_Automaton_t *aut_right,
  int isStrong)
{
  int index;
  int not_found;
  array_t *temp;
  array_t *new_ini_array;
  array_t *new_acc_array;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t* init_node;
  Sugar_Vertex_t* loop_node;
  Sugar_Automaton_t *aut_true_plus;

  /* first build Automaton T[+] */
  aut_true_plus = SugarAutomatonIdentifier(NIL(char), 0, 0);
  SugarAutomatonPlus(aut_true_plus);
  /* backup states of T[+] (initial and accepting) and get looping state */
  new_ini_array = array_dup(array_fetch(array_t*, aut_true_plus->initial_states, 0));
  array_append(new_ini_array, aut_true_plus->accept_states);
  init_node = array_fetch(Sugar_Vertex_t*, new_ini_array, 0);
  loop_node = array_fetch(Sugar_Vertex_t*, new_ini_array, 1);
  /* build Automaton as (T[+]):(r1) but with alternate edges */
  SugarAutomatonConcat(aut_true_plus, aut_left, 2);
  /* set initial states to backuped states for ... */
  new_acc_array = array_dup(aut_true_plus->accept_states);
  array_free(aut_true_plus->accept_states);
  aut_true_plus->accept_states = new_ini_array;
  /* ... normal concatination with r2, then ... */
  SugarAutomatonConcat(aut_true_plus, aut_right, 0);
  array_append(new_acc_array, aut_true_plus->accept_states);
  /* always remove new initial state from accepting states */
  temp = array_alloc(Sugar_Vertex_t*, 0);
  arrayForEachItem(Sugar_Vertex_t*, new_acc_array, index, node) {
    if (node != init_node) {
      array_insert_last(Sugar_Vertex_t*, temp, node);
    }
  }
  array_free(new_acc_array);
  new_acc_array = temp;
  /* for strong version ensure that looping state is NOT accepting */
  if (isStrong) {
    temp = array_alloc(Sugar_Vertex_t*, 0);
    arrayForEachItem(Sugar_Vertex_t*, new_acc_array, index, node) {
      if (node != loop_node) {
        array_insert_last(Sugar_Vertex_t*, temp, node);
      }
    }
    array_free(new_acc_array);
    new_acc_array = temp;
  }
  else {
    /* for weak version ensure that looping state IS accepting */
    not_found = 1;
    arrayForEachItem(Sugar_Vertex_t*, new_acc_array, index, node) {
      if (node == loop_node) { not_found = 0; break; }
    }
    if (not_found) {
      array_insert_last(Sugar_Vertex_t*, new_acc_array, loop_node);
    }
  }
  /* ... set new accepting states */
  array_free(aut_true_plus->accept_states);
  aut_true_plus->accept_states = new_acc_array;
  return aut_true_plus;
}

/**Function********************************************************************

Synopsis    [Generate automaton for formula r1|->r2!]

Description [Details about the construction can be found in the corresponding
master thesis of the author.]

SideEffects [Both old automata gets deleted. The returned automaton should
be freed by the caller.]

SeeAlso     []

******************************************************************************/
Sugar_Automaton_t*
SugarAutomatonImply(
  Sugar_Automaton_t *aut_left,
  Sugar_Automaton_t *aut_right,
  int isRightNoFormula,
  int isStrong)
{
  int index;
  array_t *dest_array;
  array_t *labl_array;
  array_t *new_acc_array;
  array_t *edge_false_array;
  Sugar_Edge_t *edge;
  Sugar_Label_t *labl;
  Sugar_Vertex_t *node;
  Sugar_Vertex_t *abort_node;
  Sugar_Automaton_t *aut_exp;
  Sugar_Automaton_t *aut_left_det;
  lsGen gen;

  /* first make left automaton deterministic (sigh) ... */
  if (aut_right != NIL(Sugar_Automaton_t)) {
    /* fprintf(nusmv_stdout, "*\n* pre autdet\n"); */
    aut_left_det = Sugar_AutomatonDeterministic(aut_left);
    /* fprintf(nusmv_stdout, "*\n* pre autfree\n"); */
    Sugar_AutomatonFree(aut_left);
    /* fprintf(nusmv_stdout, "*\n* post autfree\n"); */
  }
  else {
    /* ... but not for negation: we just need the abort edges */
    aut_left_det = aut_left; aut_left = NIL(Sugar_Automaton_t);
  }
  /* backup all states of left automaton to add abort edges */
  new_acc_array = array_alloc(Sugar_Vertex_t*, 0);
  lsForEachItem(aut_left_det->state_list, gen, node) {
    array_insert_last(Sugar_Vertex_t*, new_acc_array, node);
  }
  //  fprintf(nusmv_stdout, "*\n* mark0.1\n");
  /* add FALSE loops to accepting states if they have no outgoing edges */
  labl = ALLOC(Sugar_Label_t, 1);
  labl->name = NIL(char); labl->isNegate = 1;
  labl_array = array_alloc(Sugar_Label_t*, 1);
  array_insert(Sugar_Label_t*, labl_array, 0, labl);
  dest_array = array_alloc(Sugar_Vertex_t*, 1);
  edge_false_array = array_alloc(Sugar_Edge_t*, 0);
  //  fprintf(nusmv_stdout, "*\n* mark0.2\n");
  arrayForEachItem(Sugar_Vertex_t*, aut_left_det->accept_states, index, node) {
    if (lsLength(node->edge_out_list) == 0) {
      array_insert(Sugar_Vertex_t*, dest_array, 0, node);
      edge = SugarAutomatonNewEdge(aut_left_det, node, dest_array, labl_array, 0);
      array_insert_last(Sugar_Edge_t*, edge_false_array, edge);
    }
  }
  FREE(labl);
  array_free(labl_array);
  array_free(dest_array);
  /* create abort state with TRUE loop and add abort edges */
  abort_node = SugarAutomatonNewNode(aut_left_det);
  SugarAutomatonNewEdgeLoop(aut_left_det, abort_node);
  //  fprintf(nusmv_stdout, "*\n* mark0.3 add abort for %i states\n", array_n(new_acc_array));
  arrayForEachItem(Sugar_Vertex_t*, new_acc_array, index, node) {
    /*     fprintf(nusmv_stdout, "%i \n", index); */
    /*     printNode(node); */
    SugarAutomatonAddAbortEdge(aut_left_det, node, abort_node);
  }
  //    fprintf(nusmv_stdout, "*\n* mark0.4\n");
  /* delete previously added FALSE loops (if any created) */
  arrayForEachItem(Sugar_Edge_t*, edge_false_array, index, edge) {
    SugarAutomatonFreeEdge(edge);
  }
  array_free(edge_false_array);
  //  fprintf(nusmv_stdout, "*\n* mark1\n");

  /* for negation we just need the abort edges */
  if (aut_right == NIL(Sugar_Automaton_t)) {
    array_free(new_acc_array);
    array_insert_last(Sugar_Vertex_t*, aut_left_det->accept_states, abort_node);
  }
  else {
    /* make right automaton non-determinstic (remove alternation) */
    aut_exp = Sugar_AutomatonExpand(aut_right);
    Sugar_AutomatonFree(aut_right); aut_right = aut_exp;
    /* combine left and right automaton as {r1}:{r2} with alternate edges */
    SugarAutomatonConcat(aut_left_det, aut_right, 4);
    /* need not to extend formula of {r}(f) for infinite words */
    if (isRightNoFormula) {
      /* add self loops to accepting states of former right automaton */
      arrayForEachItem(Sugar_Vertex_t*, aut_left_det->accept_states, index, node) {
        if (lsLength(node->edge_out_list) > 0) {
          /* delete outgoing edges from accepting state */
          lsForEachItem(node->edge_out_list, gen, edge) {
            SugarAutomatonFreeEdge(edge);
          }
        }
	//	fprintf(nusmv_stdout, "*\n* mark2\n");
        SugarAutomatonNewEdgeLoop(aut_left_det, node);
      }
    }
    /* set accepting states: accept = new_accept + abort_node + accept */
    if (isStrong) {
      array_insert_last(Sugar_Vertex_t*, new_acc_array, abort_node);
      array_append(new_acc_array, aut_left_det->accept_states);
    }
    else {
      /*==--
	/* TODO: check semantics for weak version for a proper implementation
	/*==--
	/* for now weak version also states in right automaton are accepting */
      array_free(new_acc_array);
      new_acc_array = array_alloc(Sugar_Vertex_t*, 0);
      lsForEachItem(aut_left_det->state_list, gen, node) {
        array_insert_last(Sugar_Vertex_t*, new_acc_array, node);
        SugarAutomatonAddAbortEdge(aut_left_det, node, abort_node);
      }
    }
    array_free(aut_left_det->accept_states);
    aut_left_det->accept_states = new_acc_array;
  }
  return aut_left_det;
}

/**Function********************************************************************

Synopsis    []

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
static int lsCompareVertex(lsGeneric obj1, lsGeneric obj2){
  if(obj1==NULL || obj2==NULL)
    return 0;
  return ((Sugar_Vertex_t *)obj1)->id - ((Sugar_Vertex_t *) obj2)->id;
}

/**Function********************************************************************

Synopsis    []

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
static int compareEdge(void *obj1, void* obj2){
  return (*(Sugar_Edge_t **)obj1) - (*(Sugar_Edge_t **)obj2);
}

/**Function********************************************************************

Synopsis    []

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
static int compareVertex(void *obj1, void* obj2){
  if(obj1==NULL || obj2==NULL)
    return 0;
  return (*(Sugar_Vertex_t **)obj1)->id - (*(Sugar_Vertex_t **) obj2)->id;
}

/**Function********************************************************************

Synopsis    []

Description []

SideEffects []

SeeAlso     []

******************************************************************************/
static int compareLabel(void *obj1, void* obj2){
  Sugar_Label_t* l1 = *((Sugar_Label_t**)obj1);
  Sugar_Label_t* l2 = *((Sugar_Label_t**)obj2);

  if(l1->name == NULL  && l2->name == NULL){ //TRUE or FALSE
    if(l1->isNegate && ! l2->isNegate)
      return -1;
    if(!l1->isNegate && l1->isNegate)
      return 1;
    return 0;
  }
  if(l1->name == NULL)
    return -1;
  if(l2->name == NULL)
    return 1;
  
  return strcmp(l1->name, l2->name);
}

/******************************************************************************/
/* Debugging output functions
   /******************************************************************************/

#ifdef DEBUG_SUGAR

void printNode(Sugar_Vertex_t* node)
{
  Sugar_Edge_t* edge;
  lsGen gen;

  printf("Node at 0x%08x\n", node);
  printf("* ID: %d\n", node->id);
  printf("* Edge IN list:  (%d)\n", lsLength(node->edge_in_list));
  lsForEachItem(node->edge_in_list, gen, edge) { printf("   0x%08x\n", edge); }
  printf("* Edge OUT list: (%d)\n", lsLength(node->edge_out_list));
  lsForEachItem(node->edge_out_list, gen, edge) { printf("   0x%08x\n", edge); }
  printf("\n");
}

void printEdgeList(lsList edgelist){
  lsGen gen1, gen2;
  Sugar_Edge_t* edge;
  Sugar_Vertex_t* node;
  int first, index1;
  Sugar_Label_t* labl;
  
  fprintf(nusmv_stdout, " Edgelist\n");
  
  lsForEachItem(edgelist, gen1, edge) {
    first = 1;
    fprintf(nusmv_stdout, "    q%d", edge->source_vertex->id);
    lsForEachItem(edge->dest_vertex_list, gen2, node) {
      fprintf(nusmv_stdout, "%sq%d", (first)?(" -> "):(","), node->id);
      first = 0;
    }
    first = 1;
    fprintf(nusmv_stdout, "  Label: ");
    arrayForEachItem(Sugar_Label_t*, edge->label, index1, labl) {
      if (labl->name == NIL(char)) {
	fprintf(nusmv_stdout, "%s", (labl->isNegate)?"FALSE":"TRUE");
      }
      else {
	fprintf(nusmv_stdout, "%s%s'%s'", (first)?(""):("&"), (labl->isNegate)?("!"):(""), labl->name);
      }
      first = 0;
    }
    fprintf(nusmv_stdout, "\n");
  }
}

void printEdge(Sugar_Edge_t* edge)
{
  Sugar_Vertex_t* node;
  lsGen gen;

  printf("\nEdge at 0x%08x\n", edge);
  printf("* "); printLabelArray(edge->label);
  printf("* Source: 0x%08x (q%d)\n", edge->source_vertex, edge->source_vertex->id);
  printf("* Destination List: (%d)\n", lsLength(edge->dest_vertex_list));
  lsForEachItem(edge->dest_vertex_list, gen, node) { printf("    0x%08x (q%d)\n", node, node->id); }
  printf("\n");
}

void printArray(array_t* array, int isLabels)
{
  int index;
  array_t *item;

  printf("Array at 0x%08x\n", array);
  printf(" - Number of Elements:        %8d\n", array->num);
  printf(" - Size of 'data' array:      %8d\n", array->n_size);
  printf(" - Size of each array object: %8d\n", array->obj_size);

  arrayForEachItem(array_t*, array, index, item) {
    if (isLabels) {
      printf("    %2d: 0x%x08 (%s)\n", index, item, (char*)item);
    }
    else {
      printf("    %2d: 0x%x08\n", index, item);
    }
  }
  printf("\n");
}

void printNodeArray(array_t* array)
{
  int index;
  Sugar_Vertex_t* node;

  arrayForEachItem(Sugar_Vertex_t*, array, index, node) {
    printf(" q%d &", node->id);
  }
  printf("\b \n");
}

void printLabelArray(array_t* array)
{
  int index;
  Sugar_Label_t* labl;
  if(array_n(array)==0)
    printf(" label empty !! ");
    
  arrayForEachItem(Sugar_Label_t*, array, index, labl) {
    if (labl->name == NIL(char)) {
      printf(" %s ", (labl->isNegate)?("FALSE "):("TRUE "));
    }
    else {
      printf(" %s%s &", (labl->isNegate)?("!"):(""), labl->name);
    }
  }
  printf("\b \n");
}

void printEdgeArrayArray(array_t* array)
{
  int i1,i2,i3;
  array_t* temp;
  Sugar_Edge_t* edge;
  Sugar_Label_t* labl;

  arrayForEachItem(array_t*, array, i1, temp) {
    arrayForEachItem(Sugar_Edge_t*, temp, i2, edge) {
      arrayForEachItem(Sugar_Label_t*, edge->label, i3, labl) {
        if (labl->name == NIL(char)) {
          printf(" %s  ", (labl->isNegate)?("FALSE"):("TRUE"));
        }
        else {
          printf(" %s%s &", (labl->isNegate)?("!"):(""), labl->name);
        }
      }
    }
    printf("\b|");
  }
  printf("\b \n");
}

void printLabelArrayArray(array_t* array)
{
  int index1;
  int index2;
  array_t* temp;
  Sugar_Label_t* labl;

  arrayForEachItem(array_t*, array, index1, temp) {
    arrayForEachItem(Sugar_Label_t*, temp, index2, labl) {
      if (labl->name == NIL(char)) {
        printf(" %s", (labl->isNegate)?("FALSE "):("TRUE "));
      }
      else {
        printf(" %s%s |", (labl->isNegate)?("!"):(""), labl->name);
      }
    }
    printf("\b , ");
  }
  printf("\b\b \n");
}

void printNodeArrayArray(array_t* array)
{
  int index1;
  int index2;
  array_t* temp;
  Sugar_Vertex_t* node;

  arrayForEachItem(array_t*, array, index1, temp) {
    arrayForEachItem(Sugar_Vertex_t*, temp, index2, node) {
      printf(" q%d &", node->id);
    }
    printf("\b , ");
  }
  printf("\b\b \n");
}

/**Function********************************************************************

Synopsis [returns a array which containes all elements which are in
minuend but not in subtrahend. elements in the array are supposed to
be int]

Description []

SideEffects []

******************************************************************************/
/* replaced by array_sort, array_uniq
   static array_t*
   ArrayMerge(
   array_t *a1,
   array_t *a2)
   {
   int index1=0, index2=0, comparebuf, subsize;
   array_t *retval;

   a1 = array_dup(a1);
   a2 = array_dup(a2);
   retval = array_alloc(mdd_t*, array_n(minuend));

   a1 = array_dup(a1);
   a2 = array_dup(a2);

   array_sort(a1, (int (*)()) compare);
   array_sort(a2, (int (*)()) compare);
   /*
   PrintMddIdArrayAsNtkNode(minuend, Fsm_FsmReadNetwork(modelFsm));
   PrintMddIdArrayAsNtkNode(subtrahend, Fsm_FsmReadNetwork(modelFsm));
*/
/*
  size = array_n(a2);
  for(index1 = 0 ; index1 < array_n(a1) ; index1 ++){
  while((comparebuf = compare(&array_fetch(void *, a1, index1), 
  &array_fetch(void *, a2, index2)))>0
  && index2 < (size-1)) {
  index2 ++; 
  }

  if(comparebuf) {
  array_insert_last(int, retval, array_fetch(int, a1, index1));
  }
  }
  array_append(retval, a2);
  array_free(a1);
  array_free(a2);
  
  return retval;
  }
*/

#endif
