/**CHeaderFile*****************************************************************

  FileName    [sugarInt.h]

  PackageName [sugar]

  Synopsis    [Declarations for internal use.]

  Author      [Christian Josef Kargl]

  Copyright   [Copyright (c) 2003
               University of Technology Graz, Institute for Softwaretechnology]

  All rights reserved.

  Permission is hereby granted, without written agreement and without license
  or royalty fees, to use, copy, modify, and distribute this software and its
  documentation for any purpose, provided that the above copyright notice and
  the following two paragraphs appear in all copies of this software.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]

  Revision    [$Id: sugarInt.h,v 1.1 2005/12/22 23:59:08 timo Exp $]

******************************************************************************/

#ifndef _SUGARINT
#define _SUGARINT

#include <limits.h>
#include <string.h>
#include <math.h>
#include "sugar.h"
#include "utils/list.h" /*For lsList*/

/* #include "hrc.h" */
/* #include "cmd.h" */

/* #include "io.h" */
/* #include "fsm.h" */
/* #include "ntk.h" */
/* #include "ord.h" */
/* #include "part.h" */
/* #include "mc.h" */

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/
#define SUGAR_RANGE_UNUSED   (-1)
#define SUGAR_RANGE_MIN_VAL  (0)
#define SUGAR_RANGE_MAX_VAL  (INT_MAX)
#define MAX_LENGTH_OF_VAR_NAME 512 

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/**Struct**********************************************************************

  Synopsis    [Tree representing a SUGAR formula.]

  Description [Structure for a SUGAR formula. A formula is a tree. An
  internal vertex has one or two children, depending on the operator
  type: if the operator takes just one argument, then the left child
  holds the subformula, and the right child is NULL; if the operator
  takes two arguments, then both children are non-NULL. Identifier
  are special cases where the left child is a pointer to the string
  and the right child is NULL.
  
  <p>A leaf is either a boolean variable name in the network, TRUE or
  FALSE. The left and right fields are used in leaf vertices only in
  the first case above. (The left child holds a variable name while
  the right child is not used.)

  <p>Some regular expression operators have range values. The start
  value is always lower than the end value. Some formula operators
  have a weak and a strong version.]

******************************************************************************/
struct SugarFormulaStruct {
  Sugar_FormulaType type;
  Sugar_Formula_t *left;   /* left child in formula */
  Sugar_Formula_t *right;  /* right child, or null if not used */
  int numRangeStr;         /* start of range for some formulas, else -1 */
  int numRangeEnd;         /* end of range for some formulas, else -1 */
  int isStrongOperator;    /* non-zero if operator has a !, else zero */
  int refCount;            /* number of formulas referring to this
                              formula as a left or right child */       
};

/**Struct**********************************************************************

  Synopsis    [The alternating automaton for SUGAR formula]

******************************************************************************/
struct SugarAutomataStruct {
  char* name;              // the formula string
  lsList state_list;       // double linked list of vertices (Sugar_Vertex_t*)
  lsList prop_list;        // double linked list of prop. name (Sugar_Label_t*)
  lsList edge_list;        // double linked list of edges (Sugar_Edge_t*)
  array_t *initial_states; // positive boolean formula with inital states
  array_t *accept_states;  // array with pointer to vertices of accepting states
  int num_state_index;     // number count for index of new state
};

/**Struct**********************************************************************

  Synopsis    [An alternate edge for SUGAR]

  Description [This edge is used by the alternating automata by SUGAR.
  It has one source vertex and a list of destination vertices which are
  ANDed together. Similar for the label array. Each entry points to a
  proposition name and they are ANDed together.]

******************************************************************************/
struct SugarEdgeStruct {
  array_t *label;                // pointer to literals of labels (a&!b)
  Sugar_Vertex_t *source_vertex; // pointer to source vertex
  lsList dest_vertex_list;       // list of pointers to destination vertices
  lsHandle handle;               // for instant deletion operation
};

/**Struct**********************************************************************

  Synopsis    [A simple label for SUGAR]

  Description [The label has a pointer to the proposition name in the list
  of the alternating automata and a flag if the label should be negated. If
  the name pointer is NIL, the label is TRUE or FALSE. It is also used
  when creating a deterministic automaton to hold all outgoing labels of
  a node. Then the name is a string of the labels and instead of the flag
  we keep an array of all edges with that label.]

******************************************************************************/
struct SugarLabelStruct {
  char *name;   // pointer to name in prop_list of automaton, OR label string
                // of the from "LBL_$" + prop + "$" + ... + "$" + prop. 
  int isNegate; // flag if label is negated or not
};

/**Struct**********************************************************************

  Synopsis    [A simple vertex for SUGAR]

  Description [The vertex of the alternating automata for SUGAR. No special
  features, just a list to in-edges and a list for out-edges.]

******************************************************************************/
struct SugarVertexStruct {
  int id;               // number of vertex
  lsList edge_in_list;  // pointer to list of pointers to incoming edges
  lsList edge_out_list; // pointer to list of pointers to outgoing edges
  lsHandle handle;      // for instant deletion operation
  // ---- below are optional data for expanding alternation ----
  array_t *destination; // pointers to destination nodes in source automaton
  array_t *dest_accept; // pointers to already accepting destination nodes
  char *name;           // string of two pointer arrays above of the form
                        // "NODE_" + id + "_" + id + "_" + ... + "_" + id
                        // or "NODE@" + hex_address for converting to LTL
};

/**Struct**********************************************************************

  Synopsis    [A simlpe struct with two arrays.]

  Description [This struct is used while converting the nondeterministic
  automaton to a deterministic automaton. The first array holds the labels
  of the new edge and the second array holds the destination vertices.]

******************************************************************************/
struct SugarArrayPairStruct {
  array_t *label_array;
  array_t *dest_array;
};

static
char *util_strcat3(char *s1, char *s2, char *s3)
{ char *s;
  s = (char *) malloc(sizeof(char)*(strlen(s1)+strlen(s2)+strlen(s3)+1));
  sprintf(s,"%s%s%s",s1,s2,s3);
  return s;
}

static
char *util_strcat4(char *s1, char *s2, char *s3, char *s4)
{ char *s;
 s = (char *) malloc(sizeof(char) *(strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+1));
  sprintf(s,"%s%s%s%s",s1,s2,s3,s4);
  return s;
}

static
char * util_inttostr(int i)
{
  unsigned int mod, len;
  char *s;
  
  if (i == 0)
    len = 1;
  else {
    if (i < 0) {
      len = 1;
      mod = -i;
    }
    else {
      len = 0;
      mod = i;
    }
    len += floor(log10(mod)) + 1;
  }

  s = (char*) malloc(sizeof(char)*(len + 1));
  sprintf(s, "%d", i);
  
  return s;
}


/**Struct**********************************************************************

  Synopsis    [A struct for keeping the command line options.]

******************************************************************************/
/* struct SugarMcOption { */
/*   FILE *sugarFile;            /\* contains the sugar formulae *\/ */
/*   FILE *debugFile; */
/*   boolean useMore;            /\* use more as pipe for debug output *\/ */
/*   boolean simValue; */
/*   boolean printInputs; */
/*   boolean prunefair;          /\* prune the fairness *\/ */
/*   boolean iocompatible;       /\* minimization by io-compatible *\/ */
/*   boolean directsim;          /\* minimization by direct simulation rln *\/ */
/*   boolean reversesim;         /\* minimization by reverse simulation rln *\/ */
/*   Mc_DcLevel dcLevel; */
/*   Mc_GSHScheduleType schedule; */
/*   Mc_FwdBwdAnalysis direction; */
/*   Mc_VerbosityLevel verbosity; */
/*   LtlMcDbgLevel dbgLevel; */
/*   int lockstep; */
/*   int timeOutPeriod; */
/*   int noIncrementalPartition; /\* ? build the parition of MxA incrementally *\/ */
/*   int noStrengthReduction;    /\* ? use different decision procedures *\/ */
/*                               /\*   for different automaton strength *\/ */
/*   int noShuffleToTop;         /\* ? shuffle automaton vars to the top *\/ */
/* }; */

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
/* typedef struct SugarMcOption SugarMcOption_t; */

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/**Variable********************************************************************

  Synopsis    [Global flag for parsing SUGAR file.]

  Description [This flag is set to 0 before parsing a SUGAR file commences.
  If an error is found while parsing the file, this flag is set to 1.  After
  the parser returns, if this flag is set, then the global SUGAR formula array
  is freed.]

******************************************************************************/
/* extern int SugarGlobalError; */

/**Variable********************************************************************

  Synopsis    [Global pointer to the CTL formulas being created.]

  Description [This pointer is always set to the current formula
  so that the parser can free a partially constructed CTL formula
  when an error is detected. Every time the parser starts reasing
  a new formula, this variable is set to NIL(Ctlp_Fromula_t).]

******************************************************************************/
/* extern Sugar_Formula_t *SugarGlobalFormula; */

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
#define IsLabelTrue(l) (((l)->name == NIL(char)) && ((l)->isNegate == 0))
#define IsLabelFalse(l) (((l)->name == NIL(char)) && ((l)->isNegate == 1))

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

/* EXTERN LtlSet_t* SugarAutomatonLabelToLtlSet(Ltl_Automaton_t* ltl_aut, array_t* label_array); */
EXTERN Sugar_Vertex_t* SugarAutomatonNewNode(Sugar_Automaton_t *aut);
EXTERN void SugarAutomatonFreeNode(Sugar_Vertex_t *node);
EXTERN Sugar_Edge_t* SugarAutomatonNewEdge(Sugar_Automaton_t *aut, Sugar_Vertex_t *source, array_t *dest, array_t *label, int checkUnique);
EXTERN void SugarAutomatonFreeEdge(Sugar_Edge_t *edge);
EXTERN Sugar_Edge_t* SugarAutomatonNewEdgeLoop(Sugar_Automaton_t *aut, Sugar_Vertex_t *node);
EXTERN void SugarAutomatonAddAbortEdge(Sugar_Automaton_t *aut, Sugar_Vertex_t *node, Sugar_Vertex_t *abort_node);
EXTERN void SugarAutomatonFlush(Sugar_Automaton_t *aut);
EXTERN void SugarAutomatonMoveRightToLeft(Sugar_Automaton_t *aut_left, Sugar_Automaton_t *aut_right);
EXTERN array_t* SugarAutomatonMergeLabelArray(array_t *label_src_array, array_t *label_dst_array);
EXTERN array_t* SugarAutomatonMinimizeTermArray(array_t* term_array);
EXTERN Sugar_Automaton_t* SugarAutomatonIdentifier(char *name, int isNegate, int fromFormula);
EXTERN void SugarAutomatonConcat(Sugar_Automaton_t *aut_left, Sugar_Automaton_t *aut_right, int typeOverlap);
EXTERN void SugarAutomatonOr(Sugar_Automaton_t *aut_left, Sugar_Automaton_t *aut_right);
EXTERN void SugarAutomatonAndEq(Sugar_Automaton_t *aut_left, Sugar_Automaton_t *aut_right);
EXTERN void SugarAutomatonStar(Sugar_Automaton_t *aut);
EXTERN void SugarAutomatonPlus(Sugar_Automaton_t *aut);
EXTERN Sugar_Automaton_t* SugarAutomatonGoto(char *bool_prop);
EXTERN Sugar_Automaton_t* SugarAutomatonNegate(Sugar_Automaton_t *aut);
EXTERN array_t* labelDup(array_t* label);
EXTERN lsList destinationDup(lsList oldlist);
EXTERN Sugar_Automaton_t* SugarAutomatonImplyNFA(Sugar_Automaton_t *left, Sugar_Automaton_t* right);
EXTERN Sugar_Automaton_t* SugarAutomatonNegateNFA(Sugar_Automaton_t *aut);
EXTERN  conflict(array_t* fullset, array_t* partset);
EXTERN Sugar_Automaton_t* SugarAutomatonNot(Sugar_Formula_t *formula);
EXTERN Sugar_Automaton_t* SugarAutomatonAlways(Sugar_Automaton_t *aut);
EXTERN Sugar_Automaton_t* SugarAutomatonEventually(Sugar_Automaton_t *aut);
EXTERN Sugar_Automaton_t* SugarAutomatonUntil(Sugar_Automaton_t *aut_left, Sugar_Automaton_t *aut_right, int isStrong);
EXTERN Sugar_Automaton_t* SugarAutomatonImply(Sugar_Automaton_t *aut_left, Sugar_Automaton_t *aut_right, int isRightNoFormula, int isStrong);
/* EXTERN SugarMcOption_t * SugarMcOptionAlloc(void); */
/* EXTERN void SugarMcOptionFree(SugarMcOption_t *result); */
/* EXTERN int CommandSugarModelCheckIpill(Hrc_Manager_t ** hmgr, int argc, char ** argv); */
EXTERN long SugarRand(const long range);
EXTERN int SugarCompareNodeID(const void *obj1, const void *obj2);
/* EXTERN mdd_t * ltlIpillComputeFairStates(Fsm_Fsm_t *fsm, array_t *careSetArray, int verbosity, int dcLevel, Mc_GSHScheduleType schedule, Mc_FwdBwdAnalysis direction, boolean useEarlyTermination); */
/* EXTERN mdd_t * ltlIpillCheckLanguageEmptiness(Fsm_Fsm_t *modelFsm, array_t *modelCareStatesArray, Mc_LeMethod_t leMethod, int dcLevel, int dbgLevel, int printInputs, int verbosity, Mc_GSHScheduleType GSHschedule, Mc_FwdBwdAnalysis GSHdirection, int lockstep, FILE *dbgFile, int UseMore, int SimValue, char *driverName); */

/**AutomaticEnd***************************************************************/

EXTERN int SugarStringToNumber(char* string, int base);
EXTERN void SugarFileSetup(FILE* fp);  
EXTERN int SugarYyparse(void);

EXTERN long SugarRand(const long range);
EXTERN int SugarCompareNodeID(const void *obj1, const void *obj2);

#ifdef DEBUG_SUGAR
EXTERN void printNode(Sugar_Vertex_t* node);
EXTERN void printEdge(Sugar_Edge_t* edge);
EXTERN void printArray(array_t* array, int isLabels);
EXTERN void printNodeArray(array_t* array);
EXTERN void printLabelArray(array_t* array);
EXTERN void printEdgeArrayArray(array_t* array);
EXTERN void printLabelArrayArray(array_t* array);
EXTERN void printNodeArrayArray(array_t* array);
#endif

#endif /* _SUGARINT */
