/**CFile***********************************************************************

  FileName    [sugarUtil.c]

  PackageName [sugar]

  Synopsis    [Routines for manipulating SUGAR formulas.]

  Description [This file contains routines for accessing the fields of the
  SUGAR formula data structure, for printing SUGAR formulas, for reading 
  SUGAR formulas from a file, and for converting formulas to the existential
  form.]

  Author      [Christian Josef Kargl]

  Copyright   [Copyright (c) 2003
               University of Technology Graz, Institute for Softwaretechnology]

  All rights reserved.

  Permission is hereby granted, without written agreement and without license
  or royalty fees, to use, copy, modify, and distribute this software and its
  documentation for any purpose, provided that the above copyright notice and
  the following two paragraphs appear in all copies of this software.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]

  History    [Initial Version: Jun, 30. 2003
              Last Modified:   Nov, 10. 2003]

******************************************************************************/
#include "utils/utils.h"
#include "sugarInt.h"
#include "errno.h"

/* static char rcsid[] UNUSED = "$Id: sugarUtil.c,v 1.1 2005/12/22 23:59:08 timo Exp $"; */

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/**Variable********************************************************************

  Synopsis    [Array of SUGAR formulas (Sugar_Formula_t) read from a file.]

  SeeAlso     [Sugar_FormulaArrayFree]

******************************************************************************/
/* static array_t *globalSugarFormulaArray; */

/* see sugarInt.h for documentation */
/* int SugarGlobalError; */
/* Sugar_Formula_t *SugarGlobalFormula; */

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
/* static boolean SugarMcAtomicFormulaCheckSemantics(Sugar_Formula_t *formula, Ntk_Network_t *network); */

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Parses a file containing a set of CTL formulas.]

  Description [Parses a file containing a set of semicolon-ending CTL
  formulas, and returns an array of Ctlp_Formula_t representing those
  formulas.  If an error is detected while parsing the file, the routine frees
  any allocated memory and returns NULL.]

  SideEffects [Manipulates the global variables globalFormulaArray,
  CtlpGlobalError and CtlpGlobalFormula.]

  SeeAlso     [Ctlp_FormulaArrayFree Ctlp_FormulaPrint]

******************************************************************************/
/* array_t * */
/* Sugar_FileParseFormulaArray( */
/*   FILE * fp) */
/* { */
/*   st_generator *stGen; */
/*   char *name; */
/*   Sugar_Formula_t *formula; */
/*   char *flagValue; */

/*   /\* */
/*    * Initialize the global variables. */
/*    *\/ */
/*   globalSugarFormulaArray = array_alloc(Sugar_Formula_t *, 0); */
/*   SugarGlobalError = 0; */
/*   SugarGlobalFormula = NIL(Sugar_Formula_t); */
/*   //SugarMacroTable = st_init_table(strcmp,st_strhash); */
/*   SugarFileSetup(fp);         */

/*   /\* */
/*    * Parse the formula */
/*    *\/ */
/*   (void)SugarYyparse(); */

/*   /\* */
/*    * Clean up if there was an error, otherwise return the array. */
/*    *\/ */
/*   if (SugarGlobalError){ */
/*     Sugar_FormulaArrayFree(globalSugarFormulaArray); */
/*     return NIL(array_t); */
/*   } */
/*   else { */
/*     return globalSugarFormulaArray; */
/*   } */
/* } */

/**Function********************************************************************

  Synopsis    [Returns formula as a character string.]

  Description [Returns formula as a character string. All subformulas are
  delimited by parenthesis. The syntax used is the same as used by the CTL
  parser. Does nothing if passed a NULL formula.]

  SideEffects []

******************************************************************************/
char *
Sugar_FormulaConvertToString(Sugar_Formula_t * formula)
{
  char *s1        = NIL(char);
  char *s2        = NIL(char);
  char *tmpString = NIL(char);
  char *tmpString2 = NIL(char);
  char *result;

  if (formula == NIL(Sugar_Formula_t)) {
    return NIL(char);
  }
  /* The formula is a leaf. */
  if (formula->type == Sugar_IDENTIFIER_c) {
    return util_strsav((char *)(formula->left));
  }
  /* if the formula is a non-leaf, the function is called recursively */
  s1 = Sugar_FormulaConvertToString(formula->left);
  s2 = Sugar_FormulaConvertToString(formula->right);

  /*
   * TODO: reorder the cases to be in their expected frequency.
   */
  switch(formula->type) {
    case Sugar_FL_Next_c:
      if (formula->numRangeStr != SUGAR_RANGE_UNUSED) {
        tmpString = util_strcat3("[", util_inttostr(formula->numRangeStr), "](");
      }
      else {
        tmpString = util_strsav("(");
      }
      if (formula->isStrongOperator) {
        result = util_strcat4("X!", tmpString, s1, ")");
      }
      else {
        result = util_strcat4("X", tmpString, s1, ")");
      }
      break;
    case Sugar_FL_Always_c:
      result = util_strcat3("G[", s1, "]");
      break;
    case Sugar_FL_Never_c:
      result = util_strcat3("G ![", s1, "])");
      break;
    case Sugar_FL_Eventually_c:
      result = util_strcat3("F[", s1, "]");
      break;
    case Sugar_FL_Releases_c:
      tmpString = util_strcat3("[ ", s1, " V ");
      result = util_strcat3(tmpString, s2, " ]");
      break;
    case Sugar_FL_Until_c:
      if (formula->isStrongOperator) {
          tmpString = util_strcat3("[ ", s1, " U! ");
        }
      else {
          tmpString = util_strcat3("[ ", s1, " U ");
        }
      result = util_strcat3(tmpString, s2, " ]");
      break;
    case Sugar_FL_Until__c:
      if (formula->isStrongOperator) {
          tmpString = util_strcat3("[ ", s1, " U!_ ");
        }
      else {
          tmpString = util_strcat3("[ ", s1, " U_ ");
        }
      result = util_strcat3(tmpString, s2, " ]");
      break;
    case Sugar_FL_Before_c:
      break;
    case Sugar_FL_Before__c:
      break;
    case Sugar_FL_Not_c:
      result = util_strcat3("not(", s1, ")");
      break;
    case Sugar_FL_And_c:
      tmpString = util_strcat3("(", s1, ") and (");
      result = util_strcat3(tmpString, s2, ")");
      break;
    case Sugar_FL_Or_c:
      tmpString = util_strcat3("(", s1, ") or (");
      result = util_strcat3(tmpString, s2, ")");
      break;
    case Sugar_FL_Imply_c:
      tmpString = util_strcat3("(", s1, ") -> (");
      result = util_strcat3(tmpString, s2, ")");
      break;
    case Sugar_FL_Equiv_c:
      tmpString = util_strcat3("(", s1, ") <-> (");
      result = util_strcat3(tmpString, s2, ")");
      break;
    case Sugar_Seq_Always_c:
      result = util_strcat3("always {", s1, "}");
      break;
    case Sugar_Seq_Never_c:
      result = util_strcat3("never {", s1, "}");
      break;
    case Sugar_Seq_Eventually_c:
      result = util_strcat3("eventually! {", s1, "}");
      break;
    case Sugar_Seq_FLProp_c:
      tmpString = util_strcat3("{", s1, "}(");
      result = util_strcat3(tmpString, s2, ")");
      break;
    case Sugar_Seq_Imply_c:
      if (formula->isStrongOperator) {
          tmpString = util_strcat3("} |-> {", s2, "}!");
        }
      else {
          tmpString = util_strcat3("} |-> {", s2, "}");
        }
      result = util_strcat3("{", s1, tmpString);
      break;
    case Sugar_Seq_Imply2_c:
       if (formula->isStrongOperator) {
          tmpString = util_strcat3("} |=> {", s2, "}!");
        }
      else {
          tmpString = util_strcat3("} |=> {", s2, "}");
        }
      result = util_strcat3("{", s1, tmpString);
      break;
    case Sugar_Seq_Within_c:
      break;
    case Sugar_Seq_Within__c:
      break;
    case Sugar_Seq_Whilenot_c:
      break;
    case Sugar_Seq_Whilenot__c:
      break;
    case Sugar_RegExpr_Concat_c:
      result = util_strcat3(s1, ";", s2);
      break;
    case Sugar_RegExpr_Compose_c:
      tmpString = util_strcat3("{", s1, "}:{");
      result = util_strcat3(tmpString, s2, "}");
      break;
    case Sugar_RegExpr_Or_c:
      tmpString = util_strcat3("{", s1, "}|{");
      result = util_strcat3(tmpString, s2, "}");
      break;
    case Sugar_RegExpr_And_c:
      tmpString = util_strcat3("{", s1, "}&{");
      result = util_strcat3(tmpString, s2, "}");
      break;
    case Sugar_RegExpr_AndEq_c:
      tmpString = util_strcat3("{", s1, "}&&{");
      result = util_strcat3(tmpString, s2, "}");
      break;
    case Sugar_RegExpr_Plus_c:
      result = util_strcat3("{", s1, "}[+]");
      break;
    case Sugar_RegExpr_Star_c:
    case Sugar_RegExpr_Equal_c:
    case Sugar_RegExpr_Goto_c:
      tmpString = util_strsav("");
      if (formula->numRangeStr != SUGAR_RANGE_UNUSED) {
        if (formula->numRangeEnd != SUGAR_RANGE_UNUSED) {
          char* val1;
          char* val2;
          if (formula->numRangeStr != SUGAR_RANGE_MIN_VAL) {
            val1 = util_inttostr(formula->numRangeStr);
          }
          else {
            val1 = util_strsav("");
          }
          if (formula->numRangeEnd != SUGAR_RANGE_MAX_VAL) {
            val2 = util_inttostr(formula->numRangeEnd);
          }
          else {
            val2 = util_strsav("");
          }
           tmpString = util_strcat3(val1, "..", val2);
          FREE(val1);
          FREE(val2);
        }
        else {
          tmpString = util_inttostr(formula->numRangeStr);
        }
      }
      switch(formula->type) {
      case Sugar_RegExpr_Star_c:
        tmpString2 = util_strcat3("{", s1, "}[*");
        result = util_strcat3(tmpString2, tmpString, "]");
        break;
      case Sugar_RegExpr_Equal_c:
        tmpString2 = util_strcat3("{", s1, "}[=");
        result = util_strcat3(tmpString2, tmpString, "]");
        break;
      case Sugar_RegExpr_Goto_c:
        tmpString2 = util_strcat3("{", s1, "}[->");
        result = util_strcat3(tmpString2, tmpString, "]");
        break;
      }
      break;
    case Sugar_TRUE_c:
      result = util_strsav("TRUE");
      break;
    case Sugar_FALSE_c:
      result = util_strsav("FALSE");
      break;
    default:
      fail("Unexpected type");
  }
  if (s1 != NIL(char)) {
    FREE(s1);
  }
  if (s2 != NIL(char)) {
    FREE(s2);
  }
  if (tmpString != NIL(char)) {
    FREE(tmpString);
  }
  return result;
}

/**Function********************************************************************

  Synopsis    [Prints a formula to a file.]

  Description [Prints a formula to a file. All subformulas are delimited by
  parenthesis. The syntax used is the same as used by the SUGAR parser. Does
  nothing if passed a NULL formula.]

  SideEffects []

******************************************************************************/
void
Sugar_FormulaPrint(
  FILE * fp,
  Sugar_Formula_t * formula)
{
  char *tmpString;

  if (formula == NIL(Sugar_Formula_t)) {
    return;
  }
  tmpString = Sugar_FormulaConvertToString(formula);
  (void) fprintf(fp, tmpString);
  FREE(tmpString);
}

/**Function********************************************************************

  Synopsis    [Creates a SUGAR formula with the specified fields.]

  Description [Allocates a Sugar_Formula_t, and sets the 2 fields given as
  arguments. If the type is Sugar_IDENTIFIER_c, then the left and right
  fields should contain a pointer to a variable name and a pointer to a
  value respectively. Otherwise, the two fields point to subformulas.
  refCount is set to 1.]

  Comment     []

  SideEffects []

  SeeAlso     [Sugar_FormulaDecrementRefCount]

******************************************************************************/
Sugar_Formula_t *
Sugar_FormulaCreate(
  Sugar_FormulaType type,
  void * left,
  void * right)
{
  Sugar_Formula_t *formula = ALLOC(Sugar_Formula_t, 1);
  
  formula->type             = type;
  formula->left             = (Sugar_Formula_t *)left;
  formula->right            = (Sugar_Formula_t *)right;
  formula->numRangeStr      = SUGAR_RANGE_UNUSED;
  formula->numRangeEnd      = SUGAR_RANGE_UNUSED;
  formula->isStrongOperator = 0;
  formula->refCount         = 1;
  return formula;
}

/**Function********************************************************************

  Synopsis    [Increments the reference count of a formula.]

  Description [The function increments the reference count of a formula. If
  the formula is NULL, the function does nothing.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
Sugar_FormulaIncrementRefCount(
  Sugar_Formula_t *formula)
{
  if (formula != NIL(Sugar_Formula_t)) {
    ++(formula->refCount);
  }
}

/**Function********************************************************************

  Synopsis    [Decrements the reference count of a formula.]

  Description [The function decrements the reference count of formula and if
  the reference count reaches 0, the formula is freed. If the formula is NULL,
  the function does nothing. It is an error to decrement the reference count
  below 0.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
Sugar_FormulaDecrementRefCount(
  Sugar_Formula_t *formula)
{
  if (formula != NIL(Sugar_Formula_t)) {
    assert(formula->refCount > 0);
    if (--(formula->refCount) == 0)
      Sugar_FormulaFree(formula);
  }
}

/**Function********************************************************************

  Synopsis    [Frees a SUGAR formula.]

  Description [The function frees all memory associated with the formula,
  including all character strings. It also decrements the reference counts
  of its two chidren. The function does nothing if formula is NULL.]

  SideEffects []

  SeeAlso     [Sugar_FormulaArrayFree]

******************************************************************************/
void
Sugar_FormulaFree(
  Sugar_Formula_t * formula)
{
  if (formula != NIL(Sugar_Formula_t)) {

    /*
     * Free any fields that are not NULL.
     */

    if (formula->type == Sugar_IDENTIFIER_c) {
      FREE(formula->left);
    }
    else {
      if (formula->left  != NIL(Sugar_Formula_t)) {
        Sugar_FormulaDecrementRefCount(formula->left);
      }
      if (formula->right != NIL(Sugar_Formula_t)) {
        Sugar_FormulaDecrementRefCount(formula->right);
      }
    }

    FREE(formula);
  }
}

/**Function********************************************************************

  Synopsis    [Frees an array of SUGAR formulas.]

  Description [Calls Sugar_FormulaDecrementRefCount on each formula in
  formulaArray, and then frees the array itself. It is okay to call this
  function with an empty array, but it is an error to call it with a NULL
  array.]

  SideEffects []

  SeeAlso     [Sugar_FormulaFree]

******************************************************************************/
void
Sugar_FormulaArrayFree(
  array_t * formulaArray /* of Sugar_Formula_t */)
{
  int i;
  int numFormulas = array_n(formulaArray);
  Sugar_Formula_t *formula;

  assert(formulaArray != NIL(array_t));
  for (i = 0; i < numFormulas; i++) {
    formula = array_fetch(Sugar_Formula_t *, formulaArray, i);
    Sugar_FormulaDecrementRefCount(formula);
  }

  array_free(formulaArray);
}

/**Function********************************************************************

  Synopsis    [Adds formula to the end of globalSugarFormulaArray.]

  SideEffects [Manipulates the global variable globalSugarFormulaArray.]

  SeeAlso     [SugarYyparse]

******************************************************************************/
/* void */
/* Sugar_FormulaAddToGlobalArray( */
/*   Sugar_Formula_t * formula) */
/* { */
/*   array_insert_last(Sugar_Formula_t *, globalSugarFormulaArray, formula); */
/* } */

/**Function********************************************************************

  Synopsis [Perform static semantic check of sugar formula on network.]

  Description [Perform static semantic check of sugar formula on network.
  Specifically, given an identifier, check that the name is a name of a
  latch/wire in the network, and that it is of type boolean.]

  SideEffects []

******************************************************************************/
/* boolean */
/* Sugar_FormulaStaticSemanticCheckOnNetwork( */
/*   Sugar_Formula_t *formula, */
/*   Ntk_Network_t *network) */
/* { */
/*   boolean lCheck; */
/*   boolean rCheck; */

/*   if (formula == NIL(Sugar_Formula_t)) { */
/*     return TRUE; */
/*   } */

/*   if (formula->type == Sugar_IDENTIFIER_c) { */
/*     return SugarMcAtomicFormulaCheckSemantics(formula, network); */
/*   } */

/*   lCheck = Sugar_FormulaStaticSemanticCheckOnNetwork(formula->left, network); */
/*   if (!lCheck) return FALSE; */

/*   rCheck = Sugar_FormulaStaticSemanticCheckOnNetwork(formula->right, network); */
/*   if (!rCheck) return FALSE; */

/*   return TRUE; */
/* } */

/**Function********************************************************************

  Synopsis [Create a random SUGAR formula.]

  SideEffects [The returned string should be freed by the caller.]

  SeeAlso     []

******************************************************************************/
char*
Sugar_FormulaRandom(
  int maxLevel,
  int numProps,
  int makeFormula)
{
  int level;
  int phase;
  int index;
  int wheel;
  int nStr, nEnd;
  char *left, *right;
  char *temp, *range;
  char *result;

  /* generate propositionals p0, p1, ... */
  if (maxLevel < 2) {
    phase = (makeFormula<0)?1:SugarRand(2);
    index = SugarRand(numProps+1);
    if (index == numProps) {
      result = util_strsav((phase)?("TRUE"):("FALSE"));
    }
    else {
      result = util_strcat3((phase)?(""):("!"), "p", util_inttostr(index));
    }
  }
  else {
    /* create LTL formula with suffix implication */
    if (makeFormula) {
      if (maxLevel == 2) {
        temp = Sugar_FormulaRandom(1, numProps, makeFormula);
        result = util_strcat3("X!(", temp, ")");
        FREE(temp);
      }
      else {
        if (SugarRand(3) < 2) {
          wheel = SugarRand(10);
          level = SugarRand(maxLevel-2)+1;
          left = Sugar_FormulaRandom(level, numProps, (wheel<8)?1:0);
          right = Sugar_FormulaRandom(maxLevel-level-1, numProps, (wheel<9)?1:0);
          if (wheel < 2) {
            temp = util_strcat3("[(", left, ")U(");
            result = util_strcat3(temp, right, ")]");
          }
          else if (wheel < 4) {
            temp = util_strcat3("[(", left, ")V(");
            result = util_strcat3(temp, right, ")]");
          }
          else if (wheel < 6) {
            temp = util_strcat3("(", left, ")||(");
            result = util_strcat3(temp, right, ")");
          }
          else if (wheel < 8) {
            temp = util_strcat3("(", left, ")&&(");
            result = util_strcat3(temp, right, ")");
          }
          else if (wheel < 9) {
            temp = util_strcat3("{", left, "}(");
            result = util_strcat3(temp, right, ")");
          }
          else {
            temp = util_strcat4("{", left, "}|",(SugarRand(2))?("-"):("="));
            result = util_strcat4(temp, ">{", right, "}!");
          }
          FREE(left);
          FREE(right);
          FREE(temp);
        }
        else {
          wheel = SugarRand(6);
          if (wheel < 3) {
            temp = Sugar_FormulaRandom(maxLevel-1, numProps, makeFormula);
            result = util_strcat3("X!(", temp, ")");
          }
          else {
            if (wheel < 4) {
              temp = util_strsav("always");
            }
            else if (wheel < 5) {
              temp = util_strsav("eventually!");
            }
            else {
              temp = util_strsav("never");
            }
            result = util_strcat4(temp, "{", Sugar_FormulaRandom(maxLevel-1, numProps, 0), "}");
          }
          FREE(temp);
        }
      }
    }
    else {
      /* create SERE for suffix implication */
      if ((maxLevel == 2) && (SugarRand(5) == 1)) {
        nStr = SugarRand(7)+1; nEnd = SugarRand(8) + nStr;
        if (nStr < nEnd) {
          wheel = SugarRand(4);
          range = util_strcat3((wheel>0)?(util_inttostr(nStr)):(""), "..", (wheel<3)?(util_inttostr(nEnd)):(""));
        }
        else {
          range = util_strsav(util_inttostr(nStr));
        }
        temp = util_strcat3((SugarRand(2)<1)?("="):("->"), range, "");
        result = util_strcat4(Sugar_FormulaRandom(1, numProps, -1), "[", temp, "]");
        FREE(range);
        FREE(temp);
      }
      else {
        wheel = SugarRand(10);
        if (wheel < 8) {
          level = SugarRand(maxLevel);
          left = Sugar_FormulaRandom(level, numProps, 0);
          right = Sugar_FormulaRandom(maxLevel-level, numProps, 0);
          if (wheel < 4) {
            result = util_strcat3(left, ";", right);
          }
          else {
            if (wheel < 5) {
              temp = util_strcat3("{", left, "}:{");
              result = util_strcat3(temp, right, "}");
            }
            else if (wheel < 6) {
              temp = util_strcat3("{", left, "}&&{");
              result = util_strcat3(temp, right, "}");
            }
            else if (wheel < 7) {
              temp = util_strcat3("{", left, "}|{");
              result = util_strcat3(temp, right, "}");
            }
            else {
              temp = util_strcat3("{", left, "}&{");
              result = util_strcat3(temp, right, "}");
            }
            FREE(temp);
          }
          FREE(left);
          FREE(right);
        }
        else {
          if (SugarRand(3)) {
            temp = util_strsav((SugarRand(2))?("[*]"):("[+]"));
          }
          else {
            nStr = SugarRand(8); nEnd = SugarRand(8) + nStr;
            if (nStr < nEnd) {
              wheel = SugarRand(4);
              range = util_strcat3((wheel>0)?(util_inttostr(nStr)):(""), "..", (wheel<3)?(util_inttostr(nEnd)):(""));
              temp = util_strcat3("[*", range, "]"); FREE(range);
            }
            else {
              temp = util_strcat3("[*", util_inttostr(nStr), "]");
            }
          }
          result = util_strcat4("{", Sugar_FormulaRandom(maxLevel-1, numProps, 0), "}", temp);
          FREE(temp);
        }
      }
    }
  }
  return result;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Create random number within given range.]

  Description [Create a random number between 0 and range-1. Uses the
  system random number generator. The random number generator for function
  util_random() cannot be initialied proberly to get different number for
  different executions of the VIS main program.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
long
SugarRand(
  const long range)
{
  return (long)(((double)rand()/(double)(RAND_MAX)) * range);
}

/**Function********************************************************************

  Synopsis    [Compare Sugar automaton nodes by ID value.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
int
SugarCompareNodeID(
  const void *obj1,
  const void *obj2)
{
  return ((*((Sugar_Vertex_t**)obj1))->id - (*((Sugar_Vertex_t**)obj2))->id);
}

/**Function********************************************************************

  Synopsis [Perform simple static semantic checks on formula.]

  SideEffects []

  SeeAlso [Sugar_FormulaStaticSemanticCheckOnNetwork]

******************************************************************************/
/* static boolean */
/* SugarMcAtomicFormulaCheckSemantics(  */
/*   Sugar_Formula_t *formula,  */
/*   Ntk_Network_t *network) */
/* { */

/*   // ipill : for multi-value identifiers the semantic check has to be adapted */

/*   Ntk_Node_t *node; */
/*   Var_Variable_t *nodeVar; */
/*   char *NameString = (char*)formula->left; */
/*   char *eqpos=NULL, *nodeNameString=strdup(NameString), *valuename=NULL; */
/*   char eqcount=1; */
/*   int namelength=strlen(nodeNameString); */
/*   int valindex=0; */
  
/*   eqpos=strchr(NameString, '='); */
/*   if (eqpos) { */

/*     if (eqpos[1]=='=') eqcount=2; */
/*     namelength-=strlen(eqpos); */
/*     nodeNameString[namelength]=0;  */
/*     valuename=eqpos+(eqcount*sizeof(char)); */
/*   } */
   
/*   node = Ntk_NetworkFindNodeByName(network, nodeNameString); */
/*   if (!node) { */
/*     error_append("Could not find node corresponding to the name\n\t- "); */
/*     error_append(nodeNameString); */
/*     free(nodeNameString); */
/*     return FALSE; */
/*   } */
/*   nodeVar = Ntk_NodeReadVariable(node); */
  
/*   if (eqpos) { */
 
/*     valindex=Var_VariableReadIndexFromSymbolicValue(nodeVar, valuename); */
/*     if (valindex==-1) { */
    
/*       error_append("Domain of variable '"); */
/*       error_append(Var_VariableReadName(nodeVar)); */
/*       error_append("' does not include value "); */
/*       error_append(strdup(valuename)); */
/*       free(nodeNameString); */
/*       return FALSE; */
/*     } */
 
        
    
/*   } else { */
    
/*     if (Var_VariableReadNumValues(nodeVar) != 2) { */
/*       error_append("Domain of variable '"); */
/*       error_append(Var_VariableReadName(nodeVar)); */
/*       error_append("' is not boolean! (Values 0 or 1)\n"); */
/*       free(nodeNameString); */
/*       return FALSE; */
/*     } */
/*   } */

/*   free(nodeNameString); */
/*   return TRUE; */
/* } */
