/**CFile*****************************************************************

  FileName    [SymbType.c]

  PackageName [compile.symb_table]

  Synopsis    [Implementation of functions dealing with the type of 
  variables in a symbol table.]

  Description []

  SeeAlso     [SymbType.h]

  Author      [Andrei Tchaltsev]

  Copyright   [
  This file is part of the ``compile.symb_table'' package of NuSMV
  version 2.  Copyright (C) 2005 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "compileInt.h"
#include "parser/symbols.h"
#include "enc/utils/WordNumber.h"

static char rcsid[] UTIL_UNUSED = "$Id: SymbType.c,v 1.1.1.1 2005/12/20 17:06:51 timo Exp $";


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef struct SymbType_TAG {
  SymbTypeTag tag;
  node_ptr body;
  boolean isMemorySharingInstance;
} SymbType;


/*---------------------------------------------------------------------------*/
/* macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Constants declarations                                                    */
/*---------------------------------------------------------------------------*/
/* a set of constants to identify different enum-types */
enum Enum_types{ ENUM_TYPE_BOOLEAN, 
		 ENUM_TYPE_PURE_INT,
		 ENUM_TYPE_PURE_SYMBOLIC,
		 ENUM_TYPE_INT_SYMBOLIC,
                };

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Class SymbType constructor]

  Description [The tag must be a correct tag. The 'body' is the
  additional info corresponding to a particular kind of the type: for
  a enum type the body is the list of values; for "INT" or "REAL" the
  body is unused, and set to Nil; for "WORD" it is the NUMBER node
  defining the width of the type; for "WORDARRAY", the body is a pair
  of NUMBER nodes, defining the width of the address, and the width of
  the value. Set-types are used with expressions which represent a set
  values.  "NO-TYPE" is used with expressions which normally do not
  have any type such as assignments. "ERROR" type indicates an error
  (not an actual type).
  No-type, error-type and all set-types (boolean-set, integer-set,
  symbolic-set, symbolic-integer-set) should not be created with this
  constructor, but only with memory-shared function
  SymbTablePkg_..._type.  The reason behind this constrain is that
  only expressions (not variables) can have these types, therefore
  only memory-shared versions of these types are required.

  The constructor does not create a copy of the body, but just remember
  the pointer

  NB: system "reset" command destroys all node_ptr objects, including those
  used in SymbType_ptr. So destroy all symbolic types before the destruction
  of node_ptr objects, i.e. before or during "reset"]

  SideEffects        [allocate memory]

  SeeAlso            [SymbType_destroy, SymbType_end]

******************************************************************************/
SymbType_ptr SymbType_create(SymbTypeTag tag, node_ptr body)
{
  SymbType_ptr self = ALLOC(SymbType, 1);
  SYMB_TYPE_CHECK_INSTANCE(self);

  nusmv_assert(SYMB_TYPE_NONE <= tag && tag <= SYMB_TYPE_ERROR);
  self->tag = tag;
  self->isMemorySharingInstance = false;

  if (SYMB_TYPE_ENUM == tag) {
    /* these is an enum type. Find out which kind of enum this is.
       It can be a boolean enum, a pure integer enum, a pure symbolic enum 
       or integer-symbolic enum.
    */
    enum Enum_types enumType = -1; /* for sure incorrect value */

    if (body == boolean_range) { 
      /* boolean enumeration should be very specific, i.e. it should be
	 created with the variable 'boolean_range'. So in
	 compileFlatten.c in function instantiate_var all
	 possible boolean types (such as range 0..1 or
	 enum {1,0}) are converted to boolean type.
         N.B. In a enum, TRUE and FALSE constants are
	 considered as 1 and 0 (see CompileFlatten_normalise_value_list).
      */
      enumType = ENUM_TYPE_BOOLEAN;
    }
    else {
      node_ptr iter = body;
      int thereIsInt = 0;
      int thereIsSymbolic = 0;
      while (iter != Nil) {
	switch (node_get_type(car(iter))) {
	case ATOM: /* symbolic constant */
	case BIT:
	case DOT:
	case ARRAY:
	  thereIsSymbolic = 1;
	  break;
	case NUMBER: /* integer numbers */
	  thereIsInt = 1; 
	  break;

	case TRUEEXP:
	case FALSEEXP:
	  /* TRUE and FALSE cannot be in a enum, because 
	     funtion CompileFlatten_normalise_value_list normilised them out.
	  */
	  nusmv_assert(false);

	default: /* only numbers, true, false, and 
                    symbolic constants can be in the constant-list (enum-type).
		 */
	  print_node(nusmv_stderr, body); fprintf(nusmv_stderr,"\n\n");
	  nusmv_assert(false);
	}
	iter = cdr(iter);
      }/* while */
      /* there must be constant or symbolic constant in the enum */
      nusmv_assert(thereIsInt | thereIsSymbolic);
      
      enumType = (thereIsInt && thereIsSymbolic)
	        ?  ENUM_TYPE_INT_SYMBOLIC 
	        : thereIsInt
	          ? ENUM_TYPE_PURE_INT
                  : ENUM_TYPE_PURE_SYMBOLIC;
    } /* if-else (body == boolean_range) */
    /* we add additional node to remember the kind of the enum-type */
    self->body = new_node(enumType, body, Nil);
  } 
  else {
    /* except ENUM type only Word and WordArraycan can have body */
    nusmv_assert((Nil != body) == ((SYMB_TYPE_WORD == tag) ||
				   (SYMB_TYPE_WORDARRAY == tag)));
    /* usual type => just remember the body */
    self->body = body;
  }
  return self;
}


/**Function********************************************************************

  Synopsis           [Class SymbType destructor]

  Description        [Deallocate the memory. The destructor 
  does not deallocate memory from the type's body (since the
  constructor did not created the body).

  NOTE: If self is a memory sharing type instance, i.e. a type returned by 
  SymbTablePkg_..._type functions then the destructor will not delete 
  the type.]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
void SymbType_destroy(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  if (!self->isMemorySharingInstance) { /* this is a usual type instance */
    if (SYMB_TYPE_ENUM == self->tag) {
      /* for the enum-type we added one addional node => remove it now */
      node_node_setcar(self->body, Nil);
      free_node(self->body);
    }
    FREE(self);
  }
}


/**Function********************************************************************

  Synopsis           [Private class SymbType constructor
  for memory sharing type instances]

  Description        [The difference from the public constructor is that this
  constructor marks the created type as a memory sharing type. As
  result the public constructor will not be able to destroy memory
  sharing instance of a type. Use the private constructor
  SymbType_destroy_memory_sharing_type to destroy such instances.
  ]

  SideEffects        [allocate memory]

  SeeAlso            [SymbType_create, SymbType_destroy_memory_sharing_type]

******************************************************************************/
SymbType_ptr SymbType_create_memory_sharing_type(SymbTypeTag tag, node_ptr body)
{
  SymbType_ptr self = SymbType_create(tag, body);
  self->isMemorySharingInstance = true;
  return self;
}


/**Function********************************************************************

  Synopsis           [Private Class SymbType destructor
  for memory sharing instances of types.]

  Description        [The same as the public destructor SymbType_destroy
  but 'self' has to be created by private constructor 
  SymbType_create_memory_sharing_type only.]
  
  SideEffects        []

  SeeAlso            [SymbType_create_memory_sharing_type, SymbType_create]

******************************************************************************/
void SymbType_destroy_memory_sharing_type(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  /* this is a memory sharing instance */
  nusmv_assert(self->isMemorySharingInstance);

  self->isMemorySharingInstance = false;
  SymbType_destroy(self);
}


/**Function********************************************************************

  Synopsis           [Returns the tag (the kind) of the type]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
SymbTypeTag SymbType_get_tag(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return self->tag;
}


/**Function********************************************************************

  Synopsis [Returns true, if the type is a enum-type and its values are from 
  boolean range. Otherwise - returns false.]

  Description [The kind of enum-type is analysed in the constructor.]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_boolean_enum(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return SYMB_TYPE_ENUM == self->tag && 
    ENUM_TYPE_BOOLEAN == node_get_type(self->body);
}


/**Function********************************************************************

  Synopsis [Returns true, if the type is a enum-type and its value 
  are integers only. Otherwise - returns false.]

  Description [The kind of enum-type is analysed in the constructor.]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_pure_int_enum(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return SYMB_TYPE_ENUM == self->tag && 
    ENUM_TYPE_PURE_INT == node_get_type(self->body);
}


/**Function********************************************************************

  Synopsis [Returns true, if the type is a enum-type and its value 
  are symbolic constants only. Otherwise - returns false.]

  Description [The kind of enum-type is analysed in the constructor.]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_pure_symbolic_enum(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return SYMB_TYPE_ENUM == self->tag && 
    ENUM_TYPE_PURE_SYMBOLIC == node_get_type(self->body);
}


/**Function********************************************************************

  Synopsis [Returns true, if the type is a enum-type and its value 
  are symbolic AND integer constants. Otherwise - returns false.]

  Description [The kind of enum-type is analysed in the constructor.]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_int_symbolic_enum(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return SYMB_TYPE_ENUM == self->tag && 
    ENUM_TYPE_INT_SYMBOLIC == node_get_type(self->body);
}


/**Function********************************************************************

  Synopsis [Returns true, if the type is a Word type and the width of 
  the word is 1. Otherwise - returns false.]

  Description []

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_word_1(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  /* expects the body of Word type to be NUMBER */
  return SYMB_TYPE_WORD == self->tag && 1 == node_get_int(self->body);
}


/**Function********************************************************************

  Synopsis    [Returns true, if the type is one of the set-types, i.e.
  boolean-set, integer-set, symbolic-set, integer-symbolic-set, and
  false otherwise.]

  Description []

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_set(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return SYMB_TYPE_SET_BOOL == self->tag ||
    SYMB_TYPE_SET_INT == self->tag ||
    SYMB_TYPE_SET_SYMB == self->tag ||
    SYMB_TYPE_SET_INT_SYMB == self->tag;
}


/**Function********************************************************************

  Synopsis    [Returns true, if the type is a error-type, and false otherwise.]

  Description [Error type is used to indicate an error]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
boolean SymbType_is_error(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  return SYMB_TYPE_ERROR == self->tag;
}


/**Function********************************************************************

  Synopsis           [Genereate and returns a list of all possible values 
  of a particular Word type]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr SymbType_generate_all_word_values(SymbType_ptr self)
{
  /* create all the constants for the Word type */
  int width;
  unsigned long long value;
  node_ptr list;
  
  /* All assertions are checked in the function SymbType_get_word_width */

  width = SymbType_get_word_width(self);
  nusmv_assert(width <= WordNumber_max_width());
  
  
  list = Nil; 
  /* create the max value. Two left shifts are used because in C 
     shifting by a full width is not allowed
  */
  value = ~((~0ull) << (width - 1) << 1);
  ++value;
  do {
    --value;
    list = cons(find_node(NUMBER_WORD, 
			  (node_ptr)WordNumber_from_integer(value, width), 
			  Nil),
		 list);
  }
  while ( value > 0); 
  return list;
}


/**Function********************************************************************

  Synopsis           [Returns the width of a Word type]

  Description        [The given type should be Word and the 
  body of the type (given to the constructor) should be NUMBER node.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int SymbType_get_word_width(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  nusmv_assert(SYMB_TYPE_WORD == self->tag &&
	       node_get_type(self->body) == NUMBER);
  return node_get_int(self->body); /* the width */
}


/**Function********************************************************************

  Synopsis           [Returns the width of the address in a WordArray type]

  Description        [.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int SymbType_get_wordarray_awidth(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);

  nusmv_assert(SYMB_TYPE_WORDARRAY == self->tag &&
	       node_get_type(self->body) == CONS && 
	       node_get_type(car(self->body)) == NUMBER &&
	       node_get_type(cdr(self->body)) == NUMBER);
  
  return node_get_int(car(self->body)); /* the address width */
}


/**Function********************************************************************

  Synopsis           [Returns the width of the value in a WordArray type]

  Description        [.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int SymbType_get_wordarray_vwidth(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);

  nusmv_assert(SYMB_TYPE_WORDARRAY == self->tag &&
	       node_get_type(self->body) == CONS &&
	       node_get_type(car(self->body)) == NUMBER &&
	       node_get_type(car(self->body)) == NUMBER);
  
  return node_get_int(cdr(self->body)); /* the value width */
}



/**Function********************************************************************

  Synopsis    [Returns the line number where the type was declared.]

  Description [The body of the type, provided during construction, is
  a node NUMBER specifying the width of the Word or a node CONS
  specifying the address-value widths or WordArray.  This node was
  create during parsing and contains the line number of the type
  declaration.
  NB: The type should not be memory-sharing.
  NB: Virtually this function is used only in TypeChecker_is_type_wellformed
  ]

  SideEffects        []

  SeeAlso            [SymbType_create]

******************************************************************************/
int SymbType_get_word_line_number(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  nusmv_assert(( SYMB_TYPE_WORD == self->tag &&
		 node_get_type(self->body) == NUMBER
               ) ||
	       ( SYMB_TYPE_WORDARRAY == self->tag &&
		 node_get_type(self->body) == CONS));
  nusmv_assert(!self->isMemorySharingInstance); /* type is not memory-shared */
  return node_get_lineno(self->body);
}


/**Function********************************************************************

  Synopsis           [Returns the list of values of an enum type]

  Description        [The given type has to be a ENUM type.
  The return list is a list of all possible values of a enum type. This list
  was provided during construction.
  
  NB: Memory sharing types do not have particular values, since they
  are "simplified".]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr SymbType_get_enum_type_values(SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  nusmv_assert(SYMB_TYPE_ENUM == self->tag);
  nusmv_assert(!self->isMemorySharingInstance); /* type is not memory-shared */
  /* for the enum-type we added one addional node => skip it now */
  return car(self->body);
}


/**Function********************************************************************

  Synopsis            [Prints the type structure to the output stream.]

  Description         [This function is made very similar to print_node.
  If a Enum type was created with SymbType_create then all its values will be 
  printed, otherwise the type was created with SymbTablePkg_..._type
  and simplified type name (instead of actual type values) is printed.
  ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void SymbType_print (FILE* output_stream, SymbType_ptr self)
{
  SYMB_TYPE_CHECK_INSTANCE(self);
  switch (self->tag) {

  case SYMB_TYPE_NONE: 
    fprintf(output_stream, "no-type");
    break;

  case SYMB_TYPE_ENUM: 
    if (!self->isMemorySharingInstance) {
      /* this is a usual type instance => print all the values */
      if (ENUM_TYPE_BOOLEAN == node_get_type(self->body)) {
	/* we distinguish boolean type from all other Enum types */
	fprintf(output_stream, "boolean");
      }
      else { /* print all the values */
	node_ptr l = SymbType_get_enum_type_values(self);
	fprintf(output_stream, "{");
	while (l != Nil) {
	  print_node(output_stream, car(l));
	  l = cdr(l);
	  if (l != Nil) { fprintf(output_stream, ", "); }
	}
	fprintf(output_stream, "}");
      }
    } 
    else { /* this is memory sharing types (therefore, are "simplified") */
      switch(node_get_type(self->body)) {
      case ENUM_TYPE_BOOLEAN:
	fprintf(output_stream, "boolean");
	break;
      case ENUM_TYPE_PURE_INT:
	fprintf(output_stream, "integer-enum");
	break;
      case ENUM_TYPE_PURE_SYMBOLIC:
	fprintf(output_stream, "symbolic-enum");
	break;
      case ENUM_TYPE_INT_SYMBOLIC:
	fprintf(output_stream, "integer-and-symbolic-enum");
	break;
      default: nusmv_assert(false);
      }
    } /* else */
    break;

  case SYMB_TYPE_INTEGER:   
    fprintf(output_stream, "integer");
    break;;

  case SYMB_TYPE_REAL:  
    fprintf(output_stream, "real");
    break;

  case SYMB_TYPE_WORD:  
    fprintf(output_stream, "word[");
    fprintf(output_stream, "%d", SymbType_get_word_width(self));
    fprintf(output_stream, "]");
    break;

  case SYMB_TYPE_WORDARRAY:  
    fprintf(output_stream, "array ");

    fprintf(output_stream, "word[");
    fprintf(output_stream, "%d", SymbType_get_wordarray_awidth(self));
    fprintf(output_stream, "]");

    fprintf(output_stream, " of ");

    fprintf(output_stream, "word[");
    fprintf(output_stream, "%d", SymbType_get_wordarray_vwidth(self));
    fprintf(output_stream, "]");
    break;

  case SYMB_TYPE_SET_BOOL:  fprintf(output_stream, "boolean-set"); break;
  case SYMB_TYPE_SET_INT:  fprintf(output_stream, "integer-set"); break;
  case SYMB_TYPE_SET_SYMB: fprintf(output_stream, "symbolic-set"); break;
  case SYMB_TYPE_SET_INT_SYMB: fprintf(output_stream, "integer-symbolic-set");
    break;

  case SYMB_TYPE_ERROR:
    fprintf(output_stream, "error-type");
    break;

  default: nusmv_assert(false);
  }
  return;
}



