/**CFile***********************************************************************

  FileName    [TypeChecker.c]

  PackageName [compile.type_checking]

  Synopsis    [Implementation of class 'TypeChecker']

  Description []

  SeeAlso     [TypeChecker.h]

  Author      [Andrei Tchaltsev]

  Copyright   [
  This file is part of the ``compile.type_checking'' package of NuSMV
  version 2.
  Copyright (C) 2005 by ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "TypeChecker.h" 
#include "type_checkingInt.h" 

#include "compile/symb_table/symb_table.h"

#include "utils/utils.h" 
#include "utils/ustring.h" 
#include "utils/assoc.h" 
#include "parser/symbols.h" 
#include "enc/utils/WordNumber.h" 

static char rcsid[] UTIL_UNUSED = "$Id: TypeChecker.c,v 1.1.1.1 2005/12/20 17:06:52 timo Exp $";

#warning "[AC] arguably bad use of nusmv_assert -- shuold we return with error?"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/**Struct**********************************************************************

  Synopsis    [TypeChecker class definition]

  Description []

  SeeAlso     []   
  
******************************************************************************/
typedef struct TypeChecker_TAG
{
  /* -------------------------------------------------- */
  /*                  Private members                   */
  /* -------------------------------------------------- */
  Type_checking_violation_handler violationHandler; /* violation handler */
  SymbTable_ptr symbolTable; /* symbol table to look for symbols */
  hash_ptr expressionTypeMapping; /* type of checked expression's */
} TypeChecker;



/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
const SymbType_ptr nullType = SYMB_TYPE(NULL);

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void type_checker_init  ARGS((TypeChecker_ptr self,
				     Type_checking_violation_handler handler,
				     SymbTable_ptr symbolTable));
static void type_checker_deinit ARGS((TypeChecker_ptr self));


static SymbType_ptr tc_set_expression_type ARGS((TypeChecker_ptr self,
						 node_ptr expression,
						 SymbType_ptr type));
static SymbType_ptr tc_get_expression_type ARGS((TypeChecker_ptr self,
						 node_ptr expression));

static int tc_is_back_comp_type ARGS((SymbType_ptr type));

/* implicit conversion functions: to be applied on one or both operands */
static SymbType_ptr tc_right_operand_implicit_convertion 
                      ARGS((SymbType_ptr leftType, SymbType_ptr rightType));
static SymbType_ptr tc_one_operand_implicit_convertion
                      ARGS((SymbType_ptr type1, SymbType_ptr type2));
static SymbType_ptr tc_both_operands_implicit_convertion
                      ARGS((SymbType_ptr type1, SymbType_ptr type2));
static SymbType_ptr tc_implicit_convertion_to_set_type
                      ARGS((SymbType_ptr type));
static SymbType_ptr tc_convertion_from_set_type ARGS((SymbType_ptr type));

static SymbType_ptr type_checker_check_expression ARGS((TypeChecker_ptr self,
							node_ptr expression,
							node_ptr context));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [TypeChecker class constructor]

  Description        [TypeChecker class constructor. Takes 
  a violation handler, which is invoked when an expression being
  checked violated the type system. See Type_checking_violation_handler for
  more details.
  The 'symbolTable' is a symbol table to look for the type of found
  identifiers]

  SideEffects        []

  SeeAlso            [TypeChecker_destroy, Type_checking_violation_handler]   
  
******************************************************************************/
TypeChecker_ptr TypeChecker_create(Type_checking_violation_handler handler,
				   SymbTable_ptr symbolTable)
{
  TypeChecker_ptr self = ALLOC(TypeChecker, 1);
  TYPE_CHECKER_CHECK_INSTANCE(self);

  type_checker_init(self, handler, symbolTable);
  return self;
}


/**Function********************************************************************

  Synopsis           [The TypeChecker class destructor]

  Description        [The TypeChecker class destructor]

  SideEffects        []

  SeeAlso            [TypeChecker_create]   
  
******************************************************************************/
void TypeChecker_destroy(TypeChecker_ptr self)
{
  TYPE_CHECKER_CHECK_INSTANCE(self);

  type_checker_deinit(self);
  FREE(self);
}


/**Function********************************************************************

  Synopsis           [Returns the symbol table this type checker is
  associated to.]

  Description        [During its lifetime every type checker can deal only 
  with one symbol table instance (because type checker caches the
  checked expressions and their types). The symbol table is given to
  the type checker during construction, and this function returns
  this symbol table.]

  SideEffects        []

  SeeAlso            [TypeChecker_create]

******************************************************************************/
SymbTable_ptr TypeChecker_get_symbol_table(TypeChecker_ptr self)
{
  return self->symbolTable;
}

/**Function********************************************************************

  Synopsis           [The function type checks an expression and returns true
  if an expression is wellformed with respect to the type system and
  false otherwise.]

  Description        [
  The main purpose of this function is to be invoked on temprorarily
  created expressions before they are evaluated to ADD or BDD.
  This function may not be useful for expressions read from files (such as
  bodies of INVAR, SPEC, etc) since they go through flattening in the
  compilation package and type-checked there.

  NOTE: an expression may be unmodified (after parsing and compilation),
  flattened or flattened+expanded. 
  The expressions and their types are remembered by the type checker and 
  the expressions types can be obtained with TypeChecker_get_expression_type.

  See Type_checking_violation_handler for info of dealing with
  detected type violations.]

  SideEffects        []

  SeeAlso            [Type_checking_violation_handler,
  type_checker_check_expression]

******************************************************************************/
boolean TypeChecker_is_expression_wellformed(TypeChecker_ptr self,
					     node_ptr expression,
					     node_ptr context)
{
  SymbType_ptr type;
  /* Nil expressions in NuSMV often correspond to TRUE value => no violations */
  if (Nil == expression) return true;

  type = type_checker_check_expression(self, expression, context);
  if (SymbType_is_error(type)) return false;  /* an error */
  return true;
}


/**Function********************************************************************

  Synopsis           [Perform type checking of a specification]

  Description        [
  A specification is a usual (i.e. able to have a type)
  expression wrapped into a node with a specification tag such as INIT,
  INVAR, SPEC, COMPASSION, etc.
  There are two special case:
    ASSIGN can contains a list of EQDEF statements, and 
    COMPUTE can contains MIN and MAX statement.
    
  The returned value is true if no violations of the type system are detected,
  and false otherwise.

  NOTE: the expression may be unmodified (after compilation),
  flattened or flattened+expanded. 
  The expressions and their types are remembered by the type checker and 
  the expressions types can be obtained with TypeChecker_get_expression_type.

  See Type_checking_violation_handler for info of dealing with
  detected type violations.]

  SideEffects        []

  SeeAlso            [Type_checking_violation_handler,
  type_checker_check_expression]

******************************************************************************/
boolean TypeChecker_is_specification_wellformed(TypeChecker_ptr self,
						node_ptr expression)
{
  TYPE_CHECKER_CHECK_INSTANCE(self);

  /* the _whole_ expression has been already checked */
  if (nullType != tc_get_expression_type(self, expression))return true;
  /* mark this expression as checked (giving it arbitrary type) */
  tc_set_expression_type(self, expression, SymbTablePkg_boolean_enum_type());
  
  switch (node_get_type(expression)) {
    /* the expression must be bool */
  case TRANS:
  case INIT:
  case INVAR:
  case FAIRNESS:
  case JUSTICE:
  case COMPASSION:
  case SPEC:
  case LTLSPEC:
  case PSLSPEC:
  case INVARSPEC: {
    /* get the operand's type */
    SymbType_ptr type = type_checker_check_expression(self, car(expression), Nil);
    
    if (SymbType_is_error(type)) return false;  /* earlier error */
    
    /* the operand must be boolean or has no-type.
       no-type can appear in expression during predicate-normalisation
       in symbolic FSM (when assignments are pushed into expressions).
    */
    if (SymbType_is_boolean_enum(type) 
	|| SymbType_get_tag(type) == SYMB_TYPE_NONE) return true;

    /* there is violation */
    return !self->violationHandler(tc_is_back_comp_type(type)
				     ? TC_VIOLATION_TYPE_BACK_COMP 
				   : TC_VIOLATION_TYPE_MANDATORY,
				   expression, self);
  }

  case DEFINE: { /* DEFINE is artificial contract to check DEFINES body */
    /* get the operand's type */
    SymbType_ptr type = type_checker_check_expression(self, car(expression), Nil);
    
    if (SymbType_is_error(type)) return false;  /* earlier error */
    
    /* DEFINEs can have any type => do not check constrains on the type */
    return true;
  }

  case ASSIGN: { /* type check the assignment: EQDEF */
    SymbType_ptr type;
    nusmv_assert(EQDEF == node_get_type(car(expression)));

    type = type_checker_check_expression(self, car(expression), Nil);

    if (SymbType_is_error(type)) return false; /* mark the error situation */
    return true;
  }

  case COMPUTE: {
    SymbType_ptr type1;
    SymbType_ptr type2;
    node_ptr context;

    /*the expression is also wrapped in context */
    nusmv_assert(CONTEXT == node_get_type(car(expression)));
     /* get rid of COMPUTE and the context */
    context = car(car(expression));
    expression = cdr(car(expression));
    /* now the expression should be MAX or MIN */
    nusmv_assert(MINU == node_get_type(expression) || 
		 MAXU == node_get_type(expression));
    
    /* get the operands' type */
    type1 = type_checker_check_expression(self, car(expression), context);
    type2 = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return false; 
    }
  
    /* the expressions must be boolean */
    if (SymbType_is_boolean_enum(type1) && SymbType_is_boolean_enum(type2)) {
      return true;
    }
    
    return !self->violationHandler(tc_is_back_comp_type(type1) &&
				   tc_is_back_comp_type(type2) 
				     ? TC_VIOLATION_TYPE_BACK_COMP
				     : TC_VIOLATION_TYPE_MANDATORY,
				   expression, self);
  } /* COMPUTE */
    
  default: nusmv_assert(false);
    return false;
  } /* switch */
}


/**Function********************************************************************

  Synopsis           [Checks that a type is well formed.]

  Description        [
  This function is used to check the well-formedness of a type
  from a symbol table. This type should have properly created body,
  in particular, bodies should have correct line info.
  
  The constrains on a type are:
     1. word type: the width should be a NUMBER and have positive value.
        a) the width should be greater than implemenetation limit
           WordNumber_max_width() bit (since we do not use
           arbitrary-precision arithmetic and ADD structores will be
           too big otherwise)
     2. enum type: there should be no duplicate values.

  The third parameter is the variable name that type is to be checked.
  The variable name is used just to output proper message in the case 
  of a type violation.

  In the case of a type violation the violation handler obtain an
  expression CONS with variable name as left child and the body of the type
  as the right child.
  ]

  SideEffects        []

  SeeAlso            [Type_checking_violation_handler,
  type_checker_check_expression]

******************************************************************************/
boolean TypeChecker_is_type_wellformed(TypeChecker_ptr self,
				       SymbType_ptr type,
				       node_ptr varName)
{
  TYPE_CHECKER_CHECK_INSTANCE(self);
  
  switch (SymbType_get_tag(type)) {

  case SYMB_TYPE_ENUM: {
    /* check that no value is repeated */
    /* Here, we assume that 'constant_list' in parser/grammar.y makes 
       the values in the list unique (with the help of find_atom or find_node)
       but the list itself is not unique and contains correct line info.
    */
    node_ptr values = SymbType_get_enum_type_values(type);
    node_ptr iter1 = values;
    while (Nil != iter1) {
      node_ptr iter2 = cdr(iter1);

      nusmv_assert(node_get_type(iter1) == CONS); /* only lists can be here */

      while (Nil != iter2) {
	if (car(iter1) == car(iter2) && /* duplicated constant */
	    /* and this is an error */
	    self->violationHandler(TC_VIOLATION_DUPLICATE_CONSTANTS,
				   new_lined_node(CONS, varName, values,
						  node_get_lineno(values)),
				   self)) {
	  return false;
	}
	iter2 = cdr(iter2);
      }
      iter1 = cdr(iter1);
    } /* while */
    break;
  }

  case SYMB_TYPE_WORD: {
    int width = SymbType_get_word_width(type);
    if (0 < width && width <= WordNumber_max_width()) return true;
    
    if (self->violationHandler(TC_VIOLATION_INCORRECT_WORD_WIDTH,
			       new_lined_node(CONS, varName,
					    new_node(NUMBER, (node_ptr)width,Nil),
					    SymbType_get_word_line_number(type)),
			       self)) {
      return false;
    }
    break;
  }
    
  case SYMB_TYPE_WORDARRAY: {
    int awidth;
    int vwidth;

    awidth = SymbType_get_wordarray_awidth(type);
    vwidth = SymbType_get_wordarray_vwidth(type);

    if (!(0 < awidth && awidth <= WordNumber_max_width())) {
      if (self->violationHandler(TC_VIOLATION_INCORRECT_WORDARRAY_WIDTH,
				 new_lined_node(CONS, varName,
					new_node(NUMBER, (node_ptr)awidth,Nil),
					SymbType_get_word_line_number(type)),
				 self)) {
	return false;
      }
    }

    if (!(0 < vwidth && vwidth <= WordNumber_max_width())) {
      if (self->violationHandler(TC_VIOLATION_INCORRECT_WORDARRAY_WIDTH,
				 new_lined_node(CONS, varName,
					  new_node(NUMBER, (node_ptr)vwidth,Nil),
					  SymbType_get_word_line_number(type)),
				 self)) {
	return false;
      }
    }
    break;
  }
    
  case SYMB_TYPE_INTEGER:  /* (infinite-precision) integer */ 
  case SYMB_TYPE_REAL: /* (infinite-precision) rational */
    break;

  case SYMB_TYPE_NONE: /* no-type */
  case SYMB_TYPE_SET_BOOL:  /* a set of integer values */
  case SYMB_TYPE_SET_INT:  /* a set of integer values */
  case SYMB_TYPE_SET_SYMB: /* a set of symbolic values */
  case SYMB_TYPE_SET_INT_SYMB:/* a set of symbolic and integer values */
  case SYMB_TYPE_ERROR: /* indicates an error */
  default:
    nusmv_assert(false); /* a variable cannot have these types */
  }

  return true;
}


/**Function********************************************************************

  Synopsis           [Returns the type of an expression.]

  Description        [If the expression has been already type-checked
  by the same type-checker, then the type of an expression is returned.
  Otherwise, NULL is returned.
  The parameter 'context' indicate the context where expression
  has been checked. It should be exactly the same as during type checking.
  For outside user this parameter is usually Nil.
    NOTE: To type check a temporarily created expressions use 
  the function TypeChecker_is_expression_wellformed. Expressions passed 
  through compilation package are all type-checked already.
    NOTE: The returned type may be error-type indicating that 
  the expression violates the type system.
    NOTE: all returned types are the memory-sharing types
  (see SymbTablePkg_..._type). So you can compare pointers instead of 
  the type's contents.

  ]

  SideEffects        [TypeChecker_is_expression_wellformed,
  TypeChecker_is_specification_wellformed]

  SeeAlso            []

******************************************************************************/
SymbType_ptr TypeChecker_get_expression_type(TypeChecker_ptr self,
					     node_ptr expression,
					     node_ptr context)
{
  /* wrap exp into context if context is not Nil */
  if (Nil != context) expression = find_node(CONTEXT, context, expression);
  
  return tc_get_expression_type(self, expression);
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [The TypeChecker class private initializer]

  Description        [The TypeChecker class private initializer]

  SideEffects        []

  SeeAlso            [TypeChecker_create]   
  
******************************************************************************/
static void type_checker_init(TypeChecker_ptr self,
			      Type_checking_violation_handler handler,
			      SymbTable_ptr symbolTable)
{
  nusmv_assert(handler);
  nusmv_assert(symbolTable);
  /* members initialisation */
  self->violationHandler = handler;  
  self->expressionTypeMapping = new_assoc();
  self->symbolTable = symbolTable;
}


/**Function********************************************************************

  Synopsis           [The TypeChecker class private deinitializer]

  Description        [The TypeChecker class private deinitializer]

  SideEffects        []

  SeeAlso            [TypeChecker_destroy]   
  
******************************************************************************/
static void type_checker_deinit(TypeChecker_ptr self)
{
  /* members deinitialization */
  free_assoc(self->expressionTypeMapping);
}


/**Function********************************************************************

  Synopsis           [Creates the association between an expression and
  its type]

  Description        [There should be not existing associated type 
  for a given expression.

  Note: if there is (not null) context then the expression
  must be wrapped into CONTEXT expression. For example,
  for an expression exp and not null context ctx, the 
  expression provided for this function should be find_node(CONTEXT, ctx, exp).
  The expression should not be wrapped into Nil context.]

  SideEffects        []

  SeeAlso            [tc_get_expression_type]
******************************************************************************/
static SymbType_ptr tc_set_expression_type (TypeChecker_ptr self,
					    node_ptr expression,
					    SymbType_ptr type)
{
  nusmv_assert(nullType == tc_get_expression_type(self, expression));

  insert_assoc(self->expressionTypeMapping, expression, (node_ptr)type);
  return type;
}


/**Function********************************************************************

  Synopsis           [Returns 1 if there is a mapping 
  between an expression and its type, and 0 otherwise]

  Description        [If the 'type' parameter is not Nil, 
  and the mapping between the expression and its type has already been created,
  then the 'type' will be assigned to the expression's type.

  Note: if there is (not null) context then the expression
  must be wrapped into CONTEXT expression. For example,
  for an expression exp and not null context ctx, the 
  expression provided for this function should be find_node(CONTEXT, ctx, exp).
  The expression should not be wrapped into Nil context.]

  The associated type is created with tc_set_expression_type.]


  SideEffects        []

  SeeAlso            [tc_set_expression_type]
******************************************************************************/
static SymbType_ptr tc_get_expression_type(TypeChecker_ptr self,
					   node_ptr expression)
{
  return SYMB_TYPE(find_assoc(self->expressionTypeMapping, expression));
}


/**Function********************************************************************

  Synopsis           [returns 1 if the given type is "backward compatible",
  i.e. a enum or integer type.]

  Description        [We distinguish "old" types because we may want to turn
  off the type checking on these types for backward
  compatibility. Integer is also considered as "old", because an enum
  of integer values is always casted to Integer.]

  SideEffects        []

  SeeAlso            []
******************************************************************************/
static int tc_is_back_comp_type(SymbType_ptr type)
{
  return SYMB_TYPE_ENUM == SymbType_get_tag(type) ||
    SYMB_TYPE_INTEGER == SymbType_get_tag(type);
}


/**Function********************************************************************

  Synopsis           [Returns the left type, if the right
  one can be implicitly converted to the left one. NULL - otherwise]

  Description        [The implicit conversion is performed
  in accordance to the type order.
  NOTE: only memory-shared types can be given to this function.]

  SideEffects        []

  SeeAlso            []
******************************************************************************/
static SymbType_ptr 
tc_right_operand_implicit_convertion(SymbType_ptr leftType,
				     SymbType_ptr rightType)
{
  /* I expect the given types are shared. So I will not worry about memory */
  
  /* the same pointers, so, the same types, so, just return */
  if (leftType == rightType) return leftType;

  /* pointer are different => the types are different.
     Try to perform implicit convertion of the right operand to the left one.
     No need to care about word, word-array and no-type since
     they cannot be implicitly converted to any other type.
     NB: No-type is created for assignment, and symbolic (predicate-normalised)
     FSM may push assignments into case-expressions, for example.
  */


  /* integer * bool => integer */
   if (SymbType_get_tag(leftType) == SYMB_TYPE_INTEGER) {
     if (SymbType_is_boolean_enum(rightType)) return leftType;
     else return nullType;
   }
   
   /* convert right from bool to Int to decrease number of cases to check */
   if (SymbType_is_boolean_enum(rightType)) {
     rightType = SymbTablePkg_integer_type();
   }
   
   /* real * int => real */
   if (SymbType_get_tag(leftType) == SYMB_TYPE_REAL) {
     return (SymbType_get_tag(rightType) == SYMB_TYPE_INTEGER)
       ? leftType : nullType;
   }

   /* int_symb_enum * (int or symb_enum) => int_symb_enum */
   if (SymbType_is_int_symbolic_enum(leftType)) {
     return (SymbType_get_tag(rightType) == SYMB_TYPE_INTEGER ||
	     SymbType_is_pure_symbolic_enum(rightType))
       ? leftType : nullType;
   }

   /* only set-types remain */

   /* int_set * bool_set=> int_set */
   if (SymbType_get_tag(leftType) == SYMB_TYPE_SET_INT) {
     return SymbType_get_tag(rightType) == SYMB_TYPE_SET_BOOL
       ? leftType : SYMB_TYPE(NULL);
   }

   /* int_symb_set * any set-type => int_symb_set  */
   if (SymbType_get_tag(leftType) == SYMB_TYPE_SET_INT_SYMB) {
     return SymbType_is_set(rightType) ? leftType : SYMB_TYPE(NULL);
   }

   /* all other combinations are illegal */
   return nullType;
}


/**Function********************************************************************

  Synopsis           [Returns one of the given types, if the other one
  can be implicitly converted to the former one. Otherwise - Nil.]

  Description        [The implicit conversion is performed
  in accordance to the type order.
  NOTE: only memory-shared types can be given to this function.]

  SideEffects        []

  SeeAlso            []
******************************************************************************/
static SymbType_ptr 
tc_one_operand_implicit_convertion(SymbType_ptr type1, SymbType_ptr type2)
{
  SymbType_ptr returnType;

  returnType = tc_right_operand_implicit_convertion(type1, type2);
  if (returnType) return returnType; /* conversion was successful */
  /* return with swapped types and return whatever is obtained */
  return tc_right_operand_implicit_convertion(type2, type1);
}


/**Function********************************************************************

  Synopsis           [Returns the minimal type to which the both given types
  can be converted, or Nil if there is none.]

  Description        [The implicit conversion is performed
  in accordance to the type order.
  NOTE: only memory-shared types can be given to this function.]

  SideEffects        []

  SeeAlso            []
******************************************************************************/
static SymbType_ptr tc_both_operands_implicit_convertion (SymbType_ptr type1,
							  SymbType_ptr type2)
{
  /* I expect the given types are shared. So I will not worry about memory */

  /* the same pointers, so, the same types, so, just return the type */
  if (type1 == type2) return type1;

  /* pointer are different => the types are different.
     Try to perform implicit convertion of both operand to
     a minimal type.
     No need to care about word, word-array and no-type since
     they cannot be implicitly converted to any other type.
     NB: No-type is created for assignment, and (predicate-normalised)
     symbolic FSM may push assignments into case-expressions, for example.
  */

  /* cast boolean to int (to decrease the number of cases)*/
  if (SymbType_is_boolean_enum(type1)) type1 = SymbTablePkg_integer_type();
  if (SymbType_is_boolean_enum(type2)) type2 = SymbTablePkg_integer_type();

  /* (integer * bool) or (bool * integer) => integer */
  if (type1 == type2) return type1;
  
   
  /* both are [Int or Real] -> Real */
  if ( ( SymbType_get_tag(type1) == SYMB_TYPE_INTEGER ||
	 SymbType_get_tag(type1) == SYMB_TYPE_REAL ) && 
       ( SymbType_get_tag(type2) == SYMB_TYPE_INTEGER ||
	 SymbType_get_tag(type2) == SYMB_TYPE_REAL ) ) {
    return SymbTablePkg_real_type();
  }
  
  /* both are [Int or Symbolic or SymbolicInt] -> SymbolicInt */
  if ( ( SymbType_get_tag(type1) == SYMB_TYPE_INTEGER ||
	 SymbType_is_pure_symbolic_enum(type1) || 
	 SymbType_is_int_symbolic_enum(type1) ) &&
       ( SymbType_get_tag(type2) == SYMB_TYPE_INTEGER ||
	 SymbType_is_pure_symbolic_enum(type2) || 
	 SymbType_is_int_symbolic_enum(type2) ) ) {
    return SymbTablePkg_int_symbolic_enum_type();
  }

  /* only set types remain */

  /* both are [int_set or bool_set] => int_set */
  if ( (SymbType_get_tag(type1) == SYMB_TYPE_SET_BOOL ||
	SymbType_get_tag(type1) == SYMB_TYPE_SET_INT) &&
       (SymbType_get_tag(type2) == SYMB_TYPE_SET_BOOL ||
	SymbType_get_tag(type2) == SYMB_TYPE_SET_INT)) {
    return SymbTablePkg_integer_set_type();
  }
  /* both are any set type => int_symb_set */
  if (SymbType_is_set(type1) || SymbType_is_set(type2)) {
    return SymbTablePkg_integer_symbolic_set_type();
  }
  
  /* the given types are incompatible => error */
  return SYMB_TYPE(NULL);
}


/**Function********************************************************************

  Synopsis           [Returns a minimal set-type which the given type
  can be implicitly converted to, or NULL if this is impossible.]

  Description        [
  The implicit conversion is performed in accordance to the type order.
  NOTE: only memory-shared types can be given to this function.]

  SideEffects        []

  SeeAlso            []
******************************************************************************/
static SymbType_ptr tc_implicit_convertion_to_set_type (SymbType_ptr type)
{
  /* convert to set types if it is required:
     bool -> bool_set
     int -> int_set
     symb -> symb_set
     int_symb_enum -> int_symb_set
     any set type -> the same set type
     other type -> NULL
  */
  if (SymbType_is_boolean_enum(type)) {
    return SymbTablePkg_boolean_set_type();
  }
  if (SymbType_get_tag(type) == SYMB_TYPE_INTEGER) {
    return SymbTablePkg_integer_set_type();
  }
  else if (SymbType_is_pure_symbolic_enum(type)) {
    return SymbTablePkg_symbolic_set_type();
  }
  else if (SymbType_is_int_symbolic_enum(type)) {
    return SymbTablePkg_integer_symbolic_set_type();
  } 
  else if (SymbType_is_set(type)) {
    return type;
  } 
  else return nullType;
}

/**Function********************************************************************

  Synopsis           [This function is opposite to
  tc_implicit_convertion_to_set_type, i.e. if the given type is one of
  the set-types, then the type without "set" suffix is
  returned. Otherwise the time is returned without change.]

  Description        [
  More precisely the following conversion takes place:
  boolean-set -> boolean
  integer-set ->integer
  symbolic-set -> symbolic-enum
  integer-symbolic-set -> integer-symbolic-set
  another type -> the same time

  The implicit conversion is performed in accordance to the type order.
  NOTE: only memory-shared types can be given to this function.]

  SideEffects        []

  SeeAlso            [tc_implicit_convertion_to_set_type]
******************************************************************************/
static SymbType_ptr tc_convertion_from_set_type (SymbType_ptr type)
{
  if (SymbType_get_tag(type) == SYMB_TYPE_SET_BOOL) {
    return SymbTablePkg_boolean_enum_type();
  }
  if (SymbType_get_tag(type) == SYMB_TYPE_SET_INT) {
    return SymbTablePkg_integer_type();
  }
  else if (SymbType_get_tag(type) == SYMB_TYPE_SET_SYMB) {
    return SymbTablePkg_pure_symbolic_enum_type();
  }
  else if (SymbType_get_tag(type) == SYMB_TYPE_SET_INT_SYMB) {
    return SymbTablePkg_int_symbolic_enum_type();
  }
  else return type;
}


/**Function********************************************************************

  Synopsis           [Perform type checking of an expression, and returns its
  type]

  Description        [This function is the core of the expression type
  checking.
  
  The expression may by unmodified (after compilation), flattened or 
  flattened+expanded.

  The return value is the type of the expression.  If an expression
  violates the type system, the violation handler function is invoked
  (See Type_checking_violation_handler for more info) and "error" type
  is returned (not NULL).
 
  

  NB: As it is said in TypeChecker.h, the expression type checking package
  uses only memory-shared types (SymbTablePkg_..._type).
  
  NB: This function does not perform the type checking on already
  checked expressions (even if they were erroneous), but just returns
  their types (possibly error-type). This avoids multiple error and warning
  messages for the same subexpression.]

  SideEffects        []

  SeeAlso            [type_checking_violation_handler]

******************************************************************************/
static SymbType_ptr
type_checker_check_expression(TypeChecker_ptr self,
			      node_ptr expression,
			      node_ptr context)
{
  /* wrap expression into the context. This is required by 
     the facilities which remembers the type of expressions 
     and by the violation handler.
  */
  node_ptr wrappedExpr = (Nil == context) ? expression
    : find_node(CONTEXT, context, expression);
  {
    SymbType_ptr tmp = tc_get_expression_type(self, wrappedExpr);
    /* if this expression has been checked => return the associated type,
       (if the type is error-type, this expression is erroneous).
    */
    if (nullType != tmp) return tmp;
  }
  
  switch (node_get_type(expression)) {

  case CONTEXT: {
    SymbType_ptr type = type_checker_check_expression(self, cdr(expression),
						     car(expression));
    /* since before remembering the type an expression is 
       wrapped into CONTEXT, that CONTEXT can be equal to
       this CONTEXT expression, and therefore its type has 
       been remembered in the above invocation of type_checker_check_expression
    */
    if (nullType == tc_get_expression_type(self, wrappedExpr)) {
      tc_set_expression_type(self, wrappedExpr, type);
    }
    return type;
  }
    
  case TRUEEXP:
  case FALSEEXP:
    return tc_set_expression_type(self, wrappedExpr,
                                  SymbTablePkg_boolean_enum_type());
    
  case NUMBER: {
    int num = (int)car(expression);
       return tc_set_expression_type(self, wrappedExpr,
				     (num == 1 || num == 0)
				     ? SymbTablePkg_boolean_enum_type()
	                             : SymbTablePkg_integer_type());
  }

  case NUMBER_WORD: {
    /* during parsing Word constants are normilised to WordNumber_ptr
        (the left chilf).
    */
    int wordWidth = WordNumber_get_width(WORD_NUMBER(car(expression)));
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_word_type(wordWidth));
  }

  case NUMBER_FRAC:
  case NUMBER_REAL:
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_real_type());

  case TWODOTS:
    /* two-dots have two NUMBER and nothing else */
    nusmv_assert(node_get_type(car(expression)) == NUMBER && 
		 node_get_type(cdr(expression)) == NUMBER);
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_integer_set_type());


  case BIT:
#warning [AT] this kind of expressions is not checked, since it is artificially created
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_boolean_enum_type());

    /* these exprs mean a variable, a constant, a define or a parameter */
  case DOT:
  case ARRAY:
  case ATOM: {
    /* First, find out what this identifier may mean */
    node_ptr simpleName = find_atom(expression);
    /* Obtain the resolved name. CompileFlatten_resolve_name should
       not be applied to a parameter and parameter can be only ATOM.
    */
    node_ptr resolvedName = node_get_type(expression) == ATOM
      ? find_node(DOT, context, simpleName)
      : CompileFlatten_resolve_name(simpleName, context);
    
    node_ptr parameter = lookup_param_hash(resolvedName);/*only ATOM can be par*/
    boolean isVar = SymbTable_is_symbol_var(self->symbolTable, resolvedName);
    boolean isDefine = SymbTable_is_symbol_define(self->symbolTable, resolvedName);
    /* check the original ATOM expression and then the resolved name 
       for being a constant
    */
    boolean isConstant = false;
    if (ATOM == node_get_type(simpleName)) {
      isConstant = SymbTable_is_symbol_constant(self->symbolTable, simpleName);
    }
    if (!isConstant) {
      isConstant = SymbTable_is_symbol_constant(self->symbolTable, resolvedName);
    }

    /* An identifier has not been declared => error */
    if ( Nil == parameter && ! isVar && ! isDefine && ! isConstant) {
      SymbType_ptr type = self->violationHandler(TC_VIOLATION_UNDEF_IDENTIFIER,
				    wrappedExpr, self)
	? SymbTablePkg_error_type() /* ERROR */
	: SymbTablePkg_pure_symbolic_enum_type();
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* An identifier has more than one meaning => error */
    if ( ((parameter != Nil) + isVar + isDefine + isConstant) > 1 ) {
     SymbType_ptr type = self->violationHandler(TC_VIOLATION_AMBIGUOUS_IDENTIFIER,
				    wrappedExpr, self)
	? SymbTablePkg_error_type() /* ERROR */
	: SymbTablePkg_pure_symbolic_enum_type();
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* this is a constant */
    if (isConstant) return tc_set_expression_type(self, wrappedExpr,
					   SymbTablePkg_pure_symbolic_enum_type());

    /* this is a parameter */
    if (parameter != Nil) {
     SymbType_ptr type = type_checker_check_expression(self, parameter, context);
     return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* this is a define */
    if (isDefine) {
      /* the bodies of defines are cached(in BddEnc).
	 So, the type of expressions will also be correctly cached.
	 
	 Only flattened body of define is checked, since not
	 flattened body is not used anywhere.
      */
      node_ptr flatDef = SymbTable_get_define_flatten_body(self->symbolTable,
							   resolvedName);
      /* the context is Nil because expr is already flattened */
      SymbType_ptr type = type_checker_check_expression(self, flatDef, Nil);
      
      return tc_set_expression_type(self, wrappedExpr, type);
    }
    
    /* This is a variable */
    {
      SymbType_ptr type;
      nusmv_assert(isVar);
      type = SymbTable_get_var_type(self->symbolTable, resolvedName);
      /* Convert the syntactic type to a symbol type */
      /* here we share the variable's type to allow 
	 comparison of pointers instead of content of the types 
      */
      if (SymbType_get_tag(type) == SYMB_TYPE_INTEGER) {
	type = SymbTablePkg_integer_type();
      }
      else if (SymbType_get_tag(type) == SYMB_TYPE_REAL) {
	type = SymbTablePkg_real_type();
      }
      else if (SymbType_get_tag(type) == SYMB_TYPE_WORD) {
	type = SymbTablePkg_word_type(SymbType_get_word_width(type));
      }
      else if (SymbType_get_tag(type) == SYMB_TYPE_WORDARRAY) {
	int awidth = SymbType_get_wordarray_awidth(type);
	int vwidth = SymbType_get_wordarray_vwidth(type); 
	type = SymbTablePkg_wordarray_type(awidth, vwidth); 
      }
      else if (SymbType_get_tag(type) == SYMB_TYPE_ENUM) {
	if (SymbType_is_boolean_enum(type)) {
	  type = SymbTablePkg_boolean_enum_type();
	}
	else if (SymbType_is_pure_int_enum(type)) { 
	  type = SymbTablePkg_integer_type(); /* convert to Integer */
	}
	else if (SymbType_is_pure_symbolic_enum(type)) {
	  type = SymbTablePkg_pure_symbolic_enum_type();
	}
	else {
	  nusmv_assert(SymbType_is_int_symbolic_enum(type));
	  type = SymbTablePkg_int_symbolic_enum_type();
	}
      }
      else {
	nusmv_assert(false); /* unknown type */
      }

      return tc_set_expression_type(self, wrappedExpr, type);
    } /* end of isVar */    
  } /* case ATOM, DOT, ARRAY */

    /*  boolean unary operand */
  case EX:
  case AX:
  case EF:
  case AF:
  case EG:
  case AG:

  case NOT: /* "NOT" may also have Word */

  case OP_GLOBAL:
  case OP_PREC:
  case OP_NOTPRECNOT:
  case OP_FUTURE:
  case OP_NEXT:
  case OP_HISTORICAL:
  case OP_ONCE:

    /* For EBF ABF EBG ABG - ignore the number..number part (cdr) */
  case EBF:
  case ABF:
  case EBG:
  case ABG: { 
    /* get the operand's type */
    SymbType_ptr type
      = type_checker_check_expression(self, car(expression), context);

    if (SymbType_is_error(type)) { /* earlier error */
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* the operand must be boolean (or Word for NOT) */
    if (SymbType_is_boolean_enum(type) || 
	( node_get_type(expression) == NOT &&
	  SymbType_get_tag(type) == SYMB_TYPE_WORD
        ) ) {
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* is this a type error ? */
    if (self->violationHandler(
	   tc_is_back_comp_type(type) 
	      ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
	   wrappedExpr,
	   self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> keep the type what ever it is */
    return tc_set_expression_type(self, wrappedExpr, type);
  }
 
		
    /* casts: boolean -> Word[1] and Word[1] -> boolean. */
  case CAST_BOOL:
  case CAST_WORD1: {
    /* get the operands' type */
    SymbType_ptr type
      = type_checker_check_expression(self, car(expression), context);

    if (SymbType_is_error(type)) { /* earlier error */
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* if the expression is cast_bool, operand is word[1] */
    if (node_get_type(expression) == CAST_BOOL && SymbType_is_word_1(type)) {
      return tc_set_expression_type(self, wrappedExpr,
				         SymbTablePkg_boolean_enum_type());
    }
    
    /* OR the expression must be cast_word1, operand is boolean */
    if (node_get_type(expression) == CAST_WORD1 &&
	SymbType_is_boolean_enum(type)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_word_type(1));
    }
	
    /* is this a type error ? */
    if (self->violationHandler(TC_VIOLATION_TYPE_MANDATORY, wrappedExpr, self)){
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    
    /* this is not an error after all -> create the proper return type */
    return tc_set_expression_type(self, wrappedExpr,
				    node_get_type(expression) == CAST_BOOL
				    ? SymbTablePkg_boolean_enum_type()
				    : SymbTablePkg_word_type(1));
  }


    /* 'init' and 'next' do not change the type of the expression */
  case SMALLINIT:
  case NEXT: {
    /* apply find_atom on the 'init' or 'next' expression and check
       the obtained expression for being already checked. 

       The obtained expression will be the same as original one but
       some parts of NuSMV generate 'init' and 'next' expression
       with find_node, so both original and find_atom'ed expression
       should be checked
    */
    node_ptr normalisedExpr = find_atom(expression);
    SymbType_ptr type;
    /*wrap expression into context.This is required by tc_get_expression_type*/
    if (Nil != context) {
      normalisedExpr = find_node(CONTEXT, context, normalisedExpr);
    }

    type = tc_get_expression_type(self, normalisedExpr);
    /* find_atom'ed expression has been already checked. Remeber and return */
    if (nullType != type) { 
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* check the expression now */
    type = type_checker_check_expression(self, car(expression), context);
    /* associate both the original expression and the find_atom'ed one.*/
    tc_set_expression_type(self, wrappedExpr, type);
    if (normalisedExpr != wrappedExpr) {
      tc_set_expression_type(self, normalisedExpr, type);
    }
    return type;
  }


    /* binary: w[N]*w[N] or w[N]*boolean or boolean*W[N] or boolean*boolean */
  case CONCATENATION: {
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)) {/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    /* convert boolean operands to Word[1] (to simplify the code */
    if (SymbType_is_boolean_enum(type1)) type1 = SymbTablePkg_word_type(1);
    if (SymbType_is_boolean_enum(type2)) type2 = SymbTablePkg_word_type(1);

    /* both operands must be Word */
    if (SymbType_get_tag(type1) == SYMB_TYPE_WORD &&
	SymbType_get_tag(type2) == SYMB_TYPE_WORD) {
      int newWidth
	= SymbType_get_word_width(type1) + SymbType_get_word_width(type2);
      return tc_set_expression_type(self, wrappedExpr,
					SymbTablePkg_word_type(newWidth));
    }

    /* is this a type error ? */
    if (self->violationHandler(TC_VIOLATION_TYPE_MANDATORY, wrappedExpr, self)){
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    
    /* this is not an error after all -> create Word type of some width*/
    return tc_set_expression_type(self, wrappedExpr,SymbTablePkg_word_type(1));
  }

    /* first operand is Word[N], then two numbers K and L, K >=L and N>=K. */
  case BIT_SELECTION: {
    /* get the operand' type */
    SymbType_ptr type
      = type_checker_check_expression(self, car(expression), context);
    int width;
    int highBound;
    int lowBound;
    
    if (SymbType_is_error(type)) { /* earlier error */
      return tc_set_expression_type(self, wrappedExpr, type);
    }

    /* At the moment, second and third operands must be NUMBERS
       (And this is checked in flattener 
    */
    /*just consistency check*/
    nusmv_assert(COLON == node_get_type(cdr(expression)));
    nusmv_assert(NUMBER == node_get_type(car(cdr(expression))));
    nusmv_assert(NUMBER == node_get_type(cdr(cdr(expression))));
    
    /* check the first operand type */
    if (SymbType_get_tag(type) != SYMB_TYPE_WORD) { /* ERROR */
      if(self->violationHandler(TC_VIOLATION_TYPE_MANDATORY, wrappedExpr,self)){
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_error_type());
      }
      else { /* return arbitrary Word type */
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_word_type(1));
      }
    }
    width = SymbType_get_word_width(type);
    highBound = (int)car(car(cdr(expression)));
    lowBound  = (int)car(cdr(cdr(expression)));
			 
    /* checks the bit width */
    if (width <= highBound || highBound < lowBound || lowBound < 0) {/* ERROR */
      if (self->violationHandler(TC_VIOLATION_OUT_OF_WORD_WIDTH, wrappedExpr,
				 self)) {
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_error_type());
      }
      else { /* give some realistic bit specifiers */
	highBound = 0;
	lowBound = 0;
      }
    }
    /* everything is OK */
    return tc_set_expression_type(self, wrappedExpr, 
			 SymbTablePkg_word_type(highBound - lowBound + 1));
  }

  case WAWRITE: {
    /* WRITE : WORDARRAY[a,v] x WORD[a] x WORD[v] -> WORDARRAY[a,v] */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, car(cdr(expression)), context);
    SymbType_ptr type3
      = type_checker_check_expression(self, cdr(cdr(expression)), context);

    nusmv_assert(node_get_type(cdr(expression)) == WAWRITE);

    if (SymbType_is_error(type1) || SymbType_is_error(type2) ||
	SymbType_is_error(type3)) {
      /* handle error in arguments */
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* ERROR with kinds of subexpressions */
    if (SymbType_get_tag(type1) != SYMB_TYPE_WORDARRAY ||
	SymbType_get_tag(type2) != SYMB_TYPE_WORD ||
	SymbType_get_tag(type3) != SYMB_TYPE_WORD ) {
      if(self->violationHandler(TC_VIOLATION_TYPE_MANDATORY, wrappedExpr, self)){
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_error_type());
      }
      else { /* return arbitrary WordArray type */
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_wordarray_type(1,1));
      }
    }

    /* ERROR with the widths */
    if (SymbType_get_wordarray_awidth(type1) != SymbType_get_word_width(type2) ||
	SymbType_get_wordarray_vwidth(type1) != SymbType_get_word_width(type3)) {
      if (self->violationHandler(TC_VIOLATION_OUT_OF_WORDARRAY_WIDTH,
				 wrappedExpr, self)) {
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_error_type());
      }
    }

    /* No errors, then return the type of the first argument (whatever it is) */
    return tc_set_expression_type(self, wrappedExpr, type1);
  }

  case WAREAD: {
    /* WORDARRAY[a,v] x WORD[a] -> WORD[v] */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);

    if (SymbType_is_error(type1) || SymbType_is_error(type2)) {
      /* handle error in arguments */
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* ERROR with kinds of subexpressions */
    if (SymbType_get_tag(type1) != SYMB_TYPE_WORDARRAY ||
	SymbType_get_tag(type2) != SYMB_TYPE_WORD) {
      if(self->violationHandler(TC_VIOLATION_TYPE_MANDATORY, wrappedExpr, self)){
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_error_type());
      }
      else { /* return arbitrary WordArray type */
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_wordarray_type(1,1));
      }
    }

    /* ERROR with the widths */
    if (SymbType_get_wordarray_awidth(type1) != SymbType_get_word_width(type2)) {
      if (self->violationHandler(TC_VIOLATION_OUT_OF_WORDARRAY_WIDTH,
				 wrappedExpr, self)) {
	return tc_set_expression_type(self, wrappedExpr,
				      SymbTablePkg_error_type());
      }
    }

    /* No errors, then return WORD[v] */
    return tc_set_expression_type(self, wrappedExpr,
		  SymbTablePkg_word_type(SymbType_get_wordarray_vwidth(type1)));
  }


    /* arithmetic operators: b*b->i, i*i->i, r*r->r, w[n]*w[n]->w[n]
       + implicit conversion
    */
  case PLUS:
  case MINUS:
  case TIMES:
  case DIVIDE: { 
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    SymbType_ptr returnType;
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    
    /* implicit conversion of one of the operands */
    returnType = tc_one_operand_implicit_convertion(type1, type2);
    

    /*correct types are boolean, int ,b*b->i, i*i->i, r*r->r, w[n]*w[n]->w[n]*/
    if (nullType != returnType && 
	( SymbType_is_boolean_enum(returnType) ||
	  SymbType_get_tag(returnType) == SYMB_TYPE_INTEGER ||
	  SymbType_get_tag(returnType) == SYMB_TYPE_REAL ||
	   SymbType_get_tag(returnType) == SYMB_TYPE_WORD
	)) {
      /* boolean type should be converted to int */
      if (SymbType_is_boolean_enum(returnType)) {
	returnType = SymbTablePkg_integer_type();
      }
      return tc_set_expression_type(self, wrappedExpr, returnType);
    }
    
     /* is this a type error ? */
    if (self->violationHandler(
	      (tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2))
	          ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
	      wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> keep one of the types whatever it is*/
    return tc_set_expression_type(self, wrappedExpr, type1);
  }

    /* mod operator: i*i->i, w[n]*w[n]->w[n] (no implicit conversion)*/
  case MOD: { 
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    
    /* w[N] * w[N] -> w[N] */
    if ( type1 == type2 && /*types are shared => same types == same pointers*/
	 SymbType_get_tag(type1) == SYMB_TYPE_WORD ) {
      return tc_set_expression_type(self, wrappedExpr, type1);
    }
    
    /* int * int -> int, int * 2 -> bool */
    if ( type1 == type2 && /*types are shared => same types == same pointers*/
	 SymbType_get_tag(type1) == SYMB_TYPE_INTEGER ) {
      /* special case "int mod 2" -> bool */
      if (NUMBER == node_get_type(cdr(expression)) &&
	  2 == (int)car(cdr(expression))) {
	type1 =  SymbTablePkg_boolean_enum_type();
      }
      
      return tc_set_expression_type(self, wrappedExpr, type1);
    }
    
     /* is this a type error ? */
    if (self->violationHandler(
	      (tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2))
	          ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
	      wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> keep one of the types whatever it is*/
    return tc_set_expression_type(self, wrappedExpr, type1);
  }

    /* left operand is word and the right is Int or Bool */
  case LSHIFT: 
  case RSHIFT: /*case LROTATE: case RROTATE:*/ { 
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr, 
				    SymbTablePkg_error_type());
    }
    
    /* correct types are w[n]*boolean, w[n]*int */
    if (SymbType_get_tag(type1) == SYMB_TYPE_WORD &&
	( SymbType_is_boolean_enum(type2) ||
	  SymbType_get_tag(type2) == SYMB_TYPE_INTEGER)) {
      /* In future, maybe, it is worth to check that if right operand
	 is constant its value < width of Word (otherwise - warning)
      */
      return tc_set_expression_type(self, wrappedExpr, type1);
    }
    
    /* is this a type error ? */
    if (self->violationHandler(TC_VIOLATION_TYPE_MANDATORY, wrappedExpr, self)){
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> keep the left operand type */
    return tc_set_expression_type(self, wrappedExpr, type1);
  }


    /* the operands should be implicitly converted to one type,
       otherwise they must be either boolean or Word[1] 
    */
  case EQUAL:
  case NOTEQUAL: {
    /* get the operands' type */
    SymbType_ptr type1 = type_checker_check_expression(self, car(expression),
						      context);
    SymbType_ptr type2 = type_checker_check_expression(self, cdr(expression),
						      context);
    SymbType_ptr returnType;
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr, 
				    SymbTablePkg_error_type());
    }
    
    /* implicit conversion of one of the operands */
    returnType = tc_one_operand_implicit_convertion(type1, type2);

    /* returned type is enum, int, real, word or word-array
       or operands are boolean with Word[1].
       Virtually, only set types cannot be compared.
    */
    if ( (nullType != returnType &&
	   (SymbType_get_tag(returnType) == SYMB_TYPE_ENUM ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_INTEGER ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_WORD ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_WORDARRAY ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_REAL
           )
	 ) ||
	 (SymbType_is_boolean_enum(type1) && SymbType_is_word_1(type2)) ||
	 (SymbType_is_boolean_enum(type2) && SymbType_is_word_1(type1)) ){
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_boolean_enum_type());
    }
    /* is this a type error ? */
    if (self->violationHandler(
		tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
		  ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
		wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return the boolean type */
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_boolean_enum_type());
  }

    
    /* Assignemnt. The right operand should be implicitly converted to the
       left operand's type (or its set-type counterpart),
       otherwise they must be boolean with Word[1].
       The type of the whole expression is no-type.
       It is required to process EQDEF with other expressions since 
       in symbolic-FSM after predicate-normalisation EQDEF can become
       a part of case-expression in init, trans, etc, lists.

       NB: the left operand is not checked for being a variable since
       parser guarantees this.
    */
  case EQDEF: {
    /* get the operands' type */
    SymbType_ptr type1 = type_checker_check_expression(self, car(expression),
						      context);
    SymbType_ptr type2 = type_checker_check_expression(self, cdr(expression),
						      context);
    SymbType_ptr returnType;
	
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* If the  right operand is of a set-type, convert the set-type to a
       usual one.
       Convert the right operand's type then to the type of the left 
       operands.
    */
    returnType = tc_right_operand_implicit_convertion(type1,
					 tc_convertion_from_set_type(type2));

    /* returned type is any enum-type, int, real, word or
        word-array or the operands are boolean with Word[1]
        Virtually, only set types cannot be compared.
   */
    if ( (nullType != returnType &&
	  (SymbType_get_tag(returnType) == SYMB_TYPE_ENUM ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_INTEGER ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_WORD ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_WORDARRAY ||
	    SymbType_get_tag(returnType) == SYMB_TYPE_REAL
           )
	 )||
	 (SymbType_is_boolean_enum(type1) && SymbType_is_word_1(type2)) ||
	 (SymbType_is_boolean_enum(type2) && SymbType_is_word_1(type1)) ){
      return tc_set_expression_type(self, wrappedExpr, SymbTablePkg_no_type());
    }
  
    /* is this a type error ? */
    if (self->violationHandler(
		tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
		  ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
		wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return the no-type */
    return tc_set_expression_type(self, wrappedExpr, SymbTablePkg_no_type());
  }


    /*  allowed types(with implicit conversion):
	b*b, i*i, r*r, w[n]*w[n], b*w[1], w[1]*b
    */
  case LT:
  case GT:
  case LE:
  case GE: {
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    SymbType_ptr returnType;
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    
    /* implicit conversion on both operands */
    returnType = tc_one_operand_implicit_convertion(type1, type2);
    

    /* return type is not-null and arithmetic 
       or operands are boolean or Word[1]
    */
    if ( ( nullType != returnType && 
	   ( SymbType_is_boolean_enum(returnType) ||
	     SymbType_get_tag(returnType) == SYMB_TYPE_INTEGER ||
	     SymbType_get_tag(returnType) == SYMB_TYPE_REAL ||
	     SymbType_get_tag(returnType) == SYMB_TYPE_WORD )
	 ) ||
	 (SymbType_is_boolean_enum(type1) && SymbType_is_word_1(type2)) ||
	 (SymbType_is_boolean_enum(type2) && SymbType_is_word_1(type1)) ) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_boolean_enum_type());
    }
    /* is this a type error ? */
    if (self->violationHandler(
	       tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
	         ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
	       wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return the boolean type*/
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_boolean_enum_type());
  }


   /* the operator are boolean or Words of the same size */
  case OR:
  case XOR:
  case XNOR:
  case IMPLIES:
  case IFF: {
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* both operands must be boolean or Word of the same size */
    if ( (SymbType_is_boolean_enum(type1) && SymbType_is_boolean_enum(type2)) 
	 || ( SymbType_get_tag(type1) == SYMB_TYPE_WORD && type1 == type2) ) {
      return tc_set_expression_type(self, wrappedExpr, type1);
    }

    /* is this a type error ? */
    if (self->violationHandler(
	       tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
	           ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY,
	       wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> keep one of the types whatever it is*/
    return tc_set_expression_type(self, wrappedExpr, type1);
  }


    /* AND is an unusual binary operator because sometimes it is used
       as connector in expression lists.
       The distinction from usual binary logical operator:
         1. if one of the operands is Nill => it is skipped
         2. operands can be boolean as well as of no-type.
       This can happen only for artificially created expressions (not
       parsed ones)
    */
  case AND: {
    SymbType_ptr type1;
    SymbType_ptr type2;
    if (Nil == car(expression)) {
      type2 = type_checker_check_expression(self, cdr(expression), context);
      return tc_set_expression_type(self, wrappedExpr, type2);
    }

    if (Nil == cdr(expression)) {
      type1 = type_checker_check_expression(self, car(expression), context);
      return tc_set_expression_type(self, wrappedExpr, type1);
    }
    
    /* get the operands' type */
    type1 = type_checker_check_expression(self, car(expression), context);
    type2 = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* both operands must be boolean or Word of the same size */
    if ( (SymbType_is_boolean_enum(type1) && SymbType_is_boolean_enum(type2)) 
	 || ( SymbType_get_tag(type1) == SYMB_TYPE_WORD && type1 == type2) ) {
      return tc_set_expression_type(self, wrappedExpr, type1);
    }
    
    /* AND is a list connector; operands must be boolean or of no-type*/
    if ( (SymbType_is_boolean_enum(type1) ||
	  SymbType_get_tag(type1) == SYMB_TYPE_NONE) &&
	 (SymbType_is_boolean_enum(type2) ||
	  SymbType_get_tag(type2) == SYMB_TYPE_NONE) ) {
      return tc_set_expression_type(self, wrappedExpr, type1);
    }

    /* is this a type error ? */
    if (self->violationHandler(
	       tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
	           ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY,
	       wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> keep one of the types whatever it is*/
    return tc_set_expression_type(self, wrappedExpr, type1);
  }
    

    /* both operands are converted to (corresponding) set-types and
       then to the minimal common type.
       As result only the following types can be here:
       any of enum types (bool, symb, int-symb), int  or any of set type
       (i.e. integer-set, symbolic-set, integer-symbolic-set)
    */
  case UNION: {
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    SymbType_ptr returnType;
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
     return tc_set_expression_type(self, wrappedExpr,
				   SymbTablePkg_error_type());
    }
    
    /* convert operands to set types (if this is possible) and then  
       convert them to one type.
    */
    {
      SymbType_ptr tmp1 = tc_implicit_convertion_to_set_type(type1);
      SymbType_ptr tmp2 = tc_implicit_convertion_to_set_type(type2);

      if (nullType != tmp1 && nullType != tmp2) {
	returnType = tc_both_operands_implicit_convertion(tmp1, tmp2);
      }
      else returnType = nullType;
    }

    /* both types are convertable to a set type */
    if (nullType != returnType) {
      return tc_set_expression_type(self, wrappedExpr, returnType);
    }
    
    /* is this a type error ? */
    if (self->violationHandler(
		 tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
		 ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY,
		 wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return the boolean type */
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_boolean_enum_type());
  }

    /* both operands are converted to (corresponding) set-types and
       then ONE of the operand is converted to the type of the othe one.
       As result only the following types can be here:
       any of enum types (bool, symb, int-symb), int  or any of set type
       (i.e. integer-set, symbolic-set, integer-symbolic-set)
    */
  case SETIN:  { 
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    SymbType_ptr returnType;
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
        
    /* convert operands to set types (if this is possible) and then  
       convert one of the them to the other one.
    */
    {
      SymbType_ptr tmp1 = tc_implicit_convertion_to_set_type(type1);
      SymbType_ptr tmp2 = tc_implicit_convertion_to_set_type(type2);

      if (nullType != tmp1 && nullType != tmp2) {
	returnType = tc_one_operand_implicit_convertion(tmp1, tmp2);
      }
      else returnType = nullType;
    }
    
    /* Note that the types should be convertable to one of the set types (
       and be therefore, any of enum (i.e. Bool, Symb and IntSymb),
       Int, any of set types (i.e. int-set, symb-set, int-symb-set)).
       Real, Word and WordArray should not be here because
       even though assignment internally behaves the same way as 
       "in"-operatoris, assignment is never syntactically converted to 
       SETIN expression.
    */
    if (nullType != returnType) {
      return tc_set_expression_type(self, wrappedExpr,
				        SymbTablePkg_boolean_enum_type());
    }

    /* is this a type error ? */
    if (self->violationHandler(
		tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
		? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
		wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return the boolean type */
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_boolean_enum_type());
  }

    /* left expression is bool, and the union of right expression 
       and the next case (if it is not the end of list) is returned.
    */
  case CASE: {
    /* get the operands' type */
    SymbType_ptr condType
      = type_checker_check_expression(self, car(car(expression)), context);
    SymbType_ptr expType
      = type_checker_check_expression(self, cdr(car(expression)), context);
    SymbType_ptr nextCaseType;
    SymbType_ptr returnType;
    
    if (SymbType_is_error(condType) || SymbType_is_error(expType)) {/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
    
   /* left operand of ':' should be boolean */
    if ( (!SymbType_is_boolean_enum(condType)) && 
	 self->violationHandler(
		 tc_is_back_comp_type(condType) 
		   ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY,
		 Nil == context ? car(expression)
		    : find_node(CONTEXT, context, car(expression)),
		 self)) {
      return tc_set_expression_type(self, wrappedExpr, SymbTablePkg_error_type());
    }

    /* during parsing only FAILURE and CASE nodes can be in a CASE list */
    nusmv_assert(FAILURE == node_get_type(cdr(expression)) ||
		 CASE == node_get_type(cdr(expression)));

    /* this is the last element of case-list */
    if (FAILURE == node_get_type(cdr(expression))) {
      return tc_set_expression_type(self, wrappedExpr, expType);
    }

    /* there are more case-elements in the list */
    nextCaseType = type_checker_check_expression(self, cdr(expression),
						context);
    if (SymbType_is_error(nextCaseType)) {/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* both CASE expressions should be convertable to a common type.
       If one of the expressions is of a set-type then before implicit
       convertion the other expression is converted to 
       a corresponding set-type.
    */
    {
      /* if one of the type is a set-type, convert also the other one to
	 a corresponding set type
      */
      SymbType_ptr tmp1 = SymbType_is_set(nextCaseType) 
	? tc_implicit_convertion_to_set_type(expType) : expType;

      SymbType_ptr tmp2 = SymbType_is_set(expType) 
	? tc_implicit_convertion_to_set_type(nextCaseType) : nextCaseType;
      
      if (nullType != tmp1 && nullType != tmp2) {
	returnType = tc_both_operands_implicit_convertion(tmp1, tmp2);
      }
      else returnType = nullType;
    }
    
    /* we do not care which type exactly is obtained, since only
       correct type could pass the above code
    */
    if (nullType != returnType) { 
      return tc_set_expression_type(self, wrappedExpr, returnType);
    }
    
    /* is this a type error ? */
    if (self->violationHandler(
	      tc_is_back_comp_type(expType) && tc_is_back_comp_type(nextCaseType)
		 ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY, 
	      wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return one of the types  */
    return tc_set_expression_type(self, wrappedExpr, expType);
  }

    /* both operands are boolean */
  case AU:
  case EU:
  case UNTIL:
  case SINCE: {
    /* get the operands' type */
    SymbType_ptr type1
      = type_checker_check_expression(self, car(expression), context);
    SymbType_ptr type2
      = type_checker_check_expression(self, cdr(expression), context);
    
    if (SymbType_is_error(type1) || SymbType_is_error(type2)){/*earlier error*/
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }
     
    /* both operands must be boolean or Word of the same size */
    if (SymbType_is_boolean_enum(type1) && SymbType_is_boolean_enum(type2)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_boolean_enum_type());
    }
    
    /* is this a type error ? */
    if (self->violationHandler(
		 tc_is_back_comp_type(type1) && tc_is_back_comp_type(type2) 
		   ? TC_VIOLATION_TYPE_BACK_COMP : TC_VIOLATION_TYPE_MANDATORY,
		 wrappedExpr, self)) {
      return tc_set_expression_type(self, wrappedExpr,
				    SymbTablePkg_error_type());
    }

    /* this is not an error after all -> return boolean type */
    return tc_set_expression_type(self, wrappedExpr,
				  SymbTablePkg_boolean_enum_type());
  }
    
    /* ignore the number..number part and check A-U or E-U expressions */
  case ABU:
  case EBU: {
    SymbType_ptr type
      = type_checker_check_expression(self, car(expression), context);
    return tc_set_expression_type(self, wrappedExpr, type);
  }

  case FAILURE: /* FAILURE is a part of CASE-expression.
		   Should be analysed there.
		   NB: If a more complex use of FAILURE will be adopted,
		   then a new failure-type should be created, and 
		   all expression-checking procedures should take into
		   account this type.*/
    nusmv_assert(false);
  case COLON: /* COLON is a part of BIT_SELECTION or CASE. 
		 Should be analysed there 
	      */
    nusmv_assert(false);

  default: /* there are no other kinds of expressions */
    fprintf(nusmv_stderr, "*** type_checker_check_expression,");
    fprintf(nusmv_stderr, "*** UNABLE TO HANDLE EXPRESSION ");
    print_sexp(nusmv_stderr, expression);
    nusmv_assert(("UNABLE TO HANDLE EXPRESSION" == 0));
  } /* switch */
  

  nusmv_assert(false);
  return nullType;
}

/**AutomaticEnd***************************************************************/

