
/**CHeaderFile*****************************************************************

  FileName    [TypeChecker.h]

  PackageName [compile.type_checking]

  Synopsis    [Public interface of class 'TypeChecker']

  Description [ This class contains the functions performing type checking. 
  The class uses type_checking_violation_handler to deal with 
  type system violations (which may result in an error or warning)
  in the input files.
  After the type checking is performed, use function
  TypeChecker_get_expression_type to get the type of a particular
  expression.
  After the type checking is performed it is possible to obtain the
  type of an expression. Only memory-sharing types (SymbTablePkg_..._type)
  are used, so you can compare pointers instead of the type's contents.
  ]

  SeeAlso     [TypeChecker.c]

  Author      [Andrei Tchaltsev]

  Copyright   [
  This file is part of the ``compile.type_checking'' package of NuSMV 
  version 2. 
  Copyright (C) 2005 by ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

  Revision    [$Id: TypeChecker.h,v 1.1.1.1 2005/12/20 17:06:52 timo Exp $]

******************************************************************************/


#ifndef __TYPE_CHECKER_H__
#define __TYPE_CHECKER_H__


#include "compile/symb_table/SymbType.h" 
#include "compile/symb_table/SymbTable.h" 
#include "utils/utils.h" 


/*---------------------------------------------------------------------------*/
/* Type definitions                                                          */
/*---------------------------------------------------------------------------*/

/**Constant********************************************************************

  Synopsis           [A set of constants, each of which identifies
  particular kinds of possible type system violations.]

  Description        [
  These constants are one of the parameters passed to a violation
  handler when a type system violation is encountered.
  Depending on the kind of the type violation, the violation handler 
  will output an error or warning message or possibly ignore the 
  violation.
  See type_checking_violation_handler for more details.]

  These constants can be combined using bitwise OR operation. ]

  SideEffects        []

  SeeAlso            [type_checking_violation_handler]

******************************************************************************/
enum TYPE_SYSTEM_VIOLATION
  { 
    TC_VIOLATION_UNDEF_IDENTIFIER, /* undefined identifier is met */
    TC_VIOLATION_AMBIGUOUS_IDENTIFIER, /* identifier has two interpretations  */
    TC_VIOLATION_TYPE_MANDATORY,   /* the most general type system
				      violation, and this is mandatory
				      to report
				   */
    TC_VIOLATION_TYPE_BACK_COMP,   /* type system is violated, but it
				      was not an error before the
				      introduction of the type
				      checking package.
				   */
    TC_VIOLATION_OUT_OF_WORD_WIDTH, /* access to the bits outside 
				       of the Word width */
    TC_VIOLATION_INCORRECT_WORD_WIDTH, /* incorrectly formed Word type */

    TC_VIOLATION_OUT_OF_WORDARRAY_WIDTH, /* the address or
					    value word widths are not
					    the specified ones */
    TC_VIOLATION_INCORRECT_WORDARRAY_WIDTH,/* incorrectly formed WordArray type*/

    TC_VIOLATION_DUPLICATE_CONSTANTS, /* duplicate constants in an enum type */
  };


/**Type***********************************************************************

  Synopsis    [Definition of the public type for class TypeChecker]

  Description []

******************************************************************************/
typedef struct TypeChecker_TAG*  TypeChecker_ptr;


/**Type********************************************************************

  Synopsis          [The type of a type violation handler function.]

  Description       [The violation handler function is invoked when 
  an expression being checked violates the type system. 

  The violation function takes as parameters a kind of the
  type system violation (see TYPE_SYSTEM_VIOLATION), the
  expression, where this violation has been detected, and the type checker
  responsible for finding the violation.
  The function returns a boolean value indicating that the given
  kind of violation should be considered as an error, i.e. the type checking
  has to terminate. So, if the violation handler returns false value,
  the type checking will try to continue. 

  In the case of violations related to the incorrectly formed types, 
  the input expression should be CONS with the variable name 
  (the problematic type belongs to) as the left child and 
  the type's body as the left child. The line info of this expression
  should be correct (suitable for error messages).
  
  NB: Not all kinds of type system violation may be ignored.
  NB: TypeChecker is required to obtain the sub-expression types for
  output messages.
  
  The default violation handler is type_checking_violation_handler.]

  See also           [TYPE_SYSTEM_VIOLATION,
  type_checking_violation_handler]

******************************************************************************/
typedef boolean 
(*Type_checking_violation_handler) (enum TYPE_SYSTEM_VIOLATION violation,
				    node_ptr expression,
				    TypeChecker_ptr typeChecker);


/**Macros**********************************************************************

  Synopsis    [To cast and check instances of class TypeChecker]

  Description [These macros must be used respectively to cast and to check
  instances of class TypeChecker]

******************************************************************************/
#define TYPE_CHECKER(self) \
         ((TypeChecker_ptr) self)

#define TYPE_CHECKER_CHECK_INSTANCE(self) \
         (nusmv_assert(TYPE_CHECKER(self) != TYPE_CHECKER(NULL)))


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/
EXTERN TypeChecker_ptr 
TypeChecker_create ARGS((Type_checking_violation_handler handler,
			 SymbTable_ptr symbolTable));

EXTERN void TypeChecker_destroy ARGS((TypeChecker_ptr self));

EXTERN SymbTable_ptr 
TypeChecker_get_symbol_table ARGS((TypeChecker_ptr self));

EXTERN boolean TypeChecker_is_expression_wellformed
                            ARGS((TypeChecker_ptr self,
				  node_ptr expression, node_ptr context));

EXTERN boolean TypeChecker_is_specification_wellformed
                            ARGS((TypeChecker_ptr self, node_ptr expression));

EXTERN boolean TypeChecker_is_type_wellformed 
                            ARGS((TypeChecker_ptr self,
				  SymbType_ptr type, node_ptr varName));

EXTERN SymbType_ptr 
TypeChecker_get_expression_type ARGS((TypeChecker_ptr self,
				      node_ptr expression,
				      node_ptr context));


/**AutomaticEnd***************************************************************/



#endif /* __TYPE_CHECKER_H__ */
