/**CHeaderFile*****************************************************************

  FileName    [type_checking.c]

  PackageName [compile.type_checking]

  Synopsis    [the implementation of the type checking package]

  Description [See .h file for more info]

  Author      [Andrei Tchaltsev]

  Copyright   [
  This file is part of the ``compile.type_checking'' package of NuSMV version 2.
  Copyright (C) 2005 by  ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "type_checking.h"
#include "type_checkingInt.h"
#include "TypeChecker.h"

#include "parser/symbols.h"
#include "utils/ustring.h" 
#include "utils/assoc.h" 

static char rcsid[] UTIL_UNUSED = "$Id: type_checking.c,v 1.1.1.1 2005/12/20 17:06:52 timo Exp $";
/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
  
/**Variable********************************************************************

  Synopsis           [The global type checker]

  Description [An instance of a type checker is created and destroyed
  in the init and end functions of the package. During its lifetime a
  type checker can deal with only one symbol table. Moreover, type
  checker remember the expressions being already checked. As result an
  instance of a type checker can deal with only symbol table and one
  input SMV model only.]

******************************************************************************/
static TypeChecker_ptr typeChecker = TYPE_CHECKER(NULL);


/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static boolean tc_check_constrain_list ARGS((TypeChecker_ptr checker,
					     int kind,
					     node_ptr expressions));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Initialises the type checking package.]

  Description        [Can be invoked only after the symbol table has been
  created.]

  SideEffects        []

******************************************************************************/
void TypeCheckingPkg_init(SymbTable_ptr symbolTable)
{
  nusmv_assert(TYPE_CHECKER(NULL) == typeChecker); /* re-initialisation */
  typeChecker = TypeChecker_create(type_checking_violation_handler,
				   symbolTable);
}


/**Function********************************************************************

  Synopsis           [Shuts down the type checking package.]

  Description        [WARNING: All symbolic types SymbTypes
  generated during type checking will be destroyed after this package
  de-initialisation.]

  SideEffects        []

******************************************************************************/
void TypeCheckingPkg_quit()
{
  if (TYPE_CHECKER(NULL) != typeChecker) { /* package is initialised */
    TypeChecker_destroy(typeChecker);
    typeChecker = TYPE_CHECKER(NULL);
  }
}


/**Function********************************************************************

  Synopsis           [Returns the global type checker, used by 
  the this package.]

  Description        [WARNING: This function can be invoked 
  only after initialisation and before de-initialisation.]

  SideEffects        []

******************************************************************************/
TypeChecker_ptr TypeCheckingPkg_get_global_type_checker()
{
  return typeChecker;
}


/**Function********************************************************************

  Synopsis           [Checks that the types of variable decalarations
  in the layer are correctly formed. Also defines are checked to have 
  some well-formed type.]

  Description        [Constrain: the input layer should belong to the symbol
  table the type checker is associated with.

  The function iterates over all variables in the layer, and checks their type
  with function TypeChecker_is_type_wellformed.

  The function also type checks the expressions provided in
  defines (also the generated "running" defines) have some type.
  NB for developers: This is done to allow the type checker
  to remember the type of these defines (and associated constants and variable
  _process_selector_). Without this list the evaluation phase will not
  know the type of these defines (if there were not explicitly used in the 
  input text), since they can be implicitly used in ASSIGN(TRANS) contrains.

  Returns true if all the type are correctly formed and the defines
  are correct, and false otherwise. ]

  SideEffects        []

  SeeAlso            [TypeChecker_is_type_wellformed]

******************************************************************************/
boolean TypeCheckingPkg_check_layer(TypeChecker_ptr checker,
				    SymbLayer_ptr layer)
{

  boolean isOK = true;
  SymbTable_ptr table = TypeChecker_get_symbol_table(checker);
  NodeList_ptr vars = SymbLayer_get_all_vars(layer);
  ListIter_ptr iter = NodeList_get_first_iter(vars);

  /* the type checker's symbol table should contain the given layer */
  nusmv_assert(SymbTable_get_layer(table, SymbLayer_get_name(layer)) == layer);
  
  while (!ListIter_is_end(iter)) {
    node_ptr varName = NodeList_get_elem_at(vars, iter);
    SymbType_ptr type = SymbTable_get_var_type(table, varName);
    isOK = TypeChecker_is_type_wellformed(checker, type,  varName) && isOK;
    iter = ListIter_get_next(iter);
  }
  if (!isOK) return false;

  {
     NodeList_ptr defs = SymbLayer_get_defines(layer);
     ListIter_ptr iter;
     for(iter = NodeList_get_first_iter(defs); !ListIter_is_end(iter);
	 iter = ListIter_get_next(iter)) {
       node_ptr defName = NodeList_get_elem_at(defs, iter);
       /* wrap in DEFINE because it is required by the function invoked */
       isOK = TypeChecker_is_specification_wellformed(typeChecker,
						  find_node(DEFINE,defName,Nil))
	 && isOK;
     }
     
  }
  return isOK;
}


/**Function********************************************************************

  Synopsis           [checks all the module contrains are correctly typed]

  Description        [
  The module contrains are declarations INIT, INVAR, TRANS, ASSIGN,
  JUSTICE, COMPASSION.
 
  The first parameter 'checker' is a type checker to perfrom checking.
  All the remaining parameters are the sets of expressions
  constituting the bodies of the corresponding high-level
  declarations. These expressions are created during compilation and
  then passed to this function unmodified, flattened or
  flattened+expanded.  So this function is relatively specialised to
  deal with concrete data-structures created during compilation. For
  example, the expressions in the given sets are expected to be
  separated by CONS and AND.
 
  NOTE: if an expression has been flattened, then
  info about line numbers mat not be accurate.

  The type checker remebers all the checked expressions and their types,
  thus TypeChecker_get_expression_type can return the type of already checked
  expressions.
  
  The parameter 'assign' is actually the 'procs' returned 
  by Compile_FlattenHierarchy, which contains all the assignments.

  If some of expressions violates the type system, the type checker's
  violation handler is invoked. See Type_checking_violation_handler
  for more info.

  Returns false if the module contrains violate the type system, and
  otherwise true is returned.
  ]

  SideEffects        []

******************************************************************************/
boolean TypeCheckingPkg_check_constrains(TypeChecker_ptr checker,
					 node_ptr init, node_ptr trans,
					 node_ptr invar, node_ptr assign,
					 node_ptr justice, node_ptr compassion)
{
  boolean isOK = true;
  
  /* check INIT */
  if (!tc_check_constrain_list(checker, INIT, init)) isOK = false;
  
  /* check TRANS */
  if (!tc_check_constrain_list(checker, TRANS, trans)) isOK = false;
     
  /* check INVAR */
  if (!tc_check_constrain_list(checker, INVAR, invar)) isOK = false;

  /* check ASSIGN */
  if(!tc_check_constrain_list(checker, ASSIGN, assign)) isOK = false;

  /* check  JUSTICE */
  if(!tc_check_constrain_list(checker, JUSTICE, justice)) isOK = false;

  /* check  COMPASSION */
  if (!tc_check_constrain_list(checker, COMPASSION, compassion)) {
    isOK = false;
  }
  
  if (opt_verbose_level_gt(options, 2)) {
    if (isOK) fprintf(nusmv_stdout, 
		      "Successful type-checking of the module constrains\n");
  }
    
  return isOK;
}


/**Function********************************************************************

  Synopsis           [checks that the expression constituting the 
  property is correctly typed]

  Description        [
  If some of expressions violates the type system, the type checker's
  violation handler is invoked. See Type_checking_violation_handler
  for more info.

  The type checker remebers all the checked expressions and their types,
  thus TypeChecker_get_expression_type can return the type of already checked
  expressions.

  If the property violates the type system, the false value is return,
  and true value otherwise.]

  SideEffects        []

******************************************************************************/
boolean TypeCheckingPkg_check_property(TypeChecker_ptr checker,
				       Prop_ptr property)
{
  int kind;
  boolean isOK;
  node_ptr exp;

  switch (Prop_get_type(property)) {
  case Prop_NoType:  nusmv_assert(false);/* incorrect property */
  case Prop_Ctl:     kind = SPEC;      break;
  case Prop_Ltl:     kind = LTLSPEC;   break;
  case Prop_Psl:     kind = PSLSPEC;   break;
  case Prop_Invar:   kind = INVARSPEC; break;
  case Prop_Compute: kind = COMPUTE;   break;
  default:           nusmv_assert(false);
  } /* switch */

  yylineno = node_get_lineno(Prop_get_expr(property));
  exp = find_node(kind, Prop_get_expr(property), Nil);

  isOK = TypeChecker_is_specification_wellformed(typeChecker, exp);

  if (opt_verbose_level_gt(options, 2)) {
    if (isOK) {
      /* the property is not yet inserted to database => there is no index */
      fprintf(nusmv_stdout, "Successful type-checking of a property\n");
    }
  }

  return isOK;
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Static Function************************************************************

  Synopsis           [Checks a list of expressions for being well-types.
  Expressions can be unmodified (after compilation), flattened or 
  flattened+expanded.]

  Description        [The first parameter is the type checker performing 
  the type checking. The second one is the kind of SMV constrain
  containing the input expressions (like INVAR, TRANS, etc). The third
  one is a set of expressions to be checked.
  In the set the expression are separated by AND or CONS. The set also
  can be Nil. This function is specialised to deal with the output of
  the compilation phase.

  NB for developers: the list of constrain expressions collected
  during flattening, i.e. "expressions" contains the actual
  expressions only. If the expression consist of one identifier only,
  then the indentifier is normalized (with resolve-name and find-atom)
  and info about the line number is lost. Then only the wrapping AND or CONS
  node will contain the correct line.

  The return value is true if there is no type violations, and false otherwise]

  SideEffects        []

******************************************************************************/
static boolean tc_check_constrain_list(TypeChecker_ptr checker,
				       int kind,
				       node_ptr expressions)
{
  node_ptr exp;
  boolean isOK = true;

  if (Nil == expressions) return true;

  nusmv_assert(AND == node_get_type(expressions) ||
	       CONS == node_get_type(expressions));

   
  /* process the left part of the list */
  exp = car(expressions);
  if (Nil != exp && AND != node_get_type(exp) && CONS != node_get_type(exp)) {
    /* this is an actual expression. */
    yylineno = node_get_lineno(expressions);
    isOK = TypeChecker_is_specification_wellformed(typeChecker,
						   find_node(kind, exp, Nil))
      && isOK;
  }
  else {
    /* this is a subset of expressions */
    isOK = tc_check_constrain_list(checker, kind, exp) && isOK;
  }

  /* process the right part of the list */
  exp = cdr(expressions);
  if (Nil != exp && AND != node_get_type(exp) && CONS != node_get_type(exp)) {
    /* this is an actual expression  */
    yylineno = node_get_lineno(expressions);
    isOK = TypeChecker_is_specification_wellformed(typeChecker,
						   find_node(kind, exp, Nil))
      && isOK;
  }
  else {
    /* this is a subset of expressions */
    isOK = tc_check_constrain_list(checker, kind, exp) && isOK;
  }
  
  return isOK;
}


