/**CHeaderFile*****************************************************************

  FileName    [type_checking_violation_handler.c]

  PackageName [compile.type_checking]

  Synopsis    [the implementation of the type system violation handler]

  Description [A type system violation happens when a mistyped 
  expression is encountered in an input SMV module or specification.
  In this case a violation handler is invoked. Depending on the kind of
  of the violation, the violation handler may output an error message, 
  a waring message or ignore the problem.
  See the type_checking_default_violation_handler for more description.]

  Author      [Andrei Tchaltsev]

  Copyright   [
  This file is part of the ``compile'' package of NuSMV version 2. 
  Copyright (C) 2000-2005 by  ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "type_checkingInt.h"
#include "parser/symbols.h"
#include "utils/ustring.h"
#include "enc/utils/WordNumber.h"

static char rcsid[] UTIL_UNUSED = "$Id: type_checking_violation_handler.c,v 1.1.1.1 2005/12/20 17:06:52 timo Exp $";
/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static void print_operator ARGS((FILE* output_stream, node_ptr expr));
static void print_type ARGS((FILE* output_stream, node_ptr expression,
			     TypeChecker_ptr checker, node_ptr context));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [The type system violation handler.]

  Description        [The violation handler is implemented as 
  a function, which is invoked by the type checker when an expression
  being checked violates the type system.
  See the violation handler Type_checking_violation_handler
  for more explanations.

  The below function is the default violation handler, and a
  user can potentially define its own violation handler.
  
  This violation handler outputs an error and warning message to
  nusmv_stderr. The warning is output if the detected violation is
  TC_VIOLATION_TYPE_BACK_COMP or TC_VIOLATION_DUPLICATE_CONSTANTS and
  the system variable type_checking_backward_compatibility" is
  true. In this case only the false value is returned, indicating that
  this is NOT an error. Otherwise the true value is returned,
  indicating that this is an error.

  Also, if the system variable "type_check_warning_on" is false,
  warning messages are not output.

  NB: if the expression is checked in some context (context is not null) then
  before providing the expression to this function the expression should be
  wrapped into context, i.e. with find_node(CONEXT, context, expr)]

  SideEffects       []

  SeeAlso           [TYPE_SYSTEM_VIOLATION]

******************************************************************************/
EXTERN boolean
type_checking_violation_handler(enum TYPE_SYSTEM_VIOLATION violation,
				node_ptr expression,
				TypeChecker_ptr typeChecker)
{
  /* In the output message, the information about the expression
     location are output. So, make sure that the input file name and
     line number are correctly set!
  */

  boolean isError = true; /* is this error or warning */
  /* get rid of the context the expression may be wrapped in */
  node_ptr context = CONTEXT == node_get_type(expression)  
    ? car(expression) : Nil;
  node_ptr expr = CONTEXT == node_get_type(expression) 
                  ? cdr(expression) : expression;

  /* there are no other kinds of violations */
  nusmv_assert( TC_VIOLATION_UNDEF_IDENTIFIER == violation ||
		TC_VIOLATION_AMBIGUOUS_IDENTIFIER == violation ||
		TC_VIOLATION_TYPE_MANDATORY ==  violation ||
		TC_VIOLATION_TYPE_BACK_COMP == violation ||
		TC_VIOLATION_OUT_OF_WORD_WIDTH ==  violation ||
		TC_VIOLATION_OUT_OF_WORDARRAY_WIDTH ==  violation ||
		TC_VIOLATION_INCORRECT_WORD_WIDTH == violation ||
		TC_VIOLATION_INCORRECT_WORDARRAY_WIDTH == violation ||
		TC_VIOLATION_DUPLICATE_CONSTANTS == violation);
  

  /* only violation TC_VIOLATION_TYPE_BACK_COMP or 
     TC_VIOLATION_DUPLICATE_CONSTANTS and the variable
     type_checking_backward_compatibility being true, may
     make a warning from an error.
  */
  if ( (TC_VIOLATION_TYPE_BACK_COMP == violation ||
	TC_VIOLATION_DUPLICATE_CONSTANTS == violation) &&
      opt_type_checking_back_comp(options)) {
    isError = false;
  }

  if (!isError && !opt_type_checking_warning_on(options)) {
    /* this is a warning and warning messages are not allowed.
     So, do nothing, just return false (this is not an error)
    */
    return false;
  }

  if (isError) fprintf(nusmv_stderr, "\nTYPE ERROR (");
  else         fprintf(nusmv_stderr, "\nTYPE WARNING (");

  if (get_input_file(options)) {
    fprintf(nusmv_stderr, "file %s", get_input_file(options));
  }
  else fprintf(nusmv_stderr, "file stdin");

  if (node_get_lineno(expr)) fprintf(nusmv_stderr, ", line %d",
				     node_get_lineno(expr));
  fprintf(nusmv_stderr,  ") : ");
  
  switch (violation) {
  case TC_VIOLATION_UNDEF_IDENTIFIER:
    fprintf(nusmv_stderr,  "undefined identifier '");
    print_node(nusmv_stderr, expr);
    fprintf(nusmv_stderr, "'\n");
    break;
  case TC_VIOLATION_AMBIGUOUS_IDENTIFIER:
    fprintf(nusmv_stderr,  "identifier '");
    print_node(nusmv_stderr, expr);
    fprintf(nusmv_stderr, "' is ambiguous\n");
    break;
  case TC_VIOLATION_OUT_OF_WORD_WIDTH:
    /* only bit-selection may have this kind of error */
    nusmv_assert(BIT_SELECTION == node_get_type(expr));
    fprintf(nusmv_stderr,  "bit number out of the Word's width ('");
    print_operator(nusmv_stderr, expr);
    fprintf(nusmv_stderr,  "' operator)\n");
    break;

  case TC_VIOLATION_INCORRECT_WORD_WIDTH:
    nusmv_assert(CONS == node_get_type(expr));
    fprintf(nusmv_stderr, "in the declaration of '");
    print_node(nusmv_stderr, car(expr));
    fprintf(nusmv_stderr,
	    "' the Word width is not a positive number (from range [1..%d])\n",
	    WordNumber_max_width());
    break;

  case TC_VIOLATION_OUT_OF_WORDARRAY_WIDTH:
    nusmv_assert(WAWRITE == node_get_type(expr) ||
		 WAREAD == node_get_type(expr));
    if (WAWRITE == node_get_type(expr)) {
      fprintf(nusmv_stderr,
	      "in WRITE expression the width of address or value operands"
	      " is not consistent with word-array operand\n");
    }
    else if (WAREAD == node_get_type(expr)) {
      fprintf(nusmv_stderr,
	      "in READ operator the width of address operand"
	      " is not consistent with word-array operand\n");
    }
    else { 
      /* only WRITE or READ (of word-array) may have this kind of error */
      nusmv_assert(false);
    }
    break;

  case TC_VIOLATION_INCORRECT_WORDARRAY_WIDTH:
    nusmv_assert(CONS == node_get_type(expr));
    fprintf(nusmv_stderr, "in the declaration of '");
    print_node(nusmv_stderr, car(expr));
    fprintf(nusmv_stderr,
	    "' either the address or the value width are not in range [1..%d]\n",
	    WordNumber_max_width());
    break;

  case TC_VIOLATION_DUPLICATE_CONSTANTS:
    nusmv_assert(CONS == node_get_type(expr));
    fprintf(nusmv_stderr, "duplicate constants in the enum type of variable '");
    print_node(nusmv_stderr, car(expr));
    fprintf(nusmv_stderr, "'\n");
   break;

  case TC_VIOLATION_TYPE_MANDATORY:
  case TC_VIOLATION_TYPE_BACK_COMP:
    if (isError) fprintf(nusmv_stderr, "illegal ");
    else         fprintf(nusmv_stderr, "potentially incorrect ");

    switch (node_get_type(expr)) {
    case ASSIGN: /* EQDEF is checked instead of ASSIGN */
    case COMPUTE:
    case FAILURE: /* FAILURE does not participate in the type-checking */
    case ATOM: /* undeclared identifies checked earlier */
    case NUMBER:
    case NUMBER_WORD:
    case NUMBER_REAL:
    case NUMBER_FRAC:
    case TRUEEXP:
    case FALSEEXP:
    case BIT:   /* undeclared identifies checked earlier */
    case BOOLEAN:
    case SELF:
    case INTEGER:
    case REAL:
    case WORD:
    case SCALAR:
    case ARRAY:
    case DOT:
    case TWODOTS: 
    case NEXT:
    case SMALLINIT:
    /* case PROCESS: */ /*used as artificial constract. See below */ 
    case ISA:
    case DEFINE:
    case CONTEXT:
    case BDD:
    case SEMI:
    case IFTHENELSE:
      print_sexp(nusmv_stderr, expr);
      nusmv_assert(false); /* this is impossible */
      
     
      /* Binary operators */
    case CONCATENATION:
    case OR:
    case XOR:
    case XNOR:
    case IMPLIES:
    case IFF:
    case EQDEF:
    case AND:
    case RELEASES:
    case TRIGGERED:
    case UNTIL:
    case SINCE:
    case AU:
    case EU:
    case LSHIFT:
    case RSHIFT:
    case LROTATE:
    case RROTATE:
    case PLUS:
    case MINUS:
    case TIMES:
    case DIVIDE:
    case MOD:
    case EQUAL:
    case NOTEQUAL:
    case LT:
    case GT:
    case LE:
    case GE:
    case UNION:
    case SETIN:
    case MINU:
    case MAXU:
    case WAREAD:
      fprintf(nusmv_stderr, "operand types of \"");
      print_operator(nusmv_stderr, expr);
      fprintf(nusmv_stderr,"\" : ");
      print_type(nusmv_stderr, car(expr), typeChecker, context);
      fprintf(nusmv_stderr," and ");
      print_type(nusmv_stderr, cdr(expr), typeChecker, context);
      /*fprintf(nusmv_stderr,"");*/
      break;

      /* Unary operators */
    case CAST_BOOL:
    case CAST_WORD1:
    case NOT:
    case EX:
    case AX:
    case EF:
    case AF:
    case EG:
    case AG:
    case OP_GLOBAL:
    case OP_HISTORICAL:
    case OP_NEXT:
    case OP_PREC:
    case OP_NOTPRECNOT:
    case OP_FUTURE:
    case OP_ONCE:
      fprintf(nusmv_stderr,"operand type of \"");
      print_operator(nusmv_stderr, expr);
      fprintf(nusmv_stderr,"\" : ");
      print_type(nusmv_stderr, car(expr), typeChecker, context);
      break;

    /* high level unary operator */
    case INVAR:
    case TRANS:
    case INIT:
    case JUSTICE:
    case FAIRNESS:
    case COMPASSION:
    case SPEC:
    case LTLSPEC:
    case INVARSPEC:
    case PROCESS:/* PROCESS is artificial contract to check 'running' symbols */
      fprintf(nusmv_stderr,"type of ");
      print_operator(nusmv_stderr, expr);
      fprintf(nusmv_stderr," expression : ");
      print_type(nusmv_stderr, car(expr), typeChecker, context);
      break;

   case COLON: /* for CASE */
     fprintf(nusmv_stderr,"left operand type of \":\". "
	     "It should be boolean, but is ");
     print_type(nusmv_stderr, car(expr), typeChecker, context);
     break;
 
    case BIT_SELECTION:
      fprintf(nusmv_stderr,"operand types of \"[ : ]\" : ");
      print_type(nusmv_stderr, car(expr), typeChecker, context);
      fprintf(nusmv_stderr,", ");
      print_type(nusmv_stderr, car(cdr(expr)), typeChecker, context);
      fprintf(nusmv_stderr," and ");
      print_type(nusmv_stderr, cdr(cdr(expr)), typeChecker, context);
      break;

    case WAWRITE:
      fprintf(nusmv_stderr,"operand types of \"WRITE\" : ");
      print_type(nusmv_stderr, car(expr), typeChecker, context);
      fprintf(nusmv_stderr,", ");
      print_type(nusmv_stderr, car(cdr(expr)), typeChecker, context);
      fprintf(nusmv_stderr," and ");
      print_type(nusmv_stderr, cdr(cdr(expr)), typeChecker, context);
      break;

    case CASE:
      fprintf(nusmv_stderr,"types of \"case\" list elements : ");
      print_type(nusmv_stderr, cdr(car(expr)), typeChecker, context);
      fprintf(nusmv_stderr," and ");
      print_type(nusmv_stderr, cdr(expr), typeChecker, context);
      break;

    case EBF:
    case ABF:
    case EBG:
    case ABG:
      fprintf(nusmv_stderr,"operand type of \"");
      print_operator(nusmv_stderr, expr);
      fprintf(nusmv_stderr,"\" : ");
      print_type(nusmv_stderr, car(expr), typeChecker, context);
      break;
 
    case ABU:
    case EBU:
      fprintf(nusmv_stderr,"operand types of \"");
      print_operator(nusmv_stderr, expr);
      fprintf(nusmv_stderr,"\" : ");
      print_type(nusmv_stderr, car(car(expr)), typeChecker, context);
      fprintf(nusmv_stderr," and ");
      print_type(nusmv_stderr, cdr(car(expr)), typeChecker, context);
      break;


    default: /* unknown kind of an expression */
      nusmv_assert(false);
    } /* switch (node_get_type(expr)) */
    fprintf(nusmv_stderr,"\n");
    break;
    
  default: 
    nusmv_assert(false); /* unknown kind of an error */
  } /* switch (errorKind) */
  
  return isError;
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Static function*************************************************************

  Synopsis           [Just prints an expression's operator to output_stream]

  Description        [This function is the almost the same as 
  print_sexp, except this function does not print the children of the node.
  The expr must be a correct expression. 
  The function is used in printing of an error messages only.]

  SideEffects        []

  SeeAlso            [print_sexp]
******************************************************************************/
static void print_operator(FILE* output_stream, node_ptr expr)
{
  nusmv_assert((node_ptr) Nil != expr);
  switch(node_get_type(expr)){

  case CONCATENATION: fprintf(output_stream,"::"); return;
  case BIT_SELECTION: fprintf(output_stream,"[ : ]"); return;
  case CAST_BOOL:     fprintf(output_stream,"bool"); return;
  case CAST_WORD1:    fprintf(output_stream,"word1"); return;
  case AND:           fprintf(output_stream,"&"); return;
  case OR:            fprintf(output_stream,"|"); return;
  case XOR:           fprintf(output_stream,"xor"); return;
  case XNOR:          fprintf(output_stream,"xnor"); return;
  case IMPLIES:       fprintf(output_stream,"->"); return;
  case IFF:   fprintf(output_stream,"<->"); return;
  case NOT:   fprintf(output_stream,"!"); return;
  case EX:    fprintf(output_stream,"EX"); return;
  case AX:    fprintf(output_stream,"AX"); return;
  case EF:    fprintf(output_stream,"EF"); return;
  case AF:    fprintf(output_stream,"AF"); return;
  case EG:    fprintf(output_stream,"EG"); return;
  case AG:    fprintf(output_stream,"AG"); return;
  case OP_GLOBAL:     fprintf(output_stream,"G"); return;
  case OP_HISTORICAL: fprintf(output_stream,"H"); return;
  case RELEASES:      fprintf(output_stream,"V"); return;
  case TRIGGERED:     fprintf(output_stream,"T"); return;
  case UNTIL:    fprintf(output_stream,"U"); return;
  case SINCE:    fprintf(output_stream,"S"); return;
  case OP_NEXT:  fprintf(output_stream,"X"); return;
  case OP_PREC:  fprintf(output_stream,"Y"); return;
  case OP_NOTPRECNOT:    fprintf(output_stream,"Z"); return;
  case OP_FUTURE:        fprintf(output_stream,"F"); return;
  case OP_ONCE:          fprintf(output_stream,"O"); return;
  case AU:   fprintf(output_stream,"A-U"); return;
  case EU:   fprintf(output_stream,"E-U"); return;
  case EBF:  fprintf(output_stream,"EBF"); return;
  case ABF:  fprintf(output_stream,"ABF"); return;
  case EBG:  fprintf(output_stream,"EBG"); return;
  case ABG:  fprintf(output_stream,"ABG"); return;
  case ABU:  fprintf(output_stream,"ABU"); return;
  case EBU:  fprintf(output_stream,"EBU"); return;
  case LSHIFT:    fprintf(output_stream,"<<"); return;
  case RSHIFT:    fprintf(output_stream,">>"); return;
  case LROTATE:   fprintf(output_stream,"<<<"); return;
  case RROTATE:   fprintf(output_stream,">>>"); return;
  case PLUS:      fprintf(output_stream,"+"); return;
  case MINUS:     fprintf(output_stream,"-"); return;
  case TIMES:     fprintf(output_stream,"*"); return;
  case DIVIDE:    fprintf(output_stream,"/"); return;
  case MOD:       fprintf(output_stream,"mod"); return;
  case EQUAL:     fprintf(output_stream,"="); return;
  case NOTEQUAL:  fprintf(output_stream,"!="); return;
  case LT:    fprintf(output_stream,"<"); return;
  case GT:    fprintf(output_stream,">"); return;
  case LE:    fprintf(output_stream,"<="); return;
  case GE:    fprintf(output_stream,">="); return;
  case UNION: fprintf(output_stream,"union"); return;
  case SETIN: fprintf(output_stream,"in"); return;
  case COLON: fprintf(output_stream,":");return;
  case WAREAD: fprintf(output_stream,"READ");return;
  case WAWRITE: fprintf(output_stream,"WRITE");return;

  case INVAR:     fprintf(output_stream,"INVAR");return;
  case TRANS:     fprintf(output_stream,"TRANS");return;
  case INIT:      fprintf(output_stream,"INIT");return;
  case FAIRNESS:  fprintf(output_stream,"FAIRNESS");return;
  case JUSTICE:      fprintf(output_stream,"JUSTICE");return;
  case COMPASSION:      fprintf(output_stream,"COMPASSION");return;

    /* PROCESS is artificial contract to check 'running' symbols */
  case PROCESS: fprintf(output_stream,"'running'");return;

  case ASSIGN:    fprintf(output_stream,"ASSIGN");return;
  case EQDEF:     fprintf(output_stream,":=");return;
  case MINU:    fprintf(output_stream,"MIN");return;
  case MAXU:    fprintf(output_stream,"MAX");return;
  case SPEC:      fprintf(output_stream,"SPEC");return;
  case LTLSPEC:   fprintf(output_stream,"LTLSPEC");return;
  case INVARSPEC: fprintf(output_stream,"INVARSPEC");return;

    /* Below is the list of nodes that should not be encountered */
  case COMPUTE:   fprintf(output_stream,"COMPUTE");break;

  case MODULE:    fprintf(output_stream,"\n(MODULE ");break;
  case MODTYPE:   fprintf(output_stream,"\n(MODTYPE ");break;
  case CASE:      fprintf(output_stream,"\n(CASE ");break;
  case CONS:      fprintf(output_stream,"\n(CONS ");break;
  case VAR:       fprintf(output_stream,"\n(VAR ");break;
  case FAILURE:      fprintf(output_stream,"\n(FAILURE)");break;
  case ATOM:      fprintf(output_stream,"\n(ATOM ");break;
  case NUMBER:    fprintf(output_stream,"\n(NUMBER %d)",(int)car(expr));break;
  case NUMBER_WORD: fprintf(output_stream,"\n(NUMBER_WORD ");
                    WordNumber_print(output_stream, WORD_NUMBER(car(expr)));
                    fprintf(output_stream,")"); break;
  case NUMBER_REAL: fprintf(output_stream,"\n(NUMBER_REAL %s)",
			    get_text((string_ptr)car(expr)));break;
  case NUMBER_FRAC: fprintf(output_stream,"\n(NUMBER_FRAC %s)",
			    get_text((string_ptr)car(expr)));break;
  case TRUEEXP:     fprintf(output_stream,"(TRUE)");break;
  case FALSEEXP:    fprintf(output_stream,"(FALSE)");break;
  case BIT:         fprintf(output_stream,"\n(BIT ");break;
  case BOOLEAN:     fprintf(output_stream,"(BOOLEAN)");break;
  case INTEGER:     fprintf(output_stream,"(INT)");break;
  case REAL:    fprintf(output_stream,"(REAL)");break;
  case WORD:    fprintf(output_stream,"(WORD ");break;
  case SELF:    fprintf(output_stream,"(SELF)");break;
  case SCALAR:  fprintf(output_stream,"(SCALAR ");break;
  case ARRAY:   fprintf(output_stream,"\n(ARRAY ");break;
  case DOT:     fprintf(output_stream,"(DOT ");break;
  case TWODOTS:    fprintf(output_stream, "..");break;
  case NEXT:    fprintf(output_stream,"NEXT");break;
  case SMALLINIT:    fprintf(output_stream,"\n(SMALLINIT ");break;
  /*case PROCESS:    fprintf(output_stream,"\n(PROCESS ");break;*/
  case ISA:    fprintf(output_stream,"\n(ISA ");break;
  case DEFINE:    fprintf(output_stream,"\n(DEFINE ");break;
  case CONTEXT:    fprintf(output_stream,"\n(CONTEXT ");break;
  case BDD:   fprintf(output_stream,"(BDD TO BE PRINTED)");break;
  case SEMI:    fprintf(output_stream, "\n(SEMI ");break;
  case IFTHENELSE:    fprintf(output_stream, "\n (IFTHENELSE ");break;
    
    
  default:
    fprintf(nusmv_stderr, "\n%d\n", node_get_type(expr));
    nusmv_assert(false);
  }
  nusmv_assert(false);
}


/**Static function*************************************************************

  Synopsis           [Just prints out the type of an expression to
  output_stream]

  Description        [
  The expr must be a correct expression which has been already type checked 
  by 'checker'.
  The last parameter 'context' is the context where the expression
  has been found.
  The function is used in printing of an error messages only.]

  SideEffects        []

  SeeAlso            []
******************************************************************************/
static void print_type(FILE* output_stream, node_ptr expression,
		       TypeChecker_ptr checker, node_ptr context)
{
  SymbType_ptr type = TypeChecker_get_expression_type(checker,
						      expression, context);
  /* this expression must have been checked by this type checker.
     and it must not be 0 since this is an sub-expression
     of a problematic expression
  */
  nusmv_assert(type);
  SymbType_print(output_stream, type);
  return;
}
