/**CSourceFile*****************************************************************

  FileName    [AddArray.c]

  PackageName [utils]

  Synopsis    []

  Description []

  SeeAlso     [AddArray.h]

  Author      [Andrei Tchaltsev]

  Copyright   [
  This file is part of the ``enc.utils'' package of NuSMV version 2. 
  Copyright (C) 2005 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "AddArray.h"
#include "dd/dd.h" /* required by functions of word arithmetic operations */
#include "enc/operators.h" /* for word arithmetic operation */
#include "parser/symbols.h"
#include "utils/ustring.h"
#include "utils/error.h"

static char rcsid[] UTIL_UNUSED = "$Id: AddArray.c,v 1.1.1.1 2005/12/20 17:06:57 timo Exp $";

/*---------------------------------------------------------------------------*/
/* Types definition                                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro definition                                                          */
/*---------------------------------------------------------------------------*/

/**Macro  ********************************************************************

  Synopsis           [converts a type from (actually used) array_t* to our
  artificial AddArray_ptr]

  Description        [The type AddArray_ptr is used just to hide
  array_t (to enable type-checking by compilers).
  But the actuall data-structure used is array_t.]

******************************************************************************/
#define array2AddArray(array) ADD_ARRAY(array)


/**Macro  ********************************************************************

  Synopsis           [converts a type from our
  artificial AddArray_ptr to (actually used) array_t*]

  Description        [See array2AddArray]

******************************************************************************/
#define AddArray2array(addArray) ((array_t*)(addArray))


/*---------------------------------------------------------------------------*/
/* Variable definition                                                       */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static void add_array_full_adder ARGS((DdManager* dd,
				       add_ptr arg1,
				       add_ptr arg2,
				       add_ptr carry_in,
				       add_ptr* sum,
				       add_ptr* carry_out));

static void add_array_adder ARGS((DdManager* dd,
				  AddArray_ptr arg1,
				  AddArray_ptr arg2,
				  add_ptr carry_in,
				  AddArray_ptr* res,
				  add_ptr* carry_out));

static AddArray_ptr add_array_apply_on_bits ARGS((DdManager* dd,
						  NPFNN/*NPFCVT*/ op,
						  AddArray_ptr arg));

static void add_array_division_remainder ARGS((DdManager* dd,
					       AddArray_ptr arg1,
					       AddArray_ptr arg2,
					       AddArray_ptr* quotient,
					       AddArray_ptr* remainder));

static add_ptr
add_array_create_default_value_of_shift_operation ARGS((DdManager* dd,
							add_ptr number,
							int width,
							char* errMessage));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [constructor. Create an array of "number" ADDs]

  Description        [number must be positive. The index of the 
  array goes from 0 to (number - 1).]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_create(int number)
{
  nusmv_assert(number > 0);
  return array2AddArray(array_alloc(add_ptr, number));
}


/**Function********************************************************************

  Synopsis           [given an ADD create an AddArray consisting of 
  one element]

  Description        [Given ADD must already be referenced.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_from_add(add_ptr add)
{ 
  array_t* array = array_alloc(add_ptr, 1);
  array_insert(add_ptr, array, 0, add);
  return array2AddArray(array);
}


/**Function********************************************************************

  Synopsis           [create a new AddArray, a copy of a given one]

  Description        [During duplication all ADD will be referenced.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_duplicate(AddArray_ptr self)
{
  array_t* new;
  array_t* old;
  int i;
  
  old = AddArray2array(self);
  new = array_alloc(add_ptr, array_n(old));
  for (i = 0; i < array_n(old); ++i) {
    add_ptr add = array_fetch(add_ptr, old, i);
    add_ref(add);
    array_insert(add_ptr, new, i, add);
  }
  return array2AddArray(new);
}


/**Function********************************************************************

  Synopsis           [destructor of the class]

  Description        [The memory will be freed and all ADD will be
  de-referenced]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void AddArray_destroy(DdManager* dd, AddArray_ptr self)
{
  int i;
  array_t* array = AddArray2array(self);
  for (i = 0; i < array_n(array); ++i) {
    add_free(dd, array_fetch(add_ptr, array, i));
  }
  array_free(array);
}


/**Function********************************************************************

  Synopsis           [returns the size (number of elements) of the array]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
int AddArray_get_size(AddArray_ptr self)
{
  return array_n(AddArray2array(self));
}


/**Function********************************************************************

  Synopsis           [This function returns the first element of 
  the array]

  Description        [The array should contain exactly one element]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
add_ptr AddArray_get_add(AddArray_ptr self)
{
  array_t* array = AddArray2array(self);
  nusmv_assert(array_n(array) == 1);
  return array_fetch(add_ptr, array, 0);
}


/**Function********************************************************************

  Synopsis           [Returns the element number "n" from
  the array.]

  Description        ["n" can be from 0 to (size-1).
  The returned ADD is NOT referenced.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
add_ptr AddArray_get_n(AddArray_ptr self, int number)
{
  array_t* array = AddArray2array(self);
  return array_fetch(add_ptr, array, number);
}


/**Function********************************************************************

  Synopsis           [Sets the element number "number" to "add".]

  Description        [The given ADD "add" must already be referenced.
  The previous value should already be de-referenced if it is necessary.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
void AddArray_set_n(AddArray_ptr self, int number, add_ptr add)
{
  array_t* array = AddArray2array(self);
  array_insert(add_ptr, array, number, add);
}


/**Function********************************************************************

  Synopsis           [Perform the addition operations
  on two Word expressions represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The size of both arguments should be the same.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_plus(DdManager* dd, AddArray_ptr arg1,
					       AddArray_ptr arg2)
{
  AddArray_ptr res;
  add_ptr carry;
  add_ptr zero = add_zero(dd);

  add_array_adder(dd, arg1, arg2, zero, &res, &carry);

  add_free(dd, zero);
  add_free(dd, carry);
  return res;
}


/**Function********************************************************************

  Synopsis           [Perform the subtraction operations
  on two Word expressions represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The size of both arguments should be the same.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_minus(DdManager* dd, AddArray_ptr arg1,
				                AddArray_ptr arg2)
{
  AddArray_ptr res;
  add_ptr carry;

  AddArray_ptr not_arg2;
  add_ptr one = add_one(dd);

  /* arg1 - arg2 ==  arg1 + (not arg2) + 1 */
  not_arg2 = add_array_apply_on_bits(dd, node_not, arg2);
  
  add_array_adder(dd, arg1, not_arg2, one, &res, &carry);

  add_free(dd, one);
  add_free(dd, carry);
  AddArray_destroy(dd, not_arg2);
  return res;
}


/**Function********************************************************************

  Synopsis           [Perform the multiplication operations
  on two Word expressions represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The size of both arguments should be the same.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_times(DdManager* dd,
				 AddArray_ptr arg1, AddArray_ptr arg2)
{
  
  int index;
  const int N = AddArray_get_size(arg1);
  AddArray_ptr res = AddArray_create(N);

  nusmv_assert( N > 0 && N == AddArray_get_size(arg2));

  /* set the result array to (arg2 & arg1[0]), i.e. process first bit of arg1 */
  for (index = 0 ; index < N; ++index) {
    AddArray_set_n(res, index, add_apply(dd, node_and, 
					 AddArray_get_n(arg1, 0),
					 AddArray_get_n(arg2, index)));
  }

  /* A*B = ((B & A[0])<<0) +...+ ((B & A[i])<<i) +...+ ((B & A[N-1])<<N-1) */

  /* sum up all (arg2 & arg1[i]).
   Index begins with 1 because first bit has been already processed */
  for (index = 1; index < N; ++index) {
    int k;
    add_ptr carry_in;
    
    /* create an adder, i.e. (N - index) lower bits of (B & A[index])*/
    AddArray_ptr adder = AddArray_create(N - index);
    for (k = 0; k < N - index; ++ k){
      add_ptr add = add_apply(dd, node_and, 
			      AddArray_get_n(arg1, index),
			      AddArray_get_n(arg2, k));
      AddArray_set_n(adder, k, add);
    } /* for k */

    /* add the adder to the (N - index) higher bits of the result */
    carry_in = add_zero(dd); /* it will be de-referenced in the loop */

    for (k = 0; k < N - index; ++k) {
      add_ptr sum;
      add_ptr tmp_carry;
      
      add_array_full_adder(dd, AddArray_get_n(res, k + index),
			   AddArray_get_n(adder, k),
			   carry_in, &sum, &tmp_carry);
      add_free(dd, AddArray_get_n(res, k + index));
      AddArray_set_n(res, k + index, sum);
      add_free(dd, carry_in);
      carry_in = tmp_carry;
    } /* for k */

    AddArray_destroy(dd, adder);
  } /* for index */
  return res;
}


/**Function********************************************************************

  Synopsis           [Perform the division operations
  on two Word expressions represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The size of both arguments should be the same.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_divide(DdManager* dd,
				  AddArray_ptr arg1, AddArray_ptr arg2)
{
  AddArray_ptr quotient;
  AddArray_ptr remainder;

  add_array_division_remainder(dd, arg1, arg2, &quotient, &remainder);
  
  AddArray_destroy(dd, remainder);
  return quotient;
}


/**Function********************************************************************

  Synopsis           [Perform the remainder operations
  on two Word expressions represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The size of both arguments should be the same.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_mod(DdManager* dd,
			       AddArray_ptr arg1, AddArray_ptr arg2)
{
  AddArray_ptr quotient;
  AddArray_ptr remainder;

  add_array_division_remainder(dd, arg1, arg2, &quotient, &remainder);
  
  AddArray_destroy(dd, quotient);
  return remainder;
}


/**Function********************************************************************

  Synopsis           [Performs left shift operations
  on a Word expression represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The "number" argument represent 
  the number of bits to shift. "number" should have only one ADD.
  NB: The invoker should destroy the returned array.

  NB for developers:
  Every i-th bit of returned array will be:
       ITE(number=0 , arg[i],
           ITE(number=1, arg[i-1],
	    ...
             ITE(number=i, arg[0],
	      ITE(number >=0 && number <= width, zero, FAILURE)
   Does anyone know a better encoding?
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_left_shift(DdManager* dd, AddArray_ptr arg,
				      AddArray_ptr number)
{
  const int width = AddArray_get_size(arg);
  int i;
  add_ptr number_add = AddArray_get_add(number);/*number must have only one ADD*/

  AddArray_ptr res = AddArray_create(width);

  /* create the default case, i.e. the last ITE (see description) */
  add_ptr defaultCase = add_array_create_default_value_of_shift_operation(
			      dd,
			      number_add,
			      width,
			      "Right operand of left-shift is out of range");
  nusmv_assert(width > 0);
  
  /* proceed one bit at every iteration */
  for (i = 0; i < width; ++i) {
    int n;
    add_ptr bit = defaultCase;
    add_ref(bit); /* it will be de-referenced in the loop below */

    /* create all other ITEs */
    for (n = i; n >= 0; --n) {
      /* create  ITE(number=n, arg[i-n], ...) */
      add_ptr n_add = add_leaf(dd, find_node(NUMBER, (node_ptr)n, Nil));
      add_ptr number_eq_n = add_apply(dd, node_equal, number_add, n_add);
      add_ptr tmp = add_ifthenelse(dd, number_eq_n,
				   AddArray_get_n(arg, i - n),
				   bit);

      add_free(dd, n_add);
      add_free(dd, number_eq_n);
      add_free(dd, bit);
      bit = tmp;
    } /* for (n) */
    /* set the obtained bit into the returned array */
    AddArray_set_n(res, i, bit);
  } /* for (i) */

  add_free(dd, defaultCase);

  return res;
}


/**Function********************************************************************

  Synopsis           [Performs right shift operations
  on a Word expression represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The "number" argument represent 
  the number of bits to shift. "number" should have only one ADD.
  NB: The invoker should destroy the returned array.

  NB for developers:
  Every i-th bit of returned array will be:
       ITE(number=0 , arg[i],
           ITE(number=1, arg[i+1],
	    ...
             ITE(number=width-1-i, arg[width-1],
	      ITE(number >=0 && number <= width, zero, FAILURE)
   Does anyone have a better encoding?
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_right_shift(DdManager* dd, AddArray_ptr arg,
				       AddArray_ptr number)
{
  const int width = AddArray_get_size(arg);
  int i;
  add_ptr number_add = AddArray_get_add(number);/*number must have only one ADD*/

  AddArray_ptr res = AddArray_create(width);

  /* create the default case, i.e. the last ITE (see description) */
  add_ptr defaultCase = add_array_create_default_value_of_shift_operation(
			      dd,
			      number_add,
			      width,
			      "Right operand of right-shift is out of range");
  nusmv_assert(width > 0);
  
  /* proceed one bit at every iteration */
  for (i = 0; i < width; ++i) {
    int n;
    add_ptr bit = defaultCase;
    add_ref(bit); /* it will be de-referenced in the loop below */

    /* create all other ITEs */
    for (n = width - 1 - i; n >= 0; --n) {
      /* create  ITE(number=n, arg[i-n], ...) */
      add_ptr n_add = add_leaf(dd, find_node(NUMBER, (node_ptr)n, Nil));
      add_ptr number_eq_n = add_apply(dd, node_equal, number_add, n_add);
      add_ptr tmp = add_ifthenelse(dd, number_eq_n,
				   AddArray_get_n(arg, i + n),
				   bit);

      add_free(dd, n_add);
      add_free(dd, number_eq_n);
      add_free(dd, bit);
      bit = tmp;
    } /* for (n) */
    /* set the obtained bit into the returned array */
    AddArray_set_n(res, i, bit);
  } /* for (i) */

  add_free(dd, defaultCase);

  return res;
}


/**Function********************************************************************

  Synopsis           [Performs left rotate operations
  on a Word expression represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The "number" argument represent 
  the number of bits to rotate. "number" should have only one ADD.
  NB: The invoker should destroy the returned array.

  NB for developers:
  Every i-th bit  of returned array will be:
       ITE(number=0 , arg[i],
           ITE(number=1, arg[i-1],
	    ...
	     ITE(number=i, arg[0],
              ITE(number=i+1, arg[width-1],
	       ...
               ITE(number=width-1, arg[i+1],
                ITE(number=width, arg[i], FAILURE]
   Does anyone have a better encoding?
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_left_rotate(DdManager* dd, AddArray_ptr arg,
					AddArray_ptr number)
{
  const int width = AddArray_get_size(arg);
  int i;
  add_ptr number_add = AddArray_get_add(number);/*number must have only one ADD*/

  AddArray_ptr res = AddArray_create(width);

  /* create the default case, i.e. the last ITE (see description) */
  add_ptr errorCase = add_leaf(dd, 
   find_node(FAILURE,
      (node_ptr)find_string("Right operand of rotate operation is out of range"),
	     (node_ptr)(int)node_get_lineno(get_the_node())));
  
  nusmv_assert(width > 0);
  
  /* proceed one bit at every iteration */
  for (i = 0; i < width; ++i) {
    int n;
    add_ptr bit = errorCase;
    add_ref(bit); /* it will be de-referenced in the loop below */

    /* create all other ITEs */
    for (n = width ; n >= 0; --n) {
      /* create  ITE(number=n, arg[i-n], ...) */
      add_ptr n_add = add_leaf(dd, find_node(NUMBER, (node_ptr)n, Nil));
      add_ptr number_eq_n = add_apply(dd, node_equal, number_add, n_add);
      add_ptr tmp = add_ifthenelse(dd, number_eq_n,
				   AddArray_get_n(arg, 
						  i - n >= 0 
						  ? i - n 
						  : i - n + width),
				   bit);
      add_free(dd, n_add);
      add_free(dd, number_eq_n);
      add_free(dd, bit);
      bit = tmp;
    } /* for (n) */
    /* set the obtained bit into the returned array */
    AddArray_set_n(res, i, bit);
  } /* for (i) */

  add_free(dd, errorCase);

  return res;
}


/**Function********************************************************************

  Synopsis           [Performs right rotate operations
  on a Word expression represented as an array of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The "number" argument represent 
  the number of bits to rotate. "number" should have only one ADD.
  NB: The invoker should destroy the returned array.

  NB for developers:
  Every i-th bit of returned array will be:
       ITE(number=0 , arg[i],
           ITE(number=1, arg[i+1],
	    ...
	     ITE(number=width-1-i, arg[width-1],
              ITE(number=width-2-i, arg[0],
	       ...
               ITE(number=width-1, arg[i-1],
                ITE(number=width, arg[i], FAILURE]
   Does anyone have a better encoding?
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_right_rotate(DdManager* dd, AddArray_ptr arg,
					AddArray_ptr number)
{
  const int width = AddArray_get_size(arg);
  add_ptr width_add = add_leaf(dd, find_node(NUMBER, (node_ptr)width, Nil));
  add_ptr number_add = AddArray_get_add(number); /*one only ADD should be there*/
  /* create "width - number" */
  AddArray_ptr width_minus_number 
    = AddArray_from_add(add_apply(dd, node_minus, width_add, number_add));
  /* the result */
  AddArray_ptr res = AddArray_word_left_rotate(dd, arg, width_minus_number);

  add_free(dd, width_add);
  AddArray_destroy(dd, width_minus_number);
  return res;
}


/**Function********************************************************************

  Synopsis           [Performs less-then operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [the size of arguments should be the same
  The returned array will constain only one (boolean) ADD.
  NB: The invoker should destroy the returned array.
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_less(DdManager* dd,
				AddArray_ptr arg1, AddArray_ptr arg2)
{
  /* A < B 
     === NOT (A >= B) 
     === NOT ("carry bit of" (B - A - 1))
     === NOT NOT ("carry bit of" (B + 11...111 + 1 - A - 1))
     === "carry bit of" ((not A) + B)
  */

  AddArray_ptr not_arg1 = add_array_apply_on_bits(dd, node_not, arg1);
  add_ptr zero = add_zero(dd);
 
  AddArray_ptr plus;
  add_ptr carry;
 
  add_array_adder(dd, not_arg1, arg2, zero, &plus, &carry);

  AddArray_destroy(dd, plus);
  add_free(dd, zero);
  AddArray_destroy(dd, not_arg1);
  
  return AddArray_from_add(carry);
}


/**Function********************************************************************

  Synopsis           [Performs less-or-equal operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [the size of arguments should be the same
  The returned array will constain only one (boolean) ADD.
  NB: The invoker should destroy the returned array.
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_less_equal(DdManager* dd,
				      AddArray_ptr arg1, AddArray_ptr arg2)
{
  /* A <= B ===
     A < (B + 1) ==
     === "carry bit of" ((not A) + (B + 1))    (see  AddArray_word_less)
  */

  AddArray_ptr not_arg1 = add_array_apply_on_bits(dd, node_not, arg1);
  add_ptr one = add_one(dd);
 
  AddArray_ptr plus;
  add_ptr carry;
 
  add_array_adder(dd, not_arg1, arg2, one, &plus, &carry);
  
  AddArray_destroy(dd, plus);
  add_free(dd, one);
  AddArray_destroy(dd, not_arg1);
  
  return AddArray_from_add(carry);
}


/**Function********************************************************************

  Synopsis           [Performs greater-then operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [the size of arguments should be the same
  The returned array will constain only one (boolean) ADD.
  NB: The invoker should destroy the returned array.
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_greater(DdManager* dd,
				   AddArray_ptr arg1, AddArray_ptr arg2)
{
  return AddArray_word_less(dd, arg2, arg1);
}


/**Function********************************************************************

  Synopsis           [Performs greater-or-equal operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [the size of arguments should be the same
  The returned array will constain only one (boolean) ADD.
  NB: The invoker should destroy the returned array.
]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_greater_equal(DdManager* dd,
					 AddArray_ptr arg1, AddArray_ptr arg2)
{
    return AddArray_word_less_equal(dd, arg2, arg1);
}


/**Function********************************************************************

  Synopsis           [Performs equal-operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [the size of arguments should be the same
  The returned array will constain only one (boolean) ADD.
  NB: The invoker should destroy the returned array.
  ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_equal(DdManager* dd,
				 AddArray_ptr arg1, AddArray_ptr arg2)
{
  add_ptr result = add_one(dd);
  int i;
  const int width = AddArray_get_size(arg1);

  nusmv_assert(width == AddArray_get_size(arg2) && width > 0);
  
  for (i = 0; i < width; ++i) {
    add_ptr bit_equal = add_apply(dd, node_iff,
				  AddArray_get_n(arg1, i), 
				  AddArray_get_n(arg2, i));
    add_ptr tmp = add_apply(dd, node_and, bit_equal, result);
    add_free(dd, bit_equal);
    add_free(dd, result);
    result = tmp;
  }
  
  return AddArray_from_add(result);
}


/**Function********************************************************************

  Synopsis           [Performs not-equal-operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [the size of arguments should be the same
  The returned array will constain only one (boolean) ADD.
  NB: The invoker should destroy the returned array.
  ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_not_equal(DdManager* dd,
				     AddArray_ptr arg1, AddArray_ptr arg2)
{
  add_ptr result = add_zero(dd);
  int i;
  const int width = AddArray_get_size(arg1);
  
  nusmv_assert(width == AddArray_get_size(arg2) && width > 0);
  
  for (i = 0; i < width; ++i) {
    add_ptr bit_not_equal = add_apply(dd, node_xor,
				      AddArray_get_n(arg1, i), 
				      AddArray_get_n(arg2, i));
    add_ptr tmp = add_apply(dd, node_or, bit_not_equal, result);
    add_free(dd, bit_not_equal);
    add_free(dd, result);
    result = tmp;
  }
  
  return AddArray_from_add(result);
}


/**Function********************************************************************

  Synopsis           [Performs bit-selection operation
  on a Word expression represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [The high-bit and low-bit of selections
  are specified by "range". "range" must
  be ADD leafs with a RANGE node (holding two integer constants,
  and these constant must be in the range [width-1, 0]).
  NB: The invoker should destroy the returned array.
  ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_bit_selection(DdManager* dd,
					 AddArray_ptr word,
					 AddArray_ptr range)
{
  const int width = AddArray_get_size(word);
  int i;
  int highBit;
  int lowBit;
  AddArray_ptr res;

  /* convert the ADD representation of high and low bits into integers */
  {
    add_ptr range_add = AddArray_get_add(range);/*one ADD may be in range*/
    node_ptr range_node;

    nusmv_assert(add_isleaf(range_add));/* range must be a value, not ADD tree */

    range_node = add_get_leaf(dd, range_add);
    nusmv_assert(node_get_type(range_node) == RANGE &&
		 node_get_type(car(range_node)) == NUMBER &&
		 node_get_type(cdr(range_node)) == NUMBER);

    highBit = (int)car(car(range_node));
    lowBit = (int)car(cdr(range_node));
   
    nusmv_assert(0 <= lowBit && lowBit <= highBit && highBit < width);
  }

  i = highBit - lowBit + 1;
  res = AddArray_create(i);
  for ( --i; i >= 0; --i){
    add_ptr add = AddArray_get_n(word, i + lowBit);
    add_ref(add);
    AddArray_set_n(res, i, add);
  }
  return res;
}


/**Function********************************************************************

  Synopsis           [Performs concatination operation
  on two Word expressions represented as arrays of ADD.
  Every ADD corresponds to a bit of a Word expression]

  Description        [
  NB: The invoker should destroy the returned array.
  ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
AddArray_ptr AddArray_word_concatination(DdManager* dd,
					 AddArray_ptr arg1, AddArray_ptr arg2)
{
  const int width1 = AddArray_get_size(arg1);
  const int width2 = AddArray_get_size(arg2);
  int i;

  AddArray_ptr res = AddArray_create(width1 + width2);

  for (i = 0; i < width2; ++i) {
    add_ptr add = AddArray_get_n(arg2, i);
    add_ref(add);
    AddArray_set_n(res, i, add);
  }

  for (i = 0; i < width1; ++i) {
    add_ptr add = AddArray_get_n(arg1, i);
    add_ref(add);
    AddArray_set_n(res, i + width2, add);
  }
  return res;
}


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Performes the full-adder operation on the arguments arg1, arg2
  and carry_in. The returned sum-bit and carry-bit are returned in
  sum and carry_out.]

  Description [The returned ADD (sum and carry_out) are referenced.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static void add_array_full_adder (DdManager* dd,
				  add_ptr arg1,
				  add_ptr arg2,
				  add_ptr carry_in,
				  add_ptr* sum,
				  add_ptr* carry_out)
{
  add_ptr arg1_xor_arg2;
  add_ptr arg1_and_arg2;
  add_ptr arg1_xor_arg2_and_carry_in;
    
  nusmv_assert( sum != (add_ptr*)NULL && carry_out != (add_ptr*)NULL);
  
  /* sum = arg1 XOR arg2 XOR carry_in */
  arg1_xor_arg2 = add_apply(dd, node_xor, arg1, arg2);
  *sum = add_apply(dd, node_xor, arg1_xor_arg2, carry_in);
  
  /* curry_out = (arg1 and arg2) OR ((arg1 XOR arg2) AND carry_in) */
  arg1_and_arg2 = add_apply(dd, node_and, arg1, arg2);
  arg1_xor_arg2_and_carry_in =add_apply(dd, node_and, arg1_xor_arg2, carry_in);
  *carry_out =add_apply(dd, node_or, arg1_and_arg2, arg1_xor_arg2_and_carry_in);

  add_free(dd, arg1_xor_arg2);
  add_free(dd, arg1_and_arg2);
  add_free(dd, arg1_xor_arg2_and_carry_in);
  return;
}


/**Function********************************************************************

  Synopsis    [Performes the addition of two Word expressions and 
  a carry-in bit.]
  
  Description [The sum is returned by the parameter res (the invoker
  must destroy this array), and the final carry-bit is returned by the
  parameter carry_out (the ADD is referenced).
  The size of input arrays must be equal(and positive). ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static void add_array_adder(DdManager* dd, AddArray_ptr arg1, AddArray_ptr arg2,
			    add_ptr carry_in,
			    AddArray_ptr* res,  add_ptr* carry_out)
{
  const int width = AddArray_get_size(arg1);
  int i;

  nusmv_assert( res != (AddArray_ptr*)NULL && carry_out != (add_ptr*)NULL);
  nusmv_assert(AddArray_get_size(arg2) == width && width > 0);
  
  
  *res = AddArray_create(width);
  add_ref(carry_in);/* it will be de-referenced in the loop or in an outter fun */

  for (i = 0; i < width; ++i) {
    add_ptr sum;
    add_ptr tmp_carry;
    
    add_array_full_adder(dd, AddArray_get_n(arg1, i), AddArray_get_n(arg2, i),
			 carry_in, &sum, &tmp_carry);
    AddArray_set_n(*res, i, sum);
    add_free(dd, carry_in);
    carry_in = tmp_carry;
  } /* for */

  *carry_out = carry_in;
  return;
}


/**Function********************************************************************

  Synopsis    [Performs division operations on two Word expressions (which 
  are encoded as arrays of bits)]

  Description [The quotient and the remainder is returned in the 
  parameters "quotient" and "remainder" respectively.
  The invoker should free the returned arrays.

  The size of arguments should be the same (and positive).

  Every bit of the resulting arrays is wrapped in ITE which check
  the second argument (of the operation) for not being zero.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static void add_array_division_remainder(DdManager* dd,
					 AddArray_ptr arg1,
					 AddArray_ptr arg2,
					 AddArray_ptr* quotient,
					 AddArray_ptr* remainder)
{
  int index;
  const int N = AddArray_get_size(arg1);
  add_ptr isNotZero;
  AddArray_ptr quot;
  AddArray_ptr rem;

  nusmv_assert(N > 0 && AddArray_get_size(arg2) == N);

  /* Check the second argument for being NOT zero */
  {
    add_ptr tmp;
    
    for (isNotZero = add_zero(dd), index = 0; index < N; ++index) {
      tmp = add_apply(dd, node_or, isNotZero, AddArray_get_n(arg2, index));
      add_free(dd, isNotZero);
      isNotZero = tmp;
    }
  }

  /* create remainder and quotient and set the remainder to the zero value */
  quot = AddArray_create(N);
  rem = AddArray_create(N);
  for (index = 0; index < N; ++index) {
    AddArray_set_n(rem, index, add_zero(dd));
  }
  
  /* calculate the division operation */
  for (index = N - 1; index >= 0; --index) {
    add_ptr isDividable;

    /* shift remainder left by 1 bit and add the bit number 'index' from arg1 */
    {
      int k;
      add_ptr tmp;
      add_free(dd, AddArray_get_n(rem, N-1));
      for (k = N - 1; k > 0; --k) {
	AddArray_set_n(rem, k, AddArray_get_n(rem, k - 1));
      }
      tmp = AddArray_get_n(arg1, index);
      add_ref(tmp);
      AddArray_set_n(rem, 0, tmp);
    }
      
    /* calculate ADD for (remainder >= arg2), i.e. flag that 
       subtraction can be performed. Set the quotinet[index] to this value .
    */
    {
      AddArray_ptr tmp = AddArray_word_greater_equal(dd, rem, arg2);
      isDividable = AddArray_get_add(tmp);
      add_ref(isDividable);
      AddArray_destroy(dd, tmp);
      AddArray_set_n(quot, index, isDividable);
    }
    
    /* set remainder to (remainder - (arg2 AND (remainder >= arg2))) */
    {
      AddArray_ptr diff;
      AddArray_ptr guardedArg = AddArray_create(N);
      int k;
      
      for (k = 0; k < N; ++k) {
	add_ptr add = add_apply(dd, node_and, AddArray_get_n(arg2, k),
				isDividable);
	AddArray_set_n(guardedArg, k, add);
      }
      
      diff = AddArray_word_minus(dd, rem, guardedArg);

      AddArray_destroy(dd, rem);
      AddArray_destroy(dd, guardedArg);
      rem = diff;
    }
  } /* for index */
  
  /* Calculation of the quotient and the remainder is finished.
     Now guard every bit of quotient and remainder by the condition that
     the divisor is not equal to zero 
  */
  { 
    /* create an error message */
    add_ptr error = add_leaf(dd, find_node(FAILURE,
			       (node_ptr)find_string("Division by zero"),
			       (node_ptr)(int)node_get_lineno(get_the_node())));
    
    for (index = 0; index < N; ++index) {
      add_ptr guarded;
      /* quotient */
      guarded
	= add_ifthenelse(dd, isNotZero, AddArray_get_n(quot, index), error);
      add_free(dd, AddArray_get_n(quot, index));
      AddArray_set_n(quot, index, guarded);
      /* remainder */
      guarded
	= add_ifthenelse(dd, isNotZero, AddArray_get_n(rem, index), error);
      add_free(dd, AddArray_get_n(rem, index));
      AddArray_set_n(rem, index, guarded);
    }

    add_free(dd, error);
  }

  add_free(dd, isNotZero);
  *remainder = rem;
  *quotient = quot;
  return;
}


/**Function********************************************************************

  Synopsis    [Create an ADD of the default case in shift operations, i.e. ADD of
  ITE(nubmer >=0 && number <= width, zero, FAILURE).
  ]

  Description [This function is used in shift operations.
  See, for example, AddArray_word_left_shift.

  The 'number' is ADD of the number of bit the Word is shifted.
  'width' is the width of the given Word expression.
  'errMessage' is the error message to print if number is out of range,
  for example, "Right operand of left-shift is out of range".

  NB: The returned ADD is referenced.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static add_ptr add_array_create_default_value_of_shift_operation(DdManager* dd,
								 add_ptr number,
								 int width,
							        char* errMessage)
{
  add_ptr zero = add_zero(dd);
  add_ptr widthAdd = add_leaf(dd, find_node(NUMBER, (node_ptr)width, Nil));
  add_ptr errorCase = add_leaf(dd, 
			    find_node(FAILURE,
				(node_ptr)find_string(errMessage),
				(node_ptr)(int)node_get_lineno(get_the_node())));
  
  add_ptr n_ge_0 = add_apply(dd, node_ge, number, zero);
  add_ptr n_le_width = add_apply(dd, node_le, number, widthAdd);
  add_ptr n_ge_0_AND_n_le_width = add_apply(dd, node_and, n_ge_0, n_le_width);
    
  add_ptr res = add_ifthenelse(dd, n_ge_0_AND_n_le_width, zero, errorCase);

  add_free(dd, n_ge_0_AND_n_le_width);
  add_free(dd, n_le_width);
  add_free(dd, n_ge_0);
  add_free(dd, errorCase);
  add_free(dd, widthAdd);
  add_free(dd, zero);

  return res;
}

/**Function********************************************************************

  Synopsis    [A function takes an array and applies ADD's
  "add_monadic_apply" with the operation "op" on every bit of the array.]

  Description [the result of the functions is a new array
        [add_monadic_apply(op, arg[0]), 
         add_monadic_apply(op, arg[1]), 
	 ....
	 add_monadic_apply(op, arg[width-1])].

  NB: The invoker should free the returned array.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static AddArray_ptr add_array_apply_on_bits(DdManager* dd, NPFNN/*NPFCVT*/ op,
					    AddArray_ptr arg)
{
  AddArray_ptr res;
  int i;
  const int width = AddArray_get_size(arg);
  /* array should be of the positive size */
  nusmv_assert(width > 0);
  
  res = AddArray_create(width);
  for (i = 0; i < width; ++i) {
    add_ptr add = add_monadic_apply(dd, op, AddArray_get_n(arg, i));
    AddArray_set_n(res, i, add);
  }
  return res;
}



