/**CFile*****************************************************************

  FileName    [SexpFsm.c]

  PackageName [fsm.sexp]

  Synopsis    [The SexpFsm implementation]

  Description [This file contains the implementation of a wrapper of (a part 
               of) the SexpFsm API onto the NuSMV code structure.
               Only functions needed by the BddFsm are provided.]

  SeeAlso     [SexpFsm.h]

  Author      [Roberto Cavada]

  Copyright   [
  This file is part of the ``fsm.sexp'' package of NuSMV version 2. 
  Copyright (C) 2003 by ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "SexpFsm.h"
#include "sexpInt.h"

#include "utils/assoc.h"
#include "parser/symbols.h"

/* there are still some variables to be accessed there: */
#include "compile/compile.h" 
#include "compile/compileInt.h" 
#include "enc/utils/WordNumber.h" 
#include "compile/compilePredicateNormaliser.h"

static char rcsid[] UTIL_UNUSED = "$Id: SexpFsm.c,v 1.1.1.1 2005/12/20 17:06:59 timo Exp $";


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/**Type************************************************************************

  Synopsis [A fsm for a single variable. It is represented as a triple
  of Expr_ptr ]

  Description [Private structure, internally used]

  SeeAlso     []

******************************************************************************/
typedef node_ptr VarFsm_ptr;
#define VAR_FSM(x)  ((VarFsm_ptr) x)


/**Type************************************************************************

  Synopsis    [Sexp FSM, the type exported by this package]

  Description [Allows sharing of structures between derived fsm. You
  can derive a fsm from an existing one, by copying it, or by
  converting it from scalar to boolean. In general ay operation which
  takes one fsm and returns a different fsm is a derivating operation. 
  A set of FSMs which share the same structures, is called "family".]

  SeeAlso     []

******************************************************************************/
typedef struct SexpFsm_TAG {
  int* family_counter; 

  BddEnc_ptr enc;
  SymbTable_ptr symb_table;
  SymbLayer_ptr det_layer;

  VarSet_ptr vars_list;
  hash_ptr hash_var_fsm; 
  
  Expr_ptr init;
  Expr_ptr invar;
  Expr_ptr trans;
  Expr_ptr input;

  node_ptr justice;    /* list of expressions */
  node_ptr compassion; /* list of couple of expressions */
} SexpFsm;



/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
#define NODE_PTR(x)  ((node_ptr) x)  /* upcast from VarFsm_ptr */


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void sexp_fsm_create_shared_structures ARGS((SexpFsm_ptr self));
static void sexp_fsm_try_to_destroy_shared_structures ARGS((SexpFsm_ptr self));
static void sexp_fsm_copy_shared_structures ARGS((const SexpFsm_ptr self, 
						  SexpFsm_ptr copy));

static void 
sexp_fsm_hash_var_fsm_init ARGS((SexpFsm_ptr self, VarSet_ptr vars_list,
				       FlatHierarchy_ptr hierarchy));

static void sexp_fsm_init ARGS((SexpFsm_ptr self, 
				BddEnc_ptr enc, 
				SymbLayer_ptr det_layer, 
				VarSet_ptr vars_list, 
				node_ptr justice,
				node_ptr compassion));


static Expr_ptr sexp_fsm_simplify_expr ARGS((SexpFsm_ptr self, 
					     hash_ptr hash, Expr_ptr expr, 
					     const int group));
			    
static hash_ptr simplifier_hash_create ARGS((void));
static void simplifier_hash_destroy ARGS((hash_ptr hash));
static void simplifier_hash_add_expr ARGS((hash_ptr hash, 
					   Expr_ptr expr, const int group));
static boolean simplifier_hash_query_expr ARGS((hash_ptr hash, Expr_ptr expr, 
						const int group));


static void sexp_fsm_hash_var_fsm_create ARGS((SexpFsm_ptr self));
static void sexp_fsm_hash_var_fsm_destroy ARGS((SexpFsm_ptr self));
static assoc_retval 
sexp_fsm_callback_var_fsm_free ARGS((char *key, char *data, char * arg));
static VarFsm_ptr 
sexp_fsm_hash_var_fsm_lookup_var ARGS((SexpFsm_ptr self, node_ptr var));
static void 
sexp_fsm_hash_var_fsm_insert_var ARGS((SexpFsm_ptr self, 
				       node_ptr var, VarFsm_ptr varfsm));


static VarFsm_ptr var_fsm_create ARGS((Expr_ptr init, Expr_ptr invar, 
				       Expr_ptr next));

static void var_fsm_destroy ARGS((VarFsm_ptr self));
static Expr_ptr var_fsm_get_init ARGS((VarFsm_ptr self));
static Expr_ptr var_fsm_get_invar ARGS((VarFsm_ptr self));
static Expr_ptr var_fsm_get_next ARGS((VarFsm_ptr self));
static Expr_ptr var_fsm_get_input ARGS((VarFsm_ptr self));

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis [Costructor for a scalar and boolean sexp fsm. If det_layer
  is null, then a scalar fsm will be created, otherwise a boolean fsm
  will be created.]

  Description [If provided, the layer will be filled with the
  determinization variables that might derive from the
  booleanization.]

  SideEffects        []

******************************************************************************/
SexpFsm_ptr SexpFsm_create(BddEnc_ptr enc,
			   SymbLayer_ptr det_layer, 
			   VarSet_ptr vars_list, 
			   node_ptr justice,
			   node_ptr compassion,
			   FlatHierarchy_ptr hierarchy)
			   
{
  SexpFsm_ptr self;

  /* allocation: */
  self = ALLOC(SexpFsm, 1);
  SEXP_FSM_CHECK_INSTANCE(self);

  self->hash_var_fsm = (hash_ptr) NULL;
  sexp_fsm_create_shared_structures(self);

  /* initialization: */
  sexp_fsm_hash_var_fsm_init(self, vars_list, hierarchy);

  sexp_fsm_init(self, enc, det_layer, vars_list, justice, compassion);

  return self;
}



/**Function********************************************************************

  Synopsis           [Costructor for a scalar and boolean sexp fsm]

  Description        [This method is temporary exported to be used 
  from the LTL side, to compile the table fsm. 
  Warning: the variable FSMs will be not constructed, so the user cannot 
  use methods to obtain the single variable fsm information. 
  If a boolean fsm is being constructed, than all epressions must be 
  already booleanized.
  ]

  SideEffects        []

******************************************************************************/
SexpFsm_ptr 
SexpFsm_create_from_members(BddEnc_ptr enc, SymbLayer_ptr det_layer, 
			    Expr_ptr init, Expr_ptr invar,
			    Expr_ptr trans, Expr_ptr input, 
			    node_ptr justice, node_ptr compassion)
{
  SexpFsm_ptr self;

  /* allocation: */
  self = ALLOC(SexpFsm, 1);
  SEXP_FSM_CHECK_INSTANCE(self);

  self->hash_var_fsm = (hash_ptr) NULL;
  sexp_fsm_create_shared_structures(self);
  
  /* inits all private members: */
  self->enc = enc;
  self->symb_table = BaseEnc_get_symbol_table(BASE_ENC(enc));
  self->det_layer = det_layer;  
  self->vars_list = VAR_SET(NULL);

  self->init       = init;
  self->invar      = invar;
  self->trans      = trans;
  self->input      = input;
  self->justice    = justice;
  self->compassion = compassion;

  return self;
}


/**Function********************************************************************

  Synopsis           [Copy costructor]

  Description        []

  SideEffects        []

******************************************************************************/
SexpFsm_ptr SexpFsm_copy(const SexpFsm_ptr self)
{
  SexpFsm_ptr copy;

  SEXP_FSM_CHECK_INSTANCE(self);

  copy = ALLOC(SexpFsm, 1);
  SEXP_FSM_CHECK_INSTANCE(copy);

  copy->vars_list  = self->vars_list;
  copy->enc        = self->enc;
  copy->symb_table = self->symb_table;
  copy->det_layer  = self->det_layer;
  copy->init       = self->init;
  copy->invar      = self->invar;
  copy->trans      = self->trans;
  copy->input      = self->input;
  copy->justice    = self->justice;
  copy->compassion = self->compassion;
  
  sexp_fsm_copy_shared_structures(self, copy);

  return copy;
}


/**Function********************************************************************

  Synopsis           [Destructor]

  Description        []

  SideEffects        []

******************************************************************************/
void SexpFsm_destroy(SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);

  /* ------------------------------------------- */
  /* Free here all not-family structures         */
  /* ------------------------------------------- */    
  
  /* ------------------------------------------- */    

  sexp_fsm_try_to_destroy_shared_structures(self);
  FREE(self);
}


/**Function********************************************************************

  Synopsis           [Copy the Sexp FSM and perform predicate-normalisation
  on all the expressions.]

  Description        [Predicate-normalisations means that an expression is
  modified in such a way that at the end the subexpressions of a
  not-boolean expression can be only not-boolean. This is performed by
  changing boolean expression "exp" (which is a subexpression of a
  not-boolean expression) to "ITE(exp, 1, 0)", and then pushing all
  ITE up to the root of not-boolean expressions.

  Constrain: the given Sexp FSM has to be NOT boolean. Otherwise,
  it is meaningless to apply normalisation functions, since all the exporessions
  are already boolean.
  ]

  SideEffects        [SexpFsm_copy]

******************************************************************************/
SexpFsm_ptr SexpFsm_create_predicate_normalised_copy (const SexpFsm_ptr self,
					     PredicateNormaliser_ptr normaliser)
{
  SexpFsm_ptr copy;

  SEXP_FSM_CHECK_INSTANCE(self);
  /* self should not be booleanised */
  nusmv_assert(SYMB_LAYER(NULL) == self->det_layer);
  
  copy = ALLOC(SexpFsm, 1);
  SEXP_FSM_CHECK_INSTANCE(copy);

  copy->vars_list  = self->vars_list;
  copy->enc        = self->enc;
  copy->symb_table = self->symb_table;
  copy->det_layer  = self->det_layer;

  /* perform the normalisation */
  copy->init       = PredicateNormaliser_normalise_constrain(normaliser,
							     self->init);
  copy->invar      = PredicateNormaliser_normalise_constrain(normaliser,
							     self->invar);
  copy->trans      = PredicateNormaliser_normalise_constrain(normaliser,
							     self->trans);
  copy->input      = PredicateNormaliser_normalise_constrain(normaliser,
							     self->input);
  copy->justice    = PredicateNormaliser_normalise_constrain(normaliser,
							     self->justice);
  copy->compassion = PredicateNormaliser_normalise_constrain(normaliser,
							     self->compassion);

  /* shared structure must be copied since SexpFsm_scalar_to_boolean
     uses it. To my opinion ([AT]) that was quite ugly way to pass parameters 
  */
  sexp_fsm_copy_shared_structures(self, copy);
  
  return copy;
}


/**Function********************************************************************

  Synopsis           [Construct a boolean fsm from a scalar one]

  Description [Given layer will be filled with the determinization
  variables coming from the booleanization process. Returns a copy if
  self is already booleanized.]

  SideEffects        []

******************************************************************************/
SexpFsm_ptr SexpFsm_scalar_to_boolean(const SexpFsm_ptr self, 
				      SymbLayer_ptr det_layer)
{
  SexpFsm_ptr result;
  
  SEXP_FSM_CHECK_INSTANCE(self);
  SYMB_LAYER_CHECK_INSTANCE(det_layer); /* must be a valid layer */

  if (SexpFsm_is_boolean(self)) {
    result = SexpFsm_copy(self);
  }
  else {
    result = ALLOC(SexpFsm, 1);
    SEXP_FSM_CHECK_INSTANCE(result);
    result->hash_var_fsm = (hash_ptr) NULL;

    sexp_fsm_copy_shared_structures(self, result);

    sexp_fsm_init(result, self->enc, det_layer, self->vars_list, 
		  self->justice, self->compassion);
  }

  return result;
}


/**Function********************************************************************

  Synopsis           [Returns true if self if a booleanized fsm, false if 
  it is scalar]

  Description        []

  SideEffects        []

******************************************************************************/
boolean SexpFsm_is_boolean(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return (self->det_layer != SYMB_LAYER(NULL));
}


/**Function********************************************************************

  Synopsis           [Returns an Expr that collects init states for all 
  variables handled by self]

  Description        []

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_init(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return self->init;
}


/**Function********************************************************************

  Synopsis           [Returns an Expr that collects invar states for all 
  variables handled by self]

  Description        []

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_invar(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return self->invar;
}


/**Function********************************************************************

  Synopsis           [Returns an Expr that collects all next states for all 
  variables handled by self]

  Description        []

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_trans(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return self->trans;
}


/**Function********************************************************************

  Synopsis           [Returns an Expr that collects all input states for all 
  variables handled by self]

  Description        []

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_input(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  /* Currently no constraints over input are allowed, thus we return
     true to inidicate this. */
  return self->input;
}


/**Function********************************************************************

  Synopsis      [ Gets the sexp expression defining the initial state for
                  the variable "v". ]

  Description   [ Gets the sexp expression defining the initial state for
                  the variable "v". ]

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_var_init(const SexpFsm_ptr self, node_ptr var_name)
{
  VarFsm_ptr var_fsm;
  SEXP_FSM_CHECK_INSTANCE(self);

  var_fsm = find_assoc(self->hash_var_fsm, var_name);
  return var_fsm_get_init(var_fsm);
}


/**Function********************************************************************

  Synopsis      [ Gets the sexp expression defining the state constraints
                  for the variable "v". ]

  Description   [ Gets the sexp expression defining the state constraints
                  for the variable "v". ]

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_var_invar(const SexpFsm_ptr self, node_ptr var_name)
{
  VarFsm_ptr var_fsm;

  SEXP_FSM_CHECK_INSTANCE(self);
  var_fsm = find_assoc(self->hash_var_fsm, var_name);
  return var_fsm_get_invar(var_fsm);
}


/**Function********************************************************************

  Synopsis      [ Gets the sexp expression defining the transition relation
                  for the variable "v". ]

  Description   [ Gets the sexp expression defining the transition relation
                  for the variable "v". ]

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_var_trans(const SexpFsm_ptr self, node_ptr var_name)
{
  VarFsm_ptr var_fsm;

  SEXP_FSM_CHECK_INSTANCE(self);
  var_fsm = find_assoc(self->hash_var_fsm, var_name);
  return var_fsm_get_next(var_fsm);
}


/**Function********************************************************************

  Synopsis      [ Gets the sexp expression defining the input relation
                  for the variable "v". ]

  Description   []

  SideEffects        []

******************************************************************************/
Expr_ptr SexpFsm_get_var_input(const SexpFsm_ptr self, node_ptr var_name)
{
  VarFsm_ptr var_fsm;

  SEXP_FSM_CHECK_INSTANCE(self);
  var_fsm = find_assoc(self->hash_var_fsm, var_name);
  return var_fsm_get_input(var_fsm);
}




/**Function********************************************************************

  Synopsis      [ Gets the list of sexp expressions defining the set of justice 
                  constraints for this machine. ]

  Description   [ Gets the list of sexp expressions defining the set of justice 
                  constraints for this machine. ]

  SideEffects        []

******************************************************************************/
node_ptr SexpFsm_get_justice(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return self->justice;
}


/**Function********************************************************************

  Synopsis      [ Gets the list of sexp expressions defining the set of
                  compassion constraints for this machine. ]

  Description   [ Gets the list of sexp expressions defining the set of 
                  compassion constraints for this machine. ]

  SideEffects        []

******************************************************************************/
node_ptr SexpFsm_get_compassion(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return self->compassion;
}




VarSet_ptr SexpFsm_get_vars_list(const SexpFsm_ptr self)
{
  SEXP_FSM_CHECK_INSTANCE(self);
  return self->vars_list;
}


/*---------------------------------------------------------------------------*/
/* Static function definitions                                               */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis      [Call this from any constructor which creates a new family]

  Description   [Every time a fsm is created from scratch (i.e. via its 
  constructor), a new family is created. When a fsm D is derived from an 
  existing base fsm B, D belongs to the family of B]

  SideEffects        []

******************************************************************************/
static void sexp_fsm_create_shared_structures(SexpFsm_ptr self)
{
  /* a new family is allocated */
  self->family_counter = ALLOC(int, 1);
  nusmv_assert(self->family_counter != NULL);
  *(self->family_counter) = 1;

  /* ---------------------------------------- */
  /* put here all shared structure allocation */
  /* ---------------------------------------- */
  sexp_fsm_hash_var_fsm_create(self);
  
}


/**Function********************************************************************

  Synopsis      [Call this from every destructor which tries to destroy a 
  family. If the family is over, all shared structure will be destroyed]

  Description [All structures will be destroyed only if the family
  counter reaches the zero value]

  SideEffects        []

******************************************************************************/
static void sexp_fsm_try_to_destroy_shared_structures(SexpFsm_ptr self)
{
  nusmv_assert(self->family_counter != NULL); 

  *(self->family_counter) = *(self->family_counter) - 1;  
  nusmv_assert(*(self->family_counter)  >= 0);

  if ( (*(self->family_counter)) == 0 ) {
    FREE(self->family_counter);    
    self->family_counter = NULL;

    /* --------------------------------------------------------------------- */
    /* family is actually over. Put here destruction of any shared structure */
    /* --------------------------------------------------------------------- */
    sexp_fsm_hash_var_fsm_destroy(self);

  }
}


/**Function********************************************************************

  Synopsis      [Call this to copy data from fsm belonging to the same family. 
  This also increments the family counter]

  Description   []

  SideEffects        []

******************************************************************************/
static void sexp_fsm_copy_shared_structures(const SexpFsm_ptr self, 
					    SexpFsm_ptr copy)
{
  *(self->family_counter) = *(self->family_counter) + 1;  
  copy->family_counter = self->family_counter;  


  /* -------------------------------------- */
  /* put here copy of all shared structures */
  /* -------------------------------------- */

  copy->hash_var_fsm = self->hash_var_fsm;
}




/**Function********************************************************************

  Synopsis      [Initializes the vars fsm hash]

  Description   []

  SideEffects        []

******************************************************************************/
static void sexp_fsm_hash_var_fsm_init(SexpFsm_ptr self, VarSet_ptr vars_list,
				       FlatHierarchy_ptr hierarchy)
{
  node_ptr iter;

  iter = NODE_PTR(vars_list);
  while (iter != Nil) {
    int saved_yylineno = yylineno;

    VarFsm_ptr var_fsm;

    node_ptr var_name = car(iter);
    node_ptr var_name_i = find_node(SMALLINIT, var_name, Nil);
    node_ptr var_name_n = find_node(NEXT, var_name, Nil);

    node_ptr init_a  = FlatHierarchy_lookup_assign(hierarchy, var_name_i);
    node_ptr invar_a = FlatHierarchy_lookup_assign(hierarchy, var_name);
    node_ptr next_a  = FlatHierarchy_lookup_assign(hierarchy, var_name_n);

    node_ptr init_c  = FlatHierarchy_lookup_constrains(hierarchy, var_name_i);
    node_ptr invar_c = FlatHierarchy_lookup_constrains(hierarchy, var_name);
    node_ptr next_c  = FlatHierarchy_lookup_constrains(hierarchy, var_name_n);
    
    Expr_ptr init_sexp  = Expr_true();
    Expr_ptr invar_sexp = Expr_true();
    Expr_ptr trans_sexp = Expr_true();
    
    /* add all the constrains */
    if (Nil != init_c) init_sexp = Expr_and(init_sexp, EXPR(init_c));
    if (Nil != invar_c) invar_sexp = Expr_and(invar_sexp, EXPR(invar_c));
    if (Nil != next_c)  trans_sexp = Expr_and(trans_sexp, EXPR(next_c));
    
    /* add all assignments */
    if (Nil != init_a) {
      yylineno = init_a->lineno;
      init_sexp = Expr_and(init_sexp,
			   EXPR(find_node(EQDEF, var_name_i, init_a)));
    }
    
    if (Nil != invar_a) {
      yylineno = invar_a->lineno;
      invar_sexp = Expr_and(invar_sexp,
			    EXPR(new_node(EQDEF, var_name, invar_a)));
    }

    if (Nil != next_a) {
      yylineno = next_a->lineno;
      trans_sexp = Expr_and(trans_sexp,
			    EXPR(new_node(EQDEF, var_name_n, next_a)));
    }
 
    /* inserts the var fsm inside the hash table */
    var_fsm = var_fsm_create(init_sexp, invar_sexp, trans_sexp);
    sexp_fsm_hash_var_fsm_insert_var(self, var_name, var_fsm);

    iter = cdr(iter);
    yylineno = saved_yylineno;
  } /* loop */
}



/**Function********************************************************************

  Synopsis      [Initializes either the boolean or scalar sexp fsm]

  Description   []

  SideEffects        []

******************************************************************************/
static void sexp_fsm_init(SexpFsm_ptr self, 
			  BddEnc_ptr enc, 
			  SymbLayer_ptr det_layer, 
			  VarSet_ptr vars_list, 
			  node_ptr justice,
			  node_ptr compassion) 
{
  Expr_ptr init  = Expr_true();
  Expr_ptr invar = Expr_true();
  Expr_ptr next  = Expr_true();
  Expr_ptr input  = Expr_true();

  hash_ptr hash; 
  node_ptr iter;

  /* We set the verbose level to 0 and then we restore the original
     value. This because booleanization uses eval */  
  int curr_verbosity = get_verbose_level(options);
  set_verbose_level(options, 0);

  /* inits some private members */
  self->enc = enc;
  self->symb_table = BaseEnc_get_symbol_table(BASE_ENC(enc));
  self->det_layer = det_layer;
  self->vars_list = vars_list;

  
  hash = simplifier_hash_create();
  
  iter = vars_list;
  while (iter != Nil) {
    node_ptr var = car(iter);
    VarFsm_ptr varfsm = sexp_fsm_hash_var_fsm_lookup_var(self, var);
    
    if (varfsm != VAR_FSM(NULL)) {
      Expr_ptr tmp;

      /* inits */
      tmp = var_fsm_get_init(varfsm);
      if (tmp != Nil) {
	tmp = sexp_fsm_simplify_expr(self, hash, tmp, INIT);
	init = Expr_and(init, tmp);
      }

      /* invars */
      tmp = var_fsm_get_invar(varfsm);
      if (tmp != Nil) {
	tmp = sexp_fsm_simplify_expr(self, hash, tmp, INVAR);
	invar = Expr_and(invar, tmp);
      }
      
      /* next */
      tmp = var_fsm_get_next(varfsm);
      if (tmp != Nil) {
	tmp = sexp_fsm_simplify_expr(self, hash, tmp, TRANS);
	next = Expr_and(next, tmp);
      }
    }

    iter = cdr(iter);
  } /* loop */
  
  simplifier_hash_destroy(hash); 
  set_verbose_level(options, curr_verbosity);

  self->init       = init;
  self->invar      = invar;
  self->trans      = next;
  self->input      = input;

  /* Booleanizes fairness if needed: */
  if (self->det_layer != SYMB_LAYER(NULL)) {
    self->justice = expr2bexpr(self->enc, self->det_layer, justice);
    self->compassion = expr2bexpr(self->enc, self->det_layer, compassion);
  }
  else {
    self->justice    = justice;
    self->compassion = compassion;
  }

}


/**Function********************************************************************

  Synopsis           [removes duplicates from expression containing AND nodes]

  Description        [group identifies INVAR, TRANS or INIT group. 
  If booleanize is True, the expression will be also booleanized]

  SideEffects        []

******************************************************************************/
static Expr_ptr 
sexp_fsm_simplify_expr(SexpFsm_ptr self, hash_ptr hash, Expr_ptr expr, 
		       const int group)
{
  Expr_ptr result;
  
  if ((expr == EXPR(NULL)) || simplifier_hash_query_expr(hash, expr, group)) {
    result = Expr_true();
  }
  else {
    switch (node_get_type(NODE_PTR(expr))) {
    case AND:
      {
	Expr_ptr left, right;
	left  = sexp_fsm_simplify_expr(self, hash, car(NODE_PTR(expr)), group);

	right = sexp_fsm_simplify_expr(self, hash, cdr(NODE_PTR(expr)), group);

	result = Expr_and(left, right);
	break;
      }

    default:
      if (self->det_layer != SYMB_LAYER(NULL)) {
	result = EXPR(expr2bexpr(self->enc, self->det_layer, NODE_PTR(expr)));
      }
      else result = expr;

    } /* switch */
    
    simplifier_hash_add_expr(hash, expr, group);
  }
  
  return result;
}



/**Function********************************************************************

  Synopsis           [This is used when creating cluster list from vars list]

  Description        []

  SideEffects        []

******************************************************************************/
static hash_ptr simplifier_hash_create()
{
  hash_ptr result;

  result = st_init_table(st_ptrcmp, st_ptrhash);
  nusmv_assert(result != ((hash_ptr) NULL));

  return result;
}

/**Function********************************************************************

  Synopsis           [Call after sexp_fsm_cluster_hash_create]

  Description        []

  SideEffects        []

******************************************************************************/
static void simplifier_hash_destroy(hash_ptr hash)
{
  nusmv_assert(hash != (hash_ptr) NULL);
  st_free_table(hash);  
}


/**Function********************************************************************

  Synopsis           [To insert a new node in the hash]

  Description        [group is INIT, INVAR or TRANS]

  SideEffects        [The hash can change]

******************************************************************************/
static void 
simplifier_hash_add_expr(hash_ptr hash, Expr_ptr expr, const int group)
{
  int res;
  
  res = st_add_direct(hash, (char*) expr, (char*) group);
  nusmv_assert(res != ST_OUT_OF_MEM);
}


/**Function********************************************************************

  Synopsis           [Queries for an element in the hash, returns True if 
  found]

  Description        []

  SideEffects        []

******************************************************************************/
static boolean 
simplifier_hash_query_expr(hash_ptr hash, Expr_ptr expr, 
			   const int group)
{
  int hashed_group; 
  boolean result;

  result = st_lookup(hash, (char *) expr, (char **) &hashed_group);

  /* groups are checked consecutively, i.e. at first, *all* INIT expressions 
     are checked, then *all* INVAR, and then *all* TRANS. So hash_group
     will not interfere with each other
  */
  return (result && (hashed_group == group));
}


/**Function********************************************************************

  Synopsis           [Call to create the hash for var fsm inside the fsm]

  Description        [Private method, used internally]

  SideEffects        []

******************************************************************************/
static void sexp_fsm_hash_var_fsm_create(SexpFsm_ptr self) 
{
  nusmv_assert(self->hash_var_fsm == (hash_ptr) NULL);

  self->hash_var_fsm = new_assoc();  
}


/**Function********************************************************************

  Synopsis           [Call to destroy the var fsm hash]

  Description        [Private method, used internally]

  SideEffects        []

******************************************************************************/
static void sexp_fsm_hash_var_fsm_destroy(SexpFsm_ptr self) 
{
  nusmv_assert(self->hash_var_fsm != (hash_ptr) NULL);

  clear_assoc_and_free_entries(self->hash_var_fsm, 
			       sexp_fsm_callback_var_fsm_free);
}


/**Function********************************************************************

  Synopsis [Private callback that destroys a single variable fsm
  contained into the var fsm hash]

  Description        []

  SideEffects        []

******************************************************************************/
static assoc_retval sexp_fsm_callback_var_fsm_free(char *key, 
						   char *data, char * arg)
{
  VarFsm_ptr varfsm = VAR_FSM(data);

  var_fsm_destroy(varfsm);  
  return ASSOC_DELETE;
}


/**Function********************************************************************

  Synopsis [Given a variable name, returns the corresponding variable
  fsm, or NULL if not found]

  Description        []

  SideEffects        []

******************************************************************************/
static VarFsm_ptr 
sexp_fsm_hash_var_fsm_lookup_var(SexpFsm_ptr self, node_ptr var) 
{
  nusmv_assert(self->hash_var_fsm != (hash_ptr) NULL);

  return VAR_FSM( find_assoc(self->hash_var_fsm, var) );
}


/**Function********************************************************************

  Synopsis           [Adds a var fsm to the internal hash. Private.]

  Description        []

  SideEffects        []

******************************************************************************/
static void 
sexp_fsm_hash_var_fsm_insert_var(SexpFsm_ptr self, 
				 node_ptr var, VarFsm_ptr varfsm)
{
  nusmv_assert(self->hash_var_fsm != (hash_ptr) NULL);
  
  insert_assoc(self->hash_var_fsm, var, varfsm);
}




/**Function********************************************************************

  Synopsis           [Creates a var fsm]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static VarFsm_ptr var_fsm_create(Expr_ptr init, 
				 Expr_ptr invar, 
				 Expr_ptr next)
{
  return VAR_FSM( cons(NODE_PTR(init), 
		       cons(NODE_PTR(invar), NODE_PTR(next))) );
}


/**Function********************************************************************

  Synopsis           [It does not destroy the init, trans and invar nodes. 
  It destroys only the support nodes]

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static void var_fsm_destroy(VarFsm_ptr self)
{
  node_ptr node = NODE_PTR(self);

  free_node(cdr(node));
  free_node(node);
}


/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static Expr_ptr var_fsm_get_init(VarFsm_ptr self)
{
  return EXPR( car(NODE_PTR(self)) );
}


/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static Expr_ptr var_fsm_get_invar(VarFsm_ptr self)
{
  return EXPR( car(cdr(NODE_PTR(self))) );
}


/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static Expr_ptr var_fsm_get_next(VarFsm_ptr self)
{
  return EXPR( cdr(cdr(NODE_PTR(self))) );
}


/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []

******************************************************************************/
static Expr_ptr var_fsm_get_input(VarFsm_ptr self)
{
  /* Currently no constraints over input are allowed, thus we return
     true to inidicate this. */
  return Expr_true();
}


