%{
/**CFile***********************************************************************

  FileName    [aba_grammar.y]

  PackageName [parser.aba]

  Synopsis [Grammar (for Yacc and Bison) of ABA specification input
  language]

  SeeAlso     [aba_input.l]

  Author      [Tommi Junttila]

  Copyright   [Copyright (C) 2006 by Tommi Junttila. Released under GNU LGPL]

******************************************************************************/

/*# include "pslExpr.h"*/

#include <stdarg.h>
#include <stdio.h>

#include "node/node.h"
#include "parser/symbols.h"
#include "parser/psl/pslNode.h"
#include "parser/psl/pslExpr.h"
#include "utils/utils.h"
#include "utils/assoc.h"
#include "utils/str_assoc.h"
#include "utils/array.h"
#include "src/compile/compile.h"
#include "bmc/sugar.h"
#include "bmc/sugarInt.h"

extern FILE* nusmv_stderr;

EXTERN int yylex ARGS((void));

EXTERN void aba_yyerror ARGS((char* s, ...));
EXTERN int aba_error ARGS((void));

Sugar_Automaton_t *aba_parsed_aba = 0; /* the returned value of parsing */
str_hash_ptr aba_parsed_aba_map_sugarap_to_node = 0;

static hash_ptr _map_node_to_sugarap = 0;
static str_hash_ptr _map_name_to_state = 0;
static str_hash_ptr _set_of_defined_states = 0;

static void _init();

static Sugar_Vertex_t* _define_state(char *name, const int accepting);
static Sugar_Vertex_t* _get_state(char *name);

typedef struct _aba_temp_pair
{
  array_t *dest_states;
  array_t *literals;
} _aba_temp_pair;
static _aba_temp_pair *_alloc_temp_pair();
static void _free_temp_pair(_aba_temp_pair *pair);
static Sugar_Vertex_t* _current_state;
static void _add_edge(Sugar_Vertex_t* source_state, _aba_temp_pair *pair);
static Sugar_Label_t *_get_label(char *name, const int is_negated);

%}

/*%debug*/
/* %locations creates a structure storing line and column plus other
   info. Useful for printing error messages. */
/*%locations*/
/*%defines*/


%union {
  /*node_ptr node; */
  int lineno;

  /* these are news */
  int ival;
  char* fval;
  char* baseval;
  char* idname;
  array_t *array;
  void *temp_pair;
  Sugar_Label_t *sugar_label;
};

%token TKEOF            "eof"
%token TKSTRING         "string"
%token TKERROR          
%token TKSTRUCT         "struct"

%token <ival> TKNUMBER         "number"
%token <fval> TKREALNUMBER     "real number"
%token <baseval> TKBASENUMBER  "based number"

%token TKTRUE TKFALSE

%token TKABA            "ABA"
%token TKSTATE TKACCEPTING TKINITIAL

%token TKINCONTEXT

%token <idname> TKATOM
%token <idname> TKSTATENAME

%token TKVUNIT          "vunit"
%token TKVMODE          "vmode"
%token TKVPROP          "vprop"
%token TKSTRONG         "strong"
%token TKDEFPARAM       "defparam"

%token TKINHERIT        "inherit"

%token TKDOT            "dot"
%token TKENDPOINT       "endpoint"
%token TKASSIGN         "assign"
%token TKFORCE          "force"
%token TKRELEASE        "release"
%token TKPROPERTY       "property"
%token TKSEQUENCE       "sequence"
%token TKMODULE         "module"
%token TKFUNCTION       "function"
%token TKRESTRICT       "restrict"
%token TKRESTRICT_GUARANTEE "restrict_guarantee"
%token TKFORALL         "forall"
%token TKASSERT         "assert"
%token TKASSUME         "assume"
%token TKASSUME_GUARANTEE "assume_guarantee"
%token TKCOVER          "cover"
%token TKBOOLEAN        "boolean"
%token TKCASE           "case"
%token TKCASEX          "casex"
%token TKCASEZ          "casez"
%token TKELSE           "else"
%token TKENDCASE        "esac"
%token TKIF             "if"

%token TKNONDET         "$nondet"
%token TKNONDET_VECTOR  "$nondet_vector"
%token TKNONDET_RANGE   "$nondet_range"
%token TKWNONDET        "$wnondet"

%token TKBASE           
%token TKDOTDOT         ".."
%token <operator> TKPIPEMINUSGT    "|->"
%token <operator> TKPIPEEQGT       "|=>"
%token TKIDENTIFIER     "identifier"
%token TKHIERARCHICALID 
%token TKLP             "("
%token TKRP             ")"
%token TKLC             "{"
%token TKRC             "}"
%token TKLB             "["
%left TKRB              "]"
%token TKCOMMA          ","

%token TKDIEZ           "#"

%token TKTRANS          "trans"
%token TKHINT           "hint"
%token TKTEST_PINS      "test_pins"


%token <operator> TKALWAYS         "always"
%token <operator> TKNEVER          "never"
%token <operator> TKEVENTUALLYBANG "eventually!"
%token <operator> TKWITHINBANG     "within!"
%token <operator> TKWITHIN         "within"
%token <operator> TKWITHINBANG_    "within!_"
%token <operator> TKWITHIN_        "within_"
%token <operator> TKWHILENOTBANG   "whilenot!"
%token <operator> TKWHILENOT       "whilenot"
%token <operator> TKWHILENOTBANG_  "whilenot!_"
%token <operator> TKWHILENOT_      "whilenot_"
%token <operator> TKNEXT_EVENT_ABANG "next_event_a!"
%token <operator> TKNEXT_EVENT_A   "next_event_a"
%token <operator> TKNEXT_EVENT_EBANG "next_event_e!"
%token <operator> TKNEXT_EVENT_E   "next_event_e"
%token <operator> TKNEXT_EVENTBANG "next_event!"
%token <operator> TKNEXT_EVENT     "next_event"
%token <operator> TKNEXT_ABANG     "next_a!"
%token <operator> TKNEXT_EBANG     "next_e!"
%token <operator> TKNEXT_A         "next_a"
%token <operator> TKNEXT_E         "next_e"
%token <operator> TKNEXTBANG       "next!"
%token <operator> TKNEXT           "next"
%token <operator> TKNEXTfunc       "next (function)"

%token <operator> TKBEFOREBANG     "before!"
%token <operator> TKBEFORE         "before"
%token <operator> TKBEFOREBANG_    "before!_"
%token <operator> TKBEFORE_        "before_"
%token <operator> TKUNTILBANG      "until!"
%token <operator> TKUNTIL          "until"
%token <operator> TKUNTILBANG_     "until!_"
%token <operator> TKUNTIL_         "until_"

%token TKABORT          "abort"

%token TKROSE           "rose"
%token TKFELL           "fell"
%token TKPREV           "prev"

%token <operator> TKG              "G"
%token <operator> TKXBANG          "X!"
%token <operator> TKX              "X"
%token <operator> TKF              "F"
%token <operator> TKU              "U"
%token <operator> TKW              "W"
%token <operator> TKEG             "EG"
%token <operator> TKEX             "EX"
%token <operator> TKEF             "EF"
%token <operator> TKAG             "AG"
%token <operator> TKAX             "AX"
%token <operator> TKAF             "AF"
%token <operator> TKA              "A"
%token <operator> TKE              "E"
%token <operator> TKIN             "in"
%token <operator> TKUNION          "union"
%token TKQUESTIONMARK   "?"
%token TKCOLON          ":"
%token TKSEMI           ";"
%token <operator> TKPIPEPIPE       "||"
%token <operator> TKAMPERSANDAMPERSAND "&&"
%token <operator> TKMINUSGT        "->"
%token <operator> TKLTMINUSGT      "<->"
%token <operator> TKPIPE           "|"
%token <operator> TKTILDEPIPE      "~|"
%token TKOR             "or"
%token TKPOSEDGE        "posedge"
%token TKNEGEDGE        "negedge"
%token <operator> TKCARET          "^"
%token <operator> TKXOR            "xor"
%token <operator> TKCARETTILDE     "^~"
%token <operator> TKTILDECARET     "~^"
%token <operator> TKAMPERSAND      "&"
%token <operator> TKTILDEAMPERSAND "~&"
%token <operator> TKEQEQ           "=="
%token <operator> TKBANGEQ         "!="
%token <operator> TKEQEQEQ         "==="
%token <operator> TKBANGEQEQ       "!=="
%token <operator> TKEQ             "="
%token <operator> TKGT             ">"
%token <operator> TKGE             ">="
%token <operator> TKLT             "<"
%token <operator> TKLE             "<="
%token <operator> TKLTLT           "<<"
%token <operator> TKGTGT           ">>"
%token <operator> TKGTGTGT         ">>>"
%token <operator> TKLTLTLT         "<<<"
%token <operator> TKPLUS           "+"
%token <operator> TKMINUS          "-"
%token <operator> TKSPLAT          "*"
%token <operator> TKSLASH          "/"
%token <operator> TKPERCENT        "mod"
%token <operator> TKSPLATSPLAT     "**"

%token <operator> TKBANG           "!"
%token <operator> TKTILDE          "~"
%token <operator> TKLBSPLAT        "[*"
%token <operator> TKLBEQ           "[="
%token <operator> TKLBMINUSGT      "[->"
%token <operator> TKLBPLUSRB       "[+]"
%token TKSTRUDLE        "@"


/* PRECEDENCE and ASSOCIATIVITY */
/* note:  priorities should be checked carefully for consistency with verilog,
   psl documentation
 */

%left TKALWAYS TKNEVER TKG

%right TKMINUSGT TKLTMINUSGT

%left TKEVENTUALLYBANG TKWITHINBANG TKWITHIN TKWITHINBANG_
 TKWITHIN_ TKWHILENOTBANG TKWHILENOT TKWHILENOTBANG_ TKWHILENOT_
 TKNEXT_EVENT_ABANG TKNEXT_EVENT_A TKNEXT_EVENT_EBANG TKNEXT_EVENT_E
  TKNEXT_EVENTBANG TKNEXT_EVENT
 TKNEXT_ABANG TKNEXT_EBANG TKNEXT_A TKNEXT_E
 TKNEXTBANG TKNEXT TKNEXTfunc

%left TKBEFOREBANG TKBEFORE TKBEFOREBANG_ TKBEFORE_
%left TKUNTILBANG TKUNTIL TKUNTILBANG_ TKUNTIL_
%left TKABORT

%left TKROSE TKFELL TKPREV

%left TKXBANG TKX TKF TKU TKW TKEG TKEX TKEF TKAG TKAX TKAF
 TKA TKE
%left TKIN TKUNION
%right TKQUESTIONMARK TKCOLON TKSEMI
%left TKPIPEPIPE
%left TKAMPERSANDAMPERSAND
%left TKPIPE TKTILDEPIPE TKOR
%left TKPOSEDGE TKNEGEDGE
%left TKCARET TKXOR TKCARETTILDE TKTILDECARET
%left TKAMPERSAND TKTILDEAMPERSAND
%left TKEQEQ TKBANGEQ TKEQEQEQ TKBANGEQEQ TKEQ
%left TKGT TKGE TKLT TKLE
%left TKLTLT TKGTGT
%left TKGTGTGT TKLTLTLT
%left TKPLUS TKMINUS
%left TKSPLAT TKSLASH TKPERCENT TKSPLATSPLAT


%nonassoc TKBANG TKTILDE TKLBSPLAT TKLBEQ TKLBMINUSGT
%nonassoc TKLBPLUSRB
%nonassoc TKSTRUDLE

/*
%type <operator> AndOrOp

%type <psl_expr> Property
%type <psl_expr> Additional_Binary_Operators FL_Property OBE_Property
%type <psl_expr> Opt_IndexRange IndexRange ValueSet ValueRange_List ValueRange Range 

%type <psl_expr> Primary Unary_Expression Binary_Expression Conditional_Expression 
%type <psl_expr> Case_Expression case_list
%type <psl_expr> Sequence Sequence_or_Expression Sere

%type <psl_expr> Expression  Mintypmax_Expression  

%type <psl_expr> Opt_Count  Count  LowBound  HighBound
%type <psl_expr> boolean number 
%type <psl_expr> hierarchical_identifier 
%type <psl_expr> PslSpec
*/

%type <array> StateConjunction
%type <temp_pair> Conjunction
%type <sugar_label> Literal

%start ABASpec

%%

ABASpec: 
TKABA {_init(); } TKLC State_List Initial TKRC {YYACCEPT;}
;

Initial :
TKINITIAL TKLC StateConjunction_List TKRC TKSEMI { }
;

State_List :
State_List State { }
| State { }
;

State:
TKSTATE TKSTATENAME {_current_state = _define_state($2, 0); } TKLC Conjunction_List TKRC TKSEMI {}
| TKSTATE TKSTATENAME TKACCEPTING  {_current_state = _define_state($2, 1); } TKLC Conjunction_List TKRC TKSEMI {}
;

Conjunction_List :
/* empty */ {}
| Conjunction { _add_edge(_current_state, $1); }
| Conjunction TKCOMMA Conjunction_ListT { _add_edge(_current_state, $1); }
;

Conjunction_ListT :
Conjunction_ListT TKCOMMA Conjunction { _add_edge(_current_state, $3); }
| Conjunction { _add_edge(_current_state, $1); }
;

Conjunction :
Conjunction TKAMPERSAND Literal
{
  $$ = $1;
  array_insert_last(Sugar_Label_t*, ((_aba_temp_pair*)$$)->literals, $3);
}
| Conjunction TKAMPERSAND TKSTATENAME
{
  $$ = $1;
  array_insert_last(Sugar_Vertex_t*, ((_aba_temp_pair*)$$)->dest_states,
		    _get_state($3));
}
| TKSTATENAME
{
  _aba_temp_pair *pair = _alloc_temp_pair();
  array_insert_last(Sugar_Vertex_t*, pair->dest_states, _get_state($1));
  $$ = (void*)pair;
}
| Literal
{
  _aba_temp_pair *pair = _alloc_temp_pair();
  array_insert_last(Sugar_Label_t*, pair->literals, $1);
  $$ = (void*)pair;
}
;

Literal :
TKATOM {$$ = _get_label($1, 0); }
| TKTILDE TKATOM {$$ = _get_label($2, 1); }
;

StateConjunction_List :
StateConjunction_List TKCOMMA StateConjunction
{ array_insert_last(array_t*, aba_parsed_aba->initial_states, $3); }
| StateConjunction
{ array_insert_last(array_t*, aba_parsed_aba->initial_states, $1); }
;

StateConjunction :
StateConjunction TKAMPERSAND TKSTATENAME
{$$ = $1; array_insert_last(Sugar_Vertex_t*, $$, _get_state($3)); }
| TKSTATENAME
{$$ = array_alloc(Sugar_Vertex_t*, 1); array_insert_last(Sugar_Vertex_t*, $$, _get_state($1)); }
;

%%
extern int yylineno;
extern int aba_yylineno;

void aba_yyerror(char* s, ...)
{
  va_list args;

  va_start(args, s);
  fprintf(nusmv_stderr, "ABA parse error line %d(%d): ", 
	  yylineno+aba_yylineno-1,aba_yylineno);
  vfprintf(nusmv_stderr, s, args);
  fprintf(nusmv_stderr, "\n");
  va_end(args);
}

int aba_error()
{
  fprintf(nusmv_stderr, "ABA parse error line %d: \n", 
	  yylineno+aba_yylineno-1);
  return 1;
}

static void _init()
{
  aba_parsed_aba = Sugar_AutomatonCreate();
  assert(aba_parsed_aba);

  if(_map_node_to_sugarap)
    free_assoc(_map_node_to_sugarap);
  _map_node_to_sugarap = new_assoc();

  if(aba_parsed_aba_map_sugarap_to_node)
    free_str_assoc(aba_parsed_aba_map_sugarap_to_node);
  aba_parsed_aba_map_sugarap_to_node = new_str_assoc();

  if(_map_name_to_state)
    free_str_assoc(_map_name_to_state);
  _map_name_to_state = new_str_assoc();

  if(_set_of_defined_states)
     free_str_assoc(_set_of_defined_states);
  _set_of_defined_states = new_str_assoc();
}

static Sugar_Vertex_t* _define_state(char *name, const int accepting)
{
  Sugar_Vertex_t *state = 0;
  assert(aba_parsed_aba);
  state = (Sugar_Vertex_t *)find_str_assoc(_set_of_defined_states, name);
  if(state)
    {
      fprintf(nusmv_stderr, "State '%s' defined twice\n", name);
      exit(1);
      FREE(name);
      return state;
    }
  state = _get_state(name);
  assert(state);
  insert_str_assoc(_set_of_defined_states, name, (node_ptr)state);
  if(accepting)
    array_insert_last(Sugar_Vertex_t*, aba_parsed_aba->accept_states, state);

  return state;
}

static Sugar_Vertex_t* _get_state(char *name)
{
  Sugar_Vertex_t *state = 0;
  assert(aba_parsed_aba);
  state = (Sugar_Vertex_t *)find_str_assoc(_map_name_to_state, name);
  if(state)
    {
      FREE(name);
      return state;
    }
  state = SugarAutomatonNewNode(aba_parsed_aba);
  insert_str_assoc(_map_name_to_state, name, (node_ptr)state);

  return state;
}

static _aba_temp_pair *_alloc_temp_pair()
{
  _aba_temp_pair *pair = ALLOC(_aba_temp_pair, 1);
  assert(pair);
  pair->dest_states = array_alloc(Sugar_Vertex_t*, 1);
  pair->literals = array_alloc(Sugar_Label_t*, 1);
  return pair;
}

static void _free_temp_pair(_aba_temp_pair *pair)
{
  assert(pair);
  array_free(pair->dest_states);
  array_free(pair->literals);
  FREE(pair);
}

static void _add_edge(Sugar_Vertex_t* source_state, _aba_temp_pair *pair)
{
  assert(source_state);
  assert(pair);
  SugarAutomatonNewEdge(aba_parsed_aba, source_state, pair->dest_states, pair->literals, 1);
  _free_temp_pair(pair);
}

static Sugar_Label_t *_get_label(char *name, const int is_negated)
{
  Sugar_Label_t *label = 0;
  PslNode_ptr node = psl_expr_make_atom(name).psl_node;
  char *sugarap = (char*)find_assoc(_map_node_to_sugarap, node);
  if(!sugarap)
    {
      sugarap = util_strsav(name);
      lsNewEnd(aba_parsed_aba->prop_list, (lsGeneric)sugarap, 0);
      insert_assoc(_map_node_to_sugarap, node, (node_ptr)sugarap);
      insert_str_assoc(aba_parsed_aba_map_sugarap_to_node, sugarap, node);
    }
  FREE(name);
  label = ALLOC(Sugar_Label_t, 1);
  label->name = sugarap;
  label->isNegate = is_negated;
  return label;
}
