/**CFile***********************************************************************

  FileName    [smMisc.c]

  PackageName [sm]

  Synopsis    [This file contain the main routine for the batch use of NuSMV.]

  Description [This file contain the main routine for the batch use of
  NuSMV. The batch main executes the various model checking steps in a
  predefined order. After the processing of the input file than it
  return to the calling shell.]

  SeeAlso     []

  Author      [Adapted to NuSMV by Marco Roveri]

  Copyright   [
  This file is part of the ``sm'' package of NuSMV version 2. 
  Copyright (C) 1998-2001 by CMU and ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "smInt.h"
#include "set/set.h"
#include "fsm/bdd/BddFsm.h"

#include "utils/error.h"
#include "mc/mc.h"
#include "enc/enc.h"
#include "bmc/bmcUtils.h"
#include "compile/compileInt.h" /* for CompileFlatten_init_flattener */

static char rcsid[] UTIL_UNUSED = "$Id: smMisc.c,v 1.1.1.1 2005/12/20 17:07:11 timo Exp $";

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [The batch main.]

  Description [The batch main. It first read the input file, than
  flatten the hierarchy. Aftre this preliminar phase it creates the
  boolean variables necessary for the encoding and then start
  compiling the read model into BDD. Now computes the reachable states
  depending if the flag has been set. before starting verifying if the
  properties specified in the model hold or not it computes the
  fairness constraints. You can also activate the reordering and
  also choose to avoid the verification of the properties.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void smBatchMain()
{
  SymbTable_ptr st;

  /* Necessary to have standard behavior in the batch mode */
  util_resetlongjmp();

  /* Parse the input file */
  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "Parsing file %s ...", get_input_file(options));
    fflush(nusmv_stderr);
  }

  /* ================================================== */
  /*   1: Read the model                                */
  /* ================================================== */
  if (Parser_ReadSMVFromFile(get_input_file(options))) nusmv_exit(1);

  if (opt_verbose_level_gt(options, 0)) {
    fprintf(nusmv_stderr, "...done.\n");
    fflush(nusmv_stderr);
  }
 

  /* ================================================== */
  /*  2: Flatten hierarchy                              */
  /* ================================================== */
  {
    SymbLayer_ptr layer; 
    int propErr;

    /* Initializes the flattener, that must be initialized *after* the
       parsing phase */
    CompileFlatten_init_flattener(); 
    nusmv_assert(FLAT_HIERARCHY(NULL) == mainFlatHierarchy);

    st = Compile_get_global_symb_table();
    layer = SymbTable_create_layer(st, MODEL_LAYER_NAME, 
				   SYMB_LAYER_POS_BOTTOM);
  
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Flattening hierarchy...\n");
    }
  
    /* Processing of the parse tree and constructions of all the
      expressions for the state machine(s). Here the expansions are
      performed so that modules and processes are created. The
      expansion of modules is such that the formal parameters (if any)
      are replaced by the actual ones and the machine is replicated. */
    mainFlatHierarchy = Compile_FlattenHierarchy(st, layer, 
						 sym_intern("main"), Nil,
						 Nil, true);
    
    /* We store properties in the DB of properties */
    propErr = PropDb_fill(st, FlatHierarchy_get_spec(mainFlatHierarchy),
			  FlatHierarchy_get_compute(mainFlatHierarchy),
			  FlatHierarchy_get_ltlspec(mainFlatHierarchy),
			  FlatHierarchy_get_pslspec(mainFlatHierarchy),
			  FlatHierarchy_get_invarspec(mainFlatHierarchy));
    
    if (0 != propErr) return; /* error */
    
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "...done.\n");
    }
  } /* end of flattening */

  /* If the -lp option is used, list the properties and exit */
  if (opt_list_properties(options) == true) {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Listing the properties in the database...\n");
    }
    PropDb_print_all(nusmv_stdout);

    return;
  }


  /* ================================================== */
  /*  3: Builds the encodings                           */
  /* ================================================== */
  {
    BoolEnc_ptr bool_enc;
    BddEnc_ptr bdd_enc;

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Encoding variables ...");
    }

    /* creates the bool encoding, and commit the model layer, that was
       created during the flattening phase */
    Enc_init_bool_encoding();
    bool_enc = Enc_get_bool_encoding();
    BaseEnc_commit_layer(BASE_ENC(bool_enc), MODEL_LAYER_NAME);

    /* creates the bdd encoding, and again commit the model layer */
    Enc_init_bdd_encoding();
    bdd_enc = Enc_get_bdd_encoding();
    BaseEnc_commit_layer(BASE_ENC(bdd_enc), MODEL_LAYER_NAME);

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "... done.\n");
    }
  } /* end of encoding */
 

  /* Checking for the correctness of the state machines which have
     been declared. It should not mention DD manipulation. */
  Compile_InitializeBuildModel(st);


  /* ================================================== */
  /*  4: Builds the flat and bool FSMs                  */
  /* ================================================== */
  {
    SexpFsm_ptr scalar_fsm;
    SexpFsm_ptr bool_fsm;
    SymbLayer_ptr det_layer; 
    SymbLayer_ptr model_layer;

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Building Model ...");
    }

    /* ------------------------------------------------------------ */
    /*   Builds the sexp FSM of the whole read model                */
    /* ------------------------------------------------------------ */
    model_layer = SymbTable_get_layer(st, MODEL_LAYER_NAME);
    scalar_fsm = FsmBuilder_create_sexp_fsm(global_fsm_builder, 
		    Enc_get_bdd_encoding(), 
		    SYMB_LAYER(NULL), /* scalar fsm */
		    NodeList_to_node_ptr(SymbLayer_get_all_vars(model_layer)));

    Prop_master_set_scalar_sexp_fsm(scalar_fsm);

    /* ------------------------------------------------------------ */
    /*   Writes the flat scalar model                               */
    /* ------------------------------------------------------------ */
    if (get_output_flatten_model_file(options) != NIL(char)) {
      const char* layers[] = { MODEL_LAYER_NAME, NULL };
      FILE* ofileid = fopen(get_output_flatten_model_file(options), "w");

      if (ofileid == (FILE*) NULL) {
	fprintf(nusmv_stderr, "Unable to open file \"%s\".\n",
		get_output_flatten_model_file(options));
	nusmv_exit(1);
      }    
      if (opt_verbose_level_gt(options, 0)) {
	fprintf(nusmv_stderr, "Writing whole flat model into file \"%s\"...",
		get_output_flatten_model_file(options));
      }

      Compile_WriteFlatten(st, layers, ofileid, mainFlatHierarchy);

      if (opt_verbose_level_gt(options, 0)) fprintf(nusmv_stderr, "... done.\n");
      fflush(ofileid);
      fclose(ofileid);
    }

    /* ------------------------------------------------------------ */
    /*   Builds the boolean sexp FSM of the whole read model        */
    /* ------------------------------------------------------------ */
    
    /* first creates a determinizatio layer, makes it filled while
       creating the bool fsm, and then commits it to the encodings */
    det_layer = SymbTable_create_layer(st, DETERM_LAYER_NAME, 
				       SYMB_LAYER_POS_BOTTOM);

    bool_fsm = SexpFsm_scalar_to_boolean(scalar_fsm, det_layer);
    Prop_master_set_bool_sexp_fsm(bool_fsm);
    
    BaseEnc_commit_layer(BASE_ENC(Enc_get_bool_encoding()), 
			 DETERM_LAYER_NAME);
    BaseEnc_commit_layer(BASE_ENC(Enc_get_bdd_encoding()), 
			 DETERM_LAYER_NAME);
    
    /* ------------------------------------------------------------ */
    /*   Writes the flat boolean model (if requested)               */
    /* ------------------------------------------------------------ */
    if (get_output_boolean_model_file(options) != NIL(char)) {
      FILE* ofileid;

      ofileid = fopen(get_output_boolean_model_file(options), "w");
      if (ofileid == (FILE*)NULL) {
	fprintf(nusmv_stderr, "Unable to open file \"%s\".\n",
		get_output_flatten_model_file(options));
	nusmv_exit(1);
      }
      if (opt_verbose_level_gt(options, 0)) {
	fprintf(nusmv_stderr, "Writing whole boolean model into file \"%s\"...",
		get_output_boolean_model_file(options));
      }

      Compile_WriteFlattenBool(Enc_get_bdd_encoding(), bool_fsm, ofileid, 
			       mainFlatHierarchy);
    
      if (opt_verbose_level_gt(options, 0)) fprintf(nusmv_stderr, "... done.\n");
      fflush(ofileid);
      fclose(ofileid);
    }

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "... done.\n");
    }
  } /* end of build model */


  /* ================================================== */
  /*  5.1 BMC starts                                    */
  /* ================================================== */
  if (opt_bmc_mode(options) == true) {
    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Entering BMC mode...\n");
    }
    
    /* Initializes the bmc package, and commits both the model and the
       determinization layers: */
    Bmc_Init();

    if (get_prop_no(options) != -1) {
      int prop_no = get_prop_no(options);
      Prop_ptr prop;

      if (opt_verbose_level_gt(options, 0)) {
        fprintf(nusmv_stderr, "Verifying property %d...\n", prop_no);
      }
      
      if ( (prop_no < 0) || (prop_no >= PropDb_get_size()) ) {
        fprintf(nusmv_stderr, 
                "Error: \"%d\" is not a valid property index\n", prop_no);
        nusmv_exit(1);
      }

      prop = PropDb_get_prop_at_index(prop_no);
      
      switch (Prop_get_type(prop)) {
      case Prop_Ltl:
	{
	  int rel_loop; 
	  
	  /* skip if -ils option is given */
	  if (opt_ignore_ltlspec(options)) break;

	  rel_loop = Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options), 
						     NULL);
	  Bmc_GenSolveLtl(prop, get_bmc_pb_length(options), rel_loop, 
			  /*increasing length*/ TRUE , 
			  TRUE, BMC_DUMP_NONE, NULL);
	  break;
	} 
	
      case Prop_Psl:
	{
	  int rel_loop = Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options), 
							 NULL);
          /* skip if -ips option is given */
	  if (opt_ignore_pslspec(options)) break;

	  Bmc_check_psl_property(prop, false, false, false, 
				 get_bmc_pb_length(options), rel_loop);
	  break;
	}
	
      case Prop_Invar: 
	/* skip if -ii option is given */
	if (opt_ignore_invar(options)) break;
	
        Bmc_GenSolveInvar(prop, TRUE, BMC_DUMP_NONE, NULL);
	break;
	
      default:
        fprintf(nusmv_stderr, 
                "Error: only LTL and INVAR properties can be checked in BMC mode\n");
        nusmv_exit(1);
      } /* switch on type */

    }
    else {
      /* Checks all ltlspecs, invarspecs and pslspecs */
      
      if (! opt_ignore_ltlspec(options)) {
	lsList props; 
	lsGen  iterator; 
	Prop_ptr prop;
	int rel_loop;

	if (opt_verbose_level_gt(options, 0)) {
	  fprintf(nusmv_stderr, "Verifying the LTL properties...\n");
	}

	
	props = PropDb_get_props_of_type(Prop_Ltl);	  
        nusmv_assert(props != LS_NIL);
	
        lsForEachItem(props, iterator, prop) {
          rel_loop = Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options), 
						     NULL);

          Bmc_GenSolveLtl(prop, get_bmc_pb_length(options), rel_loop, 
                          /*increasing length*/ TRUE, TRUE, BMC_DUMP_NONE, NULL);
        }

        lsDestroy(props, NULL); /* the list is no longer needed */
      }

      if (! opt_ignore_pslspec(options)) {
	lsList props; 
	lsGen  iterator; 
	Prop_ptr prop;
	int rel_loop = 
	  Bmc_Utils_ConvertLoopFromString(get_bmc_pb_loop(options), NULL);

	if (opt_verbose_level_gt(options, 0)) {
	  fprintf(nusmv_stderr, "Verifying the PSL properties...\n");
	}
	
	props = PropDb_get_props_of_type(Prop_Psl);	  
        nusmv_assert(props != LS_NIL);
	
        lsForEachItem(props, iterator, prop) {
	  if (Prop_is_psl_ltl(prop)) {
	    Bmc_check_psl_property(prop, false, false, false, 
				   get_bmc_pb_length(options), rel_loop);
	  }
        }

        lsDestroy(props, NULL); /* the list is no longer needed */
      }

      if (! opt_ignore_invar(options)) {
        lsList props;
        lsGen  iterator; 
        Prop_ptr prop;
	
	if (opt_verbose_level_gt(options, 0)) {
	  fprintf(nusmv_stderr, "Verifying the INVAR properties...\n");
	}
      
	props = PropDb_get_props_of_type(Prop_Invar);
        nusmv_assert(props != LS_NIL);
        
        lsForEachItem(props, iterator, prop) {
          Bmc_GenSolveInvar(prop, TRUE, BMC_DUMP_NONE, NULL);
        }

        lsDestroy(props, NULL); /* the list is no longer needed */
      }
    }

    return;
  } /* end of BMC */


  /* ================================================== */
  /*  5.2 BDD-based model checking starts               */
  /* ================================================== */
  
  /* Builds the BDD FSM of the whole read model if COI is disabled */
  if (opt_cone_of_influence(options) == false) {
    BddFsm_ptr bdd_fsm;

    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "Creating the BDD fsm... ");
    }

    bdd_fsm = FsmBuilder_create_bdd_fsm(global_fsm_builder, 
					Enc_get_bdd_encoding(), 
					Prop_master_get_scalar_sexp_fsm(), 
					get_partition_method(options));    
    Prop_master_set_bdd_fsm(bdd_fsm);
    
    if (opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, " done.\n");
    }
  }
  
  /* Reachability */
  if (opt_cone_of_influence(options) == false) {
    /* Builds the OBDD representing Reachable States Set */
    if (opt_forward_search(options)) {
      
      if (opt_verbose_level_gt(options, 0)) {
        fprintf(nusmv_stderr,"Setting reachable states on...");
      }

      set_use_reachable_states(options);
      
      if (opt_verbose_level_gt(options, 0)) fprintf(nusmv_stderr,"...done\n");
    }
  }
  
  /* checks the fsm if required */
  if (opt_check_fsm(options) == true) {
    if (opt_cone_of_influence(options) == false) {
      BddFsm_check_machine(Prop_master_get_bdd_fsm());
    }
    else {
      fprintf(nusmv_stderr, "WARNING: Check for totality of the transition relation cannot currently\n");
      fprintf(nusmv_stderr, "performed in batch mode if the cone of influence reduction has been enabled.\n");
    }
  }

  /* Ordering file dump and exit only if ordering is disabled: */
  if ( !opt_reorder(options) 
       && !is_default_order_file(options)
       && !util_is_string_null(get_output_order_file(options)))  {
    VarOrderingType dump_type;
    if (opt_write_order_dumps_bits(options)) dump_type = DUMP_BITS;
    else dump_type = DUMP_DEFAULT;

    BddEnc_write_var_ordering(Enc_get_bdd_encoding(), 
			      get_output_order_file(options), 
			      dump_type);
    return;
  }

  if (get_prop_no(options) != -1) {
    int prop_no = get_prop_no(options);

    if (opt_verbose_level_gt(options, 0)) {
      fprintf(nusmv_stderr, "Verifying property %d...\n", prop_no);
    }
      
    if ( (prop_no < 0) || (prop_no >= PropDb_get_size()) ) {
      fprintf(nusmv_stderr, 
              "Error: \"%d\" is not a valid property index\n", prop_no);
      nusmv_exit(1);
    }

    PropDb_verify_prop_at_index(prop_no);
  }
  else {
    /* Evaluates the Specifications */
    if (!opt_ignore_spec(options)) {
      PropDb_verify_all_type(Prop_Ctl);
    }

    if (!opt_ignore_compute(options)) {
      PropDb_verify_all_type(Prop_Compute);
    }

    /* Evaluates the LTL specifications */
    if (!opt_ignore_ltlspec(options)) {
      PropDb_verify_all_type(Prop_Ltl);
    }

    /* Evaluates the PSL specifications */
    if (!opt_ignore_pslspec(options)) {
      PropDb_verify_all_type(Prop_Psl);
    }

    /* Evaluates CHECKINVARIANTS */
    if (!opt_ignore_invar(options)) {
      PropDb_verify_all_type(Prop_Invar);
    }

  }

  /* Reporting of statistical information. */
  if (opt_verbose_level_gt(options, 0)) {  
    print_usage(nusmv_stdout);
  }

  /* Computing and Reporting of the Effect of Reordering */
  if (opt_reorder(options)) {
    fprintf(nusmv_stdout, "\n========= starting reordering ============\n");
    dd_reorder(dd_manager, get_reorder_method(options), DEFAULT_MINSIZE);
    fprintf(nusmv_stdout, "\n========= after reordering ============\n");
    if (opt_verbose_level_gt(options, 0)) {  
      print_usage(nusmv_stdout);
    }
  }
  
  /* Ordering file dump: */
  if ( opt_reorder(options) 
       || (!is_default_order_file(options) 
	   && !util_is_string_null(get_output_order_file(options))) ) {
    VarOrderingType dump_type;
    if (opt_write_order_dumps_bits(options)) dump_type = DUMP_BITS;
    else dump_type = DUMP_DEFAULT;

    BddEnc_write_var_ordering(Enc_get_bdd_encoding(), 
			      get_output_order_file(options), 
			      dump_type);
  }

  if (opt_print_reachable(options) == true) {
    /* Reporting of Reachable States */
    if (opt_cone_of_influence(options) == false) {
      BddFsm_print_reachable_states_info(Prop_master_get_bdd_fsm(), 
					 false, /* do not print states */
					 nusmv_stdout);
    }
    else {
      fprintf(nusmv_stderr, "WARNING: Statistics of reachable states is not currently available\n");
      fprintf(nusmv_stderr, "in batch mode if cone of influence reduction has been enabled.\n");
    }
  }
}

/**Function********************************************************************

  Synopsis           [Prints usage statistic.]

  Description        [Prints on <code>nusmv_stdout</code> usage
  statistics, i.e. the amount of memory used, the amount of time
  spent, the number of BDD nodes allocated and the size of the model
  in BDD.]

  SideEffects        []

  SeeAlso            [compilePrintBddModelStatistic]
******************************************************************************/
void print_usage(FILE * file)
{
  fprintf(nusmv_stdout, "######################################################################\n");
  util_print_cpu_stats(file);
  fprintf(file, "######################################################################\n");
  fprintf(file, "BDD statistics\n");
  fprintf(file, "--------------------\n");
  fprintf(file, "BDD nodes allocated: %d\n", get_dd_nodes_allocated(dd_manager));
  fprintf(file, "--------------------\n");
  if (opt_cone_of_influence(options) == false) {
    BddFsm_ptr bdd_fsm = Prop_master_get_bdd_fsm();
    if (bdd_fsm != BDD_FSM(NULL)) {
      BddFsm_print_info(bdd_fsm, file);    
    }
  }
  else {
    fprintf(nusmv_stderr, "WARNING: Model Statistics is not currently available\n");
    fprintf(nusmv_stderr, "if cone of influence reduction has been enabled.\n");
  }
}



/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []
******************************************************************************/
void restore_nusmv_stdout()
{
  if (nusmv_stdout != def_nusmv_stdout) {
    fclose(nusmv_stdout);
    nusmv_stdout = def_nusmv_stdout;
  }
}



/**Function********************************************************************

  Synopsis           []

  Description        []

  SideEffects        []

  SeeAlso            []
******************************************************************************/
void restore_nusmv_stderr()
{
  if (nusmv_stderr != def_nusmv_stderr) {
    fclose(nusmv_stderr);
    nusmv_stderr = def_nusmv_stderr;
  }  
}

