/*
 * The BMC algorithms from Heljanko,Junttila&Latvala CAV 2005 paper
 * Copyright 2005 by Tommi Junttila
 * Released under GNU Lesser General Public License (LGPL)
 */


#include "bmcTJ.h"

#include "bmcBmc.h"
#include "bmcInt.h"
#include "bmcGen.h"
#include "bmcTableau.h"
#include "bmcConv.h"
#include "bmcDump.h"
#include "bmcModel.h"
#include "bmcWff.h"
#include "bmcUtils.h"
#include "bmcTJIncrSat.h"
#include "bmcSatTrace.h"
#include "trace/Trace.h"
#include "trace/TraceManager.h"

#include "dag/dag.h"

#include "node/node.h"
#include "be/be.h"
#include "mc/mc.h" /* for print_spec */

#include "sat/sat.h" /* for solver and result */
#include "sat/SatSolver.h"
#include "sat/SatIncSolver.h"

#include "compile/symb_table/SymbType.h"
#include "compile/symb_table/SymbLayer.h"

#include "enc/enc.h"
#include "enc/be/BeEnc.h"
#include "parser/symbols.h" /* for ATOM */
#include "utils/ustring.h"
#include "utils/assoc.h"

EXTERN node_ptr boolean_range;

static const char* _SNH_text = "%s:%d: Should not happen";
static const char* _SNYI_text = "%s:%d: Something not yet implemented\n";

#ifdef BENCHMARKING
#include <unistd.h>
#include <sys/times.h>
struct tms time_bmc_start;
struct tms time_gen_start;
struct tms time_solver_start;
struct tms time_temp;
static double time_diff(struct tms *start, struct tms *stop)
{
  return ((((double)stop->tms_utime + (double)stop->tms_stime) -
	   ((double)start->tms_utime + (double)start->tms_stime)) /
	  (double)CLK_TCK);
}
#endif



/*
 * Predefinition for a modified cex printer
 */
static Trace_ptr _generate_and_print_cntexample(BeEnc_ptr be_enc, 
						Bmc_IncrSat_ptr solver, 
						BddEnc_ptr bdd_enc, 
						const int k, 
						const char* trace_name,
						be_ptr l_var);





/*
 * Auxiliary wrapper stuff needed to use both the standard and my incremental
 * SAT solver interfaces
 */
typedef struct _MetaSolver_TAG
{
  BeEnc_ptr        be_enc;

  int              using_volatile_group;

  Bmc_IncrSat_ptr  TJ_solver;

  SatIncSolver_ptr S_solver;
  SatSolverGroup   S_permanent_group;
  SatSolverGroup   S_volatile_group;
} _MetaSolver;

static _MetaSolver *_MS_create(BeEnc_ptr be_enc,
			       const int use_TJ_solver)
{
  assert(be_enc);

  _MetaSolver *ms = ALLOC(_MetaSolver, 1);
  ms->be_enc = be_enc;
  ms->using_volatile_group = 0;
  ms->TJ_solver = 0;
  ms->S_solver = 0;
  if(use_TJ_solver)
    {
      ms->TJ_solver = Bmc_IncrSat_init(get_sat_solver(options));
      if(ms->TJ_solver == 0)
	{
	  fprintf(nusmv_stderr,
		  "Incremental sat solver '%s' is not available.\n", 
		  get_sat_solver(options));
	  FREE(ms);
	  return 0;
	}
    }
  else
    {
      ms->S_solver = Sat_CreateIncSolver(get_sat_solver(options));
      if(ms->S_solver == SAT_INC_SOLVER(NULL)) {
	fprintf(nusmv_stderr,
		"Incremental sat solver '%s' is not available.\n", 
		get_sat_solver(options));
	FREE(ms);
	return 0;
      }
      ms->S_permanent_group =
	SatSolver_get_permanent_group(SAT_SOLVER(ms->S_solver));
    }
  return ms;
}
static void _MS_destroy(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_release(ms->TJ_solver); ms->TJ_solver = 0;
    }
  else
    {
      SatIncSolver_destroy(ms->S_solver); ms->S_solver = 0;
    }
}
static void _MS_goto_permanent_group(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  assert(ms->using_volatile_group);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_pop_frame(ms->TJ_solver);
    }
  else
    {
      SatIncSolver_destroy_group(ms->S_solver, ms->S_volatile_group);
    }
  ms->using_volatile_group = 0;
}
static void _MS_goto_volatile_group(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);
  assert(!ms->using_volatile_group);
  if(ms->TJ_solver)
    {
      Bmc_IncrSat_new_frame(ms->TJ_solver);
    }
  else
    {
      ms->S_volatile_group = SatIncSolver_create_group(ms->S_solver);
    }
  ms->using_volatile_group = 1;
}
static void _MS_force_true(_MetaSolver *ms,
			   be_ptr be_constraint)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);

  if(ms->TJ_solver)
    {
      Bmc_IncrSat_insert(ms->TJ_solver, be_constraint);
      Bmc_IncrSat_force_true(ms->TJ_solver, be_constraint);
    }
  else
    {
      SatSolver_ptr solver = SAT_SOLVER(ms->S_solver);
      Be_Cnf_ptr cnf = Be_ConvertToCnf(BeEnc_get_be_manager(ms->be_enc),
				       be_constraint);
      if(ms->using_volatile_group)
	{
	  SatSolver_add(solver, cnf, ms->S_volatile_group);
	  SatSolver_set_polarity(solver, cnf, 1, ms->S_volatile_group);
	}
      else
	{
	  SatSolver_add(solver, cnf, ms->S_permanent_group);
	  SatSolver_set_polarity(solver, cnf, 1, ms->S_permanent_group);
	}
      Be_Cnf_Delete(cnf);
    }
}
static void _MS_force_constraint_list(_MetaSolver *ms, lsList constraints)
{
  lsGen  iterator;
  be_ptr be_constraint;
  
  lsForEachItem(constraints, iterator, be_constraint)
    _MS_force_true(ms, be_constraint);
}
static Bmc_IncrSat_result _MS_solve(_MetaSolver *ms)
{
  assert(ms);
  assert(ms->be_enc);
  assert(!(ms->TJ_solver && ms->S_solver));
  assert(ms->TJ_solver || ms->S_solver);

  if(ms->TJ_solver)
    {
      return Bmc_IncrSat_solve(ms->TJ_solver);
    }
  else
    {
      return SatSolver_solve_all_groups(SAT_SOLVER(ms->S_solver));
    }
}





/*
 *
 * Auxiliary data structure code for a set of node_ptr
 *
 */
static hash_ptr _set_create();
static void     _set_destroy(hash_ptr hash);
static void     _set_insert(hash_ptr hash, node_ptr bexp);
static int      _set_is_in(hash_ptr hash, node_ptr bexp);

static hash_ptr _set_create()
{
  hash_ptr hash = new_assoc();
  nusmv_assert(hash != (hash_ptr) NULL);
}

static void _set_destroy(hash_ptr hash)
{
  if(hash != (hash_ptr) NULL)
    free_assoc(hash);
}

static void _set_insert(hash_ptr hash, node_ptr bexp)
{
  nusmv_assert(hash != (hash_ptr) NULL);
  nusmv_assert((node_ptr)1 != (node_ptr)NULL);
  insert_assoc(hash, bexp, (node_ptr)1);
}

static int _set_is_in(hash_ptr hash, node_ptr bexp)
{
  node_ptr result;
  nusmv_assert(hash != (hash_ptr) NULL);
  result = find_assoc(hash, bexp);
  if(result == (node_ptr)NULL)
    return 0;
  return 1;
}





/*
 * Auxiliary node printing routine
 */
static void _my_print_node(FILE *out,
			   const char *prefix,
			   node_ptr node,
			   const char *postfix)
{
  if(prefix) fprintf(out, "%s", prefix);
  print_node(out, node);
  if(postfix) fprintf(out, "%s", postfix);
}





/*
 * Auxiliary routine: print the list of node_ptr
 */
static void _print_node_list(FILE *out, lsList l)
{
  char *sep = "";
  lsGen  iterator; 
  node_ptr node;

  lsForEachItem(l, iterator, node)
    {
      _my_print_node(out, sep, node, "");
      sep = ",";
    }
}





/*
 * An auxiliary routine to add a new state variable with 'name' to 'layer'
 */
static node_ptr _add_new_state_variable(SymbLayer_ptr layer, char *name)
{
  SymbType_ptr symbolicType;
  node_ptr node;
  
  node = find_node(DOT, Nil, sym_intern(name));
  assert(node);

  if(!SymbLayer_can_declare_var(layer, node))
    error_redefining(node); 
    
  symbolicType = SymbType_create(SYMB_TYPE_ENUM, boolean_range);
  SymbLayer_declare_state_var(layer, node, symbolicType);
  return node;
}





/*
 *
 * The structure maintaining variable information needed in the BMC procedure
 *
 */
typedef struct HJL_state_vars_struct_TAG
{
  /* lsLists below are lists of node_ptr */

  /* System state variables occurring in transition relation */
  lsList   transition_state_vars;

  /* Auxiliary variables needed in the PLTL BMC encoding */
  lsList   translation_vars_pd0;  /* [[f]]_i^0 vars */
  lsList   translation_vars_pdx;  /* [[f]]_i^d, d > 0 vars */
  lsList   translation_vars_aux;  /* <<F f>>_i and <<G f>>_i vars */
  node_ptr l_var;                 /* The loop selector variable */
  node_ptr LoopExists_var;        /* The loop exists variable */
  
  /* System state variables occurring in formula */
  lsList   formula_state_vars;

  /* System input variables occurring in formula */
  lsList   formula_input_vars;

  /* System variables for the simple path constraint.
   * Union of transition_state_vars, formula_state_vars, formula_input_vars */
  lsList   simple_path_system_vars;
} HJL_state_vars_struct;



/*
 * Create an empty HJL_state_vars_struct
 */
static HJL_state_vars_struct *_HJL_state_vars_create()
{
  HJL_state_vars_struct *HJL_state_vars = ALLOC(HJL_state_vars_struct, 1);
  assert(HJL_state_vars);

  HJL_state_vars->transition_state_vars = lsCreate();
  HJL_state_vars->l_var = NULL;
  HJL_state_vars->LoopExists_var = NULL;
  HJL_state_vars->translation_vars_pd0 = lsCreate();
  HJL_state_vars->translation_vars_pdx = lsCreate();
  HJL_state_vars->translation_vars_aux = lsCreate();
  HJL_state_vars->formula_state_vars = lsCreate();
  HJL_state_vars->formula_input_vars = lsCreate();
  HJL_state_vars->simple_path_system_vars = lsCreate();

  return HJL_state_vars;
}



/*
 * Print a HJL_state_vars_struct 'HJL_state_vars' to 'out'
 */
static void _HJL_state_vars_print(HJL_state_vars_struct *HJL_state_vars,
				  FILE* out)
{
  lsGen  iterator; 
  node_ptr node;
  char *sep;

  fprintf(out, "The %d state variables in transition relation are: ",
	  lsLength(HJL_state_vars->transition_state_vars));
  _print_node_list(out, HJL_state_vars->transition_state_vars);
  fprintf(out, "\n");

  fprintf(out, "The l variable is: ");
  print_node(out, HJL_state_vars->l_var);
  fprintf(out, "\n");

  fprintf(out, "The LoopExists variable is: ");
  print_node(out, HJL_state_vars->LoopExists_var);
  fprintf(out, "\n");

  fprintf(out, "Translation state variables at depth 0 are: ");
  _print_node_list(out, HJL_state_vars->translation_vars_pd0);
  fprintf(out, "\n");

  fprintf(out, "Translation state variables at depth d>0 are: ");
  _print_node_list(out, HJL_state_vars->translation_vars_pdx);
  fprintf(out, "\n");

  fprintf(out, "Translation state variables for aux translation are: ");
  _print_node_list(out, HJL_state_vars->translation_vars_aux);
  fprintf(out, "\n");

  fprintf(out, "State variables in formula are: ");
  _print_node_list(out, HJL_state_vars->formula_state_vars);
  fprintf(out, "\n");

  fprintf(out, "Input variables in formula are: ");
  _print_node_list(out, HJL_state_vars->formula_input_vars);
  fprintf(out, "\n");

  fprintf(out, "The %d system variables in the simple path constraint are: ",
	  lsLength(HJL_state_vars->simple_path_system_vars));
  _print_node_list(out, HJL_state_vars->simple_path_system_vars);
  fprintf(out, "\n");
}





/*
 * Make the be formula "\land_{v \in V} s_i = s_j"
 * using the variables V given in the list 'vars'
 */
static be_ptr _equal_vectors_formula(const BeEnc_ptr be_enc,
				     lsList vars,
				     const unsigned int i,
				     const unsigned int j)
{
  lsGen    iterator;
  node_ptr node_sv;

  be_ptr be_constraint;

  Be_Manager_ptr be_mgr;

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Initialize constraint */
  be_constraint = Be_Truth(be_mgr);

  lsForEachItem(vars, iterator, node_sv)
    {
      /* Get state variable sv at time i */
      be_ptr be_sv_i = BeEnc_name_to_timed(be_enc, node_sv, i);
      /* Get state variable sv at time j */
      be_ptr be_sv_j = BeEnc_name_to_timed(be_enc, node_sv, j);

      be_constraint = Be_And(be_mgr,
			     be_constraint,
			     Be_Iff(be_mgr, be_sv_i, be_sv_j));
    }

  return be_constraint;
}





/*
 * Routines for the state indexing scheme
 * State 0 is the L state and 1 is the E state
 * The first real state is state 2.
 */
static int _L_state() {return 0; }
static int _E_state() {return 1; }
static int _real_k(int k) {return k+2; }
static unsigned int _model_k(int k) {assert(k >= 2); return k-2; }
static char* _real_k_string(const unsigned int k_real)
{
  char *str = ALLOC(char, 32);
  if(k_real == _L_state()) sprintf(str, "L");
  else if(k_real == _E_state()) sprintf(str, "E");
  else sprintf(str, "%u", _model_k(k_real));
  return str;
}





/*
 * Get the initial state and invariant constraint at time i.
 * Similar to Bmc_Model_GetInit0(be_fsm) except for the i argument.
 */
static be_ptr Bmc_TJ_GetInitI(const BeFsm_ptr be_fsm, const int i)
{
  BeEnc_ptr be_enc = BeFsm_get_be_encoding(be_fsm);
  be_ptr init = BeEnc_untimed_expr_to_timed(be_enc,
			      Be_And(BeEnc_get_be_manager(be_enc), 
				     BeFsm_get_init(be_fsm),
				     BeFsm_get_invar(be_fsm)),
					    i);
  return init;
}





/*
 * The data structure holding information for each subformula f
 */
typedef struct _HJL_node_info_struct
{
  unsigned int past_depth;   /* The past depth of the subformula */

  /* The variables for [[f]]^d if translated via variables, 0 otherwise */
  array_t*     trans_vars;   /* node_ptr[past_depth] of node_ptr */
  /* The bes for [[f]]_i^d */
  array_t*     trans_bes;    /* array_t [time] array_t [past_depth] of be_ptr*/

  /* The variable for the auxiliary <<F f>> translation if needed */
  node_ptr     aux_F_node;
  /* The bes for <<F f>>_i */
  array_t*     aux_F_trans;  /* be_ptr aux_F_trans[time] */

  /* The variable for the auxiliary <<G f>> translation if needed */
  node_ptr     aux_G_node;
  /* The bes for <<G f>>_i */
  array_t*     aux_G_trans;  /* be_ptr aux_G_trans[time] */
} _HJL_node_info;


/*
 * Allocate a new _HJL_node_info structure and initialize it
 */
static _HJL_node_info *_alloc_info()
{
  _HJL_node_info *info = ALLOC(_HJL_node_info, 1);
  info->past_depth = 0;
  info->trans_vars = 0;
  info->trans_bes = array_alloc(array_t*, 1);
  array_insert(array_t*, info->trans_bes, 0, 0);
  info->aux_F_node = 0;
  info->aux_F_trans = 0;
  info->aux_G_node = 0;
  info->aux_G_trans = 0;
  return info;
}


/*
 * Deallocate a _HJL_node_info structure given as 'r'
 */
static enum st_retval _HJL_node_info_free_func(char *k, char *r, char *a)
{
  int i;
  if(r) 
    {
      _HJL_node_info* info = (_HJL_node_info*)r;
      if(info->trans_vars)
	{
	  array_free(info->trans_vars);
	  info->trans_vars = 0;
	}
      if(info->trans_bes)
	{
	  for(i = 0; i < array_n(info->trans_bes); i++) {
	    array_t* a = array_fetch(array_t*, info->trans_bes, i);
	    if(a) {
	      array_free(a);
	      array_insert(array_t*, info->trans_bes, i, 0);
	    }
	  }
	  array_free(info->trans_bes); info->trans_bes = 0;
	}
      if(info->aux_F_trans)
	{
	  array_free(info->aux_F_trans);
	  info->aux_F_trans = 0;
	}
      if(info->aux_G_trans)
	{
	  array_free(info->aux_G_trans);
	  info->aux_G_trans = 0;
	}
      FREE(info); info = 0;
    }
  return ST_DELETE;
}





/*
 * Compute the variables that occur in the formula ltlspec.
 * ltlspec must be in NNF.
 */
static lsList _HJL_find_formula_vars(node_ptr ltlspec)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  lsList   formula_vars = 0;
  int i;

  /* Debug output */
  if (opt_verbose_level_ge(options, 5))
    _my_print_node(nusmv_stderr,"Finding vars occurring in ",ltlspec,"\n");
  
  visit_cache = _set_create();
  formula_vars = lsCreate();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);
  
  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node, lsf, rsf;
      int has_unprocessed_children;

      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);

      /* Already visited? */
      if(_set_is_in(visit_cache, node)) {
	if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	  internal_error(_SNH_text,__FILE__,__LINE__);
	continue;
      }

      /* Traverse children and build info */
      lsf = car(node);
      rsf = cdr(node);      
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{
	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	  lsNewBegin(formula_vars, (lsGeneric)node, LS_NH);
	  break;
	  
	case TRUEEXP:
	case FALSEEXP:
	  break;

	case AND:
	case OR:
	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	case UNTIL:
	case RELEASES:
	case TRIGGERED:
	case SINCE:
	  if(!_set_is_in(visit_cache, lsf)) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(!_set_is_in(visit_cache, rsf)) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  break;
	  
	case NOT:
	case OP_GLOBAL:
	case OP_FUTURE:
	case OP_NEXT:
	case OP_HISTORICAL:
	case OP_ONCE:
	case OP_PREC:
	case OP_NOTPRECNOT:
	  if(!_set_is_in(visit_cache, lsf)) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  break;
	  
	default:
	  print_node(nusmv_stderr, node);
	  internal_error("%s:%d: TJ: Something not implemented",
			 __FILE__, __LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;
      
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);

      _set_insert(visit_cache, node);
    }
      
  lsDestroy(unprocessed_nodes, 0);
  _set_destroy(visit_cache);

  return formula_vars;
}





/*
 * Auxiliary routines for printing the subformula variable association
 */
static void _HJL_print_varmap(FILE *out, node_ptr node, _HJL_node_info *info)
{
  assert(node);
  assert(info);
  if(opt_verbose_level_ge(options, 2))
    {
      unsigned int d;
      for(d = 0; d <= info->past_depth; d++)
	{
	  _my_print_node(out, "[[", node, "]]");
	  fprintf(out, "^%u = ", d);
	  _my_print_node(out, "", array_fetch(node_ptr, info->trans_vars, d), ";\n");
	}
    }
}
static void _HJL_print_Gvarmap(FILE *out, node_ptr var, node_ptr formula)
{
  assert(var);
  assert(formula);
  if(opt_verbose_level_ge(options, 2))
    {
      _my_print_node(nusmv_stderr, "<<G ", formula, ">> = ");
      _my_print_node(nusmv_stderr, "", var, ";\n");
    }
}
static void _HJL_print_Fvarmap(FILE *out, node_ptr var, node_ptr formula)
{
  assert(var);
  assert(formula);
  if(opt_verbose_level_ge(options, 2))
    {
      _my_print_node(nusmv_stderr, "<<F ", formula, ">> = ");
      _my_print_node(nusmv_stderr, "", var, ";\n");
    }
}





/*
 * Returns a fresh state variable allocated to layer with name #LTL_t'index'.
 * index is increased by one.
 */
static node_ptr _1_fresh_state_var(SymbLayer_ptr layer, unsigned int *index)
{
  char new_var_name[16];
  node_ptr new_var = 0;

  assert(index);
  
  sprintf(new_var_name, "#LTL_t%u", *index);
  *index = *index + 1;
  new_var = _add_new_state_variable(layer, new_var_name);
  return new_var;
}

/*
 * Returns n fresh state variables allocated to layer with name #LTL_t'index'.
 * index is increased by n.
 */
static array_t* _n_fresh_state_vars(SymbLayer_ptr layer,
				    const unsigned int n,
				    unsigned int *index)
{
  array_t* array;
  unsigned int i;

  assert(n > 0);
  assert(index);

  array = array_alloc(node_ptr, n);
  assert(array);

  for(i = 0; i < n; i++)
    {
      array_insert(node_ptr, array, i, _1_fresh_state_var(layer, index));
    }
  return array;
}





/*
 * Allocates info->pastdepth+1 new state variables for the main translation
 * in info->trans_vars.
 * state_vars_formula_pd0, state_vars_formula_pdx and new_var_index are
 * updated accordingly
 */
static void _allocate_trans_vars(_HJL_node_info *info,
				 SymbLayer_ptr layer,
				 lsList state_vars_formula_pd0,
				 lsList state_vars_formula_pdx,
				 int *new_var_index)
{
  unsigned int d;

  assert(info);
  assert(info->trans_vars == 0);

  info->trans_vars = _n_fresh_state_vars(layer,
					 info->past_depth + 1,
					 new_var_index);

  /* Update state_vars_formula_pd0 and state_vars_formula_pdx */
  lsNewBegin(state_vars_formula_pd0,
	     (lsGeneric)array_fetch(node_ptr,info->trans_vars,0),
	     LS_NH);
  for(d = 1; d <= info->past_depth; d++)
    lsNewBegin(state_vars_formula_pdx,
	       (lsGeneric)array_fetch(node_ptr,info->trans_vars,d),
	       LS_NH);
}



/*
 * Associates each subformula node of ltlspec with a _HJL_LTL_info
 * Returns a hash from node_ptr to _HJL_LTL_info*.
 * New state variables named #LTL_t'i' can be allocate to 'layer'.
 * The new state vars are inserted in state_vars_formula_??? appropriately.
 */
static hash_ptr _HJL_init_LTL_info(node_ptr ltlspec,
				   SymbLayer_ptr layer,
				   lsList state_vars_formula_pd0,
				   lsList state_vars_formula_pdx,
				   lsList state_vars_formula_aux,
				   const int opt_force_state_vars,
				   const int opt_do_virtual_unrolling)
{
  hash_ptr info_map = 0;
  lsList   unprocessed_nodes = 0;
  unsigned int d;
  unsigned int new_var_index = 0;

  assert(state_vars_formula_pd0);
  assert(state_vars_formula_pdx);
  assert(state_vars_formula_aux);

  /* Debug output */
  if (opt_verbose_level_ge(options, 4))
    _my_print_node(nusmv_stderr, "Computing the past depth information and"
		   " allocating variables for ", ltlspec, "\n");

  info_map = new_assoc();
  nusmv_assert(info_map != (hash_ptr)NULL);

  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);
  
  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node = 0, lsf = 0, rsf = 0;
      _HJL_node_info *info = 0, *lsf_info = 0, *rsf_info = 0;
      int has_unprocessed_children;

      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);

      /* Is info already computed? */
      if(find_assoc(info_map, node) != 0)
	{
	  /* Remove from unprocessed stack */
	  if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	    internal_error(_SNH_text,__FILE__,__LINE__);
	  continue;
	}
      
      /* Traverse children and build info */
      lsf = car(node);
      rsf = cdr(node);      
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{
	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	case TRUEEXP:
	case FALSEEXP:
	  /* Base case: no children */
	  info = _alloc_info();
	  info->past_depth = 0;
	  if(opt_force_state_vars)
	    {
	      /* Formula variable translation */
	      _allocate_trans_vars(info, layer, state_vars_formula_pd0,
				   state_vars_formula_pdx, &new_var_index);
	      _HJL_print_varmap(nusmv_stderr, node, info);
	    }
	  else
	    {
	      /* Definitional translation */
	      info->trans_vars = 0;
	    }
	  insert_assoc(info_map, node, (node_ptr)info);
	  break;


	case AND:
	case OR:
	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(!rsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = max(lsf_info->past_depth, rsf_info->past_depth);
	  else
	    info->past_depth = 0;

	  if(opt_force_state_vars)
	    {
	      /* Formula variable translation */
	      _allocate_trans_vars(info, layer, state_vars_formula_pd0,
				   state_vars_formula_pdx, &new_var_index);
	      _HJL_print_varmap(nusmv_stderr, node, info);
	    }
	  else
	    {
	      /* Definitional translation */
	      info->trans_vars = 0;
	    }
	  insert_assoc(info_map, node, (node_ptr)info);
	  break;

	  
	case NOT:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = lsf_info->past_depth;
	  else
	    info->past_depth = 0;
	  assert(info->past_depth == 0);

	  if(opt_force_state_vars)
	    {
	      /* Formula variable translation */
	      _allocate_trans_vars(info, layer, state_vars_formula_pd0,
				   state_vars_formula_pdx, &new_var_index);
	      _HJL_print_varmap(nusmv_stderr, node, info);
	    }
	  else
	    {
	      /* Definitional translation */
	      info->trans_vars = 0;
	    }
	  insert_assoc(info_map, node, (node_ptr)info);
	  break;

	  
	case OP_FUTURE:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = lsf_info->past_depth;
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);

	  /* Aux_F_node for the subformula */
	  if(lsf_info->aux_F_node == 0)
	    {
	      lsf_info->aux_F_node = _1_fresh_state_var(layer, &new_var_index);
	      lsNewBegin(state_vars_formula_aux,
			 (lsGeneric)lsf_info->aux_F_node, LS_NH);
	      _HJL_print_Fvarmap(nusmv_stderr, lsf_info->aux_F_node, lsf);
	    }
	  break;


	case OP_GLOBAL:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = lsf_info->past_depth;
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);

	  /* Aux_G_node for the subformula */
	  if(lsf_info->aux_G_node == 0)
	    {
	      lsf_info->aux_G_node = _1_fresh_state_var(layer, &new_var_index);
	      lsNewBegin(state_vars_formula_aux,
			 (lsGeneric)lsf_info->aux_G_node, LS_NH);
	      _HJL_print_Gvarmap(nusmv_stderr, lsf_info->aux_G_node, lsf);
	    }
	  break;


	case OP_NEXT:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = lsf_info->past_depth;
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);

	  /* Subformula referred in the L and k+1 state,
	     make sure it can be free i.e. force it to be translated via var */
	  if(lsf_info->trans_vars == 0)
	    {
	      _allocate_trans_vars(lsf_info, layer, state_vars_formula_pd0,
				   state_vars_formula_pdx, &new_var_index);
	      _HJL_print_varmap(nusmv_stderr, lsf, lsf_info);
	    }
	  break;


	case OP_HISTORICAL:
	case OP_ONCE:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = lsf_info->past_depth + 1;
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);
	  break;



	case OP_PREC:
	case OP_NOTPRECNOT:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = lsf_info->past_depth + 1;
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);

	  /* Subformula referred in the E state, make sure it can be free */
	  if(lsf_info->trans_vars == 0)
	    {
	      _allocate_trans_vars(lsf_info, layer, state_vars_formula_pd0,
				   state_vars_formula_pdx, &new_var_index);
	      _HJL_print_varmap(nusmv_stderr, lsf, lsf_info);
	    }
	  break;


	case UNTIL:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(!rsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = max(lsf_info->past_depth, rsf_info->past_depth);
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);

	  /* Aux_F_node for the right subformula */
	  if(rsf_info->aux_F_node == 0) {
	    rsf_info->aux_F_node = _1_fresh_state_var(layer, &new_var_index);
	    lsNewBegin(state_vars_formula_aux,
		       (lsGeneric)rsf_info->aux_F_node, LS_NH);
	    _HJL_print_Fvarmap(nusmv_stderr, rsf_info->aux_F_node, rsf);
	  }
	  break;


	case RELEASES:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(!rsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = max(lsf_info->past_depth, rsf_info->past_depth);
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);

	  /* Aux_G_node for the right subformula */
	  if(rsf_info->aux_G_node == 0) {
	    rsf_info->aux_G_node = _1_fresh_state_var(layer, &new_var_index);
	    lsNewBegin(state_vars_formula_aux,
		       (lsGeneric)rsf_info->aux_G_node, LS_NH);
	    _HJL_print_Gvarmap(nusmv_stderr, rsf_info->aux_G_node, rsf);
	  }
	  break;


	case TRIGGERED:
	case SINCE:
	  lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	  rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	  if(!lsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(!rsf_info) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(has_unprocessed_children)
	    break;

	  info = _alloc_info();

	  if(opt_do_virtual_unrolling)
	    info->past_depth = max(lsf_info->past_depth, rsf_info->past_depth)+1;
	  else
	    info->past_depth = 0;

	  _allocate_trans_vars(info, layer, state_vars_formula_pd0,
			       state_vars_formula_pdx, &new_var_index);
	  _HJL_print_varmap(nusmv_stderr, node, info);
	  insert_assoc(info_map, node, (node_ptr)info);
	  break;

	  
	default:
	  internal_error(_SNYI_text, __FILE__, __LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;
      
      assert(find_assoc(info_map, node) != 0);

      /* Remove from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);

      /* Debug output */
      if (opt_verbose_level_ge(options, 6))
	{
	  _my_print_node(nusmv_stderr, " The past depth of ", node, " is ");
	  fprintf(nusmv_stderr, "%u\n", info->past_depth);
	}
    }
      
  lsDestroy(unprocessed_nodes, 0);

  return info_map;
}





/*
 * Initialize trans_bes[i][d], 0<=d<=pd, to
 *   (i) the formula [[f]]_i^d for definitionally translated subformulae
 *   (ii) the [[f]]_i^d be variable for variable translated subformulae
 */
static void _HJL_init_state_vector(const BeEnc_ptr be_enc,
				   const node_ptr ltlspec,
				   const hash_ptr info_map,
				   const unsigned int i_real)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  unsigned int d;
  char *i_str = _real_k_string(i_real);
  Be_Manager_ptr be_mgr;

  nusmv_assert(info_map != (hash_ptr)NULL);


  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);


  /* Debug output */
  if (opt_verbose_level_ge(options, 2))
    {
      fprintf(nusmv_stderr, "Initializing state vector at time %s\n", i_str);
      fflush(nusmv_stderr);
    }
  
  
  visit_cache = _set_create();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);
  
  while(lsLength(unprocessed_nodes) > 0)
    {
      /* Nodes for the current formula and its subformulae */
      node_ptr f_node, lsf_node, rsf_node;
      /* The corresponding node info structures */
      _HJL_node_info *f_info, *lsf_info, *rsf_info;
      array_t *past_array, *lsf_past_array, *rsf_past_array;
      int has_unprocessed_children;

      /* Get the current formula to f_node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&f_node, LS_NH) != LS_OK ||
	 !f_node)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get the info structure of the formula */
      f_info = (_HJL_node_info*)find_assoc(info_map, f_node);
      if(!f_info)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get past_array */
      assert(f_info->trans_bes);
      while(array_n(f_info->trans_bes) <= i_real)
	array_insert(array_t*, f_info->trans_bes, array_n(f_info->trans_bes), 0);
      past_array = array_fetch(array_t*, f_info->trans_bes, i_real);


      /* Already visited? */
      if(_set_is_in(visit_cache, f_node))
	{
	  assert(past_array);
	  /* Remove node from unprocessed stack */
	  if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&f_node) != LS_OK)
	    internal_error(_SNH_text,__FILE__,__LINE__);
	  continue;
	}
      

      /* Initialize past_array[i] if not already done */
      if(!past_array)
	{
	  past_array = array_alloc(be_ptr, f_info->past_depth+1);
	  for(d = 0; d <= f_info->past_depth; d++)
	    array_insert(be_ptr, past_array, d, 0);
	  array_insert(array_t*, f_info->trans_bes, i_real, past_array);
	}

      for(d = 0; d <= f_info->past_depth; d++)
	assert(array_fetch(be_ptr, past_array, d) == 0);      

      /* Traverse children and build past_array */
      lsf_node = car(f_node);
      rsf_node = cdr(f_node);
      has_unprocessed_children = 0;
      switch(node_get_type(f_node))
	{
	  
	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	  {
	    be_ptr be_result = 0;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");
	    
	    assert(f_info->past_depth == 0);

	    if(f_info->trans_vars != 0)
	      {
		/* main_array[i][0] = [[p]]_i */
		node_ptr node_p = array_fetch(node_ptr, f_info->trans_vars, 0);
		assert(node_p);
		be_result = BeEnc_name_to_timed(be_enc, node_p, i_real);
		assert(be_result);
		array_insert(be_ptr, past_array, 0, be_result);
		    
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  at depth 0: ");
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    else
	      {
		/* Definitional translation */
		/* main_array[i][0] = p_i */
		be_result = BeEnc_name_to_timed(be_enc, f_node, i_real);
		assert(be_result);
		array_insert(be_ptr, past_array, 0, be_result);
		
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[p]]_%s^0 := p_%s: ",i_str,i_str);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }

	case TRUEEXP:
	  {
	    be_ptr be_result = 0;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");
	    
	    assert(f_info->past_depth == 0);

	    if(f_info->trans_vars != 0)
	      {
		/* main_array[i][0] = [[TRUE]]_i */
		node_ptr node_aux = array_fetch(node_ptr, f_info->trans_vars, 0);
		assert(node_aux);
		be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		assert(be_result);
		array_insert(be_ptr, past_array, 0, be_result);
		
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  at depth %u: ", d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    else
	      {
		/* Definitional translation */
		/* main_array[i][0] = TRUE */
		be_result = Be_Truth(be_mgr);
		array_insert(be_ptr, past_array, 0, be_result);
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[TRUE]]_%s^0 := TRUE: ", i_str);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    
	    break;
	  }

	case FALSEEXP:
	  {
	    be_ptr be_result = 0;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");
	    
	    assert(f_info->past_depth == 0);

	    if(f_info->trans_vars != 0)
	      {
		/* Formula variable translation */
		/* main_array[i][0] = [[FALSE]]_i */
		node_ptr node_aux = array_fetch(node_ptr, f_info->trans_vars, 0);
		assert(node_aux);
		be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		assert(be_result);
		array_insert(be_ptr, past_array, 0, be_result);
		    
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  at depth 0: ");
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    else
	      {
		/* Definitional translation */
		/* main_array[i][0] = FALSE */
		be_result = Be_Falsity(be_mgr);
		array_insert(be_ptr, past_array, 0, be_result);
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[FALSE]]_%s^0 := FALSE: ", i_str);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }


	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  {
	    internal_error("%s:%d: Formula not in NNF\n", __FILE__,__LINE__);
	    break;
	  }
	  
	case OR:
	  {
	    unsigned int d;

	    if(!_set_is_in(visit_cache, lsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf_node);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf_node);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(array_t*,rsf_info->trans_bes,i_real);
	    assert(rsf_past_array);
	    
	    if(f_info->trans_vars != 0)
	      {
		/* Formula variable translation */
		for(d = 0; d <= f_info->past_depth; d++)
		  {
		    /* main_array[i][d] = [[f | g]]_i^d */
		    node_ptr node_aux;
		    be_ptr   be_result;
		    node_aux = array_fetch(node_ptr, f_info->trans_vars, d);
		    assert(node_aux);
		    be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		    assert(be_result);
		    array_insert(be_ptr, past_array, d, be_result);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  at depth %u: ", d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    else
	      {
		for(d = 0; d <= f_info->past_depth; d++)
		  {
		    /* Definitional translation */
		    /* [[f | g]]_i^d := [[f]]_i^d | [[g]]_i^d */
		    be_ptr be_lsf, be_rsf, be_result;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_lsf = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_lsf);
		    be_rsf = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_rsf);
		    be_result = Be_Or(be_mgr, be_lsf, be_rsf);
		    assert(array_fetch(be_ptr, past_array, d) == 0);
		    array_insert(be_ptr, past_array, d, be_result);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr,
			      "  [[f | g]]_%s^%d := [[f]]_%s^%d | [[g]]_%s^%d: ",
			      i_str, d, i_str, d_lsf, i_str, d_rsf);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");	
		    }
		  }
	      }

	    break;
	  }


	case AND:
	  {
	    unsigned int d;

	    if(!_set_is_in(visit_cache, lsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf_node);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf_node);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(array_t*,rsf_info->trans_bes,i_real);
	    assert(rsf_past_array);
	    
	    if(f_info->trans_vars != 0)
	      {
		/* Formula variable translation */
		for(d = 0; d <= f_info->past_depth; d++)
		  {
		    /* main_array[i][d] = [[f & g]]_i^d */
		    node_ptr node_aux;
		    be_ptr   be_result;
		    node_aux = array_fetch(node_ptr, f_info->trans_vars, d);
		    assert(node_aux);
		    be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		    assert(be_result);
		    array_insert(be_ptr, past_array, d, be_result);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  at depth %u: ", d);
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
			fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    else
	      {
		for(d = 0; d <= f_info->past_depth; d++)
		  {
		    /* Definitional translation */
		    /* [[f | g]]_i^d := [[f]]_i^d & [[g]]_i^d */
		    be_ptr be_lsf, be_rsf, be_result;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_lsf = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_lsf);
		    be_rsf = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_rsf);
		    be_result = Be_And(be_mgr, be_lsf, be_rsf);
		    assert(array_fetch(be_ptr, past_array, d) == 0);
		    array_insert(be_ptr, past_array, d, be_result);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[f & g]]_%s^%d := [[f]]_%s^%d & [[g]]_%s^%d: ",
			      i_str, d, i_str, d_lsf, i_str, d_rsf);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }



	case NOT:
	  {
	    unsigned int d;
	    if(!_set_is_in(visit_cache, lsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(f_info->past_depth == 0);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf_node);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");

	    if(f_info->trans_vars != 0)
	      {
		/* Formula variable translation */
		for(d = 0; d <= f_info->past_depth; d++)
		  {
		    /* main_array[i][d] = [[!f]]_i^d */
		    node_ptr node_aux;
		    be_ptr   be_result;
		    node_aux = array_fetch(node_ptr, f_info->trans_vars, d);
		    assert(node_aux);
		    be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		    assert(be_result);
		    array_insert(be_ptr, past_array, d, be_result);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  at depth %u: ", d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    else
	      {
		/* Definitional translation */
		for(d = 0; d <= f_info->past_depth; d++)
		  {
		    /* [[!f]]_i^d := ![[f]]_i^d */
		    be_ptr be_lsf, be_result;
		    be_lsf = array_fetch(be_ptr, lsf_past_array, d);
		    assert(be_lsf);
		    be_result = Be_Not(be_mgr, be_lsf);
		    assert(array_fetch(be_ptr, past_array, d) == 0);
		    array_insert(be_ptr, past_array, d, be_result);

		    if(opt_verbose_level_ge(options, 6))
		      fprintf(nusmv_stderr, "  [!f]]_%s^%d := ![[f]]_%s^%d: ",
			      i_str, d, i_str, d),
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
			fprintf(nusmv_stderr, "\n");
		  }
	      }
	    break;
	  }


	case UNTIL:
	case RELEASES:
	case TRIGGERED:
	case SINCE:
	  {
	    unsigned int d;
	    if(!_set_is_in(visit_cache, lsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(f_info->trans_vars);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");

	    for(d = 0; d <= f_info->past_depth; d++)
	      {
		/* main_array[i][d] = [[f OP g]]_i^d */
		node_ptr node_aux;
		be_ptr   be_result;
		node_aux = array_fetch(node_ptr, f_info->trans_vars, d);
		assert(node_aux);
		be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		assert(be_result);
		array_insert(be_ptr, past_array, d, be_result);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  at depth %u: ", d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }

	case OP_FUTURE:
	case OP_GLOBAL:
	case OP_NEXT:
	case OP_HISTORICAL:
	case OP_ONCE:
	case OP_PREC:
	case OP_NOTPRECNOT:
	  {
	    unsigned int d;
	    if(!_set_is_in(visit_cache, lsf_node)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf_node, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(f_info->trans_vars);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", f_node, "\n");

	    for(d = 0; d <= f_info->past_depth; d++)
	      {
		/* main_array[i][d] = [[OP f]]_i^d */
		node_ptr node_aux;
		be_ptr   be_result;
		node_aux = array_fetch(node_ptr, f_info->trans_vars, d);
		assert(node_aux);
		be_result = BeEnc_name_to_timed(be_enc, node_aux, i_real);
		assert(be_result);
		array_insert(be_ptr, past_array, d, be_result);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  at depth %u: ", d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }
	  
	default:
	  print_node(stderr, f_node);
	  internal_error("%s:%d: TJ: Something not implemented",
			 __FILE__, __LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;
      
      /* Remove node from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&f_node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Mark visited */
      _set_insert(visit_cache, f_node);
    }

      
  lsDestroy(unprocessed_nodes, 0);
  _set_destroy(visit_cache); visit_cache = 0;
  FREE(i_str); i_str = 0;
}




/*
 * Create the base constraints.
 * Return a list of be_ptr for the created constraints.
 * Create the following constraints:
 *   !LoopExists => ([[f]]_L^d == FALSE)
 *   LoopExists => ([[Ff]]_E^pd(Gf) => <<Ff>>_E)
 *   LoopExists => ([[Gf]]_E^pd(Gf) <= <<Gf>>_E)
 *   LoopExists => ([[fUg]]_E^pd(Gf) => <<Fg>>_E)
 *   LoopExists => ([[fRg]]_E^pd(Gf) <= <<Gg>>_E)
 * If do_optimization is true, then create the following constraints:
 *   [[p]]_E^d <=> p_E
 *   [[TRUE]]_E^0 <=> TRUE
 *   [[FALSE]]_E^0 <=> FALSE
 *   [[f | g]]_E^d <=> [[f]]_E^d | [[g]]_E^d
 *   [[f & g]]_E^d <=> [[f]]_E^d & [[g]]_E^d
 *   [[!f]]_E^d <=> ![[f]]_E^d
 *   [[Ff]]_E^d <=> [[f]]_E^d | [[Ff]]_L^min(d+1,pd(f))
 *   [[Ff]]_E^d+1 => [[Ff]]_E^d
 *   <<Ff>>_E => [[Ff]]_E^pd(Ff)
 *   [[Gf]]_E^d <=> [[f]]_E^d & [[Gf]]_L^min(d+1,pd(f))
 *   [[Gf]]_E^d => [[Gf]]_E^d+1
 *   [[Gf]]_E^pd(Gf) => <<Gf>>_E
 *   [[fUg]]_E^d <=> [[g]]_E^d | ([[f]]_E^d  & [[fUg]]_L^min(d+1,pd(fUg)))
 *   [[fRg]]_E^d <=> [[g]]_E^d & ([[f]]_E^d  | [[fRg]]_L^min(d+1,pd(fUg)))
 *   [[Xf]]_E^d <=> [[f]]_L^min(d+1,pd(f))
 *   [[Hf]]_E^d+1 => [[Hf]]_E^d 
 *   [[Of]]_E^d => [[Of]]_E^d+1
 */
static lsList _HJL_unroll_base(const node_ptr ltlspec,
			       const BeEnc_ptr be_enc,
			       const hash_ptr info_map,
			       const be_ptr be_LoopExists,
			       const int do_optimization)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  lsList   created_constraints = 0;
  Be_Manager_ptr be_mgr;

  nusmv_assert(info_map != (hash_ptr)NULL);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);

  /* Create list for the constraints */
  created_constraints = lsCreate();

  /* Debug output */
  if(opt_verbose_level_gt(options, 1))
    {
      fprintf(nusmv_stderr, "Creating the base constraints\n");
      fflush(nusmv_stderr);
    }
  
  visit_cache = _set_create();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);

  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node, lsf, rsf;
      _HJL_node_info *info, *lsf_info, *rsf_info;
      array_t *E_past_array = 0, *L_past_array = 0;
      int has_unprocessed_children;

      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get info */
      info = (_HJL_node_info*)find_assoc(info_map, node);
      if(!info)
	internal_error(_SNH_text,__FILE__,__LINE__);

      /* Get past_array */
      /* _HJL_init_state_vector should have allocated trans_bes[i] before */
      if(info->trans_bes)
	{
	  assert(array_n(info->trans_bes) >= 2);
	  L_past_array = array_fetch(array_t*, info->trans_bes, _L_state());
	  assert(L_past_array);
	  E_past_array = array_fetch(array_t*, info->trans_bes, _E_state());
	  assert(E_past_array);
	}

      
      /* Already build? */
      if(_set_is_in(visit_cache, node))
	{
	  /* Remove node from unprocessed stack */
	  if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	    internal_error(_SNH_text,__FILE__,__LINE__);
	  continue;
	}


     /* Traverse children and build info */
      lsf = car(node);
      rsf = cdr(node);      
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{

	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	  {
	    if(do_optimization)
	      {
		/* Optimization:
		   Add [[p]]_E^d <=> p_E */
		be_ptr be_result = 0;
		assert(info->past_depth == 0);
		if(info->trans_vars == 0)
		  {
		    /* Definitional translation already done by
		       state vector init */
		    assert(array_fetch(be_ptr, E_past_array, 0) != 0);
		  }
		else
		  {
		    /* State variable translation
		     * Build [[p]]_E^0 <=> p_E */
		    be_ptr be_p_E_0, be_realp_E;
		    be_p_E_0 = array_fetch(be_ptr, E_past_array, 0);
		    assert(be_p_E_0);
		    be_realp_E = BeEnc_name_to_timed(be_enc, node, _E_state());
		    assert(be_realp_E);
		    
		    be_result = Be_Iff(be_mgr, be_p_E_0, be_realp_E);
		    assert(be_result);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[p]]_E^0 <=> p_E: ");
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }

	case TRUEEXP:
	  {
	    if(do_optimization)
	      {
		be_ptr be_result;
		
		assert(info->past_depth == 0);
		
		if(info->trans_vars == 0)
		  {
		    /* Definitional translation already done by
		       state vector init */
		    assert(array_fetch(be_ptr, E_past_array, 0) != 0);
		  }
		else
		  {
		    /* Formula variable translation */
		    /* Build [[TRUE]]_E^0 <=> TRUE */
		    be_ptr be_TRUE_E_0 = array_fetch(be_ptr, E_past_array, 0);
		    assert(be_TRUE_E_0);
		    
		    be_result = Be_Iff(be_mgr, be_TRUE_E_0, Be_Truth(be_mgr));
		    assert(be_result);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[TRUE]]_E^0 <=> TRUE: ");
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }

	case FALSEEXP:
	  {
	    if(do_optimization)
	      {
		be_ptr be_result;
		
		assert(info->past_depth == 0);
		
		if(info->trans_vars == 0)
		  {
		    /* Definitional translation already done by
		       state vector init */
		    assert(array_fetch(be_ptr, E_past_array, 0) != 0);
		  }
		else
		  {
		    /* Formula variable translation */
		    /* Build [[FALSE]]_E^0 <=> FALSE */
		    be_ptr be_FALSE_E_0 = array_fetch(be_ptr, E_past_array, 0);
		    assert(be_FALSE_E_0);
		    
		    be_result = Be_Iff(be_mgr,be_FALSE_E_0,Be_Falsity(be_mgr));
		    assert(be_result);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[FALSE]]_E^0 <=> FALSE: ");
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }

	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  {
	    internal_error("%s:%d: Formula not in NNF\n",__FILE__,__LINE__);
	    break;
	  }

	case OR:
	  {
	    unsigned int d;
	    array_t *past_array_f_E, *past_array_g_E;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    past_array_f_E = array_fetch(array_t*, lsf_info->trans_bes,
					 _E_state());
	    assert(past_array_f_E);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    past_array_g_E = array_fetch(array_t*, rsf_info->trans_bes,
					 _E_state());
	    assert(past_array_g_E);
	    
	    if(do_optimization) {
	      /* Add [[f | g]]_E^d <=> [[f]]_E^d | [[g]]_E^d */
	      if(info->trans_vars == 0)
		{
		  /* Definitional translation already done by
		     state vector init */
		  for(d = 0; d <= info->past_depth; d++)
		    assert(array_fetch(be_ptr, E_past_array, d) != 0);
		}
	      else
		{
		  for(d = 0; d <= info->past_depth; d++)
		    {
		      be_ptr be_fORg_E_d, be_f_E_d, be_g_E_d, be_result;
		      const unsigned int d_lsf = min(d, lsf_info->past_depth);
		      const unsigned int d_rsf = min(d, rsf_info->past_depth);
		      
		      be_fORg_E_d = array_fetch(be_ptr, E_past_array, d);
		      assert(be_fORg_E_d);
		      be_f_E_d = array_fetch(be_ptr, past_array_f_E, d_lsf);
		      assert(be_f_E_d);
		      be_g_E_d = array_fetch(be_ptr, past_array_g_E, d_rsf);
		      assert(be_g_E_d);
		      be_result = Be_Iff(be_mgr, be_fORg_E_d,
					 Be_Or(be_mgr, be_f_E_d, be_g_E_d));
		      assert(be_result);
		      
		      /* Save the created constraint */
		      lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		      
		      if(opt_verbose_level_ge(options, 6)) {
			fprintf(nusmv_stderr,
				"  [[f | g]]_E^%d <=> [[f]]_E^%d | [[g]]_E^%d: ",
				d, d_lsf, d_rsf);
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
			fprintf(nusmv_stderr, "\n");
		      }
		    }
		}
	    }

	    break;
	  }


	case AND:
	  {
	    unsigned int d;
	    array_t *past_array_f_E, *past_array_g_E;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    past_array_f_E = array_fetch(array_t*, lsf_info->trans_bes,
					 _E_state());
	    assert(past_array_f_E);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    past_array_g_E = array_fetch(array_t*, rsf_info->trans_bes,
					 _E_state());
	    assert(past_array_g_E);
	    
	    if(do_optimization) {
	      /* Add [[f & g]]_E^d <=> [[f]]_E^d & [[g]]_E^d */
	      if(info->trans_vars == 0)
		{
		  /* Definitional translation already done by
		     state vector init */
		  for(d = 0; d <= info->past_depth; d++)
		    assert(array_fetch(be_ptr, E_past_array, d) != 0);
		}
	      else
		{
		  for(d = 0; d <= info->past_depth; d++)
		    {
		      be_ptr be_fANDg_E_d, be_f_E_d, be_g_E_d, be_result;
		      const unsigned int d_lsf = min(d, lsf_info->past_depth);
		      const unsigned int d_rsf = min(d, rsf_info->past_depth);
		      
		      be_fANDg_E_d = array_fetch(be_ptr, E_past_array, d);
		      assert(be_fANDg_E_d);
		      be_f_E_d = array_fetch(be_ptr, past_array_f_E, d_lsf);
		      assert(be_f_E_d);
		      be_g_E_d = array_fetch(be_ptr, past_array_g_E, d_rsf);
		      assert(be_g_E_d);
		      be_result = Be_Iff(be_mgr, be_fANDg_E_d,
					 Be_And(be_mgr, be_f_E_d, be_g_E_d));
		      assert(be_result);
		      
		      /* Save the created constraint */
		      lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		      
		      if(opt_verbose_level_ge(options, 6)) {
			fprintf(nusmv_stderr,
				"  [[f & g]]_E^%d <=> [[f]]_E^%d & [[g]]_E^%d: ",
				d, d_lsf, d_rsf);
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
			fprintf(nusmv_stderr, "\n");
		      }
		    }
		}
	    }

	    break;
	  }

	case NOT:
	  {
	    unsigned int d;
	    array_t* past_array_f_E;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->past_depth == 0);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    past_array_f_E = array_fetch(array_t*, lsf_info->trans_bes,
					 _E_state());
	    assert(past_array_f_E);
	    
	    if(do_optimization) {
	      /* Add [[!f]]_E^d <=> ![[f]]_E^d */
	      if(info->trans_vars == 0)
		{
		  /* Definitional translation already done by
		     state vector init */
		  for(d = 0; d <= info->past_depth; d++)
		    assert(array_fetch(be_ptr, E_past_array, d) != 0);
		}
	      else
		{
		  /* State variable translation */
		  for(d = 0; d <= info->past_depth; d++)
		    {
		      be_ptr be_notf_E_d, be_f_E_d, be_result;
		      be_notf_E_d = array_fetch(be_ptr, E_past_array, d);
		      assert(be_notf_E_d);
		      be_f_E_d = array_fetch(be_ptr, past_array_f_E, d);
		      assert(be_f_E_d);
		      be_result = Be_Iff(be_mgr, be_notf_E_d,
					 Be_Not(be_mgr, be_f_E_d));
		      assert(be_result);
		      
		      /* Save the created constraint */
		      lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		      
		      if(opt_verbose_level_ge(options, 6)) {
			fprintf(nusmv_stderr, "  [!f]]_E^%d <=> ![[f]]_E^%d: ",
				d, d);
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
			fprintf(nusmv_stderr, "\n");
		      }
		    }
		}
	    }

	    break;
	  }


	case OP_FUTURE:
	  {
	    unsigned int  d;
	    array_t*      lsf_E_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    assert(lsf_info->past_depth == info->past_depth);
	    lsf_E_past_array = array_fetch(array_t*,
					   lsf_info->trans_bes, _E_state());
	    assert(lsf_E_past_array);
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", node, "\n");

	    {
	      /* Add LoopExists => ([[Ff]]_E^pd(Gf) => <<Ff>>_E) */
	      be_ptr be_Ff_E_pd, be_auxFf_E, be_result;
	      be_Ff_E_pd = array_fetch(be_ptr, E_past_array, info->past_depth);
	      assert(be_Ff_E_pd);
	      assert(lsf_info->aux_F_node);
	      be_auxFf_E = BeEnc_name_to_timed(be_enc,
					       lsf_info->aux_F_node,
					       _E_state());
	      assert(be_auxFf_E);
	      be_result = Be_Implies(be_mgr,
				     be_LoopExists,
				     Be_Implies(be_mgr,
						be_Ff_E_pd,
						be_auxFf_E));
	      
	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	      
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr,
			"  LoopExists => ([[Ff]]_E^%d => <<Ff>>_E): ",
			info->past_depth);
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }


	    if(do_optimization)
	      {
		/* Optimization.
		 * Add [[Ff]]_E^d <=> [[f]]_E^d | [[Ff]]_L^min(d+1,pd(f)) */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    be_ptr be_aux, be_lsf, be_aux_next, be_result;
		    be_aux = array_fetch(be_ptr, E_past_array, d);
		    assert(be_aux);
		    be_lsf = array_fetch(be_ptr, lsf_E_past_array, d);
		    assert(be_lsf);
		    be_aux_next = array_fetch(be_ptr, L_past_array,
					      min(d+1, info->past_depth));
		    assert(be_aux_next);
		    be_result = Be_Iff(be_mgr, be_aux,
				       Be_Or(be_mgr, be_lsf, be_aux_next));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[Ff]]_E^%d <=> [[f]]_E^%d | [[Ff]]_L^min(%d,%d): ", d, d, d+1, info->past_depth);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
		
	    if(do_optimization)
	      {
		/* Optimization:
		   Add [[Ff]]_E^d+1 => [[Ff]]_E^d */
		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Ff_E_d, be_Ff_E_dP1, be_result;
		    be_Ff_E_d = array_fetch(be_ptr, E_past_array, d);
		    assert(be_Ff_E_d);
		    be_Ff_E_dP1 = array_fetch(be_ptr, E_past_array, d+1);
		    assert(be_Ff_E_dP1);
		    be_result = Be_Implies(be_mgr, be_Ff_E_dP1, be_Ff_E_d);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, " [[Ff]]_E^%u => [[Ff]]_E^%u: ",
			      d+1, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    if(do_optimization)
	      {
		/* Optimization: Add <<Ff>>_E => [[Ff]]_E^pd(Ff) */
		be_ptr be_Ff_E_pd, be_auxFf_E, be_result;
		be_Ff_E_pd = array_fetch(be_ptr,E_past_array,info->past_depth);
		assert(be_Ff_E_pd);
		assert(lsf_info->aux_F_node);
		be_auxFf_E = BeEnc_name_to_timed(be_enc,
						 lsf_info->aux_F_node,
						 _E_state());
		assert(be_auxFf_E);
		be_result = Be_Implies(be_mgr, be_auxFf_E, be_Ff_E_pd);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
		
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr,
			"   <<Ff>>_E => [[Ff]]_E^%d: ",
			info->past_depth);
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }	      
	    }


	    break;
	  }


	case OP_GLOBAL:
	  {
	    unsigned int  d;
	    array_t*      lsf_E_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    assert(lsf_info->past_depth == info->past_depth);
	    lsf_E_past_array = array_fetch(array_t*,
					   lsf_info->trans_bes, _E_state());
	    assert(lsf_E_past_array);
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", node, "\n");

	    {
	      /* Add LoopExists => ([[Gf]]_E^pd(Gf) <= <<Gf>>_E) */
	      be_ptr be_Gf_E_pd, be_auxGf_E, be_result;
	      be_Gf_E_pd = array_fetch(be_ptr, E_past_array, info->past_depth);
	      assert(be_Gf_E_pd);
	      assert(lsf_info->aux_G_node);
	      be_auxGf_E = BeEnc_name_to_timed(be_enc,
					       lsf_info->aux_G_node,
					       _E_state());
	      assert(be_auxGf_E);
	      be_result = Be_Implies(be_mgr,
				     be_LoopExists,
				     Be_Implies(be_mgr,
						be_auxGf_E,
						be_Gf_E_pd));
	      
	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr,
			"  LoopExists => ([[Gf]]_E^%d <= <<Gf>>_E^%d): ",
			info->past_depth, info->past_depth);
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }	      
	    }
	    
	    if(do_optimization)
	      {
		/* Optimization:
		   Add [[Gf]]_E^d <=> [[f]]_E^d & [[Gf]]_L^min(d+1,pd(f)) */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    be_ptr be_aux, be_lsf, be_aux_next, be_result;
		    be_aux = array_fetch(be_ptr, E_past_array, d);
		    assert(be_aux);
		    be_lsf = array_fetch(be_ptr, lsf_E_past_array, d);
		    assert(be_lsf);
		    be_aux_next = array_fetch(be_ptr, L_past_array,
					      min(d+1, info->past_depth));
		    assert(be_aux_next);
		    be_result = Be_Iff(be_mgr, be_aux,
				       Be_And(be_mgr, be_lsf, be_aux_next));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[Gf]]_E^%u <=> [[f]]_E^%u & [[Gf]]_L^min(%u+1,%u))", d, d, d, info->past_depth);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    if(do_optimization)
	      {
		/* Optimization:
		   Add [[Gf]]_E^d => [[Gf]]_E^d+1 */
		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Gf_E_d, be_Gf_E_dP1, be_result;
		    be_Gf_E_d = array_fetch(be_ptr, E_past_array, d);
		    assert(be_Gf_E_d);
		    be_Gf_E_dP1 = array_fetch(be_ptr, E_past_array, d+1);
		    assert(be_Gf_E_dP1);
		    be_result = Be_Implies(be_mgr, be_Gf_E_d, be_Gf_E_dP1);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, " [[Gf]]_E^%u => [[Gf]]_E^%u: ",
			      d, d+1);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }


	    if(do_optimization)
	      {
		/* Optimization: Add [[Gf]]_E^pd(Gf) => <<Gf>>_E */
		be_ptr be_Gf_E_pd, be_auxGf_E, be_result;
		be_Gf_E_pd = array_fetch(be_ptr,E_past_array,info->past_depth);
		assert(be_Gf_E_pd);
		assert(lsf_info->aux_G_node);
		be_auxGf_E = BeEnc_name_to_timed(be_enc,
						 lsf_info->aux_G_node,
						 _E_state());
		assert(be_auxGf_E);
		be_result = Be_Implies(be_mgr, be_Gf_E_pd, be_auxGf_E);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
		
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr,
			"  [[Gf]]_E^%d => <<Gf>>_E): ",
			info->past_depth);
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }	      
	    }

	    break;
	  }


	case UNTIL:
	  {
	    unsigned int  d;
	    array_t*      past_array_f_E;
	    array_t*      past_array_g_E;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    past_array_f_E = array_fetch(array_t*,
					 lsf_info->trans_bes, _E_state());
	    assert(past_array_f_E);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);	    
	    past_array_g_E = array_fetch(array_t*,
					 rsf_info->trans_bes, _E_state());
	    assert(past_array_g_E);


	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", node, "\n");

	    {
	      /* Add LoopExists => ([[fUg]]_E^pd(Gf) => <<Fg>>_E) */
	      be_ptr be_fUg_E_pd, be_auxFg_E, be_result;
	      be_fUg_E_pd = array_fetch(be_ptr,E_past_array,info->past_depth);
	      assert(be_fUg_E_pd);
	      assert(rsf_info->aux_F_node);
	      be_auxFg_E = BeEnc_name_to_timed(be_enc,
					       rsf_info->aux_F_node,
					       _E_state());
	      assert(be_auxFg_E);
	      be_result = Be_Implies(be_mgr,
				     be_LoopExists,
				     Be_Implies(be_mgr,
						be_fUg_E_pd,
						be_auxFg_E));
	      
	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	      
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr,
			"  LoopExists => ([[fUg]]_E^%u => <<Fg>>_E): ",
			info->past_depth);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
	      }
	    }


	    if(do_optimization)
	      {
		/* Optimization: [[fUg]]_E^d <=>
		 * [[g]]_E^d | ([[f]]_E^d  & [[fUg]]_L^min(d+1,pd(fUg))) */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    be_ptr be_fUg_E_d, be_f_E_d, be_g_E_d, be_fUg_L_dP1;
		    be_ptr be_result;

		    be_fUg_E_d = array_fetch(be_ptr, E_past_array, d);
		    assert(be_fUg_E_d);

		    be_f_E_d = array_fetch(be_ptr, past_array_f_E,
					   min(d, lsf_info->past_depth));
		    assert(be_f_E_d);

		    be_g_E_d = array_fetch(be_ptr, past_array_g_E,
					   min(d, rsf_info->past_depth));
		    assert(be_g_E_d);

		    be_fUg_L_dP1 = array_fetch(be_ptr, L_past_array,
					       min(d+1, info->past_depth));
		    assert(be_fUg_L_dP1);

		    be_result = Be_Iff(be_mgr,
				       be_fUg_E_d,
				       Be_Or(be_mgr,
					     be_g_E_d,
					     Be_And(be_mgr,
						    be_f_E_d,
						    be_fUg_L_dP1)));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[fUg]]_E^%u <=> [[g]]_E^%u | ([[f]]_E^%u & [[fUg]]_L^min(%u,%u)): ",
			      d, min(d, rsf_info->past_depth),
			      min(d, lsf_info->past_depth),
			      d+1, info->past_depth);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }


	case RELEASES:
	  {
	    unsigned int  d;
	    array_t*      past_array_f_E;
	    array_t*      past_array_g_E;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    past_array_f_E = array_fetch(array_t*,
					 lsf_info->trans_bes, _E_state());
	    assert(past_array_f_E);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);	    
	    past_array_g_E = array_fetch(array_t*,
					 rsf_info->trans_bes, _E_state());
	    assert(past_array_g_E);


	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", node, "\n");

	    {
	      /* Add LoopExists => ([[fRg]]_E^pd(Gf) <= <<Gg>>_E) */
	      be_ptr be_fRg_E_pd, be_auxGg_E, be_result;
	      be_fRg_E_pd = array_fetch(be_ptr,E_past_array,info->past_depth);
	      assert(be_fRg_E_pd);
	      assert(rsf_info->aux_G_node);
	      be_auxGg_E = BeEnc_name_to_timed(be_enc,
					       rsf_info->aux_G_node,
					       _E_state());
	      assert(be_auxGg_E);
	      be_result = Be_Implies(be_mgr,
				     be_LoopExists,
				     Be_Implies(be_mgr,
						be_auxGg_E,
						be_fRg_E_pd));
	      
	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	      
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr,
			"  LoopExists => ([[fRg]]_E^%u <= <<Gg>>_E): ",
			info->past_depth);
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }


	    if(do_optimization)
	      {
		/* Optimization: [[fRg]]_E^d <=>
		 * [[g]]_E^d & ([[f]]_E^d  | [[fRg]]_L^min(d+1,pd(fUg))) */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    be_ptr be_fRg_E_d, be_f_E_d, be_g_E_d, be_fRg_L_dP1;
		    be_ptr be_result;

		    be_fRg_E_d = array_fetch(be_ptr, E_past_array, d);
		    assert(be_fRg_E_d);

		    be_f_E_d = array_fetch(be_ptr, past_array_f_E,
					   min(d, lsf_info->past_depth));
		    assert(be_f_E_d);

		    be_g_E_d = array_fetch(be_ptr, past_array_g_E,
					   min(d, rsf_info->past_depth));
		    assert(be_g_E_d);

		    be_fRg_L_dP1 = array_fetch(be_ptr, L_past_array,
					       min(d+1, info->past_depth));
		    assert(be_fRg_L_dP1);

		    be_result = Be_Iff(be_mgr,
				       be_fRg_E_d,
				       Be_And(be_mgr,
					      be_g_E_d,
					      Be_Or(be_mgr,
						    be_f_E_d,
						    be_fRg_L_dP1)));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[fRg]]_E^%u <=> [[g]]_E^%u & ([[f]]_E^%u | [[fRg]]_L^min(%u,%u)): ",
			      d, min(d, rsf_info->past_depth),
			      min(d, lsf_info->past_depth),
			      d+1, info->past_depth);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }


	case OP_NEXT:
	  {
	    unsigned int  d;
	    array_t*      lsf_L_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    assert(lsf_info->past_depth == info->past_depth);
	    lsf_L_past_array = array_fetch(array_t*,
					   lsf_info->trans_bes, _L_state());
	    assert(lsf_L_past_array);
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, " ", node, "\n");

	    if(do_optimization)
	      {
		/* Add [[Xf]]_E^d <=> [[f]]_L^min(d+1,pd(f)) */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    be_ptr be_aux, be_lsf_next, be_result;
		    be_aux = array_fetch(be_ptr, E_past_array, d);
		    assert(be_aux);
		    be_lsf_next = array_fetch(be_ptr, lsf_L_past_array,
					      min(d+1, info->past_depth));
		    assert(be_lsf_next);
		    be_result = Be_Iff(be_mgr, be_aux, be_lsf_next);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr,
			      " [[Xf]]_E^%u <=> [[f]]_L^min(%u+1,%u): ",
			      d, d, lsf_info->past_depth);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }


	case OP_HISTORICAL:
	  {
	    unsigned int  d;
	    
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    if(do_optimization)
	      {
		/* Optimization:
		   Add [[Hf]]_E^d+1 => [[Hf]]_E^d */
		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Hf_E_d, be_Hf_E_dP1, be_result;
		    be_Hf_E_d = array_fetch(be_ptr, E_past_array, d);
		    assert(be_Hf_E_d);
		    be_Hf_E_dP1 = array_fetch(be_ptr, E_past_array, d+1);
		    assert(be_Hf_E_dP1);
		    be_result = Be_Implies(be_mgr, be_Hf_E_dP1, be_Hf_E_d);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, " [[Hf]]_E^%u => [[Hf]]_E^%u: ",
			      d+1, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }

	case OP_ONCE:
	  {
	    unsigned int  d;
	    
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);

	    if(do_optimization)
	      {
		/* Optimization:
		   Add [[Of]]_E^d => [[Of]]_E^d+1 */
		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Of_E_d, be_Of_E_dP1, be_result;
		    be_Of_E_d = array_fetch(be_ptr, E_past_array, d);
		    assert(be_Of_E_d);
		    be_Of_E_dP1 = array_fetch(be_ptr, E_past_array, d+1);
		    assert(be_Of_E_dP1);
		    be_result = Be_Implies(be_mgr, be_Of_E_d, be_Of_E_dP1);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, " [[Of]]_E^%u => [[Of]]_E^%u: ",
			      d, d+1);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }


	case OP_PREC:
	case OP_NOTPRECNOT:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);
	    break;
	  }

	case TRIGGERED:
	case SINCE:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    assert(info->trans_bes);
	    assert(L_past_array);
	    assert(E_past_array);
	    break;
	  }

	default:
	  print_node(stderr, node);
	  internal_error("%s:%d: TJ: Something not implemented",
			 __FILE__, __LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;
      
      /* Remove node from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Mark visited */
      _set_insert(visit_cache, node);


      /*
       * Constraints common to all subformulae
       */
      if(info->trans_vars)
	{
	  /***
	   * Formula variable translation
	   * Add !LoopExists => ([[f]]_L^d == FALSE)
	   */
	  int d;
	  for(d = 0; d <= info->past_depth; d++)
	    {
	      be_ptr be_f_L_d, be_result;
	      be_f_L_d = array_fetch(be_ptr, L_past_array, d);
	      assert(be_f_L_d);
	      be_result = Be_Implies(be_mgr,
				     Be_Not(be_mgr, be_LoopExists),
				     Be_Not(be_mgr, be_f_L_d));
	  
	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	      
	      if(opt_verbose_level_ge(options, 6)) {
		_my_print_node(nusmv_stderr, " f: ", node, "\n");
		fprintf(nusmv_stderr,
			"  !LoopExists => ([[f]]_L^%d <=> FALSE): ", d);
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }
	}

    }
      
  lsDestroy(unprocessed_nodes, 0);
  _set_destroy(visit_cache); visit_cache = 0;

  return created_constraints;
}



/*
 * Build InLoop_i stuff
 * Define InLoop_i = (InLoop_{i-1} || l_i)
 * Return the constraint InLoop_{i-1} => !l_i  (or 0 when i=0)
 */
static be_ptr _HJL_build_InLoop_i(const unsigned int i_model,
				  const BeEnc_ptr be_enc,
				  const HJL_state_vars_struct *HJL_state_vars,
				  array_t *InLoop_array)
{
  Be_Manager_ptr be_mgr;
  be_ptr be_l_i = 0;
  be_ptr be_InLoop_i = 0;
  be_ptr be_InLoop_i_minus_1 = 0;
  be_ptr be_constraint = 0;
  
  assert(be_enc);
  assert(HJL_state_vars);
  assert(InLoop_array);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  if(array_n(InLoop_array) <= i_model)
    array_insert(be_ptr, InLoop_array, i_model, 0);
    
  if(i_model == 0)
    {
      /* Build InLoop_0 := FALSE */
      be_InLoop_i = array_fetch(be_ptr, InLoop_array, i_model);
      if(be_InLoop_i == 0)
	{
	  be_InLoop_i = Be_Falsity(be_mgr);
	  array_insert(be_ptr, InLoop_array, i_model, be_InLoop_i);
	  if(opt_verbose_level_ge(options, 2))
	    fprintf(nusmv_stderr, "Constructed InLoop_%d := FALSE",
		    i_model, i_model-1, i_model);
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, ": ");
	    Be_DumpSexpr(be_mgr, be_InLoop_i, nusmv_stderr);
	  }
	  if(opt_verbose_level_ge(options, 2))
	    fprintf(nusmv_stderr, "\n");
	}
      return 0;
    }

  /* Get previous InLoop */
  be_InLoop_i_minus_1 = array_fetch(be_ptr, InLoop_array, i_model-1);
  assert(be_InLoop_i_minus_1 != 0);
    
  /* Get l_i */
  be_l_i = BeEnc_name_to_timed(be_enc,HJL_state_vars->l_var,_real_k(i_model));
  
  /* Get the constraint InLoop_{i-1} => !l_i */
  be_constraint = Be_Implies(be_mgr,
			     be_InLoop_i_minus_1,
			     Be_Not(be_mgr, be_l_i));
  
  if(opt_verbose_level_ge(options, 2))
    fprintf(nusmv_stderr, "Created the constraint (InLoop_%d => !l_%d)",
	    i_model-1, i_model);
  if(opt_verbose_level_ge(options, 6)) {
    fprintf(nusmv_stderr, ": ");
    Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
  }
  if(opt_verbose_level_ge(options, 2))
    fprintf(nusmv_stderr, "\n");

  /* Build InLoop_i := (InLoop_{i-1} || l_i) */
  be_InLoop_i = array_fetch(be_ptr, InLoop_array, i_model);
  if(be_InLoop_i == 0)
    {
      be_InLoop_i = Be_Or(be_mgr, be_InLoop_i_minus_1, be_l_i);
      array_insert(be_ptr, InLoop_array, i_model, be_InLoop_i);
    
      if(opt_verbose_level_ge(options, 2))
	fprintf(nusmv_stderr, "Constructed InLoop_%d := InLoop_%d | l_%d",
		i_model, i_model-1, i_model);
      if(opt_verbose_level_ge(options, 6)) {
	fprintf(nusmv_stderr, ": ");
	Be_DumpSexpr(be_mgr, be_InLoop_i, nusmv_stderr);
      }
      if(opt_verbose_level_ge(options, 2))
	fprintf(nusmv_stderr, "\n");
    }

  return be_constraint;
}





/*
 * Add and forces the constraints to be true in the current frame of the solver
 */
static void _force_constraint_list(Bmc_IncrSat_ptr solver, lsList constraints)
{
    lsGen  iterator;
    be_ptr be_constraint;

    lsForEachItem(constraints, iterator, be_constraint)
      {
	Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
      }
}





/*
 * Force SimplePath_{i,k} for each 0<=i<k
 */
static lsList _HJL_SimplePaths(const unsigned int k,
			       const BeEnc_ptr be_enc,
			       const HJL_state_vars_struct *HJL_state_vars,
			       array_t *InLoop_array)
{
  Be_Manager_ptr be_mgr;
  unsigned int j;
  lsList created_constraints = lsCreate();

  assert(be_enc);
  assert(InLoop_array);
  
  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  for(j = 0; j < k; j++)
    {
      be_ptr be_equal_pd0_translation_vars;
      be_ptr be_equal_pdx_translation_vars;
      be_ptr be_equal_aux_translation_vars;
      be_ptr be_InLoop_k, be_InLoop_j;
      be_ptr be_equal_InLoops;
      be_ptr be_constraint;
      be_ptr be_c2;
      
      be_ptr be_equal_system_vars =
	_equal_vectors_formula(be_enc,
			       HJL_state_vars->simple_path_system_vars,
			       _real_k(j), _real_k(k));

      be_equal_pd0_translation_vars =
	_equal_vectors_formula(be_enc,
			       HJL_state_vars->translation_vars_pd0,
			       _real_k(j), _real_k(k));
      
      be_equal_pdx_translation_vars =
	_equal_vectors_formula(be_enc,
			       HJL_state_vars->translation_vars_pdx,
			       _real_k(j), _real_k(k));
      
      be_equal_aux_translation_vars =
	_equal_vectors_formula(be_enc,
			       HJL_state_vars->translation_vars_aux,
			       _real_k(j), _real_k(k));
      
      be_InLoop_k = array_fetch(be_ptr, InLoop_array, k);
      assert(be_InLoop_k);
      
      be_InLoop_j = array_fetch(be_ptr, InLoop_array, j);
      assert(be_InLoop_j);
      
      be_equal_InLoops = Be_Iff(be_mgr, be_InLoop_j, be_InLoop_k);
      
      /* (s_j != s_k) */
      be_constraint = Be_Not(be_mgr, be_equal_system_vars);
      /* (s_j != s_k) | (InLoop_j != InLoop_k) */
      be_constraint = Be_Or(be_mgr,
			    be_constraint,
			    Be_Not(be_mgr, be_equal_InLoops));
      /* (s_j != s_k) | (InLoop_j != InLoop_k) | ([[f]]_j^0 != [[f]]_k^0) */
      be_constraint = Be_Or(be_mgr,
			    be_constraint,
			    Be_Not(be_mgr,
				   be_equal_pd0_translation_vars));

      be_c2 = Be_And(be_mgr,
		     Be_And(be_mgr, be_InLoop_j, be_InLoop_k),
		     Be_Or(be_mgr,
			   Be_Not(be_mgr,
				  be_equal_pdx_translation_vars),
			   Be_Not(be_mgr,
				  be_equal_aux_translation_vars)
			   )
		     );
      be_constraint = Be_Or(be_mgr, be_constraint, be_c2);
      
      lsNewEnd(created_constraints, be_constraint, LS_NH);
      
      if(opt_verbose_level_ge(options, 2)) {
	fprintf(nusmv_stderr, "Created SimplePath_{%d,%d}", j, k);
	if(opt_verbose_level_ge(options, 6)) {
	  fprintf(nusmv_stderr, ": ");
	  Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
	}
	fprintf(nusmv_stderr, "\n"); 
      }
    }
  return created_constraints;
}





/*
 * Create the k-invariant constraints for propositional operators at time i
 * Return a list of be_ptrs for the created constraints
 */
static lsList _HJL_unroll_invariant_propositional(const unsigned int i_model,
						  const node_ptr ltlspec,
						  const BeEnc_ptr be_enc,
						  const hash_ptr info_map,
						  const be_ptr be_InLoop_i,
						  const be_ptr be_l_i,
						  const int do_optimization)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  lsList   created_constraints = 0;
  Be_Manager_ptr be_mgr;
  const unsigned int i_real = _real_k(i_model);
  unsigned int d;
  
  /* Some verbose stuff */
  char *str_debug1 = " Translating formula ";
  char str_debug2[32];
  sprintf(str_debug2, " at timestep %u\n", i_model);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Create list for the constraints */
  created_constraints = lsCreate();

  /* Verbose output */
  if(opt_verbose_level_ge(options, 2))
    fprintf(nusmv_stderr,
	    "Unrolling k-invariant propositional stuff at time %u\n",
	    i_model);
  
  
  visit_cache = _set_create();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);
  
  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node, lsf, rsf;
      _HJL_node_info *info, *lsf_info, *rsf_info;
      array_t *past_array, *lsf_past_array, *rsf_past_array;
      int has_unprocessed_children;
      
      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get info */
      info = (_HJL_node_info*)find_assoc(info_map, node);
      if(!info)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get past_array, i.e., info->trans_bes[i] */
      /* _HJL_init_state_vector should have allocated trans_bes[i] before */
      assert(info->trans_bes);
      assert(array_n(info->trans_bes) >= i_real+1);
      past_array = array_fetch(array_t*, info->trans_bes, i_real);
      assert(past_array);

      /* Already translated? */
      if(_set_is_in(visit_cache, node))
	{
	  /* info->trans_bes[i][0...pd] should have been built */
	  for(d = 0; d <= info->past_depth; d++)
	    assert(array_fetch(be_ptr, past_array, d) != 0);
	  /* Remove node from unprocessed stack */
	  if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	    internal_error(_SNH_text,__FILE__,__LINE__);
	  continue;
	}


      /* Traverse children and make k-invariant constraints */
      lsf = car(node);
      rsf = cdr(node);      
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{

	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	  {
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    assert(info->past_depth == 0);

	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		assert(array_fetch(be_ptr, past_array, 0) != 0);
	      }
	    else
	      {
		/* State variable translation
		 * Build [[p]]_i^0 <=> p_i
		 */
		be_ptr be_p_i_0 = 0, be_realp_i = 0, be_result = 0;

		be_p_i_0 = array_fetch(be_ptr, past_array, 0);
		assert(be_p_i_0);
		be_realp_i = BeEnc_name_to_timed(be_enc, node, i_real);
		assert(be_realp_i);
		be_result = Be_Iff(be_mgr, be_p_i_0, be_realp_i);
		assert(be_result);

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[p]]_%d^0 <=> p_%d: ",
			  i_model, i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }	    
	    break;
	  }


	case TRUEEXP:
	  {
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    assert(info->past_depth == 0);

	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		assert(array_fetch(be_ptr, past_array, 0) != 0);
	      }
	    else
	      {
		/*
		 * Build [[TRUE]]_i^0 <=> TRUE
		 */
		be_ptr be_TRUE_i_0 = 0, be_true = 0, be_result = 0;

		be_TRUE_i_0 = array_fetch(be_ptr, past_array, 0);
		assert(be_TRUE_i_0);
		be_true = Be_Truth(be_mgr);
		assert(be_true);
		be_result = Be_Iff(be_mgr, be_TRUE_i_0, be_true);
		assert(be_result);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[TRUE]]_%d^0 <=> TRUE: ", i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }


	case FALSEEXP:
	  {
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    assert(info->past_depth == 0);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		assert(array_fetch(be_ptr, past_array, 0) != 0);
	      }
	    else
	      {
		/*
		 * Build [[FALSE]]_i^0 <=> FALSE
		 */
		be_ptr be_FALSE_i_0 = 0, be_false = 0, be_result = 0;

		be_FALSE_i_0 = array_fetch(be_ptr, past_array, 0);
		assert(be_FALSE_i_0);
		be_false = Be_Falsity(be_mgr);
		assert(be_false);		
		be_result = Be_Iff(be_mgr, be_FALSE_i_0, be_false);
		assert(be_result);

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
		
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr,"  [[FALSE]]_%d^0 <=> FALSE: ",i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }


	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  {
	    internal_error("%s:%d: Formula not in NNF\n",__FILE__,__LINE__);
	    break;
	  }



	case OR:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(array_t*,rsf_info->trans_bes,i_real);
	    assert(rsf_past_array);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		for(d = 0; d <= info->past_depth; d++)
		  assert(array_fetch(be_ptr, past_array, d) != 0);
	      }
	    else
	      {
		for(d = 0; d <= info->past_depth; d++)
		  {
		    /* [[f | g]]_i^d <=> [[f]]_i^d | [[g]]_i^d */
		    be_ptr be_fORg_i_d, be_f_i_d, be_g_i_d, be_result;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_fORg_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_fORg_i_d);
		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_g_i_d);
		    be_result = Be_Iff(be_mgr, be_fORg_i_d,
				       Be_Or(be_mgr, be_f_i_d, be_g_i_d));
		    assert(be_result);

		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr,
			      "  [[f | g]]_%d^%d <=> [[f]]_%d^%d | [[g]]_%d^%d: ",
			      i_model, d, i_model, d_lsf, i_model, d_rsf);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }


	case AND:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(array_t*,rsf_info->trans_bes,i_real);
	    assert(rsf_past_array);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		for(d = 0; d <= info->past_depth; d++)
		  assert(array_fetch(be_ptr, past_array, d) != 0);
	      }
	    else
	      {
		for(d = 0; d <= info->past_depth; d++)
		  {
		    /* [[f & g]]_i^d <=> [[f]]_i^d & [[g]]_i^d */
		    be_ptr be_fANDg_i_d, be_f_i_d, be_g_i_d, be_result;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_fANDg_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_fANDg_i_d);
		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_g_i_d);
		    be_result = Be_Iff(be_mgr,
				       be_fANDg_i_d,
				       Be_And(be_mgr, be_f_i_d, be_g_i_d));
		    assert(be_result);

		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[f & g]]_%d^%d <=> [[f]]_%d^%d & [[g]]_%d^%d: ",
			      i_model, d, i_model, d_lsf, i_model, d_rsf);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }



	case NOT:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    assert(info->past_depth == 0);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		for(d = 0; d <= info->past_depth; d++)
		  assert(array_fetch(be_ptr, past_array, d) != 0);
	      }
	    else
	      {
		/* State variable translation */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    /* Add [[!f]]_i^d <=> ![[f]]_i^d */
		    be_ptr be_notf_i_d, be_f_i_d, be_result;
		    be_notf_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_notf_i_d);
		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		    assert(be_f_i_d);
		    be_result = Be_Iff(be_mgr, be_notf_i_d,
				       Be_Not(be_mgr, be_f_i_d));
		    assert(be_result);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6))
		      fprintf(nusmv_stderr, "  [!f]]_%d^%d <=> ![[f]]_%d^%d: ",
			      i_model, d, i_model, d),
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
			fprintf(nusmv_stderr, "\n");
		  }
	      }

	    break;
	  }


	case OP_NEXT:
	case OP_FUTURE:
	case OP_GLOBAL:
	case OP_ONCE:
	case OP_HISTORICAL:
	case OP_PREC:
	case OP_NOTPRECNOT:
	  {
	    /* Unary temporal operators */
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;
	    
	    /* For each future and past formula f, info->trans_bes[i][d] should
	       already point to a be variable for [[f]]_i^d*/
	    for(d = 0; d <= info->past_depth; d++)
	      assert(array_fetch(be_ptr, past_array, d) != 0);
	    
	    /* Not a propositional operator, do nothing */
	    
	    break;
	  }


	case UNTIL:
	case RELEASES:
	case SINCE:
	case TRIGGERED:
	  {
	    /* Binary temporal operators */
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* For each future and past formula f, info->trans_bes[i][d] should
	       already point to a be variable for [[f]]_i^d */

	    for(d = 0; d <= info->past_depth; d++)
	      assert(array_fetch(be_ptr, past_array, d) != 0);

	    /* Not a propositional operator, do nothing */

	    break;
	  }

	default:
	  print_node(stderr, node);
	  internal_error(_SNYI_text,__FILE__,__LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;

      /* Do the auxiliary translations if necessary */
      if(info->aux_F_node)
	{
	  be_ptr be_aux_i, be_aux_i_minus_1, be_f_i_pd;
	  be_ptr be_result;

	  if(opt_verbose_level_ge(options, 6))
	    _my_print_node(nusmv_stderr, " Doing <<F f>> translation for ",
			   node, "\n");

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_F_node, i_real);
	  assert(be_aux_i);

	  if(i_model == 0)
	    {
	      /* <<F f>>_0 <=> FALSE */
	      be_result = Be_Iff(be_mgr, be_aux_i, Be_Falsity(be_mgr));

	      if(opt_verbose_level_ge(options, 6))
		fprintf(nusmv_stderr, "  <<F f>>_0 <=> FALSE: "),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }
	  else
	    {	    
	      /* <<F f>>_i <=> <<F f>>_{i-1} | (InLoop_i & [[f]]_i^PD(f)) */
	      be_aux_i_minus_1 = BeEnc_name_to_timed(be_enc,
						     info->aux_F_node,
						     i_real-1);
	      assert(be_aux_i_minus_1);
	      be_f_i_pd = array_fetch(be_ptr, past_array, info->past_depth);
	      assert(be_f_i_pd);
	      be_result = Be_Iff(be_mgr,
				 be_aux_i,
				 Be_Or(be_mgr,
				       be_aux_i_minus_1,
				       Be_And(be_mgr,
					      be_InLoop_i,
					      be_f_i_pd)));

	      if(opt_verbose_level_ge(options, 6))
		fprintf(nusmv_stderr,
			"  <<F f>>_%u <=> <<F f>>_%u | (InLoop_%u & [[f]]_%u^%d): ", i_model, i_model-1, i_model, i_model, info->past_depth),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }

	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	}
      if(info->aux_F_node && do_optimization)
	{
	  /* Optimization: add <<Ff>>_i => <<Ff>>_E */
	  be_ptr be_aux_i, be_aux_E;
	  be_ptr be_result;

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_F_node, i_real);
	  assert(be_aux_i);

	  be_aux_E = BeEnc_name_to_timed(be_enc, info->aux_F_node, _E_state());
	  assert(be_aux_E);

	  be_result = Be_Implies(be_mgr, be_aux_i, be_aux_E);

	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	  
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "  <<F f>>_%u <=> <<F f>>_E: ", i_model);
	    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}

      if(info->aux_G_node)
	{
	  be_ptr be_aux_i, be_aux_i_minus_1, be_f_i_pd;
	  be_ptr be_result;

	  if(opt_verbose_level_ge(options, 6))
	    _my_print_node(nusmv_stderr, " Doing <<G f>> translation for ",
			   node, "\n");

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_G_node, i_real);
	  assert(be_aux_i);

	  if(i_model == 0)
	    {
	      /* <<G f>>_0 <=> TRUE */
	      be_result = Be_Iff(be_mgr, be_aux_i, Be_Truth(be_mgr));

	      if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr, "  <<G f>>_0 <=> TRUE: "),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }
	  else
	    {	    
	      /* <<G f>>_i <=> <<G f>>_{i-1} & (!InLoop_i | [[f]]_i^PD(f)) */
	      be_aux_i_minus_1 = BeEnc_name_to_timed(be_enc,
						     info->aux_G_node,
						     i_real-1);
	      assert(be_aux_i_minus_1);
	      be_f_i_pd = array_fetch(be_ptr, past_array, info->past_depth);
	      assert(be_f_i_pd);
	      be_result = Be_Iff(be_mgr,
				 be_aux_i,
				 Be_And(be_mgr,
					be_aux_i_minus_1,
					Be_Or(be_mgr,
					      Be_Not(be_mgr, be_InLoop_i),
					      be_f_i_pd)));

	      if(opt_verbose_level_ge(options, 6))
		fprintf(nusmv_stderr, "  <<G f>>_%u <=> <<G f>>_%u & (!InLoop_%u | [[f]]_%u^%d): ", i_model, i_model-1, i_model, i_model, info->past_depth),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }
	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	}
      if(info->aux_G_node && do_optimization)
	{
	  /* Optimization: add <<Gf>>_E => <<Gf>>_i */
	  be_ptr be_aux_i, be_aux_E;
	  be_ptr be_result;

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_G_node, i_real);
	  assert(be_aux_i);

	  be_aux_E = BeEnc_name_to_timed(be_enc, info->aux_G_node, _E_state());
	  assert(be_aux_E);
	  
	  be_result = Be_Implies(be_mgr, be_aux_E, be_aux_i);

	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	  
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "  <<F f>>_E <=> <<F f>>_%u: ", i_model);
	    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}

      
      if(info->trans_vars)
	{
	  /*
	   * Add (l_i => ([[f]]_L^d <=> [[f]]_i^d)) for
	   * formula variable translated subformulae
	   */
	  unsigned int d;
	  array_t *f_L_past_array;

	  f_L_past_array =
	    array_fetch(array_t*, info->trans_bes, _L_state());
	  assert(f_L_past_array);
	  
	  if(opt_verbose_level_ge(options, 6))
	    _my_print_node(nusmv_stderr," Doing binding to L for ",node,"\n");
	  
	  for(d = 0; d <= info->past_depth; d++)
	    {
	      be_ptr be_f_L_d, be_f_i_d, be_result;

	      be_f_i_d = array_fetch(be_ptr, past_array, d);
	      assert(be_f_i_d);

	      be_f_L_d = array_fetch(be_ptr, f_L_past_array, d);
	      assert(be_f_L_d);

	      be_result = Be_Implies(be_mgr,
				     be_l_i,
				     Be_Iff(be_mgr, be_f_L_d, be_f_i_d));

	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

	      if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr,
			  "  l_%u => ([[f]]_L_%d <=> [[f]]_%u^%d): ",
			  i_model, d, i_model, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	    }
	}

      
      /* Remove node from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Mark visited */
      _set_insert(visit_cache, node);
    }
      
  lsDestroy(unprocessed_nodes, 0); unprocessed_nodes = 0;
  _set_destroy(visit_cache); visit_cache = 0;

  return created_constraints;
}



/*
 * Create the k-invariant constraints for future temporal operators at time i
 * Return a list of be_ptrs for the created constraints
 */
static lsList _HJL_unroll_invariant_f(const unsigned int i_model,
				      const node_ptr ltlspec,
				      const BeEnc_ptr be_enc,
				      const hash_ptr info_map,
				      const be_ptr be_InLoop_i,
				      const be_ptr be_l_i,
				      const int do_optimization)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  lsList   created_constraints = 0;
  Be_Manager_ptr be_mgr;
  const unsigned int i_real = _real_k(i_model);
  unsigned int d;

  /* Some verbose stuff */
  char *str_debug1 = " Translating formula ";
  char str_debug2[32];
  sprintf(str_debug2, " at timestep %u\n", i_model);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Create list for the constraints */
  created_constraints = lsCreate();

  /* Verbose output */
  if(opt_verbose_level_ge(options, 2))
    {
      fprintf(nusmv_stderr, "Unrolling k-invariant future stuff at time %u\n",
	      i_model);
      fflush(nusmv_stderr);
    }
  

  visit_cache = _set_create();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);
  
  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node, lsf, rsf;
      _HJL_node_info *info, *lsf_info, *rsf_info;
      array_t *past_array, *lsf_past_array, *rsf_past_array;
      int has_unprocessed_children;
      
      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get info */
      info = (_HJL_node_info*)find_assoc(info_map, node);
      if(!info)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get past_array, i.e., info->trans_bes[i] */
      /* _HJL_init_state_vector should have built trans_bes[i] before */
      assert(info->trans_bes);
      assert(array_n(info->trans_bes) >= i_real+1);
      past_array = array_fetch(array_t*, info->trans_bes, i_real);
      assert(past_array);


      /* Already translated? */
      if(_set_is_in(visit_cache, node))
	{
	  /* info->trans_bes[i][0...pd] should have been built */
	  for(d = 0; d <= info->past_depth; d++)
	    assert(array_fetch(be_ptr, past_array, d) != 0);
	  /* Remove node from unprocessed stack */
	  if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	    internal_error(_SNH_text,__FILE__,__LINE__);
	  continue;
	}


      /* Traverse children and make k-invariant constraints */
      lsf = car(node);
      rsf = cdr(node);      
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{

	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	  {
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    assert(info->past_depth == 0);

	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		assert(array_fetch(be_ptr, past_array, 0) != 0);
	      }
	    else
	      {
		/* State variable translation
		 * Build [[p]]_i^0 <=> p_i
		 */
		be_ptr be_p_i_0 = 0, be_realp_i = 0, be_result = 0;

		be_p_i_0 = array_fetch(be_ptr, past_array, 0);
		assert(be_p_i_0);
		be_realp_i = BeEnc_name_to_timed(be_enc, node, i_real);
		assert(be_realp_i);
		be_result = Be_Iff(be_mgr, be_p_i_0, be_realp_i);
		assert(be_result);

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[p]]_%d^0 <=> p_%d: ",
			  i_model, i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }	    
	    break;
	  }


	case TRUEEXP:
	  {
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    assert(info->past_depth == 0);

	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		assert(array_fetch(be_ptr, past_array, 0) != 0);
	      }
	    else
	      {
		/*
		 * Build [[TRUE]]_i^0 <=> TRUE
		 */
		be_ptr be_TRUE_i_0 = 0, be_true = 0, be_result = 0;

		be_TRUE_i_0 = array_fetch(be_ptr, past_array, 0);
		assert(be_TRUE_i_0);
		be_true = Be_Truth(be_mgr);
		assert(be_true);
		be_result = Be_Iff(be_mgr, be_TRUE_i_0, be_true);
		assert(be_result);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[TRUE]]_%d^0 <=> TRUE: ", i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }


	case FALSEEXP:
	  {
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    assert(info->past_depth == 0);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		assert(array_fetch(be_ptr, past_array, 0) != 0);
	      }
	    else
	      {
		/*
		 * Build [[FALSE]]_i^0 <=> FALSE
		 */
		be_ptr be_FALSE_i_0 = 0, be_false = 0, be_result = 0;

		be_FALSE_i_0 = array_fetch(be_ptr, past_array, 0);
		assert(be_FALSE_i_0);
		be_false = Be_Falsity(be_mgr);
		assert(be_false);		
		be_result = Be_Iff(be_mgr, be_FALSE_i_0, be_false);
		assert(be_result);

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
		
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr,"  [[FALSE]]_%d^0 <=> FALSE: ",i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }
	    break;
	  }


	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  {
	    internal_error("%s:%d: Formula not in NNF\n",__FILE__,__LINE__);
	    break;
	  }



	case OR:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(array_t*,rsf_info->trans_bes,i_real);
	    assert(rsf_past_array);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		for(d = 0; d <= info->past_depth; d++)
		  assert(array_fetch(be_ptr, past_array, d) != 0);
	      }
	    else
	      {
		for(d = 0; d <= info->past_depth; d++)
		  {
		    /* [[f | g]]_i^d <=> [[f]]_i^d | [[g]]_i^d */
		    be_ptr be_fORg_i_d, be_f_i_d, be_g_i_d, be_result;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_fORg_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_fORg_i_d);
		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_g_i_d);
		    be_result = Be_Iff(be_mgr, be_fORg_i_d,
				       Be_Or(be_mgr, be_f_i_d, be_g_i_d));
		    assert(be_result);

		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr,
			      "  [[f | g]]_%d^%d <=> [[f]]_%d^%d | [[g]]_%d^%d: ",
			      i_model, d, i_model, d_lsf, i_model, d_rsf);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }


	case AND:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(array_t*,rsf_info->trans_bes,i_real);
	    assert(rsf_past_array);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		for(d = 0; d <= info->past_depth; d++)
		  assert(array_fetch(be_ptr, past_array, d) != 0);
	      }
	    else
	      {
		for(d = 0; d <= info->past_depth; d++)
		  {
		    /* [[f & g]]_i^d <=> [[f]]_i^d & [[g]]_i^d */
		    be_ptr be_fANDg_i_d, be_f_i_d, be_g_i_d, be_result;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_fANDg_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_fANDg_i_d);
		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_g_i_d);
		    be_result = Be_Iff(be_mgr,
				       be_fANDg_i_d,
				       Be_And(be_mgr, be_f_i_d, be_g_i_d));
		    assert(be_result);

		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[f & g]]_%d^%d <=> [[f]]_%d^%d & [[g]]_%d^%d: ",
			      i_model, d, i_model, d_lsf, i_model, d_rsf);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }



	case NOT:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    assert(info->past_depth == 0);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    if(info->trans_vars == 0)
	      {
		/* Definitional translation already done by
		   state vector init */
		for(d = 0; d <= info->past_depth; d++)
		  assert(array_fetch(be_ptr, past_array, d) != 0);
	      }
	    else
	      {
		/* State variable translation */
		for(d = 0; d <= info->past_depth; d++)
		  {
		    /* Add [[!f]]_i^d <=> ![[f]]_i^d */
		    be_ptr be_notf_i_d, be_f_i_d, be_result;
		    be_notf_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_notf_i_d);
		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		    assert(be_f_i_d);
		    be_result = Be_Iff(be_mgr, be_notf_i_d,
				       Be_Not(be_mgr, be_f_i_d));
		    assert(be_result);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6))
		      fprintf(nusmv_stderr, "  [!f]]_%d^%d <=> ![[f]]_%d^%d: ",
			      i_model, d, i_model, d),
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
			fprintf(nusmv_stderr, "\n");
		  }
	      }

	    break;
	  }


	case OP_NEXT:
	  {
	    array_t*      lsf_next_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_next_past_array = array_fetch(array_t*, lsf_info->trans_bes,
					      i_real + 1);
	    assert(lsf_next_past_array);
	    
	    for(d = 0; d <= info->past_depth; d++)
	      {
		/* Add [[Xf]]_i^d <=> [[f]]_{i+1}^d */
		be_ptr be_aux, be_lsf_next, be_result;
		be_aux = array_fetch(be_ptr, past_array, d);
		assert(be_aux);
		be_lsf_next = array_fetch(be_ptr, lsf_next_past_array, d);
		assert(be_lsf_next);
		be_result = Be_Iff(be_mgr, be_aux, be_lsf_next);
		assert(be_result);

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr, "  [Xf]]_%d^%d <=> [[f]]_%d^%d: ",
			  i_model, d, i_model+1, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	      }
	    break;
	  }


	case OP_FUTURE:
	  {
	    array_t*      next_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);
	    
	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);
	    
	    next_past_array = array_fetch(array_t*, info->trans_bes,
					  i_real+1);
	    assert(next_past_array);
	    
	    for(d = 0; d <= info->past_depth; d++)
	      {
		/* Add [[F f]]_i^d <=> [[f]]_i^d | [[F f]]_{i+1}^d */
		be_ptr be_aux, be_lsf, be_aux_next, be_result;
		be_aux = array_fetch(be_ptr, past_array, d);
		assert(be_aux);
		be_lsf = array_fetch(be_ptr, lsf_past_array, d);
		assert(be_lsf);
		be_aux_next = array_fetch(be_ptr, next_past_array, d);
		assert(be_aux_next);
		be_result = Be_Iff(be_mgr, be_aux,
				   Be_Or(be_mgr, be_lsf, be_aux_next));

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr,
			  " [[F f]]_%d^%d <=> [[f]]_%d^%d | [[F f]]_%d^%d: ",
			  i_model, d, i_model, d, i_model+1, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	      }


	    if(do_optimization)
	      {
		/* Optimization: <<Ff>>_i => [[Ff]]_i^pd(Ff) */
		be_ptr be_Ff_i_pd, be_auxFf_i, be_result;

		assert(lsf_info->aux_F_node);
		be_auxFf_i = BeEnc_name_to_timed(be_enc,
						 lsf_info->aux_F_node,
						 i_real);
		assert(be_auxFf_i);

		be_Ff_i_pd = array_fetch(be_ptr, past_array, info->past_depth);
		assert(be_Ff_i_pd);
		
		be_result = Be_Implies(be_mgr, be_auxFf_i, be_Ff_i_pd);
		    
		/* Save the created constraint */
		lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  <<Ff>>_%u => [[Ff]]_%u^%u: ",
			  i_model, i_model, info->past_depth);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }


	    if(do_optimization)
	      {
		/* Optimization: InLoop_i => ([[Ff]]_i^d+1 => [[Ff]]_i^d) */
		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Ff_i_dP1, be_Ff_i_d, be_result;
		    
		    be_Ff_i_dP1 = array_fetch(be_ptr, past_array, d+1);
		    assert(be_Ff_i_dP1);
		    
		    be_Ff_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_Ff_i_d);
		    
		    be_result = Be_Implies(be_mgr,
					   be_InLoop_i,
					   Be_Implies(be_mgr,
						      be_Ff_i_dP1,
						      be_Ff_i_d));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6))
		      fprintf(nusmv_stderr, "  InLoop_%u => ([[Ff]]_%u^%d => [[Ff]]_%u^%d): ",
			      i_model, i_model, d+1, i_model, d),
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
			fprintf(nusmv_stderr, "\n");
		  }
	      }
	    

	    if(do_optimization)
	      {
		/* Optimization: [[Ff]]_E^d => [[Ff]]_i^d */
		array_t* past_array_Ff_E =
		  array_fetch(array_t*, info->trans_bes, _E_state());
		assert(past_array_Ff_E);

		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Ff_i_d, be_Ff_E_d, be_result;
		    
		    be_Ff_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_Ff_i_d);
		    
		    be_Ff_E_d = array_fetch(be_ptr, past_array_Ff_E, d);
		    assert(be_Ff_E_d);
		    
		    be_result = Be_Implies(be_mgr, be_Ff_E_d, be_Ff_i_d);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[Ff]]_E^%u => [[Ff]]_%u^%u: ",
			      d, i_model, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }


	case OP_GLOBAL:
	  {
	    array_t*      next_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(array_t*,lsf_info->trans_bes,i_real);
	    assert(lsf_past_array);

	    next_past_array = array_fetch(array_t*,info->trans_bes,i_real+1);
	    assert(next_past_array);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		/* Add [[Gf]]_i^d <=> [[f]]_i^d & [[Gf]]_{i+1}^d */
		be_ptr be_Gf_i_d, be_f_i_d, be_Gf_ip1_d, be_result;
		be_Gf_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_Gf_i_d);
		be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		assert(be_f_i_d);
		be_Gf_ip1_d = array_fetch(be_ptr, next_past_array, d);
		assert(be_Gf_ip1_d);
		be_result = Be_Iff(be_mgr, be_Gf_i_d,
				   Be_And(be_mgr, be_f_i_d, be_Gf_ip1_d));

		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr,
			  "  [[Gf]]_%u^%d <=> [[f]]_%u^%d & [[Gf]]_%u^%d: ",
			  i_model, d, i_model, d, i_model+1, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	      }


	    if(do_optimization)
	      {
		/* Optimization: [[Gf]]_i^pd(Ff) => <<Gf>>_i */
		be_ptr be_Gf_i_pd, be_auxGf_i, be_result;

		assert(lsf_info->aux_G_node);
		be_auxGf_i = BeEnc_name_to_timed(be_enc,
						 lsf_info->aux_G_node,
						 i_real);
		assert(be_auxGf_i);

		be_Gf_i_pd = array_fetch(be_ptr, past_array, info->past_depth);
		assert(be_Gf_i_pd);
		
		be_result = Be_Implies(be_mgr, be_Gf_i_pd, be_auxGf_i);
		    
		/* Save the created constraint */
		lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		
		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  [[Gf]]_%u^%u => <<Gf>>_%u: ",
			  i_model, info->past_depth, i_model);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }


	    if(do_optimization)
	      {
		/* Optimization: InLoop_i => ([[Gf]]_i^d => [[Gf]]_i^d+1) */
		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Gf_i_d, be_Gf_i_dP1, be_result;
		    
		    be_Gf_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_Gf_i_d);
		    
		    be_Gf_i_dP1 = array_fetch(be_ptr, past_array, d+1);
		    assert(be_Gf_i_dP1);
		    
		    be_result = Be_Implies(be_mgr,
					   be_InLoop_i,
					   Be_Implies(be_mgr,
						      be_Gf_i_d,
						      be_Gf_i_dP1));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6))
		      fprintf(nusmv_stderr, "  InLoop_%u => ([[Gf]]_%u^%d => [[Gf]]_%u^%d): ",
			      i_model, i_model, d, i_model, d+1),
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
			fprintf(nusmv_stderr, "\n");
		  }
	      }


	    if(do_optimization)
	      {
		/* Optimization: [[Gf]]_i^d => [[Gf]]_E^d */
		array_t* past_array_Gf_E =
		  array_fetch(array_t*, info->trans_bes, _E_state());
		assert(past_array_Gf_E);

		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Gf_i_d, be_Gf_E_d, be_result;
		    
		    be_Gf_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_Gf_i_d);
		    
		    be_Gf_E_d = array_fetch(be_ptr, past_array_Gf_E, d);
		    assert(be_Gf_E_d);
		    
		    be_result = Be_Implies(be_mgr, be_Gf_i_d, be_Gf_E_d);
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  [[Gf]]_%u^%u => [[Gf]]_E^%u: ",
			      i_model, d, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }
	    break;
	  }


	case UNTIL:
	  {
	    array_t*      next_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;
	    
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(be_ptr, rsf_info->trans_bes, i_real);
	    assert(rsf_past_array);
	    
	    next_past_array = array_fetch(array_t*, info->trans_bes, i_real+1);
	    assert(next_past_array);
	    
	    for(d = 0; d <= info->past_depth; d++)
	      {
		/* Add [[f U g]]_i^d <=>
		 *     [[g]]_i^d | ([[f]]_i^d & [[f U g]]_{i+1}^d)
		 */
		be_ptr be_fUg_i_d, be_f_i_d, be_g_i_d, be_fUg_iP1_d, be_result;

		be_fUg_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_fUg_i_d);
		be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		assert(be_f_i_d);
		be_g_i_d = array_fetch(be_ptr, rsf_past_array, d);
		assert(be_g_i_d);
		be_fUg_iP1_d = array_fetch(be_ptr, next_past_array, d);
		assert(be_fUg_iP1_d);
		be_result = Be_Iff(be_mgr,
				   be_fUg_i_d,
				   Be_Or(be_mgr,
					 be_g_i_d,
					 Be_And(be_mgr,
						be_f_i_d,
						be_fUg_iP1_d)));
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr,
			  " [[f U g]]_%u^%d <=>"
			  " [[g]]_%u^%d | ([[f]]_%u^%d & [[f U g]]_%u^%d): ",
			  i_model, d, i_model, d, i_model, d, i_model+1, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	      }
	    
	    break;
	  }


	case RELEASES:
	  {
	    array_t*      next_past_array;

	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;
	    	
	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);
	    
	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(be_ptr, rsf_info->trans_bes, i_real);
	    assert(rsf_past_array);
	    
	    next_past_array = array_fetch(array_t*,info->trans_bes,i_real+1);
	    assert(next_past_array);
	    
	    for(d = 0; d <= info->past_depth; d++)
	      {
		/* Add [[f R g]]_i^d <=>
		 *     [[g]]_i^d & ([[f]]_i^d | [[f R g]]_{i+1}^d)
		 */
		be_ptr be_fRg_i_d, be_f_i_d, be_g_i_d, be_fRg_iP1_d, be_result;

		be_fRg_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_fRg_i_d);
		be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		assert(be_f_i_d);
		be_g_i_d = array_fetch(be_ptr, rsf_past_array, d);
		assert(be_g_i_d);
		be_fRg_iP1_d = array_fetch(be_ptr, next_past_array, d);
		assert(be_fRg_iP1_d);
		be_result = Be_Iff(be_mgr,
				   be_fRg_i_d,
				   Be_And(be_mgr,
					  be_g_i_d,
					  Be_Or(be_mgr,
						be_f_i_d,
						be_fRg_iP1_d)));
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr,
			  " [[f R g]]_%u^%d <=>"
			  " [[g]]_%u^%d & ([[f]]_%u^%d | [[f R g]]_%u^%d): ",
			  i_model, d, i_model, d, i_model, d, i_model+1, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	      }

	    break;
	  }


	case OP_ONCE:
	case OP_HISTORICAL:
	case OP_PREC:
	case OP_NOTPRECNOT:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* For each past operator, info->trans_bes[i][d] should
	       point to a be variable for [[f]]_i^d */
	    for(d = 0; d <= info->past_depth; d++)
	      assert(array_fetch(be_ptr, past_array, d) != 0);

	    /* Not a future operator, do nothing */

	    break;
	  }


	case SINCE:
	case TRIGGERED:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* For each past operator, info->trans_bes[i][d] should
	       point to a be variable for [[f]]_i^d */

	    for(d = 0; d <= info->past_depth; d++)
	      assert(array_fetch(be_ptr, past_array, d) != 0);

	    /* Not a future operator, do nothing */

	    break;
	  }


	default:
	  print_node(stderr, node);
	  internal_error(_SNYI_text,__FILE__,__LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;

      /* Do the auxiliary translations if necessary */
      if(info->aux_F_node)
	{
	  be_ptr be_aux_i, be_aux_i_minus_1, be_f_i_pd;
	  be_ptr be_result;

	  if(opt_verbose_level_ge(options, 6))
	    _my_print_node(nusmv_stderr, " Doing <<F f>> translation for ",
			   node, "\n");

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_F_node, i_real);
	  assert(be_aux_i);

	  if(i_model == 0)
	    {
	      /* <<F f>>_0 <=> FALSE */
	      be_result = Be_Iff(be_mgr, be_aux_i, Be_Falsity(be_mgr));

	      if(opt_verbose_level_ge(options, 6))
		fprintf(nusmv_stderr, "  <<F f>>_0 <=> FALSE: "),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }
	  else
	    {	    
	      /* <<F f>>_i <=> <<F f>>_{i-1} | (InLoop_i & [[f]]_i^PD(f)) */
	      be_aux_i_minus_1 = BeEnc_name_to_timed(be_enc,
						     info->aux_F_node,
						     i_real-1);
	      assert(be_aux_i_minus_1);
	      be_f_i_pd = array_fetch(be_ptr, past_array, info->past_depth);
	      assert(be_f_i_pd);
	      be_result = Be_Iff(be_mgr,
				 be_aux_i,
				 Be_Or(be_mgr,
				       be_aux_i_minus_1,
				       Be_And(be_mgr,
					      be_InLoop_i,
					      be_f_i_pd)));

	      if(opt_verbose_level_ge(options, 6))
		fprintf(nusmv_stderr,
			"  <<F f>>_%u <=> <<F f>>_%u | (InLoop_%u & [[f]]_%u^%d): ", i_model, i_model-1, i_model, i_model, info->past_depth),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }

	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	}
      if(info->aux_F_node && do_optimization)
	{
	  /* Optimization: add <<Ff>>_i => <<Ff>>_E */
	  be_ptr be_aux_i, be_aux_E;
	  be_ptr be_result;

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_F_node, i_real);
	  assert(be_aux_i);

	  be_aux_E = BeEnc_name_to_timed(be_enc, info->aux_F_node, _E_state());
	  assert(be_aux_E);

	  be_result = Be_Implies(be_mgr, be_aux_i, be_aux_E);

	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	  
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "  <<F f>>_%u <=> <<F f>>_E: ", i_model);
	    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}

      if(info->aux_G_node)
	{
	  be_ptr be_aux_i, be_aux_i_minus_1, be_f_i_pd;
	  be_ptr be_result;

	  if(opt_verbose_level_ge(options, 6))
	    _my_print_node(nusmv_stderr, " Doing <<G f>> translation for ",
			   node, "\n");
	  
	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_G_node, i_real);
	  assert(be_aux_i);

	  if(i_model == 0)
	    {
	      /* <<G f>>_0 <=> TRUE */
	      be_result = Be_Iff(be_mgr, be_aux_i, Be_Truth(be_mgr));

	      if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr, "  <<G f>>_0 <=> TRUE: "),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }
	  else
	    {	    
	      /* <<G f>>_i <=> <<G f>>_{i-1} & (!InLoop_i | [[f]]_i^PD(f)) */
	      be_aux_i_minus_1 = BeEnc_name_to_timed(be_enc,
						     info->aux_G_node,
						     i_real-1);
	      assert(be_aux_i_minus_1);
	      be_f_i_pd = array_fetch(be_ptr, past_array, info->past_depth);
	      assert(be_f_i_pd);
	      be_result = Be_Iff(be_mgr,
				 be_aux_i,
				 Be_And(be_mgr,
					be_aux_i_minus_1,
					Be_Or(be_mgr,
					      Be_Not(be_mgr, be_InLoop_i),
					      be_f_i_pd)));

	      if(opt_verbose_level_ge(options, 6))
		fprintf(nusmv_stderr, "  <<G f>>_%u <=> <<G f>>_%u & (!InLoop_%u | [[f]]_%u^%d): ", i_model, i_model-1, i_model, i_model, info->past_depth),
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		  fprintf(nusmv_stderr, "\n");
	    }
	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	}
      if(info->aux_G_node && do_optimization)
	{
	  /* Optimization: add <<Gf>>_E => <<Gf>>_i */
	  be_ptr be_aux_i, be_aux_E;
	  be_ptr be_result;

	  be_aux_i = BeEnc_name_to_timed(be_enc, info->aux_G_node, i_real);
	  assert(be_aux_i);

	  be_aux_E = BeEnc_name_to_timed(be_enc, info->aux_G_node, _E_state());
	  assert(be_aux_E);
	  
	  be_result = Be_Implies(be_mgr, be_aux_E, be_aux_i);

	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	  
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "  <<F f>>_E <=> <<F f>>_%u: ", i_model);
	    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}

      
      if(info->trans_vars)
	{
	  /*
	   * Add (l_i => ([[f]]_L^d <=> [[f]]_i^d)) for
	   * formula variable translated subformulae
	   */
	  unsigned int d;
	  array_t *f_L_past_array;

	  f_L_past_array =
	    array_fetch(array_t*, info->trans_bes, _L_state());
	  assert(f_L_past_array);
	  
	  if(opt_verbose_level_ge(options, 6))
	    _my_print_node(nusmv_stderr," Doing binding to L for ",node,"\n");
	  
	  for(d = 0; d <= info->past_depth; d++)
	    {
	      be_ptr be_f_L_d, be_f_i_d, be_result;

	      be_f_i_d = array_fetch(be_ptr, past_array, d);
	      assert(be_f_i_d);

	      be_f_L_d = array_fetch(be_ptr, f_L_past_array, d);
	      assert(be_f_L_d);

	      be_result = Be_Implies(be_mgr,
				     be_l_i,
				     Be_Iff(be_mgr, be_f_L_d, be_f_i_d));

	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

	      if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr,
			  "  l_%u => ([[f]]_L_%d <=> [[f]]_%u^%d): ",
			  i_model, d, i_model, d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
		    fprintf(nusmv_stderr, "\n");
	    }
	}

      
      /* Remove node from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Mark visited */
      _set_insert(visit_cache, node);
    }
      
  lsDestroy(unprocessed_nodes, 0); unprocessed_nodes = 0;
  _set_destroy(visit_cache); visit_cache = 0;

  return created_constraints;
}



/*
 * Create the k-invariant constraints at time i
 * Return a list of be_ptrs for the created constraints
 */
static lsList _HJL_unroll_invariant_p(const unsigned int i_model,
				      const node_ptr ltlspec,
				      const BeEnc_ptr be_enc,
				      const hash_ptr info_map,
				      const be_ptr be_InLoop_i,
				      const be_ptr be_l_i,
				      const int do_optimization)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  lsList   created_constraints = 0;
  Be_Manager_ptr be_mgr;
  const unsigned int i_real = _real_k(i_model);
  unsigned int d;

  /* Some verbose stuff */
  char *str_debug1 = " Translating formula ";
  char str_debug2[32];
  sprintf(str_debug2, " at timestep %u\n", i_model);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Create list for the constraints */
  created_constraints = lsCreate();

  /* Debug output */
  if(opt_verbose_level_gt(options, 1))
    {
      fprintf(nusmv_stderr, "Unrolling k-invariant past stuff at time %u\n",
	      i_model);
      fflush(nusmv_stderr);
    }
  

  visit_cache = _set_create();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);
  
  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node, lsf, rsf;
      _HJL_node_info *info, *lsf_info, *rsf_info;
      array_t *past_array, *lsf_past_array, *rsf_past_array;
      int has_unprocessed_children;
      
      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get info */
      info = (_HJL_node_info*)find_assoc(info_map, node);
      if(!info)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get past_array */
      /* _HJL_init_state_vector should have built trans_bes[i] before */
      assert(info->trans_bes);
      assert(array_n(info->trans_bes) >= i_real+1);
      past_array = array_fetch(array_t*, info->trans_bes, i_real);
      assert(past_array);


      /* Already translated? */
      if(_set_is_in(visit_cache, node))
	{
	  for(d = 0; d <= info->past_depth; d++)
	    assert(array_fetch(be_ptr, past_array, d) != 0);
	  /* Remove node from unprocessed stack */
	  if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	    internal_error(_SNH_text,__FILE__,__LINE__);
	  continue;
	}


      /* Traverse children and make k-invariant constraints */
      lsf = car(node);
      rsf = cdr(node);      
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{
	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	  {
	    assert(info->past_depth == 0);
	    break;

	    /* Not a past time operator, do nothing */
	  }


	case TRUEEXP:
	case FALSEEXP:
	  {
	    assert(info->past_depth == 0);
	    break;

	    /* Not a past time operator, do nothing */
	  }


	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  {
	    internal_error("%s:%d: Formula not in NNF\n",__FILE__,__LINE__);
	    break;
	  }


	case OR:
	case AND:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* Not a past time operator, do nothing */
	    break;
	  }


	case NOT:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* Not a past time operator, do nothing */
	    break;
	  }


	case OP_NEXT:
	case OP_FUTURE:
	case OP_GLOBAL:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* Not a past time operator, do nothing */
	    break;
	  }


	case UNTIL:
	case RELEASES:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    /* Not a past time operator, do nothing */
	    break;
	  }


	case OP_ONCE:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		be_ptr be_Of_i_d, be_result;
		const unsigned int d_lsf = min(d, lsf_info->past_depth);

		be_Of_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_Of_i_d);

		if(i_model == 0)
		  {
		    /* Add [[O f]]_0^d <=> [[f]]_0^0 */
		    be_ptr be_f_0_0 = array_fetch(be_ptr, lsf_past_array, 0);
		    assert(be_f_0_0);
		    be_result = Be_Iff(be_mgr, be_Of_i_d, be_f_0_0);
		  }
		else if(i_model >= 1 && d == 0)
		  {
		    /* Add [[O f]]_i^0 <=> [[f]]_i^0 | [[O f]]_{i-1}^0*/
		    be_ptr be_f_i_d, be_Of_iM1_d;
		    array_t* past_array_Of_iM1;

		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);

		    past_array_Of_iM1 = array_fetch(array_t*, info->trans_bes,
						    i_real - 1);
		    assert(past_array_Of_iM1);
		    be_Of_iM1_d = array_fetch(be_ptr, past_array_Of_iM1, d);
		    assert(be_Of_iM1_d);

		    be_result = Be_Iff(be_mgr,
				       be_Of_i_d,
				       Be_Or(be_mgr, be_f_i_d, be_Of_iM1_d));
		  }
		else if(i_model >= 1 && d > 0)
		  {
		    /*
		     * Add [[O f]]_i^d <=>
		     * [[f]]_i^d | ITE(l_i,[[O f]]_E^{d-1},[[O f]]_{i-1}^d)
		     */
		    be_ptr be_prev, be_E, be_f_i_d;
		    array_t* prev_past_array;
		    array_t* E_past_array;

		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);

		    prev_past_array =
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(prev_past_array);
		    be_prev = array_fetch(be_ptr, prev_past_array, d);
		    assert(be_prev);
		    
		    E_past_array =
		      array_fetch(array_t*, info->trans_bes, _E_state());
		    assert(E_past_array);
		    be_E = array_fetch(be_ptr, E_past_array, d-1);
		    assert(be_E);
		    
		    be_result = Be_Iff(be_mgr,
				       be_Of_i_d,
				       Be_Or(be_mgr,
					     be_f_i_d,
					     Be_Ite(be_mgr,
						    be_l_i, be_E, be_prev)));
		  }
		else
		  internal_error(_SNH_text,__FILE__,__LINE__);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  Adding constraint at depth %u: ",d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }

	    if(do_optimization)
	      {
		/* Optimization: InLoop_i => ([[Of]]_i^d => [[Of]]_i^d+1) */
		/* Optimization: InLoop_i => ([[Of]]_i^d => [[Of]]_E^d) */
		array_t* past_array_Of_E =
		  array_fetch(array_t*, info->trans_bes, _E_state());
		assert(past_array_Of_E);

		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Of_i_d, be_Of_i_dP1, be_Of_E_d, be_result;
		    
		    be_Of_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_Of_i_d);

		    be_Of_i_dP1 = array_fetch(be_ptr, past_array, d+1);
		    assert(be_Of_i_dP1);
		    
		    be_result = Be_Implies(be_mgr,
					   be_InLoop_i,
					   Be_Implies(be_mgr,
						      be_Of_i_d,
						      be_Of_i_dP1));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6))
		      fprintf(nusmv_stderr, "  InLoop_%u => ([[Of]]_%d^%d => [[Of]]_%d^%d): ",
			      i_model, i_model, d, i_model, d+1),
			Be_DumpSexpr(be_mgr, be_result, nusmv_stderr),
			fprintf(nusmv_stderr, "\n");


		    be_Of_E_d = array_fetch(be_ptr, past_array_Of_E, d);
		    assert(be_Of_E_d);
		    be_result = Be_Implies(be_mgr,
					   be_InLoop_i,
					   Be_Implies(be_mgr,
						      be_Of_i_d,
						      be_Of_E_d));

		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);

		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  InLoop_%u => ([[Of]]_%d^%d => [[Of]]_E^%d): ",
			      i_model, i_model, d, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }


	case OP_HISTORICAL:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		be_ptr be_Hf_i_d, be_lsf, be_result;
		const unsigned int d_lsf = min(d, lsf_info->past_depth);

		be_Hf_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_Hf_i_d);

		be_lsf = array_fetch(be_ptr, lsf_past_array, d_lsf);
		assert(be_lsf);

		if(i_model == 0)
		  {
		    /* Add [[H f]]_0^0 <=> [[f]]_0^0 */
		    be_ptr be_f_0_0 = array_fetch(be_ptr, lsf_past_array, 0);
		    assert(be_f_0_0);
		    be_result = Be_Iff(be_mgr, be_Hf_i_d, be_f_0_0);
		  }
		else if(i_model >= 1 && d == 0)
		  {
		    /* Add [[H f]]_i^0 <=> [[f]]_i^0 & [[H f]]_{i-1}^0*/
		    be_ptr be_prev;
		    array_t* prev_past_array =
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(prev_past_array);
		    be_prev = array_fetch(be_ptr, prev_past_array, d);
		    assert(be_prev);
		    be_result = Be_Iff(be_mgr,
				       be_Hf_i_d,
				       Be_And(be_mgr, be_lsf, be_prev));
		  }
		else if(i_model >= 1 && d > 0)
		  {
		    /*
		     * Add [[H f]]_i^d <=>
		     * [[f]]_i^d & ITE(l_i,[[H f]]_E^{d-1},[[H f]]_{i-1}^d)
		     */
		    be_ptr be_prev, be_E;
		    array_t* prev_past_array;
		    array_t* E_past_array;

		    prev_past_array =
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(prev_past_array);
		    be_prev = array_fetch(be_ptr, prev_past_array, d);
		    assert(be_prev);
		    
		    E_past_array =
		      array_fetch(array_t*, info->trans_bes, _E_state());
		    assert(E_past_array);
		    be_E = array_fetch(be_ptr, E_past_array, d-1);
		    assert(be_E);
		    
		    be_result = Be_Iff(be_mgr,
				       be_Hf_i_d,
				       Be_And(be_mgr,
					      be_lsf,
					      Be_Ite(be_mgr,
						     be_l_i, be_E, be_prev)));
		  }
		else
		  internal_error(_SNH_text,__FILE__,__LINE__);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  Adding constraint at depth %u: ",d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }

	    if(do_optimization)
	      {
		/* Optimization: InLoop_i => ([[Hf]]_i^d+1 => [[Hf]]_i^d) */
		/* Optimization: InLoop_i => ([[Hf]]_E^d => [[Hf]]_i^d) */
		array_t* past_array_Hf_E =
		  array_fetch(array_t*, info->trans_bes, _E_state());
		assert(past_array_Hf_E);

		for(d = 0; d < info->past_depth; d++)
		  {
		    be_ptr be_Hf_i_d, be_Hf_i_dP1, be_Hf_E_d, be_result;
		    
		    be_Hf_i_d = array_fetch(be_ptr, past_array, d);
		    assert(be_Hf_i_d);
		    
		    be_Hf_i_dP1 = array_fetch(be_ptr, past_array, d+1);
		    assert(be_Hf_i_dP1);
		    
		    be_result = Be_Implies(be_mgr,
					   be_InLoop_i,
					   Be_Implies(be_mgr,
						      be_Hf_i_dP1,
						      be_Hf_i_d));
		    
		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  InLoop_%u => ([[Hf]]_%d^%d => [[Hf]]_%d^%d): ",
			      i_model, i_model, d+1, i_model, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }


		    be_Hf_E_d = array_fetch(be_ptr, past_array_Hf_E, d);
		    assert(be_Hf_E_d);

		    be_result = Be_Implies(be_mgr,
					   be_InLoop_i,
					   Be_Implies(be_mgr,
						      be_Hf_E_d,
						      be_Hf_i_d));

		    /* Save the created constraint */
		    lsNewBegin(created_constraints,(lsGeneric)be_result,LS_NH);
		    
		    if(opt_verbose_level_ge(options, 6)) {
		      fprintf(nusmv_stderr, "  InLoop_%u => ([[Hf]]_E^%d => [[Hf]]_%d^%d): ",
			      i_model, d, i_model, d);
		      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		      fprintf(nusmv_stderr, "\n");
		    }
		  }
	      }

	    break;
	  }


	case SINCE:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);

	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(be_ptr, rsf_info->trans_bes, i_real);
	    assert(rsf_past_array);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		be_ptr be_fSg_i_d, be_result;

		be_fSg_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_fSg_i_d);
		
		if(i_model == 0)
		  {
		    /* Add [[f S g]]_0^d <=> [[g]]_0^0 */
		    be_ptr be_g_0_0 = array_fetch(be_ptr, rsf_past_array, 0);
		    assert(be_g_0_0);
		    be_result = Be_Iff(be_mgr, be_fSg_i_d, be_g_0_0);
		  }
		else if(i_model >= 1 && d == 0)
		  {
		    /* Add [[f S g]]_i^0 <=>
		       [[g]]_i^0 | ([[f]]_i^0 & [[f S g]]_{i-1}^0)*/
		    be_ptr be_f_i_d, be_g_i_d, be_fSg_iM1_d;
		    array_t* past_array_fSg_iM1;

		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		    assert(be_f_i_d);
		    
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d);
		    assert(be_g_i_d);

		    past_array_fSg_iM1 = 
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(past_array_fSg_iM1);
		    be_fSg_iM1_d = array_fetch(be_ptr, past_array_fSg_iM1, d);
		    assert(be_fSg_iM1_d);

		    be_result = Be_Iff(be_mgr,
				       be_fSg_i_d,
				       Be_Or(be_mgr,
					     be_g_i_d,
					     Be_And(be_mgr,
						    be_f_i_d,
						    be_fSg_iM1_d)));
		  }
		else if(i_model >= 1 && d > 0)
		  {
		    /*
		     * Add [[f S g]]_i^d <=>
		     * [[g]]_i^d | ([[f]]_i_d &
		     *              ITE(l_i,[[fSg]]_E^{d-1},[[fSg]]_{i-1}^d))
		     */
		    be_ptr be_f_i_d, be_g_i_d, be_fSg_E_dM1, be_fSg_iM1_d;
		    array_t* past_array_fSg_iM1;
		    array_t* past_array_fSg_E;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);
		    
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_g_i_d);

		    past_array_fSg_E = 
		      array_fetch(array_t*, info->trans_bes, _E_state());
		    assert(past_array_fSg_E);
		    be_fSg_E_dM1 = array_fetch(be_ptr, past_array_fSg_E, d-1);
		    assert(be_fSg_E_dM1);

		    past_array_fSg_iM1 = 
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(past_array_fSg_iM1);
		    be_fSg_iM1_d = array_fetch(be_ptr, past_array_fSg_iM1, d);
		    assert(be_fSg_iM1_d);
		    
		    be_result = Be_Iff(be_mgr,
				       be_fSg_i_d,
				       Be_Or(be_mgr,
					     be_g_i_d,
					     Be_And(be_mgr,
						    be_f_i_d,
						    Be_Ite(be_mgr,
							   be_l_i,
							   be_fSg_E_dM1,
							   be_fSg_iM1_d))));
		  }
		else
		  internal_error(_SNH_text,__FILE__,__LINE__);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  Adding constraint at depth %u: ",d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }


	    break;
	  }


	case TRIGGERED:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(!_set_is_in(visit_cache, rsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)rsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);

	    rsf_info = (_HJL_node_info*)find_assoc(info_map, rsf);
	    assert(rsf_info);
	    rsf_past_array = array_fetch(be_ptr, rsf_info->trans_bes, i_real);
	    assert(rsf_past_array);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		be_ptr be_fTg_i_d, be_result;

		be_fTg_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_fTg_i_d);
		
		if(i_model == 0)
		  {
		    /* Add [[f T g]]_0^d <=> [[g]]_0^0 */
		    be_ptr be_g_0_0 = array_fetch(be_ptr, rsf_past_array, 0);
		    assert(be_g_0_0);
		    be_result = Be_Iff(be_mgr, be_fTg_i_d, be_g_0_0);
		  }
		else if(i_model >= 1 && d == 0)
		  {
		    /* Add [[f T g]]_i^0 <=>
		       [[g]]_i^0 & ([[f]]_i^0 | [[f T g]]_{i-1}^0)*/
		    be_ptr be_f_i_d, be_g_i_d, be_fTg_iM1_d;
		    array_t* past_array_fTg_iM1;

		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d);
		    assert(be_f_i_d);
		    
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d);
		    assert(be_g_i_d);

		    past_array_fTg_iM1 = 
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(past_array_fTg_iM1);
		    be_fTg_iM1_d = array_fetch(be_ptr, past_array_fTg_iM1, d);
		    assert(be_fTg_iM1_d);

		    be_result = Be_Iff(be_mgr,
				       be_fTg_i_d,
				       Be_And(be_mgr,
					      be_g_i_d,
					      Be_Or(be_mgr,
						    be_f_i_d,
						    be_fTg_iM1_d)));
		  }
		else if(i_model >= 1 && d > 0)
		  {
		    /*
		     * Add [[f T g]]_i^d <=>
		     * [[g]]_i^d & ([[f]]_i_d |
		     *              ITE(l_i,[[fTg]]_E^{d-1},[[fTg]]_{i-1}^d))
		     */
		    be_ptr be_f_i_d, be_g_i_d, be_fTg_E_dM1, be_fTg_iM1_d;
		    array_t* past_array_fTg_iM1;
		    array_t* past_array_fTg_E;
		    const unsigned int d_lsf = min(d, lsf_info->past_depth);
		    const unsigned int d_rsf = min(d, rsf_info->past_depth);

		    be_f_i_d = array_fetch(be_ptr, lsf_past_array, d_lsf);
		    assert(be_f_i_d);
		    
		    be_g_i_d = array_fetch(be_ptr, rsf_past_array, d_rsf);
		    assert(be_g_i_d);

		    past_array_fTg_E = 
		      array_fetch(array_t*, info->trans_bes, _E_state());
		    assert(past_array_fTg_E);
		    be_fTg_E_dM1 = array_fetch(be_ptr, past_array_fTg_E, d-1);
		    assert(be_fTg_E_dM1);

		    past_array_fTg_iM1 = 
		      array_fetch(array_t*, info->trans_bes, i_real-1);
		    assert(past_array_fTg_iM1);
		    be_fTg_iM1_d = array_fetch(be_ptr, past_array_fTg_iM1, d);
		    assert(be_fTg_iM1_d);
		    
		    be_result = Be_Iff(be_mgr,
				       be_fTg_i_d,
				       Be_And(be_mgr,
					      be_g_i_d,
					      Be_Or(be_mgr,
						    be_f_i_d,
						    Be_Ite(be_mgr,
							   be_l_i,
							   be_fTg_E_dM1,
							   be_fTg_iM1_d))));
		  }
		else
		  internal_error(_SNH_text,__FILE__,__LINE__);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6)) {
		  fprintf(nusmv_stderr, "  Adding constraint at depth %u: ",d);
		  Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		  fprintf(nusmv_stderr, "\n");
		}
	      }


	    break;
	  }


	case OP_PREC:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		be_ptr be_Yf_i_d, be_result;
		const unsigned int d_lsf = min(d, lsf_info->past_depth);

		be_Yf_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_Yf_i_d);

		if(i_model == 0)
		  {
		    /* Add [[Y f]]_0^d <=> FALSE */
		    be_result = Be_Iff(be_mgr, be_Yf_i_d, Be_Falsity(be_mgr));
		  }
		else if(i_model >= 1 && d == 0)
		  {
		    /* Add [[Y f]]_i^0 <=> [[f]]_{i-1}^0 */
		    be_ptr be_f_iM1_0;
		    array_t* past_array_f_iM1 =
		      array_fetch(array_t*, lsf_info->trans_bes, i_real-1);
		    assert(past_array_f_iM1);
		    be_f_iM1_0 = array_fetch(be_ptr, past_array_f_iM1, d);
		    assert(be_f_iM1_0);
		    be_result = Be_Iff(be_mgr, be_Yf_i_d, be_f_iM1_0);
		  }
		else if(i_model >= 1 && d > 0)
		  {
		    /*
		     * Add [[Y f]]_i^d <=>
		     * ITE(l_i,[[f]]_E^{d-1},[[f]]_{i-1}^min(d,pd(f)))
		     */
		    be_ptr be_f_iM1_d, be_f_E_dM1;
		    array_t* past_array_f_iM1;
		    array_t* past_array_f_E;

		    past_array_f_E =
		      array_fetch(array_t*, lsf_info->trans_bes, _E_state());
		    assert(past_array_f_E);
		    be_f_E_dM1 = array_fetch(be_ptr, past_array_f_E, d-1);
		    assert(be_f_E_dM1);
		    
		    past_array_f_iM1 =
		      array_fetch(array_t*, lsf_info->trans_bes, i_real-1);
		    assert(past_array_f_iM1);
		    be_f_iM1_d = array_fetch(be_ptr, past_array_f_iM1, d_lsf);
		    assert(be_f_iM1_d);
		    
		    be_result = Be_Iff(be_mgr,
				       be_Yf_i_d,
				       Be_Ite(be_mgr,
					      be_l_i,
					      be_f_E_dM1,
					      be_f_iM1_d));
		  }
		else
		  internal_error(_SNH_text,__FILE__,__LINE__);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr, "  Adding constraint at depth %u: ", d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr), fprintf(nusmv_stderr, "\n");
	      }

	    break;
	  }


	case OP_NOTPRECNOT:
	  {
	    if(!_set_is_in(visit_cache, lsf)) {
	      lsNewBegin(unprocessed_nodes, (lsGeneric)lsf, LS_NH);
	      has_unprocessed_children = 1;
	    }
	    if(has_unprocessed_children)
	      break;

	    lsf_info = (_HJL_node_info*)find_assoc(info_map, lsf);
	    assert(lsf_info);
	    lsf_past_array = array_fetch(be_ptr, lsf_info->trans_bes, i_real);
	    assert(lsf_past_array);

	    if(opt_verbose_level_ge(options, 6))
	      _my_print_node(nusmv_stderr, str_debug1, node, str_debug2);

	    for(d = 0; d <= info->past_depth; d++)
	      {
		be_ptr be_Zf_i_d, be_result;
		const unsigned int d_lsf = min(d, lsf_info->past_depth);

		be_Zf_i_d = array_fetch(be_ptr, past_array, d);
		assert(be_Zf_i_d);

		if(i_model == 0)
		  {
		    /* Add [[Z f]]_0^d <=> TRUE */
		    be_result = Be_Iff(be_mgr, be_Zf_i_d, Be_Truth(be_mgr));
		  }
		else if(i_model >= 1 && d == 0)
		  {
		    /* Add [[Z f]]_i^0 <=> [[f]]_{i-1}^0 */
		    be_ptr be_f_iM1_0;
		    array_t* lsf_prev_past_array =
		      array_fetch(array_t*, lsf_info->trans_bes, i_real-1);
		    assert(lsf_prev_past_array);
		    be_f_iM1_0 = array_fetch(be_ptr, lsf_prev_past_array, d);
		    assert(be_f_iM1_0);
		    be_result = Be_Iff(be_mgr, be_Zf_i_d, be_f_iM1_0);
		  }
		else if(i_model >= 1 && d > 0)
		  {
		    /*
		     * Add [[Z f]]_i^d <=>
		     * ITE(l_i,[[f]]_E^{d-1},[[f]]_{i-1}^min(d,pd(f)))
		     */
		    be_ptr be_f_iM1_d, be_f_E_dM1;
		    array_t* lsf_prev_past_array;
		    array_t* lsf_E_past_array;

		    lsf_E_past_array =
		      array_fetch(array_t*, lsf_info->trans_bes, _E_state());
		    assert(lsf_E_past_array);
		    be_f_E_dM1 = array_fetch(be_ptr, lsf_E_past_array, d-1);
		    assert(be_f_E_dM1);
		    
		    lsf_prev_past_array =
		      array_fetch(array_t*, lsf_info->trans_bes, i_real-1);
		    assert(lsf_prev_past_array);
		    be_f_iM1_d = array_fetch(be_ptr,lsf_prev_past_array,d_lsf);
		    assert(be_f_iM1_d);
		    
		    be_result = Be_Iff(be_mgr,
				       be_Zf_i_d,
				       Be_Ite(be_mgr,
					      be_l_i,
					      be_f_E_dM1,
					      be_f_iM1_d));
		  }
		else
		  internal_error(_SNH_text,__FILE__,__LINE__);
		
		/* Save the created constraint */
		lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

		if(opt_verbose_level_ge(options, 6))
		  fprintf(nusmv_stderr, "  Adding constraint at depth %u: ", d),
		    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr), fprintf(nusmv_stderr, "\n");
	      }

	    break;
	  }



	default:
	  print_node(stderr, node);
	  internal_error(_SNYI_text,__FILE__,__LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;
      
      /* Remove node from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Mark visited */
      _set_insert(visit_cache, node);
    }
      
  lsDestroy(unprocessed_nodes, 0);
  _set_destroy(visit_cache); visit_cache = 0;

  return created_constraints;
}





/*
 * Create the formula specific k-dependent constraints
 * Return a list of be_ptrs for the created constraints
 */
static lsList _HJL_formula_dependent(const unsigned int k_model,
				     const node_ptr ltlspec,
				     const BeEnc_ptr be_enc,
				     const hash_ptr info_map)
{
  hash_ptr visit_cache = 0;
  lsList   unprocessed_nodes = 0;
  lsList   created_constraints = 0;
  Be_Manager_ptr be_mgr;

  assert(ltlspec);
  assert(be_enc);
  assert(info_map);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);

  /* Create list for the created constraints */
  created_constraints = lsCreate();

  /* Debug output */
  if(opt_verbose_level_gt(options, 1)) {
    fprintf(nusmv_stderr,
	    "Creating the formula specific k-dependent constraints for k=%d\n",
	    k_model);
    fflush(nusmv_stderr);
  }
  

  visit_cache = _set_create();
  unprocessed_nodes = lsCreate();
  lsNewBegin(unprocessed_nodes, (lsGeneric)ltlspec, LS_NH);

  
  while(lsLength(unprocessed_nodes) > 0)
    {
      node_ptr node;
      _HJL_node_info *info;
      array_t *f_E_past_array, *f_L_past_array;
      int has_unprocessed_children;

      /* Get node */
      if(lsFirstItem(unprocessed_nodes, (lsGeneric*)&node, LS_NH) != LS_OK ||
	 !node)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Get info */
      info = (_HJL_node_info*)find_assoc(info_map, node);
      if(!info)
	internal_error(_SNH_text,__FILE__,__LINE__);

      /* Get past_array */
      /* _HJL_make_state_vector should have built trans_bes[i] before */
      assert(info->trans_bes);
      assert(array_n(info->trans_bes) >= 2);
      f_L_past_array = array_fetch(array_t*, info->trans_bes, _L_state());
      assert(f_L_past_array);
      f_E_past_array = array_fetch(array_t*, info->trans_bes, _E_state());
      assert(f_E_past_array);

      /* Traverse children */
      has_unprocessed_children = 0;
      switch(node_get_type(node))
	{
	case ATOM:
	case BIT:
	case DOT:
	case ARRAY:
	case TRUEEXP:
	case FALSEEXP:
	  /* Leaf nodes */
	  break;

	case XOR:
	case XNOR:
	case IMPLIES:
	case IFF:
	  internal_error("%s:%d: Formula not in NNF\n",__FILE__,__LINE__);
	  break;

	case NOT:
	case OP_NEXT:
	case OP_PREC:
	case OP_NOTPRECNOT:
	case OP_ONCE:
	case OP_HISTORICAL:
	case OP_GLOBAL:
	case OP_FUTURE:
	  /* Unary operators */
	  if(!_set_is_in(visit_cache, car(node))) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)car(node), LS_NH);
	    has_unprocessed_children = 1;
	  }
	  break;

	case OR:
	case AND:
	case TRIGGERED:
	case SINCE:
	case RELEASES:
	case UNTIL:
	  /* Binary operators */
	  if(!_set_is_in(visit_cache, car(node))) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)car(node), LS_NH);
	    has_unprocessed_children = 1;
	  }
	  if(!_set_is_in(visit_cache, cdr(node))) {
	    lsNewBegin(unprocessed_nodes, (lsGeneric)cdr(node), LS_NH);
	    has_unprocessed_children = 1;
	  }
	  break;

	default:
	  print_node(stderr, node);
	  internal_error("%s:%d: TJ: Something not implemented",
			 __FILE__, __LINE__);
	  break;
	}
      if(has_unprocessed_children)
	continue;
      
      /* Remove node from unprocessed stack */
      if(lsDelBegin(unprocessed_nodes, (lsGeneric*)&node) != LS_OK)
	internal_error(_SNH_text,__FILE__,__LINE__);
      
      /* Mark visited */
      _set_insert(visit_cache, node);


      if(info->trans_vars) {
	/*
	 * Add [[f]]_E^d <=> [[f]]_k^d
	 */
	int d;
	array_t *f_k_past_array =
	  array_fetch(array_t*, info->trans_bes, _real_k(k_model));
	assert(f_k_past_array);

	for(d = 0; d <= info->past_depth; d++)
	  {
	    be_ptr be_f_E_d, be_f_k_d, be_result;
	    be_f_E_d = array_fetch(be_ptr, f_E_past_array, d);
	    assert(be_f_E_d);
	    be_f_k_d = array_fetch(be_ptr, f_k_past_array, d);
	    assert(be_f_k_d);
	    be_result = Be_Iff(be_mgr, be_f_E_d, be_f_k_d);
	    
	    /* Save the created constraint */
	    lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	    
	    if(opt_verbose_level_ge(options, 6)) {
	      _my_print_node(nusmv_stderr, " f: ", node, "\n");
	      fprintf(nusmv_stderr, "  ([[f]]_E^%d <=> [[f]]_%u^%d): ",
		      d, k_model, d);
	      Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	      fprintf(nusmv_stderr, "\n");
	    }
	  }
      }


      if(info->trans_vars)
	{
	  /*
	   * Add [[f]]_{k+1}^d <=> [[f]]_L^min(d+1,pd(f))
	   */
	  int d;
	  array_t *f_kP1_past_array =
	    array_fetch(array_t*, info->trans_bes, _real_k(k_model+1));
	  assert(f_kP1_past_array);
	  
	  for(d = 0; d <= info->past_depth; d++)
	    {
	      be_ptr be_f_kP1_d, be_f_L_dP1, be_result;
	      be_f_kP1_d = array_fetch(be_ptr, f_kP1_past_array, d);
	      assert(be_f_kP1_d);
	      be_f_L_dP1 = array_fetch(be_ptr, f_L_past_array,
				       min(d+1,info->past_depth));
	      assert(be_f_L_dP1);
	      be_result = Be_Iff(be_mgr, be_f_kP1_d, be_f_L_dP1);
	    
	      /* Save the created constraint */
	      lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);
	      
	      if(opt_verbose_level_ge(options, 6)) {
		_my_print_node(nusmv_stderr, " f: ", node, "\n");
		fprintf(nusmv_stderr, "  ([[f]]_%d^%d <=> [[f]]_L^%d): ",
			k_model+1, d, min(d+1,info->past_depth));
		Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }
	}
      

      if(info->aux_F_node)
	{
	  /* Add <<Ff>>_E <=> <<Ff>>_k */
	  be_ptr be_auxFf_E, be_auxFf_k, be_result;

	  be_auxFf_E = BeEnc_name_to_timed(be_enc, 
					   info->aux_F_node,
					   _E_state());
	  assert(be_auxFf_E);

	  be_auxFf_k = BeEnc_name_to_timed(be_enc,
					   info->aux_F_node,
					   _real_k(k_model));
	  assert(be_auxFf_k);

	  be_result = Be_Iff(be_mgr, be_auxFf_E, be_auxFf_k);
	  
	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "  <<Ff>>_E <=> <<Ff>>_%d: ", k_model);
	    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}


      if(info->aux_G_node)
	{
	  /* Add <<Gf>>_E <=> <<Gf>>_k */
	  be_ptr be_auxGf_E, be_auxGf_k, be_result;

	  be_auxGf_E = BeEnc_name_to_timed(be_enc,
					   info->aux_G_node,
					   _E_state());
	  assert(be_auxGf_E);

	  be_auxGf_k = BeEnc_name_to_timed(be_enc,
					   info->aux_G_node,
					   _real_k(k_model));
	  assert(be_auxGf_k);

	  be_result = Be_Iff(be_mgr, be_auxGf_E, be_auxGf_k);
	  
	  /* Save the created constraint */
	  lsNewBegin(created_constraints, (lsGeneric)be_result, LS_NH);

	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "  <<Gf>>_E <=> <<Gf>>_%d: ", k_model);
	    Be_DumpSexpr(be_mgr, be_result, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	}

    }
      
  lsDestroy(unprocessed_nodes, 0); unprocessed_nodes = 0;
  _set_destroy(visit_cache); visit_cache = 0;

  return created_constraints;
}


/*
 * Unroll future fragment from previous_k+1 to new_k
 * Unroll past fragment   from previous_k+2 to new_k+1
 */
static lsList _HJL_unroll_invariant(const int previous_k,
				    const int new_k,
				    const BeEnc_ptr be_enc,
				    const HJL_state_vars_struct *HJL_state_vars,
				    array_t * const InLoop_array,
				    const node_ptr bltlspec,
				    const hash_ptr info_map,
				    const int opt_do_optimization)
{
  int i;
  lsList created_constraints = lsCreate();

  assert(previous_k < new_k);

  for(i = max(previous_k + 1, 0); i <= new_k + 1; i++)
    {
      be_ptr be_InLoop_i;
      be_ptr be_l_i;

      be_InLoop_i = array_fetch(be_ptr, InLoop_array, i);
      assert(be_InLoop_i);
      
      be_l_i = BeEnc_name_to_timed(be_enc, HJL_state_vars->l_var, _real_k(i));
      assert(be_l_i);
      
      if(i <= new_k)
	{
	  /* Future fragment */
	  lsList new_constraints = 
	    _HJL_unroll_invariant_f(i, bltlspec,
				    be_enc, info_map,
				    be_InLoop_i, be_l_i,
				    opt_do_optimization);
	  lsJoin(created_constraints, new_constraints, 0);
	  lsDestroy(new_constraints, 0);
	}
      if(i >= previous_k + 2)
	{
	  /* Past fragment */
	  lsList new_constraints = 
	    _HJL_unroll_invariant_p(i, bltlspec,
				    be_enc, info_map,
				    be_InLoop_i, be_l_i,
				    opt_do_optimization);
	  lsJoin(created_constraints, new_constraints, 0);
	  lsDestroy(new_constraints, 0);
	}
    }
  return created_constraints;
}


/*
 * Create the constraint l_{k+1} <=> FALSE
 * Create the constraint s_E = s_k
 * Create the constraint LoopExists <=> InLoop_k
 * Create the formula specific k-dependent constraints
 */
static lsList _HJL_dependent(const int k,
			     const node_ptr bltlspec,
			     const BeEnc_ptr be_enc,
			     const HJL_state_vars_struct *HJL_state_vars,
			     array_t *InLoop_array,
			     const be_ptr be_LoopExists,
			     const hash_ptr info_map)
{
  Be_Manager_ptr be_mgr;
  lsList created_constraints = lsCreate();

  assert(k >= 0);
  assert(bltlspec);
  assert(be_enc);
  assert(InLoop_array);
  assert(info_map);

  /* Get be manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);


  /*
   * Add and force the constraint l_{k+1} <=> FALSE
   */
  {
    be_ptr be_l_kP1 = BeEnc_name_to_timed(be_enc,
					  HJL_state_vars->l_var,
					  _real_k(k+1));
    assert(be_l_kP1);
    be_ptr be_constraint = Be_Iff(be_mgr, be_l_kP1, Be_Falsity(be_mgr));
    assert(be_constraint);

    lsNewEnd(created_constraints, be_constraint, LS_NH);

    if(opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "Forced (l_%d <=> FALSE)", k+1);
      if(opt_verbose_level_ge(options, 6)) {
	fprintf(nusmv_stderr, ": ");
	Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
      }
      fprintf(nusmv_stderr, "\n");
    }
  }


  /*
   * Add and force the constraint s_E = s_k
   */
  {
    be_ptr be_constraint =
      _equal_vectors_formula(be_enc,
			     HJL_state_vars->simple_path_system_vars,
			     _E_state(), _real_k(k));
    
    lsNewEnd(created_constraints, be_constraint, LS_NH);

    if(opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "Forced (s_E = s_%d)", k);
      if(opt_verbose_level_ge(options, 6)) {
	fprintf(nusmv_stderr, ": ");
	Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
      }
      fprintf(nusmv_stderr, "\n");
    }
  }


  /*
   * Add and force the constraint LoopExists <=> InLoop_k
   */
  {
    be_ptr be_InLoop_k, be_constraint;
    be_InLoop_k = array_fetch(be_ptr, InLoop_array, k);
    assert(be_InLoop_k);
    be_constraint = Be_Iff(be_mgr, be_LoopExists, be_InLoop_k);

    lsNewEnd(created_constraints, be_constraint, LS_NH);
    
    if(opt_verbose_level_gt(options, 1)) {
      fprintf(nusmv_stderr, "Forced (LoopExists <=> InLoop_%d)", k);
      if(opt_verbose_level_ge(options, 6)) {
	fprintf(nusmv_stderr, ": ");
	Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
      }
      fprintf(nusmv_stderr, "\n");
    }
  }

  
  /*
   * Add and force the formula specific k-dependent constraints
   */
  {
    lsList new_constraints = 
      _HJL_formula_dependent(k, bltlspec, be_enc, info_map);
    lsJoin(created_constraints, new_constraints, 0);
    lsDestroy(new_constraints, 0);
  }
  return created_constraints;
}





/*
 * Return Booleanized, negated and NNFed formula with fairness constraints
 */
static node_ptr _HJL_make_boolean_formula(Prop_ptr ltlprop)
{
  node_ptr fltlspec = (node_ptr)NULL;

  fltlspec = detexpr2bexpr(Enc_get_bdd_encoding(), Prop_get_expr(ltlprop));

  fltlspec = Bmc_Wff_MkNot(fltlspec);
        
  /*
   * Add fairness constraints to the formula
   */
  {
    SexpFsm_ptr sexp_fsm = Prop_get_bool_sexp_fsm(ltlprop);
    assert(sexp_fsm != (SexpFsm_ptr)NULL);
    
    node_ptr justice_list = SexpFsm_get_justice(sexp_fsm);
    while(!is_list_empty(justice_list))
      {
	fltlspec = Bmc_Wff_MkAnd(fltlspec, Bmc_Wff_MkGlobally(Bmc_Wff_MkEventually(car(justice_list))));
	justice_list = cdr(justice_list);
      }
    
    node_ptr compassion_list = SexpFsm_get_compassion(sexp_fsm);
    if(!is_list_empty(compassion_list))
      internal_error("%s:%d: Compassion not handled",__FILE__,__LINE__);
  }
  
  return Bmc_Wff_MkNnf(fltlspec);
}


/*
 * Find state and input variables that occurr in the formula.
 * Build the list of system variables for simple path constraints.
 *
 * HJL_state_vars->formula_state_vars will have the state vars occurring
 *   in the formula bltlspec
 * HJL_state_vars->formula_input_vars will have the input vars occurring
 *   in the formula bltlspec
 * HJL_state_vars->simple_path_system_vars will be the union of
 *   HJL_state_vars->transition_state_vars,
 *   HJL_state_vars->formula_state_vars, and
 *   HJL_state_vars->formula_input_vars
 */
static void _HJL_find_relevant_vars(HJL_state_vars_struct *HJL_state_vars,
				    BeFsm_ptr be_fsm,
				    node_ptr bltlspec)
{
  int i = 0;
  lsGen  iterator;
  node_ptr node = 0;
  BeEnc_ptr be_enc = 0;
  SymbTable_ptr st = 0;
  lsList f_vars = 0;

  assert(HJL_state_vars);

  /* Get be encoding */
  be_enc = BeFsm_get_be_encoding(be_fsm);
  assert(be_enc);

  /* Get symbol table */
  st = BaseEnc_get_symbol_table(BASE_ENC(be_enc));
  assert(st);

  /* Clean lists if not empty */
  if(lsLength(HJL_state_vars->formula_state_vars) > 0)
    {
      lsDestroy(HJL_state_vars->formula_state_vars, 0);
      HJL_state_vars->formula_state_vars = lsCreate();
    }
  if(lsLength(HJL_state_vars->formula_input_vars) > 0)
    {
      lsDestroy(HJL_state_vars->formula_input_vars, 0);
      HJL_state_vars->formula_input_vars = lsCreate();
    }
  if(lsLength(HJL_state_vars->simple_path_system_vars) > 0)
    {
      lsDestroy(HJL_state_vars->simple_path_system_vars, 0);
      HJL_state_vars->simple_path_system_vars = lsCreate();
    }
  
  /* Get all the variables occurring in the formula */
  f_vars = _HJL_find_formula_vars(bltlspec);
  assert(f_vars);
  
  /* Classify the variables to state and input ones */
  lsForEachItem(f_vars, iterator, node)
    {
      if(SymbTable_is_symbol_state_var(st, node))
	lsNewEnd(HJL_state_vars->formula_state_vars, (lsGeneric)node, LS_NH);
      else if(SymbTable_is_symbol_input_var(st, node))
	lsNewEnd(HJL_state_vars->formula_input_vars, (lsGeneric)node, LS_NH);
      else
	internal_error("%s:%d: Unknown variable type (not state or input)",
		       __FILE__, __LINE__);
    }
  /* Release list */
  lsDestroy(f_vars, 0); f_vars = 0;
  
  
  
  /* Build simple_path_system_vars */
  {
    hash_ptr union_set = _set_create();
    
    lsForEachItem(HJL_state_vars->transition_state_vars, iterator, node)
      {
	if(!_set_is_in(union_set, node))
	  {
	    _set_insert(union_set, node);
	      lsNewEnd(HJL_state_vars->simple_path_system_vars,
		       (lsGeneric)node, LS_NH);
	  }
      }
    lsForEachItem(HJL_state_vars->formula_state_vars, iterator, node)
      {
	if(!_set_is_in(union_set, node))
	  {
	    _set_insert(union_set, node);
	    lsNewEnd(HJL_state_vars->simple_path_system_vars,
		     (lsGeneric)node, LS_NH);
	  }
      }
    lsForEachItem(HJL_state_vars->formula_input_vars, iterator, node)
      {
	if(!_set_is_in(union_set, node))
	  {
	    _set_insert(union_set, node);
	    lsNewEnd(HJL_state_vars->simple_path_system_vars,
		     (lsGeneric)node, LS_NH);
	  }
      }

    _set_destroy(union_set);
  }
}



/*
 * MS stands for Meta Solver interface
 * Define this if you want a possibility to use both the standard and
 * my incremental SAT solver interface.
 * If this is not defined, the algorithms use my interface.
 */
#define MS


/*
 * The main routine
 * If MS (above) is defined, then setting opt_use_standard_solver
 *   makes the routine to use the standard incremental SAT solver interface
 */
int Bmc_HJL_zigzag_incr(Prop_ptr ltlprop,
			const int max_k,
			const int opt_do_virtual_unrolling,
			const int opt_do_completeness_check,
			const int opt_use_standard_solver)
{
  node_ptr bltlspec;  /* Its booleanization */
  BeFsm_ptr be_fsm; /* The corresponding be fsm */
  BeEnc_ptr be_enc;
  BoolEnc_ptr bool_enc;
  Be_Manager_ptr be_mgr;


  /* Structure for holding state variable information */
  HJL_state_vars_struct *HJL_state_vars = 0;

#ifdef MS
  _MetaSolver *zolver = 0;
#else
  /* incremental sat solver instance */
  Bmc_IncrSat_ptr solver = 0;
#endif

  /* Symbol table and layer stuff needed to allocate new state vars */
  SymbTable_ptr symb_table = (SymbTable_ptr)NULL;
  SymbLayer_ptr ltl_layer = (SymbLayer_ptr)NULL;

  int current_k = 0;
  int previous_k = -1; /* used to create Be of execution from time 0 */
  boolean found_solution = 0;
  boolean formula_forced_to_true = 0;
  
  /* Associate each subformula node_ptr with _HJL_node_info */
  hash_ptr info_map = 0;

  /* The LoopExists variable */
  be_ptr be_LoopExists = 0;
  
  /* Array of be_ptr,
     InLoop_array[i] has the be for InLoop_i (where i is a model index) */
  array_t* InLoop_array = 0;

  /* Force each subformula to have its own state variable.
   * Costly, only for debugging/cross validation */
  const int opt_force_state_vars = 0;
  
  const int opt_do_optimization = 1;


#ifdef BENCHMARKING
  times(&time_bmc_start);
#endif  

  /* checks that a property was selected: */
  nusmv_assert(ltlprop != PROP(NULL));

  /* checks that a property was selected: */
  nusmv_assert(ltlprop != (Prop_ptr)NULL);
  
  /* checks if it has already been checked: */
  if (Prop_get_status(ltlprop) != Prop_Unchecked)
    {
      /* aborts this check */
      return 0;
    }
  
  
  if (opt_cone_of_influence(options) == true)
    {
      Prop_apply_coi_for_bmc(ltlprop, global_fsm_builder);
    }
  
  be_fsm = Prop_get_be_fsm(ltlprop);
  if (be_fsm == (BeFsm_ptr) NULL)
    {
      Prop_set_fsm_to_master(ltlprop);
      be_fsm = Prop_get_be_fsm(ltlprop);
      nusmv_assert(be_fsm != (BeFsm_ptr) NULL);
    }
  

  /* Get be encoding */
  be_enc = BeFsm_get_be_encoding(be_fsm);
  assert(be_enc);

  /* Get Boolean encoding */
  bool_enc = BoolEncClient_get_bool_enc(BOOL_ENC_CLIENT(be_enc));
  assert(bool_enc);

  /*
   * Booleanized, negated and NNFed formula with fairness constraints
   */
  bltlspec = _HJL_make_boolean_formula(ltlprop);
  assert(bltlspec);

  
  /* Find manager */
  be_mgr = BeEnc_get_be_manager(be_enc);
  assert(be_mgr);


  /*
   * Incremental SAT solver construction
   */
#ifdef MS
  zolver = _MS_create(be_enc, opt_use_standard_solver == 0);
  if(zolver == 0)
    /* Something went wrong */
    return 1;
#else
  solver = Bmc_IncrSat_init(get_sat_solver(options));
  if(solver == 0)
    {
      fprintf(nusmv_stderr, "Incremental sat solver '%s' is not available.\n", 
	      get_sat_solver(options));
      return 1;
    }
#endif

  /*
   * Initialize the state variable structure
   */
  HJL_state_vars = _HJL_state_vars_create();
  assert(HJL_state_vars);

  if (opt_verbose_level_gt(options, 4))
    _HJL_state_vars_print(HJL_state_vars, nusmv_stderr);


  /*
   * Get the state variables occurring in the transition relation and
   * add them to HJL_state_vars->transition_state_vars
   */
  {
    lsGen iterator; 
    be_ptr be_var = 0;
    hash_ptr node_seen = 0;
    lsList tr_be_vars = Bmc_Utils_get_vars_list_for_uniqueness(be_enc,
							       ltlprop);
    node_seen = _set_create();
    lsForEachItem(tr_be_vars, iterator, be_var)
      {
	node_ptr node = BeEnc_var_to_name(be_enc, be_var);
	if(!_set_is_in(node_seen, node))
	  {
	    lsNewEnd(HJL_state_vars->transition_state_vars,
		     (lsGeneric)node, LS_NH);
	    _set_insert(node_seen, node);
	  }
      }
    _set_destroy(node_seen); node_seen = 0;
    lsDestroy(tr_be_vars, 0); tr_be_vars = 0;
  }


  /* Get the symbol table */
  symb_table = Compile_get_global_symb_table();

  /* Add a new layer for translation variables */
  ltl_layer = SymbTable_create_layer(symb_table, "LTL translation vars",
				     SYMB_LAYER_POS_TOP);

  /* Allocate the l_i variable */
  HJL_state_vars->l_var = _add_new_state_variable(ltl_layer, "#LTL_l");

  /* Allocate the LoopExists variable */
  HJL_state_vars->LoopExists_var =
    _add_new_state_variable(ltl_layer, "#LTL_LoopExists");


  /*
   * Associates each subformula node of ltlspec with a _HJL_LTL_info
   * Returns a hash from node_ptr to _HJL_LTL_info*.
   * New state variables named #LTL_t'i' can be allocate to 'layer'.
   * The new state vars are inserted in state_vars_formula_??? appropriately.
   */
  info_map = _HJL_init_LTL_info(bltlspec,
				ltl_layer,
				HJL_state_vars->translation_vars_pd0,
				HJL_state_vars->translation_vars_pdx,
				HJL_state_vars->translation_vars_aux,
				opt_force_state_vars,
				opt_do_virtual_unrolling);



  /*
   * After introducing all new variables, commit ltl_layer
   */
  BaseEnc_commit_layer(BASE_ENC(bool_enc), SymbLayer_get_name(ltl_layer));
  BaseEnc_commit_layer(BASE_ENC(be_enc), SymbLayer_get_name(ltl_layer));
  BaseEnc_commit_layer(BASE_ENC(Enc_get_bdd_encoding()),
		       SymbLayer_get_name(ltl_layer));


  /*
   * Find state and input variables that occurr in the formula.
   * Build the list of system variables for simple path constraints.
   *
   * HJL_state_vars->formula_state_vars will have the state vars occurring
   *   in the formula bltlspec
   * HJL_state_vars->formula_input_vars will have the input vars occurring
   *   in the formula bltlspec
   * HJL_state_vars->simple_path_system_vars will be the union of
   *   HJL_state_vars->transition_state_vars,
   *   HJL_state_vars->formula_state_vars, and
   *   HJL_state_vars->formula_input_vars
   */
  _HJL_find_relevant_vars(HJL_state_vars, be_fsm, bltlspec);


  if (opt_verbose_level_gt(options, 4))
    _HJL_state_vars_print(HJL_state_vars, nusmv_stderr);


  /*
   * Create the LoopExists variable be
   * Simulate untimed variable by using only time index 0.
   */
  be_LoopExists = BeEnc_name_to_timed(be_enc,HJL_state_vars->LoopExists_var,0);
  

  /*
   * Initialize the InLoop array
   */
  InLoop_array = array_alloc(be_ptr, 1);
  
  
  /*
   * Initialize state vectors of aux states L and E
   */
  _HJL_init_state_vector(be_enc, bltlspec, info_map, _L_state());
  _HJL_init_state_vector(be_enc, bltlspec, info_map, _E_state());

  

  /*
   * Insert the initial condition into the sat solver permanently
   */
  {
    be_ptr be_init = Bmc_TJ_GetInitI(be_fsm, _real_k(0));
#ifdef MS
    _MS_force_true(zolver, be_init);
#else
    Bmc_IncrSat_insert(solver, be_init);
    Bmc_IncrSat_force_true(solver, be_init);
#endif

    if(opt_verbose_level_ge(options, 6)) {
      fprintf(nusmv_stderr, "Forced initial condition: ");
      Be_DumpSexpr(be_mgr, be_init, nusmv_stderr);
      fprintf(nusmv_stderr, "\n");
    }
  }



  /*
   * Make base stuff
   */
  {
    lsList new_constraints = _HJL_unroll_base(bltlspec,
					      be_enc, info_map, be_LoopExists,
					      opt_do_optimization);
    /* Force constraints to be true in the fixed frame */
#ifdef MS
    _MS_force_constraint_list(zolver, new_constraints);
#else
    _force_constraint_list(solver, new_constraints);
#endif
    lsDestroy(new_constraints, 0); new_constraints = 0;
  }
  


  /*
   * Start problem generation:
   */
  previous_k = -1000; /* Start from deep prehistory (at least <= -2) */
  for(current_k = 0;
      current_k <= max_k && !found_solution;
      current_k = current_k + 1)
    {
      int i = 0;
      Bmc_IncrSat_result satResult;
      
      if(opt_verbose_level_gt(options, 0)) {
	fprintf(nusmv_stderr,
		"\nGenerating problem with bound %d, all possible loopbacks...\n",
                current_k);
      }


      
      /********************************************************************
       *
       * The k-invariant part of the translation
       *
       ********************************************************************/

      
      /*
       * Formula state vector initialization
       * Initialize new state vectors up to current_k+1
       * Assumes that state vectors are already initialized up to previous_k+1
       */
      for(i = max(previous_k + 2, 0); i <= current_k + 1; i++)
	{
	  _HJL_init_state_vector(be_enc,bltlspec,info_map,_real_k(i));
	}


      /*
       * Unroll the model transition relation up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      for(i = max(previous_k, 0); i < current_k; i++)
	{
	  be_ptr be_TR = Bmc_Model_GetUnrolling(be_fsm, 
						_real_k(i),
						_real_k(i+1));
#ifdef MS
	  _MS_force_true(zolver, be_TR);
#else
	  Bmc_IncrSat_insert(solver, (Rbc_t*)be_TR);
	  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_TR);
#endif
	  
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "Forced T(%d,%d)", i, i+1);
	  if(opt_verbose_level_ge(options, 7)) {
	    fprintf(nusmv_stderr, ": ");
	    Be_DumpSexpr(be_mgr, be_TR, nusmv_stderr);
	  }
	  if(opt_verbose_level_gt(options, 1))
	    fprintf(nusmv_stderr, "\n");
	}


      /*
       * Unroll the (l_i => (s_{i-1} = s_E)) constraint up to current_k
       * Force it to the fixed frame 0
       * Assumes that it has already been unrolled up to previous_k
       */
      for(i = max(previous_k + 1, 0); i <= current_k; i++)
	{
	  be_ptr be_l_i;
	  be_ptr be_constraint;

	  be_l_i = BeEnc_name_to_timed(be_enc,
				       HJL_state_vars->l_var,
				       _real_k(i));
	  
	  if(i == 0)
	    {
	      /* l_0 <=> FALSE */
	      be_constraint = Be_Iff(be_mgr, be_l_i, Be_Falsity(be_mgr));
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr, "Created (l_0 <=> false): ");
		Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }
	  else
	    {
	      be_constraint =
		_equal_vectors_formula(be_enc,
				       HJL_state_vars->simple_path_system_vars,
				       _real_k(i-1),_E_state());
	      be_constraint = Be_Implies(be_mgr, be_l_i, be_constraint);
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr, "Created (l_%d => (s_%d = s_E)): ",
			current_k, current_k-1);
		Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }
#ifdef MS
	  _MS_force_true(zolver, be_constraint);
#else
	  Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	  Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
	}


      /*
       * Define InLoop_i := (InLoop_{i-1} || l_i)
       * and build constraint InLoop_{i-1} => !l_i up to current_k + 1
       * Assumes that they have already been build up to previous_k + 1
       */
      for(i = max(previous_k + 1, 0); i <= current_k+1; i++)
	{
	  be_ptr be_constraint = _HJL_build_InLoop_i(i, 
						     be_enc,
						     HJL_state_vars,
						     InLoop_array);
	  if(be_constraint)
	    {
#ifdef MS
	      _MS_force_true(zolver, be_constraint);
#else
	      Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	      Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
	    }
	}


      if(opt_do_optimization)
	{
	  /*
	   * Unroll the (l_i => LoopExists) constraint up to current_k
	   * Force 'em true in the fixed frame 0
	   * Assumes that this has already been done upto previous_k
	   */
	  for(i = max(previous_k + 1, 0); i <= current_k; i++)
	    {
	      be_ptr be_l_i;
	      be_ptr be_constraint;
	      
	      be_l_i = BeEnc_name_to_timed(be_enc,
					   HJL_state_vars->l_var,
					   _real_k(i));
	      
	      be_constraint = Be_Implies(be_mgr, be_l_i, be_LoopExists);
	      
#ifdef MS
	      _MS_force_true(zolver, be_constraint);
#else
	      Bmc_IncrSat_insert(solver, (Rbc_t*)be_constraint);
	      Bmc_IncrSat_force_true(solver, (Rbc_t*)be_constraint);
#endif
	      
	      if(opt_verbose_level_ge(options, 6)) {
		fprintf(nusmv_stderr, "Created (l_%d => LoopExists): ",
			current_k);
		Be_DumpSexpr(be_mgr, be_constraint, nusmv_stderr);
		fprintf(nusmv_stderr, "\n");
	      }
	    }
	}


      /*
       * Unroll future fragment from previous_k+1 to current_k
       * Unroll past fragment   from previous_k+2 to current_k+1
       */
      {
	lsList new_constraints = _HJL_unroll_invariant(previous_k,
						       current_k,
						       be_enc,
						       HJL_state_vars,
						       InLoop_array,
						       bltlspec,
						       info_map,
						       opt_do_optimization);
#ifdef MS
	_MS_force_constraint_list(zolver, new_constraints);
#else
	_force_constraint_list(solver, new_constraints);
#endif
	lsDestroy(new_constraints, 0); new_constraints = 0;
      }

      /*
       * Force negated NNF formula to true if not already done
       */
      if(!formula_forced_to_true)
	{
	  _HJL_node_info* info;
	  array_t*         past_array;
	  be_ptr           be_f_0_0;
	  
	  /* Get info */
	  info = (_HJL_node_info*)find_assoc(info_map, bltlspec);
	  assert(info);
	  
	  /* Get past_array */
	  assert(info->trans_bes);
	  assert(array_n(info->trans_bes) > 0);
	  past_array = array_fetch(array_t*, info->trans_bes, _real_k(0));
	  assert(past_array);
	  
	  be_f_0_0 = array_fetch(be_ptr, past_array, 0);
	  assert(be_f_0_0);

#ifdef MS
	  _MS_force_true(zolver, be_f_0_0);
#else	  
	  Bmc_IncrSat_insert(solver, be_f_0_0);
	  Bmc_IncrSat_force_true(solver, be_f_0_0);
#endif
	  
	  if(opt_verbose_level_ge(options, 6)) {
	    fprintf(nusmv_stderr, "Forced [[f]]_0^0 to true: ");
	    Be_DumpSexpr(be_mgr, be_f_0_0, nusmv_stderr);
	    fprintf(nusmv_stderr, "\n");
	  }
	  formula_forced_to_true = true;
	}


      if(opt_do_completeness_check)
	{
	  /*
	   * Unroll the SimplePath constraints
	   */
	  for(i = max(previous_k+1, 0); i <= current_k; i++)
	    {
	      lsList new_constraints = 
		_HJL_SimplePaths(i, be_enc, HJL_state_vars, InLoop_array);
#ifdef MS
	      _MS_force_constraint_list(zolver, new_constraints);
#else
	      _force_constraint_list(solver, new_constraints);
#endif
	      lsDestroy(new_constraints, 0); new_constraints = 0;
	    }

#ifdef BENCHMARKING
	  fprintf(nusmv_stderr, ":START:benchmarking Solving\n");
	  times(&time_solver_start);
#endif  
	  

	  /*
	   * Actually solve the problem
	   */
#ifdef MS
	  satResult = _MS_solve(zolver);
#else
	  satResult = Bmc_IncrSat_solve(solver);
#endif
      
	  
#ifdef BENCHMARKING
	  {
	    times(&time_temp);
	    fprintf(nusmv_stderr, ":UTIME = %.4f secs.\n",
		    time_diff(&time_solver_start, &time_temp));
	    fprintf(nusmv_stderr, ":STOP:benchmarking Solving\n");
	    fprintf(nusmv_stderr, "TJ: completeness, k=%d, solver time=%.4f, "
		    "cumulative time=%.4f.\n",
		    current_k,
		    time_diff(&time_solver_start, &time_temp),
		    time_diff(&time_bmc_start, &time_temp));
	  }
#endif
	  
	  
	  /* Processes the result: */
	  switch (satResult)
	    {
	    case BMC_INCRSAT_SATISFIABLE_PROBLEM:
	      {
		int i;
		
		fprintf(nusmv_stderr,
			"-- could not terminate yet with bound %d",
			current_k);
		
		if (opt_verbose_level_gt(options, 2)) {
		  fprintf(nusmv_stderr, " for "); 		  
		  print_spec(nusmv_stderr, ltlprop);
		}
		fprintf(nusmv_stderr, "\n");		
		break;
	      }
	    case BMC_INCRSAT_UNSATISFIABLE_PROBLEM:
	      fprintf(nusmv_stderr, "-- ");
	      print_spec(nusmv_stderr, ltlprop);
	      fprintf(nusmv_stderr, "  is true\n");
	      Prop_set_status(ltlprop, Prop_True);
	      
	      found_solution = true;
	      break;
	      
	    case BMC_INCRSAT_INTERNAL_ERROR:
	      internal_error("Sorry, solver answered with a fatal Internal " 
			     "Failure during problem solving.\n");
	      break;
	      
	    case BMC_INCRSAT_TIMEOUT:
	    case BMC_INCRSAT_MEMOUT:
	      internal_error("Sorry, solver ran out of resources and aborted "
			     "the execution.\n");
	      break;
	      
	    default:
	      internal_error("Bmc_GenSolveLtlTJ: Unexpected value in satResult");
	      
	    } /* switch */
	  
	  
	  
	  if(found_solution)
	    break;

	} /* if(opt_do_completeness_check) */


      /********************************************************************
       *
       * The k-dependent part of the translation
       *
       ********************************************************************/


      if(opt_verbose_level_gt(options, 1))
	fprintf(nusmv_stderr, "Adding k-dependent stuff\n");
      

      /*
       * Go to the volatile frame 1
       */
#ifdef MS
      _MS_goto_volatile_group(zolver);
#else
      Bmc_IncrSat_new_frame(solver);
#endif


      /*
       * Add k-dependent constraints
       */
      {
	lsList new_constraints = _HJL_dependent(current_k,
						bltlspec,
						be_enc,
						HJL_state_vars,
						InLoop_array,
						be_LoopExists,
						info_map);
#ifdef MS
	_MS_force_constraint_list(zolver, new_constraints);
#else
	_force_constraint_list(solver, new_constraints);
#endif
	lsDestroy(new_constraints, 0); new_constraints = 0;
      }
      

#ifdef BENCHMARKING
      fprintf(nusmv_stderr, ":START:benchmarking Solving\n");
      times(&time_solver_start);
#endif  


      /*
       * Actually solve the problem
       */
#ifdef MS
      satResult = _MS_solve(zolver);
#else
      satResult = Bmc_IncrSat_solve(solver);
#endif


#ifdef BENCHMARKING
      {
	times(&time_temp);
	fprintf(nusmv_stderr, ":UTIME = %.4f secs.\n",
		time_diff(&time_solver_start, &time_temp));
	fprintf(nusmv_stderr, ":STOP:benchmarking Solving\n");
	fprintf(nusmv_stderr, "TJ: counterexample, k=%d, solver time=%.4f, "
		"cumulative time=%.4f.\n",
		current_k,
		time_diff(&time_solver_start, &time_temp),
		time_diff(&time_bmc_start, &time_temp));
      }
#endif


     /* Processes the result: */
      switch (satResult)
	{
	case BMC_INCRSAT_UNSATISFIABLE_PROBLEM:
	  {
	    fprintf(nusmv_stderr,
		    "-- no counterexample found with bound %d",
		  current_k);
	    if (opt_verbose_level_gt(options, 2)) {
	      fprintf(nusmv_stderr, " for "); 		  
	      print_spec(nusmv_stderr, ltlprop);
	    }
	    fprintf(nusmv_stderr, "\n");
	    
	    break;
	  }
	case BMC_INCRSAT_SATISFIABLE_PROBLEM:
	  fprintf(nusmv_stderr, "-- ");
	  print_spec(nusmv_stderr, ltlprop);
	  fprintf(nusmv_stderr, "  is false\n");
	  Prop_set_status(ltlprop, Prop_False);
	  
	  found_solution = true;
	  
	  if (opt_counter_examples(options)) {
#ifdef MS
	    if(opt_use_standard_solver)
	      {
		/* Counter-examples with standard incremental SAT solver
		   not supported yet */
	      }
	    else
	      {
		Trace_ptr trace = 
		  _generate_and_print_cntexample(be_enc,
						 zolver->TJ_solver,
						 Enc_get_bdd_encoding(), 
						 current_k, 
						 "BMC Counterexample",
						 HJL_state_vars->l_var);
		Prop_set_trace(ltlprop, Trace_get_id(trace));
	      }
#else
	    Trace_ptr trace = 
	      _generate_and_print_cntexample(be_enc,
					     solver,
					     Enc_get_bdd_encoding(), 
					     current_k, 
					     "BMC Counterexample",
					     HJL_state_vars->l_var);
	    Prop_set_trace(ltlprop, Trace_get_id(trace));
#endif
	  }
	  
	  break;
	  
	case BMC_INCRSAT_INTERNAL_ERROR:
	  internal_error("Sorry, solver answered with a fatal Internal " 
			 "Failure during problem solving.\n");
	  break;
	  
	case BMC_INCRSAT_TIMEOUT:
	case BMC_INCRSAT_MEMOUT:
	  internal_error("Sorry, solver ran out of resources and aborted "
			 "the execution.\n");
	  break;
	  
	default:
	  internal_error("Bmc_GenSolveLtlTJ: Unexpected value in satResult");
	  
	} /* switch */



      if(opt_verbose_level_gt(options, 1))
	fprintf(nusmv_stderr, "Deleting k-dependent stuff\n");


      /*
       * Release the volatile frame 1
       */
#ifdef MS
      _MS_goto_permanent_group(zolver);
#else
      Bmc_IncrSat_pop_frame(solver);
#endif

      /*
       * Remember the current_k
       */
      previous_k = current_k;
    }


  /*
   * Deallocate info_map
   */
  clear_assoc_and_free_entries(info_map, _HJL_node_info_free_func);
  info_map = 0;
  

  /*
   * Remove extra encoding layers
   */
  BaseEnc_remove_layer(BASE_ENC(Enc_get_bdd_encoding()),
		       SymbLayer_get_name(ltl_layer));
  BaseEnc_remove_layer(BASE_ENC(be_enc), SymbLayer_get_name(ltl_layer));
  BaseEnc_remove_layer(BASE_ENC(bool_enc), SymbLayer_get_name(ltl_layer));
  SymbTable_remove_layer(symb_table, ltl_layer);
  ltl_layer = 0;

  /*
   * Destroy the sat solver instance
   */
#ifdef MS
  _MS_destroy(zolver); zolver = 0;
#else
  Bmc_IncrSat_release(solver); solver = 0;
#endif
  
  return 0;
}


/*
 * Similar to Bmc_Utils_generate_and_print_cntexample but modified
 * for my incremental SAT solver interface
 */
static Trace_ptr _generate_and_print_cntexample(BeEnc_ptr be_enc, 
						Bmc_IncrSat_ptr solver, 
						BddEnc_ptr bdd_enc, 
						const int k, 
						const char* trace_name,
						be_ptr l_var)
{
  BmcSatTrace_ptr sat_trace;
  Trace_ptr trace;
  lsList be_model;
  node_ptr path;
  int bmc_tr;

  assert(l_var);

  /*	  
  The original
  be_model = Be_CnfModelToBeModel(BeEnc_get_be_manager(be_enc), 
				  SatSolver_get_model(solver));
  is replaced by the following:
  */
  be_model = lsCreate();
  {
    int i = 0;
    /* For each time step ... */
    for(i = _real_k(0); i <= _real_k(k); i++)
      {
	/* get values of state and input variables */
	int index = BeEnc_get_first_untimed_var_index(be_enc, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
	while(BeEnc_is_var_index_valid(be_enc, index))
	  {
	    be_ptr var;
	    be_ptr var_real_i;
	    int result;

	    var = BeEnc_index_to_var(be_enc, index);
	    var_real_i = BeEnc_var_to_timed(be_enc, var, i);

	    result = Bmc_IncrSat_get_solution_value(solver, var_real_i);
	    if(result == 0 || result == 1)
	      {
		be_ptr var_model_i =
		  BeEnc_var_to_timed(be_enc, var, _model_k(i));
		int rbc_index = BeEnc_var_to_index(be_enc, var_model_i);
		if(result == 0)
		  rbc_index = -rbc_index;
		lsNewEnd(be_model, (lsGeneric)rbc_index, LS_NH);
	      }

	    /* Go to next var */
	    index = BeEnc_get_next_var_index(be_enc, index, BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
	  }
      }
  }

  /* As no real be_prob was ever generated,
     it is replaced by a dummy, non-constant be */
  sat_trace = BmcSatTrace_create(l_var, be_model);
  /*sat_trace = BmcSatTrace_create(be_prob, be_model);*/
  path = NodeList_to_node_ptr(
	    BmcSatTrace_get_symbolic_model(sat_trace, be_enc, k));
	  
  trace = Trace_create_from_state_input_list(bdd_enc,
					     trace_name,
					     TRACE_TYPE_CNTEXAMPLE,
					     path);

  bmc_tr = TraceManager_register_trace(global_trace_manager, trace);
	  
  /* Print the trace using default plugin */
  fprintf(nusmv_stdout,
	  "-- as demonstrated by the following execution sequence\n");

  /* New code: find and print the loopback point */
  {
    int i = 0;
    int loopback = -1;
    for(i = 0; i <= k; i++)
      {
	be_ptr l_var_real_i = BeEnc_name_to_timed(be_enc, l_var, _real_k(i));
	int result = Bmc_IncrSat_get_solution_value(solver, l_var_real_i);
	assert(result == 0 || result == 1);
	if(result == 1)
	  {
	    assert(loopback == -1);
	    loopback = i;
	  }
      }
    if(loopback != -1)
      {
	/* NuSMV prints traces so that first state is numbered 1 */
	fprintf(nusmv_stdout,
		"   the loop starts at state %d (that is redundantly printed also as state %d)\n", loopback+1-1, k+1);
      }
    else
      {
	fprintf(nusmv_stdout, "   the execution has no loop\n");
      }
  }

  TraceManager_execute_plugin(global_trace_manager, 
		      TraceManager_get_default_plugin(global_trace_manager),
		      bmc_tr);

  BmcSatTrace_destroy(&sat_trace);
  lsDestroy(be_model, 0);

  return trace;
}
