#include "bmcInt.h"
#include "sugarInt.h"
#include "parser/psl/pslExpr.h"
#include "parser/psl/pslNode.h"
#include "parser/psl/psl_grammar.h"
#include "parser/symbols.h"
#include "utils/assoc.h"

#include "bmcPslToSugar.h"

/**Function********************************************************************

  Synopsis           [Given a PSL property, create the corresponding sugar 
  property. The sugar property can use the sugar to alternating automata 
  construction by TU Graz. The conversion is implemented as a simple recursive
  that generates the new data structure by a case analysis.]

  Description [Generate a sugar formula data structure from a psl property 
  data structure]

  SideEffects        []

  SeeAlso            [Sugar_AutomatonGenerate]

******************************************************************************/
static void parseCount(Sugar_Formula_t *f,PslNode_ptr expr);

static Sugar_Formula_t * _sereToSugar(const PslNode_ptr expr,
				      const str_hash_ptr ap_table,
				      const hash_ptr at);

static
Sugar_Formula_t *_fl2Sugar(const PslNode_ptr expr,
			   const str_hash_ptr ap_table,
			   const hash_ptr already_translated);

Sugar_Formula_t *Bmc_pslToSugar(PslNode_ptr expr, str_hash_ptr ap_table)
{
  hash_ptr already_translated = new_assoc();
  Sugar_Formula_t *result = _fl2Sugar(expr, ap_table, already_translated);
  free_assoc(already_translated);
  return result;
}

static int apcount = 0;

static
Sugar_Formula_t *_fl2Sugar(const PslNode_ptr expr,
			   const str_hash_ptr ap_table,
			   const hash_ptr already_translated)
{
  PslOp op; 
  Sugar_Formula_t *result = 0;

  nusmv_assert(expr != PSL_NULL);

  fprintf(nusmv_stderr, "Translating: "); PslNode_print(nusmv_stderr, expr); fprintf(nusmv_stderr, "\n");

  op = psl_node_get_op(expr); 
  
  switch (op) {

  case TRUEEXP:
  case TKTRUE:
    result = Sugar_FormulaCreate(Sugar_TRUE_c,NULL,NULL);
    result->isStrongOperator = 1;
    break;

  case FALSEEXP:
  case TKFALSE:
    result = Sugar_FormulaCreate(Sugar_FALSE_c,NULL,NULL);
    result->isStrongOperator = 1;
    break;
    
    /*Atomic propositions*/
  case BIT:
  case DOT:
  case ARRAY:
  case NOT:
  case AND:
  case OR:
  case IFF:
  case TKIN:
  case TKSPLAT:
  case TKSLASH:
  case TKPERCENT:
  case TKEQEQ:
  case TKEQ:
  case TKBANGEQ:
  case TKLT:
  case TKLE:
  case TKGT:
  case TKGE:
  case TKATOM:
  case TKIDENTIFIER:
  case TKARRAY:
    {
      char *ap = (char *)find_assoc(already_translated, expr);
      if(ap)
	{
	  ap = strdup(ap);
	}
      else
	{
	  ap = util_strcat3("p","_",util_inttostr(apcount++));
	  nusmv_assert(find_str_assoc(ap_table, ap)==NULL);
	  insert_str_assoc(ap_table, ap, expr);
	  insert_assoc(already_translated, expr, (node_ptr)ap);
	}
      result = Sugar_FormulaCreate(Sugar_IDENTIFIER_c, ap, NULL);
      result->isStrongOperator = 1;
      break;
    }



    /*Temporal operators should not appear*/
  case OP_NEXT:
  case OP_GLOBAL:
  case OP_FUTURE:
  case UNTIL:
  case RELEASES:
  case IMPLIES:
  case ATOM:
  case EX:
  case AX:
  case EF:
  case AF:
  case EG:
  case AG:
  case EE:
  case AA:
  case EBF:
  case EBG:
  case ABF:
  case ABG:
  case BUNTIL:
  case MMIN:
  case MMAX:
    nusmv_assert(false);
    /*End Be operators*/
    
    
  case TKCONTEXT:
    /*Should have been eliminated before conversion to sugar*/
    nusmv_assert(false);
    break;


    /* numerical expressions operators */
  case TKNUMBER:
  case TKMINUS:
  case TKPLUS:
    nusmv_assert(false);

  case TKUNION:
  case TKCARET:
  case TKDOTDOT:
    nusmv_assert(false);

  case TKCASE:
    /*TODO*/
    nusmv_assert(false);
/*     { */
/*       /\* "case" block is fully exploited here (handles also "ite") *\/ */
/*       int ret;  */
/*       PslNode_ptr i;  */
/*       ret = fprintf(f, "(case ");  */
/*       for (i=expr; psl_node_is_case(i); i=psl_node_get_case_next(i)) */
/*         ret = ret &&  */
/*           PslNode_print(f, psl_node_get_case_cond(i)) && */
/*           fprintf(f, ":") && */
/*           PslNode_print(f, psl_node_get_case_then(i)) && */
/*           fprintf(f, "; ");  */
/*       return ret && fprintf(f, " endcase)");  */
/*     } */

  case TKREPLPROP:
    /*should have been eliminated before conversion to sugar*/
    nusmv_assert(false);

  case TKFORALL: 
    /*should have been eliminated before conversion to sugar*/
    nusmv_assert(false);
             
  case TKCONS:
    /*Unknown, TODO*/
    nusmv_assert(false);

/*     if (psl_node_get_right(expr)) { */
/*       return PslNode_print(f, psl_node_get_right(expr)) && */
/*         fprintf(f, ", ") && */
/*         PslNode_print(f, psl_node_get_left(expr));  */
/*     } */
/*     else return PslNode_print(f, psl_node_get_left(expr));  */

  case TKBOOLEAN:
    /*TODO, no current interpretation*/
    nusmv_assert(false);
    
  case TKBANG:
    /* Logical not */
    result = Sugar_FormulaCreate(Sugar_FL_Not_c,
				 _fl2Sugar(psl_node_get_left(expr), ap_table,
					   already_translated),
				 NULL);
    break;
    
  case TKAMPERSAND:
    /* Logical and */
    result = Sugar_FormulaCreate(Sugar_FL_And_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    break;
    
  case TKPIPE:
    /* Logical or */
    result = Sugar_FormulaCreate(Sugar_FL_Or_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    break;

  case TKMINUSGT:
    /* Logical implication */
    result = Sugar_FormulaCreate(Sugar_FL_Imply_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    break;

  case TKLTMINUSGT:
    /* Logical equivalence */
    result = Sugar_FormulaCreate(Sugar_FL_Equiv_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    break;


    /* temporal operators */
  case TKF:
    /* finally/eventually */
    result = Sugar_FormulaCreate(Sugar_FL_Eventually_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 NULL);
    break;

  case TKG:
    /* globally */
    result = Sugar_FormulaCreate(Sugar_FL_Always_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 NULL);
    break;
    
  case TKU:
    /* strong until */
    result = Sugar_FormulaCreate(Sugar_FL_Until_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 1;
    break;
    
  case TKW:
    /* weak until */
    result = Sugar_FormulaCreate(Sugar_FL_Until_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 0;
    break;



  case TKEG:
  case TKEX:
  case TKEF:
  case TKAG:
  case TKAX:
  case TKAF:
  case TKA:
  case TKE:
    /*Branching time operators are not supported*/
    nusmv_assert(false);

    
  case TKBEFOREBANG:
    result = Sugar_FormulaCreate(Sugar_FL_Before_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 1;
    break;
    
  case TKBEFORE:
    result = Sugar_FormulaCreate(Sugar_FL_Before_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 0;
    break;
    
  case TKBEFOREBANG_:
    result = Sugar_FormulaCreate(Sugar_FL_Before__c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 1;
    break;
    
  case TKBEFORE_:
    result = Sugar_FormulaCreate(Sugar_FL_Before__c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 0;
    break;
    
  case TKUNTILBANG:
    result = Sugar_FormulaCreate(Sugar_FL_Until_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 1;
    break;
    
  case TKUNTIL:
    result = Sugar_FormulaCreate(Sugar_FL_Until_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 0;
    break;
    
  case TKUNTILBANG_:
    result = Sugar_FormulaCreate(Sugar_FL_Until__c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 1;
    break;
    
  case TKUNTIL_:
    result = Sugar_FormulaCreate(Sugar_FL_Until__c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 _fl2Sugar(psl_node_get_right(expr),
					   ap_table, already_translated));
    result->isStrongOperator = 0;
    break;
    
  case TKALWAYS:
    if(psl_node_is_sere(psl_node_get_left(expr)))
      {
	/* always sere */
	result = Sugar_FormulaCreate(Sugar_Seq_Always_c,
				     _sereToSugar(psl_node_get_left(expr),
						  ap_table,
						  already_translated),
				     NULL);
      }
    else
      {
	/* always fl_property */
	result = Sugar_FormulaCreate(Sugar_FL_Always_c,
				     _fl2Sugar(psl_node_get_left(expr),
					       ap_table, already_translated),
				     NULL);
      }
    break;


  case TKNEVER:
    if(psl_node_is_sere(psl_node_get_left(expr)))
      {
	/* never sere */
	result = Sugar_FormulaCreate(Sugar_Seq_Never_c,
				     _sereToSugar(psl_node_get_left(expr),
						  ap_table,
						  already_translated),
				     NULL);
      }
    else
      {
	/* never fl_property */
	result = Sugar_FormulaCreate(Sugar_FL_Never_c,
				     _fl2Sugar(psl_node_get_left(expr),
					       ap_table, already_translated),
				     NULL);
      }
    break;

  case TKEVENTUALLYBANG:
    if(psl_node_is_sere(psl_node_get_left(expr)))
      {
	/* eventually! sere */
	result = Sugar_FormulaCreate(Sugar_Seq_Eventually_c,
				     _sereToSugar(psl_node_get_left(expr),
						  ap_table,
						  already_translated),
				     NULL);
      }
    else
      {
	result = Sugar_FormulaCreate(Sugar_FL_Eventually_c,
				     _fl2Sugar(psl_node_get_left(expr),
					       ap_table, already_translated),
				     NULL);
      }
    break;

  case TKWITHINBANG:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {      
      /*Add strong FLAG*/
      Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Within_c,
						_fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						_fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
      result->isStrongOperator=1;
      return result;
    } 

  case TKWITHIN:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
      Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Within_c,
						_fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						_fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
      /*Check weak FLAG*/      
      result->isStrongOperator=0;
      return result;
    } 


  case TKWITHINBANG_:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
      Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Within__c,
						_fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						_fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
      /*Check strong FLAG*/      
      result->isStrongOperator=1;
      return result;
    } 


  case TKWITHIN_:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
	Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Within__c,
						  _fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						  _fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
	/*Check weak FLAG*/      
	result->isStrongOperator=0;
	return result;
    } 
    
  case TKWHILENOT:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
	Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Whilenot_c,
						  _fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						  _fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
	/*Check weak FLAG*/      
	result->isStrongOperator=0;
	return result;
    } 
  case TKWHILENOTBANG:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
	Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Whilenot_c,
						  _fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						  _fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
	/*Check strong FLAG*/      
	result->isStrongOperator=1;
	return result;
    }     
  case TKWHILENOT_:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
	Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Whilenot__c,
						  _fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						  _fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
	/*Check weak FLAG*/      
	result->isStrongOperator=0;
	return result;
    }     
 
  case TKWHILENOTBANG_:
    /*Implementation based on guess and not supported by sugar! Do not use!*/
    nusmv_assert(false);
    {
	Sugar_Formula_t *result=Sugar_FormulaCreate(Sugar_Seq_Whilenot__c,
						  _fl2Sugar(psl_node_get_left(expr), ap_table, already_translated),
						  _fl2Sugar(psl_node_get_right(expr), ap_table, already_translated));
	/*Check strong FLAG*/      
	result->isStrongOperator=1;
	return result;
    }     
    
  case TKX:
  case TKXBANG:
  case TKNEXT:
  case TKNEXTBANG: 
    result = Sugar_FormulaCreate(Sugar_FL_Next_c,
				 _fl2Sugar(psl_node_get_left(expr),
					   ap_table, already_translated),
				 NULL);
    if(op == TKXBANG || op == TKNEXTBANG) 
      result->isStrongOperator = 1;
    else
      result->isStrongOperator = 0;
    /* simple or extended version? */
    if (psl_node_get_right(expr))
      {
	/* NEXT[number](fl_property) */
	PslNode_ptr when;
	assert(psl_node_get_op(psl_node_get_right(expr)) == TKCOLON);
	assert(!psl_node_get_right(psl_node_get_right(expr)));
	when = psl_node_get_left(psl_node_get_right(expr));
	assert(when);
	assert(psl_node_get_op(when) == TKNUMBER);
	assert(psl_node_get_left(when));
	result->numRangeStr=(int)psl_node_get_left(when);
      }
    break;

  case TKNEXT_A:
  case TKNEXT_ABANG:
  case TKNEXT_E:
  case TKNEXT_EBANG:
    /*Currently not supported*/
    nusmv_assert(false);

  case TKNEXT_EVENTBANG:
  case TKNEXT_EVENT:
  case TKNEXT_EVENT_ABANG:
  case TKNEXT_EVENT_A:
  case TKNEXT_EVENT_EBANG:
  case TKNEXT_EVENT_E:
    /*Currently we not support these*/
    nusmv_assert(false);


    /* sere operators */
  case TKPIPEMINUSGT:
    {
      PslNode_ptr left = psl_node_get_left(psl_node_get_left(expr));
      PslNode_ptr right = psl_node_get_right(psl_node_get_left(expr));
      assert(psl_node_is_sere(left));
      
      if(psl_node_is_sere(right))
	/* sere |-> sere */
	result = Sugar_FormulaCreate(Sugar_Seq_Imply_c,
				     _sereToSugar(left, ap_table,
						  already_translated),
				     _sereToSugar(right, ap_table,
						  already_translated));
      else
	result = Sugar_FormulaCreate(Sugar_Seq_FLProp_c,
				     _sereToSugar(left, ap_table,
						  already_translated),
				     _fl2Sugar(right,
					       ap_table, already_translated)
				     );
      
      /* Check if with ! */
      if(psl_node_get_right(expr))
	result->isStrongOperator = 1;
      else
	result->isStrongOperator = 0;
    }
    break;

  case TKPIPEEQGT:
    {
      PslNode_ptr left = psl_node_get_left(psl_node_get_left(expr));
      PslNode_ptr right = psl_node_get_right(psl_node_get_left(expr));
      assert(psl_node_is_sere(left));
      assert(psl_node_is_sere(right));

      result = Sugar_FormulaCreate(Sugar_Seq_Imply2_c,
				   _sereToSugar(left,
						ap_table, already_translated),
				   _sereToSugar(right,
						ap_table, already_translated));
      /* Check if with ! */
      if(psl_node_get_right(expr))
	result->isStrongOperator = 1;
      else
	result->isStrongOperator = 0;
    }
    break;

  case TKTILDE:
    nusmv_assert(false); 
    break;

  case TKAMPERSANDAMPERSAND:
  case TKPIPEPIPE:
    /*Should not appear here, their meaning is ambigous*/
    nusmv_assert(false);
    break;

  case TKABORT:
    /*Unsupported*/
    nusmv_assert(false);



  default:
    fprintf(nusmv_stderr, 
	    "_fl2Sugar: operator type not supported \"%d\"\n", op); 
    nusmv_assert(false); 
    break; 
  }
  
  assert(result);
  return result;
}

static void
parseCount(Sugar_Formula_t *f, PslNode_ptr expr)
{
  switch(psl_node_get_op(expr)) {
  case TKNUMBER:
    f->numRangeStr=(int) psl_node_get_left(expr);
    break;
  case TKDOTDOT:
    parseCount(f, psl_node_get_left(expr));
    parseCount(f, psl_node_get_left(expr));
    break;
  default:
    nusmv_assert(false);

  }
  return;
}


static Sugar_Formula_t * _sereToSugar(const PslNode_ptr expr,
				      const str_hash_ptr ap_table,
				      const hash_ptr at)
{
  PslOp op = psl_node_get_op(expr);
 
#if 0 
  fprintf(nusmv_stderr, "Sere translating: ");
  PslNode_print(nusmv_stderr, expr);
  fprintf(nusmv_stderr, "\n");
#endif

  switch(op)
    {
    case TKTRUE:
    case TRUEEXP:
      return Sugar_FormulaCreate(Sugar_TRUE_c,NULL,NULL);
      
    case TKFALSE:
    case FALSEEXP:
      return Sugar_FormulaCreate(Sugar_FALSE_c,NULL,NULL);

    case BIT:
    case DOT:
    case ARRAY:
    case NOT:
    case AND:
    case OR:
    case IFF:
    case TKIN:
    case TKSPLAT:
    case TKSLASH:
    case TKPERCENT:
    case TKEQEQ:
    case TKEQ:
    case TKBANGEQ:
    case TKLT:
    case TKLE:
    case TKGT:
    case TKGE:
    case TKATOM:
    case TKIDENTIFIER:
    case TKARRAY:
      assert(PslNode_is_propositional(expr));
      {
	Sugar_Formula_t * result = 0;
	char *ap = (char *)find_assoc(at, expr);
	if(ap)
	  {
	    ap = strdup(ap);
	  }
	else
	  {
	    ap=util_strcat3("p","_",util_inttostr(apcount++));
	    nusmv_assert(find_str_assoc(ap_table, ap)==NULL);
	    insert_str_assoc(ap_table, ap, expr);
	    insert_assoc(at, expr, (node_ptr)ap);
	  }
	result = Sugar_FormulaCreate(Sugar_IDENTIFIER_c, ap, NULL);
	result->isStrongOperator = 0;
	return result;
      }


    case TKSERECOMPOUND:
    case TKSERE:
      assert(!psl_node_get_right(expr)); 
      return _sereToSugar(psl_node_get_left(expr),ap_table,at);
      
    case TKSERECONCAT:
      return Sugar_FormulaCreate(Sugar_RegExpr_Concat_c,
				 _sereToSugar(psl_node_get_left(expr),
					      ap_table, at),
				 _sereToSugar(psl_node_get_right(expr),
					      ap_table, at));

    case TKSEREFUSION:
      return Sugar_FormulaCreate(Sugar_RegExpr_Compose_c,
				 _sereToSugar(psl_node_get_left(expr),
					      ap_table,at),
				 _sereToSugar(psl_node_get_right(expr),
					      ap_table,at));
      
    case TKSEREREPEATED:
      {
	nusmv_assert(psl_node_get_left(expr)); 
	
	/* see psl_expr_make_repeated_sere() above for parse-tree structure */
	Sugar_Formula_t *temp;
	{
	  PslOp sop = psl_node_get_op(psl_node_get_left(expr)); 
	  switch (sop) {
	  case TKLBSPLAT: 
	    {
	      temp=Sugar_FormulaCreate(Sugar_RegExpr_Star_c,
				       NULL,
				       NULL); 
	      if (psl_node_get_left(psl_node_get_left(expr))) { 				     
		temp->left=_sereToSugar(psl_node_get_left(psl_node_get_left(expr)),ap_table,at);
	      }
	      if(psl_node_get_right(expr)) {
		parseCount(temp,psl_node_get_right(expr));
	      }
	      return temp;
	    }
	    
	  case TKLBPLUSRB: 
	    nusmv_assert(psl_node_get_left(psl_node_get_left(expr))); 
	    {    
	      return Sugar_FormulaCreate(Sugar_RegExpr_Plus_c,
					 _sereToSugar(psl_node_get_left(psl_node_get_left(expr)),ap_table,at),
				       NULL);
	    }
	  case TKLBEQ: 
	    nusmv_assert(psl_node_get_right(expr));
	    nusmv_assert(psl_node_get_left(psl_node_get_left(expr)));
	    {
	      temp=Sugar_FormulaCreate(Sugar_RegExpr_Equal_c,
				       _sereToSugar(psl_node_get_left(psl_node_get_left(expr)),ap_table,at),
				       NULL);
	      parseCount(temp,psl_node_get_right(expr));
	      return temp;
	    }
	    
	  case TKLBMINUSGT:
	    nusmv_assert(psl_node_get_right(expr));
	    nusmv_assert(psl_node_get_left(psl_node_get_left(expr)));
	    {
	      temp=Sugar_FormulaCreate(Sugar_RegExpr_Equal_c,
				       _sereToSugar(psl_node_get_left(psl_node_get_left(expr)),ap_table,at),
				       NULL);
	      parseCount(temp,psl_node_get_right(expr));
	      return temp;
	    }
	    
	    

	  default: 
	    fprintf(nusmv_stderr, "operator type not supported " \
		    "in repeated sere\"%d\"\n", op); 
	    nusmv_exit(1); 
	  }
	}
      }
      
    case TKAMPERSAND:
      return Sugar_FormulaCreate(Sugar_RegExpr_And_c,
				 _sereToSugar(psl_node_get_left(expr),
					      ap_table,at),
				 _sereToSugar(psl_node_get_right(expr),
					      ap_table,at));
      
    case TKAMPERSANDAMPERSAND:
      return Sugar_FormulaCreate(Sugar_RegExpr_AndEq_c,
				 _sereToSugar(psl_node_get_left(expr),
					      ap_table, at),
				 _sereToSugar(psl_node_get_right(expr),
					      ap_table, at));
      
    case TKPIPE:
      return Sugar_FormulaCreate(Sugar_RegExpr_Or_c,
				 _sereToSugar(psl_node_get_left(expr),
					      ap_table, at),
				 _sereToSugar(psl_node_get_right(expr),
					      ap_table, at));
    default:
      fprintf(nusmv_stderr, 
	      "_sereToSugar: operator type not supported \"%d\"\n", op);
      assert(false); 
      break; 
    }
  
  assert(false);
  return 0;
}
